#pragma once

#include "keymap.h"

class	CSkkRuleTreeNodeOutput {
public:
	enum {
		NTYPE_INVALIDVALUE	= -1,
		NTYPE_MACRO,
		NTYPE_STRING,
		NTYPE_STRINGPAIR,
	} ;
public:
	CSkkRuleTreeNodeOutput () {
		return ;
	}
	virtual	~CSkkRuleTreeNodeOutput () {
		return ;
	}
	virtual int		iGetType () const {
		return	NTYPE_INVALIDVALUE ;
	}
} ;

class	CSkkRuleTreeNodeOutputStringPair : public CSkkRuleTreeNodeOutput {
private:
	LPWSTR		m_strHira ;
	int			m_nHira ;
	LPWSTR		m_strKata ;
	int			m_nKata ;
public:
	CSkkRuleTreeNodeOutputStringPair () : CSkkRuleTreeNodeOutput (), m_strHira (NULL), m_nHira (0), m_strKata (NULL), m_nKata (0) {
		return ;
	}
	virtual	~CSkkRuleTreeNodeOutputStringPair () {
		if (m_strHira != NULL) {
			delete[]	m_strHira ;
			m_strHira	= NULL ;
		}
		if (m_strKata != NULL) {
			delete[]	m_strKata ;
			m_strKata	= NULL ;
		}
		return ;
	}
	virtual int		iGetType () const {
		return	NTYPE_STRINGPAIR ;
	}
	virtual LPCWSTR		pGetHiragana (int* pnLength) const {
		if (pnLength != NULL)
			*pnLength	= m_nHira ;
		return	m_strHira ;
	}
	virtual LPCWSTR		pGetKatakana (int* pnLength) const {
		if (pnLength != NULL)
			*pnLength	= m_nKata ;
		return	m_strKata ;
	}
	static CSkkRuleTreeNodeOutput*	pCreateInstance (LPCWSTR wstrHira, int nHira, LPCWSTR wstrKata, int nKata) ;
} ;

class	CSkkRuleTreeNodeOutputString : public CSkkRuleTreeNodeOutput {
private:
	LPWSTR		m_strValue ;
	int			m_nValue ;
public:
	CSkkRuleTreeNodeOutputString () : CSkkRuleTreeNodeOutput (), m_strValue (NULL), m_nValue (0) {
		return ;
	}
	virtual	~CSkkRuleTreeNodeOutputString () {
		if (m_strValue != NULL) {
			delete[]	m_strValue ;
			m_strValue	= NULL ;
		}
		return ;
	}
	virtual int		iGetType () const {
		return	NTYPE_STRING ;
	}
	virtual	LPCWSTR		pGetString (int* pnLength) const {
		if (pnLength != NULL)
			*pnLength	= m_nValue ;
		return	m_strValue ;
	}
	static CSkkRuleTreeNodeOutput*	pCreateInstance (LPCWSTR pwString, int nLength) ;
} ;

class	CSkkRuleTreeNodeOutputMacro : public CSkkRuleTreeNodeOutput {
private:
	int			m_nMacro ;
public:
	CSkkRuleTreeNodeOutputMacro () : CSkkRuleTreeNodeOutput (), m_nMacro (NFUNC_NOFUNCTION) {
		return ;
	}
	virtual	~CSkkRuleTreeNodeOutputMacro () {
		return ;
	}
	virtual int		iGetType () const {
		return	NTYPE_MACRO ;
	}
	virtual	int			iGetFunctionCode () const {
		return	m_nMacro ;
	}
	static CSkkRuleTreeNodeOutput*	pCreateInstance (int nFuncNo) ;
} ;

/*	Rule 1{̖؂ȂāA̖؁cтɂǂ̂ł͂ȂH
 *	oh ̖ o Ԃ1{̖؂ł͕\łȂŁA
 *	ʂ̖؂ɉt΁Achar-previous ȂĂ(A
 *	uv邩ǂ̂ǁAGfB^Ȃ͂
 *	IꏊɁuv邩ǂ͕Ȃ)ǂȂ̂ł͂ȂH
 *
 *	؂𕪊 traverse Ƃł͂܂B(؂ǂ蒼Ƃ͑)
 */
class	CSkkRuleTreeNode {
private:
	enum {
		SYSCHAR_INVALID		= -1,
		SYSCHAR_DEFAULT		= 65536,
		SYSCHAR_VOWEL,
		SYSCHAR_CONSONANT,
	} ;
private:
	int						m_nCH ;
	int						m_iRule ;
	LPWSTR					m_pPrefix ;			/* ԁB*/
	int						m_nPrefix ;
	LPWSTR					m_pNextState ;		/* ԁB*/
	int						m_nNextState ;
	int						m_nNextRule ;		/* ̖؁B*/

	CSkkRuleTreeNodeOutput*	m_pOutput ;

	CSkkRuleTreeNode*		m_pChild ;			/* qB*/
	CSkkRuleTreeNode*		m_pBrother ;		/* ZB*/

public:
	CSkkRuleTreeNode () : m_nCH (0), m_pPrefix (NULL), m_nPrefix (0), m_pNextState (NULL), m_nNextState (0), m_pOutput (NULL), m_pChild (NULL), m_pBrother (NULL) {
		return ;
	}
	virtual	~CSkkRuleTreeNode () {
		if (m_pPrefix != NULL) {
			delete[]	m_pPrefix ;
			m_pPrefix	= NULL ;
		}
		if (m_pNextState != NULL) {
			delete[]	m_pNextState ;
			m_pNextState	= NULL ;
		}
		if (m_pOutput != NULL) {
			delete	m_pOutput ;
			m_pOutput	= NULL ;
		}
		return ;
	} 

	virtual	int							iGetChar () const {
		return	m_nCH ;
	}

	virtual	LPCWSTR						pGetNextState (int* pnLength, int* pnNextRule) const {
		if (pnLength != NULL)
			*pnLength	= m_nNextState ;
		if (pnNextRule != NULL)
			*pnNextRule	= m_nNextRule ;
		return	m_pNextState ;
	}

	virtual	const CSkkRuleTreeNode*		pGetChild () const {
		return	m_pChild ;
	}

	virtual	CSkkRuleTreeNode*			pGetChild () {
		return	m_pChild ;
	}

	virtual	void	vSetChild (CSkkRuleTreeNode* pChild) {
		m_pChild	= pChild ;
		return ;
	}

	virtual	const CSkkRuleTreeNode*		pGetBrother () const {
		return	m_pBrother ;
	}

	virtual	CSkkRuleTreeNode*			pGetBrother () {
		return	m_pBrother ;
	}

	virtual	void	vSetBrother (CSkkRuleTreeNode* pBrother) {
		m_pBrother	= pBrother ;
		return ;
	}

	virtual	CSkkRuleTreeNodeOutput*		pGetOutput () {
		return	m_pOutput ;
	}

	virtual	const CSkkRuleTreeNodeOutput*		pGetOutput () const {
		return	m_pOutput ;
	}

	virtual	void	vSetOutput (CSkkRuleTreeNodeOutput* pOutput) {
		if (m_pOutput == pOutput)
			return ;
		if (m_pOutput != NULL) 
			delete	m_pOutput ;
		m_pOutput	= pOutput ;
		return ;
	}

	virtual	void	vDebugOut () const ;

	static	CSkkRuleTreeNode*		pCreateInstance (int iRule, int nChar, LPCWSTR pwPrefix, int nPrefixLen, LPCWSTR pwNext, int nNextLen, int iNextRule, CSkkRuleTreeNodeOutput* pOutput) ;

	static	LPCWSTR					pGetPrefix (CSkkRuleTreeNode* pNode, int* pnPrefix) ;
	static	LPCWSTR					pGetPrefix (const CSkkRuleTreeNode* pNode, int* pnPrefix) ;
	static	CSkkRuleTreeNode*		pSelectBranch (CSkkRuleTreeNode* pNode, int iCH) ;
	static	const CSkkRuleTreeNode*	pSelectBranch (const CSkkRuleTreeNode* pNode, int iCH) ;
	static	CSkkRuleTreeNode*		pGetBranchList (CSkkRuleTreeNode* pNode) ;
	static	const CSkkRuleTreeNode*	pGetBranchList (const CSkkRuleTreeNode* pNode) ;

	static	int						iSearchTree (CSkkRuleTreeNode* pTree, LPCWSTR pwString, int nString, CSkkRuleTreeNode** ppResult) ;
	static	BOOL					bAddRule (CSkkRuleTreeNode** ppTree, int iTree, LPCWSTR wstrPrefix, int nPrefixLen, LPCWSTR wstrNext, int nNextLen, int nNextTree, LPCWSTR wstrHira, int nHiraLen, LPCWSTR wstrKata, int nKataLen) ;
	static	BOOL					bAddRule (CSkkRuleTreeNode** ppTree, int iTree, LPCWSTR wstrPrefix, int nPrefixLen, LPCWSTR wstrNext, int nNextLen, int nNextTree, LPCWSTR wstrString, int nString) ;
	static	BOOL					bAddRule (CSkkRuleTreeNode** ppTree, int iTree, LPCWSTR wstrPrefix, int nPrefixLen, LPCWSTR wstrNext, int nNextLen, int nNextTree, int nMacro) ;
	static	BOOL					bDeleteRule (CSkkRuleTreeNode**	ppTree, LPCWSTR pString, int nString) ;
	static	void					vDestroyTree (CSkkRuleTreeNode* pNode) ;

private:
//	static	int						iSearchTre (CSkkRuleTreeNode* pTree, LPCWSTR pString, int nString, CSkkRuleTreeNode** ppResult) ;

	static	BOOL					bMatchChar (int nSystemChar, int nInputChar, int* pnScore) ;
	static	int						iGetRuleCharacter (LPCWSTR* ppwPtr, LPCWSTR pwPtrEnd) ;
} ;

/*


BOOL	TSkkRuleTree_bInitializeDefault (struct CSkkRuleTreeNode**) ;
BOOL	TSkkRuleTreeNode_bAddRuleSP		(struct CSkkRuleTreeNode**, LPCWSTR, int, LPCWSTR, int, LPCWSTR, int, LPCWSTR, int) ;
BOOL	TSkkRuleTreeNode_bAddRuleS		(struct CSkkRuleTreeNode**, LPCWSTR, int, LPCWSTR, int, LPCWSTR, int) ;
BOOL	TSkkRuleTreeNode_bAddRuleM		(struct CSkkRuleTreeNode**, LPCWSTR, int, LPCWSTR, int, int) ;

void	TSkkRuleTreeNode_vDestroyTree	(struct CSkkRuleTreeNode*) ;
LPCWSTR	TSkkRuleTreeNode_pGetPrefix		(const struct CSkkRuleTreeNode*, int*) ;
struct CSkkRuleTreeNode*				TSkkRuleTreeNode_pSelectBranch	(struct CSkkRuleTreeNode*, int) ;
LPCWSTR	TSkkRuleTreeNode_pGetNextState	(const struct CSkkRuleTreeNode*, int*) ;
struct CSkkRuleTreeNode*				TSkkRuleTreeNode_pGetBranchList (struct CSkkRuleTreeNode*) ;

const struct CSkkRuleTreeNodeOutput*	TSkkRuleTreeNode_pGetOutput		(const struct CSkkRuleTreeNode*) ;

struct CSkkRuleTreeNodeOutput*			TSkkRuleTreeNodeOutput_pCreateStringPairInstance	(LPCWSTR, int, LPCWSTR, int) ;
struct CSkkRuleTreeNodeOutput*			TSkkRuleTreeNodeOutput_pCreateMacroInstance			(int) ;
struct CSkkRuleTreeNodeOutput*			TSkkRuleTreeNodeOutput_pCreateStringInstance		(LPCWSTR, int) ;
void	TSkkRuleTreeNodeOutput_vDestroy		(struct CSkkRuleTreeNodeOutput*) ;

int		TSkkRuleTreeNodeOutput_iGetType		(const struct CSkkRuleTreeNodeOutput*) ;
BOOL	TSkkRuleTreeNodeOutput_bGetMacro	(const struct CSkkRuleTreeNodeOutput*, int* pnFuncNo) ;
BOOL	TSkkRuleTreeNodeOutput_bGetString	(const struct CSkkRuleTreeNodeOutput*, LPCWSTR*, int*) ;
BOOL	TSkkRuleTreeNodeOutput_bGetLString	(const struct CSkkRuleTreeNodeOutput*, LPCWSTR*, int*) ;
BOOL	TSkkRuleTreeNodeOutput_bGetRString	(const struct CSkkRuleTreeNodeOutput*, LPCWSTR*, int*) ;



*/