#if !defined (UNITTEST)
#include "globals.h"
#else
#include "StdAfx.h"
#endif
#include "TSearchSession.h"
#include "TPacket.h"
#include "TProtocol.h"
#include "jstring.h"
#include "debug.h"

/*========================================================================
 *	structures
 */
struct CTS4NumIndex {
	int		m_nIndex ;	/* #4 pos */
	int		m_nPos ;	/* candidate pos */
	int		m_nLen ;	/* candidate length */
} ;

class CTS4Candidate {
private:
	int						m_nKeyIndex ;		/* S4 keyword */
	int						m_nResultIndex ;	/* S4 result */
	int						m_nResultLength ;	/* S4 result length */
	CTS4Candidate*			m_pNext ;

public:
	CTS4Candidate () : m_nKeyIndex (-1), m_nResultIndex (-1), m_nResultLength (0), m_pNext (NULL) {
		return ;
	}
	CTS4Candidate (int nKeyIndex, int nResultIndex, int nResultLength) : m_nKeyIndex (nKeyIndex), m_nResultIndex (nResultIndex), m_nResultLength (nResultLength), m_pNext (NULL) {
		return ;
	}

	virtual void	vSetNext (CTS4Candidate* pNode) {
		m_pNext	= pNode ;
		return ;
	}
	virtual CTS4Candidate*	pGetNext () {
		return	m_pNext ;
	}
	virtual CTS4Candidate*	pGetNext () const {
		return	m_pNext ;
	}
	virtual int				iGetKeyIndex () const {
		return	m_nKeyIndex ;
	}
	virtual int				iGetResultIndex () const {
		return	m_nResultIndex ;
	}
	virtual int				iGetResultLength () const {
		return	m_nResultLength ;
	}
} ;

/*	template ācƍsƂ낾Aꂾ hash value  class ɂ
 *	operator== ...ƁAhash index 𐶐 method ǉĂƁc
 *	ρBC Ȃ炻p function  function pointer œn̂c...
 */
class CTSearchSessionCandidate {
private:
	int							m_nIndex ;
	CTSearchSessionCandidate*	m_pPrev ;	/* tێ prev, next */
	CTSearchSessionCandidate*	m_pNext ;
	CTSearchSessionCandidate*	m_pLeft ;	/* hash table  entry Ƃ left, right */
	CTSearchSessionCandidate*	m_pRight ;
	CTSearchSessionCandidate*	m_pRef ;
	CTS4Candidate*				m_pRefS4 ;

public:
	CTSearchSessionCandidate () :
		m_nIndex (-1),
		m_pPrev (NULL),		m_pNext  (NULL),
		m_pLeft (NULL),		m_pRight (NULL),
		m_pRef  (NULL),		m_pRefS4 (NULL) {
		return ;
	}
	CTSearchSessionCandidate (int nIndex) :
		m_nIndex (nIndex),
		m_pPrev (NULL),		m_pNext  (NULL),
		m_pLeft (NULL),		m_pRight (NULL),
		m_pRef  (NULL),		m_pRefS4 (NULL) {
		return ;
	}
	virtual		~CTSearchSessionCandidate () {
		return ;
	}
	virtual CTSearchSessionCandidate*	pGetPrevious () {
		return	m_pPrev ;
	}
	virtual CTSearchSessionCandidate*	pGetNext () {
		return	m_pNext ;
	}
	virtual CTSearchSessionCandidate*	pGetLeft () {
		return	m_pLeft ;
	}
	virtual CTSearchSessionCandidate*	pGetRight () {
		return	m_pRight ;
	}
	virtual const CTSearchSessionCandidate*	pGetPrevious () const {
		return	m_pPrev ;
	}
	virtual const CTSearchSessionCandidate*	pGetNext () const {
		return	m_pNext ;
	}
	virtual const CTSearchSessionCandidate*	pGetLeft () const {
		return	m_pLeft ;
	}
	virtual const CTSearchSessionCandidate*	pGetRight () const {
		return	m_pRight ;
	}
	virtual int	iGetValue () const {
		return	m_nIndex ;
	}
	virtual void	vSetNext (CTSearchSessionCandidate* pNode) {
		m_pNext	= pNode ;
	}
	virtual void	vSetPrevious (CTSearchSessionCandidate* pNode) {
		m_pPrev	= pNode ;
	}
	virtual void	vSetLeft (CTSearchSessionCandidate* pNode) {
		m_pLeft	= pNode ;
	}
	virtual void	vSetRight (CTSearchSessionCandidate* pNode) {
		m_pRight	= pNode ;
	}
	virtual	void	vSetLink (CTSearchSessionCandidate* pNode) {
		m_pRef	= pNode ;
	}
	virtual	CTSearchSessionCandidate*	pGetLink () {
		return	m_pRef ;
	}
	virtual void	vAddLink2 (CTS4Candidate* pNodeToBeLinked) {
		if (pNodeToBeLinked == NULL)
			return ;

		if (m_pRefS4 == NULL) {
			m_pRefS4	= pNodeToBeLinked ;
		} else {
			CTS4Candidate*	pPrevNode ;
			CTS4Candidate*	pNodeS4 ;

			pPrevNode	= NULL ;
			pNodeS4		= pNodeToBeLinked ;
			while (pNodeS4 != NULL) {
				pPrevNode	= pNodeS4 ;
				pNodeS4		= pNodeS4->pGetNext () ;
			}
			pPrevNode->vSetNext (m_pRefS4) ;
			m_pRefS4	= pNodeToBeLinked ;
		}
	}
	virtual CTS4Candidate*	pGetLink2 () {
		return	m_pRefS4 ;
	}
	virtual const CTS4Candidate*	pGetLink2 () const {
		return	m_pRefS4 ;
	}
} ;

/*========================================================================
 *	prototypes
 */
static	BOOL	_bNumExp			(LPCWSTR strNum, int nNumLen, int nType, CVarbuffer < WCHAR, 16 >*) ;
static	BOOL	_bNumJisx0208Latin	(LPCWSTR strNum, int nNumLen, CVarbuffer < WCHAR, 16 >*) ;
static	BOOL	_bNumType2Kanji		(LPCWSTR strNum, int nNumLen, CVarbuffer < WCHAR, 16 >*) ;
static	BOOL	_bNumType3Kanji		(LPCWSTR strNum, int nNumLen, CVarbuffer < WCHAR, 16 >*) ;
static	BOOL	_bNumType5Kanji		(LPCWSTR strNum, int nNumLen, CVarbuffer < WCHAR, 16 >*) ;
static	BOOL	_bNumType4Kanji		(LPCWSTR strNum, int nNumLen, CVarbuffer < WCHAR, 16 >*) ;
static	BOOL	_bNumType9Shogi		(LPCWSTR strNum, int nNumLen, CVarbuffer < WCHAR, 16 >*) ;
static	BOOL	_bIntNump			(LPCWSTR strNum, int nNumLen) ;
static	BOOL	_bIntOrFloatNump	(LPCWSTR strNum, int nNumLen) ;

/*========================================================================
 *	static global variables
 */
static	LPCWSTR		_wstrNumericUnit [] = {
	L"",	L"",	L"",	L"",	L"",	L"R",	L"",	L"a",
	L"",	L"",	L"",	L"",	L"P͍",	L"m_",
	L"ߗR",	L"svc",	L"ʑ吔",
} ;

static	LPCWSTR		_wstrKanjiDigits = L"ZOlܘZ" ;

static	LPCWSTR		_wstrNumericUnitType5 [] = {
	L"",	L"",	L"",	L"",	L"",	L"R",	L"",	L"a",
	L"",	L"",	L"",	L"",	L"P͍",	L"m_",
	L"ߗR",	L"svc",	L"ʑ吔",
} ;

static	LPCWSTR		_wstrKanjiDigitsType5 = L"QlޘZ" ;

/*========================================================================
 *	class CTSearchSession
 */
CTSearchSession::CTSearchSession ()
{
	int	i ;

	m_pCurCandidate		= NULL ;
	m_pTopCandidate		= NULL ;
	m_pLastCandidate	= NULL ;

	/*	hash table ̒gOɏĂB*/
	for (i = 0 ; i < CANDIDATEHASHSIZE ; i ++)
		m_rpTblCandidate [i]	= NULL ;

	m_rszKeyword [0]	= L'\0' ;
	m_nwstrKeyword		= 0 ;
	m_nCandidate		= 0 ;
	m_nSearchCount		= 0 ;
	m_nParsePosition	= 0 ;

	m_rszNumericList [0] = m_rszNumericList [1] = L'\0' ;
	return ;
}

CTSearchSession::CTSearchSession (LPCWSTR wstrKeyword, int nKeyword)
{
	int	i ;

	/*	[ށA̕cBvgR͍͌ߑłɂĂB*/
	m_nwstrKeyword		= 0 ;
	m_pCurCandidate		= NULL ;
	m_pTopCandidate		= NULL ;
	m_pLastCandidate	= NULL ;

	/*	hash table ̒gOɏĂB*/
	for (i = 0 ; i < CANDIDATEHASHSIZE ; i ++)
		m_rpTblCandidate [i]	= NULL ;

	nKeyword			= (nKeyword > MAXKEYWORDLEN)? MAXKEYWORDLEN : nKeyword ;
	memcpy (m_rszKeyword, wstrKeyword, nKeyword * sizeof (WCHAR)) ;
	m_nwstrKeyword		= nKeyword ;
	m_nCandidate		= 0 ;
	m_nSearchCount		= 0 ;
	m_nParsePosition	= 0 ;

	m_rszNumericList [0] = m_rszNumericList [1] = L'\0' ;
	return ;
}

CTSearchSession::~CTSearchSession ()
{
	m_pCurCandidate		= NULL ;
	m_pTopCandidate		= NULL ;
	m_pLastCandidate	= NULL ;
	_bCleanupHashTable () ;
	return ;
}

LPCWSTR
CTSearchSession::pGetKeyword (int* pnWord) 
{
	if (pnWord != NULL)
		*pnWord	= m_nwstrKeyword ;
	return	m_rszKeyword ;
}

LPCWSTR
CTSearchSession::pGetKeyword (int* pnWord) const
{
	if (pnWord != NULL)
		*pnWord	= m_nwstrKeyword ;
	return	m_rszKeyword ;
}

LPCWSTR
CTSearchSession::pGetCandidate ()
{
	int		nPosition ;

	if (m_pCurCandidate == NULL) {
		/*	肦ȂԁB*/
		if (m_pTopCandidate != NULL)
			return	NULL ;

		/*	Session pẮAB*/
		while (m_nSearchCount >= 0) {
			if (! _bSearch ())
				return	NULL ;
			if (m_pTopCandidate != NULL)
				break ;
		}
		/*	͑SĎsĂB*/
		if (m_pTopCandidate == NULL)
			return	NULL ;
		m_pCurCandidate	= m_pTopCandidate ;
	}
	
	nPosition		= m_pCurCandidate->iGetValue () ;
	ASSERT (0 <= nPosition && nPosition < m_vbufCandidate.iGetUsage ()) ;
	return	m_vbufCandidate.pGetBuffer () + nPosition ;
}

LPCWSTR
CTSearchSession::pGetReferCandidate ()
{
	CTSearchSessionCandidate*	pRef	= NULL ; 
	int							nPosition ;

	/*	̏ꍇɂ́AuvȂĎꏟȂƂ͂ȂB
	 */
	if (m_pCurCandidate == NULL) 
		return	NULL ;

	pRef	= m_pCurCandidate->pGetLink () ;
	if (pRef == NULL)
		return	NULL ;

	nPosition		= pRef->iGetValue () ;
	ASSERT (0 <= nPosition && nPosition < m_vbufCandidate.iGetUsage ()) ;
	return	m_vbufCandidate.pGetBuffer () + nPosition ;
}

/*	#4 ̑Ή InsertCandidate ƂƂs邱ƂlB
 *	ɑΉȂ΂ȂȂB
 *	Ή鎞΁AȂƂ󋵂݂ (#4 o^AVȂĂ
 *	ƂvƂBA#4 ̌⒆ɏoƂԂ́c[XP[X
 *	lƖȂ񂾂ǁB)
 */
BOOL
CTSearchSession::bInsertCandidate (
	LPCWSTR			wstrCandidate,
	int				nstrCandidate)
{
	CTSearchSessionCandidate*	pEntry ;
	BOOL	fRetval ;

	if (nstrCandidate <= 0 || wstrCandidate == NULL)
		return	FALSE ;

	if (m_pCurCandidate == NULL &&
		m_pTopCandidate != NULL) 
		return	FALSE ;
	
	if (! m_vbufCandidate.bAdd (wstrCandidate, nstrCandidate) ||
		! m_vbufCandidate.bAdd (L"", 1))
		return	FALSE ;

	pEntry	= _pCreateCandidate (m_nParsePosition, nstrCandidate, &fRetval) ;
	/*	̃p[Yʒu̓obt@̍ŌɂȂƎvB*/
	m_nParsePosition	= m_vbufCandidate.iGetUsage () ;

	/*	[ށB͂܂ȁB*/
	if (pEntry == NULL)
		return	fRetval ;

	if (m_pCurCandidate != NULL) {
		CTSearchSessionCandidate*	pNextEntry ;

		pNextEntry	= m_pCurCandidate->pGetNext	() ;
		m_pCurCandidate->vSetNext (pEntry) ;
		if (pNextEntry != NULL) {
			pNextEntry->vSetPrevious (pEntry) ;
			pEntry->vSetNext (pNextEntry) ;
		} else {
			m_pLastCandidate	= pEntry ;
		}
		pEntry->vSetPrevious (m_pCurCandidate) ;
	} else {
		m_pTopCandidate		= pEntry ;
		m_pCurCandidate		= pEntry ;
		m_pLastCandidate	= pEntry ;
	}
	m_nCandidate	++ ;

	/*	Ȃ萔lo^\̂ŁAӂB
	 */
	if (_bNumericp ()) {
		BOOL	bConverted	= FALSE ;

		if (_bNumConvert (pEntry, m_rszNumericList, NULL, &bConverted) && bConverted) {
			_bRemoveCandidate (pEntry) ;

			/*()
			 *	Create ̃^C~O pEntry  register Ă̂ŁADestroy Ă΂ȂĂ
			 *	Session j̃^C~O Destroy sB
			 */
		}
	}
	return	TRUE ;
}

BOOL
CTSearchSession::bInsertCandidateWithS4Mapping (
	LPCWSTR				wstrCandidate,
	int					nstrCandidate,
	const CTS4Mapping*	plstS4Mapping)
{
	CTSearchSessionCandidate*	pEntry ;
	BOOL	fRetval ;

	if (plstS4Mapping == NULL)
		return	bInsertCandidate (wstrCandidate, nstrCandidate) ;

	if (nstrCandidate <= 0 || wstrCandidate == NULL)
		return	FALSE ;

	if (m_pCurCandidate == NULL && m_pTopCandidate != NULL) 
		return	FALSE ;
	
	if (! m_vbufCandidate.bAdd (wstrCandidate, nstrCandidate) ||
		! m_vbufCandidate.bAdd (L"", 1))
		return	FALSE ;

	pEntry	= _pCreateCandidate (m_nParsePosition, nstrCandidate, &fRetval) ;

	/*	̃p[Yʒu̓obt@̍ŌɂȂƎvB*/
	m_nParsePosition	= m_vbufCandidate.iGetUsage () ;

	/*	[ށB͂܂ȁB*/
	if (pEntry == NULL)
		return	fRetval ;

	if (m_pCurCandidate != NULL) {
		CTSearchSessionCandidate*	pNextEntry ;

		pNextEntry	= m_pCurCandidate->pGetNext () ;
		m_pCurCandidate->vSetNext (pEntry) ;
		if (pNextEntry != NULL) {
			pNextEntry->vSetPrevious (pEntry) ;
			pEntry->vSetNext (pNextEntry) ;
		} else {
			m_pLastCandidate	= pEntry ;
		}
		pEntry->vSetPrevious (m_pCurCandidate) ;
	} else {
		m_pTopCandidate		= pEntry ;
		m_pCurCandidate		= pEntry ;
		m_pLastCandidate	= pEntry ;
	}
	m_nCandidate	++ ;

	/*	Ȃ萔lo^\̂ŁAӂB
	 */
	if (_bNumericp ()) {
		BOOL	bConverted	= FALSE ;

		if (_bNumConvert (pEntry, m_rszNumericList, plstS4Mapping, &bConverted) && bConverted) {
			_bRemoveCandidate (pEntry) ;
			/*()
			 *	Create ̃^C~O pEntry  register Ă̂ŁADestroy Ă΂ȂĂ
			 *	Session j̃^C~O Destroy sB
			 */
		}
	}
	return	TRUE ;
}

BOOL
CTSearchSession::bRemoveCandidate ()
{
	return	_bRemoveCandidate (m_pCurCandidate) ;
}

/* 	#4 ϊ̂߂ɕKvB_pRef  #4 ̓Ăϊʂw悤
 *	ݒ肷B
 *
 *	Ⴆ΁Au17v -> u#vŌ -> u#4v-> #4 ϊ
 *	s -> u\v ƂȂ킯A
 *	u\v̕ #4 Ƃ񂪕KvɂȂB(͓o^
 *	̎)
 */
BOOL
CTSearchSession::bLinkCandidate (
	CTSearchSessionCandidate*	pLink)
{
	if (m_pCurCandidate == NULL)
		return	FALSE ;
	m_pCurCandidate->vSetLink (pLink) ;
	return	TRUE ;
}

CTSearchSessionCandidate*
CTSearchSession::pGetCurrentPoint ()
{
	return	m_pCurCandidate ;
}

int
CTSearchSession::iGetNumberOfCandidate (
	BOOL*	pfContinue)
{
	if (pfContinue != NULL)
		*pfContinue	= (m_nSearchCount >= 0) ;
	return	m_nCandidate ;
}

BOOL
CTSearchSession::bNextCandidate ()
{
	CTSearchSessionCandidate*	pCandidate ;

	if (m_pCurCandidate == NULL)
		return	FALSE ;
	pCandidate		= m_pCurCandidate->pGetNext () ;
	if (pCandidate == NULL) {
		/*	Session pẮAB*/
		while (m_nSearchCount >= 0) {
			if (! _bSearch ())
				return	FALSE ;
			pCandidate	= m_pCurCandidate->pGetNext () ;
			if (pCandidate != NULL)
				break ;
		}
		if (pCandidate == NULL)
			return	FALSE ;
	}
	m_pCurCandidate	= pCandidate ;
	return	TRUE ;
}

BOOL
CTSearchSession::bPreviousCandidate ()
{
	CTSearchSessionCandidate*	pCandidate ;

	if (m_pCurCandidate == NULL)
		return	FALSE ;
	pCandidate		= m_pCurCandidate->pGetPrevious () ;
	if (pCandidate == NULL)
		return	FALSE ;
	m_pCurCandidate	= pCandidate ;
	return	TRUE ;
}

BOOL
CTSearchSession::bRewind ()
{
	m_pCurCandidate	= m_pTopCandidate ;
	return	TRUE ;
}

BOOL
CTSearchSession::bNumericp () const
{
	/* KvB*/
	return	_bNumericp () ;
}

LPCWSTR
CTSearchSession::pGetNumericList () const
{
	return	_bNumericp ()? m_rszNumericList : NULL ;
}

CTS4Candidate*
CTSearchSession::pGetNumericLink ()
{
	CTS4Candidate*	pRef	= NULL ; 

	/*	̏ꍇɂ́AuvȂĎꏟȂƂ͂ȂB
	 */
	if (m_pCurCandidate == NULL) 
		return	NULL ;

	pRef	= m_pCurCandidate->pGetLink2 () ;
	if (pRef == NULL)
		return	NULL ;

	return	pRef ;
}

LPCWSTR
CTSearchSession::pGetNumericKeyword (
	CTS4Candidate*			pNumCand,
	int*					pnKeywordLen)
{
	LPCWSTR	pwKeyword ;

	if (pNumCand == NULL)
		return	NULL ;
	/* error check */
	if (pNumCand->iGetKeyIndex () < 0 || pNumCand->iGetKeyIndex () >= m_vbufCandidate.iGetUsage ())
		return	NULL ;

	pwKeyword	= m_vbufCandidate.pGetBuffer () + pNumCand->iGetKeyIndex () ;
	if (pnKeywordLen != NULL)
		*pnKeywordLen	= wcslen (pwKeyword) ;
	return	pwKeyword ;
}

LPCWSTR
CTSearchSession::pGetNumericResult (
	CTS4Candidate*			pNumCand,
	int*					pnResultLen)
{
	if (pNumCand == NULL)
		return	NULL ;

	/* error check */
	if (pNumCand->iGetResultIndex () < 0 || (pNumCand->iGetResultIndex () + pNumCand->iGetResultLength ()) > m_vbufCandidate.iGetUsage ())
		return	NULL ;
	if (pnResultLen != NULL)
		*pnResultLen	= pNumCand->iGetResultLength () ;
	
	return	m_vbufCandidate.pGetBuffer () + pNumCand->iGetResultIndex () ;
}

CTS4Candidate*
CTSearchSession::pGetNextNumericLink (
	CTS4Candidate*			pNumCand)
{
	if (pNumCand == NULL)
		return	NULL ;
	return	pNumCand->pGetNext () ;
}

/*===============================================================
 */

BOOL
CTSearchSession::_bUpdateCandidateList (
	int		nPosition)
{
	int		nEndPos, iSlashCheck, nCount, nPrevPos ;
	BOOL		fDoubleQuoteCheck ;
	LPWSTR		ptr ;
#if defined (DBG) || defined (DEBUG)
	int					nFound = 0 ;
#endif

	ptr					= m_vbufCandidate.pGetBuffer () + nPosition ;
	nEndPos				= m_vbufCandidate.iGetUsage () ;
	fDoubleQuoteCheck	= FALSE ;
	nCount				= 0 ;
	iSlashCheck			= 0 ;
	if (nPosition == 0) {
		while (nPosition < nEndPos && ! _bIsSeparateChar (*ptr)) {
			ptr			++ ;
			nPosition	++ ;
		}
	}
	nPrevPos			= nPosition ;

	while (nPosition < nEndPos) {
		/* _uNEH[g𖳎_ŊmFB*/
		if (iSlashCheck <= 0 && *ptr == L'\\')
			iSlashCheck	= 2 ;
		if (iSlashCheck <= 0) {
			switch (*ptr){
				/* Double Quote ̊Ԃ͗vӁBobNXbV̎ł͂Ȃ_u
				 * NEH[gɈ͂܂ꂽ͈̕(ƂƂ
				 * Bɂ̒ slash ĂĂ邱Ƃɂ
				 * B*/  
			case 0x22:
				if (!fDoubleQuoteCheck){
					/* _uNEH[g̒ɂȂ΁A̕
					 * 񂪊JnƂɂȂB*/
					fDoubleQuoteCheck	= TRUE ;
				} else {
					/* _uNEH[g̒Ń_uNEH[gtcB 
					 * RA_uNEH[g̒𔲂ƂɂȂB*/
					fDoubleQuoteCheck 	= FALSE ;
				}
				/* J[\ʒüړB*/
				ptr			++ ;
				nPosition	++ ;
				nCount		++ ;
				continue ;
				
				/* ̏I[tĂ܂ꍇ̏B*/
			case L'\n' :
			case L'\r' :
			case L'\0' :
				/* ɏI[̓_uNEH[g̒낤낤
				 * łȂcB_uNEH[g̒ɓoꂵG[
				 * Ỏ҂łȂB*/
				goto exit_loop ;

			default :
				/* ^͌ƌ̊ԂɕK݂B*/
				if (_bIsSeparateChar (*ptr)) {
					/* _uNEH[g̒ɓoꂵ̂ł΁A{̋@\͖B*/
					if (fDoubleQuoteCheck)
						break ;
				
					/* ̈ʒu̕͏I[č\܂BF؂ȊOɖ
					 * ĂȂ̂BȂ NUL ł\B*/
					*ptr	= L'\0' ;
					/* Vm[hB*/
					if (nCount > 0 && nPrevPos < nPosition) {
						CTSearchSessionCandidate*	pNewEntry ;
						BOOL	bRetval ;

						/*	(annotation) ݂\B\cł邩
						 *	ۂ͒uĂāȀȂ΁B
						 */
						pNewEntry	= _pCreateCandidate (nPrevPos, nCount, &bRetval) ;
						if (pNewEntry != NULL)
							(void) _bInsertCandidateLast (pNewEntry) ;
#if defined (DBG) || defined (DEBUG)
						nFound ++ ;
#endif
					}
					/* ꏊyсA|C^₷B*/
					ptr			++ ;
					nPosition	++ ;
					/* O(܂荡̏ꏊ)A"/" tꏊ pPos ɋLB*/
					nPrevPos	= nPosition ;
					nCount		= 0 ;
					continue ;
				} else {
					break ;
				}
			}
		}
		ptr			++ ;
		nPosition	++ ;
		nCount		++ ;
		if (iSlashCheck > 0)
			iSlashCheck	-- ;
	}
  exit_loop:
#if defined (DBG) || defined (DEBUG)
	DEBUGPRINTF ((TEXT ("_UpdateCandidateList (found: %d)\n"), nFound)) ;
#endif
	return	nPrevPos ;
}

CTSearchSessionCandidate*
CTSearchSession::_pCreateCandidate (
	int			nPosition,
	int			nCount,
	BOOL*		pfRetval)	/* ɑ݂ĂǂB*/
{
	CTSearchSessionCandidate*	pEntry ;
	LPCWSTR	wstring ;

	wstring	= m_vbufCandidate.pGetBuffer () + nPosition ;
	if (_bIsExistCandidatep (wstring, nCount)) {
		if (pfRetval)
			*pfRetval	= TRUE ;
		return	NULL ;
	}
	pEntry	= new CTSearchSessionCandidate (nPosition) ;
	if (pEntry == NULL) {
		if (pfRetval)
			*pfRetval	= FALSE ;
		return	NULL ;
	}
	if (! _bRegisterCandidate (pEntry)) {
		delete	pEntry ;
		if (pfRetval)
			*pfRetval	= FALSE ;
		return	NULL ;
	}
	if (pfRetval)
		*pfRetval	= FALSE ;
	return	pEntry ;
}

BOOL
CTSearchSession::_bInsertCandidateLast (
	CTSearchSessionCandidate*	pEntry)
{
	if (m_pLastCandidate != NULL) 
		m_pLastCandidate->vSetNext (pEntry) ;
	if (m_pTopCandidate == NULL)
		m_pTopCandidate	= pEntry ;

	pEntry->vSetPrevious (m_pLastCandidate) ;
	m_pLastCandidate	= pEntry ;
	m_nCandidate	++ ;
	return	TRUE ;
}

CTSearchSessionCandidate*
CTSearchSession::_pGetCandidate (
	LPCWSTR						wstring,
	int							nstring)
{
	CTSearchSessionCandidate*	pEntry ;
	int		nIndex, nPosition, nszBuffer, nCompare ;
	LPCWSTR	wszBuffer ;

	nIndex		= (unsigned int)*wstring * nstring % CANDIDATEHASHSIZE ;
	ASSERT (0 <= nIndex && nIndex < CANDIDATEHASHSIZE) ;
	pEntry		= m_rpTblCandidate [nIndex] ;
	if (pEntry == NULL)
		return	NULL ;
	wszBuffer	= m_vbufCandidate.pGetBuffer () ;
	nszBuffer	= m_vbufCandidate.iGetUsage () ;
	while (pEntry != NULL) {
		nPosition	= pEntry->iGetValue () ;
		ASSERT (0 <= nPosition && nPosition < nszBuffer) ;
		nCompare	= wcsncmp (wstring, wszBuffer + nPosition, nstring) ;
		if (nCompare == 0) {
			if (*(wszBuffer + nPosition + nstring) == L'\0') 
				return	pEntry ;
			nCompare	= 1 ;
		}
		if (nCompare > 0) {
			pEntry	= pEntry->pGetRight () ;
		} else {
			pEntry	= pEntry->pGetLeft () ;
		}
	}
	return	NULL ;
}


BOOL
CTSearchSession::_bIsExistCandidatep (
	LPCWSTR	wstring,
	int		nstring)
{
	CTSearchSessionCandidate*	pEntry ;
	int		nIndex, nPosition, nszBuffer, nCompare ;
	LPCWSTR	wszBuffer ;

	nIndex		= (unsigned int)*wstring * nstring % CANDIDATEHASHSIZE ;
	ASSERT (0 <= nIndex && nIndex < CANDIDATEHASHSIZE) ;
	pEntry		= m_rpTblCandidate [nIndex] ;
	if (pEntry == NULL)
		return	FALSE ;
	wszBuffer	= m_vbufCandidate.pGetBuffer () ;
	nszBuffer	= m_vbufCandidate.iGetUsage () ;
	while (pEntry != NULL) {
		nPosition	= pEntry->iGetValue () ;
		ASSERT (0 <= nPosition && nPosition < nszBuffer) ;
		nCompare	= wcsncmp (wstring, wszBuffer + nPosition, nstring) ;
		if (nCompare == 0) {
			if (*(wszBuffer + nPosition + nstring) == L'\0') 
				return	TRUE ;
			nCompare	= 1 ;
		}
		if (nCompare > 0) {
			pEntry	= pEntry->pGetRight () ;
		} else {
			pEntry	= pEntry->pGetLeft  () ;
		}
	}
	return	FALSE ;
}

BOOL
CTSearchSession::_bRegisterCandidate (
	CTSearchSessionCandidate*	pNewEntry)
{
	CTSearchSessionCandidate*	pEntry ;
	int		nIndex, nPosition, nstring, nszBuffer, nCompare ;
	LPCWSTR	wszBuffer ;
	LPCWSTR	wstring ;

	wszBuffer	= m_vbufCandidate.pGetBuffer () ;
	nszBuffer	= m_vbufCandidate.iGetUsage () ;
	wstring		= wszBuffer + pNewEntry->iGetValue () ;
	nstring		= wcslen (wstring) ;
	nIndex		= (unsigned int)*wstring * nstring % CANDIDATEHASHSIZE ;
	ASSERT (0 <= nIndex && nIndex < CANDIDATEHASHSIZE) ;
	pEntry		= m_rpTblCandidate [nIndex] ;
	if (pEntry == NULL) {
		m_rpTblCandidate [nIndex]	= pNewEntry ;
		return	TRUE ;
	}

	while (pEntry != NULL) {
		nPosition	= pEntry->iGetValue () ;
		ASSERT (0 <= nPosition && nPosition < nszBuffer) ;
		nCompare	= wcsncmp (wstring, wszBuffer + nPosition, nstring) ;
		if (nCompare == 0) {
			if (*(wszBuffer + nPosition + nstring) == L'\0') 
				return	TRUE ;
			nCompare	= 1 ;
		}
		if (nCompare > 0) {
			if (pEntry->pGetRight () == NULL) {
				pEntry->vSetRight (pNewEntry) ;
				return	TRUE ;
			}
			pEntry	= pEntry->pGetRight () ;
		} else {
			if (pEntry->pGetLeft () == NULL) {
				pEntry->vSetLeft (pNewEntry) ;
				return	TRUE ;
			}
			pEntry	= pEntry->pGetLeft  () ;
		}
	}
	return	FALSE ;
}

BOOL
CTSearchSession::_bCleanupHashTable ()
{
	int		i ;

	for (i = 0 ; i < CANDIDATEHASHSIZE ; i ++) {
		if (m_rpTblCandidate [i] != NULL)
			_bCleanupHashTableSub (m_rpTblCandidate [i]) ;
	}
	return	TRUE ;
}

BOOL
CTSearchSession::_bCleanupHashTableSub (
	CTSearchSessionCandidate*	pEntry)
{
	ASSERT (pEntry != NULL) ;

	if (pEntry->pGetLeft () != NULL) 
		_bCleanupHashTableSub (pEntry->pGetLeft ()) ;
	if (pEntry->pGetRight () != NULL)
		_bCleanupHashTableSub (pEntry->pGetRight ()) ;
	delete	pEntry ;
	return	TRUE ;
}

BOOL
CTSearchSession::_bIsSeparateChar (
	WCHAR		wch) const
{
	return	(wch == L'/') ;
}

BOOL
CTSearchSession::_bInsertCandidateNext (
	CTSearchSessionCandidate*	pNode,
	CTSearchSessionCandidate*	pNewNode)
{
	if (pNewNode == NULL)
		return	FALSE ;

	if (pNode == NULL) {
		pNewNode->vSetNext (m_pTopCandidate) ;
		pNewNode->vSetPrevious (NULL) ;
		if (m_pTopCandidate != NULL) {
			m_pTopCandidate->vSetPrevious (pNewNode) ;
			m_pTopCandidate		= pNewNode ;
		} else {
			m_pCurCandidate		= pNewNode ;
			m_pTopCandidate		= pNewNode ;
			m_pLastCandidate	= pNewNode ;
		}
	} else {
		CTSearchSessionCandidate*	pNextNode ;

		pNextNode	= pNode->pGetNext () ;
		pNode->vSetNext (pNewNode) ;
		pNewNode->vSetPrevious (pNode) ;
		pNewNode->vSetNext (pNextNode) ;
		if (pNextNode != NULL) {
			pNextNode->vSetPrevious (pNewNode) ;
		} else {
			/*if (m_pLastCandidate == pNode) */
			m_pLastCandidate	= pNewNode ;
		}
	}
	m_nCandidate	++ ;
	return	TRUE ;
}

BOOL
CTSearchSession::_bRemoveCandidate (
	CTSearchSessionCandidate*	pEntry)
{
	CTSearchSessionCandidate*	pPrevEntry ;
	CTSearchSessionCandidate*	pNextEntry ;

	if (pEntry == NULL)
		return	FALSE ;

	pNextEntry	= pEntry->pGetNext () ;
	pPrevEntry	= pEntry->pGetPrevious () ;

	if (pPrevEntry != NULL) {
		pPrevEntry->vSetNext (pNextEntry) ;
	} else {
		m_pTopCandidate	= pNextEntry ;
	}
	if (pNextEntry != NULL) {
		pNextEntry->vSetPrevious (pPrevEntry) ;
	} else {
		m_pLastCandidate	= pPrevEntry ;
	}
	if (pEntry == m_pCurCandidate) {
		if (pNextEntry != NULL) {
			m_pCurCandidate	= pNextEntry ;
		} else {
			m_pCurCandidate	= pPrevEntry ;
		}
	}
#if defined (DBG) || defined (DEBUG)
	if (m_nCandidate <= 0) {
		DEBUGPRINTF ((TEXT ("Something wrong in RemoveCandiate (%d < 0)\n"), m_nCandidate)) ;
	}
#endif
	m_nCandidate	-- ;
	return	TRUE ;
}

BOOL
CTSearchSession::_bMultipleNumConvert ()
{
	CTSearchSessionCandidate*	pNode ;
	CTSearchSessionCandidate*	pNextNode ;
	BOOL						bConverted ;

	pNode	= m_pTopCandidate ;
	while (pNode != NULL) {
		pNextNode	= pNode->pGetNext () ;
		/* ł Candidate  num-convert B*/
		bConverted	= FALSE ;
		if (_bNumConvert (pNode, m_rszNumericList, NULL, &bConverted) && bConverted) {
			_bRemoveCandidate (pNode) ;
		}
		pNode		= pNextNode ;
	}
	return	TRUE ;
}

BOOL
CTSearchSession::_bNumericp () const
{
	return	FALSE ;
}

BOOL
CTSearchSession::_bNumConvert (
	CTSearchSessionCandidate*	pEntry,
	LPCWSTR						wstrNumList,
	const CTS4Mapping*			plstS4Mapping,
	BOOL*						pbConverted)
{
	CVarbuffer < CVarbuffer < WCHAR, 16 >*, 16 >	vbuf ;
	CVarbuffer < int, 16 >			vbufPos ;
	CVarbuffer < WCHAR, 16 >		vbufDest ;
	CVarbuffer < CTS4NumIndex, 16 >	vbufS4Index ;
	CVarbuffer < int, 16 >			vbufNumIndex ;
	CVarbuffer < WCHAR, 16 >*		pvbuf ;
	CVarbuffer < WCHAR, 16 >**		ppvbuf ;
	LPCWSTR		wstrSrc ;
	int			nSrcLen ;
	LPCWSTR		pSrc, pSrcEnd, pNum ;
	int			i, nNumIndex ;
	int*		pnNumIndex ;
	const CTS4Mapping*	pS4Node ;
	BOOL		bRetval		= FALSE ;
	BOOL		bConverted	= FALSE ;

	if (pEntry == NULL)
		return	FALSE ;

	wstrSrc	= m_vbufCandidate.pGetBuffer () + pEntry->iGetValue () ;
	nSrcLen	= wcslen (wstrSrc) ;

	pSrc		= wstrSrc ;
	pSrcEnd		= wstrSrc + nSrcLen ;
	pNum		= wstrNumList ;
	nNumIndex	= 0 ;
	pS4Node		= plstS4Mapping ;

	while (pSrc < pSrcEnd) {
		if (*pSrc == L'#') {
			pSrc	++ ;
			if (pSrc < pSrcEnd && L'0' <= *pSrc && *pSrc <= L'9') {
				int		nNumType ;

				/* number WJB*/
				nNumType	= *pSrc - L'0';
				/*  #4 ϊꍇɂ͓1ɂȂȂB*/
				if (*pNum != L'\0') {
					int		nNumLen ;

					nNumLen	= wcslen (pNum) ;
					pvbuf	= new CVarbuffer < WCHAR, 16 > ;
					if (pvbuf == NULL)
						goto	exit_func ;

					/*	#4 ̓WJ@ɗ^Ăꍇɂ́AsȂB
					 */
					if (nNumType == 4										&& 
						pS4Node != NULL										&& 
						pS4Node->iGetResultLength () > 0					&&
						!wcsncmp (pNum, pS4Node->pGetKeyword (), nNumLen)	&&
						pS4Node->pGetKeyword () [nNumLen] == L'\0') {
						if (! pvbuf->bAdd (pS4Node->pGetResult (), pS4Node->iGetResultLength ()) ||
							! pvbuf->bAdd (L"\0", 2)) {
							delete	pvbuf ;
							goto	exit_func ;
						}
					} else {
						if (! _bNumExp (pNum, nNumLen, nNumType, pvbuf)) {
							delete	pvbuf ;
							goto	exit_func ;
						}
					}
					if (nNumType == 4 && 
						pS4Node != NULL) {
						pS4Node	= pS4Node->pGetNext () ;
					}
					if (! vbuf.bAdd (&pvbuf, 1)) {
						delete	pvbuf ;
						goto	exit_func ;
					}
					pNum	+= nNumLen + 1 ;
					nNumIndex	++ ;
				}
			}
		} 
		pSrc	++ ;
	}
	if (vbuf.iGetUsage () <= 0) {
		*pbConverted	= FALSE ;
		goto	exit_func ;
	}

	if (! vbufNumIndex.bRequire (nNumIndex))
		goto	exit_func ;

	pnNumIndex	= vbufNumIndex.pGetBuffer () ;
	pNum		= wstrNumList ;
	for (i = 0 ; i < nNumIndex ; i ++) {
		if (pNum == NULL || *pNum == L'\0') {
			pnNumIndex [i]	= -1 ;
		} else {
			int		nNumLen, nPosition ;
			nNumLen			= wcslen (pNum) ;
			nPosition		= m_vbufCandidate.iGetUsage () ;
			if (! m_vbufCandidate.bAdd (pNum, nNumLen + 1))
				goto	exit_func ;
			pnNumIndex [i]	= nPosition ;
			pNum	+= nNumLen + 1 ;
		}
	}

	bRetval	= TRUE ;

	/* ̗񋓕oăeXgcBɂ邩ȁB*/
	{
		int*			pPosTop ;
		int*			pPosEnd ;
		int*			pPos, nCount ;
		CVarbuffer < WCHAR, 16 >**	ppvbufTop ;

		if (! vbufPos.bRequire (vbuf.iGetUsage ()))
			goto	exit_func ;

		pPosTop	= vbufPos.pGetBuffer () ;
		pPosEnd	= pPosTop + vbufPos.iGetUsage () ;
		pPos	= pPosTop ;
		while (pPos < pPosEnd)
			*pPos ++	= 0 ;
		ppvbufTop	= vbuf.pGetBuffer () ;

		for ( ; ; ) {
			pSrc		= wstrSrc ;
//			pSrcEnd		= wstrSrc + nSrcLen ;
			vbufDest.vClear () ;
			vbufS4Index.vClear () ;
			pPos		= pPosTop ;
			ppvbuf		= ppvbufTop ;

			/*	lWJ邽߂ɁAIWi̕ϊʂXLB
			 *
			 *	wstrNumList:	nul ŋ؂ꂽl̃XgBIWi̕ϊʂ
			 *					߂鎞ɐ؂oĂB
			 *					``123\0456\0\0'' ̂悤ȊB 
			 *	pSrc:			IWi̕ϊ
			 *					Ⴆ΁ApSrc ɂ ``#2#4'' ̂悤ɓĂB
			 *	*ppvbuf:		X̐l̓WJʂĂB
			 *					 wstrNumList ̗ႾƁA
			 *					ppvbuf [0] -> 123 ̓WJ (by #2)
			 *					ppvbuf [1] -> 456 ̓WJ (by #4)
			 *	pDest:			#2  #4  *ppvbuf ̒̌ŒuʂB
			 *					``123STU'' ƂB
			 */
			while (pSrc < pSrcEnd) {
				LPCWSTR	pSrcBack ;

				pSrcBack	= pSrc ;
				while (pSrc < pSrcEnd && *pSrc != L'#') 
					pSrc	++ ;
				if (pSrcBack < pSrc) {
					if (! vbufDest.bAdd (pSrcBack, pSrc - pSrcBack)) 
						goto	exit_func ;
				}
				if (pSrc < pSrcEnd && *pSrc == L'#') {
					pSrc	++ ;
					if (pSrc < pSrcEnd && L'0' <= *pSrc && *pSrc <= L'9') {
						if (pPos < pPosEnd) {
							LPCWSTR	pNum ;
							int		nNum, nPos ;

							pNum	= (*ppvbuf)->pGetBuffer () ;
							nNum	= (*ppvbuf)->iGetUsage () ;
							nPos	= *pPos ;
							if (nNum <= 0 || *(pNum + nPos) == L'\0') {
								/* Empty */
							} else {
								int		n ;

								/*	 #4 A#4 ɂĉɕϊꂽ̂
								 *	ʂ@ޕKvBށAʎ̂ pDest ɓ
								 *	SȌ`œ邩͕ȂAۑĂ
								 *	ǂH
								 *	[ށB
								 *
								 *	struct {
								 *		int		m_nResultPos ;
								 *		int		m_nResultLen ;
								 *	}
								 */
								n	= wcslen (pNum + nPos) ;
								if (*pSrc == L'4') {
									struct CTS4NumIndex	s4n ;
									s4n.m_nIndex	= pnNumIndex [ppvbuf - ppvbufTop] ;
									s4n.m_nPos		= vbufDest.iGetUsage () ;
									s4n.m_nLen		= n ;
									if (! vbufS4Index.bAdd (&s4n, 1)) {
										/* out of memory */
										goto	exit_func ;
									}
								}
								if (! vbufDest.bAdd (pNum + nPos, n)) {
									/* out of memory */
									goto	exit_func ;
								}
							}
							pPos	++ ;
							ppvbuf	++ ;
						}
						pSrc ++ ;
					}
				}
			}

			/* separator  nul KvB*/
			nCount		= vbufDest.iGetUsage () ;
			/* Session  Add */
			if (nCount > 0) {
				CTSearchSessionCandidate*	pNewEntry ;
				LPCWSTR	pWord ;
				int		nWordLen ;
				BOOL	bRetval ;
				int		nPosition ;

				if (! vbufDest.bAdd (L"", 1)) 
					goto	exit_func ;

				nPosition	= m_vbufCandidate.iGetUsage () ;
				pWord		= vbufDest.pGetBuffer () ;
				nWordLen	= vbufDest.iGetUsage  () ;

				pNewEntry	= _pGetCandidate (pWord, nWordLen - 1) ;
				if (pNewEntry == NULL) {
					if (! m_vbufCandidate.bAdd (pWord, nWordLen)) {
						/*  error  critical */
						goto	exit_func ;
					}
					/* nPosition  pDest - wbuf o^ĂB*/
					pNewEntry	= _pCreateCandidate (nPosition, nWordLen, &bRetval) ;
					if (bRetval || pNewEntry == NULL) 
						goto	exit_func ;
					if (! _bInsertCandidateNext (pEntry, pNewEntry)) {
						delete	pNewEntry ;
						goto	exit_func ;
					}
					/*	WJꂽʂIWiւ̃N\ĂB
					 *	͎o^̎ɉ痈̂ȂȂȂ
					 *	悤ɁB
					 */
					pNewEntry->vSetLink (pEntry) ;
				}
				bConverted	= TRUE ;

				{
					struct CTS4NumIndex*	pS4Index ;
					CTS4Candidate*			pNode ;
					CTS4Candidate*			pTop	= NULL ;
					int						nS4Indices, nBase ;

					pS4Index	= (struct CTS4NumIndex*) vbufS4Index.pGetBuffer () ;
					nS4Indices	= vbufS4Index.iGetUsage () ;
					nBase		= pNewEntry->iGetValue () ;
					while (nS4Indices > 0) {
						if (pS4Index->m_nIndex >= 0) {
#if defined (DEBUG) || defined (_DEBUG)
							{
								WCHAR	buf1 [256], buf2 [256] ;
								LPCWSTR	ptr ;
								int		n ;
								ptr	= m_vbufCandidate.pGetBuffer () ;
#if __STDC_WANT_SECURE_LIB__
								wcsncpy_s (buf1, MYARRAYSIZE (buf1), ptr + pS4Index->m_nIndex, MYARRAYSIZE (buf1) - 1) ;
#else
								wcsncpy (buf1, ptr + pS4Index->m_nIndex, MYARRAYSIZE (buf1) - 1) ;
#endif
								buf1 [MYARRAYSIZE (buf1) - 1]	= L'\0' ;
								n	= MIN (pS4Index->m_nLen, MYARRAYSIZE (buf2) - 1) ;
#if __STDC_WANT_SECURE_LIB__
								wcsncpy_s (buf2, MYARRAYSIZE (buf2), ptr + nBase + pS4Index->m_nPos, n) ;
#else
								wcsncpy (buf2, ptr + nBase + pS4Index->m_nPos, n) ;
#endif
								buf2 [n]	= L'\0' ;
								DebugPrintf (TEXT ("#4 ǉ: P: \"%s\"(pos:%d) :\"%s\" (pos:%d, len:%d)\n"), buf1, pS4Index->m_nIndex, buf2, pS4Index->m_nPos + nBase, pS4Index->m_nLen) ;
							}
#endif
							pNode	= new CTS4Candidate (pS4Index->m_nIndex, nBase + pS4Index->m_nPos, pS4Index->m_nLen) ; 
							if (pTop != NULL) {
								pTop->vSetNext (pNode) ;
							} else {
								pTop	= pNode ;
							}
						}
						pS4Index	++ ;
						nS4Indices	-- ;
					}
					/*	͒ǉɂȂB
					 *	#4 ̏ꍇɂ́A#4 ł邱ƂɉāAĕ]̂o
					 *	Kv邩B
					 *	() 1 -> #4 ->  ̏ꍇALink1 ɂ "#4" ALink2 ɂ "1" 
					 *	B
					 */
					pNewEntry->vAddLink2 (pTop) ;
				}
			}

			/* ɑ鏈B*/
			pPos		= pPosTop ;
			ppvbuf		= vbuf.pGetBuffer () ;
			while (pPos < pPosEnd) {
				LPCWSTR	pNum ;
				int		nNum, nPos ;
				pNum	= (*ppvbuf)->pGetBuffer () ;
				nNum	= (*ppvbuf)->iGetUsage () ;
				nPos	= *pPos ;
				if (nNum <= 0 || *(pNum + nPos) == L'\0') {
					*pPos	= 0 ;
					/* continue */
				} else {
					int		nLen, nNewPos ;

					nLen	= wcslen (pNum + nPos) ;
					nNewPos	= nPos + nLen + 1 ;
					if (nNewPos >= nNum || *(pNum + nNewPos) == L'\0') {
						*pPos	= 0 ;
						/* continue */
					} else {
						*pPos	= nNewPos ;
						break ;
					}
				}
				pPos	++ ;
				ppvbuf	++ ;
			}
			if (pPos >= pPosEnd)
				break ;
		}
	}
	*pbConverted	= bConverted ;

exit_func:
	ppvbuf	= vbuf.pGetBuffer () ;
	for (i = 0 ; i < vbuf.iGetUsage () ; i ++) {
		if (ppvbuf [i] != NULL) {
			delete	ppvbuf [i] ;
			ppvbuf [i]	= NULL ;
		}
	}
	return	bRetval ;
}

/*
(defun skk-num-exp (num type)
  "ascii  (string)  NUM  TYPE ɏ]ϊAϊ̕ԂB
TYPE ͉L̒ʂB
0 -> ϊ
1 -> Sp֕ϊ
2 -> ֕ϊ (ʎȂ)
3 -> ֕ϊ (ʎ)
4 -> ̐̂̂L[ɂĎČ
5 ->  (`ȂǂŎgp镶gp) ֕ϊ (ʎ)
9 -> Ŏgp鐔 (\"Rl\" Ȃ) ɕϊ"
  (save-match-data
    (let ((fun (cdr (assq type skk-num-type-alist))))
      (when fun
	(funcall fun num)))))
	*/
BOOL
_bNumExp (
	LPCWSTR						strNum,
	int							nNumLen,
	int							nType,
	CVarbuffer < WCHAR, 16 >*	pvbuf)
{
	switch (nType) {
	case	0:
		if (pvbuf->bAdd (strNum, nNumLen) && pvbuf->bAdd (L"\0", 1))
			return	TRUE ;
		return	FALSE ;
	case	1:
		return	_bNumJisx0208Latin (strNum, nNumLen, pvbuf) ;
	case	2:
		return	_bNumType2Kanji (strNum, nNumLen, pvbuf) ;
	case	3:
		return	_bNumType3Kanji (strNum, nNumLen, pvbuf) ;
	case	4:	/* ̓eXgłȂB*/
		return	_bNumType4Kanji (strNum, nNumLen, pvbuf) ;
	case	5:
		return	_bNumType5Kanji (strNum, nNumLen, pvbuf) ;
	case	9:
		return	_bNumType9Shogi (strNum, nNumLen, pvbuf) ;
	default:
		break ;
	}
	return	TRUE ;
}

/*
(defun skk-num-jisx0208-latin (num)
  "ascii  NUM Sp̕ɕϊAϊ̕ԂB
Ⴆ \"45\"  \"ST\" ɕϊB"
  (let ((candidate
	 (mapconcat (function
		     (lambda (c)
		       (skk-num-get-suuji c skk-num-alist-type1)))
		    num "")))
    (unless (string= candidate "")
      candidate)))*/
BOOL
_bNumJisx0208Latin (
	LPCWSTR			strNum,
	int				nNumLen,
	CVarbuffer < WCHAR, 16 >*	pvbuf)
{
	/*  mapping table  customize l邩ȂBӁB*/
	static LPCWSTR	strMapTable	= L"OPQRSTUVWX" ;
	LPCWSTR		pSrc,  pSrcEnd ;
	WCHAR		wch ;

	if (! _bIntOrFloatNump (strNum, nNumLen))
		return	TRUE ;

	pSrc		= strNum ;
	pSrcEnd		= strNum + nNumLen ;
	while (pSrc < pSrcEnd) {
		if (L'0' <= *pSrc && *pSrc <= L'9') {
			wch	= *(strMapTable + (*pSrc - L'0')) ;
		} else if (L'.' == *pSrc) {
			wch	= L'D' ;
		} else {
			wch	= *pSrc ;
		}
		if (! pvbuf->bAdd (&wch, 1))
			return	FALSE ;
		pSrc	++ ;
	}
	/* NUL NUL ŏI邱ƁB*/
	return	pvbuf->bAdd (L"\0", 1) ;
}

/*
(defun skk-num-type2-kanji (num)
  "ascii  NUM ̕ɕϊAϊ̕ԂB
Ⴆ΁A\"45\"  \"l\" ɕϊB"
  (save-match-data
    (when (skk-num-int-p num)
      (let ((candidate
	     (mapconcat (function
			 (lambda (c)
			   (skk-num-get-suuji
			    c
			    skk-num-alist-type2)))
			num "")))
	(unless (string= candidate "")
	  candidate))))) */
BOOL
_bNumType2Kanji (
	LPCWSTR						strNum,
	int							nNumLen,
	CVarbuffer < WCHAR, 16 >*	pvbuf)
{
	/*  mapping table  customize l邩ȂBӁB*/
	static LPCWSTR	strMapTable	= L"ZOlܘZ" ;
	LPCWSTR		pSrc,  pSrcEnd ;
	WCHAR		wch ;

	if (! _bIntNump (strNum, nNumLen))
		return	TRUE ;

	pSrc		= strNum ;
	pSrcEnd		= strNum + nNumLen ;
	while (pSrc < pSrcEnd) {
		if (L'0' <= *pSrc && *pSrc <= L'9') {
			wch	= *(strMapTable + (*pSrc - L'0')) ;
		} else {
			wch	= *pSrc ;
		}
		if (! pvbuf->bAdd (&wch, 1))
			return	FALSE ;
		pSrc	++ ;
	}

	/* NUL NUL ŏI邱ƁB*/
	return	pvbuf->bAdd (L"\0", 1) ;
}

/*
(defun skk-num-type3-kanji (num)
  "ascii  NUM ̕ɕϊ (ʎ)Aϊ̕
ԂBႦ \"1021\"  \"\\" ɕϊB"
  (save-match-data
    (when (skk-num-int-p num)
      ;; _܂܂Ȃ
      (skk-num-to-kanji num 'type3))))
	  */
BOOL
_bNumType3Kanji (
	LPCWSTR						strNum,
	int							nNumLen,
	CVarbuffer < WCHAR, 16 >*	pvbuf)
{
	LPCWSTR		pSrc,  pSrcEnd ;
	LPWSTR		pDest, pDestEnd ;
	WCHAR		wbuf [256] ;

	if (! _bIntNump (strNum, nNumLen))
		return	TRUE ;

	/* 疳ʑ吔܂ł͕\B*/
	if (nNumLen >= sizeof (_wstrNumericUnit) / sizeof (_wstrNumericUnit [0]) * 4)
		return	TRUE ;

	pSrc		= strNum ;
	pSrcEnd		= strNum + nNumLen ;
	pDest		= wbuf ;
	pDestEnd	= wbuf + sizeof (wbuf) / sizeof (wbuf [0]) ;

	if (nNumLen == 1 && *strNum == L'0') {
		if (pDest < pDestEnd)
			*pDest ++	= _wstrKanjiDigits [0] ;
	} else {
		int		nDigitNumber, nLength ;
		BOOL	fSilent ;

		nLength			= nNumLen ;
		nDigitNumber	= ((nLength - 1) / 4) - 1 ;
		/* 傫߂鐔łH */
		if (nLength < 17){
			fSilent	= TRUE ;
			nLength	= nNumLen ;
			while (pSrc < pSrcEnd && pDest < pDestEnd) {
				if (L'1' < *pSrc || (*pSrc != L'0' && (nLength % 4) == 1)) {
					if (pDest < pDestEnd) 
						*pDest		++	= _wstrKanjiDigits [*pSrc - L'0'] ;
				} 
				if (*pSrc != L'0'){
					if (nLength % 4 == 2){
						if (pDest < pDestEnd)
							*pDest		++	= L'\' ;
					} else if (nLength % 4 == 3){
						if (pDest < pDestEnd)
							*pDest		++	= L'S' ;
					} else if (nLength % 4 == 0){
						if (pDest < pDestEnd)
							*pDest		++	= L'' ;
					}
					fSilent	= FALSE ;
				}
				/* ʂB*/
				if (!fSilent || *pSrc != L'0'){
					int		nDigitNumber	= nLength - 1 ;

					if (nDigitNumber % 4 == 0){
						nDigitNumber	= nDigitNumber / 4 - 1 ;
						if (nDigitNumber >= 0){
							int		n, nMax ;

							n		= wcslen (_wstrNumericUnit [nDigitNumber]) ;
							nMax	= pDestEnd - pDest ;
							if (0 < n) {
								if (nMax < n) 
									n	= nMax ;
								memcpy (pDest, _wstrNumericUnit [nDigitNumber], n * sizeof (WCHAR)) ;
								pDest	+= n ;
							}
							fSilent	= TRUE ;
						}
					}
				}
				nLength	-- ;
				pSrc	++ ;
			}
		}
		/* ``疜'' Ƃ\݂΁A``疜'' ɕϊBresult ǂݒĂ͔̑cB*/
		pSrc	= wbuf ;
		pSrcEnd	= pDest ;
		nLength	= pSrcEnd - pSrc ;
		while (pSrc < pSrcEnd) {
			if (*pSrc == L'' && (pSrc + 1) < pSrcEnd && *(pSrc + 1) == L'') {
				int		nPosition, nMove ;

				if (pSrc > wbuf) {
					LPCWSTR	pCheck ;

					pCheck	= _wstrKanjiDigits + 1 ;
					while (*pCheck != L'\0' && *pCheck != *(pSrc - 1))
						pCheck	++ ;
					if (*pCheck != L'\0') {
						break ;
					}
				} 

				/* 擪 ``'' ǉKvB*/
				nPosition	= pSrc - wbuf ;
				if (nLength >= sizeof (wbuf) / sizeof (wbuf [0])) {
					nMove	= sizeof (wbuf) / sizeof (wbuf [0]) - nPosition - 1 ;
				} else {
					nMove	= nLength - nPosition ;
					pDest	++ ;
				}
//				printf ("position = %d, move = %d\n", nPosition, nMove) ;
				if (nMove > 0)
					memmove (wbuf + nPosition + 1, wbuf + nPosition, nMove * sizeof (WCHAR)) ;
				wbuf [nPosition]	= L'' ;
				break ;
			}
			pSrc	++ ;
		}
	}
	if (! pvbuf->bAdd (wbuf, pDest - wbuf))
		return	FALSE ;

	/* NUL NUL ŏI邱ƁB*/
	return	pvbuf->bAdd (L"\0", 1) ;
}


/*
(defconst skk-num-alist-type5
  '((ju . "E") (hyaku . "S") (sen . "")
    (man . "") (oku . "") (cho . "") (kei . "")
    (?0 . "") (?1 . "") (?2 . "") (?3 . "Q")
    (?4 . "l") (?5 . "") (?6 . "Z") (?7 . "")
    (?8 . "") (?9 . "") (?\  . ""))
	*/
BOOL
_bNumType5Kanji (
	LPCWSTR						strNum,
	int							nNumLen,
	CVarbuffer < WCHAR, 16 >*	pvbuf)
{
	LPCWSTR		pSrc,  pSrcEnd ;

	/* [0-9]* łȂ΃G[B*/
	if (! _bIntNump (strNum, nNumLen))
		return	TRUE ;

	/* 疳ʑ吔܂ł͕\B*/
	if (nNumLen >= sizeof (_wstrNumericUnit) / sizeof (_wstrNumericUnit [0]) * 4)
		return	TRUE ;

	pSrc		= strNum ;
	pSrcEnd		= strNum + nNumLen ;

	if (nNumLen == 1 && *strNum == L'0') {
		if (! pvbuf->bAdd (&_wstrKanjiDigitsType5 [0], 1))
			return	FALSE ;
	} else {
		int		nDigitNumber, nLength ;
		BOOL	fSilent ;

		nLength			= nNumLen ;
		nDigitNumber	= ((nLength - 1) / 4) - 1 ;
		/* 傫߂鐔łH */
		if (nLength < 17){
			fSilent	= TRUE ;
			nLength	= nNumLen ;
			while (pSrc < pSrcEnd) {
				if (L'1' < *pSrc || (*pSrc != L'0' && (nLength % 4) == 1)) {
					if (! pvbuf->bAdd (&_wstrKanjiDigitsType5 [*pSrc - L'0'], 1))
						return	FALSE ;
				} 
				if (*pSrc != L'0'){
					if (nLength % 4 == 2){
						if (! pvbuf->bAdd (L"E", 1))
							return	FALSE ;
					} else if (nLength % 4 == 3){
						if (! pvbuf->bAdd (L"S", 1))
							return	FALSE ;
					} else if (nLength % 4 == 0){
						if (! pvbuf->bAdd (L"", 1))
							return	FALSE ;
					}
					fSilent	= FALSE ;
				}
				/* ʂB*/
				if (!fSilent || *pSrc != L'0'){
					int		nDigitNumber	= nLength - 1 ;

					if (nDigitNumber % 4 == 0){
						nDigitNumber	= nDigitNumber / 4 - 1 ;
						if (nDigitNumber >= 0){
							LPCWSTR	wstrDigit	= _wstrNumericUnitType5 [nDigitNumber] ;

							if (! pvbuf->bAdd (wstrDigit, wcslen (wstrDigit)))
								return	FALSE ;
							fSilent	= TRUE ;
						}
					}
				}
				nLength	-- ;
				pSrc	++ ;
			}
		}
	}
	/* NUL NUL ŏI邱ƁB*/
	return	pvbuf->bAdd (L"\0", 1) ;
}

BOOL
_bNumType4Kanji (
	LPCWSTR						strNum,
	int							nNumLen,
	CVarbuffer < WCHAR, 16 >*	pvbuf)
{
	CTSearchSession*	pSession ;
	LPCWSTR				wstrResult ;
	BOOL				bRetval ;

	/*	#4 ϊ̒ #4 ϊ܂܂ĂāAloop ɂȂȂ悤
	 *	numeric-conversion ͒~B
	 */
	pSession	= new CTHenkanSession (strNum, nNumLen, SEARCH_OKURI_NASHI, FALSE, FALSE) ;
	if (pSession == NULL)
		return	FALSE ;

	bRetval	= TRUE ;
	do {
		/* ŏ Candidate ͂Ȃ GetCandidate ł́H */
		wstrResult	= pSession->pGetCandidate () ;
		if (wstrResult == NULL)
			break ;
		if (! pvbuf->bAdd (wstrResult, wcslen (wstrResult)) || ! pvbuf->bAdd (L"\0", 1)) {
			bRetval	= FALSE ;
			goto	exit_func ;
		}
	} while (pSession->bNextCandidate ()) ;

	if (! pvbuf->bAdd (L"\0", 1))
		bRetval	= FALSE ;

exit_func:
	delete	pSession ;
	return	bRetval ;
}

BOOL
_bNumType9Shogi (
	LPCWSTR						strNum,
	int							nNumLen,
	CVarbuffer < WCHAR, 16 >*	pvbuf)
{
	/*  mapping table  customize l邩ȂBӁB*/
	static LPCWSTR	strMapTable1	= L"OPQRSTUVWX" ;
	static LPCWSTR	strMapTable2	= L"ZOlܘZ" ;

	if (nNumLen != 2 || ! _bIntNump (strNum, nNumLen))
		return	FALSE ;

	if ((L'0' <= strNum [0] && strNum [0] <= L'9') && 
	    (L'0' <= strNum [1] && strNum [1] <= L'9')) {
		WCHAR	wbuf [3] ;
		wbuf [0]	= strMapTable1 [strNum [0] - L'0'] ;
		wbuf [1]	= strMapTable2 [strNum [1] - L'0'] ;
		wbuf [2]	= L'\0' ;
		if (! pvbuf->bAdd (wbuf, 3))
			return	FALSE ;
	} else {
		return	FALSE ;
	}
	return	TRUE ;
}

BOOL
_bIntNump (
	LPCWSTR		strNum,
	int			nNumLen)
{
	while (nNumLen > 0) {
		if (! (L'0' <= *strNum && *strNum <= L'9'))
			return	FALSE ;
		strNum	++ ;
		nNumLen	-- ;
	}
	return	TRUE ;
}

BOOL
_bIntOrFloatNump (
	LPCWSTR		strNum,
	int			nNumLen)
{
	while (nNumLen > 0) {
		if (! (L'0' <= *strNum && *strNum <= L'9') && *strNum != L'.')
			return	FALSE ;
		strNum	++ ;
		nNumLen	-- ;
	}
	return	TRUE ;
}

/*========================================================================
 *	艼Ȃ̒ʏϊ Session
 */
CTHenkanSession::CTHenkanSession (
	LPCWSTR				wstrKeyword,
	int					nKeyword,
	int					iSearchType,
	BOOL				bNumericConversion,
	BOOL				bNumericFloat) :
	CTSearchSession ()
{
	BOOL					bCurrentlyNumericConverting	= FALSE ;

	/*	IWĩL[[hLB(lϊ̗LɊւ炸c)
	 */
	nKeyword		= (nKeyword > MYARRAYSIZE (m_rszNoConvKeyword))? MYARRAYSIZE (m_rszNoConvKeyword) : nKeyword ;
	memcpy (m_rszNoConvKeyword, wstrKeyword, nKeyword * sizeof (WCHAR)) ;
	m_nwstrNoConvKeyword	= nKeyword ;

	if (bNumericConversion) {
		m_nwstrKeyword	= iSkkNumComputeHenkanKey (wstrKeyword, nKeyword, m_rszKeyword, MYARRAYSIZE (m_rszKeyword), m_rszNumericList, MYARRAYSIZE (m_rszNumericList), bNumericFloat) ;
		m_bNumericFloat	= bNumericFloat ;

		/* ۂɐlϊsĂ邩ǂ AutoOkuriProcess ɉe^B*/
		if (m_nwstrKeyword != nKeyword || wcsncmp (wstrKeyword, m_rszKeyword, nKeyword) != 0) {
			bCurrentlyNumericConverting	= TRUE ;

			if (iSearchType == SEARCH_OKURI_AUTO)
				iSearchType	= SEARCH_OKURI_NASHI ;
		}
	} else {
		memcpy (m_rszKeyword, m_rszNoConvKeyword, m_nwstrNoConvKeyword * sizeof (WCHAR)) ;
		m_nwstrKeyword	= m_nwstrNoConvKeyword ;
		m_bNumericFloat	= FALSE ;
	}
	m_bNumericConversion	= bCurrentlyNumericConverting ;

	iSearchType			= (iSearchType != SEARCH_OKURI_AUTO && iSearchType != SEARCH_OKURI_NASHI)? SEARCH_OKURI_NASHI : iSearchType ;
	m_iSearchType		= iSearchType ;
	return ;
}

BOOL
CTHenkanSession::_bNumericp () const 
{
	return	m_bNumericConversion && m_rszNumericList [0] != L'\0' ;
}

BOOL
CTHenkanSession::_bSearch ()
{
	BOOL	bRetval ;

	if (m_bNumericConversion) {
		int	nSearchCount ;

		/*	lϊȂ[hŌB̎Am_nSearchCount XVĂ܂̂ŁA
		 *	oĂB
		 */
		m_bNumericConversion	= FALSE ;
		nSearchCount			= m_nSearchCount ;
		/*	{͎̑ŐlėǂǂƂƂɂȂ̂cB
		 */
		bRetval	= _bSearch (m_rszNoConvKeyword, m_nwstrNoConvKeyword) ;
		m_bNumericConversion	= TRUE ;
		m_nSearchCount			= nSearchCount ;
		if (! bRetval)
			return	FALSE ;
	}
	return	_bSearch (m_rszKeyword, m_nwstrKeyword) ;
}

BOOL
CTHenkanSession::_bSearch (
	LPCWSTR		pwKeyword,
	int			nwKeyword)
{
	CTPacket	packet ;
	HANDLE		hPipe ;
	BOOL		fResult	= FALSE ;
	WORD		woTotalResult ;
	int			nMajor, nMinor, nSize, nRecv ;
	LPCWSTR		strResult ;
	int			nResult ;
	int			nNext	= -1 ;
	
	hPipe	= _hOpenServer () ;
	if (hPipe == INVALID_HANDLE_VALUE)
		return	FALSE ;
	
	if (! packet.bSetHeader (SKKISERV_PROTO_SEARCH_EX, m_iSearchType)	||
		! packet.bAddCard32 (m_nSearchCount)							||
		! packet.bAddString (pwKeyword, nwKeyword)						||
		! packet.bAddString (NULL, 0)									||
		! packet.bSetLength ()											||
		! _bSend (hPipe, &packet))
		goto	exit_func ;
	
	DEBUGPRINTF ((TEXT ("_bSearch (): 1\n"))) ;
	if (! _bRecv (hPipe, &packet) ||
		! packet.bGetHeader (&nMajor, &nMinor, &nSize))
		goto	exit_func ;
	DEBUGPRINTF ((TEXT ("_Search (Major:%d, Minor:%d): 2\n"), nMajor, nMinor)) ;
	nRecv	= packet.iGetDataSize () ;
	if (nMajor != SKKISERV_PROTO_SEARCH_REPLY)
		goto	exit_func ;
	if (! packet.bGetCard16 (SKKISERV_PROTO_HEADER_SIZE, &woTotalResult))
		goto	exit_func ;
	
	if ((int)(woTotalResult * sizeof (WCHAR) + 2 + SKKISERV_PROTO_HEADER_SIZE) > nSize)
		goto	exit_func ;
	
	strResult	= (LPCWSTR)(packet.pGetData () + SKKISERV_PROTO_HEADER_SIZE + 2) ;
	nResult		= (nRecv - SKKISERV_PROTO_HEADER_SIZE - 2) / sizeof (WCHAR) ;
	nResult		= (nResult > woTotalResult)? woTotalResult : nResult ;

#if defined (DEBUG) || defined (_DEBUG)
	{
		WCHAR	buf [512], buf2 [512] ;
		int		n ;

		n		= (nResult > MYARRAYSIZE (buf)-1)? (MYARRAYSIZE(buf)-1) : nResult ;
		memcpy (buf, strResult, sizeof (WCHAR) * n) ;
		buf [n]	= L'\0' ;
		n		= wnsprintfW (buf2, MYARRAYSIZE (buf2) - 1, L"_bSearch (count:%d, result:\"%s\", len:%d, pos:%d)\n", m_nSearchCount, buf, nResult, m_nParsePosition) ;
		buf2 [n]	= L'\0' ;
		OutputDebugStringW (buf2) ;
	}
#endif
	/*	m_nParsePosition  #4 ϊ̎ɑ삳ĂB\AdvB
	 */
	if (woTotalResult > 0) {
		int	nPosition	= m_vbufCandidate.iGetUsage () ;
		if (! m_vbufCandidate.bAdd (strResult, nResult))
			goto	exit_func ;
		m_nParsePosition	= _bUpdateCandidateList (nPosition) ;
	}
	DEBUGPRINTF ((TEXT ("_Search (pos=%d, top:%p, cur:%p): 6\n"), m_nParsePosition, m_pTopCandidate, m_pCurCandidate)) ;
	fResult		= TRUE ;
	nNext		= (signed char)nMinor ;
	if (0 <= nNext && nNext <= m_nSearchCount)
		nNext	= -1 ;

	/* Œǉꂽ̐lWJsB*/
	if (m_bNumericConversion) 
		_bMultipleNumConvert () ;
	
  exit_func:
	/* session ̏Íc͈񌟍Ȃ̂ŁAŏIB*/
	m_nSearchCount	= nNext ;
	DisconnectNamedPipe (hPipe) ;
	CloseHandle (hPipe) ;
	return	fResult ;
}

#if 0
BOOL
CTHenkanSession::_bSearch ()
{
	CTPacket			packet ;
	HANDLE		hPipe ;
	BOOL				fResult	= FALSE ;
	WORD				woTotalResult ;
	int					nMajor, nMinor, nSize, nRecv ;
	LPCWSTR	strResult ;
	int		nResult ;
	int		nNext	= -1 ;
	
	hPipe	= _hOpenServer () ;
	if (hPipe == INVALID_HANDLE_VALUE)
		return	FALSE ;
	
	if (! packet.bSetHeader (SKKISERV_PROTO_SEARCH_EX, m_iSearchType)	||
		! packet.bAddCard32 (m_nSearchCount)							||
		! packet.bAddString (m_rszKeyword, m_nwstrKeyword)				||
		! packet.bAddString (NULL, 0)									||
		! packet.bSetLength ()											||
		! _bSend (hPipe, &packet))
		goto	exit_func ;
	
	DEBUGPRINTF ((TEXT ("_bSearch (): 1\n"))) ;
	if (! _bRecv (hPipe, &packet) ||
		! packet.bGetHeader (&nMajor, &nMinor, &nSize))
		goto	exit_func ;
	DEBUGPRINTF ((TEXT ("_Search (Major:%d, Minor:%d): 2\n"), nMajor, nMinor)) ;
	nRecv	= packet.iGetDataSize () ;
	if (nMajor != SKKISERV_PROTO_SEARCH_REPLY)
		goto	exit_func ;
	if (! packet.bGetCard16 (SKKISERV_PROTO_HEADER_SIZE, &woTotalResult))
		goto	exit_func ;
	
	if ((int)(woTotalResult * sizeof (WCHAR) + 2 + SKKISERV_PROTO_HEADER_SIZE) > nSize)
		goto	exit_func ;
	
	strResult	= (LPCWSTR)(packet.pGetData () + SKKISERV_PROTO_HEADER_SIZE + 2) ;
	nResult		= (nRecv - SKKISERV_PROTO_HEADER_SIZE - 2) / sizeof (WCHAR) ;
	nResult		= (nResult > woTotalResult)? woTotalResult : nResult ;

#if defined (DEBUG) || defined (_DEBUG)
	{
		WCHAR	buf [512], buf2 [512] ;
		int		n ;

		n		= (nResult > MYARRAYSIZE (buf)-1)? nResult : (MYARRAYSIZE(buf)-1) ;
		memcpy (buf, strResult, sizeof (WCHAR) * n) ;
		buf [n]	= L'\0' ;
		n		= wnsprintfW (buf2, MYARRAYSIZE (buf2) - 1, L"_bSearch (count:%d, result:\"%s\", len:%d, pos:%d)\n", m_nSearchCount, buf, nResult, m_nParsePosition) ;
		buf2 [n]	= L'\0' ;
		OutputDebugStringW (buf2) ;
	}
#endif
	/*	m_nParsePosition  #4 ϊ̎ɑ삳ĂB\AdvB
	 */
	if (woTotalResult > 0) {
		int	nPosition	= m_vbufCandidate.iGetUsage () ;
		if (! m_vbufCandidate.bAdd (strResult, nResult))
			goto	exit_func ;
		m_nParsePosition	= _bUpdateCandidateList (nPosition) ;
	}
	DEBUGPRINTF ((TEXT ("_Search (pos=%d, top:%p, cur:%p): 6\n"), m_nParsePosition, m_pTopCandidate, m_pCurCandidate)) ;
	fResult		= TRUE ;
	nNext		= (signed char)nMinor ;
	if (0 <= nNext && nNext <= m_nSearchCount)
		nNext	= -1 ;

	/* Œǉꂽ̐lWJsB*/
	if (m_bNumericConversion) 
		_bMultipleNumConvert () ;
	
  exit_func:
	/* session ̏Íc͈񌟍Ȃ̂ŁAŏIB*/
	m_nSearchCount	= nNext ;
	DisconnectNamedPipe (hPipe) ;
	CloseHandle (hPipe) ;
	return	fResult ;
}
#endif

/*===============================================================
 */
CTOkuriHenkanSession::CTOkuriHenkanSession (
	LPCWSTR			wstrKeyword,
	int				nKeywordLen,
	LPCWSTR			wstrOkuriChar,
	int				nOkuriCharLen,
	int				iSearchType,
	BOOL			bNumericConversion,
	BOOL			bNumericFloat) :
	CTSearchSession (wstrKeyword, nKeywordLen)
{
	if (bNumericConversion) {
		m_nwstrKeyword		= iSkkNumComputeHenkanKey (wstrKeyword, nKeywordLen, m_rszKeyword, MYARRAYSIZE (m_rszKeyword), m_rszNumericList, MYARRAYSIZE (m_rszNumericList), bNumericFloat) ;
		m_bNumericFloat	= bNumericFloat ;
	} else {
		int			nKeyword ;
		nKeyword		= (nKeywordLen > MYARRAYSIZE (m_rszKeyword))? MYARRAYSIZE (m_rszKeyword) : nKeywordLen ;
		memcpy (m_rszKeyword, wstrKeyword, nKeyword * sizeof (WCHAR)) ;
		m_nwstrKeyword	= nKeyword ;
		m_bNumericFloat	= FALSE ;
	}
	m_bNumericConversion	= bNumericConversion ;

	if (nOkuriCharLen > 0) {
		nOkuriCharLen	= (nOkuriCharLen > MYARRAYSIZE (m_rszOkuriChar))? MYARRAYSIZE (m_rszOkuriChar) : nOkuriCharLen ;
		memcpy (m_rszOkuriChar, wstrOkuriChar, nOkuriCharLen * sizeof (WCHAR)) ;
		m_nOkuriCharLen	= nOkuriCharLen ;
		m_iSearchType	= (iSearchType != SEARCH_OKURI_NASHI && iSearchType != SEARCH_OKURI_AUTO)? iSearchType : SEARCH_OKURI_ARI ;
	} else {
		m_nOkuriCharLen	= 0 ;
		m_iSearchType	= SEARCH_OKURI_ARI ;
	}
	return ;
}

BOOL
CTOkuriHenkanSession::_bSearch ()
{
	CTPacket			packet ;
	HANDLE		hPipe ;
	BOOL				fResult	= FALSE ;
	WORD				woTotalResult ;
	int					nMajor, nMinor, nSize, nRecv ;
	LPCWSTR	strResult ;
	int		nResult ;
	int		nNext	= -1 ;
	
	hPipe	= _hOpenServer () ;
	if (hPipe == INVALID_HANDLE_VALUE)
		return	FALSE ;
	
	if (! packet.bSetHeader	(SKKISERV_PROTO_SEARCH_EX, m_iSearchType)	||
		! packet.bAddCard32	(m_nSearchCount)							||
		! packet.bAddString	(m_rszKeyword, m_nwstrKeyword)				||
		! packet.bAddString	(m_rszOkuriChar, m_nOkuriCharLen)			||
		! packet.bSetLength ()											||
		! _bSend (hPipe, &packet))
		goto	exit_func ;
	
	if (! _bRecv (hPipe, &packet) ||
		! packet.bGetHeader (&nMajor, &nMinor, &nSize))
		goto	exit_func ;
	DEBUGPRINTF ((TEXT ("_Search (Major:%d, Minor:%d): 2\n"), nMajor, nMinor)) ;
	nRecv	= packet.iGetDataSize () ;
	if (nMajor != SKKISERV_PROTO_SEARCH_REPLY)
		goto	exit_func ;
	if (! packet.bGetCard16 (SKKISERV_PROTO_HEADER_SIZE, &woTotalResult))
		goto	exit_func ;
	
	if ((int)(woTotalResult * sizeof (WCHAR) + 2 + SKKISERV_PROTO_HEADER_SIZE) > nSize)
		goto	exit_func ;
	
	strResult	=  (LPCWSTR)(packet.pGetData () + SKKISERV_PROTO_HEADER_SIZE + 2) ;
	nResult		= (nRecv - SKKISERV_PROTO_HEADER_SIZE - 2) / sizeof (WCHAR) ;
	nResult		= (nResult > woTotalResult)? woTotalResult : nResult ;
	if (woTotalResult > 0) {
		int	nPosition	= m_vbufCandidate.iGetUsage () ;
		if (! m_vbufCandidate.bAdd (strResult, nResult))
			goto	exit_func ;
		m_nParsePosition	= _bUpdateCandidateList (nPosition) ;
	}
	fResult		= TRUE ;
	nNext		= (signed char)nMinor ;
	if (0 <= nNext && nNext <= m_nSearchCount)
		nNext	= -1 ;
	
  exit_func:
	/* session ̏Íc͈񌟍Ȃ̂ŁAŏIB*/
	m_nSearchCount	= nNext ;
	DisconnectNamedPipe (hPipe) ;
	CloseHandle (hPipe) ;
	return	fResult ;
}


/*===============================================================
 */
CTCompletionSession::CTCompletionSession (
	LPCWSTR			wstrKeyword,
	int				nKeywordLen) :
	CTSearchSession (wstrKeyword, nKeywordLen)
{
	return ;
}

BOOL
CTCompletionSession::_bSearch ()
{
	CTPacket			packet ;
	HANDLE		hPipe ;
	BOOL				fResult	= FALSE ;
	WORD				woTotalResult ;
	int					nMajor, nMinor, nSize, nRecv ;
	LPCWSTR	strResult ;
	int		nResult ;
	
	hPipe	= _hOpenServer () ;
	if (hPipe == INVALID_HANDLE_VALUE)
		return	FALSE ;
	
	if (! packet.bSetHeader (SKKISERV_PROTO_COMPLETION, 1) ||
		! packet.bAddString (m_rszKeyword, m_nwstrKeyword) ||
		! packet.bSetLength () ||
		! _bSend (hPipe, &packet))
		goto	exit_func ;
	
	if (! _bRecv (hPipe, &packet) ||
		! packet.bGetHeader (&nMajor, &nMinor, &nSize))
		goto	exit_func ;
	nRecv	= packet.iGetDataSize () ;
	if (nMajor != SKKISERV_PROTO_COMPLETION_REPLY)
		goto	exit_func ;
	if (! packet.bGetCard16 (SKKISERV_PROTO_HEADER_SIZE, &woTotalResult))
		goto	exit_func ;
	
	if ((int)(woTotalResult * sizeof (WCHAR) + 2 + SKKISERV_PROTO_HEADER_SIZE) > nSize)
		goto	exit_func ;
	
	strResult	= (LPCWSTR)(packet.pGetData () + SKKISERV_PROTO_HEADER_SIZE + 2) ;
	nResult		= (nRecv - SKKISERV_PROTO_HEADER_SIZE - 2) / sizeof (WCHAR) ;
	nResult		= (nResult > woTotalResult)? woTotalResult : nResult ;
	if (woTotalResult > 0) {
		int	nPosition	= m_vbufCandidate.iGetUsage () ;
		if (! m_vbufCandidate.bAdd (strResult, nResult))
			goto	exit_func ;
		m_nParsePosition	= _bUpdateCandidateList (nPosition) ;
	}
	fResult			= TRUE ;
	
  exit_func:
	/* session ̏Íc͈񌟍Ȃ̂ŁAŏIB*/
	m_nSearchCount	= -1 ;
	DisconnectNamedPipe (hPipe) ;
	CloseHandle (hPipe) ;
	return	fResult ;
}

BOOL
CTCompletionSession::_bIsSeparateChar (
	WCHAR		wch) const
{
	return	(wch == L' ') ;
}

/*===============================================================
 */
CTKakuteiHistoryCompletionSession::CTKakuteiHistoryCompletionSession () :
	CTSearchSession ()
{
	return ;
}

BOOL
CTKakuteiHistoryCompletionSession::_bSearch ()
{
	CTPacket			packet ;
	HANDLE				hPipe ;
	BOOL				fResult	= FALSE ;
	WORD				woTotalResult ;
	int					nMajor, nMinor, nSize, nRecv ;
	LPCWSTR				strResult ;
	int					nResult ;

	hPipe	= _hOpenServer () ;
	if (hPipe == INVALID_HANDLE_VALUE)
		return	FALSE ;
	
	if (! packet.bSetHeader (SKKISERV_PROTO_GETKAKUTEIHISTORY, 1) ||
		! packet.bSetLength () ||
		! _bSend (hPipe, &packet))
		goto	exit_func ;

	if (! _bRecv (hPipe, &packet) ||
		! packet.bGetHeader (&nMajor, &nMinor, &nSize))
		goto	exit_func ;

	nRecv	= packet.iGetDataSize () ;
	if (nMajor != SKKISERV_PROTO_GETKAKUTEIHISTORY_REPLY)
		goto	exit_func ;
	if (! packet.bGetCard16 (SKKISERV_PROTO_HEADER_SIZE, &woTotalResult))
		goto	exit_func ;
	
	if ((int)(woTotalResult * sizeof (WCHAR) + 2 + SKKISERV_PROTO_HEADER_SIZE) > nSize)
		goto	exit_func ;
	
	strResult	= (LPCWSTR)(packet.pGetData () + SKKISERV_PROTO_HEADER_SIZE + 2) ;
	nResult		= (nRecv - SKKISERV_PROTO_HEADER_SIZE - 2) / sizeof (WCHAR) ;
	nResult		= (nResult > woTotalResult)? woTotalResult : nResult ;
	if (woTotalResult > 0) {
		int	nPosition	= m_vbufCandidate.iGetUsage () ;
		if (! m_vbufCandidate.bAdd (strResult, nResult))
			goto	exit_func ;
		m_nParsePosition	= _bUpdateCandidateList (nPosition) ;
	}
	fResult			= TRUE ;
	
  exit_func:
	/* session ̏Íc͈񌟍Ȃ̂ŁAŏIB*/
	m_nSearchCount	= -1 ;
	DisconnectNamedPipe (hPipe) ;
	CloseHandle (hPipe) ;
	return	fResult ;
}

BOOL
CTKakuteiHistoryCompletionSession::_bIsSeparateChar (
	WCHAR						wch) const
{
	return	(wch == L' ') ;
}

