/*	langbar.cpp
 *
 *	Language bar ui code
 */
#include "globals.h"
#include "skimic.h"
#include "resource.h"
#include "skkimmgr.h"
#include "IME\ImeDoc.h"

#define LANGBAR_ITEM_DESC	L"̓[h" // max 32 chars! ށAő32B

typedef struct {
	const WCHAR*	pchDesc ;
	int				nMode ;
	void			(*pfnHandler)(CSkkImeTextService* _this) ;
}	TSFLBMENUINFO ;

static	const TSFLBMENUINFO	c_rgMenuItems []	= {
	{ L"Ђ炪",			IMECMODE_HIRAGANA,		CSkkImeTextService::_Menu_ToHiragana },
	{ L"SpJ^Ji",		IMECMODE_KATAKANA,		CSkkImeTextService::_Menu_ToZenkata },
	{ L"Spp",			IMECMODE_ZENKAKU,		CSkkImeTextService::_Menu_ToZenei },
	{ L"pJ^Ji",		IMECMODE_HANKANA,		CSkkImeTextService::_Menu_ToJisx0201 },
	{ L"pp(pJi)",IMECMODE_HANKANA_ROMAN,	CSkkImeTextService::_Menu_ToJisx0201Roman },
	{ L"pp",			IMECMODE_ASCII,			CSkkImeTextService::_Menu_ToHanei },
	{ L"ړ",			IMECMODE_OFF,			CSkkImeTextService::_Menu_ToDirect },
	{ NULL,					-1,						NULL },
	{ L"LZ",		-1,						NULL }
} ;

/*	IME ̓̓[h{^(o[) GUID
 */
static	const GUID	c_guidImeItemButtonCMode	= {
	0x5d97b50b, 0x1b86, 0x4577, { 0xaa, 0xf6, 0xbd, 0x76, 0xe5, 0xb, 0x18, 0xec }
} ;

#define	SKKIME_LANGBARITEMSINK_COOKIE	0x0fab0fab

enum {
	MENU_ITEM_INDEX_CANCEL		= -1,
	MENU_ITEM_INDEX_HIRAGANA	= 0,
	MENU_ITEM_INDEX_KATAKANA,
	MENU_ITEM_INDEX_ZENEI,
	MENU_ITEM_INDEX_HANKANA,
	MENU_ITEM_INDEX_HANKANA_ROMAN,
	MENU_ITEM_INDEX_ASCII,
	MENU_ITEM_INDEX_DIRECT
} ;

/*	CLangBarItemCModeButton: Conversion Mode Language Bar Item Button
 */
class	CLangBarItemCModeButton : public ITfLangBarItemButton,
								  public ITfSource
{
public:
	CLangBarItemCModeButton (CSkkImeTextService*	pSkkIme) ;
	~CLangBarItemCModeButton () ;

	// IUnknown
	STDMETHODIMP	QueryInterface (REFIID riid, void **ppvObj) ;
	STDMETHODIMP_(ULONG)	AddRef (void) ;
	STDMETHODIMP_(ULONG)	Release (void) ;
	
	// ITfLangBarItem
	STDMETHODIMP	GetInfo (TF_LANGBARITEMINFO *pInfo) ;
	STDMETHODIMP	GetStatus (DWORD *pdwStatus) ;
	STDMETHODIMP	Show (BOOL fShow) ;
	STDMETHODIMP	GetTooltipString (BSTR *pbstrToolTip) ;
	
	// ITfLangBarItemButton
	STDMETHODIMP	OnClick (TfLBIClick click, POINT pt, const RECT *prcArea) ;
	STDMETHODIMP	InitMenu (ITfMenu *pMenu) ;
	STDMETHODIMP	OnMenuSelect (UINT wID) ;
	STDMETHODIMP	GetIcon (HICON *phIcon) ;
	STDMETHODIMP	GetText (BSTR *pbstrText) ;
	
	// ITfSource
	STDMETHODIMP	AdviseSink(REFIID riid, IUnknown *punk, DWORD *pdwCookie);
	STDMETHODIMP	UnadviseSink(DWORD dwCookie);

	STDMETHODIMP	Update () ;

private:
	CSkkImeTextService*		m_pSkkIme ;
	ITfLangBarItemSink*		_pLangBarItemSink ;
	TF_LANGBARITEMINFO		_tfLangBarItemInfo ;
	LONG					_cRef ;
} ;

/*
 */
CLangBarItemCModeButton::CLangBarItemCModeButton (
	CSkkImeTextService*		pSkkIme)
{
	DllAddRef () ;

	_tfLangBarItemInfo.clsidService	= c_clsidSkkImeTextService ;
	_tfLangBarItemInfo.guidItem		= c_guidConversionModeItemButton ;
	_tfLangBarItemInfo.dwStyle		= TF_LBI_STYLE_BTN_MENU | TF_LBI_STYLE_TEXTCOLORICON | TF_LBI_STYLE_SHOWNINTRAY ;
	_tfLangBarItemInfo.ulSort		= 2 ;
	SafeStringCopy (_tfLangBarItemInfo.szDescription, MYARRAYSIZE (_tfLangBarItemInfo.szDescription), LANGBAR_ITEM_DESC) ;

	m_pSkkIme			= pSkkIme ;
	m_pSkkIme->AddRef () ;
	_pLangBarItemSink	= NULL ;
	_cRef				= 1 ;
	return ;
}

CLangBarItemCModeButton::~CLangBarItemCModeButton ()
{
	DllRelease () ;
	m_pSkkIme->Release () ;
	return ;
}

STDAPI
CLangBarItemCModeButton::QueryInterface (
	REFIID			riid,
	void**			ppvObj)
{
	if (ppvObj == NULL)
		return	E_INVALIDARG ;

	*ppvObj	= NULL ;
	if (IsEqualIID (riid, IID_IUnknown) ||
		IsEqualIID (riid, IID_ITfLangBarItem) ||
		IsEqualIID (riid, IID_ITfLangBarItemButton)) {
		*ppvObj	= (ITfLangBarItemButton *)this ;
	} else if (IsEqualIID (riid, IID_ITfSource)) {
		*ppvObj	= (ITfSource *)this ;
	}
	if (*ppvObj != NULL) {
		AddRef () ;
		return	S_OK ;
	}
	return	E_NOINTERFACE ;
}

STDAPI_(ULONG)
CLangBarItemCModeButton::AddRef ()
{
	return	++ _cRef ;
}

STDAPI_(ULONG)
CLangBarItemCModeButton::Release ()
{
	LONG	cr	= -- _cRef ;

	if (_cRef == 0) {
		delete	this ;
	}
	return	cr ;
}

STDAPI
CLangBarItemCModeButton::GetInfo (
	TF_LANGBARITEMINFO*		pInfo)
{
	if (pInfo == NULL)
		return	E_INVALIDARG ;

	*pInfo	= _tfLangBarItemInfo ;
	return	S_OK ;
}

STDAPI
CLangBarItemCModeButton::Show (
	BOOL					fShow)
{
    return	E_NOTIMPL ;
	UNREFERENCED_PARAMETER (fShow) ;
}

STDAPI
CLangBarItemCModeButton::GetStatus (
	DWORD*					pdwStatus)
{
	if (pdwStatus == NULL)
		return	E_INVALIDARG ;

	DEBUGPRINTF ((TEXT ("CLangBarItemCModeButton::GetStatus ()\n"))) ;
	*pdwStatus	= (m_pSkkIme->_IsKeyboardDisabled ())? TF_LBI_STATUS_DISABLED : 0 ;
	return	S_OK ;
}

/*	Button  tooltip ԂBԂl SysAllocString ɂ
 *	mۂ̈ɏKvB SysFreeString 
 *	̂́AĂяo̐ӔCłB
 */
STDAPI
CLangBarItemCModeButton::GetTooltipString (
	BSTR*					pbstrToolTip)
{
	if (pbstrToolTip == NULL)
		return	E_INVALIDARG ;

	*pbstrToolTip	= SysAllocString (LANGBAR_ITEM_DESC) ;
	return	(*pbstrToolTip == NULL)? E_OUTOFMEMORY : S_OK ;
}

/*	ITfLangBarItemButton::OnClick
 *
 *	 method ̓[Uo[ TF_LBI_STYLE_BTN_BUTTON ܂
 *	 TF_LBI_STYLE_BTN_TOGGLE X^CĂ{^̏Ń}
 *	EXNbNɌĂяoB
 *	{^ item  TF_LBI_STYLE_BTN_BUTTON X^CȂ
 *	̂ȂA method gȂB
 *(*)
 *	̏󋵂ł͓ɉKv͂Ȃ̂ŁAS_OK 𑦕ԂB
 */
STDAPI
CLangBarItemCModeButton::OnClick (
	TfLBIClick				click,
	POINT					pt,
	const RECT*				prcArea)
{
	return	S_OK ;
	UNREFERENCED_PARAMETER (click) ;
	UNREFERENCED_PARAMETER (pt) ;
	UNREFERENCED_PARAMETER (prcArea) ;
}

/*	ITfLangBarItemButton::InitMenu
 *
 *	 method  TF_LBI_STYLE_BTN_MENU X^Co[̃{^
 *	o[{^ɑ΂ĕ\ menu item ǉėLɂ邽
 *	ɌĂяoB
 */
STDAPI
CLangBarItemCModeButton::InitMenu (
	ITfMenu*				pMenu)
{
	register CSkkImeMgr*	pImeMgr ;
	register CImeDoc*		pDoc ;
	register int		i, nCMode ;
	register DWORD		dwFlag ;
	register LPCWSTR	wstrDesc ;
	register ULONG		nstrDesc ;

	if (pMenu == NULL)
		return	E_INVALIDARG ;

	pImeMgr	= m_pSkkIme->_GetCurrentIME () ;
	if (pImeMgr == NULL)
		return	E_FAIL ;

	if (! m_pSkkIme->_IsKeyboardOpen ()) {
		nCMode	= IMECMODE_OFF ;
	} else {
		pDoc	= pImeMgr->GetDocument () ;
		if (pDoc == NULL)
			return	E_FAIL ;

		nCMode	= pDoc->iGetConversionMode () ;
	}
	for (i = 0 ; i < MYARRAYSIZE (c_rgMenuItems) ; i ++) {
		wstrDesc		= c_rgMenuItems [i].pchDesc ;
		if (wstrDesc != NULL) {
			nstrDesc	= wcslen (wstrDesc) ;
			dwFlag		= (c_rgMenuItems [i].nMode == nCMode)? TF_LBMENUF_CHECKED : 0 ;
		} else {
			nstrDesc	= 0 ;
			dwFlag		= TF_LBMENUF_SEPARATOR ;
		}
		pMenu->AddMenuItem (i, dwFlag, NULL, NULL, wstrDesc, nstrDesc, NULL) ;
	}
	return	S_OK ;
}

STDAPI
CLangBarItemCModeButton::OnMenuSelect (
	UINT					wID)
{
	if (wID >= MYARRAYSIZE (c_rgMenuItems))
		return	E_FAIL ;

	/*	NULL ̏ꍇ Cancel ƎvƂɂB*/
	if (c_rgMenuItems [wID].pfnHandler != NULL) {
		c_rgMenuItems [wID].pfnHandler (m_pSkkIme) ;
	    if (_pLangBarItemSink != NULL)
			_pLangBarItemSink->OnUpdate (TF_LBI_ICON) ;
	}
	return	S_OK ;
}

STDAPI
CLangBarItemCModeButton::GetIcon (
	HICON*					phIcon)
{
	CSkkImeMgr*	pImeMgr ;
	CImeDoc*	pDoc ;
	LPCTSTR		str ;

	DEBUGPRINTF ((TEXT ("CLangBarItemCModeButton::GetIcon ()\n"))) ;

	if (phIcon == NULL)
		return	E_INVALIDARG ;

	pImeMgr	= m_pSkkIme->_GetCurrentIME () ;
	if (pImeMgr == NULL)
		return	E_FAIL ;

	if (m_pSkkIme->_IsKeyboardDisabled ()) {
		str	= TEXT ("IDI_CMODE_HIRAGANA") ;
	} else if (! m_pSkkIme->_IsKeyboardOpen ()) {
		str	= TEXT ("IDI_CMODE_DIRECT") ;
	} else {
		pDoc	= pImeMgr->GetDocument () ;
		if (pDoc == NULL)
			return	E_FAIL ;

		switch (pDoc->iGetConversionMode ()) {
		case	IMECMODE_ZENKAKU:
			str	= TEXT ("IDI_CMODE_ZENEI") ;
			break ;
		case	IMECMODE_ASCII:
			str	= TEXT ("IDI_CMODE_HANEI") ;
			break ;
		case	IMECMODE_HANKANA:
			str	= TEXT ("IDI_CMODE_HANKATA") ;
			break ;
		case	IMECMODE_HANKANA_ROMAN:
			str	= TEXT ("IDI_CMODE_HANROMA") ;
			break ;
		case	IMECMODE_KATAKANA:
			str	= TEXT ("IDI_CMODE_ZENKATA") ;
			break ;
		case	IMECMODE_HIRAGANA:
			str	= TEXT ("IDI_CMODE_HIRAGANA") ;
			break ;
		default:
			/*	ړ͔͂̑Y݂ȁB*/
			str	= TEXT ("IDI_CMODE_DIRECT") ;
			break ;
		}
	}
    *phIcon	= (HICON)LoadImage (g_hInst, str, IMAGE_ICON, 16, 16, LR_SHARED) ;
    return (*phIcon != NULL) ? S_OK : E_FAIL ;
}

STDAPI
CLangBarItemCModeButton::GetText (
	BSTR*			pbstrText)
{
	if (pbstrText == NULL)
		return	E_INVALIDARG ;

	*pbstrText	= SysAllocString (LANGBAR_ITEM_DESC) ;
	return	(*pbstrText == NULL)? E_OUTOFMEMORY : S_OK ;
}

STDAPI
CLangBarItemCModeButton::AdviseSink (
	REFIID			riid,
	IUnknown*		punk,
	DWORD*			pdwCookie)
{
    if (!IsEqualIID (IID_ITfLangBarItemSink, riid))
        return	CONNECT_E_CANNOTCONNECT ;

    if (_pLangBarItemSink != NULL)
        return	CONNECT_E_ADVISELIMIT ;

    if (punk->QueryInterface (IID_ITfLangBarItemSink, (void **)&_pLangBarItemSink) != S_OK) {
        _pLangBarItemSink	= NULL ;
        return	E_NOINTERFACE ;
    }

    *pdwCookie	= SKKIME_LANGBARITEMSINK_COOKIE ;
    return	S_OK ;
}

STDAPI
CLangBarItemCModeButton::UnadviseSink (
	DWORD			dwCookie)
{
    if (dwCookie != SKKIME_LANGBARITEMSINK_COOKIE)
        return	CONNECT_E_NOCONNECTION ;

    if (_pLangBarItemSink == NULL)
        return	CONNECT_E_NOCONNECTION ;

    _pLangBarItemSink->Release () ;
    _pLangBarItemSink	= NULL ;

    return	S_OK ;
}

STDAPI
CLangBarItemCModeButton::Update ()
{
    if (_pLangBarItemSink == NULL)
        return	CONNECT_E_NOCONNECTION ;

	_pLangBarItemSink->OnUpdate (TF_LBI_ICON | TF_LBI_STATUS) ;
	return	S_OK ;
}

/*========================================================================*
 *	public function interface
 */
BOOL
CSkkImeTextService::_InitCModeLangBarItem ()
{
	ITfLangBarItemMgr*	pLangBarItemMgr	= NULL ;
	ITfLangBarItem*		pItem			= NULL ;
	BOOL				bRet ;
	HRESULT				hr ;

	if (! _IsShowLangBarItem (REGKEY_SHOWINPUTMODEICON)) 
		return	TRUE ;
	if (m_pThreadMgr == NULL)
		return	FALSE ;
	if (m_pThreadMgr->QueryInterface (IID_ITfLangBarItemMgr, (void**)&pLangBarItemMgr) != S_OK) 
		return	FALSE ;

	/*	IME ̃{^폜B(IE  IME  TSF ̗Ă΂Ă܂̂ŁAIME ̃{^
	 *	o[ɎcĂ܂Ƃ邽)
	 */
	pItem	= NULL ;
	if (SUCCEEDED (pLangBarItemMgr->GetItem (c_guidImeItemButtonCMode, &pItem)) && pItem != NULL) {
		DEBUGPRINTF ((TEXT ("[tsf] RemoveItem: IME-ButtonCMode\n"))) ;
		pLangBarItemMgr->RemoveItem (pItem) ;
		pItem->Release () ;
	}

	/* TSF ̃{^ǉB*/
	bRet			= FALSE ;
	m_pCModeLangBarItem	= new CLangBarItemCModeButton (this) ;
	if (m_pCModeLangBarItem == NULL)
		goto	Exit ;

	hr	= pLangBarItemMgr->AddItem (m_pCModeLangBarItem) ;
	if (FAILED (hr)) 
		goto	Exit ;

	DEBUGPRINTF ((TEXT ("[tsf] AddItem: TSF-ButtonCMode\n"))) ;
	bRet	= TRUE ;
Exit:
	pLangBarItemMgr->Release () ;
	return	bRet ;
}

void
CSkkImeTextService::_UninitCModeLangBarItem ()
{
	ITfLangBarItemMgr*	pLangBarItemMgr ;

	if (m_pCModeLangBarItem == NULL)
		return ;
	if (m_pThreadMgr != NULL) {
		if (m_pThreadMgr->QueryInterface (IID_ITfLangBarItemMgr, (void**)&pLangBarItemMgr) == S_OK) {
			DEBUGPRINTF ((TEXT ("[tsf] RemoveItem: IME-ButtonCMode\n"))) ;
			pLangBarItemMgr->RemoveItem (m_pCModeLangBarItem) ;
			pLangBarItemMgr->Release () ;
		}
	}
	m_pCModeLangBarItem->Release () ;
	m_pCModeLangBarItem	= NULL ;
	return ;
}

void
CSkkImeTextService::_UpdateCModeLangBarItem ()
{
	ITfLangBarItemMgr*	pLangBarItemMgr ;

	if (m_pThreadMgr == NULL)
		return ;
	if (m_pThreadMgr->QueryInterface (IID_ITfLangBarItemMgr, (void**)&pLangBarItemMgr) != S_OK)
		return ;
	if (_IsShowLangBarItem (REGKEY_SHOWINPUTMODEICON) && m_pCModeLangBarItem != NULL) 
		m_pCModeLangBarItem->Update () ;
	DEBUGPRINTF ((TEXT ("[tsf] UpdateItem: IME-ButtonCMode\n"))) ;
	pLangBarItemMgr->Release () ;
	return ;
}




