/*	langbar.cpp
 *
 *	Language bar ui code
 */
#include "globals.h"
#include "skimic.h"
#include "resource.h"

#define LANGBAR_ITEM_DESC	L"͕"	// max 32 chars! ށAő32B
#define	SKKIME_LANGBARITEMSINK_COOKIE	0x0fab0fad

typedef struct {
	const WCHAR*	pchDesc ;
	DWORD			(*pfnGetFlag)(CSkkImeTextService* _this) ;
	void			(*pfnHandler)(CSkkImeTextService* _this) ;
}	TSFLBMENUINFO ;

static	const TSFLBMENUINFO	c_rgMenuItems []	= {
	{ L"wv(&H)",
	  NULL,
	  CSkkImeTextService::_Menu_Help },
	{ L"vpeB(&R)",
	  NULL,
	  CSkkImeTextService::_Menu_Property },
	{ L"ĕϊ(&C)",
	  CSkkImeTextService::_Menu_GetNormalFlag,
	  CSkkImeTextService::_Menu_Reconversion },
	{ NULL,
	  NULL,
	  NULL },
	{ L"L[{[h̕\/\",
	  CSkkImeTextService::_Menu_GetToggleKeyboardFlag,
	  CSkkImeTextService::_Menu_ToggleShowKeyboard },
	{ NULL,
	  NULL,
	  NULL },
	{ L"LZ",
	  NULL,	
	  NULL }
} ;

/*	IME IME{^(o[) GUID
 */
static	const GUID	c_guidImeItemButtonIME	= {
	0xf0e6f1f4, 0xb3aa, 0x4ef6, { 0x96, 0xd7, 0x33, 0x53, 0xf5, 0x80, 0x37, 0x30 }
} ;

enum {
	MENU_ITEM_INDEX_CANCEL		= -1,
	MENU_ITEM_INDEX_HELP,
	MENU_ITEM_INDEX_PROPERTY,
	MENU_ITEM_INDEX_RECONVRESION,
	MENU_ITEM_INDEX_SHOWKEYBOARD,
} ;

/*	c{^̐ class Kvƍl̂ǂB
 *	ォǉȂƂȂȁB
 */
class	CLangBarItemIMEButton : public ITfLangBarItemButton,
									public ITfSource
{
public:
	CLangBarItemIMEButton (CSkkImeTextService* pSkkIme) ;
	~CLangBarItemIMEButton () ;

	// IUnknown
	STDMETHODIMP QueryInterface (REFIID riid, void **ppvObj) ;
	STDMETHODIMP_(ULONG) AddRef (void) ;
	STDMETHODIMP_(ULONG) Release (void) ;
	
	// ITfLangBarItem
	STDMETHODIMP GetInfo (TF_LANGBARITEMINFO *pInfo) ;
	STDMETHODIMP GetStatus (DWORD *pdwStatus) ;
	STDMETHODIMP Show (BOOL fShow) ;
	STDMETHODIMP GetTooltipString (BSTR *pbstrToolTip) ;
	
	// ITfLangBarItemButton
	STDMETHODIMP OnClick (TfLBIClick click, POINT pt, const RECT *prcArea) ;
	STDMETHODIMP InitMenu (ITfMenu *pMenu) ;
	STDMETHODIMP OnMenuSelect (UINT wID) ;
	STDMETHODIMP GetIcon (HICON *phIcon) ;
	STDMETHODIMP GetText (BSTR *pbstrText) ;
	
	// ITfSource
	STDMETHODIMP AdviseSink(REFIID riid, IUnknown *punk, DWORD *pdwCookie);
	STDMETHODIMP UnadviseSink(DWORD dwCookie);

	STDMETHODIMP	Update () ;

private:
	CSkkImeTextService*		m_pSkkIme ;
	ITfLangBarItemSink*		_pLangBarItemSink ;
	TF_LANGBARITEMINFO		_tfLangBarItemInfo ;
	LONG					_cRef ;
} ;

CLangBarItemIMEButton::CLangBarItemIMEButton (
	CSkkImeTextService*		pSkkIme)
{
	DllAddRef () ;

	_tfLangBarItemInfo.clsidService	= c_clsidSkkImeTextService ;
	_tfLangBarItemInfo.guidItem		= c_guidToolItemButton ;
	_tfLangBarItemInfo.dwStyle		= TF_LBI_STYLE_BTN_MENU | TF_LBI_STYLE_SHOWNINTRAY ;
	_tfLangBarItemInfo.ulSort		= 1 ;
	SafeStringCopy (_tfLangBarItemInfo.szDescription, MYARRAYSIZE (_tfLangBarItemInfo.szDescription), LANGBAR_ITEM_DESC) ;

	m_pSkkIme			= pSkkIme ;
	m_pSkkIme->AddRef () ;
	_pLangBarItemSink	= NULL ;
	_cRef				= 1 ;
	return ;
}

CLangBarItemIMEButton::~CLangBarItemIMEButton ()
{
	DllRelease () ;
	m_pSkkIme->Release () ;
	return ;
}

STDAPI
CLangBarItemIMEButton::QueryInterface (
	REFIID			riid,
	void**			ppvObj)
{
	if (ppvObj == NULL)
		return	E_INVALIDARG ;

	*ppvObj	= NULL ;
	if (IsEqualIID (riid, IID_IUnknown) ||
		IsEqualIID (riid, IID_ITfLangBarItem) ||
		IsEqualIID (riid, IID_ITfLangBarItemButton)) {
		*ppvObj	= (ITfLangBarItemButton *)this ;
	} else if (IsEqualIID (riid, IID_ITfSource)) {
		*ppvObj	= (ITfSource *)this ;
	}
	if (*ppvObj != NULL) {
		AddRef () ;
		return	S_OK ;
	}
	return	E_NOINTERFACE ;
}

STDAPI_(ULONG)
CLangBarItemIMEButton::AddRef ()
{
	return	++ _cRef ;
}

STDAPI_(ULONG)
CLangBarItemIMEButton::Release ()
{
	LONG	cr	= -- _cRef ;

	if (_cRef == 0) {
		delete	this ;
	}
	return	cr ;
}

STDAPI
CLangBarItemIMEButton::GetInfo (
	TF_LANGBARITEMINFO*		pInfo)
{
	if (pInfo == NULL)
		return	E_INVALIDARG ;

	*pInfo	= _tfLangBarItemInfo ;
	return	S_OK ;
}

STDAPI
CLangBarItemIMEButton::Show (
	BOOL					fShow)
{
    return	E_NOTIMPL ;
	UNREFERENCED_PARAMETER (fShow) ;
}

STDAPI
CLangBarItemIMEButton::GetStatus (
	DWORD*					pdwStatus)
{
	if (pdwStatus == NULL)
		return	E_INVALIDARG ;

	*pdwStatus	= (m_pSkkIme->_IsKeyboardDisabled ())? TF_LBI_STATUS_DISABLED : 0 ;
	return	S_OK ;
}

/*	Button  tooltip ԂBԂl SysAllocString ɂ
 *	mۂ̈ɏKvB SysFreeString 
 *	̂́AĂяo̐ӔCłB
 */
STDAPI
CLangBarItemIMEButton::GetTooltipString (
	BSTR*					pbstrToolTip)
{
	if (pbstrToolTip == NULL)
		return	E_INVALIDARG ;

	*pbstrToolTip	= SysAllocString (LANGBAR_ITEM_DESC) ;
	return	(*pbstrToolTip == NULL)? E_OUTOFMEMORY : S_OK ;
}

/*	ITfLangBarItemButton::OnClick
 *
 *	 method ̓[Uo[ TF_LBI_STYLE_BTN_BUTTON ܂
 *	 TF_LBI_STYLE_BTN_TOGGLE X^CĂ{^̏Ń}
 *	EXNbNɌĂяoB
 *	{^ item  TF_LBI_STYLE_BTN_BUTTON X^CȂ
 *	̂ȂA method gȂB
 *(*)
 *	̏󋵂ł͓ɉKv͂Ȃ̂ŁAS_OK 𑦕ԂB
 */
STDAPI
CLangBarItemIMEButton::OnClick (
	TfLBIClick				click,
	POINT					pt,
	const RECT*				prcArea)
{
	return	S_OK ;
	UNREFERENCED_PARAMETER (click) ;
	UNREFERENCED_PARAMETER (pt) ;
	UNREFERENCED_PARAMETER (prcArea) ;
}

/*	ITfLangBarItemButton::InitMenu
 *
 *	 method  TF_LBI_STYLE_BTN_MENU X^Co[̃{^
 *	o[{^ɑ΂ĕ\ menu item ǉėLɂ邽
 *	ɌĂяoB
 */
STDAPI
CLangBarItemIMEButton::InitMenu (
	ITfMenu*				pMenu)
{
	register int		i ;
	register DWORD		dwFlag ;
	register LPCWSTR	wstrDesc ;
	register ULONG		nstrDesc ;

	if (pMenu == NULL)
		return	E_INVALIDARG ;

	for (i = 0 ; i < MYARRAYSIZE (c_rgMenuItems) ; i ++) {
		wstrDesc		= c_rgMenuItems [i].pchDesc ;
		if (wstrDesc != NULL) {
			nstrDesc	= wcslen (wstrDesc) ;
			dwFlag		= (c_rgMenuItems [i].pfnGetFlag != NULL)? (c_rgMenuItems [i].pfnGetFlag)(m_pSkkIme) : 0 ;
		} else {
			nstrDesc	= 0 ;
			dwFlag		= TF_LBMENUF_SEPARATOR ;
		}
		pMenu->AddMenuItem (i, dwFlag, NULL, NULL, wstrDesc, nstrDesc, NULL) ;
	}
	return	S_OK ;
}

STDAPI
CLangBarItemIMEButton::OnMenuSelect (
	UINT					wID)
{
	if (wID >= MYARRAYSIZE (c_rgMenuItems))
		return	E_FAIL ;

	/*	NULL ̏ꍇ Cancel ƎvƂɂB*/
	if (c_rgMenuItems [wID].pfnHandler != NULL) {
		c_rgMenuItems [wID].pfnHandler (m_pSkkIme) ;
//		m_pSkkIme->_UpdateLangBarItem (c_guidToolItemButton) ;
	}
	return	S_OK ;
}

STDAPI
CLangBarItemIMEButton::GetIcon (
	HICON*					phIcon)
{
	if (phIcon == NULL)
		return	E_INVALIDARG ;

    *phIcon	= (HICON)LoadImage (g_hInst, TEXT ("IDI_SKKIME"), IMAGE_ICON, 16, 16, LR_SHARED) ;
    return (*phIcon != NULL) ? S_OK : E_FAIL ;
}

STDAPI
CLangBarItemIMEButton::GetText (
	BSTR*			pbstrText)
{
	if (pbstrText == NULL)
		return	E_INVALIDARG ;

	*pbstrText	= SysAllocString (LANGBAR_ITEM_DESC) ;
	return	(*pbstrText == NULL)? E_OUTOFMEMORY : S_OK ;
}

STDAPI
CLangBarItemIMEButton::AdviseSink (
	REFIID			riid,
	IUnknown*		punk,
	DWORD*			pdwCookie)
{
    if (!IsEqualIID (IID_ITfLangBarItemSink, riid))
        return	CONNECT_E_CANNOTCONNECT ;

    if (_pLangBarItemSink != NULL)
        return	CONNECT_E_ADVISELIMIT ;

    if (punk->QueryInterface (IID_ITfLangBarItemSink, (void **)&_pLangBarItemSink) != S_OK) {
        _pLangBarItemSink	= NULL ;
        return	E_NOINTERFACE ;
    }

    *pdwCookie	= SKKIME_LANGBARITEMSINK_COOKIE ;
    return	S_OK ;
}

STDAPI
CLangBarItemIMEButton::UnadviseSink (
	DWORD			dwCookie)
{
    if (dwCookie != SKKIME_LANGBARITEMSINK_COOKIE)
        return	CONNECT_E_NOCONNECTION ;

    if (_pLangBarItemSink == NULL)
        return	CONNECT_E_NOCONNECTION ;

    _pLangBarItemSink->Release () ;
    _pLangBarItemSink	= NULL ;

    return	S_OK ;
}

STDAPI
CLangBarItemIMEButton::Update ()
{
    if (_pLangBarItemSink == NULL)
        return	CONNECT_E_NOCONNECTION ;

	_pLangBarItemSink->OnUpdate (TF_LBI_STATUS) ;
	return	S_OK ;
}

/*========================================================================*
 *	public function interface
 */
BOOL
CSkkImeTextService::_InitIMELangBarItem ()
{
	ITfLangBarItemMgr*	pLangBarItemMgr	= NULL ;
	ITfLangBarItem*		pItem			= NULL ;
	BOOL				bRet ;
	HRESULT				hr ;

	if (! _IsShowLangBarItem (REGKEY_SHOWIMEICON)) 
		return	TRUE ;

	if (m_pThreadMgr == NULL)
		return	FALSE ;

	if (m_pThreadMgr->QueryInterface (IID_ITfLangBarItemMgr, (void**)&pLangBarItemMgr) != S_OK) 
		return	FALSE ;

	/*	IME ̃{^폜B(IE  IME  TSF ̗Ă΂Ă܂̂ŁAIME ̃{^
	 *	o[ɎcĂ܂Ƃ邽)
	 *
	 *	Vista ł́Ã{^͕svȂ悤BKeyboard {^̃ACR`ύX
	 *	߂Ă悤B> MS-IME
	 */
	pItem	= NULL ;
	if (SUCCEEDED (pLangBarItemMgr->GetItem (c_guidImeItemButtonIME, &pItem)) && pItem != NULL) {
		DEBUGPRINTF ((TEXT ("[tsf] RemoveItem: IME-ButtonIME\n"))) ;
		pLangBarItemMgr->RemoveItem (pItem) ;
		pItem->Release () ;
	}

	bRet			= FALSE ;
	m_pIMELangBarItem	= new CLangBarItemIMEButton (this) ;
	if (m_pIMELangBarItem == NULL)
		goto	Exit ;

	/*	ǂʂ process ŒǉĂ Button ܂ĂȂƂ炵B
	 *	̂ AddItem G[Ԃ\̂ŁA RemoveItem Ă
	 *	ǉB
	 */
	hr	= pLangBarItemMgr->AddItem (m_pIMELangBarItem) ;
	if (FAILED (hr)) 
		goto	Exit ;
	DEBUGPRINTF ((TEXT ("[tsf] AddItem: TSF-ButtonIME\n"))) ;
	bRet	= TRUE ;

Exit:
	pLangBarItemMgr->Release () ;
	return	bRet ;
}

void
CSkkImeTextService::_UninitIMELangBarItem ()
{
	ITfLangBarItemMgr*	pLangBarItemMgr ;
	
	if (m_pIMELangBarItem == NULL)
		return ;
	
	if (m_pThreadMgr != NULL) {
		if (m_pThreadMgr->QueryInterface (IID_ITfLangBarItemMgr, (void**)&pLangBarItemMgr) == S_OK) {
			pLangBarItemMgr->RemoveItem (m_pIMELangBarItem) ;
			pLangBarItemMgr->Release () ;
		}
	}
	m_pIMELangBarItem->Release () ;
	m_pIMELangBarItem	= NULL ;
	return ;
}

void
CSkkImeTextService::_UpdateIMELangBarItem ()
{
	ITfLangBarItemMgr*	pLangBarItemMgr ;

	if (m_pThreadMgr == NULL)
		return ;
	if (m_pThreadMgr->QueryInterface (IID_ITfLangBarItemMgr, (void**)&pLangBarItemMgr) != S_OK)
		return ;

	if (_IsShowLangBarItem (REGKEY_SHOWIMEICON) && m_pIMELangBarItem != NULL) 
		m_pIMELangBarItem->Update () ;
	pLangBarItemMgr->Release () ;
	return ;
}

