//
// globals.h
//
// Global variable declarations.
//
#pragma once

#include <windows.h>
#include <shlwapi.h>
#include <wchar.h>
#include <ole2.h>
#include <olectl.h>
#include <assert.h>
#include "msctf.h"
#include "ctffunc.h"

/*	function prototypes
 */
LONG	DllAddRef () ;
LONG	DllRelease () ;

void	InsertTextAtSelection	(TfEditCookie ec, ITfContext *pContext, const WCHAR *pchText, ULONG cchText) ;
BOOL	AdviseSink				(IUnknown *pSource, IUnknown *pSink, REFIID riid, DWORD *pdwCookie) ;
void	UnadviseSink			(IUnknown *pSource, DWORD *pdwCookie) ;
BOOL	AdviseSingleSink		(TfClientId tfClientId, IUnknown *pSource, IUnknown *pSink, REFIID riid) ;
void	UnadviseSingleSink		(TfClientId tfClientId, IUnknown *pSource, REFIID riid) ;
BOOL	IsRangeCovered			(TfEditCookie ec, ITfRange *pRangeTest, ITfRange *pRangeCover) ;
BOOL	GetGUIDAtomFromGUID		(REFGUID refguid, DWORD* pdw) ;

void	DebugPrintf	(LPCTSTR strFormat,...) ;
void	DebugPrintfW(LPCWSTR strFormat,...) ;
void	vDebugPrintfToFile (LPCTSTR strFormat, ...) ;

#undef	DEBUGLOGFILE_ENABLE
#define	DEBUGFILENAME		TEXT("skimic.log")

#if !defined (MYARRAYSIZE)
#define	MYARRAYSIZE(myarray)		(sizeof(myarray)/sizeof(myarray[0]))
#endif
#if !defined (ASSERT)
#if defined (DBG) || defined (DEBUG)
#define	ASSERT(x)	assert(x)
#else
#define	ASSERT(x)	/*(x)*/
#endif
#endif
#if defined (NOALIGNOF)
#define	__alignof(sometype)		my_alignof(sizeof(sometype))

__inline	int	my_alignof (int iSize) {
	/* 4byte alignment? */
	switch (iSize) {
		case	1:
		case	2:
			return	iSize ;
		case	3:
		case	4:
			return	4 ;
		case	5:
		case	6:
		case	7:
		case	8:
			return	8 ;
		default:
			return	16 ;	/* ... */
	}
}
#endif

/*	ނށALANGIDc SKKIME  LANGID ƈv
 *	ƂKvȂ̂H
 *
 *       +-----------------------+-------------------------+
 *       |     Sublanguage ID    |   Primary Language ID   |
 *       +-----------------------+-------------------------+
 *        15                   10 9                       0   bit
 */
#define	SKKIME_LANGID			MAKELANGID(LANG_JAPANESE, SUBLANG_DEFAULT)
#define	SKKIME_KEYBOARDLAYOUT	TEXT("E0110411")	

#define	SKKIME_DESC				TEXT("SKKIME ver. 1.5")
#define	SKKIME_DESC_W			L"SKKIME ver. 1.5"
#define	SKKIME_DESC_A			"SKKIME ver. 1.5"
#define	SKKIME_MODEL			TEXT("Apartment")

//#define LANGBAR_ITEM_DESC	L"̓[h" // max 32 chars! ށAő32B

#define SKKIME_ICON_INDEX  8

#define	REGKEY_SHOWKEYBRDICON		L"ShowKeyboardIcon"
#define	REGKEY_SHOWIMEICON			L"ShowInputMethodIcon"
#define	REGKEY_SHOWINPUTMODEICON	L"ShowInputModeIcon"

#define SafeRelease(punk)       \
{                               \
    if ((punk) != NULL)         \
    {                           \
        (punk)->Release();      \
    }                           \
}                   

#define SafeReleaseClear(punk)  \
{                               \
    if ((punk) != NULL)         \
    {                           \
        (punk)->Release();      \
        (punk) = NULL;          \
    }                           \
}                   

#if defined (DEBUG) || defined (DBG) || defined (_DEBUG)
#define	DEBUGPRINTF(text)	DebugPrintf##text
#define	DEBUGPRINTFW(text)	DebugPrintfW##text
#else
#define	DEBUGPRINTF(text)	/*text*/
#define	DEBUGPRINTFW(text)	/*text*/
#endif

#if !defined (MIN)
#define	MIN(lvalue,rvalue)	(((lvalue) < (rvalue))? (lvalue) : (rvalue))
#endif
#if !defined (MAX)
#define	MAX(lvalue,rvalue)	(((lvalue) > (rvalue))? (lvalue) : (rvalue))
#endif

//+---------------------------------------------------------------------------
//
// SafeStringCopy
//
// Copies a string from one buffer to another.  wcsncpy does not always
// null-terminate the destination buffer; this function does.
//----------------------------------------------------------------------------

inline void SafeStringCopy (WCHAR *pchDst, ULONG cchMax, const WCHAR *pchSrc)
{
    if (cchMax > 0) {
//#if !defined(WIN64) && defined (_MSC_VER) && (_MSC_VER >= 1400) 
#if __STDC_WANT_SECURE_LIB__
		wcsncpy_s(pchDst, cchMax, pchSrc, _TRUNCATE) ;
#else
		wcsncpy(pchDst, pchSrc, cchMax) ;
#endif
		pchDst[cchMax-1] = '\0';
	}
}

/*========================================================================
 *	global variables
 */
extern HINSTANCE		g_hInst ;
extern LONG				g_cRefDll ;
extern CRITICAL_SECTION	g_cs ;
extern const CLSID		c_clsidSkkImeTextService ;
extern const GUID		c_guidSkkImeProfile ;
extern const GUID		c_guidConversionModeItemButton ;
extern const GUID		c_guidToolItemButton ;
extern const GUID		c_guidSkkImeDisplayAttribute ;
extern const GUID		c_guidSkkImeDisplayAttributeInput ;
extern const GUID		c_guidSkkImeDisplayAttributeTargetConverted	;
extern const GUID		c_guidSkkImePreservedKeyToggleIME ;

