#include "globals.h"
#include "skimic.h"
#include "skkimmgr.h"
#include "editsess.h"

#if !defined (MAXCOMPLEN)
#define	MAXCOMPLEN	256
#endif

class CTReconvCandidateString ;
class CTReconvCandidateList ;
class CTReconvEnumTfCandidates ;

class CTReconvCandidateString : public ITfCandidateString {
private:
	WCHAR			_wszText [MAXCOMPLEN] ;
	int				_nIndex ;

    LONG			_cRef ;     // COM ref count

public:
	CTReconvCandidateString (LPCWSTR wstrText, int nIndex) {
		lstrcpynW (_wszText, wstrText, MAXCOMPLEN) ;
		_nIndex	= nIndex ;
		_cRef	= 1 ;
		return ;
	}

	virtual	~CTReconvCandidateString ()	{
	}

	STDMETHODIMP	QueryInterface (REFIID riid, void** ppvObj) {
		if (ppvObj == NULL)
			return	E_INVALIDARG ;

		*ppvObj	= NULL ;
		if (IsEqualIID (riid, IID_IUnknown) ||
			IsEqualIID (riid, IID_ITfCandidateString)) {
			*ppvObj	= (ITfCandidateString *)this ;
		}
		if (*ppvObj) {
			AddRef () ;
			return	S_OK ;
		}
		return	E_NOINTERFACE ;
	}

    STDMETHODIMP_(ULONG) AddRef(void) {
        return	++ _cRef ;
    }

    STDMETHODIMP_(ULONG) Release(void) {
        LONG cr = --_cRef ;
		
        assert(_cRef >= 0) ;
		
        if (_cRef == 0) {
            delete	this ;
        }
        return cr ;
    }

	STDMETHODIMP	GetString (BSTR* pbstr) {
		if (pbstr == NULL)
			return	E_INVALIDARG ;
		*pbstr	= SysAllocString (_wszText) ;
		return	(*pbstr != NULL)? S_OK : E_OUTOFMEMORY ;
	}
	STDMETHODIMP	GetIndex (ULONG* pnIndex) {
		if (pnIndex == NULL)
			return	E_INVALIDARG ;
		*pnIndex	= (ULONG) _nIndex ;
		return	S_OK ;
	}
} ;

class CTReconvEnumTfCandidates : public IEnumTfCandidates {
private:
	ITfCandidateList*		_pCandList ;
	ULONG					_nPos ;
    LONG					_cRef ;     // COM ref count

public:
	CTReconvEnumTfCandidates (ITfCandidateList* pCandList) {
		assert (pCandList != NULL) ;
		_pCandList	= pCandList ;
		_pCandList->AddRef () ;
		_nPos		= 0 ;
		_cRef		= 1 ;
		return ;
	}

	~CTReconvEnumTfCandidates () {
		if (_pCandList != NULL)
			_pCandList->Release () ;
		_pCandList	= NULL ;
	}

	STDMETHODIMP	QueryInterface (REFIID riid, void** ppvObj) {
		if (ppvObj == NULL)
			return	E_INVALIDARG ;

		*ppvObj	= NULL ;
		if (IsEqualIID (riid, IID_IUnknown) ||
			IsEqualIID (riid, IID_IEnumTfCandidates)) {
			*ppvObj	= (IEnumTfCandidates *)this ;
		}
		if (*ppvObj) {
			AddRef () ;
			return	S_OK ;
		}
		return	E_NOINTERFACE ;
	}

	STDMETHODIMP	Clone (IEnumTfCandidates** ppEnum) {
		CTReconvEnumTfCandidates*	pEnum ;
		
		if (ppEnum == NULL)
			return	S_OK ;	/* ? */

		pEnum	= new CTReconvEnumTfCandidates (_pCandList) ;
		if (pEnum == NULL)
			return	E_OUTOFMEMORY ;

		pEnum->_nPos	= _nPos ;
		*ppEnum			= pEnum ;
		return	S_OK ;
	}

	/*	݈ʒu  ulCount  Candidate mۂāAmۂłʂ
	 *	pcFetched ɓĕԂB
	 */
	STDMETHODIMP	Next (ULONG ulCount, ITfCandidateString** ppCand, ULONG* pcFetched) {
		ULONG	nCnt, nFetched, n ;
		HRESULT	hr ;

		if (ulCount > 0 && ppCand == NULL)
			return	E_INVALIDARG ;

		if (FAILED (_pCandList->GetCandidateNum (&nCnt)))
			return	S_FALSE ;

		nFetched	= ((_nPos + ulCount) > nCnt)? (nCnt - _nPos) : ulCount ;
		if (pcFetched != NULL)
			*pcFetched	= nFetched ;
		for (n = 0 ; n < nFetched ; n ++) {
			hr	= _pCandList->GetCandidate (_nPos + n, ppCand + n) ;
			if (FAILED (hr))
				return	hr ;
		}
		_nPos	+= nFetched ;	/* ͕Kvȓ삩H */
		return	S_OK ;
	}

	STDMETHODIMP	Reset () {
		_nPos	= 0 ;
		return	S_OK ;
	}

	STDMETHODIMP	Skip (ULONG ulCount) {
		ULONG	nNewPos	= _nPos + ulCount ;
		ULONG	nCnt ;

		if (FAILED (_pCandList->GetCandidateNum (&nCnt)))
			return	S_FALSE ;
		if (nNewPos >= nCnt) {
			_nPos	= nCnt - 1 ;
			return	S_FALSE ;
		}
		_nPos	= nNewPos ;
		return	S_OK ;
	}

    STDMETHODIMP_(ULONG) AddRef(void) {
        return	++ _cRef ;
    }

    STDMETHODIMP_(ULONG) Release(void) {
        LONG cr = --_cRef ;
		
        assert(_cRef >= 0) ;
		
        if (_cRef == 0) {
            delete	this ;
        }
        return cr ;
    }
} ;

/*	tϊ̃Xg\z邱Ƃ݂łȂ̂ŁA^ꂽ range
 *	̂܂ܓ邱ƂɂȂB
 */
class CTReconvCandidateList : public ITfCandidateList {
private:
	WCHAR				_wszCandidate [MAXCOMPLEN] ;
    LONG				_cRef ;     // COM ref count
	ULONG				_nResult ;
	TfCandidateResult	_imcr ;

public:
	CTReconvCandidateList (LPCWSTR wstrTEXT) {
		lstrcpynW (_wszCandidate, wstrTEXT, MYARRAYSIZE (_wszCandidate)) ;
		_nResult	= (ULONG) -1 ;
		_imcr		= CAND_CANCELED ;
		_cRef		= 1 ;
	}

	virtual			~CTReconvCandidateList () {
	}

	STDMETHODIMP	QueryInterface (REFIID riid, void** ppvObj) {
		if (ppvObj == NULL)
			return	E_INVALIDARG ;

		*ppvObj	= NULL ;
		if (IsEqualIID (riid, IID_IUnknown) ||
			IsEqualIID (riid, IID_ITfCandidateList)) {
			*ppvObj	= (ITfCandidateList*) this ;
		}
		if (*ppvObj) {
			AddRef () ;
			return	S_OK ;
		}
		return	E_NOINTERFACE ;
	}

	STDMETHODIMP	EnumCandidates (IEnumTfCandidates** ppEnum) {
		if (ppEnum == NULL)
			return	E_INVALIDARG ;
		*ppEnum	= new CTReconvEnumTfCandidates (this) ;
		return	(*ppEnum != NULL)? S_OK : E_OUTOFMEMORY ;
	}

	STDMETHODIMP	GetCandidate (ULONG nIndex, ITfCandidateString** pCand) {
		if (nIndex != 0)
			return	E_INVALIDARG ;
		if (pCand == NULL)
			return	E_INVALIDARG ;
		*pCand	= new CTReconvCandidateString (_wszCandidate, nIndex) ;
		return	(*pCand != NULL)? S_OK : E_OUTOFMEMORY ;
	}

	STDMETHODIMP	GetCandidateNum (ULONG* pnCnt) {
		if (pnCnt == NULL)
			return	E_INVALIDARG ;
		*pnCnt	= 1 ;
		return	S_OK ;
	}

	STDMETHODIMP	SetResult (ULONG nIndex, TfCandidateResult imcr) {
		if (imcr != CAND_CANCELED && nIndex != 0)
			return	E_FAIL ;
		_nResult	= nIndex ;
		_imcr		= imcr ;
		return	S_OK ;
	}

    STDMETHODIMP_(ULONG) AddRef(void) {
        return	++ _cRef ;
    }

    STDMETHODIMP_(ULONG) Release(void) {
        LONG cr = --_cRef ;
		
        assert(_cRef >= 0) ;
		
        if (_cRef == 0) {
            delete	this ;
        }
        return cr ;
    }

	ULONG	GetResult (TfCandidateResult* pResult) {
		if (pResult != NULL)
			*pResult	= _imcr ;
		return	_nResult ;
	}
} ;

class CQueryReconversionSession : public CEditSessionBase
{
public:
	CQueryReconversionSession (CSkkImeTextService* pTSF, ITfContext* pContext, ITfRange* pRange) :
		CEditSessionBase (pContext)
	{
		_pTSF	= pTSF ;
		_pTSF->AddRef () ;
		_pRange	= pRange ;
		_pRange->AddRef () ;
	}

	~CQueryReconversionSession ()
	{
		_pTSF->Release () ;
		_pTSF	= NULL ;
		_pRange->Release () ;
		_pRange	= NULL ;
	}
	
	STDMETHODIMP	DoEditSession (TfEditCookie ec) ;

	HRESULT	GetResult (ITfRange** ppNewRange)
	{
		return	_pRange->Clone (ppNewRange) ;
	}

private:
	CSkkImeTextService*	_pTSF ;
	ITfRange*			_pRange ;
} ;

class CGetReconversionEditSession : public CEditSessionBase
{
public:
	CGetReconversionEditSession (CSkkImeTextService* pTSF, ITfContext* pContext, ITfRange* pRange) :
		CEditSessionBase (pContext)
	{
		DEBUGPRINTF ((TEXT ("CGetReconversionEditSession::constructor\n"))) ;
		_pTSF	= pTSF ;
		_pTSF->AddRef () ;
		_pRange	= pRange ;
		_pRange->AddRef () ;

		_wszTEXT [0]	= L'\0' ;
		_nTEXT	= 0 ;
	}

	~CGetReconversionEditSession ()
	{
		DEBUGPRINTF ((TEXT ("CGetReconversionEditSession::destructor\n"))) ;
		_pTSF->Release () ;
		_pTSF	= NULL ;
		_pRange->Release () ;
		_pRange	= NULL ;
	}

	STDMETHODIMP	DoEditSession (TfEditCookie ec) ;

	CTReconvCandidateList*	GetCandidateList ()
	{
#if defined (DEBUG) || defined (DBG)
		OutputDebugStringW (L"CGetReconversionEditSession::GetCandidateList\n") ;
		OutputDebugStringW (_wszTEXT) ;
#endif
		return	new CTReconvCandidateList (_wszTEXT) ;
	}

private:
	CSkkImeTextService*	_pTSF ;
	ITfRange*			_pRange ;
	WCHAR				_wszTEXT [MAXCOMPLEN] ;
	int					_nTEXT ;
} ;

class CReconvertEditSession : public CEditSessionBase
{
public:
	CReconvertEditSession (CSkkImeTextService* pTSF, ITfContext* pContext, CTReconvCandidateList* pCandList) : 
		CEditSessionBase (pContext)
	{
		_pTSF		= pTSF ;
		_pTSF->AddRef () ;
		_pCandList	= pCandList ;
		_pCandList->AddRef () ;
	}

	~CReconvertEditSession ()
	{
		_pTSF->Release () ;
		_pTSF		= NULL ;
		_pCandList->Release () ;
		_pCandList	= NULL ;
	}

	STDMETHODIMP	DoEditSession (TfEditCookie ec) ;

private:
	CSkkImeTextService*		_pTSF ;
	CTReconvCandidateList*	_pCandList ;
} ;

STDAPI
CQueryReconversionSession::DoEditSession (
	register TfEditCookie		ec)
{
	BOOL	fEmpty ;
	WCHAR	wszBuffer [256] ;
	ULONG	cch ;

	if (FAILED (_pRange->IsEmpty (ec, &fEmpty)))
		return	E_FAIL ;
	if (fEmpty) {
		DEBUGPRINTF ((TEXT ("CQueryReconversionSession/Range is empty.\n"))) ;
	} else {
		if (FAILED (_pRange->GetText (ec, 0, wszBuffer, MYARRAYSIZE (wszBuffer) - 1, &cch))) {
			DEBUGPRINTF ((TEXT ("CQueryReconversionSession/GetText failed.\n"))) ;
		} else {
			wszBuffer [cch]	= L'\0' ;
			DEBUGPRINTF ((TEXT ("CQueryReconversionSession/Range is \"%s\"(%d)\n"), wszBuffer, cch)) ;
		}
	}
	return	S_OK ;
}

/*	Range ͈͓̔ɂ镶ȄsĂB
 */
STDAPI
CGetReconversionEditSession::DoEditSession (
	register TfEditCookie		ec)
{
	ULONG	cch	= 0 ;

	if (FAILED (_pRange->GetText (ec, 0, _wszTEXT, MAXCOMPLEN - 1, &cch))) {
		DEBUGPRINTF ((TEXT ("CGetReconversionEditSession::DoEditSession failed.\n"))) ;
		return	E_FAIL ;
	}
	_nTEXT	= cch ;
	_wszTEXT [cch]	= L'\0' ;
#if defined (DEBUG) || defined (DBG)
	DEBUGPRINTF ((TEXT ("CGetReconversionEditSession::DoEditSession\n"))) ;
	OutputDebugStringW (_wszTEXT) ;
#endif
	return	S_OK ;
}

/*	Reconvert 鎞̏B
 */
STDAPI
CReconvertEditSession::DoEditSession (
	register TfEditCookie		ec)
{
	CSkkImeMgr*			pIME	= _pTSF->_GetCurrentIME () ;
	ULONG				nResult ;
	TfCandidateResult	imcr ;
	ITfCandidateString*	pCandStr ;
	BSTR				bstrCand ;
	HRESULT				hr ;

	nResult	= _pCandList->GetResult (&imcr) ;
	DEBUGPRINTF ((TEXT ("CReconversionEditSession/Result is \"%d\"\n"), nResult)) ;
	if (nResult == (ULONG)-1) {
		/*	xIꂽƂ͂ȂƂƂA
		 *	 dialog box łoāAǂ̌₪ǂ
		 *	[Uɐq˂KvB
		 *
		 *	A͌1ȂāA͂ꂽ̂
		 *	܂܂Ȃ̂ŁÂ悤ɐݒ肷B
		 */
		nResult	= 0 ;
		imcr	= CAND_FINALIZED ;
	} else {
		/*	xȂIꂽƂB
		 */
		if (imcr != CAND_FINALIZED) {
			/*	ǁA[U̓LZƉ߂B
			 */
			hr	= S_OK ;
			goto	exit_func ;
		}
	}

	/*	IꂽҏW\ɖ߂B
	 */
	hr	= _pCandList->GetCandidate (nResult, &pCandStr) ;
	if (FAILED (hr))
		goto	exit_func ;
	if (SUCCEEDED (pCandStr->GetString (&bstrCand))) {
		OutputDebugStringW (bstrCand) ;
		hr	= pIME->OnReconvertSession (m_pContext, ec, bstrCand) ;
		SysFreeString (bstrCand) ;
	}
	pCandStr->Release () ;
  exit_func:
	return	hr ;
}

/*========================================================================*
 */
STDAPI
CSkkImeTextService::QueryRange (
	ITfRange*			pRange,
	ITfRange**			ppNewRange,
	BOOL*				pfConvertable)
{
	CQueryReconversionSession*	pQuerySession ;
	ITfContext*	pContext ;
	HRESULT		hr ;

	DEBUGPRINTF ((TEXT ("ITfReconvert::QueryRange()\n"))) ;

	/*	{͔͈͒Ȃ΂ȂȂcB
	 */
	if (pRange == NULL) {
		DEBUGPRINTF ((TEXT ("ITfReconvert::QueryRange(): pRange == NULL\n"))) ;
		return	E_INVALIDARG ;
	}
	if (FAILED (pRange->GetContext (&pContext)))
		return	E_INVALIDARG ;

	pQuerySession	= new CQueryReconversionSession (this, pContext, pRange) ;
	if (pQuerySession == NULL)
		return	E_OUTOFMEMORY ;

	pContext->RequestEditSession (m_tfClientId, pQuerySession, TF_ES_SYNC | TF_ES_READ, &hr) ;
	if (SUCCEEDED (hr)) {
		hr	= pQuerySession->GetResult (ppNewRange) ;
		if (FAILED (hr))
			goto	exit_func ;
	}
	pQuerySession->Release () ;
	if (pfConvertable != NULL)
		*pfConvertable	= TRUE ;
	hr	= S_OK ;
  exit_func:
	return	hr ;
}

STDAPI
CSkkImeTextService::GetReconversion (
	ITfRange*			pRange,
	ITfCandidateList**	ppCandList)
{
	CGetReconversionEditSession*	pEditSession ;
	ITfContext*	pContext ;
	HRESULT	hr ;

	DEBUGPRINTF ((TEXT ("ITfReconvert::GetReconversion()\n"))) ;

	if (pRange == NULL || ppCandList == NULL)
		return	E_INVALIDARG ;

	if (FAILED (pRange->GetContext (&pContext)))
		return	E_INVALIDARG ;

	pEditSession	= new CGetReconversionEditSession (this, pContext, pRange) ;
	if (pEditSession == NULL)
		return	E_OUTOFMEMORY ;

	if (m_pRCandidateList != NULL) {
		m_pRCandidateList->Release () ;
		m_pRCandidateList	= NULL ;
	}

	pContext->RequestEditSession (m_tfClientId, pEditSession, TF_ES_SYNC | TF_ES_READ, &hr) ;
	if (SUCCEEDED (hr)) {
		CTReconvCandidateList*	pCandList ;

		pCandList	= pEditSession->GetCandidateList () ;
		if (pCandList == NULL) {
			hr	= E_OUTOFMEMORY ;
			goto	exit_func ;
		}
		m_pRCandidateList	= pCandList ;
		m_pRCandidateList->AddRef () ;
		*ppCandList	= pCandList ;
	}

  exit_func:
	pEditSession->Release () ;
	return	hr ;
}

/*
 *	SetResult ĂȂ́A̒Ń_CAOoČÎ
 *	̂悤B
 *	ASetResult  CandidateList ̒ǔ₪ǂvƑI΂Ă
 *	΁A̗pȋCB͎ȂƂȂȁB
 *
 *	Query  GetCandidateList ꂸɁA Reconvert Ă΂\
 *	ȁB̏ꍇɂ͍lȂƂȂB
 */
STDAPI
CSkkImeTextService::Reconvert (
	ITfRange*			pRange)
{
	CReconvertEditSession*	pEditSession ;
	ITfContext*	pContext ;
	HRESULT		hr ;

	DEBUGPRINTF ((TEXT ("ITfReconvert::Reconvert()\n"))) ;

	if (pRange == NULL)
		return	E_INVALIDARG ;

	if (FAILED (pRange->GetContext (&pContext)))
		return	E_INVALIDARG ;

	/*	x GetCandidateList Ă΂ĂȂꍇɂ́A Range ɍݒ
	 *	Ă镶DB
	 */
	if (m_pRCandidateList == NULL) {
		ITfCandidateList*	pCandList	= NULL ;

		hr	= GetReconversion (pRange, &pCandList) ;
		if (FAILED (hr))
			return	hr ;
		pCandList->Release () ;
	}

	pEditSession	= new CReconvertEditSession (this, pContext, m_pRCandidateList) ;
	if (pEditSession == NULL)
		return	E_OUTOFMEMORY ;

	if (m_pRCandidateList != NULL) {
		m_pRCandidateList->Release () ;
		m_pRCandidateList	= NULL ;
	}
	pContext->RequestEditSession (m_tfClientId, pEditSession, TF_ES_ASYNCDONTCARE | TF_ES_READWRITE, &hr) ;
	pEditSession->Release () ;
	return	hr ;
}


/*========================================================================*
 */
BOOL
CSkkImeTextService::_InitReconversion ()
{
	m_pRCandidateList	= NULL ;
	return	TRUE ;
}

void
CSkkImeTextService::_UninitReconversion ()
{
	SafeReleaseClear (m_pRCandidateList) ;
	m_pRCandidateList	= NULL ;
	return ;
}

