#include "globals.h"
#include "skimic.h"
#include "skkimmgr.h"
#include "uistatus.h"
#include "IME\ImeDoc.h"
#include "IME\ImeConfig.h"
#include "CandListUI.h"
#include "ToolTipUI.h"
#include "editsess.h"

#define	DEFAULT_MINIBUF_CHARWIDTH	80

/*========================================================================
 *	prototypes
 */
HFONT	_CheckNativeCharset (HDC hDC) ;

/*========================================================================
 *	class CGetTextExtentEditSession
 */
class CGetTextExtentEditSession : public CEditSessionBase
{
public:
    CGetTextExtentEditSession (CSkkImeTextService* pTSF, ITfContext* pContext, ITfContextView* pContextView, ITfRange* pRange, RECT* pRect) : CEditSessionBase (pContext)
    {
		m_pTSF				= pTSF ;
        m_pContextView		= pContextView ;
        m_pRange			= pRange ;
		m_pRect				= pRect ;
    }

    // ITfEditSession
    STDMETHODIMP	DoEditSession (TfEditCookie ec) ;

private:
	CSkkImeTextService*	m_pTSF ;
    ITfContextView*		m_pContextView ;
    ITfRange*			m_pRange ;
	RECT*				m_pRect ;
} ;

STDAPI
CGetTextExtentEditSession::DoEditSession (TfEditCookie ec)
{
	RECT					rc, rcTemp ;
    BOOL					fClipped ;
	HRESULT					hr ;

	if (FAILED (m_pContextView->GetScreenExt (&rcTemp)))
		return	E_FAIL ;

	hr	= m_pContextView->GetTextExt (ec, m_pRange, &rc, &fClipped) ;
	if (hr == S_OK && rc.top < rc.bottom) {
		//rc.left		= rcTemp.left ;
		//rc.right	= rcTemp.right ;
		if (rc.top > rcTemp.bottom || rc.bottom < rcTemp.top) {
			DEBUGPRINTF ((TEXT ("_pContextView->GetTextExt(l:%d,r:%d,t:%d,b:%d,clip:%d) fail?\n"), rc.left, rc.right, rc.top, rc.bottom, fClipped)) ;
			rc.top		= rcTemp.top ;
			rc.bottom	= rcTemp.bottom ;
			if (m_pRect)
				*m_pRect	= rcTemp ;
		} else {
			DEBUGPRINTF ((TEXT ("_pContextView->GetTextExt(l:%d,r:%d,t:%d,b:%d,clip:%d)\n"), rc.left, rc.right, rc.top, rc.bottom, fClipped)) ;
			if (m_pRect)
				*m_pRect	= rc ;
		}
	    return	S_OK ;
	} else {
		DEBUGPRINTF ((TEXT ("_pContextView->GetTextExt(l:%d,r:%d,t:%d,b:%d, clip:%d) failed(0x%x),(range:%p)\n"),
			rc.left, rc.right, rc.top, rc.bottom,
			hr, m_pRange, fClipped)) ;
		if (m_pRect)
			*m_pRect	= rcTemp ;
		return	S_OK ;
	}
}

/*========================================================================
 *	class CSkkImeMgr
 */
CSkkImeMgr::CSkkImeMgr (CSkkImeTextService* pTSF)
{
	m_pTSF						= pTSF ;
	m_pTSF->AddRef () ;
	m_pDoc						= NULL ;
	m_dwCookieTextLayoutSink	= TF_INVALID_COOKIE ;
	m_pContext					= NULL ;
	m_cRef						= 1 ;
	DllAddRef () ;
	return ;
}

CSkkImeMgr::~CSkkImeMgr ()
{
	if (m_pDoc != NULL) {
		delete	m_pDoc ;
		m_pDoc	= NULL ;
	}
	m_pTSF->Release () ;
	m_pTSF	= NULL ;
	DllRelease () ;
	return ;
}

STDAPI
CSkkImeMgr::QueryInterface (REFIID riid, void** ppvObj)
{
	if (ppvObj == NULL)
		return	E_INVALIDARG ;

	*ppvObj	= NULL ;
    if (IsEqualIID (riid, IID_IUnknown) ||
		IsEqualIID (riid, IID_ITfTextLayoutSink)) {
        *ppvObj = (ITfTextLayoutSink *)this ;
    }

    if (*ppvObj) {
        AddRef () ;
        return	S_OK ;
    }
    return	E_NOINTERFACE ;
}

STDAPI_(ULONG)
CSkkImeMgr::AddRef ()
{
    return	++m_cRef ;
}

STDAPI_(ULONG)
CSkkImeMgr::Release ()
{
    LONG	cr = --m_cRef ;

    assert (m_cRef >= 0) ;

    if (m_cRef == 0) {
        delete	this ;
    }
    return	cr ;
}

STDAPI
CSkkImeMgr::OnLayoutChange (
	ITfContext*			pContext,
	TfLayoutCode		lcode,
	ITfContextView*		pContextView)
{
	if (pContext != m_pContext)
		return	S_OK ;

	DEBUGPRINTF ((TEXT ("CSkkImeMgr::OnLayoutChange (%d)\n"), lcode)) ;
    switch (lcode) {
	case	TF_LC_CREATE:
	case	TF_LC_CHANGE:
		if (pContext != NULL && m_pTSF != NULL)
			_UpdateUIElements (pContext, m_pTSF->_GetClientId ()) ;
		break ;

	case TF_LC_DESTROY:
		_CloseUIElements () ;
		break ;
	default:
		break ;
    }
    return	S_OK ;
	UNREFERENCED_PARAMETER (pContextView) ;
}

HRESULT
CSkkImeMgr::AdviseTextLayoutSink (ITfContext* pContext)
{
    HRESULT		hr ;
    ITfSource*	pSource	= NULL ;

	if (m_pContext == pContext)
		return	S_OK ;

	if (m_pContext != NULL) 
		UnadviseTextLayoutSink () ;

	if (pContext == NULL) 
		return	S_FALSE ;

    hr	= E_FAIL ;
    if (FAILED (pContext->QueryInterface (IID_ITfSource, (void **)&pSource)))
        goto	Exit ;

    if (FAILED (pSource->AdviseSink (IID_ITfTextLayoutSink, (ITfTextLayoutSink *)this, &m_dwCookieTextLayoutSink)))
        goto	Exit ;

    hr	= S_OK ;
	m_pContext	= pContext ;
	m_pContext->AddRef () ;
Exit:
    if (pSource != NULL)
        pSource->Release () ;
    return	hr ;
}

HRESULT
CSkkImeMgr::UnadviseTextLayoutSink ()
{
    HRESULT		hr ;
    ITfSource*	pSource	= NULL ;

    hr	= E_FAIL ;
    if (m_pContext == NULL)
        goto	Exit ;

    if (FAILED (m_pContext->QueryInterface (IID_ITfSource, (void **)&pSource)))
        goto	Exit ;

    if (FAILED (pSource->UnadviseSink (m_dwCookieTextLayoutSink)))
        goto	Exit;

    hr	= S_OK ;
Exit:
	if (m_pContext != NULL) {
		m_pContext->Release () ;
		m_pContext	= NULL ;
	}
	m_dwCookieTextLayoutSink	= TF_INVALID_COOKIE ;
    if (pSource != NULL)
        pSource->Release () ;
    return	hr ;
}

/*========================================================================
 */
BOOL
CSkkImeMgr::_Init ()
{
	DEBUGPRINTF ((TEXT ("CSkkImeMgr::_Init ()\n"))) ;

	m_pDoc	= new CImeDoc () ;
	if (m_pDoc == NULL)
		return	FALSE ;

	if (! m_pDoc->bInit (m_pTSF->_GetIMEConfig ())) {
		delete	m_pDoc ;
		m_pDoc	= NULL ;
		return	FALSE ;
	}
	return	TRUE ;
}

void
CSkkImeMgr::_Uninit ()
{
	DEBUGPRINTF ((TEXT ("CSkkImeMgr::_Uninit ()\n"))) ;

	_CloseUIElements () ;

	if (m_pDoc != NULL) {
		delete	m_pDoc ;
		m_pDoc	= NULL ;
	}
	return ;
}

/*	L[}bvɏ]̂؂B
 *	L[Dǂ́A
 *		1. tH[JX Top Buffer ɑ݂邩ۂ 
 *			Yes -> 2. ̃`FbN
 *			No  -> L[D
 *		2. vtBNX邩
 *			Yes -> L[D (̌AL[ԂȂ)
 *			No  -> 3. ̃`FbN
 *		3. L[}bv (SKK//Jĩ[hɈˑ邪)
 *			L[̓J[\̈ړɊւ̂łāAҏW̃eLXg𑀍삷
 *			̂ł͂ȂH
 *			Yes -> L[vZXȂB
 *			No  -> L[DB
 */
int
CSkkImeMgr::QueryProcessKey (
	WPARAM			wParam,		/* [in] */
	LPARAM			lParam,		/* [in] */
	DWORD			dwMask)		/* [in] */
{
	struct TMSG	msg ;
	BOOL		fEaten = FALSE, fToggleIME = FALSE ;

	/*	Control ̂݁AShift ݂̂͒DȂB*/
	if (wParam == VK_CONTROL || wParam == VK_SHIFT || wParam == VK_MENU) {
		return	SKKIME_NOT_PROCESS_KEY ;
	}

	memset (&msg, 0, sizeof (msg)) ;
	msg.m_nMessage	= WM_KEYDOWN ;
	msg.m_wParam	= wParam ;
	msg.m_lParam	= lParam ;
	msg.m_rParam	= dwMask ;
	msg.m_nTime		= GetTickCount () ;

#if defined (DEBUGLOGFILE_ENABLE)
	vDebugPrintfToFile (TEXT ("[%d] CSkkImeMgr::QueryProcessKey: (wParam:0x%0x, lParam:0x%0x, mask:0x%0x)\r\n"),
		GetTickCount (), (unsigned int)wParam, (unsigned int)lParam, msg.m_rParam) ;
#endif

	if (m_pDoc->bQueryToggleIMEEvent (&msg, &fToggleIME) && fToggleIME) {
#if defined (DEBUGLOGFILE_ENABLE)
		vDebugPrintfToFile (TEXT ("[%d] CSkkImeMgr::QueryProcessKey: (wParam:0x%0x, lParam:0x%0x, mask:0x%0x) -> ToggleIME\r\n"),
			GetTickCount (), (unsigned int)wParam, (unsigned int)lParam, msg.m_rParam) ;
#endif
		return	SKKIME_PROCESS_KEY_TOGGLEIME ;
	}

	/*	̑ɂĂ͔YݒB͎̑󂯎ׂH
	 *
	 *	 if (m_pTSF->_IsComposing () || _IsDuringConversion ()) 
	 *	return	TRUE ;
	 */
	if (m_pTSF->_IsComposing ()) {
#if defined (DEBUGLOGFILE_ENABLE)
		vDebugPrintfToFile (TEXT ("[%d] CSkkImeMgr::QueryProcessKey: (wParam:0x%0x, lParam:0x%0x, mask:0x%0x) -> Composing\r\n"),
			GetTickCount (), (unsigned int)wParam, (unsigned int)lParam, msg.m_rParam) ;
#endif
		return	SKKIME_PROCESS_KEY ;
	}

	if (! m_pDoc->bQueryFilterEvent (&msg, &fEaten)) {
#if defined (DEBUGLOGFILE_ENABLE)
		vDebugPrintfToFile (TEXT ("[%d] CSkkImeMgr::QueryProcessKey: (wParam:0x%0x, lParam:0x%0x, mask:0x%0x) -> Not Process\r\n"),
			GetTickCount (), (unsigned int)wParam, (unsigned int)lParam, msg.m_rParam) ;
#endif
		return	SKKIME_NOT_PROCESS_KEY ;
	}
#if defined (DEBUGLOGFILE_ENABLE)
	vDebugPrintfToFile (TEXT ("[%d] CSkkImeMgr::QueryProcessKey: (wParam:0x%0x, lParam:0x%0x, mask:0x%0x) -> Process?(%d)\r\n"),
		GetTickCount (), (unsigned int)wParam, (unsigned int)lParam, msg.m_rParam, fEaten) ;
#endif
	return	fEaten? SKKIME_PROCESS_KEY : SKKIME_NOT_PROCESS_KEY ;
}

/*	DoEditSession ̒ĂяoB
 */
HRESULT
CSkkImeMgr::OnEditSession (
	ITfContext*		pContext,	/* [in] edit context */
	TfEditCookie	ec,			/* [in] edit cookie */
	UINT			uMessage,
	WPARAM			wParam,		/* [in] virtual keycode */
	LPARAM			lParam,		/* [in] */
	DWORD			dwMask,		/* [in] */
	BOOL*			pfEaten)	/* [out] Is this key event eaten? */
{
	struct TMSG	msg ;
	int		nCMode ;
	BOOL	fToggleIME	= FALSE ;
	HRESULT	hr			= S_OK ;
	const struct TMSG*	pUnprocessMsg	= NULL ;
	BOOL	bEaten ;

	memset (&msg, 0, sizeof (msg)) ;
	msg.m_nMessage	= uMessage ;
	msg.m_wParam	= wParam ;
	msg.m_lParam	= lParam ;
	msg.m_rParam	= dwMask ;
	msg.m_nTime		= GetTickCount () ;

#if defined (DEBUGLOGFILE_ENABLE)
	vDebugPrintfToFile (TEXT ("[%d] CSkkImeMgr::OnEditSession: (wParam:0x%0x, lParam:0x%0x, mask:0x%0x)\r\n"), 
		GetTickCount (), (unsigned int)wParam, (unsigned int)lParam, msg.m_rParam) ;
#endif

	if (! m_pDoc->bQueryToggleIMEEvent (&msg, &fToggleIME))
		fToggleIME	= FALSE ;
	if (fToggleIME) {
		BOOL	fOpen	= m_pTSF->_IsKeyboardOpen () ;
		m_pTSF->_SetKeyboardOpen (fOpen? FALSE : TRUE) ;
		if (fOpen) {
			BOOL	fComposing ;
			ITfComposition*	pComposition ;

			fComposing	= m_pTSF->_IsComposing (&pComposition) ;
			if (fComposing) {
				m_pTSF->_TerminateCompositionInContext (pContext) ;
			}
			if (m_pDoc != NULL)
				m_pDoc->vClear (TRUE) ;
		}
		*pfEaten	= TRUE ;
		return	S_OK ;
	}

	/*	filter O̕ϊ[hLĂB
	 */
	nCMode	= m_pDoc->iGetConversionMode () ;
	if (! m_pDoc->bFilterEvent (&msg))
		return	S_FALSE ;
	bEaten			= TRUE ;
	pUnprocessMsg	= m_pDoc->pGetUnprocessEvent () ;
	if (pUnprocessMsg != NULL) {
		if (pUnprocessMsg->m_nMessage == msg.m_nMessage &&
			pUnprocessMsg->m_wParam   == msg.m_wParam &&
			pUnprocessMsg->m_lParam   == msg.m_lParam) {
			bEaten	= FALSE ;
		}
	}
	*pfEaten	= bEaten ;

#if defined (DEBUG)
	if (pfEaten != NULL)
		DEBUGPRINTF ((TEXT ("CSkkImeMgr::OnEditSession (Eaten:%d)\n"), (int)*pfEaten)) ;
#endif
	if (! _UpdateCompose (pContext, ec)) {
		DEBUGPRINTF ((TEXT ("CSkkImeMgr::_UpdateCompose failed.\n"))) ;
		hr	= E_FAIL ;
	}
	if (! _UpdateDisplayAttribute (pContext, ec)) {
		DEBUGPRINTF ((TEXT ("CSkkImeMgr::_UpdateDisplayAttribute failed.\n"))) ;
		hr	= E_FAIL ;
	}

	/*	ϊ[hXVĂ΁ALangBar ɔfB
	 *	܂ACompartment ɔfB
	 */
	if (nCMode != m_pDoc->iGetConversionMode ()) {
#if defined (TF_CONVERSIONMODE_NATIVE)
		unsigned int	uTFCMode	= (unsigned int)-1 ;

		switch (m_pDoc->iGetConversionMode ()) {
		case	IMECMODE_ASCII:
			uTFCMode	= TF_CONVERSIONMODE_ALPHANUMERIC ;
			break ;
		case	IMECMODE_ZENKAKU:
			uTFCMode	= TF_CONVERSIONMODE_ALPHANUMERIC | TF_CONVERSIONMODE_FULLSHAPE ;
			break ;
		case	IMECMODE_HANKANA:
			uTFCMode	= TF_CONVERSIONMODE_NATIVE | TF_CONVERSIONMODE_KATAKANA ;
			break ;
		case	IMECMODE_HANKANA_ROMAN:
			uTFCMode	= TF_CONVERSIONMODE_NATIVE | TF_CONVERSIONMODE_ALPHANUMERIC ;
			break ;
		case	IMECMODE_KATAKANA:
			uTFCMode	= TF_CONVERSIONMODE_NATIVE | TF_CONVERSIONMODE_KATAKANA | TF_CONVERSIONMODE_FULLSHAPE ;
			break ;
		case	IMECMODE_HIRAGANA:
			uTFCMode	= TF_CONVERSIONMODE_NATIVE | TF_CONVERSIONMODE_FULLSHAPE ;
			break ;
		default:
			/* error */
			break ;
		}
		if (uTFCMode != (unsigned int)-1) 
			(void) m_pTSF->_SetKeyboardInputModeConversion (uTFCMode) ;
#endif
		m_pTSF->_UpdateLangBarItem (c_guidConversionModeItemButton) ;
	}
	/*	SentenseMode ͏ TF_SENTENCEMODE_NONE */

	/*	ToolTipUI, CandidateListUI ̍XVB*/
	if (FAILED (_UpdateUIElements (pContext, ec))) {
		DEBUGPRINTF ((TEXT ("CSkkImeMgr::_UpdateStatusWindow failed.\n"))) ;
		hr	= E_FAIL ;
	}
	return	hr ;
}

HRESULT
CSkkImeMgr::OnClearSession (
	TfEditCookie		ec,
	ITfComposition*		pComposition,
	BOOL				bFullClear)
{
	BOOL	fRetval ;
	ITfRange*		pRange ;
	
	DEBUGPRINTF ((TEXT ("CSkkImeTextService::OnClearSession ()\n"))) ;

	if (pComposition != NULL &&
		pComposition->GetRange (&pRange) == S_OK) {
		fRetval	= pRange->SetText (ec, 0, L"", 0) == S_OK ;
		pRange->Release () ;
		pRange	= NULL ;
	}
	if (m_pDoc != NULL) {
		/*	̏ł oh ̓ԂێłȂ̂ŌӖɂ
		 *	ȂȂBA"IME ON","COMPOSITION TEXTȂ" Ԃŕsӂ
		 *	OnClearSession ĂяoāA̓[hωĂ܂悤
		 *	Ԃ邱Ƃ͂łB
		 */
		int	nMode	= m_pDoc->iGetConversionMode () ;
		m_pDoc->vClear (bFullClear) ;
		m_pDoc->bSetConversionMode (nMode) ;
	}
	_CloseUIElements () ;
	return	S_OK ;
}

HRESULT
CSkkImeMgr::OnToggleImeSession (
	ITfContext*		pContext,	/* [in] edit context */
	TfEditCookie	ec)			/* [in] edit cookie */
{
	BOOL	fOpen ;

	if (m_pTSF == NULL)
		return	E_FAIL ;

	fOpen	= m_pTSF->_IsKeyboardOpen () ;
	m_pTSF->_SetKeyboardOpen (fOpen? FALSE : TRUE) ;
	if (fOpen) {
		BOOL	fComposing ;
		ITfComposition*	pComposition ;

		fComposing	= m_pTSF->_IsComposing (&pComposition) ;
		if (fComposing) {
			m_pTSF->_TerminateCompositionInContext (pContext) ;
		}
		if (m_pDoc != NULL)
			m_pDoc->vClear (TRUE) ;
	}
	return	S_OK ;
	UNREFERENCED_PARAMETER (ec) ;
}

HRESULT
CSkkImeMgr::OnReconvertSession (
	ITfContext*	pContext,	/* [in] edit context */
	TfEditCookie	ec,			/* [in] edit cookie */
	LPCWSTR		wstrText)
{
	int		nCMode ;

	if (m_pDoc == NULL || m_pTSF->_IsKeyboardDisabled ())
		return	E_FAIL ;

	nCMode	= m_pDoc->iGetConversionMode () ;
	m_pDoc->bSetReconvertText (wstrText, lstrlenW (wstrText)) ;
	_UpdateCompose (pContext, ec) ;
	_UpdateDisplayAttribute (pContext, ec) ;
	if (nCMode != m_pDoc->iGetConversionMode ()) 
		m_pTSF->_UpdateLangBarItem (c_guidConversionModeItemButton) ;

	if (! m_pTSF->_IsKeyboardOpen ()) {
		m_pTSF->_SetKeyboardOpen (TRUE) ;
	}
	_UpdateUIElements (pContext, ec) ;
	return	S_OK ;
}

HRESULT
CSkkImeMgr::OnChangeInputModeSession (
	ITfContext*		pContext,		/* [in] edit context */
	TfEditCookie	ec,				/* [in] edit cookie */
	UINT			uConversionMode,
	UINT			uSentenceMode,
	BOOL			bUpdateCompartment)
{
	HRESULT	hr			= S_OK ;
#if defined (TF_CONVERSIONMODE_NATIVE)
	int		nCMode ;

	switch (uConversionMode) {
	case	TF_CONVERSIONMODE_ALPHANUMERIC:
		nCMode	= IMECMODE_ASCII ;
		break ;
	case	TF_CONVERSIONMODE_ALPHANUMERIC | TF_CONVERSIONMODE_FULLSHAPE:
		nCMode	= IMECMODE_ZENKAKU ;
		break ;
	case	TF_CONVERSIONMODE_NATIVE | TF_CONVERSIONMODE_KATAKANA:
		nCMode	= IMECMODE_HANKANA ;
		break ;
	case	TF_CONVERSIONMODE_NATIVE | TF_CONVERSIONMODE_ALPHANUMERIC:
		nCMode	= IMECMODE_HANKANA_ROMAN ;
		break ;
	case	TF_CONVERSIONMODE_NATIVE | TF_CONVERSIONMODE_KATAKANA | TF_CONVERSIONMODE_FULLSHAPE:
		nCMode	= IMECMODE_KATAKANA ;
			break ;
	case	TF_CONVERSIONMODE_NATIVE | TF_CONVERSIONMODE_FULLSHAPE:
		nCMode	= IMECMODE_HIRAGANA ;
		break ;
	default:
		/*	̏ꍇ Compartment F߂Ȃ̂ŁAB
		 *	 SetValue ƕK E_UNEXPECTED ̃G[߂̂ Compartment 
		 *	Ƃ͂łȂB
		 */
		if (! bUpdateCompartment)
			return	S_OK ;
		nCMode	= -1 ;
	}
	/*	filter O̕ϊ[hLĂB
	 */
	if (m_pDoc->iGetConversionMode () == nCMode)
		return	S_FALSE ;
	if (m_pDoc->bIsStatusActivep () || m_pDoc->bShowHenkanCandidatesModep() || m_pDoc->bInputByCodeOrMenuModep (NULL))
		return	S_FALSE ;
	if (! m_pDoc->bSetConversionMode (nCMode))
		return	S_FALSE ;

	if (! _UpdateCompose (pContext, ec)) {
		DEBUGPRINTF ((TEXT ("CSkkImeMgr::_UpdateCompose failed.\n"))) ;
		hr	= E_FAIL ;
	}
	if (! _UpdateDisplayAttribute (pContext, ec)) {
		DEBUGPRINTF ((TEXT ("CSkkImeMgr::_UpdateDisplayAttribute failed.\n"))) ;
		hr	= E_FAIL ;
	}

	/*	ϊ[hXVĂ΁ALangBar ɔfB
	 *	܂ACompartment ɔfB
	 */
	if (bUpdateCompartment) {
		unsigned int	uTFCMode	= (unsigned int)-1 ;

		switch (m_pDoc->iGetConversionMode ()) {
		case	IMECMODE_ASCII:
			uTFCMode	= TF_CONVERSIONMODE_ALPHANUMERIC ;
			break ;
		case	IMECMODE_ZENKAKU:
			uTFCMode	= TF_CONVERSIONMODE_ALPHANUMERIC | TF_CONVERSIONMODE_FULLSHAPE ;
			break ;
		case	IMECMODE_HANKANA:
			uTFCMode	= TF_CONVERSIONMODE_NATIVE | TF_CONVERSIONMODE_KATAKANA ;
			break ;
		case	IMECMODE_KATAKANA:
			uTFCMode	= TF_CONVERSIONMODE_NATIVE | TF_CONVERSIONMODE_KATAKANA | TF_CONVERSIONMODE_FULLSHAPE ;
			break ;
		case	IMECMODE_HIRAGANA:
			uTFCMode	= TF_CONVERSIONMODE_NATIVE | TF_CONVERSIONMODE_FULLSHAPE ;
			break ;
		default:
			/* error */
			break ;
		}
		if (uTFCMode != (unsigned int)-1) 
			(void) m_pTSF->_SetKeyboardInputModeConversion (uTFCMode) ;
	}
	m_pTSF->_UpdateLangBarItem (c_guidConversionModeItemButton) ;

	/*	SentenseMode ͏ TF_SENTENCEMODE_NONE */
	if (bUpdateCompartment) {
		(void) m_pTSF->_SetKeyboardInputModeSentence (TF_SENTENCEMODE_NONE) ;
	}

	/*	ToolTipUI, CandidateListUI ̍XVB
	 */
	if (FAILED (_UpdateUIElements (pContext, ec))) {
		DEBUGPRINTF ((TEXT ("CSkkImeMgr::_UpdateStatusWindow failed.\n"))) ;
		hr	= E_FAIL ;
	}
#else
	m_pTSF->_UpdateLangBarItem (c_guidConversionModeItemButton) ;
	if (FAILED (_UpdateUIElements (pContext, ec))) {
		DEBUGPRINTF ((TEXT ("CSkkImeMgr::_UpdateStatusWindow failed.\n"))) ;
		hr	= E_FAIL ;
	}
#endif
	return	hr ;
	UNREFERENCED_PARAMETER (uSentenceMode) ;
}

CImeDoc*
CSkkImeMgr::GetDocument ()
{
	return		m_pDoc ;
}

void
CSkkImeMgr::OnSetFocus (
	BOOL			fFocus)
{
	_PopupUIElements (fFocus) ;
	return ;
}

BOOL
CSkkImeMgr::OnKeyboardOpen (
	BOOL			fOpen)
{
	if (fOpen) {
		/*	Keyboard  open ɂȂɁuȁv[hɖ߂ݒ肪Lǂ
		 *	B
		 */
		if (CImeConfig::bKanaModeWhenOpenp (m_pTSF->_GetIMEConfig ())) {
			if (m_pDoc == NULL)
				return	FALSE ;
			if (m_pDoc->iGetConversionMode () != IMECMODE_HIRAGANA)
				m_pDoc->bSetConversionMode (IMECMODE_HIRAGANA) ;
		}
	} else {
		_CloseUIElements () ;
		if (m_pDoc != NULL) {
			int	nMode	= m_pDoc->iGetConversionMode () ;
			m_pDoc->vClear (TRUE) ;
			m_pDoc->bSetConversionMode (nMode) ;
		}
	}
	return	TRUE ;
}

/*========================================================================*
 */
BOOL
CSkkImeMgr::_UpdateCompose (
	ITfContext*	pContext,
	TfEditCookie	ec)
{
    TF_SELECTION		tfSelection ;
	ULONG				cFetched ;
	ITfComposition*		pComposition ;
	WCHAR				wszPText [512] ;	/* [Amagic number ... */
	WCHAR				wszReadingText [512] ;
	LPCWSTR	wstrPText ;
	LPWSTR		wpDest ;
	int					nPText, nRawShift, nRawCursor ;
	int					nReadingShift, nReadingText ;
	int		n, nShift, nCursor, nDest ;
	LONG				cch ;
	ITfRange*			pRange		= NULL ;
	ITfProperty*		pProp 		= NULL ;
	ITfProperty*		pPropAttr	= NULL ;
	BOOL				fComposing ;
	BOOL				fRetval		= FALSE ;
	BOOL				fContinue ;

    if (pContext->GetSelection (ec, TF_DEFAULT_SELECTION, 1, &tfSelection, &cFetched) != S_OK ||
		cFetched != 1)
        return	FALSE ;

	fComposing	= m_pTSF->_IsComposing (&pComposition) ;
	wstrPText	= m_pDoc->pGetPreeditText (&nPText) ;

	/*	Composition String ŁAMinibuffer LłȂꍇɂ
	 *	Composition IĔB
	 *
	 *	if (nPText <= 0 __&& ! m_pDoc->bIsStatusActivep ()__) {
	 *	AȂ status activep  comment outH Wed Oct 29 23:26:43 2003
	 */
	if (nPText <= 0 && ! (m_pDoc->bIsStatusActivep () || m_pDoc->bShowHenkanCandidatesModep() || m_pDoc->bInputByCodeOrMenuModep (NULL))) {
		if (fComposing) 
			m_pTSF->_TerminateCompositionInContext (pContext) ;
		fRetval	= TRUE ;
		goto	Exit ;
	}
	if (! fComposing) {
		if (! m_pTSF->_StartComposition (pContext))
			goto	Exit ;
		(void) m_pTSF->_IsComposing (&pComposition) ;
	}

	/*	Composition ̃eLXg̑B
	 */
	if (pComposition == NULL || pComposition->GetRange (&pRange) != S_OK || pRange == NULL) 
		goto	Exit ;

	/*	attribute  clear ĂB*/
    if (SUCCEEDED (pContext->GetProperty (GUID_PROP_ATTRIBUTE, &pPropAttr) && pPropAttr != NULL)) {
		pPropAttr->Clear (ec, pRange) ;
		pPropAttr->Release () ;
	}

#if defined (buggy_code_using_prop_reading) || 1
	if (SUCCEEDED (pContext->GetProperty (GUID_PROP_READING, &pProp) && pProp != NULL)) {
		pProp->Clear (ec, pRange) ;
	} else {
		pProp	= NULL ;
	}
#endif

	/*	Composition ̃eLXg̓Am肳̂̑B
	 *	StartRange ړāAړŊOꂽ͈͂ɂ̂m
	 *	ƎvB
	 */
	if (! m_pDoc->bQueryUpdateContext (&nRawShift, &nRawCursor, &fContinue)) 
		goto	Exit ;

#if defined (DEBUGLOGFILE_ENABLE)
	vDebugPrintfToFile (TEXT ("[%d] UpdateCompose: RawShift(%d), RawCursor(%d), Continue(%d), Text(\"%s\"), TextLen(%d)\r\n"),
		GetTickCount (), nRawShift, nRawCursor, fContinue, wstrPText, nPText) ;
#endif

	nReadingShift	= nRawShift ;
	nReadingText	= m_pDoc->iGetReadingText (wszReadingText, MYARRAYSIZE (wszReadingText) - 1, nRawShift, &nReadingShift) ;
	wszReadingText [nReadingText]	= L'\0' ;

//	if (nPText <= 0 && m_pDoc->bIsStatusActivep ()) {
	if (nPText <= 0 && (m_pDoc->bRecursiveEditp () || m_pDoc->bInputByCodeOrMenuModep (NULL) || m_pDoc->bShowHenkanCandidatesModep ())) {
		/*	Fri Oct 31 23:26:15 2003
		 *		Ȃ茵AǉB̐ŚA
		 *		j-input-by-code-or-menu sԂł́A
		 *		composition ݂ȂƎvāAterminate composition
		 *		Ă΂ĂȂ߁B
		 */
		/*	Wed Aug 31 09:45:03 2005
		 *		c芸ēx폜B̎̈Ӑ}͖YĂ܂Ăc
		if (pRange->SetText (ec, 0, L" ", 1) != S_OK) 
			goto	Exit ;
		 *	폜R OfficeXP ̍sŃXy[X}ƁASetText (L" ")
		 *	 E_FAIL ĕςɂȂ邽߁B
		 */
		nShift		= 0 ;
		nCursor		= 0 ;
		if (pRange->SetText (ec, 0, L" ", 1) != S_OK) {
			/*	ŃG[ĂAUpdateContext ͌ĂяoȂ΂ȂȂB
			 *	łȂ΁A Update Ƃ[vāǍ
			 *	Xe[gɎsB(Fri Dec 09 11:36:30 2005)
			 */
			(void) m_pDoc->bUpdateContext () ;
			goto	Exit ;
		}
	} else {
		/*	nꂽeLXg̒Ɋ܂܂sR[h̖
		 *	΂ȂȂBL'\n'  0x0D, 0x0A ւƓWJƓ
		 *	J[\ʒuCKvB
		 */
		wpDest		= wszPText ;
		nDest		= MYARRAYSIZE (wszPText) ;
		nShift		= nRawShift ;
		nCursor		= nRawCursor ;
		for (n = 0 ; n < nPText ; n ++) {
			if (n == nRawShift) 
				nShift	= wpDest - wszPText ;
			if (n == nRawCursor) 
				nCursor	= wpDest - wszPText ;
			if (*wstrPText == L'\n') {
				*wpDest ++	= 0x0D ;
				nDest -- ;
				/*	0x0D, 0x0A ɒ̂͗ǂȂ炵B
				 *	Outlook2000 ͖ȂAInternet Explorer6
				 *	Ŗ肪B
				 *	䂦Ɏ̃R[h comment ɂĂB
				 *
				 *	if (nDest > 1) {
				 *		*wpDest ++	= 0x0D ;
				 *		*wpDest ++	= 0x0A ;
				 *		nDest -= 2 ;
				 *	}
				 */
			} else {
				*wpDest ++	= *wstrPText ;
				nDest -- ;
			}
			wstrPText	++ ;
		}
#if defined (DEBUG) || defined (_DEBUG)
		{
			WCHAR	buf [256] ;
			int		n ;

			n	= wpDest - wszPText ;
			memcpy (buf, wszPText, sizeof (WCHAR) * n) ;
			buf [n]	= L'\0' ;
			DebugPrintf (TEXT ("pRange->SetText (\"%s\"(len:%d))\n"), buf, (int)(wpDest - wszPText)) ;
		}
#endif
		/*	Word 2003 Ő}\ꂽꍇAeLXg{bNX̍ŏ̈ꕶ Shift+Ƃ
		 *	uȂƂvoꍇɁAĂ܂肪邪c
		 *	wordŏ1͐ΉpƉ肵ĂۂCcB
		 */
		nPText	= wpDest - wszPText ;
		if (pRange->SetText (ec, 0, wszPText, nPText) != S_OK) {
			/*	ŃG[ĂAUpdateContext ͌ĂяoȂ΂ȂȂB
			 *	łȂ΁A Update Ƃ[vāǍ
			 *	Xe[gɎsB(Fri Dec 09 11:36:30 2005)
			 */
			(void) m_pDoc->bUpdateContext () ;
			goto	Exit ;
		}
	}

	if (! m_pDoc->bUpdateContext ())
		goto	Exit ;

	DEBUGPRINTF ((TEXT ("nShift = %d/%d, nCursor = %d/%d, fContinute = %d\n"),
				  nShift, nRawShift, nCursor, nRawCursor, fContinue)) ;

	if (pRange->ShiftStart (ec, nShift, &cch, NULL) != S_OK)
		goto	Exit ;

	if (pComposition->ShiftStart (ec, pRange) != S_OK ||
		pComposition->ShiftEnd   (ec, pRange) != S_OK)
		goto	Exit ;

	/*	ҏW̃eLXg LANGID ݒ肵Ă݂Bꂪ MSWORD ŏۂɘa
	 *	̂낤H (Thu Jul 26 20:32:31 2007)
	 */
#if ! defined (no_prop_langid)
	{
		ITfProperty*	pPropLang	= NULL ;

		if (SUCCEEDED (pContext->GetProperty (GUID_PROP_LANGID, &pPropLang)) && pPropLang != NULL) {
			VARIANT		var ;
			HRESULT		hr ;

			VariantInit (&var) ;
			var.vt		= VT_I4 ;
			var.ulVal	= SKKIME_LANGID ;
			hr			= pPropLang->SetValue (ec, pRange, &var) ;
			pPropLang->Release () ;
		}
	}
#endif

	tfSelection.range->ShiftStartToRange (ec, pRange, TF_ANCHOR_START) ;
	tfSelection.range->Collapse (ec, TF_ANCHOR_START) ;
	tfSelection.range->ShiftStart (ec, nCursor, &cch, NULL) ;
	pContext->SetSelection (ec, 1, &tfSelection) ;

#if defined (buggy_code_using_prop_reading) || 1
	if (pProp != NULL && nReadingText > nReadingShift) {
		VARIANT		var ;
		HRESULT		hr ;

		VariantInit (&var) ;
		var.vt		= VT_BSTR ;
		var.bstrVal	= SysAllocString (wszReadingText + nReadingShift) ;
		hr	= pProp->SetValue (ec, pRange, &var) ;
		if (FAILED (hr)) {
#if defined (DEBUG) || defined (DBG)
			LPVOID lpMsgBuf	= NULL ;
			FormatMessage(
				FORMAT_MESSAGE_ALLOCATE_BUFFER |
				FORMAT_MESSAGE_FROM_SYSTEM |
				FORMAT_MESSAGE_IGNORE_INSERTS,
				NULL,
				hr,
				MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // ̌
				(LPTSTR) &lpMsgBuf,
				0,
				NULL) ;
			if (lpMsgBuf != NULL)
				OutputDebugString ((LPCTSTR)lpMsgBuf) ;
			LocalFree (lpMsgBuf) ;
			DEBUGPRINTF ((TEXT ("pProp->SetValue() failed (0x%lx)\n"), (DWORD)hr)) ;
			if (hr == E_INVALIDARG) {
				DEBUGPRINTF ((TEXT ("E_INVALIDARG\n"))) ;
			}
#endif
		}
		VariantClear (&var) ;
	}
#endif

	DEBUGPRINTF ((TEXT ("CSkkImeMgr::_UpdateCompose () - 8\n"))) ;
	if (! fContinue)
		m_pTSF->_TerminateComposition (ec) ;
	fRetval	= TRUE ;

  Exit:
	if (pProp != NULL)
		pProp->Release () ;
	if (pRange != NULL)
		pRange->Release () ;
	tfSelection.range->Release () ;
	DEBUGPRINTF ((TEXT ("CSkkImeMgr::_UpdateCompose () - leave(%d)\n"), fRetval)) ;
	return	fRetval ;
}

HRESULT
CSkkImeMgr::_UpdateDisplayAttribute (
	ITfContext*	pContext,
	TfEditCookie	ec)
{
	ITfComposition*			pComposition ;
	ITfRange*				pRange			= NULL ;
	ITfRange*				pRangeC			= NULL ;
	BOOL					fComposing, fConverted ;
	int						nBegin, nEnd ;
	HRESULT					hr	= E_FAIL ;

	DEBUGPRINTF ((TEXT ("CSkkImeMgr::_UpdateDisplayAttribute ()\n"))) ;


	fComposing	= m_pTSF->_IsComposing (&pComposition) ;
	if (! fComposing)
		goto	Exit ;
	if ((hr = pComposition->GetRange (&pRange)) != S_OK) 
		goto	Exit ;

	fConverted	= m_pDoc->bGetConvertedRegion (&nBegin, &nEnd) ;
	pRangeC		= NULL ;
	if (fConverted && nBegin < nEnd && nEnd > 0) {
		LONG	cch ;
		if (FAILED (pRange->Clone (&pRangeC)))
			goto	Exit ;
		if (FAILED (pRangeC->ShiftStart (ec, nBegin, &cch, NULL)) ||
			FAILED (pRangeC->Collapse (ec, TF_ANCHOR_START)) ||
			FAILED (pRangeC->ShiftEnd (ec, nEnd - nBegin, &cch, NULL)))
			goto	Exit ;
	}
	if (pRangeC != NULL) {
		hr	= m_pTSF->_SetCompositionDisplayAttributes (ec, pRange, pRangeC) ;
	} else {
		/*	Converted ȂĂ Input Text ݂͑̂ŁACB
		 */
		hr	= m_pTSF->_SetCompositionDisplayAttributes (ec, pRange, NULL) ;
	}
  Exit:
	if (pRange != NULL)
		pRange->Release () ;
	if (pRangeC != NULL)
		pRangeC->Release () ;
    return	hr ;
	UNREFERENCED_PARAMETER (pContext) ;
}

HRESULT
CSkkImeMgr::_QueryFocus (
	ITfContext*	pContext,
	TfEditCookie	ec)
{
    ITfRange*		pRangeComposition ;
    TF_SELECTION	tfSelection ;
	HRESULT			hr	= S_OK ;
	ULONG			cFetched ;
	BOOL			fCovered ;
	ITfComposition*	pComposition ;

//	DEBUGPRINTF ((TEXT ("CSkkImeMgr::_QueryFocus() -- (1) \n"))) ;

	if (! m_pTSF->_IsComposing (&pComposition))
		return	S_OK ;

//	DEBUGPRINTF ((TEXT ("CSkkImeMgr::_QueryFocus() -- (2) \n"))) ;

    if (pContext->GetSelection (ec, TF_DEFAULT_SELECTION, 1, &tfSelection, &cFetched) != S_OK ||
		cFetched != 1)
        return	S_FALSE ;
	
//	DEBUGPRINTF ((TEXT ("CSkkImeMgr::_QueryFocus(%p) -- (3) \n"), pComposition)) ;

    if (pComposition->GetRange (&pRangeComposition) == S_OK) {
        fCovered	= IsRangeCovered (ec, tfSelection.range, pRangeComposition) ;
        pRangeComposition->Release () ;
		
		DEBUGPRINTF ((TEXT ("CSkkImeMgr::_QueryFocus() -- (4) \n"))) ;
        if (!fCovered) {
            hr	= S_FALSE; // don't eat the key, it's outside our composition
            goto	Exit ;
        }
    }
//	DEBUGPRINTF ((TEXT ("CSkkImeMgr::_QueryFocus() -- (5) \n"))) ;
 Exit:
//	DEBUGPRINTF ((TEXT ("CSkkImeMgr::_QueryFocus() -- (6) \n"))) ;
    tfSelection.range->Release () ;
	return	hr ;
}

/*========================================================================
 *	UIElement ̑Ɋւ method
 */
HRESULT
CSkkImeMgr::_UpdateUIElements (
	ITfContext*		pContext,
	TfEditCookie	ec)
{
	if (m_pTSF == NULL || m_pDoc == NULL)
		return	E_FAIL ;

	/*	SetText  Range ɃeLXgݒ肵āAApplication ɐ߂Ă΂炭Ȃ
	 *	Range ̈ʒu͓ȂBAȂ Minibuffer Window \悤ȑ
	 *	(j-input-Ƃ) sƁA\ʒuĂ܂B
	 *	ɂ AdviseTextLayoutSink 𗘗pāAApplication ̌vZIĂʒu
	 *	𓾂邵ȂBAeLXgݒ肳ĂȂ΍sȂ̂
	 *	Minibuffer Window \鎞ɂ͉RłeLXgKvc(悤)
	 *	UpdateCompose  dummy  SetText (L" ", 1) ͂łB(͂ȂsȂ̂
	 *	Application  Text Ȃ̂ Range ̈ʒuvZȂ񂩂`͂Ȃƍl̂
	 *	Ȃ͂ȂcB)
	 *	ܘ_AWordPad ̂悤ɃeLXgĂȂĂȏꏊԂĂ̂
	 *	̂cʓIł͂Ȃ悤B
	 *	(*) Context ĵƂ^͎cĂB̎ŗǂ̂
	 *	sB
	 */
	AdviseTextLayoutSink (pContext) ;

	/* \ʒu̍XV͂(\/\Ɋւ炸) */
	_LayoutUIElements (pContext, ec) ;

	/*	XVꂽgǂƂ́AUIElements ̒gɔCB(Page, Count
	 *	ύXꂽƂǂƂB
	 */	
	_UpdateCandidateListUI () ;
	_UpdateToolTipUI () ;
	return	S_OK ;
}

HRESULT
CSkkImeMgr::_CloseUIElements ()
{
	CSkkImeCandidateListUIElement*	pCandListUI ;
	CSkkImeToolTipUIElement*		pToolTipUI ;
	BOOL	bShow ;

	if (m_pTSF == NULL || m_pDoc == NULL)
		return	E_FAIL ;

	pToolTipUI	= m_pTSF->_GetToolTipUI () ;
	if (pToolTipUI != NULL) {
		if (pToolTipUI->_IsActivep (&bShow)) {
			m_pTSF->_EndToolTipUI () ;
			pToolTipUI->_Close () ;
		}
	}
	pCandListUI	= m_pTSF->_GetCandidateListUI () ;
	if (pCandListUI != NULL) {
		if (pCandListUI->_IsActivep (&bShow)) {
			m_pTSF->_EndCandidateListUI () ;
			pCandListUI->_Close () ;
		}
	}
	UnadviseTextLayoutSink () ;
	return	S_OK ;
}

HRESULT
CSkkImeMgr::_PopupUIElements (
	BOOL		bFocus) 
{
	CSkkImeCandidateListUIElement*	pCandListUI ;
	CSkkImeToolTipUIElement*		pToolTipUI ;
	BOOL	bShow ;

	if (m_pTSF == NULL || m_pDoc == NULL)
		return	E_FAIL ;

	pToolTipUI	= m_pTSF->_GetToolTipUI () ;
	if (pToolTipUI != NULL) {
		if (pToolTipUI->_IsActivep (&bShow)) {
			if (bShow)
				pToolTipUI->_Popup (bFocus) ;
		}
	}
	pCandListUI	= m_pTSF->_GetCandidateListUI () ;
	if (pCandListUI != NULL) {
		if (pCandListUI->_IsActivep (&bShow)) {
			if (bShow)
				pCandListUI->_Popup (bFocus) ;
			m_pTSF->_EndCandidateListUI () ;
		}
	}
	return	S_OK ;
}

/*	ł́AUIElement ̕\ʒuƑ傫𒲐邾ŁA\/\͕ύXȂB
 *	Update ƂȂB
 *	܂A UIElement  UIElementId gĎ擾 UIElement ΏۂƂĂȂ
 *	̂́A\ʒu̒KvȂ͎̂O UIElement ̏ꍇłB
 */
HRESULT
CSkkImeMgr::_UpdateToolTipUI ()
{
	CSkkImeToolTipUIElement*	pToolTipUI ;
	BOOL				bComposing ;
	ITfComposition*		pComposition	= NULL ;
	HRESULT	hr	= S_OK ;
	BOOL	bShow ;

	if (m_pTSF == NULL)
		return	E_FAIL ;

	pToolTipUI	= m_pTSF->_GetToolTipUI () ;
	if (pToolTipUI == NULL)
		return	S_FALSE ;

	/*	Wed Aug 31 09:50:02 2005
	 *		Composing łȂԂ UIStatus  Popup ƁAeLXgҏWĂȂԂ
	 *		Popup cĂ܂̂ł܂ǂȂ (design IɂAITfRange ̒lIɂ)B
	 *---
	 *	Sat Mar 17 11:47:02 2007
	 *		ToolTip \邩ǂ́AMessage 邩ǂōsB⃊Xg\
	 *		ĂԂŁAɃ}EXJ[\킹 ToolTip \邱Ƃ
	 *		悤ɂĂ邪ÃLbN̔́Ał͂łȂ̂ŁB(Window Message
	 *		āA}EXJ[\ȂƂȂc)
	 */
	bComposing	= m_pTSF->_IsComposing (&pComposition) ;
	if (bComposing && m_pDoc->bHaveMessagep ()) {
		LPCWSTR	pwMessage ;
		int		iMessageLen ;
		DWORD	dwId ;

		pwMessage	= m_pDoc->pGetMessage (&iMessageLen) ;
		pToolTipUI->_SetText (pwMessage, iMessageLen) ;

		/*	Active ǂ̔f́AOpen ꂽǂ1_̂݁B
		 *	ꂪƎUIȂ̂ Application UIȂ͕̂sB
		 *
		 *	ɓꂽ bShow ̂ǂ炩̂B
		 */
		if (! pToolTipUI->_IsActivep (&bShow)) {
			hr	= m_pTSF->_BeginToolTipUI (&dwId, &bShow) ;
			if (hr != E_NOINTERFACE && FAILED (hr))
				return	hr ;
			if (hr == E_NOINTERFACE) {
				dwId	= -1 ;
				bShow	= TRUE ;
			}
			hr	= pToolTipUI->_Open (dwId, bShow) ;
			if (FAILED (hr))
				return	hr ;
		} else {
			pToolTipUI->_Update () ;
		}
		m_pTSF->_UpdateToolTipUI () ;
	} else {
		if (pToolTipUI->_IsActivep (&bShow)) {
			m_pTSF->_EndToolTipUI () ;
			pToolTipUI->_Close () ;
		}
	}
	return	hr ;
}

HRESULT
CSkkImeMgr::_UpdateCandidateListUI ()
{
	CSkkImeCandidateListUIElement*	pCandListUI ;
	BOOL				bComposing ;
	ITfComposition*		pComposition	= NULL ;
	HRESULT	hr	= S_OK ;
	BOOL	bShow ;

	if (m_pTSF == NULL)
		return	E_FAIL ;

	pCandListUI	= m_pTSF->_GetCandidateListUI () ;
	if (pCandListUI == NULL)
		return	S_FALSE ;

	/*	Wed Aug 31 09:50:02 2005
	 *		Composing łȂԂ UIStatus  Popup ƁAeLXgҏWĂȂԂ
	 *		Popup cĂ܂̂ł܂ǂȂ (design IɂAITfRange ̒lIɂ)B
	 *---
	 *	ToolTipUI  CandidateListUIAǂ Composing ̂ȂAœn̂
	 *	Ȃ񂶂ȂH
	 */
	bComposing	= m_pTSF->_IsComposing (&pComposition) ;
	if (bComposing && (m_pDoc->bIsStatusActivep () || m_pDoc->bShowHenkanCandidatesModep() || m_pDoc->bInputByCodeOrMenuModep (NULL))) {

		if (! pCandListUI->_IsActivep (&bShow)) {
			DWORD	dwId ;

			hr	= m_pTSF->_BeginCandidateListUI (&dwId, &bShow) ;
			if (hr == E_NOINTERFACE) {
				dwId	= (DWORD)-1 ;
				bShow	= TRUE ;
			} else {
				if (FAILED (hr))
					return	hr ;
			}
			hr	= pCandListUI->_Open (dwId, bShow) ;
			if (FAILED (hr))
				return	hr ;
		} else {
			pCandListUI->_Update () ;
		}
		m_pTSF->_UpdateCandidateListUI () ;
	} else {
		if (pCandListUI->_IsActivep (&bShow)) {
			m_pTSF->_EndCandidateListUI () ;
			pCandListUI->_Close () ;
		}
	}
	return	hr ;
}

HRESULT
CSkkImeMgr::_LayoutUIElements (
	ITfContext*				pContext,
	TfEditCookie			ec)
{
	ITfComposition*			pComposition	= NULL ;
	const IMECANDIDATES*	pMyCand			= NULL ;
	RECT					rcUI, rcTarget ;
	HRESULT					hr				= E_FAIL ;
	BOOL					bTargetRect		= FALSE ;
	CSkkImeCandidateListUIElement*	pCandListUI ;
	CSkkImeToolTipUIElement*		pToolTipUI ;

	if (pContext == NULL)
		return	E_FAIL ;
	if (! m_pTSF->_IsComposing (&pComposition) || pComposition == NULL) 
		return	E_FAIL ;
	{
		ITfRange*			pRange	= NULL ;

		if (SUCCEEDED (pComposition->GetRange (&pRange)) && pRange != NULL) {
			hr	= _GetRangeRect (pContext, ec, pRange, &rcTarget) ;
			if (hr == S_OK && rcTarget.top < rcTarget.bottom)
				bTargetRect	= TRUE ;
			pRange->Release () ;
		}
	}
	if (! bTargetRect) {
		DEBUGPRINTF ((TEXT ("_GetRangeRect all failed.\n"))) ;
		rcTarget.left	= 0 ;
		rcTarget.top	= 0 ;
		rcTarget.right	= 1 ;
		rcTarget.bottom	= 1 ;
	}

	_CalculateUIElementLayout (&rcUI, &rcTarget) ;

	pCandListUI	= m_pTSF->_GetCandidateListUI () ;
	if (pCandListUI != NULL)
		pCandListUI->_MoveWindow (rcUI.left, rcUI.top, rcUI.right, rcUI.bottom) ;

	/* ToolTipUI ̈ʒu CandListUI ̕\/\ŕωB*/
	if (m_pDoc != NULL && m_pDoc->bHaveMessagep ()) {
		pToolTipUI	= m_pTSF->_GetToolTipUI () ;
		if (pToolTipUI != NULL) {
			pMyCand	= m_pDoc->pGetStatusText () ;
			if (pMyCand != NULL && pMyCand->bIsEnabledp ()) {
				pToolTipUI->_MoveWindow (rcUI.left, rcUI.top + rcUI.bottom, 0, 0) ;
			} else {
				pToolTipUI->_MoveWindow (rcUI.left, rcUI.top, 0, 0) ;
			}
		}
	}
	return	hr ;
}

HRESULT
CSkkImeMgr::_GetRangeRect (
	ITfContext*				pContext,
	TfEditCookie			ec,
	ITfRange*				pRange,
	RECT*					pRect)
{
	CGetTextExtentEditSession*	pEditSession	= NULL ;
	ITfContextView*				pContextView	= NULL ;
	HRESULT			hr, hrTargetRect ;

	if (pContext == NULL || pRange == NULL || pRect == NULL)
		return	E_INVALIDARG ;

	hr	= pContext->GetActiveView (&pContextView) ;
	if (FAILED (hr))
		return	hr ;

	pEditSession	= new CGetTextExtentEditSession (m_pTSF, pContext, pContextView, pRange, pRect) ;
	if (pEditSession == NULL) {
		hr	= E_FAIL ;
		goto	out ;
	}
	hr	= pContext->RequestEditSession (ec, pEditSession, TF_ES_SYNC | TF_ES_READ, &hrTargetRect) ;
	pEditSession->Release () ;
	if (hr == S_OK) 
		hr	= hrTargetRect ;
out:
	pContextView->Release () ;
	return	hr ;
}

HRESULT
CSkkImeMgr::_CalculateUIElementLayout (
	RECT*				prcDest,
	const RECT*			prcSrc)
{
	const IMECANDIDATES*	pMyCand ;
	HDC					hDC ;
	HFONT				hOldFont ;
	RECT				rcUI, rcWork ;
	int					nWidth, nHeight, nFontHeight, nFontWidth ;
	int					nFrameWidth, nFrameHeight, nDefaultWidth, nDefaultHeight ;
	int					iStyle ;
	TEXTMETRIC			tm ;
	SIZE				sz ;

	memset (&rcUI, 0, sizeof (rcUI)) ;

	/*	ʑŜ̃TCY𓾂B*/
	SystemParametersInfo (SPI_GETWORKAREA, 0, &rcWork, 0) ;

	hDC		= CreateDC (TEXT ("DISPLAY"), NULL, NULL, NULL) ;
	if (hDC == NULL) {
		if (prcDest != NULL) {
			prcDest->left	= 0 ;
			prcDest->top	= 0 ;
			prcDest->right	= rcWork.right ;
			prcDest->bottom	= 18 ;
		}
		return	E_FAIL ;	// ̏ꍇ͌vZs\B
	}

	hOldFont	= (HFONT) _CheckNativeCharset (hDC) ;
	if (GetTextMetrics (hDC, &tm)) {
		nFontWidth	= tm.tmAveCharWidth ;
		nFontHeight	= tm.tmHeight ;
	} else {
		GetTextExtentPoint32W (hDC, L"", 1, &sz) ;
		nFontWidth	= (sz.cx + 1) / 2 ;
		nFontHeight	= sz.cy ;
	}
	nFrameWidth		= GetSystemMetrics (SM_CXDLGFRAME) * 2 ;
	nFrameHeight	= GetSystemMetrics (SM_CYDLGFRAME) * 2 ;
	nDefaultWidth	= nFontWidth * DEFAULT_MINIBUF_CHARWIDTH + nFrameWidth ;
	nDefaultWidth	= (nDefaultWidth > rcWork.right)? rcWork.right : nDefaultWidth ;
	nDefaultHeight	= nFontHeight + nFrameHeight ;
	nDefaultHeight	= (nDefaultHeight > rcWork.bottom)? rcWork.bottom : nDefaultHeight ;

	nWidth	= (prcSrc != NULL && nDefaultWidth < (prcSrc->right - prcSrc->left))? (prcSrc->right - prcSrc->left) : nDefaultWidth ;
	nHeight	= 0 ;

	if ((nWidth + nFrameWidth) > (rcWork.right - rcWork.left)) 
		nWidth	= rcWork.right - rcWork.left - nFrameWidth ;
	if (nWidth <= 0) {
		nWidth		= prcSrc->right - prcSrc->left ;
		nHeight		= prcSrc->bottom - prcSrc->top ;
		goto	skip_calc ;
	}

	pMyCand	= m_pDoc->pGetStatusText () ;
	iStyle	= (pMyCand != NULL)? pMyCand->m_iStyle : IMECANDSTYLE_UNUSED ;
	switch (iStyle) {
	case	IMECANDSTYLE_READ:
		{
			WCHAR	bufText [1024] ;
			int		nText ;
			SIZE	sz ;

			nText	= m_pDoc->iGetCandidateText (bufText, MYARRAYSIZE (bufText), NULL, NULL, pMyCand) ;
			if (nText > 0) {
				GetTextExtentPoint32W (hDC, bufText, nText, &sz) ;
				if ((sz.cx + nFrameWidth) > nWidth) {
					int	n	= nWidth - nFrameWidth ;
					nHeight	= (sz.cx + n - 1) / n * sz.cy + nFrameHeight ;
				} else {
					nHeight	= sz.cy + nFrameHeight ;
				}
				nHeight	= (nHeight < nDefaultHeight)? nDefaultHeight : nHeight ;
			} else {
				nHeight	= nDefaultHeight ;
			}
		}
		break ;
	case	IMECANDSTYLE_CODE0:
	case	IMECANDSTYLE_CODE1:
		nHeight	= nDefaultHeight ;
		break ;	

	case	IMECANDSTYLE_MINIBUFFERTEXT:
	default:
		{
			LPCWSTR		pwSText ;
			int			nSTextLen, iCursorPos ;
			SIZE		sz ;

			if (pMyCand != NULL && pMyCand->m_iCount == 1 && iStyle == IMECANDSTYLE_MINIBUFFERTEXT) {
				if (! m_pDoc->bGetStatusCursor (&iCursorPos))
					iCursorPos	= -1 ;
				pwSText		= (LPCWSTR)pMyCand->m_vbufCandidate.pGetBuffer () + *(pMyCand->m_vbufCandidateIndex.pGetBuffer () + 0) ;
				nSTextLen	= lstrlenW (pwSText) ;
			} else {
				if (! m_pDoc->bHaveMessagep ()) 
					break ;
				pwSText		= m_pDoc->pGetMessage (&nSTextLen) ;
				iCursorPos	= -1 ;
			}
			/* Minibuffer Text ̏ꍇB*/
			if (_GetStatusTextSize (hDC, pwSText, nSTextLen, iCursorPos, nWidth - nFrameWidth, &sz)) {
				nWidth	= sz.cx + nFrameWidth ;
				nHeight	= sz.cy + nFrameHeight ;
			}
		}
		break ;
	}

	/*	\ʒu̒BMulti-Screen lāB
	 */
skip_calc:
	{
		POINT	pt, ptOffset, ptBase ;

		pt.x	= prcSrc->left ;
		pt.y	= prcSrc->bottom + 1 ;
		if (rcWork.right > 0) {
			ptOffset.x	= pt.x % rcWork.right ;
			if (ptOffset.x < 0)
				ptOffset.x	= rcWork.right + ptOffset.x ;
			ptBase.x	= pt.x - ptOffset.x ;
			if ((ptOffset.x + nWidth) > rcWork.right){
				ptOffset.x	= rcWork.right - nWidth ;
			if (ptOffset.x < 0)
				ptOffset.x	= 0 ;
			}
			pt.x		= ptBase.x + ptOffset.x ;
		}
		if (rcWork.bottom > 0) {
			ptOffset.y	= pt.y % rcWork.bottom ;
			if (ptOffset.y < 0)
				ptOffset.y	= rcWork.bottom + ptOffset.y ;
			ptBase.y	= pt.y - ptOffset.y ;
			if ((ptOffset.y + nHeight) > rcWork.bottom){
				ptOffset.y = rcWork.bottom - nHeight ;
			}
			if (ptOffset.y < 0)
				ptOffset.y	= 0 ;
			pt.y		= ptBase.y + ptOffset.y ;
		}
		rcUI.left	= pt.x ;
		rcUI.top	= pt.y ;
		rcUI.right	= nWidth ;
		rcUI.bottom	= nHeight ;
	}

	DEBUGPRINTF ((TEXT ("View(l:%d, t:%d, r:%d, b:%d), WorkArea(l:%d, t:%d, r:%d, b:%d), w:%d, h:%d\n"),
				  prcSrc->left, prcSrc->top, prcSrc->right, prcSrc->bottom,
				  rcWork.left, rcWork.top, rcWork.right, rcWork.bottom,
				  nFontWidth, nFontHeight)) ;
	if (prcDest != NULL)
		*prcDest	= rcUI ;
	DeleteObject (SelectObject (hDC, hOldFont)) ;
	DeleteDC (hDC) ;
	return	S_OK ;
}

BOOL
CSkkImeMgr::_GetStatusTextSize (
	HDC						hDC,
	LPCWSTR					pwSText,
	int						nSTextLen,
	int						iCursorPos,
	int						nDefaultWidth,
	SIZE*					pSZ)
{
	/*	Key 1 ɕt ``X:'' 2ASeparator ̋󔒂 1 A
	 *	̕\ 1 ` 2 A
	 *	āANELEMENTS (pCandidateKey) * (2 + 1 + 1) ܂ (2 + 1 + 2)
	 *	B
	 */
	RECT			rc ;
	int				x, y, nDY ;
	TEXTMETRIC		tm ;
	BOOL			bCalculate ;
	int				iCalculateCount ;

	rc.left			= 0 ;
	rc.top			= 0 ;
	rc.right		= nDefaultWidth ;
	iCalculateCount	= 3 ;

	/* Minibuffer Window ̑傫𓾂B*/
	GetTextMetrics (hDC, &tm) ;
	nDY			= tm.tmHeight ;

	do {
		LPCWSTR		pText ;
		int			nText, nTextPos ;
		BOOL		bCursorPos ;

		pText		= pwSText ;
		nText		= nSTextLen ;
		nTextPos	= 0 ;
		y			= rc.top ;
		x			= rc.left ;
		bCursorPos	= FALSE ;
		bCalculate	= FALSE ;
		while (0 < nText) {
			int		nChar, iSX ;
			BOOL	bCRLF ;
			SIZE	sz ;

			bCRLF	= FALSE ;
			iSX		= x ;
			for (nChar = 1 ; nChar <= nText ; nChar ++) {
				if ((nTextPos + nChar - 1) == iCursorPos && *(pText + nChar - 1) == L'|') {
					break ;	/* cursor */
				}
				GetTextExtentPointW (hDC, pText, nChar, &sz) ;
				if ((x + sz.cx) >= rc.right) {
					if (nChar == 1 && x == rc.left) {
						/*	width ύXčŏB
						 *	A[vɓȂ悤ɍővZ񐔂
						 *	ݒ肵ĂB
						 */
						rc.right	= rc.left + sz.cx + 1 ;
						if (iCalculateCount <= 0)
							return	FALSE ;
						bCalculate	= TRUE ;
						iCalculateCount	-- ;
					}
					bCRLF	= TRUE ;
					break ;
				}
			}
			nChar	-- ;
			if (nChar <= 0)
				break ;
			/* eLXg\B*/
			GetTextExtentPointW (hDC, pText, nChar, &sz) ;
			x	+= sz.cx ;
			if (bCRLF) {
				/*	܂ԂL\B*/
				GetTextExtentPointW (hDC, L"\\", 1, &sz) ;
				x	+= sz.cx ;
			}

			/* KvȂJ[\\B*/
			/*if (nTextPos <= iCursorPos && iCursorPos <= (nTextPos + nChar)){*/
			if ((nTextPos + nChar) == iCursorPos) {
				/* J[\̒O܂ŕ`悳ĂԁB*/
				if ((nChar + 1) >= nText) {
				} else {
					bCursorPos		= TRUE ;
				}
				nChar	++ ;	/* skip: J[\ʒuɑ}Ă``|'' */
			} else {
				if (bCursorPos) {
					bCursorPos	= FALSE ;
				}
			}
			pText		+= nChar ;
			nTextPos	+= nChar ;
			nText		-= nChar ;
			if (bCRLF) {
				y		+= nDY ;
				x		= rc.left ;
			}
		}
	}	while (bCalculate) ;

	if (pSZ != NULL) {
		pSZ->cx	= (y > rc.top || iCursorPos != -1)? rc.right : x ;
		pSZ->cy	= y + nDY ;
	}
	return	TRUE ;
}

/*========================================================================*
 */
BOOL
CSkkImeTextService::_InitSkkImeMgr ()
{
    BOOL		fThreadfocus ;
    ITfSource*	pSource	= NULL ;

	if (m_pConfig == NULL) {
		m_pConfig	= CImeConfig::pCreateInstance () ;
	} else {
		m_pConfig->vUpdate () ;
	}

	m_pSkkIme	= new CSkkImeMgr (this) ;
	if (m_pSkkIme == NULL)
		return	FALSE ;

	if (! m_pSkkIme->_Init ())
		goto	ExitError ;

    // we also need a thread focus sink
	if (m_pThreadMgr != NULL) {
		if (m_pThreadMgr->QueryInterface(IID_ITfSource, (void **)&pSource) != S_OK) {
			pSource	= NULL ;
			goto	ExitError ;
		}

		if (pSource->AdviseSink(IID_ITfThreadFocusSink, (ITfThreadFocusSink *)this, &m_dwThreadFocusSinkCookie) != S_OK) {
			// make sure we don't try to Unadvise m_dwThreadFocusSinkCookie later
			m_dwThreadFocusSinkCookie	= TF_INVALID_COOKIE ;
			goto	ExitError ;
		}
		pSource->Release () ;
		pSource	= NULL ;

		// we may need to display the snoop window right now
		// our thread focus sink won't be called until something changes,
		// so we need to check the current state.

		if (m_pThreadMgr->IsThreadFocus(&fThreadfocus) == S_OK && fThreadfocus) {
			OnSetThreadFocus () ;
		}
	}
    return	TRUE ;

ExitError:
    SafeRelease(pSource);
    _UninitSkkImeMgr () ;
    return FALSE;
}

void
CSkkImeTextService::_UninitSkkImeMgr ()
{
    ITfSource*	pSource ;

    if (m_pSkkIme != NULL) {
        m_pSkkIme->_Uninit () ;
		m_pSkkIme->Release () ;
		m_pSkkIme	= NULL ;
    }
    if (m_dwThreadFocusSinkCookie != TF_INVALID_COOKIE) {
		if (m_pThreadMgr != NULL) {
			if (m_pThreadMgr->QueryInterface(IID_ITfSource, (void **)&pSource) == S_OK) {
				pSource->UnadviseSink(m_dwThreadFocusSinkCookie);
				pSource->Release();
			}
		}
        m_dwThreadFocusSinkCookie = TF_INVALID_COOKIE;
    }
	return ;
}

/*========================================================================
 *	....
 */

HFONT
_CheckNativeCharset (
	HDC	hDC) 
{
	BOOL		bDiffCharSet	= FALSE ;
	HFONT		hOldFont ;
	LOGFONT	lfFont ;

	hOldFont	= (HFONT) GetCurrentObject (hDC, OBJ_FONT) ;
	GetObject (hOldFont, sizeof(LOGFONT), &lfFont) ;

	if (lfFont.lfCharSet != SHIFTJIS_CHARSET/*NATIVE_CHARSET*/) {
		bDiffCharSet			= TRUE ;
		lfFont.lfWeight			= FW_NORMAL ;
		lfFont.lfCharSet		= SHIFTJIS_CHARSET ; //NATIVE_CHARSET
		lfFont.lfFaceName[0]	= TEXT ('\0') ;
		SelectObject (hDC, CreateFontIndirect (&lfFont)) ;
	} else {
		hOldFont				= NULL ;
	}
	return	hOldFont ;
}


