/* # skkinput (Simple Kana-Kanji Input)
 *
 * This file is part of skkinput.
 * Copyright (C) 2002
 * Takashi SAKAMOTO (PXG01715@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "local.h"
#include <stdio.h>
#include "lispmgrp.h"
#include "cstring.h"
#include "kanji.h"

#define	lispEntity_GetBufferPtr(ptr)	((TLispBuffer *)((TLispEntity *)(ptr) + 1))

static	BOOL			lispBuffer_initialize	(TLispManager*, TLispEntity*) ;
static	BOOL			lispBuffer_uninitialize (TLispManager*, TLispEntity*) ;
static	TLispEntity*	lispBuffer_getMarker	(TLispEntity*, int) ;
static	BOOL			lispBuffer_getMarkerPosition (TLispManager*, TLispEntity*, int, int*) ;

/*	buffer entity 쐬B
 */
BOOL
lispMgr_CreateBuffer (
	register TLispManager*			pLispMgr,
	register TLispEntity** const	ppEntityRet)
{
	TLispEntity*			pEntity ;
	register TLispBuffer*	pBuffer ;
	register int			i ;
	TLispEntity**			ppEntSymbol ;
	int						nEntSymbol ;
	TLispEntity*			pEntEmpty ;

	ASSERT (pLispMgr    != NULL) ;
	ASSERT (ppEntityRet != NULL) ;

#if defined (DEBUG_LV99)
	fprintf (stderr, "Garbage collecting ...") ;
	fflush (stderr) ;
	lispMgr_CollectGarbage (pLispMgr) ;
	fprintf (stderr, "done.\n") ;
#endif
	/*	GeBeBƂĕKvȃ̈mۂB*/
	if (TFAILED (lispMgr_AllocateEntity (pLispMgr, sizeof (TLispBuffer), &pEntity))) 
		return	FALSE ;

	/*	GeBeBƂďB*/
	pEntity->m_iType	= LISPENTITY_BUFFER ;
	pBuffer				= lispEntity_GetBufferPtr (pEntity) ;

	/*	obt@[J̃V{͂ȂB*/
	for (i = 0 ; i < SIZE_LISP_BIND_TABLE ; i ++) {
		pBuffer->m_apVariableTable [i]	= NULL ;
	}
	/*	OXgɂȂĂB*/
	pBuffer->m_pNextBuffer		= pEntity ;
	pBuffer->m_pPrevBuffer		= pEntity ;
	pBuffer->m_pEntKeymap		= NULL ;
	pBuffer->m_pEntName			= NULL ;
	pBuffer->m_fModified		= FALSE ;
	pBuffer->m_nTick			= 0 ;
	
	/*	obt@ɂ͈؂̃}[JȂĂȂB*/
	pBuffer->m_lstMarkers		= NULL ;
	for (i = 0 ; i < NUM_BUFFER_MUST_MARKERS ; i ++) 
		pBuffer->m_apMustMarkers [i]	= NULL ;

	/*	GeBeBXgɂȂB*/
	(void) lispMgr_RegisterMisc (pLispMgr, pEntity) ;

	/*	obt@B*/
	if (TFAILED (lispBuffer_initialize (pLispMgr, pEntity))) {
		lispMgr_DestroyEntity (pLispMgr, pEntity) ;
		return	FALSE ;
	}

	/*	ftHg̃[Jϐ쐬Bl͂ȂB*/
	(void) lispMgr_GetLocalSymbols (pLispMgr, &ppEntSymbol, &nEntSymbol) ;
	(void) lispMgr_CreateEmpty (pLispMgr, &pEntEmpty) ;
	while (nEntSymbol -- > 0) {
#if defined (DEBUG)
		fprintf (stderr, "Symbol(") ;
		lispEntity_Print (pLispMgr, *ppEntSymbol) ;
		fprintf (stderr, ") <-- nil\n") ;
#endif
		lispBuffer_MakeSymbolValue (pLispMgr, pEntity, *ppEntSymbol) ;
		lispBuffer_SetSymbolValue  (pLispMgr, pEntity, *ppEntSymbol, pEntEmpty) ;
		ppEntSymbol	++ ;
	}

	*ppEntityRet	= pEntity ;
	return	TRUE ;
}

void
lispMgr_DestroyBuffer (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer)
{
	register TLispBuffer*	pBuffer ;

	ASSERT (pEntBuffer     != NULL) ;
#if defined (DEBUG)
	fprintf (stderr, "lispMgr_DestroyBuffer (%p, %p)\n", pLispMgr, pEntBuffer) ;
#endif
	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (pBuffer->m_pEntKeymap != NULL)
		lispEntity_Release (pLispMgr, pBuffer->m_pEntKeymap) ;
	pBuffer->m_pEntKeymap	= NULL ;

	/*	}[JĂKvB*/
	lispBuffer_uninitialize (pLispMgr, pEntBuffer) ;
	return ;
}

/*
 *[@\]
 *	obt@Ԃɖ߂BȂ킿c
 *	obt@ŕҏWĂSďAobt@[J
 *	ϐSĉA}[JԂɖ߂B
 */
BOOL
lispBuffer_Clear (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer)
{
	if (TFAILED (lispBuffer_uninitialize (pLispMgr, pEntBuffer)) ||
		TFAILED (lispBuffer_initialize   (pLispMgr, pEntBuffer)))
		return	FALSE ;
	return	TRUE ;
}

BOOL
lispBuffer_SetNext (
	register TLispEntity* pEntBuffer,
	register TLispEntity* pEntNextBuffer)
{
	register TLispBuffer*	pBuffer ;

	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	ASSERT ((pEntNextBuffer != NULL &&
			 pEntNextBuffer->m_iType == LISPENTITY_BUFFER) ||
			(pEntNextBuffer == NULL)) ;

	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	pBuffer->m_pNextBuffer	= pEntNextBuffer ;
	return	TRUE ;
}

BOOL
lispBuffer_GetNext (
	register TLispEntity*			pEntBuffer,
	register TLispEntity** const	ppReturn)
{
	register TLispBuffer*	pBuffer ;

	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	ASSERT (ppReturn != NULL) ;

	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	*ppReturn	= pBuffer->m_pNextBuffer ;
	return	TRUE ;
}

BOOL
lispBuffer_SetPrevious (
	register TLispEntity*	pEntBuffer,
	register TLispEntity*	pEntPrevBuffer)
{
	register TLispBuffer*	pBuffer ;

	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	ASSERT ((pEntPrevBuffer != NULL &&
			 pEntPrevBuffer->m_iType == LISPENTITY_BUFFER) ||
			(pEntPrevBuffer == NULL)) ;

	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	pBuffer->m_pPrevBuffer	= pEntPrevBuffer ;
	return	TRUE ;
}

BOOL
lispBuffer_GetPrevious (
	register TLispEntity*			pEntBuffer,
	register TLispEntity** const	ppReturn)
{
	register TLispBuffer*	pBuffer ;

	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	ASSERT (ppReturn != NULL) ;

	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	*ppReturn	= pBuffer->m_pPrevBuffer ;
	return	TRUE ;
}

BOOL
lispBuffer_MakeSymbolValue (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register TLispEntity*	pSymbol)
{
	register TLispBuffer*	pBuffer ;

	ASSERT (pLispMgr != NULL) ;
	ASSERT (pEntBuffer  != NULL) ;
	ASSERT (pSymbol  != NULL) ;
	
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;

	return	lispBindTable_MakeEntry (pLispMgr, pBuffer->m_apVariableTable, ARRAYSIZE (pBuffer->m_apVariableTable), pSymbol, NULL) ;
}

/*
 *	ɃV{킯ɂ͂ȂcB
 */
BOOL
lispBuffer_SetSymbolValue (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register TLispEntity*	pSymbol,
	register TLispEntity*	pValue)
{
	register TLispBuffer*	pBuffer ;

	ASSERT (pLispMgr   != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pSymbol    != NULL) ;
	/*ASSERT (hValue  != NULL) ;*/

	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;

	return	lispBindTable_SetEntryValue (pLispMgr, pBuffer->m_apVariableTable, ARRAYSIZE (pBuffer->m_apVariableTable), pSymbol, pValue) ;
}

BOOL
lispBuffer_SetSymbolValueWithName (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register const Char*	pName,
	register const int		nName,
	register TLispEntity*	pValue)
{
	register TLispBuffer*	pBuffer ;
	TLispEntity*	pEntSymbol ;

	ASSERT (pLispMgr   != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pName != NULL && nName > 0) ;
	/*ASSERT (hValue  != NULL) ;*/

	if (TFAILED (lispMgr_InternSymbol (pLispMgr, pName, nName, &pEntSymbol)))
		return	FALSE ;

	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;

	return	lispBindTable_SetEntryValue (pLispMgr, pBuffer->m_apVariableTable, ARRAYSIZE (pBuffer->m_apVariableTable), pEntSymbol, pValue) ;
}

BOOL
lispBuffer_GetSymbolValue (
	register TLispManager*			pLispMgr,
	register TLispEntity*			pEntBuffer,
	register TLispEntity*			pSymbol,
	register TLispEntity** const	ppReturn)
{
	register TLispBuffer*	pBuffer ;

	ASSERT (pLispMgr   != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pSymbol    != NULL) ;
	/*ASSERT (hValue  != NULL) ;*/

	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	return	lispBindTable_GetEntryValue (pLispMgr, pBuffer->m_apVariableTable, ARRAYSIZE (pBuffer->m_apVariableTable), pSymbol, ppReturn) ;
}

BOOL
lispBuffer_GetSymbolValueWithName (
	register TLispManager*			pLispMgr,
	register TLispEntity*			pEntBuffer,
	register const Char*			pName,
	register const int				nName,
	register TLispEntity** const	ppReturn)
{
	register TLispBuffer*	pBuffer ;
	TLispEntity*	pEntSymbol ;

	ASSERT (pLispMgr   != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pName != NULL && nName > 0) ;
	/*ASSERT (hValue  != NULL) ;*/

	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;

	if (TFAILED (lispMgr_InternSymbol (pLispMgr, pName, nName, &pEntSymbol)))
		return	FALSE ;

	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	return	lispBindTable_GetEntryValue (pLispMgr, pBuffer->m_apVariableTable, ARRAYSIZE (pBuffer->m_apVariableTable), pEntSymbol, ppReturn) ;
}


/*
 *	}[Jobt@̊ǗɒuB̗ŔAobt@j
 *	̃obt@w}[Ĵ݂͂܂B
 *	̃obt@wĂ}[J͔j悤ɂȂƂȂ
 *	̂ł͂Ȃ낤AƁB
 *	ƁÃobt@ɑĂ}[J͑}Ƃ̑ňړȂ
 *	ƂȂB
 */
BOOL
lispBuffer_AddMarker (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register TLispEntity*	pEntMarker)
{
	TLispEntity*	pEntPrevBuffer ;
	register TLispBuffer*	pBuffer ;
	
	ASSERT (pLispMgr      != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntMarker != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;

	/*	}[JƂĂ̓𔲂oB*/
	if (TFAILED (lispMarker_GetBufferPosition (pLispMgr, pEntMarker, &pEntPrevBuffer, NULL))) {
#if defined (DEBUG)
		fprintf (stderr, "[buffer] This marker has no buffer.\n") ;
#endif
		return	FALSE ;
	}

	/*	ɂ̃obt@ɂȂĂꍇɂ͖ȂB*/
	if (pEntPrevBuffer == pEntBuffer) {
#if defined (DEBUG)
		fprintf (stderr, "[buffer] This marker has already been inserted.\n") ;
#endif
		return	TRUE ;
	}
	if (pEntPrevBuffer != NULL) 
		lispBuffer_RemoveMarker (pLispMgr, pEntMarker) ;
	/*	}[J̃obt@ƈʒuύXB*/
	lispMarker_SetBufferPosition (pLispMgr, pEntMarker, pEntBuffer, 1) ;
	
	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	/*	obt@̃}[JNȂB*/
	if (pBuffer->m_lstMarkers == NULL) {
		pBuffer->m_lstMarkers	= pEntMarker ;
		lispMarker_SetNext (pLispMgr, pEntMarker, NULL) ;
	} else {
		lispMarker_SetPrevious (pLispMgr, pBuffer->m_lstMarkers, pEntMarker) ;
		lispMarker_SetNext (pLispMgr, pEntMarker, pBuffer->m_lstMarkers) ;
		pBuffer->m_lstMarkers	= pEntMarker ;
	}
	lispMarker_SetPrevious (pLispMgr, pEntMarker, NULL) ;
	return	TRUE ;
}

BOOL
lispBuffer_RemoveMarker (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntMarker)
{
	TLispEntity*	pEntBuffer ;
	TLispEntity*	pEntPrevMarker ;
	TLispEntity*	pEntNextMarker ;
	register TLispBuffer*	pBuffer ;
	
	ASSERT (pLispMgr      != NULL) ;
	ASSERT (pEntMarker != NULL) ;

	/*	}[JƂĂ̓𔲂oB*/
	if (TFAILED (lispMarker_GetBufferPosition (pLispMgr, pEntMarker, &pEntBuffer, NULL)) ||
		pEntBuffer == NULL)
		return	FALSE ;
	
	/*	}[J̃obt@ƈʒuύXB*/
	lispMarker_SetBufferPosition (pLispMgr, pEntMarker, NULL, 0) ;
	
	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	/*	obt@̃}[JNȂB*/
	lispMarker_GetPrevious (pLispMgr, pEntMarker, &pEntPrevMarker) ;
	lispMarker_GetNext (pLispMgr, pEntMarker, &pEntNextMarker) ;
	
	if (pEntPrevMarker != NULL) {
		lispMarker_SetNext (pLispMgr, pEntPrevMarker, pEntNextMarker) ;
	} else {
		pBuffer->m_lstMarkers	= pEntNextMarker ;
	}
	if (pEntNextMarker != NULL) {
		lispMarker_SetPrevious (pLispMgr, pEntNextMarker, pEntPrevMarker) ;
	}
	return	TRUE ;
}

BOOL
lispBuffer_SetKeymap (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register TLispEntity*	pEntKeymap)
{
	register TLispBuffer*	pBuffer ;
	ASSERT (pLispMgr       != NULL) ;
	ASSERT (pEntBuffer     != NULL) ;
	ASSERT (pEntKeymap     != NULL) ;
	lispEntity_AddRef (pLispMgr, pEntKeymap) ;
	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (pBuffer->m_pEntKeymap != NULL)
		lispEntity_Release (pLispMgr, pBuffer->m_pEntKeymap) ;
	pBuffer->m_pEntKeymap	= pEntKeymap ;
	return	TRUE ;
}

BOOL
lispBuffer_GetKeymap (
	register TLispEntity*	pEntBuffer,
	register TLispEntity**	ppEntKeymapRet)
{
	register TLispBuffer*	pBuffer ;
	ASSERT (pEntBuffer     != NULL) ;
	ASSERT (ppEntKeymapRet != NULL) ;
	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	*ppEntKeymapRet	= pBuffer->m_pEntKeymap ;
	return	TRUE ;
}

BOOL
lispBuffer_SetPrompt (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register const Char*	pString,
	register int			nString)
{
	register TLispBuffer*	pBuffer ;
	int						nBufferTop, nBufferEnd ;
	TBufStringMarker		mk ;

	ASSERT (pLispMgr   != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pString    != NULL || nString == 0) ;

	if (nString <= 0)
		return	TRUE ;

	lispBuffer_PointBufferTop (pLispMgr, pEntBuffer, &nBufferTop) ; 
	lispBuffer_PointBufferEnd (pLispMgr, pEntBuffer, &nBufferEnd) ;

	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (TFAILED (TBufString_SolveMarker (&pBuffer->m_Text, nBufferTop - 1, &mk))) {
#if defined (DEBUG)
		fprintf (stderr, "lispBuffer_SetPrompt () : SolveMarker error!\n") ;
#endif
		return	FALSE ;
	}
	if (TFAILED (TBufStringMarker_InsertString (&mk, pString, nString))) {
#if defined (DEBUG)
		fprintf (stderr, "lispBuffer_SetPrompt () : InsertString error!\n") ;
#endif
		return	FALSE ;
	}
	/*	J[\ړB*/
	lispBuffer_MoveMarkerWithInsertion (pLispMgr, pEntBuffer, nBufferTop - 1, nString) ;
	pBuffer->m_nTick	++ ;
	return	TRUE ;
}

/*
 *	}[J̃J[\̈ʒuԂB
 */
BOOL
lispBuffer_Point (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int*			pnPoint)
{
	return	lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_POINT, pnPoint) ;
}

BOOL
lispBuffer_PointBufferEnd (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int*			pnPoint)
{
	return	lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BLAST, pnPoint) ;
}

/*
 *	ҏW\ȗ̈̐擪ʒuԂB
 */
BOOL
lispBuffer_PointMax (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int*			pnPoint)
{
	return	lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_ELAST, pnPoint) ;
}

BOOL
lispBuffer_PointBufferTop (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int*			pnPoint)
{
	return	lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BTOP, pnPoint) ;
}

/*
 *	ҏW\ȗ̈̍ŏIʒuԂB
 */
BOOL
lispBuffer_PointMin (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int*			pnPoint)
{
	return	lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_ETOP, pnPoint) ;
}

BOOL
lispBuffer_PointMarker (
	register TLispManager*			pLispMgr,
	register TLispEntity*			pEntBuffer,
	register TLispEntity** const	ppMarkerReturn)
{
	ASSERT (pLispMgr       != NULL) ;
	ASSERT (pEntBuffer  != NULL) ;
	ASSERT (ppMarkerReturn != NULL) ;
	*ppMarkerReturn	= lispBuffer_getMarker (pEntBuffer, MARKER_POINT) ;
	return	TRUE ;
	UNREFERENCED_PARAMETER (pLispMgr) ;
}

/*
 *	}[N̈ʒuԂB
 */
BOOL
lispBuffer_MarkMarker (
	register TLispManager*			pLispMgr,
	register TLispEntity*			pEntBuffer,
	register TLispEntity** const	ppMarkerReturn)
{
	ASSERT (pLispMgr       != NULL) ;
	ASSERT (pEntBuffer  != NULL) ;
	ASSERT (ppMarkerReturn != NULL) ;
	*ppMarkerReturn	= lispBuffer_getMarker (pEntBuffer, MARKER_MARK) ;
	return	TRUE ;
	UNREFERENCED_PARAMETER (pLispMgr) ;
}

BOOL
lispBuffer_SetName (
	register TLispManager*		pLispMgr,
	register TLispEntity*		pEntBuffer,
	register TLispEntity*		pEntName)
{
	register TLispBuffer*	pBuffer ;
	ASSERT (pLispMgr   != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;

	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (pBuffer->m_pEntName != NULL) 
		lispEntity_Release (pLispMgr, pBuffer->m_pEntName) ;
	pBuffer->m_pEntName	= pEntName ;
	if (pEntName != NULL)
		lispEntity_AddRef (pLispMgr, pEntName) ;
	return	TRUE ;
}

BOOL
lispBuffer_GetName (
	register TLispManager*		pLispMgr,
	register TLispEntity*		pEntBuffer,
	register TLispEntity**		ppEntReturn)
{
	register TLispBuffer*	pBuffer ;
	ASSERT (pLispMgr   != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;

	pBuffer			= lispEntity_GetBufferPtr (pEntBuffer) ;
	*ppEntReturn	= pBuffer->m_pEntName ;
	return	TRUE ;
	UNREFERENCED_PARAMETER (pLispMgr) ;
}

/*
 *	obt@ɓ/ҏWĂeLXg𔲂o߂̃}[J𓾂B
 *	ǃANZXɂ̓}[JKvȂ̂ŁB
 */
BOOL
lispBuffer_GetString (
	register TLispManager*		pLispMgr,
	register TLispEntity*		pEntBuffer,
	register TBufStringMarker*	pMarker,
	register int* const			pnLength)
{
	register TLispBuffer*	pBuffer ;
	int				nTop, nLast ;

	ASSERT (pLispMgr      != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pMarker       != NULL) ;
	ASSERT (pnLength      != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	
	/*	obt@̐擪Ō܂ł𕶎ƂĔoBnarrowing 
	 *	ʂɒӂ邱ƁB*/
	if (TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BLAST, &nLast)) ||
		TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BTOP,  &nTop)))
		return	FALSE ;
	ASSERT (nTop > 0 && nLast > 0) ;
	ASSERT (nTop <= nLast) ;
	
	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (TFAILED (TBufString_SolveMarker (&pBuffer->m_Text, nTop - 1, pMarker)))
		return	FALSE ;
	*pnLength	= nLast - nTop ;
	return	TRUE ;
}

/*
 *	obt@̕ҏW\łȂGAɊ܂܂镶(vvgƂ)
 *	܂񂾃g[^̕ԂBNarrowing ̉e͓R󂯂ȂB
 *()
 *	obt@̍ŌɉĂꍇɂ͍̎ł͑Ήł
 *	Bg[^̕JEgĂȂ̂ŁB
 */
BOOL
lispBuffer_GetFullString (
	register TLispManager*		pLispMgr,
	register TLispEntity*		pEntBuffer,
	register TBufStringMarker*	pMarker,
	register int* const			pnLength)
{
	register TLispBuffer*	pBuffer ;
	int				nLast ;

	ASSERT (pLispMgr      != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pMarker       != NULL) ;
	ASSERT (pnLength      != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	
	/*	obt@̐擪Ō܂ł𕶎ƂĔoBnarrowing 
	 *	ʂɒӂ邱ƁB*/
	if (TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BLAST, &nLast)))
		return	FALSE ;
	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (TFAILED (TBufString_SolveMarker (&pBuffer->m_Text, 0, pMarker)))
		return	FALSE ;
	*pnLength	= nLast - 1 ;	/* marker  1 炾 - 1 B*/
	return	TRUE ;
}

/*
 *	obt@̕ҏW\̈ nStartPos  nEndPos Ɍ肷B
 *	Aobt@̂ƂƂ̑傫
 */
BOOL
lispBuffer_Narrow (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int			nStartPos,
	register int			nEndPos)
{
	register TLispBuffer*	pBuffer ;
	int				nBufferTop, nBufferLast ;
	
	ASSERT (pLispMgr      != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	
	/*	narrowing ̈̊Jnʒu͏Iʒu菬ė~B*/
	if (nStartPos > nEndPos)
		return	FALSE ;

	/*	obt@͈̔͂𓾂B*/
	if (TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BTOP,  &nBufferTop)) ||
		TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BLAST, &nBufferLast)))
		return	FALSE ;
	/*	narrowing ͈͂obt@̊OG[ɂB*/
	if (nStartPos < nBufferTop || nEndPos > nBufferLast)
		return	FALSE ;
	
	/*	ҏW\̈킷}[JړB*/
	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (TFAILED (lispMarker_SetBufferPosition (pLispMgr, pBuffer->m_apMustMarkers [MARKER_ETOP],  pEntBuffer, nStartPos)) ||
		TFAILED (lispMarker_SetBufferPosition (pLispMgr, pBuffer->m_apMustMarkers [MARKER_ELAST], pEntBuffer, nEndPos)))
		return	FALSE ;
	pBuffer->m_nTick	++ ;
	return	TRUE ;
}

BOOL
lispBuffer_Widen (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer)
{
	register TLispBuffer*	pBuffer ;
	int				nBufferTop, nBufferLast ;
	
	ASSERT (pLispMgr      != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	
	/*	obt@͈̔͂𓾂B*/
	if (TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BTOP,  &nBufferTop)) ||
		TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BLAST, &nBufferLast)))
		return	FALSE ;

	/*	ҏW\̈킷}[Jobt@̐擪ƍŌɈړB*/
	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (TFAILED (lispMarker_SetBufferPosition (pLispMgr, pBuffer->m_apMustMarkers [MARKER_ETOP],  pEntBuffer, nBufferTop)) ||
		TFAILED (lispMarker_SetBufferPosition (pLispMgr, pBuffer->m_apMustMarkers [MARKER_ELAST], pEntBuffer, nBufferLast)))
		return	FALSE ;
	pBuffer->m_nTick	++ ;
	return	TRUE ;
}

BOOL
lispBuffer_InsertFileContents (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int			nPosition,
	register KFILE*			pFile,
	register int*			pnReadCount)
{
	register TLispBuffer*	pBuffer ;
	int					nEditLast, nEditTop, nBufferLast ;
	TBufStringMarker	mk ;
	register Char		cc ;
	register int		nReadCount	= 0 ;

	ASSERT (pLispMgr   != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	ASSERT (pFile      != NULL) ;

	if (TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_ETOP,  &nEditTop)) ||
		TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_ELAST, &nEditLast)) ||
		TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BLAST, &nBufferLast)) ||
		nEditTop > nPosition || nPosition > nEditLast) {
#if defined (DEBUG)
		fprintf (stderr, "[buffer] Bad Marker Position.\n") ;
#endif
		return	FALSE ;
	}

	ASSERT (nEditTop  <= nEditLast) ;
	ASSERT (nEditLast <= nBufferLast) ;

	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (TFAILED (TBufString_SolveMarker (&pBuffer->m_Text, nPosition - 1, &mk)))
		return	FALSE ;

	while (cc = KFile_Getc (pFile), cc != EOF) {
		if (TFAILED (TBufStringMarker_InsertChar (&mk, cc, 1)))
			break ;
		nReadCount	++ ;
	}

	/*	}[JړKvB*/
	lispBuffer_MoveMarkerWithInsertion (pLispMgr, pEntBuffer, nPosition, nReadCount) ;
	pBuffer->m_fModified	= TRUE ;
	pBuffer->m_nTick	++ ;

	if (pnReadCount != NULL)
		*pnReadCount			= nReadCount ;
	return	(cc == EOF)? TRUE : FALSE ;
}

/*
 *	obt@ɕ nCount }B
 */
BOOL
lispBuffer_InsertChar (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int			nPosition,
	register const Char		cc,
	register int			nCount)
{
	register TLispBuffer*	pBuffer ;
	int				nEditLast, nEditTop, nBufferLast ;
	TBufStringMarker	mk ;
	
	ASSERT (pLispMgr      != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;

	if (nCount <= 0)
		return	TRUE ;
	
	if (TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_ETOP,  &nEditTop)) ||
		TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_ELAST, &nEditLast)) ||
		TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BLAST, &nBufferLast))) {
#if defined (DEBUG)
		fprintf (stderr, "[buffer] Bad Marker Position.\n") ;
#endif
		return	FALSE ;
	}
	
	ASSERT (nEditTop  <= nEditLast) ;
	ASSERT (nEditLast <= nBufferLast) ;
	
	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (nEditTop <= nPosition && nPosition <= nEditLast) {
		if (TFAILED (TBufString_SolveMarker (&pBuffer->m_Text, nPosition - 1, &mk)))
			return	FALSE ;
		if (TFAILED (TBufStringMarker_InsertChar (&mk, cc, nCount)))
			return	FALSE ;

		/*	}[JړKvB*/
		lispBuffer_MoveMarkerWithInsertion (pLispMgr, pEntBuffer, nPosition, nCount) ;
		pBuffer->m_fModified	= TRUE ;
		pBuffer->m_nTick	++ ;
	}
	return	TRUE ;
}

/*
 *	 delete-char ̓obt@ narrow 
 */
BOOL
lispBuffer_DeleteChar (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int			nPosition,
	register int			nCount)
{
	register TLispBuffer*	pBuffer ;
	int						nEditTop, nEditLast ;
	TBufStringMarker	mk ;
	
	ASSERT (pLispMgr      != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (pEntBuffer->m_iType == LISPENTITY_BUFFER) ;
	
	if (TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_ETOP,  &nEditTop)) ||
		TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_ELAST, &nEditLast)))
		return	FALSE ;
	
	ASSERT (nEditTop  <= nEditLast) ;

	if (nEditTop > nPosition || (nPosition + nCount) > nEditLast)
		return	FALSE ;
	if (nCount <= 0)
		return	TRUE ;

	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (TFAILED (TBufString_SolveMarker (&pBuffer->m_Text, nPosition - 1, &mk)))
		return	FALSE ;
	if (TFAILED (TBufStringMarker_DeleteChar (&mk, nCount)))
		return	FALSE ;
	/*	}[JړKvB*/
	lispBuffer_MoveMarkerWithDeletion (pLispMgr, pEntBuffer, nPosition, nCount) ;
	pBuffer->m_fModified	= TRUE ;
	pBuffer->m_nTick	++ ;
	return	TRUE ;
}

BOOL
lispBuffer_InsertString (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int			nPosition,
	register const Char*	pString,
	register int			nLength)
{
	register TLispBuffer*	pBuffer ;
	int						nEditLast, nEditTop, nBufferLast ;
	TBufStringMarker		mk ;
	
	ASSERT (pLispMgr   != NULL) ;
	ASSERT (pEntBuffer != NULL) ;

	if (nLength <= 0)
		return	TRUE ;

	ASSERT (pString    != NULL) ;
	
	if (TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_ETOP,  &nEditTop)) ||
		TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_ELAST, &nEditLast)) ||
		TFAILED (lispBuffer_getMarkerPosition (pLispMgr, pEntBuffer, MARKER_BLAST, &nBufferLast)))
		return	FALSE ;
	
	ASSERT (nEditTop  <= nEditLast) ;
	ASSERT (nEditLast <= nBufferLast) ;
	
#if defined (DEBUG)
	lispBuffer_Print (pLispMgr, pEntBuffer) ;
#endif
	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	if (nEditTop <= nPosition && nPosition <= nEditLast) {
		if (TFAILED (TBufString_SolveMarker (&pBuffer->m_Text, nPosition - 1, &mk)))
			return	FALSE ;
		if (TFAILED (TBufStringMarker_InsertString (&mk, pString, nLength)))
			return	FALSE ;

		/*	}[JړKvB*/
		lispBuffer_MoveMarkerWithInsertion (pLispMgr, pEntBuffer, nPosition, nLength) ;
		pBuffer->m_fModified	= TRUE ;
		pBuffer->m_nTick	++ ;
	}
#if defined (DEBUG)
	lispBuffer_Print (pLispMgr, pEntBuffer) ;
#endif
	return	TRUE ;
}

/*	buffer  modified flag QƂB
 */
BOOL
lispBuffer_GetModifiedp (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer)
{
	register TLispBuffer*	pBuffer ;
	
	ASSERT (pLispMgr != NULL) ;
	ASSERT (pEntBuffer != NULL) ;

	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	return	pBuffer->m_fModified ;
	UNREFERENCED_PARAMETER (pLispMgr) ;
}

/*	buffer  modified flag ݒ肷Bmodified flag ͒ڂ͏CłȂB
 *	̊֐oR邱ƁB
 */
void
lispBuffer_SetModifiedp (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register BOOL			fModified)
{
	register TLispBuffer*	pBuffer ;
	
	ASSERT (pLispMgr != NULL) ;
	ASSERT (pEntBuffer != NULL) ;

	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	pBuffer->m_fModified	= fModified ;
	pBuffer->m_nTick	++ ;
	return ;
	UNREFERENCED_PARAMETER (pLispMgr) ;
}

/*
 *	obt@eLXgʒu nPosition  nCount }ꍇ
 *	}[J͕̓킷B
 *	ۂ̑}Ƃ͂ł͉lȂB
 */
BOOL
lispBuffer_MoveMarkerWithInsertion (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int			nPosition,
	register int			nCount)
{
	register TLispBuffer*	pBuffer ;
	register TLispEntity*	pEntMarker ;
	TLispEntity*			pEntNextMarker ;
#if defined (DEBUG_LV99)
	register int			i = 0 ;
#endif

	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	pEntMarker	= pBuffer->m_lstMarkers ;
	
#if defined (DEBUG_LV99)
	fprintf (stderr, "move-marker-with-insertion (%d, %d)\n", nPosition, nCount) ;
#endif
	while (pEntMarker != NULL) {
#if defined (DEBUG_LV99)
		fprintf (stderr, "(%d) ", i ++) ;
#endif
		lispMarker_MoveWithInsertion (pLispMgr, pEntMarker, nPosition, nCount) ;
		lispMarker_GetNext (pLispMgr, pEntMarker, &pEntNextMarker) ;
		pEntMarker	= pEntNextMarker ;
	}
	return	TRUE ;
}

/*
 *	obt@eLXgʒu nPosition  nCount 폜ꍇ
 *	}[J͕̓킷B
 *	ۂ̑}Ƃ͂ł͉lȂB
 */
BOOL
lispBuffer_MoveMarkerWithDeletion (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int			nPosition,
	register int			nCount)
{
	TLispEntity*	pEntMarker ;
	TLispEntity*	pEntNextMarker ;
	TLispBuffer*	pBuffer ;

	pBuffer			= lispEntity_GetBufferPtr (pEntBuffer) ;
	pEntMarker	= pBuffer->m_lstMarkers ;
	
	/*	obt@ɂȂĂ}[JԂɈړcB
	 *	̕@ł̓}[J̐ƂƂĂxcB
	 *	܂AQƂȂ}[J̓QVQV\Ȃ̂ŁcB*/
	while (pEntMarker != NULL) {
		lispMarker_MoveWithDeletion (pLispMgr, pEntMarker, nPosition, nCount) ;
		lispMarker_GetNext (pLispMgr, pEntMarker, &pEntNextMarker) ;
		pEntMarker	= pEntNextMarker ;
	}
	return	TRUE ;
}

/*	lbuffer.c */
int
lispBuffer_GetTick (
	register TLispEntity*	pEntBuffer)
{
	register TLispBuffer*	pBuffer ;
	ASSERT (pEntBuffer != NULL) ;
	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	return	pBuffer->m_nTick ;
}

/*
 *	fobOp̊֐B
 *	obt@̒gSɕ\B
 */
BOOL
lispBuffer_Print (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer)
{
	TLispEntity*		pEntMarker ;
	TLispEntity*		pEntNextMarker ;
	TLispBuffer*		pBuffer ;
	KANJISTATEMACHINE	ksm ;
	int					n, i, nLength ;
	char				achBuf [16] ;
	static char*		apMarkerName []	= {
		"BTOP", "ETOP", "POINT", "ELAST", "BLAST", "MARK",
	} ;
	TBufStringMarker	mk ;
	Char				cc ;

	ASSERT (pLispMgr   != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	fprintf (stderr, "buffer(%p) ---\n", pEntBuffer) ;
	pBuffer		= lispEntity_GetBufferPtr (pEntBuffer) ;
	pEntMarker	= pBuffer->m_lstMarkers ;
	for (i = 0 ; i < NUM_BUFFER_MUST_MARKERS ; i ++) {
		fprintf (stderr, "(%5s) ", apMarkerName [i]) ;
		if (pBuffer->m_apMustMarkers [i] == NULL) {
			fprintf (stderr, "no-exist\n") ;
		} else {
			lispMarker_Print (pLispMgr, pBuffer->m_apMustMarkers [i]) ;
			fprintf (stderr, "\n") ;
		}
	}
	i			= 0 ;
	while (pEntMarker != NULL) {
		fprintf (stderr, "(%d) ", i ++) ;
		lispMarker_Print (pLispMgr, pEntMarker) ;
		fprintf (stderr, "\n") ;
		lispMarker_GetNext (pLispMgr, pEntMarker, &pEntNextMarker) ;
		pEntMarker	= pEntNextMarker ;
	}

	if (TFAILED (TBufString_SolveMarker (&pBuffer->m_Text, 0, &mk)))
		return	FALSE ;

	lispBuffer_PointBufferEnd (pLispMgr, pEntBuffer, &nLength) ;
	nLength		-- ;
	fprintf (stderr, "buffer-string(%d) = \"", nLength) ;
	InitializeKanjiFiniteStateMachine (&ksm, KCODING_SYSTEM_SHIFTJIS) ;
	while (nLength > 0) {
		cc	= TBufStringMarker_GetChar (&mk) ;
		n	= RtransferKanjiFiniteStateMachine (&ksm, cc, achBuf) ;
		achBuf [n]	= '\0' ;
		fprintf (stderr, "%s", achBuf) ;
		TBufStringMarker_Forward (&mk, 1) ;
		nLength	-- ;
	}
	fprintf (stderr, "\"\n") ;
	return	TRUE ;
}

/*	private functions */
/*
 *	obt@Ƀ}[J(K݂)t肷B
 */
BOOL
lispBuffer_initialize (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer)
{
	register TLispBuffer*	pBuffer ;
	BOOL			afInsertionType [NUM_BUFFER_MUST_MARKERS - 1]	= {
		FALSE,	/* buffer top */
		FALSE,	/* edit   top */
		TRUE,	/* point */
		TRUE,	/* edit   end */
		TRUE,	/* buffer end */
	} ;
	TLispEntity*	pEntMarker ;
	register int	i ;
	
	ASSERT (pLispMgr      != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	
	pBuffer			= lispEntity_GetBufferPtr (pEntBuffer) ;

	if (TFAILED (TBufString_Initialize (&pBuffer->m_Text)))
		return	FALSE ;

	/*	obt@ɕK݂}[J쐬B*/
	for (i = 0 ; i < NUM_BUFFER_MUST_MARKERS - 1 ; i ++) {
		if (TFAILED (lispMgr_CreateMarker (pLispMgr, &pEntMarker)))
			return	FALSE ;
		lispEntity_AddRef (pLispMgr, pEntMarker) ;
		lispMarker_SetInsertionType (pLispMgr, pEntMarker, afInsertionType [i]) ;
		/*	}[Jobt@̊ǗɒǉB*/
		lispBuffer_AddMarker (pLispMgr, pEntBuffer, pEntMarker) ;
		pBuffer->m_apMustMarkers [i]	= pEntMarker ;
	}
	/*	obt@wȂ}[JB*/
	if (TFAILED (lispMgr_CreateMarker (pLispMgr, &pEntMarker)))
		return	FALSE ;
	lispEntity_AddRef (pLispMgr, pEntMarker) ;
	lispMarker_SetInsertionType (pLispMgr, pEntMarker, FALSE) ;
	pBuffer->m_apMustMarkers [i]	= pEntMarker ;
	pBuffer->m_nTick	++ ;
	return	TRUE ;
}

BOOL
lispBuffer_uninitialize (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer)
{
	register TLispBuffer*	pBuffer ;
	register TLispEntity*	pEntMarker ;
	register TLispEntity**	ppEntMustMarker ;
	register int			i ;

	pBuffer	= lispEntity_GetBufferPtr (pEntBuffer) ;

	/*	obt@[JϐB*/
	lispBindTable_Destroy (pLispMgr, pBuffer->m_apVariableTable, SIZE_LISP_BIND_TABLE) ;
	pEntMarker	= pBuffer->m_lstMarkers ;
	/*	K{}[JȊOB*/
	while (pEntMarker != NULL) {
		TLispEntity*	pEntNextMarker ;
		lispMarker_GetNext (pLispMgr, pEntMarker, &pEntNextMarker) ;
		lispBuffer_RemoveMarker (pLispMgr, pEntMarker) ;
		pEntMarker	= pEntNextMarker ;
	}
	/*	K{}[JB*/
	ppEntMustMarker	= pBuffer->m_apMustMarkers ;
	for (i = 0 ; i < NUM_BUFFER_MUST_MARKERS ; i ++) {
		if (*ppEntMustMarker != NULL) {
			lispEntity_Release (pLispMgr, *ppEntMustMarker) ;
			*ppEntMustMarker = NULL ;
		}
		ppEntMustMarker	++ ;
	}
	TBufString_Uninitialize (&pBuffer->m_Text) ;
	pBuffer->m_nTick	++ ;
	return	TRUE ;
}

TLispEntity*
lispBuffer_getMarker (
	register TLispEntity*	pEntBuffer,
	register int			nMarker)
{
	register TLispBuffer*	pBuffer ;
	
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (MARKER_BTOP <= nMarker && nMarker < NUM_BUFFER_MUST_MARKERS) ;

	pBuffer			= lispEntity_GetBufferPtr (pEntBuffer) ;
	return	pBuffer->m_apMustMarkers [nMarker] ;
}

BOOL
lispBuffer_getMarkerPosition (
	register TLispManager*	pLispMgr,
	register TLispEntity*	pEntBuffer,
	register int			nMarker,
	register int*			pnRetval)
{
	TLispEntity*	pMarkerBuffer ;
	TLispEntity*	pMarker ;
	
	ASSERT (pLispMgr      != NULL) ;
	ASSERT (pEntBuffer != NULL) ;
	ASSERT (MARKER_BTOP <= nMarker && nMarker < NUM_BUFFER_MUST_MARKERS) ;
	ASSERT (pnRetval      != NULL) ;

	pMarker			= lispBuffer_getMarker (pEntBuffer, nMarker) ;
	if (pMarker == NULL)
		return	FALSE ;
	if (TSUCCEEDED (lispMarker_GetBufferPosition (pLispMgr, pMarker, &pMarkerBuffer, pnRetval)) &&
		pMarkerBuffer == pEntBuffer) 
		return	TRUE ;
	return	FALSE ;
}

