/* # skkinput (Simple Kana-Kanji Input)
 *
 * This file is part of skkinput.
 * Copyright (C) 2002
 * Takashi SAKAMOTO (PXG01715@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "local.h"
#include <stdio.h>
#include <assert.h>
#include "lmachinep.h"

static	int	lispMachineState_keyBindingLookupKey (TLispMachine*, TLispEntity*, TLispEntity*, TLispEntity*, int, BOOL, BOOL, TLispEntity**) ;

/*
 *	(keymap CHARTABLE . ALIST) Ƃ`̐VL[}bvĕԂB
 *	CHARTABLE ͎ۂɂ ASCII ɔꂽ char-table Ȃ̂A
 *	vector őpĂBALIST  function key  mouse event Ȃ
 *	input stream ɂ鑼̉𔛂 assoc-list łB
 *	SẴGg͍ŏ nil ɁAucommand undefinedvӖĂ
 *	邪AĂB
 *
 *	optional  STRING ̓L[}bṽj[^ĂBx-popup-menu
 *	Ŏg炵B
 *
 *	(make-keymap &optional STRING)
 */
TLMRESULT
lispMachineState_MakeKeymap (
	register TLispMachine* pLM)
{
	register TLispManager*	pLispMgr ;
	TLispEntity*	pArglist ;
	TLispEntity*	pString ;
	TLispEntity*	pKeymap ;

	assert (pLM != NULL) ;
	pLispMgr	= pLM->m_pLispMgr ;
	assert (pLispMgr != NULL) ;

	lispMachineCode_GetLReg (pLM, LM_LREG_ACC, &pArglist) ;
	if (TFAILED (lispEntity_GetCar (pLispMgr, pArglist, &pString))) {
		lispMachineCode_SetError (pLM) ;
		return	LMR_RETURN ;
	}
	if (TFAILED (lispMgr_CreateKeymap (pLispMgr, pString, &pKeymap))) 
		return	LMR_ERROR ;
	lispMachineCode_SetLReg (pLM, LM_LREG_ACC, pKeymap) ;
	return	LMR_RETURN ;
}

/*
 *	(make-sparse-keymap &optional STRING)
 */
TLMRESULT
lispMachineState_MakeSparseKeymap (
	register TLispMachine* pLM)
{
	register TLispManager*	pLispMgr ;
	TLispEntity*	pArglist ;
	TLispEntity*	pString ;
	TLispEntity*	pKeymap ;

	assert (pLM != NULL) ;
	pLispMgr	= pLM->m_pLispMgr ;
	assert (pLispMgr != NULL) ;

	lispMachineCode_GetLReg (pLM, LM_LREG_ACC, &pArglist) ;
	if (TFAILED (lispEntity_GetCar (pLispMgr, pArglist, &pString))) {
		lispMachineCode_SetError (pLM) ;
		return	LMR_RETURN ;
	}
	if (TFAILED (lispMgr_CreateSparseKeymap (pLispMgr, pString, &pKeymap))) 
		return	LMR_ERROR ;
	lispMachineCode_SetLReg (pLM, LM_LREG_ACC, pKeymap) ;
	return	LMR_RETURN ;
}

TLMRESULT
lispMachineState_CopyKeymap (
	register TLispMachine* pLM)
{
	register TLispManager*	pLispMgr ;
	TLispEntity*	pEntArglist ;
	TLispEntity*	pEntKeymap ;
	TLispEntity*	pEntRetval ;

	assert (pLM != NULL) ;
	pLispMgr	= pLM->m_pLispMgr ;
	assert (pLispMgr != NULL) ;
	lispMachineCode_GetLReg (pLM, LM_LREG_ACC, &pEntArglist) ;
	lispEntity_GetCar (pLispMgr, pEntArglist, &pEntKeymap) ;
	if (TFAILED (lispMgr_CopyKeymap (pLispMgr, pEntKeymap, &pEntRetval))) {
		lispMachineCode_SetError (pLM) ;
	} else {
		lispMachineCode_SetLReg (pLM, LM_LREG_ACC, pEntRetval) ;
	}
	return	LMR_RETURN ;
}

/*
 *	(keymapp OBJECT)
 *
 *	keymapp  built-in function łB
 *	 OBJECT L[}bvł΁At ԂB
 *	L[}bv (keymap . ALIST) ܂́Å֐̒`L[}bvł
 *	V{ł(̌҂͈͂̕Ӗsc)B
 *	ALIST ̗vf (CHAR . DEFN) ܂ (SYMBOL . DEFN) ̂悤Ȍ`Ă
 *	Bvector ̎́cB
 */
TLMRESULT
lispMachineState_Keymapp (
	register TLispMachine*	pLM)
{
	register TLispManager*	pLispMgr ;
	TLispEntity*	pArglist ;
	TLispEntity*	pEntity ;
	TLispEntity*	pRetval ;

	assert (pLM != NULL) ;

	pLispMgr	= pLM->m_pLispMgr ;
	assert (pLispMgr != NULL) ;

	/*
	 *	̒iKł͈ Eval ͍ςłB
	 *	̓XgƂȂāAACC ɓĂB
	 */
	lispMachineCode_GetLReg (pLM, LM_LREG_ACC, &pArglist) ;

	if (TFAILED (lispEntity_GetCar (pLispMgr, pArglist, &pEntity))) {
		lispMachineCode_SetError (pLM) ;
		return	LMR_RETURN ;
	}
	if (TSUCCEEDED (lispEntity_Keymapp (pLispMgr, pEntity))) {
		lispMgr_CreateT   (pLispMgr, &pRetval) ;
	} else {
		lispMgr_CreateNil (pLispMgr, &pRetval) ;
	}
	lispMachineCode_SetLReg (pLM, LM_LREG_ACC, pRetval) ;
	return	LMR_RETURN ;
}

/*
 *
 */
TLMRESULT
lispMachineState_LookupKey (
	register TLispMachine*	pLM)
{
	register TLispManager*	pLispMgr ;
	TLispEntity*	pArglist ;
	TLispEntity*	pKey ;
	TLispEntity*	pKeymap ;
	TLispEntity*	pFunction ;
	TLispEntity*	pAcceptDefault ;
	BOOL			fAcceptDefault ;
	int				nMatch ;
	int				nLength ;

	assert (pLM != NULL) ;
	pLispMgr	= pLM->m_pLispMgr ;
	assert (pLispMgr != NULL) ;

	/*
	 *	̒iKł͈ Eval ͍ςłB
	 *	̓XgƂȂāAACC ɓĂB
	 */
	lispMachineCode_GetLReg (pLM, LM_LREG_ACC, &pArglist) ;
	assert (pArglist != NULL) ;

	if (TFAILED (lispEntity_GetCar  (pLispMgr, pArglist, &pKeymap)) ||
		TFAILED (lispEntity_GetCdr  (pLispMgr, pArglist, &pArglist)) ||
		TFAILED (lispEntity_GetCar  (pLispMgr, pArglist, &pKey)) ||
		TFAILED (lispEntity_GetCadr (pLispMgr, pArglist, &pAcceptDefault))) {
		lispMachineCode_SetError (pLM) ;
		return	LMR_RETURN ;
	}
	if (TFAILED (lispEntity_Nullp (pLispMgr, pAcceptDefault))) {
		fAcceptDefault	= TRUE ;
	} else {
		fAcceptDefault	= FALSE ;
	}
	/*	ACCEPT-DEFAULT ͂܂ȂB*/
	nMatch	= 0 ;
	if (TFAILED (lispKeymap_Lookup (pLispMgr, pKeymap, pKey, fAcceptDefault, FALSE, &nMatch, &pFunction))) 
		lispMachineCode_SetError (pLM) ;

	(void) lispEntity_GetLength (pLispMgr, pKey, &nLength) ;
	if (nLength == nMatch) {
		/*	100% vꍇɂ́AԂꂽ function B*/
		lispMachineCode_SetLReg (pLM, LM_LREG_ACC, pFunction) ;
	} else {
		TLispEntity*	hInteger ;
		
		/*	vԂB*/
		if (TFAILED (lispMgr_CreateInteger (pLispMgr, nMatch, &hInteger))) 
			return	LMR_ERROR ;
		lispMachineCode_SetLReg (pLM, LM_LREG_ACC, hInteger) ;
	}
	return	LMR_RETURN ;
}

/*
 *	(define-key KEYMAP KE DEF)
 *
 *	 KEYMAP, KEY, DEFBKEYMAP ̒ L[V[PX KEY 
 *	DEF ƂĒ`BKEYMAP  keymap łBKEY ͕܂
 *	 symbol ╶̂̃xNgłB
 *	ASCII  (127 𒴂)  vector gĂȂ܂߂
 *	ƂłBDEF  key ̒`ł鉽łB
 *	nil  keymap ɂ
 *	command
 *	string
 *	keymap
 *	symbol
 *	cons (STRING . DEFN)
 *	... 
 */
TLMRESULT
lispMachineState_DefineKey (
	register TLispMachine* pLM)
{
	register TLispManager*	pLispMgr ;
	TLispEntity*	pArglist ;
	TLispEntity*	pKEYMAP ;
	TLispEntity*	pKEY ;
	TLispEntity*	pDEF ;

	assert (pLM != NULL) ;
	pLispMgr	= pLM->m_pLispMgr ;
	assert (pLispMgr != NULL) ;
	lispMachineCode_GetLReg (pLM, LM_LREG_ACC, &pArglist) ;
	assert (pArglist != NULL) ;

	if (TFAILED (lispEntity_GetCar  (pLispMgr, pArglist, &pKEYMAP)) ||
		TFAILED (lispEntity_GetCdr  (pLispMgr, pArglist, &pArglist)) ||
		TFAILED (lispEntity_GetCar  (pLispMgr, pArglist, &pKEY)) ||
		TFAILED (lispEntity_GetCadr (pLispMgr, pArglist, &pDEF))) {
		lispMachineCode_SetError (pLM) ;
		return	LMR_RETURN ;
	}
	/*	pKEY  VECTOR ̏ꍇ͍͍lĂȂB*/
	if (TSUCCEEDED (lispEntity_Vectorp (pLispMgr, pKEY))) {
		TLispEntity**	ppEntKeySeq ;
		int				nKeySeq ;

		(void) lispEntity_GetVectorValue (pLispMgr, pKEY, &ppEntKeySeq, &nKeySeq) ;
		if (TFAILED (lispKeymap_DefineKeyWithVector (pLispMgr, pKEYMAP, ppEntKeySeq, nKeySeq, pDEF)))
			lispMachineCode_SetError (pLM) ;
	} else {
		const Char*		pString ;
		int				nString ;

		if (TFAILED (lispEntity_GetStringValue (pLispMgr, pKEY, &pString, &nString))) {
			lispMachineCode_SetError (pLM) ;
			return	LMR_RETURN ;
		}
		if (TFAILED (lispKeymap_DefineKey (pLispMgr, pKEYMAP, pString, nString, pDEF))) 
			lispMachineCode_SetError (pLM) ;
	}
	lispMachineCode_SetLReg (pLM, LM_LREG_ACC, pDEF) ;
	return	LMR_RETURN ;
}

/*
 *	(current-minor-mode-maps)
 *
 *	current-buffer  minor-modes ɊĂĂ keymap ̃Xg
 *	ԂB
 *
 *	minor-mode-map-alist Ɉv̂oāA
 *	(keymap1 keymap2 ...) Ƃ list B
 */
TLMRESULT
lispMachineState_CurrentMinorModeMaps (register TLispMachine* pLM)
{
	register TLispManager*	pLispMgr ;
	TLispEntity*	pEntKeymaps ;

	assert (pLM != NULL) ;
	pLispMgr	= pLM->m_pLispMgr ;
	assert (pLispMgr != NULL) ;
	if (TFAILED (lispMachineCode_CurrentMinorModeMaps (pLM, &pEntKeymaps))) {
		lispMachineCode_SetError (pLM) ;
	} else {
		lispMachineCode_SetLReg  (pLM, LM_LREG_1, pEntKeymaps) ;
	}
	return	LMR_RETURN ;
}

/*
 *	(use-local-map KEYMAP)
 *
 *	local keymap Ƃ KEYMAP IB KEYMAP  nil Ȃ
 *	local keymap ͖ɂ邱ƂӖB
 */
TLMRESULT
lispMachineState_UseLocalMap (register TLispMachine* pLM)
{
	register TLispManager*	pLispMgr ;
	TLispEntity*	pEntArglist ;
	TLispEntity*	pEntKeymap ; 
	TLispEntity*	pEntCurBuffer ;

	assert (pLM != NULL) ;
	pLispMgr	= pLM->m_pLispMgr ;
	assert (pLispMgr != NULL) ;
	lispMachineCode_GetLReg (pLM, LM_LREG_ACC, &pEntArglist) ;
	if (TFAILED (lispEntity_GetCar (pLispMgr, pEntArglist, &pEntKeymap)) ||
		TFAILED (lispMachineCode_GetCurrentBuffer (pLM, &pEntCurBuffer))) {
		lispMachineCode_SetError (pLM) ;
		return	LMR_RETURN ;
	}
	lispBuffer_SetKeymap (pLispMgr, pEntCurBuffer, pEntKeymap) ;
	lispMachineCode_SetLReg (pLM, LM_LREG_ACC, pEntKeymap) ;
	return	LMR_RETURN ;
}

/*
 *	(current-local-map)
 *
 *	current-buffer  local keymap ԂBĂȂ
 *	nil ԂB
 */
TLMRESULT
lispMachineState_CurrentLocalMap (register TLispMachine* pLM)
{
	register TLispManager*	pLispMgr ;
	TLispEntity*	pEntArglist ;
	TLispEntity*	pEntCurBuffer ;
	TLispEntity*	pEntKeymap ;

	assert (pLM != NULL) ;
	pLispMgr	= pLM->m_pLispMgr ;
	assert (pLispMgr != NULL) ;
	lispMachineCode_GetLReg (pLM, LM_LREG_ACC, &pEntArglist) ;
	if (TFAILED (lispMachineCode_GetCurrentBuffer (pLM, &pEntCurBuffer))) {
		lispMachineCode_SetError (pLM) ;
	} else {
		if (TFAILED (lispBuffer_GetKeymap (pEntCurBuffer, &pEntKeymap)) ||
			pEntKeymap == NULL) 
			lispMgr_CreateNil (pLispMgr, &pEntKeymap) ;
		lispMachineCode_SetLReg (pLM, LM_LREG_ACC, pEntKeymap) ;
	}
	return	LMR_RETURN ;
}

/*
 *	(where-is-internal DEFINITION &optional KEYMAP FIRSTONLY NOINDIRECT)
 *
 *	DEFINITION ĂяoL[̃XgԂB
 *	 KEYMAP  nil łȂ΁A KEYMAP  global keymap 
 *	B
 *	 KEYMAP  nil Ȃ΁A active  keymap ̑SĂB
 *	 KEYMAP  keymap ̃Xgł΁A keymap 
 *	B
 *
 *	3Ԗڂ̈ FIRSTONLY  nil łȂ΁Atŏ̃L[V[P
 *	XԂB(SẴL[V[PXłȂ)
 *	 FIRSTONLY V{ `non-ascii' Ȃ΁Aꂪł邩ɂ
 *	炸ŏɌtoCfBOԂB
 *	FIRSTONLY ʂ non-nil value ł΁AASCII ̃V[PXIԁB
 *	j[oCfBO͋ۂB
 *	4Ԗڂ̈ NOINDIRECT  non-nil Ȃ΁ÃL[}bv܂ slot 
 *	ւ̊ԐڎQƂɏ]ȂB͊ԐړIȒ`邱Ƃ\ɂB
 *	If optional 4th arg NOINDIRECT is non-nil, don't follow indirections
 *	to other keymaps or slots.  This makes it possible to search for an
 *	indirect definition itself.
 *
 *	current-minor-mode-maps, current-local-map, global-map
 */
TLMRESULT
lispMachineState_WhereIsInternal (register TLispMachine* pLM)
{
	TLispEntity*	pEntBuffer ;

	/*	not-implemented */
	lispMachineCode_GetCurrentBuffer (pLM, &pEntBuffer) ;
	return	LMR_RETURN ;
}

/*	built-in function
 *		(key-binding KEY &optional ACCEPT-DEFAULT)
 *
 *	݂ keymap  KEY ɑ΂R}hԂBKEY ͕܂
 *	xNg܂̓L[Xg[N̗łBL[oCh͑A֐
 *	`V{ł낤B
 *
 *	RAkey-binding  default ̃oCfBOƂē삷 t 
 *	΂oCfBO𖳎B
 */
TLMRESULT
lispMachineState_KeyBinding (
	register TLispMachine*	pLM)
{
	register TLispManager*	pLispMgr	= pLM->m_pLispMgr ;
	TLispEntity*	pEntArglist ;
	TLispEntity*	pEntKey ;
	TLispEntity*	pEntAcceptDefault ;
	TLispEntity*	pEntKeymap ;
	TLispEntity*	pEntFunc ;
	TLispEntity*	pEntSymGlobalMap ;
	TLispEntity*	pEntGlobalKeymap ;
	TLispEntity*	pEntCurBuffer ;
	BOOL			fAcceptDefault ;
	int				nLength ;

	lispMachineCode_GetLReg (pLM, LM_LREG_ACC, &pEntArglist) ;
	lispEntity_GetCar  (pLispMgr, pEntArglist, &pEntKey) ;
	lispEntity_GetCadr (pLispMgr, pEntArglist, &pEntAcceptDefault) ;
	fAcceptDefault	= !lispEntity_Nullp (pLispMgr, pEntAcceptDefault) ;

	if (TFAILED (lispEntity_GetLength (pLispMgr, pEntKey, &nLength))) {
		lispMachineCode_SetError (pLM) ;
		return	LMR_RETURN ;
	}

	lispMachineCode_GetCurrentBuffer (pLM, &pEntCurBuffer) ;
	assert (pEntCurBuffer != NULL) ;
	(void) lispBuffer_GetKeymap (pEntCurBuffer, &pEntKeymap) ;
	
	pEntSymGlobalMap	= lispMgr_GetReservedEntity (pLispMgr, LISPMGR_INDEX_GLOBAL_MAP) ;
	if (TFAILED (lispMachine_GetCurrentSymbolValue (pLM, pEntSymGlobalMap, &pEntGlobalKeymap)) ||
		TSUCCEEDED (lispEntity_Voidp (pLispMgr, pEntGlobalKeymap))) 
		pEntGlobalKeymap	= NULL ;
	(void) lispMachineState_keyBindingLookupKey (pLM, pEntKeymap, pEntGlobalKeymap, pEntKey, nLength, FALSE, fAcceptDefault, &pEntFunc) ;
	lispMachineCode_SetLReg (pLM, LM_LREG_ACC, pEntFunc) ;
	return	LMR_RETURN ;
}

/*	built-in function
 *		(j-key-binding KEY &optional ACCEPT-DEFAULT)
 *
 *	݂ keymap  KEY ɑ΂R}hԂBKEY ͕܂
 *	xNg܂̓L[Xg[N̗łBL[oCh͑A֐
 *	`V{ł낤B
 *
 *	RAkey-binding  default ̃oCfBOƂē삷 t 
 *	΂oCfBO𖳎B
 *
 *
 *	key-binding Ƃ̓̈Ⴂ́Awindow-proc Ɠl̃L[
 *	ƂłBAႦ΁A\C-\S-t  bind 𒲂ׂ鎞ɂ 
 *	\C-\S-t  bind ̎ \C-t  bind B
 */
TLMRESULT
lispMachineState_JKeyBinding (
	register TLispMachine*	pLM)
{
	register TLispManager*	pLispMgr	= pLM->m_pLispMgr ;
	TLispEntity*	pEntArglist ;
	TLispEntity*	pEntKey ;
	TLispEntity*	pEntAcceptDefault ;
	TLispEntity*	pEntKeymap ;
	TLispEntity*	pEntFunc ;
	TLispEntity*	pEntSymGlobalMap ;
	TLispEntity*	pEntGlobalKeymap ;
	TLispEntity*	pEntCurBuffer ;
	BOOL			fAcceptDefault ;
	int				nLength ;
	register int	nMatch ;

	lispMachineCode_GetLReg (pLM, LM_LREG_ACC, &pEntArglist) ;
	lispEntity_GetCar  (pLispMgr, pEntArglist, &pEntKey) ;
	lispEntity_GetCadr (pLispMgr, pEntArglist, &pEntAcceptDefault) ;
	fAcceptDefault	= !lispEntity_Nullp (pLispMgr, pEntAcceptDefault) ;

	if (TFAILED (lispEntity_GetLength (pLispMgr, pEntKey, &nLength))) {
		lispMachineCode_SetError (pLM) ;
		return	LMR_RETURN ;
	}

	lispMachineCode_GetCurrentBuffer (pLM, &pEntCurBuffer) ;
	assert (pEntCurBuffer != NULL) ;
	(void) lispBuffer_GetKeymap (pEntCurBuffer, &pEntKeymap) ;
	
	pEntSymGlobalMap	= lispMgr_GetReservedEntity (pLispMgr, LISPMGR_INDEX_GLOBAL_MAP) ;
	if (TFAILED (lispMachine_GetCurrentSymbolValue (pLM, pEntSymGlobalMap, &pEntGlobalKeymap)) ||
		TSUCCEEDED (lispEntity_Voidp (pLispMgr, pEntGlobalKeymap))) 
		pEntGlobalKeymap	= NULL ;
#if defined (DEBUG)
	fprintf (stderr, "key(%d, %d) = ", nLength, fAcceptDefault) ;
	lispEntity_Print (pLispMgr, pEntKey) ;
	fprintf (stderr, "\n") ;
#endif
	nMatch	= lispMachineState_keyBindingLookupKey (pLM, pEntKeymap, pEntGlobalKeymap, pEntKey, nLength, FALSE, fAcceptDefault, &pEntFunc) ;
#if defined (DEBUG)
	fprintf (stderr, "[1] key-bind(%d) = ", nMatch) ;
	lispEntity_Print (pLispMgr, pEntFunc) ;
	fprintf (stderr, "\n") ;
#endif
	if (nMatch <= 0 || TSUCCEEDED (lispEntity_Nullp (pLispMgr, pEntFunc))) {
		nMatch	= lispMachineState_keyBindingLookupKey (pLM, pEntKeymap, pEntGlobalKeymap, pEntKey, nLength, TRUE, fAcceptDefault, &pEntFunc) ;
#if defined (DEBUG)
		fprintf (stderr, "[2] key-bind(%d) = ", nMatch) ;
		lispEntity_Print (pLispMgr, pEntFunc) ;
		fprintf (stderr, "\n") ;
#endif
	}
	lispMachineCode_SetLReg (pLM, LM_LREG_ACC, pEntFunc) ;
	return	LMR_RETURN ;
}

int
lispMachineState_keyBindingLookupKey (
	register TLispMachine*	pLM,
	register TLispEntity*	pEntLocalKeymap,
	register TLispEntity*	pEntGlobalKeymap,
	register TLispEntity*	pEntKey,
	register int			nLength,
	register BOOL		fNearMatch,
	register BOOL		fAcceptDefault,
	register TLispEntity**	ppEntRetval)
{
	register TLispManager*	pLispMgr	= pLM->m_pLispMgr ;
	TLispEntity*	pEntFunc ;
	int				nMatch ;

	nMatch		= 0 ;
	pEntFunc	= NULL ;
	if (pEntLocalKeymap != NULL &&
		TSUCCEEDED (lispKeymap_Lookup (pLispMgr, pEntLocalKeymap, pEntKey, fAcceptDefault, fNearMatch, &nMatch, &pEntFunc)) &&
		TFAILED (lispEntity_Nullp (pLispMgr, pEntFunc)) &&
		nMatch > 0 && (nMatch == nLength || fAcceptDefault)) {
#if defined (DEBUG)
		fprintf (stderr, "[a] pEntFunc(%p), nMatch(%d), fAcceptDefault(%d), nLength(%d)\n",
				 pEntFunc, nMatch, fAcceptDefault, nLength) ;
#endif
		goto	keyhit ;
	}

	nMatch	= 0 ;
	if (pEntGlobalKeymap != NULL &&
		TFAILED (lispEntity_Voidp (pLispMgr, pEntGlobalKeymap))) {
		/*	L[obt@ɂ܂Ă͂păL[}bvB*/
		lispKeymap_Lookup (pLispMgr, pEntGlobalKeymap, pEntKey, fAcceptDefault, fNearMatch, &nMatch, &pEntFunc) ;
#if defined (DEBUG)
		fprintf (stderr, "[b] pEntFunc(%p), nMatch(%d), fAcceptDefault(%d), nLength(%d)\n",
				 pEntFunc, nMatch, fAcceptDefault, nLength) ;
#endif
		if (pEntFunc == NULL || nMatch <= 0 || (nMatch != nLength && !fAcceptDefault)) 
			lispMgr_CreateNil (pLispMgr, &pEntFunc) ;
	} else {
#if defined (DEBUG)
		fprintf (stderr, "pEntGlobalKeymap (%p)\n", pEntGlobalKeymap) ;
#endif
		/*	ɂqbgȂ΁A undefined łB*/
		lispMgr_CreateNil (pLispMgr, &pEntFunc) ;
	}

 keyhit:
	if (ppEntRetval != NULL)
		*ppEntRetval	= pEntFunc ;
	return	nMatch ;
}


