#include "local.h"
#include <stdio.h>
#include <stdlib.h>
#include "jisyo.h"
#include "localjisyo.h"
#include "cstring.h"
#include "kstring.h"
#include "kfile.h"

#define	OKURIARI_STR	";; okuri-ari entries."
#define	OKURINASI_STR	";; okuri-nasi entries."

/*	vg^Cv錾B*/
static	BOOL	skkLocalJisyo_search		(SkkJisyo*, const Char*, int, const Char*, int, int, TVarbuffer*) ;
static	BOOL	skkLocalJisyo_record		(SkkJisyo*, const Char*, int, BOOL, const Char*, int) ;
static	BOOL	skkLocalJisyo_purge			(SkkJisyo*, const Char*, int, BOOL, const Char*, int) ;
static	BOOL	skkLocalJisyo_recordEx		(SkkJisyo*, const Char*, int, const Char*, int, const Char*, int, BOOL) ;
static	BOOL	skkLocalJisyo_purgeEx		(SkkJisyo*, const Char*, int, const Char*, int, const Char*, int, BOOL) ;
static	BOOL	skkLocalJisyo_save			(SkkJisyo*) ;
static	BOOL	skkLocalJisyo_completion	(SkkJisyo*, const Char*, int, TVarbuffer*) ;
static	BOOL	skkLocalJisyo_destroy		(SkkJisyo*) ;
static	LPCTSTR	skkLocalJisyo_getPath		(SkkJisyo*) ;

static	BOOL	skkLocalJisyo_load				(SkkLocalJisyo*, LPCTSTR) ;
static	BOOL	skkLocalJisyo_initJisyo			(SkkLocalJisyo*) ;
static	int		skkLocalJisyo_findAnsiAtLineTop	(SkkLocalJisyo*, int, const char*) ;
static	int		skkLocalJisyo_findCandidate		(SkkLocalJisyo*, int, int, const Char*, int) ;
static	BOOL	skkLocalJisyo_findCompletion	(TBufStringMarker*, TVarbuffer*) ;
static	BOOL	skkLocalJisyo_prepareDirectory	(LPCTSTR) ;
static	BOOL	skkLocalJisyo_moveLineToTop		(SkkLocalJisyo*, int, BOOL) ;
static	int		skkLocalJisyo_findOkuriCandidate(SkkLocalJisyo*, int, int, const Char*, int, int*) ;

static	BOOL	skkLocalJisyo_bDeleteOkuriEntry	(SkkLocalJisyo*, int, const Char*, int, const Char*, int, int*) ;
static	BOOL	skkLocalJisyo_bOkuriEntryExistp	(SkkLocalJisyo*, int, const Char*, int) ;
static	int		skkLocalJisyo_iCompareMarkerWithString	(TBufStringMarker*, const Char*, int) ;
static	BOOL	skkLocalJisyo_bSkipNextOkuriEntry		(TBufStringMarker*) ;

static	BOOL	skkLocalJisyo_bVerify		(SkkLocalJisyo*) ;


/*	VTBL
 */
static SkkJisyoFunc		sLocalJisyoProcTbl	= {
	skkLocalJisyo_search,
	skkLocalJisyo_completion,
	skkLocalJisyo_recordEx,
	skkLocalJisyo_purgeEx,
	skkLocalJisyo_save,
	skkLocalJisyo_destroy,
	skkLocalJisyo_getPath,
} ;

SkkJisyo*
SkkLocalJisyo_Create (
	const Char*		pFileName,
	int				nFileName)
{
	TCHAR			strPath [PATH_MAX + 1] ;
	SkkJisyo*		pJisyo ;
	SkkLocalJisyo*	pLocalJisyo ;

	if (nFileName > PATH_MAX)
		return	NULL ;

#if defined (UNICODE)
	internal2wstr (strPath, ARRAYSIZE (strPath) - 1, pFileName, nFileName) ;
#else
	cstrtostr (strPath, pFileName, nFileName) ;
#endif
	strPath [nFileName]	= '\0' ;
#if defined (DEBUG) || 1
	DebugPrintfToFile (TEXT ("SkkLocalJisyo_Create (localjisyo => \"%s\")\n"), strPath) ;
#endif

	pJisyo	= MALLOC (sizeof (SkkJisyo) + sizeof (SkkLocalJisyo)) ;
	if (pJisyo == NULL)
		return	NULL ;

	pJisyo->m_pVtbl		= &sLocalJisyoProcTbl ;
	pLocalJisyo			= (SkkLocalJisyo *)(pJisyo + 1) ;
	memcpy (pLocalJisyo->m_tszPath, strPath, sizeof (TCHAR) * PATH_MAX) ;
	pLocalJisyo->m_nOkuriAriPos		= 0 ;
	pLocalJisyo->m_nOkuriNasiPos	= 0 ;
	pLocalJisyo->m_pNext			= NULL ;
	if (TFAILED (TBufString_Initialize (&pLocalJisyo->m_bufJisyo))) {
		FREE (pJisyo) ;
		return	NULL ;
	}

	if (TFAILED (skkLocalJisyo_load (pLocalJisyo, strPath))) {
		/* t@C݂Ȃ΁AƂɍB*/
		/* xobt@B*/
		TBufString_Uninitialize (&pLocalJisyo->m_bufJisyo) ;
		TBufString_Initialize   (&pLocalJisyo->m_bufJisyo) ;
		skkLocalJisyo_initJisyo (pLocalJisyo) ;
	}
	pLocalJisyo->m_fModified		= FALSE ;
	return	pJisyo ;
}

SkkJisyo*
SkkLocalJisyo_CreateT (
	LPCTSTR		strPath)
{
	int			nstrPath ;
	SkkJisyo*		pJisyo ;
	SkkLocalJisyo*	pLocalJisyo ;

	if (strPath == NULL)
		return	NULL ;
	nstrPath	= lstrlen (strPath) ;
	if (nstrPath > PATH_MAX)
		return	NULL ;

#if defined (DEBUG) || 1
	DebugPrintfToFile (TEXT ("SkkLocalJisyo_CreateT (localjisyo => \"%s\")\n"), strPath) ;
#endif
	pJisyo	= MALLOC (sizeof (SkkJisyo) + sizeof (SkkLocalJisyo)) ;
	if (pJisyo == NULL)
		return	NULL ;

	pJisyo->m_pVtbl		= &sLocalJisyoProcTbl ;
	pLocalJisyo			= (SkkLocalJisyo *)(pJisyo + 1) ;
	lstrcpy (pLocalJisyo->m_tszPath, strPath) ;
	/*memcpy (pLocalJisyo->m_tszPath, strPath, sizeof (TCHAR) * PATH_MAX) ;*/
	pLocalJisyo->m_nOkuriAriPos		= 0 ;
	pLocalJisyo->m_nOkuriNasiPos	= 0 ;
	pLocalJisyo->m_pNext			= NULL ;
	if (TFAILED (TBufString_Initialize (&pLocalJisyo->m_bufJisyo))) {
		FREE (pJisyo) ;
		return	NULL ;
	}

	if (TFAILED (skkLocalJisyo_load (pLocalJisyo, strPath))) {
		/* t@C݂Ȃ΁AƂɍB*/
		/* xobt@B*/
		TBufString_Uninitialize (&pLocalJisyo->m_bufJisyo) ;
		TBufString_Initialize   (&pLocalJisyo->m_bufJisyo) ;
		skkLocalJisyo_initJisyo (pLocalJisyo) ;
	}
	return	pJisyo ;
}

/*	auto-okuri-process ŗpBskk-okuri-search-subr-original ɑB
 *	ƁAlϊꍇɂ auto-okuri-process ͖ɂȂ̂A
 *	ӂ邱ƁB
 *	cǂ͌鑤̏񂪕KvƂƂcB
 */
BOOL
SkkLocalJisyo_bOkuriSearch (
	SkkJisyo*		pSkkJisyo,
	const Char*		pKey,
	int				nKey,
	const Char*		pOkurigana,
	int				nOkurigana,
	int				nOkuriType,
	TVarbuffer*		pvbuf)
{
	SkkLocalJisyo*		pJisyo ;
	int					nStartPos, nEndPos, nPos, iMatchLen ;
	TBufStringMarker	mk ;
	Char				cc, chPrev = -1 ;
	BOOL				bDuringAnnotation ;

	pJisyo		= (SkkLocalJisyo *)(pSkkJisyo + 1) ;
	nStartPos	= pJisyo->m_nOkuriAriPos ;
	nEndPos		= pJisyo->m_nOkuriNasiPos ;

	/*	cɂ߂Ȃc
	 *	1. key ̐擪Ǝ̒PrB
	 *	2. uAt@xbg+Xy[Xv̕܂ŗāAkey ƈقȂ邱ƂB
	 *	3. key ̎ƈقȂĂ镔̕(must )̃[}\L̃[}\L
	 *	   vB
	 *	 3 iKz ok ł͂Ȃ낤B
	 *
	 *	3ڂ̕] /[..]/ ŎsĂ悤BƁAz, j ̂悤 z, j ̂悤ȗh炬
	 *	̂ŁA1tIIƂ킯ɂ͂Ȃ悤B
	 */
	for ( ; ; ) {
		nPos		= skkLocalJisyo_findOkuriCandidate (pJisyo, nStartPos, nEndPos, pKey, nKey, &iMatchLen) ;
		if (nPos < 0) {
			/*	Not found - ȏ㑶݂Ȃ or S݂ȂB
			 */
			break ;
		}
		if (iMatchLen < 1 || iMatchLen >= nKey) {
			/* 蓾ȂƎvA͂̍s skip B*/
			goto	exit_head ;
		}

		if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk) ))
			return	FALSE ;

		while (cc = TBufStringMarker_GetChar (&mk), cc != (Char) -1 && cc != '/' && cc != '\n') {
			TBufStringMarker_Forward (&mk, 1) ;
		}
		if (cc != '/')
			return	FALSE ;

		pOkurigana	= pKey + iMatchLen ;
		nOkurigana	= nKey - iMatchLen ;
		bDuringAnnotation	= FALSE ;
		do {
			if (TFAILED (TBufStringMarker_Forward (&mk, 1)))
				break ;
			cc = TBufStringMarker_GetChar (&mk) ;
			switch (cc) {
			case	'[':
				if (chPrev == '/' && ! bDuringAnnotation) {
					/*	艼vTB*/
					static Char	_srSeparator []	= { ']', '/', '[' } ;
					int			nCount ;

					for ( ; ; ) {
						nCount	= skkLocalJisyo_iCompareMarkerWithString (&mk, pOkurigana, nOkurigana) ;
						if (nCount < nOkurigana) {
							cc = TBufStringMarker_GetChar (&mk) ;
							if (cc == '/') {
								/*	found */
								chPrev	= cc ;
								while (cc = TBufStringMarker_GetChar (&mk), cc != (Char)-1 && cc != '\n') {
									switch (cc) {
									case	']':
										if (cc == ']' && chPrev == '/' && ! bDuringAnnotation)
											goto	exit_head ;
										break ;
									case	'/':
										if (! bDuringAnnotation) {
											/*	艼ɒǉĂȂ΂ȂȂB
											 *	͌ɑ艼̏񂪂Ȃ̂ŁB
											 *	ǉYƁAႦ ``Ureshii'' ŌƂB
											 *	 ``//'' Ԃ
											 *		ꂵ -> 
											 *	ƁuvĂ܂B]ނׂ́A
											 *		ꂵ -> 
											 *	ł낤B
											 */
											if (TFAILED (TVarbuffer_Add (pvbuf, pOkurigana, nOkurigana)))
												return	FALSE ;
										}
										bDuringAnnotation	= FALSE ;
										break ;
									case	';':
										/*	艼ɒǉĂȂ΂ȂȂcA
										 *	Annotation ܂Ă\B
										 *
										 *	[/;قɂ/] Ƃ₾ꍇɁAl
										 *	``'' ǉ ``/;قɂ炵/''
										 *	Ƃʂ߂āA
										 *		ꂵ -> 
										 *	߂A``قɂ炵'' ƂʂɂȂĂ܂B
										 *
										 *	lisp ϊH c܂ł͒mȂcB
										 */
										if (TFAILED (TVarbuffer_Add (pvbuf, pOkurigana, nOkurigana)))
											return	FALSE ;
										bDuringAnnotation	= TRUE ;
										break ;
									default:
										break ;
									}
									if (TFAILED (TVarbuffer_Add (pvbuf, &cc, 1)))
										return	FALSE ;
									if (TFAILED (TBufStringMarker_Forward (&mk, 1)))
										break ;
									chPrev	= cc ;
								}
							}
						}
						/*  [...] TB*/
						for ( ; ; ) {
							while (cc = TBufStringMarker_GetChar (&mk), cc != '/' && cc != '\n' && cc != (Char)-1) {
								if (TFAILED (TBufStringMarker_Forward (&mk, 1)))
									goto	exit_head ;
							}
							if (cc != '/')
								goto	exit_head ;
							if (TFAILED (TBufStringMarker_Forward (&mk, 1)))
								goto	exit_head ;

							/* ؂ ``]/['' ɂȂĂ锤B*/
							nCount	= skkLocalJisyo_iCompareMarkerWithString (&mk, _srSeparator, ARRAYSIZE (_srSeparator)) ;
							if (nCount == 0)
								break ;
						}
					}
				}
				break ;
			case	'/':
				bDuringAnnotation	= FALSE ;
				break ;
			case	';':
				bDuringAnnotation	= TRUE ;
				break ;
			default:
				break ;
			}
			chPrev	= cc ;
		}	while (cc != (Char) -1 && cc != '\n') ;

exit_head:
		/* }[J̈ʒu߂B*/
		if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk) ))
			return	FALSE ;
		while (cc = TBufStringMarker_GetChar (&mk), cc != '\n' && cc != (Char)-1) {
			TBufStringMarker_Forward (&mk, 1) ;
			nPos	++ ;
		}
		if (cc == (Char) -1)
			break ;
		nStartPos	= nPos + 1 ;	/* ``\n'' ̎̕ɁB*/
	}
	return	TRUE ;
	UNREFERENCED_PARAMETER (nOkuriType) ;
}

/*========================================================================
 *	X̎̑B
 *========================================================================*/
BOOL
skkLocalJisyo_search (
	SkkJisyo*		pSkkJisyo,
	const Char*		pKey,
	int				nKey,
	const Char*		pOkurigana,
	int				nOkurigana,
	int				nOkuriType,
	TVarbuffer*		pvbuf)
{
	SkkLocalJisyo*		pJisyo ;
	int					nStartPos, nEndPos, nPos ;
	TBufStringMarker	mk ;
	Char				cc, chPrev = -1 ;
	BOOL				bDuringAnnotation ;

	pJisyo		= (SkkLocalJisyo *)(pSkkJisyo + 1) ;

	/* auto-okuri Ał͂Ȃ nOkurigana > 0 ׂH */
	if (nOkuriType != SEARCH_OKURI_NASHI && nOkuriType != SEARCH_OKURI_AUTO) {
		nStartPos	= pJisyo->m_nOkuriAriPos ;
		nEndPos		= pJisyo->m_nOkuriNasiPos ;	
	} else {
		nStartPos	= pJisyo->m_nOkuriNasiPos ;
		nEndPos		= INT_MAX ;	
	}
	nPos		= skkLocalJisyo_findCandidate (pJisyo, nStartPos, nEndPos, pKey, nKey) ;
	if (nPos < 0) {
		/*	Not found */
		return	TRUE ;
	}
	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk) ))
		return	FALSE ;
	while (cc = TBufStringMarker_GetChar (&mk), 
		cc != (Char) -1 && cc != '/' && cc != '\n') {
		TBufStringMarker_Forward (&mk, 1) ;
		nPos	++ ;
	}
	if (cc != '/')
		return	FALSE ;

	/*	艼̌@ okuri-strict Ȃ okuri-strict-precedence ̏ꍇɂ́A
	 *	艼Ƀ}b`(݂̂Ԃ|ɗ)
	 */
	if ((nOkuriType == SEARCH_OKURI_STRICT || nOkuriType == SEARCH_OKURI_STRICT_PRECEDENCE) && nOkurigana > 0) {
		bDuringAnnotation	= FALSE ;
		do {
			if (TFAILED (TBufStringMarker_Forward (&mk, 1)))
				break ;
			cc = TBufStringMarker_GetChar (&mk) ;
			switch (cc) {
			case	'[':
				if (chPrev == '/' && ! bDuringAnnotation) {
					/*	艼vTB*/
					static Char	_srSeparator []	= { ']', '/', '[' } ;
					int			nCount ;

					for ( ; ; ) {
						nCount	= skkLocalJisyo_iCompareMarkerWithString (&mk, pOkurigana, nOkurigana) ;
						if (nCount == 0) {
							/* found */
							cc = TBufStringMarker_GetChar (&mk) ;
							if (cc == '/') {
								/*	found */
								chPrev				= cc ;
								while (cc = TBufStringMarker_GetChar (&mk), cc != (Char)-1 && cc != '\n') {
									switch (cc) {
									case	']':
										if (cc == ']' && chPrev == '/' && ! bDuringAnnotation)
											goto	exit_head ;
										break ;
									case	'/':
										bDuringAnnotation	= FALSE ;
										break ;
									case	';':
										bDuringAnnotation	= TRUE ;
										break ;
									default:
										break ;
									}
									if (TFAILED (TVarbuffer_Add (pvbuf, &cc, 1)))
										return	FALSE ;
									if (TFAILED (TBufStringMarker_Forward (&mk, 1)))
										break ;
									chPrev	= cc ;
								}
								goto	exit_head ;
							}
						}
						for ( ; ; ) {
							while (cc = TBufStringMarker_GetChar (&mk), cc != '/' && cc != '\n' && cc != (Char)-1) {
								if (TFAILED (TBufStringMarker_Forward (&mk, 1)))
									goto	exit_head ;
							}
							if (cc != '/')
								goto	exit_head ;
							if (TFAILED (TBufStringMarker_Forward (&mk, 1)))
								goto	exit_head ;
							nCount	= skkLocalJisyo_iCompareMarkerWithString (&mk, _srSeparator, ARRAYSIZE (_srSeparator)) ;
							if (nCount == 0)
								break ;
						}
					}
				}
				break ;
			case	'/':
				bDuringAnnotation	= FALSE ;
				break ;
			case	';':
				bDuringAnnotation	= TRUE ;
				break ;
			default:
				break ;
			}
			chPrev	= cc ;
		}	while (cc != (Char) -1 && cc != '\n') ;

exit_head:
		/*	STRICT ƁAȊǑ͎̂ċB
		 */
		if (nOkuriType == SEARCH_OKURI_STRICT)
			return	TRUE ;

		/* }[J̈ʒu߂B*/
		if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk) ))
			return	FALSE ;

		cc = TBufStringMarker_GetChar (&mk) ;
	}

	/* ʏ팟ŌtɒǉB*/
	chPrev				= '\0' ;
	bDuringAnnotation	= FALSE ;
	while (cc != (Char) -1 && cc != '\n') {
		if (TFAILED (TVarbuffer_Add (pvbuf, &cc, 1)))
			return	FALSE ;
		if (TFAILED (TBufStringMarker_Forward (&mk, 1)))
			break ;
		cc = TBufStringMarker_GetChar (&mk) ;

		switch (cc) {
		case	'[':
			/*	艼ĂĂ ``/['' ͏ILB
			 */
			if (nOkuriType != SEARCH_OKURI_NASHI && nOkuriType != SEARCH_OKURI_AUTO && chPrev == '/' && ! bDuringAnnotation) {
				goto	exit_loop ;
			}
			break ;
		case	'/':
			bDuringAnnotation	= FALSE ;
			break ;
		case	';':
			bDuringAnnotation	= TRUE ;
			break ;
		default:
			break ;
		}
		chPrev	= cc ;
	}
exit_loop:
	return	TRUE ;
}

BOOL
skkLocalJisyo_completion (
	SkkJisyo*		pSkkJisyo,
	const Char*		pKey,
	int				nKey,
	TVarbuffer*		pvbuf)
{
	SkkLocalJisyo*		pJisyo	= (SkkLocalJisyo *)(pSkkJisyo + 1) ;
	int					nPos ;
	TBufStringMarker	mk ;
	Char				cc ;
	static const Char	chSEPARATOR	= ' ' ;	/* separator  space */

	nPos	= pJisyo->m_nOkuriNasiPos ;
	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk) ))
		return	FALSE ;

	while (cc = TBufStringMarker_GetChar (&mk), cc != (Char) -1) {
		if (cc != ';'){
			const Char*	ptr		= pKey ;
			int			nptr	= nKey ;

			while (nptr > 0){
				if (cc == (Char) -1 || cc != *ptr)
					break ;
				ptr		++ ;
				nptr	-- ;
				TBufStringMarker_Forward (&mk, 1) ;
				cc	= TBufStringMarker_GetChar (&mk) ;
			}
			if (nptr == 0) {
				if (TFAILED (TVarbuffer_Add (pvbuf, &chSEPARATOR, 1)) ||
					TFAILED (TVarbuffer_Add (pvbuf, pKey, nKey)) ||
					TFAILED (skkLocalJisyo_findCompletion (&mk, pvbuf)))
					return	FALSE ;
				continue ;
			}
		}
		/*	s܂ŃXLbvB*/
		while (cc != '\n' && cc != '\r' && cc != (Char) -1) {
			TBufStringMarker_Forward (&mk, 1) ;
			cc	= TBufStringMarker_GetChar (&mk) ;
		}
		/*	s̊Ԃ̓XLbvB*/
		while (cc == '\n' || cc == '\r') {
			TBufStringMarker_Forward (&mk, 1) ;
			cc	= TBufStringMarker_GetChar (&mk) ;
		}
		/*	obt@̍Ō܂œBH */
		if (cc == (Char) -1)
			break ;
	}
	if (TVarbuffer_GetUsage (pvbuf) > 0) 
		return	TVarbuffer_Add (pvbuf, &chSEPARATOR, 1) ;
	return	TRUE ;
}

/*	ϊʂ̓o^sBauto-okuri ɂ͑ΉȂB
 */
BOOL
skkLocalJisyo_record (
	SkkJisyo*		pSkkJisyo,
	const Char*	pKey,
	int			nKey,
	BOOL			fOkuri,
	const Char*	pResult,
	int			nResult)
{
	SkkLocalJisyo*	pJisyo	= (SkkLocalJisyo *)(pSkkJisyo + 1) ;
	int	nStartPos, nEndPos, nPos, nInsert, nQuote ;
	Char	cc ;
	BOOL	fRetval	= FALSE ;
	TVarbuffer			vbufWord ;
	TBufStringMarker	mk ;

	DEBUGPRINTF ((TEXT ("SkkLocalJisyo_Record (%p, %p, %d, %d, %p, %d)\n"),
				 pJisyo, pKey, nKey, fOkuri, pResult, nResult)) ;

	if (TFAILED (TVarbuffer_Initialize (&vbufWord, sizeof (Char))))
		return	FALSE ;

	/*	͕ quote B*/
	nQuote		= quoteCstring (NULL, pResult, nResult) ;
	if (nQuote > 0) {
		Char*	strWord ;
		
		if (TFAILED (TVarbuffer_Require (&vbufWord, nQuote)))
			return	FALSE ;

		strWord	= TVarbuffer_GetBuffer (&vbufWord) ;
		(void) quoteCstring (strWord, pResult, nResult) ;
		pResult	= strWord ;
		nResult	= nQuote ;
	}

	/*	key  result  check  unicode ɂKv邩
	 *	check B
	 *	 internal char ɕϊ unicode 瓮Ȃ
	 *	邩ǂA画fłB
	 */
	if (internalStringContainsUnicodeCharp (pKey, nKey) ||
		internalStringContainsUnicodeCharp (pResult, nResult))
		pJisyo->m_fUnicode	= TRUE ;

	nStartPos	= fOkuri? pJisyo->m_nOkuriAriPos  : pJisyo->m_nOkuriNasiPos ;
	nEndPos		= fOkuri? pJisyo->m_nOkuriNasiPos : INT_MAX ;
	nPos		= skkLocalJisyo_findCandidate (pJisyo, nStartPos, nEndPos, pKey, nKey) ;
	if (nPos < 0) {
		/*	̍s̐擪ɂ insert đvB*/
		nPos	= nStartPos ;
		if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk) ))
			goto	exit_record ;
		while (cc = TBufStringMarker_GetChar (&mk), 
			   cc != (Char) -1 && cc != '\n') {
			TBufStringMarker_Forward (&mk, 1) ;
			nPos	++ ;
		}
		if (cc == (Char) -1) {
			DEBUGPRINTF ((TEXT ("Fatal, unexpected end of buffer.\n"))) ;
			goto	exit_record ;
		}
		TBufStringMarker_Forward (&mk, 1) ;
		nPos	++ ;

		/*	`` //\n'' B*/
		TBufStringMarker_InsertString (&mk, pKey, nKey) ;
		TBufStringMarker_InsertChar   (&mk, ' ', 1) ;
		TBufStringMarker_InsertChar   (&mk, '/', 1) ;
		TBufStringMarker_InsertString (&mk, pResult, nResult) ;
		TBufStringMarker_InsertChar   (&mk, '/', 1) ;
		TBufStringMarker_InsertChar   (&mk, '\n', 1) ;
		nInsert	= nKey + 1 + 1 + nResult + 1 + 1 ;
	} else {
		const Char*	ptr ;
		int			nptr ;

		if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk) ))
			goto	exit_record ;
		while (cc = TBufStringMarker_GetChar (&mk), 
			   cc != (Char) -1 && cc != '/' && cc != '[' && cc != '\n') 
			TBufStringMarker_Forward (&mk, 1) ;
		if (cc != '/') {
			DEBUGPRINTF ((TEXT ("Fatal, separator not found.\n"))) ;
			goto	exit_record ;
		}
		
		/*	܂擪ɍ̕ϊʂ}B*/
		TBufStringMarker_InsertChar   (&mk, '/', 1) ;
		TBufStringMarker_InsertString (&mk, pResult, nResult) ;
		nInsert	= 1 + nResult ;

		/*	̈ʒu '/' ł͂܂ĂBs܂ł
		 *	ϊʂ邩ǂB*/
		TBufStringMarker_Forward (&mk, 1) ;
		do {
			ptr		= pResult ;
			nptr	= nResult ;
			/*	܂ '/' ܂łrBł '[' ͍B*/
			while (cc = TBufStringMarker_GetChar (&mk), 
				   cc != (Char)-1 && /*cc != '[' &&*/ cc != '/' && cc != '\n' &&
				   cc == *ptr && nptr > 0) {
				TBufStringMarker_Forward (&mk, 1) ;
				ptr		++ ;
				nptr	-- ;
			}
			if (nptr == 0) {
				cc = TBufStringMarker_GetChar (&mk) ;
				if (cc == '/') {
					/*	found */
					TBufStringMarker_Backward (&mk, nResult + 1) ;
					TBufStringMarker_DeleteChar (&mk, nResult + 1) ;
					nInsert	-= (nResult + 1) ;
					break ;
				}
			}
			/*	 '/' ܂œǂݐi߂B*/
			while (cc != '/' && cc != (Char)-1 && cc != '\n') {
				TBufStringMarker_Forward (&mk, 1) ;
				cc = TBufStringMarker_GetChar (&mk) ;
			}
			/*	'/' łԁAǂݔ΂B*/
			while (cc == '/') {
				TBufStringMarker_Forward (&mk, 1) ;
				cc = TBufStringMarker_GetChar (&mk) ;
			}
			/*	'/' ̒ '[' ͏IӖƂB*/
		} while (cc != (Char) -1 && cc != '\n' && cc != '[') ;

		/*	nKey ̂́A
		 *	 nPos 
		 *		ւ񂩂 /ϊ/
		 *	           
		 *	́uvƁu/v̊Ԃ̋󔒂wB
		 */
		fRetval	= skkLocalJisyo_moveLineToTop (pJisyo, nPos - nKey, fOkuri) ;
	}
	if (fOkuri)
		pJisyo->m_nOkuriNasiPos	+= nInsert ;
	pJisyo->m_fModified	= TRUE ;
	fRetval	= TRUE ;
 exit_record:
	TVarbuffer_Uninitialize (&vbufWord) ;
	return	fRetval ;
}

/*	ϊʂ̓o^sBauto-okuri ɂ͑ΉȂB
 */
BOOL
skkLocalJisyo_purge (
	SkkJisyo*		pSkkJisyo,
	const Char*		pKey,
	int				nKey,
	BOOL			fOkuri,
	const Char*		pResult,
	int				nResult)
{
	SkkLocalJisyo*	pJisyo	= (SkkLocalJisyo *)(pSkkJisyo + 1) ;
	int	nStartPos, nEndPos, nPos, nRemove, nCandPos, nLineTop ;
	Char	cc ;
	TBufStringMarker	mk ;

	nStartPos	= fOkuri? pJisyo->m_nOkuriAriPos  : pJisyo->m_nOkuriNasiPos ;
	nEndPos		= fOkuri? pJisyo->m_nOkuriNasiPos : INT_MAX ;
	nPos		= skkLocalJisyo_findCandidate (pJisyo, nStartPos, nEndPos, pKey, nKey) ;
	if (nPos < 0) 
		return	FALSE ;

	nLineTop	= nPos - nKey ;
	assert (nLineTop >= 0) ;

	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk)))
		return	FALSE ;
	while (cc = TBufStringMarker_GetChar (&mk), 
		   cc != (Char) -1 && cc != '/' && cc != '[' && cc != '\n') {
		TBufStringMarker_Forward (&mk, 1) ;
		nPos	++ ;
	}
	if (cc != '/')
		return	FALSE ;

	/*	̈ʒu '/' ł͂܂ĂBs܂ł
	 *	ϊʂ邩ǂB*/
	TBufStringMarker_Forward (&mk, 1) ;
	nPos	++ ;
	nCandPos	= nPos ;

	do {
		const Char*	ptr		= pResult ;
		int			nptr	= nResult ;

		while (cc = TBufStringMarker_GetChar (&mk), 
			   cc != (Char)-1 && cc != '/' && cc == *ptr && nptr > 0) {
			TBufStringMarker_Forward (&mk, 1) ;
			ptr		++ ;
			nptr	-- ;
		}
		if (nptr == 0) {
			cc = TBufStringMarker_GetChar (&mk) ;
			if (cc == '/') 
				break ;
		}
		while (cc != '/' && cc != (Char)-1 && cc != '\n') {
			TBufStringMarker_Forward (&mk, 1) ;
			cc = TBufStringMarker_GetChar (&mk) ;
		}
		if (cc == '/')
			TBufStringMarker_Forward (&mk, 1) ;
	} while (cc != (Char) -1 && cc != '\n' && cc != '[') ;

	if (cc == (Char) -1 || cc == '\n' || cc == '[')
		return	FALSE ;

	/*	foundBo^Ăϊʂ폜B*/
	TBufStringMarker_Backward   (&mk, nResult + 1) ;
	TBufStringMarker_DeleteChar (&mk, nResult + 1) ;
	nRemove	= nResult + 1 ;

	/*	SĂ̒Pꂪ폜ꂽ̂ǂ`FbNB*/
	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nCandPos, &mk)))
		return	FALSE ;
	
	/*	 '/' or '[' or '\n' Ȃ͊ĂB*/
	cc	= TBufStringMarker_GetChar (&mk) ;
	if (cc == '/' || cc == '[' || cc == '\n') {
		TBufStringMarker	mkLineTop ;
		int		nLineLen ;

		if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nLineTop, &mkLineTop)))
			return	FALSE ;

		mk			= mkLineTop ;
		nLineLen	= 0 ;
		while (cc = TBufStringMarker_GetChar (&mk), 
			   cc != (Char) -1 && cc != '\n') {
			TBufStringMarker_Forward (&mk, 1) ;
			nLineLen	++ ;
		}
		if (cc == '\n') 
			nLineLen	++ ;

		TBufStringMarker_DeleteChar (&mkLineTop, nLineLen) ;
		nRemove	+= nLineLen ;
	}
	if (fOkuri)
		pJisyo->m_nOkuriNasiPos	-= nRemove ;
	pJisyo->m_fModified	= TRUE ;
	return	TRUE ;
}

/*	ށAokurigana  length ő肠肩ǂ̔͂łȂcB skk-auto-okuri-eary-process
 *	 okurigana Ȃ̂ŁAokurigana length=0 ɂȂĂ܂B
 */
BOOL
skkLocalJisyo_recordEx (
	SkkJisyo*		pSkkJisyo,
	const Char*		pKey,
	int				nKey,
	const Char*		pResult,
	int				nResult,
	const Char*		pOkurigana,
	int				nOkurigana,
	BOOL			bOkuri)
{
	SkkLocalJisyo*	pJisyo	= (SkkLocalJisyo *)(pSkkJisyo + 1) ;
	int	nStartPos, nEndPos, nPos, nInsert, nQuote ;
	Char	cc ;
	BOOL	fRetval	= FALSE ;
	TVarbuffer			vbufWord ;
	TBufStringMarker	mk ;

	DEBUGPRINTF ((TEXT ("SkkLocalJisyo_RecordEx (%p, %p, %d, %d, %p, %d)\n"),
				 pJisyo, pKey, nKey, nOkurigana, pResult, nResult)) ;

	if (TFAILED (TVarbuffer_Initialize (&vbufWord, sizeof (Char))))
		return	FALSE ;

	/*	͕ quote B*/
	nQuote		= quoteCstring (NULL, pResult, nResult) ;
	if (nQuote > 0) {
		Char*	strWord ;
		
		if (TFAILED (TVarbuffer_Require (&vbufWord, nQuote)))
			return	FALSE ;

		strWord	= TVarbuffer_GetBuffer (&vbufWord) ;
		(void) quoteCstring (strWord, pResult, nResult) ;
		pResult	= strWord ;
		nResult	= nQuote ;
	}

	/*	key  result  check  unicode ɂKv邩
	 *	check B
	 *	 internal char ɕϊ unicode 瓮Ȃ
	 *	邩ǂA画fłB
	 */
	if (internalStringContainsUnicodeCharp (pKey, nKey) ||
		internalStringContainsUnicodeCharp (pResult, nResult))
		pJisyo->m_fUnicode	= TRUE ;

	/*	艼ϊƂĂA艼ȂƂ̂ŁAʂɑ艼
	 *	ϊǂ̃tbOnB
	 */
	if (! bOkuri)
		nOkurigana	= 0 ;
	nStartPos	= bOkuri? pJisyo->m_nOkuriAriPos  : pJisyo->m_nOkuriNasiPos ;
	nEndPos		= bOkuri? pJisyo->m_nOkuriNasiPos : INT_MAX ;
	nPos		= skkLocalJisyo_findCandidate (pJisyo, nStartPos, nEndPos, pKey, nKey) ;
	if (nPos < 0) {
		/*	̍s̐擪ɂ insert đvB*/
		nPos	= nStartPos ;
		if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk) ))
			goto	exit_record ;
		while (cc = TBufStringMarker_GetChar (&mk), 
			   cc != (Char) -1 && cc != '\n') {
			TBufStringMarker_Forward (&mk, 1) ;
			nPos	++ ;
		}
		if (cc == (Char) -1) {
			DEBUGPRINTF ((TEXT ("Fatal, unexpected end of buffer.\n"))) ;
			goto	exit_record ;
		}
		TBufStringMarker_Forward (&mk, 1) ;
		nPos	++ ;

		/*	`` //\n'' B*/
		nInsert	= 0 ;
		TBufStringMarker_InsertString (&mk, pKey, nKey) ;
		TBufStringMarker_InsertChar   (&mk, ' ', 1) ;
		TBufStringMarker_InsertChar   (&mk, '/', 1) ;
		TBufStringMarker_InsertString (&mk, pResult, nResult) ;
		nInsert	+= nKey + 1 + 1 + nResult ;

		/* [<okuri-char>/<result>/] */
		if (nOkurigana > 0) {
			TBufStringMarker_InsertChar   (&mk, '/', 1) ;
			TBufStringMarker_InsertChar   (&mk, '[', 1) ;
			TBufStringMarker_InsertString (&mk, pOkurigana, nOkurigana) ;
			TBufStringMarker_InsertChar   (&mk, '/', 1) ;
			TBufStringMarker_InsertString (&mk, pResult, nResult) ;
			TBufStringMarker_InsertChar   (&mk, '/', 1) ;
			TBufStringMarker_InsertChar   (&mk, ']', 1) ;
			nInsert	+= 1 + 1 + nOkurigana + 1 + nResult + 1 + 1 ;
		}
		TBufStringMarker_InsertChar   (&mk, '/', 1) ;
		TBufStringMarker_InsertChar   (&mk, '\n', 1) ;
		nInsert	+= 1 + 1 ;
	} else {
		const Char*	ptr ;
		int			nptr ;

		if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk) ))
			goto	exit_record ;
		while (cc = TBufStringMarker_GetChar (&mk), 
			   cc != (Char) -1 && cc != '/' && cc != '[' && cc != '\n') 
			TBufStringMarker_Forward (&mk, 1) ;
		if (cc != '/') {
			DEBUGPRINTF ((TEXT ("Fatal, separator not found.\n"))) ;
			goto	exit_record ;
		}
		
		/*	܂擪ɍ̕ϊʂ}B*/
		TBufStringMarker_InsertChar   (&mk, '/', 1) ;
		TBufStringMarker_InsertString (&mk, pResult, nResult) ;
		nInsert	= 1 + nResult ;

		/*	̈ʒu '/' ł͂܂ĂBs܂łɁuݓo^Ă̂Ɠ
		 *	ϊʁv邩ǂB
		 *	΍Ȃ΂ȂȂB
		 */
		TBufStringMarker_Forward (&mk, 1) ;
		do {
			ptr		= pResult ;
			nptr	= nResult ;
			/*	܂ '/' ܂łrBł '[' ͍B*/
			while (cc = TBufStringMarker_GetChar (&mk), 
				   cc != (Char)-1 && /*cc != '[' &&*/ cc != '/' && cc != '\n' &&
				   cc == *ptr && nptr > 0) {
				TBufStringMarker_Forward (&mk, 1) ;
				ptr		++ ;
				nptr	-- ;
			}
			if (nptr == 0) {
				cc = TBufStringMarker_GetChar (&mk) ;
				if (cc == '/') {
					/*	found */
					TBufStringMarker_Backward   (&mk, nResult + 1) ;
					TBufStringMarker_DeleteChar (&mk, nResult + 1) ;
					nInsert	-= (nResult + 1) ;
					break ;
				}
			}
			/*	 '/' ܂œǂݐi߂B*/
			while (cc != '/' && cc != (Char)-1 && cc != '\n') {
				TBufStringMarker_Forward (&mk, 1) ;
				cc = TBufStringMarker_GetChar (&mk) ;
			}
			/*	'/' łԁAǂݔ΂B*/
			while (cc == '/') {
				TBufStringMarker_Forward (&mk, 1) ;
				cc = TBufStringMarker_GetChar (&mk) ;
			}
			/*	'/' ̒ '[' ͏IӖƂB*/
		} while (cc != (Char) -1 && cc != '\n' && cc != '[') ;

		/*	艼݂̂Ȃ΁A"/[" ܂Ői߂B
		 */
		if (nOkurigana > 0 && cc != (Char)-1) {
			if (cc != '\n' && cc != '[') {
				for ( ; ; ) {
					while (cc == '/') {
						TBufStringMarker_Forward (&mk, 1) ;
						cc = TBufStringMarker_GetChar (&mk) ;
					}
					if (cc == (Char)-1 || cc == '\n' || cc == '[') 
						break ;
					while (cc != '/' && cc != '\n' && cc != (Char)-1) {
						TBufStringMarker_Forward (&mk, 1) ;
						cc = TBufStringMarker_GetChar (&mk) ;
					}
				}
			}
			if (cc == '[') {
				/* ̏ꍇɂ /[...]/ mFčsKvB*/
				do {
					TBufStringMarker_Forward (&mk, 1) ;

					ptr		= pOkurigana ;
					nptr	= nOkurigana ;
					/*	܂ '/' ܂łrBł '[' ͍B*/
					while (cc = TBufStringMarker_GetChar (&mk), 
						   cc != (Char)-1 && /*cc != '[' &&*/ cc != '/' && cc != '\n' &&
						   cc == *ptr && nptr > 0) {
						TBufStringMarker_Forward (&mk, 1) ;
						ptr		++ ;
						nptr	-- ;
					}
					if (nptr == 0) {
						cc = TBufStringMarker_GetChar (&mk) ;
						if (cc == '/') {
							/* 艼݂̂ŁA result ݂邩ׂB*/
							do {
								ptr		= pResult ;
								nptr	= nResult ;
								/*	܂ '/' ܂łrBł '[' ͍B*/
								while (cc = TBufStringMarker_GetChar (&mk), 
									   cc != (Char)-1 && /*cc != '[' &&*/ cc != '/' && cc != '\n' &&
									   cc == *ptr && nptr > 0) {
									TBufStringMarker_Forward (&mk, 1) ;
									ptr		++ ;
									nptr	-- ;
								}
								if (nptr == 0) {
									cc = TBufStringMarker_GetChar (&mk) ;
									if (cc == '/') {
										/* ǉ̕Kv͂ȂB*/
										goto	exit_loop ;
									}
								}
								/*	 '/' ܂œǂݐi߂B*/
								while (cc != '/' && cc != (Char)-1 && cc != '\n') {
									TBufStringMarker_Forward (&mk, 1) ;
									cc = TBufStringMarker_GetChar (&mk) ;
								}
								/*	'/' łԁAǂݔ΂B*/
								while (cc == '/') {
									TBufStringMarker_Forward (&mk, 1) ;
									cc = TBufStringMarker_GetChar (&mk) ;
								}
							} while (cc != (Char) -1 && cc != '\n' && cc != ']' && cc != '[') ;

							/* [<okuri>/.../] ɂȂ̂ŁAŌɒǉB*/
							if (cc == ']') {
								TBufStringMarker_InsertString (&mk, pResult, nResult) ;
								TBufStringMarker_InsertChar   (&mk, '/', 1) ;
								nInsert	+= nResult + 1 ;
							}
							goto	exit_loop ;
						}
					}
					/*	 '[' ܂œǂݐi߂B*/
					while (cc != '[' && cc != (Char)-1 && cc != '\n') {
						TBufStringMarker_Forward (&mk, 1) ;
						cc = TBufStringMarker_GetChar (&mk) ;
					}
				}	while (cc == '[') ;
			}
			if (cc != (Char)-1) {
				/* s܂Ői߂B*/
				while (cc != '\n' && cc != (Char)-1) {
					TBufStringMarker_Forward (&mk, 1) ;
					cc = TBufStringMarker_GetChar (&mk) ;
				}
				if (cc == '\n') {
					/* VKɒǉBA"/\n" ̌`ɂȂĂȂƑʖځB*/
					TBufStringMarker_Backward (&mk, 1) ;
					cc = TBufStringMarker_GetChar (&mk) ;
					if (cc == '/') {
						TBufStringMarker_Forward (&mk, 1) ;
						TBufStringMarker_InsertChar   (&mk, '[', 1) ;
						TBufStringMarker_InsertString (&mk, pOkurigana, nOkurigana) ;
						TBufStringMarker_InsertChar   (&mk, '/', 1) ;
						TBufStringMarker_InsertString (&mk, pResult, nResult) ;
						TBufStringMarker_InsertChar   (&mk, '/', 1) ;
						TBufStringMarker_InsertChar   (&mk, ']', 1) ;
						TBufStringMarker_InsertChar   (&mk, '/', 1) ;
						nInsert	+= 1 + nOkurigana + 1 + nResult + 1 + 1 + 1 ;
					}
				}
			}
		}
exit_loop:

		/*	nKey ̂́A
		 *	 nPos 
		 *		ւ񂩂 /ϊ/
		 *	           
		 *	́uvƁu/v̊Ԃ̋󔒂wB
		 */
		fRetval	= skkLocalJisyo_moveLineToTop (pJisyo, nPos - nKey, bOkuri) ;
	}
	if (bOkuri)
		pJisyo->m_nOkuriNasiPos	+= nInsert ;
	pJisyo->m_fModified	= TRUE ;
	fRetval	= TRUE ;
 exit_record:
	TVarbuffer_Uninitialize (&vbufWord) ;
	return	fRetval ;
}

/*	ϊʂ̓o^sBauto-okuri ɂ͑ΉȂB
 */
BOOL
skkLocalJisyo_purgeEx (
	SkkJisyo*		pSkkJisyo,
	const Char*		pKey,
	int				nKey,
	const Char*		pResult,
	int				nResult,
	const Char*		pOkurigana,
	int				nOkurigana,
	BOOL			bOkuri)
{
	SkkLocalJisyo*	pJisyo	= (SkkLocalJisyo *)(pSkkJisyo + 1) ;
	int		nStartPos, nEndPos, nPos, nRemove, nCandPos, nLineTop ;
	Char	cc ;
	TBufStringMarker	mk ;

	if (! bOkuri)
		nOkurigana	= 0 ;

	nStartPos	= bOkuri? pJisyo->m_nOkuriAriPos  : pJisyo->m_nOkuriNasiPos ;
	nEndPos		= bOkuri? pJisyo->m_nOkuriNasiPos : INT_MAX ;
	nPos		= skkLocalJisyo_findCandidate (pJisyo, nStartPos, nEndPos, pKey, nKey) ;
	if (nPos < 0) 
		return	FALSE ;

	nLineTop	= nPos - nKey ;
	assert (nLineTop >= 0) ;

	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk)))
		return	FALSE ;
	while (cc = TBufStringMarker_GetChar (&mk), 
		   cc != (Char) -1 && cc != '/' && cc != '[' && cc != '\n') {
		TBufStringMarker_Forward (&mk, 1) ;
		nPos	++ ;
	}
	if (cc != '/')
		return	FALSE ;

	/*	艼폜 nOkurigana > 0 ̏ꍇɂ́A[艼/P1/P2/.../] 
	 *	/.../ ̊Ԃ폜ɗ߂B
	 *	
	 */
	if (bOkuri && nOkurigana > 0) {
		if (skkLocalJisyo_bDeleteOkuriEntry (pJisyo, nPos, pOkurigana, nOkurigana, pResult, nResult, &nRemove))
			goto	exit_func ;
	} else {
		nRemove		= 0 ;
	}

	/*	̈ʒu '/' ł͂܂ĂBs܂ł
	 *	ϊʂ邩ǂB*/
	TBufStringMarker_Forward (&mk, 1) ;
	nPos		++ ;
	nCandPos	= nPos ;

	do {
		const Char*	ptr		= pResult ;
		int			nptr	= nResult ;

		while (cc = TBufStringMarker_GetChar (&mk), 
			   cc != (Char)-1 && cc != '/' && cc == *ptr && nptr > 0) {
			TBufStringMarker_Forward (&mk, 1) ;
			ptr		++ ;
			nptr	-- ;
		}
		if (nptr == 0) {
			cc = TBufStringMarker_GetChar (&mk) ;
			if (cc == '/') 
				break ;
		}
		while (cc != '/' && cc != (Char)-1 && cc != '\n') {
			TBufStringMarker_Forward (&mk, 1) ;
			cc = TBufStringMarker_GetChar (&mk) ;
		}
		if (cc == '/')
			TBufStringMarker_Forward (&mk, 1) ;
	} while (cc != (Char) -1 && cc != '\n' && cc != '[') ;

	if (cc == (Char) -1 || cc == '\n' || cc == '[')
		return	FALSE ;

	/*	foundBo^Ăϊʂ폜B*/
	TBufStringMarker_Backward   (&mk, nResult + 1) ;
	TBufStringMarker_DeleteChar (&mk, nResult + 1) ;
	nRemove	+= nResult + 1 ;

	/*	SĂ̒Pꂪ폜ꂽ̂ǂ`FbNB*/
	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nCandPos, &mk)))
		return	FALSE ;
	
	/*	 '/' or '[' or '\n' Ȃ͊ĂB*/
	cc	= TBufStringMarker_GetChar (&mk) ;
	if (cc == '/' || cc == '[' || cc == '\n') {
		TBufStringMarker	mkLineTop ;
		int		nLineLen ;

		if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nLineTop, &mkLineTop)))
			return	FALSE ;

		mk			= mkLineTop ;
		nLineLen	= 0 ;
		while (cc = TBufStringMarker_GetChar (&mk), 
			   cc != (Char) -1 && cc != '\n') {
			TBufStringMarker_Forward (&mk, 1) ;
			nLineLen	++ ;
		}
		if (cc == '\n') 
			nLineLen	++ ;

		TBufStringMarker_DeleteChar (&mkLineTop, nLineLen) ;
		nRemove	+= nLineLen ;
	}
exit_func:
	if (bOkuri)
		pJisyo->m_nOkuriNasiPos	-= nRemove ;
	pJisyo->m_fModified	= TRUE ;
	return	TRUE ;
}

BOOL
skkLocalJisyo_save (
	SkkJisyo*		pSkkJisyo)
{
	SkkLocalJisyo*	pJisyo	= (SkkLocalJisyo *)(pSkkJisyo + 1) ;
	FILE*		fp ;
	TBufStringMarker	mk ;
	char				buffer [1024] ;
	Char		cc ;
	char*		ptr ;
	int		nptr, nctext ;
	KANJISTATEMACHINE	ksm ;

	assert (pJisyo != NULL) ;

	if (! pJisyo->m_fModified)
		return	TRUE ;

	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, 0, &mk))) 
		return	FALSE ;

#if __STDC_WANT_SECURE_LIB__
	if (_tfopen_s (&fp, pJisyo->m_tszPath, TEXT ("wb")) != 0) {
		skkLocalJisyo_prepareDirectory (pJisyo->m_tszPath) ;
		if (_tfopen_s (&fp, pJisyo->m_tszPath, TEXT("wb")) != 0) {
			return	FALSE ;
		}
	}
#else
	fp		= _tfopen (pJisyo->m_tszPath, TEXT("wb")) ;
	if (fp == NULL) {
		skkLocalJisyo_prepareDirectory (pJisyo->m_tszPath) ;
		fp	= _tfopen (pJisyo->m_tszPath, TEXT("wb")) ;
		if (fp == NULL) {
			return	FALSE ;
		}
	}
#endif

	/*	 unicode ŎȂ΂ȂȂfB*/
	if (pJisyo->m_fUnicode) {
		/*	unicode ɂ̂ȂA0xFF 0xFE ŊJnKvB*/
		InitializeKanjiFiniteStateMachine (&ksm, KCODING_SYSTEM_UNICODE) ;
		buffer [0]	= (char)0xFF ;
		buffer [1]	= (char)0xFE ;
		fwrite (buffer, sizeof (char), 2, fp) ;
	} else {
		InitializeKanjiFiniteStateMachine (&ksm, KCODING_SYSTEM_ISO2022JP2) ;
	}
	ptr		= buffer ;
	nptr	= 0 ;
	/*	ށAcoding-system ǂ悤H */
	while (cc = TBufStringMarker_GetChar (&mk), cc != EOF) {
		TBufStringMarker_Forward (&mk, 1) ;
		nctext	= RtransferKanjiFiniteStateMachine (&ksm, cc, ptr) ;
		ptr		+= nctext ;
		nptr	+= nctext ;
		if (nptr < (sizeof (buffer) - 16)) {
			fwrite (buffer, 1, nptr, fp) ;
#if defined (DEBUG)
			fwrite (buffer, 1, nptr, stdout) ;
#endif
			ptr		= buffer ;
			nptr	= 0 ;
		}
	}
	if (nptr > 0) {
		fwrite (buffer, 1, nptr, fp) ;
#if defined (DEBUG)
		fwrite (buffer, 1, nptr, stdout) ;
#endif
	}
	fclose (fp) ;
	pJisyo->m_fModified	= FALSE ;
	return	TRUE ;
}

BOOL
skkLocalJisyo_destroy (
	SkkJisyo*		pSkkJisyo)
{
	SkkLocalJisyo*	pJisyo	= (SkkLocalJisyo *)(pSkkJisyo + 1) ;

	TBufString_Uninitialize (&pJisyo->m_bufJisyo) ;
	FREE (pSkkJisyo) ;
	return	TRUE ;
}

LPCTSTR
skkLocalJisyo_getPath (
	SkkJisyo*			pSkkJisyo)
{
	SkkLocalJisyo*	pJisyo	= (SkkLocalJisyo *)(pSkkJisyo + 1) ;

	return	pJisyo->m_tszPath ;
}

/*
 */
BOOL
skkLocalJisyo_load (
	SkkLocalJisyo*		pJisyo,
	LPCTSTR			pFileName)
{
	KFILE				kf ;
	TBufStringMarker	mk ;
	Char		cc ;
	BOOL		fResult ;
	int		nOkuriAri, nOkuriNasi ;
#if defined (DEBUG)
	int		nCount ;
#endif

	assert (pJisyo != NULL) ;
	assert (pFileName != NULL) ;

	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, 0, &mk) ))
		return	FALSE ;

	KFile_Init (&kf) ;
	if (TFAILED (KFile_Open (&kf, pFileName, -1))) 
		return	FALSE ;
	KFile_Rewind (&kf) ;

	/*	Unicode ꍇɂ́Asave ̎ɂ unicode ɂȂ悤A
	 *	LĂB
	 */
	pJisyo->m_fUnicode	= (KFile_GetCodingSystem (&kf) == KCODING_SYSTEM_UNICODE) ;
	
	fResult	= TRUE ;
#if defined (DEBUG)
	nCount	= 0 ;
#endif
	while (cc = KFile_Getc (&kf), cc != EOF) {
		if (TFAILED (TBufStringMarker_InsertChar (&mk, cc, 1))) {
			DEBUGPRINTF ((TEXT ("skkLocalJisyo_load: insert-char failed.\n"))) ;
			break ;
		}
#if defined (DEBUG)
		nCount	++ ;
#endif
	}
	fResult	= (cc == EOF) ;
	KFile_Close (&kf) ;
#if defined (DEBUG)
	DEBUGPRINTF ((TEXT ("skkLocalJisyo_load: read-chars = %d\n"), nCount)) ;
#endif

	/*	JnʒuƑȂJnʒuTB2݂Ȃ
	 *	ꍇɁu͉ĂvƂ݂ȂB
	 */
	nOkuriAri	= skkLocalJisyo_findAnsiAtLineTop (pJisyo, 0, OKURIARI_STR) ;
	nOkuriNasi	= skkLocalJisyo_findAnsiAtLineTop (pJisyo, nOkuriAri, OKURINASI_STR) ;
	DEBUGPRINTF ((TEXT ("OkuriAri = %d, OkuriNasi = %d\n"), nOkuriAri, nOkuriNasi)) ;
	if (nOkuriAri < 0 || nOkuriNasi < 0) 
		return	FALSE ;

	pJisyo->m_nOkuriAriPos	= nOkuriAri ;
	pJisyo->m_nOkuriNasiPos	= nOkuriNasi ;
	return	TRUE ;
}

/*	
 */
int
skkLocalJisyo_findAnsiAtLineTop (
	SkkLocalJisyo*	pJisyo,
	int			nStartPos,
	const char*	pString)
{
	TBufStringMarker		mk ;
	int			nPos ;
	const char*	ptr ;
	Char			cc, chTop ;

	assert (pJisyo  != NULL) ;
	assert (pString != NULL) ;

	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nStartPos, &mk)))
		return	-1 ;

	/*	s當񂪈v邩ǂ𔻒肷B*/
	nPos	= nStartPos ;
	chTop	= Char_MakeAscii (*pString) ;
	while (cc = TBufStringMarker_GetChar (&mk), cc != (Char) -1) {
		if (cc == chTop) {
			TBufStringMarker_Forward (&mk, 1) ;
			nPos	++ ;

			ptr		= pString + 1 ;
			while (*ptr != '\0') {
				cc	= TBufStringMarker_GetChar (&mk) ;
				DEBUGPRINTF ((TEXT ("findAnsiAtLineTop (%d): cc = 0x%08x, *ptr = %d\n"), nPos, (unsigned int)cc, (int)*ptr)) ;
				if (Char_DifferenceAscii (cc, *ptr) != 0)
					break ;
				TBufStringMarker_Forward (&mk, 1) ;
				nPos	++ ;
				ptr		++ ;
			}
			if (*ptr == '\0') {
				/*	tB*/
				return	nPos ;
			}
		}

		/*	̍s܂ňړB*/
		while (cc != (Char)-1 && cc != '\n') {
			TBufStringMarker_Forward (&mk, 1) ;
			nPos	++ ;
			cc = TBufStringMarker_GetChar (&mk) ;
		}
		TBufStringMarker_Forward (&mk, 1) ;
		nPos	++ ;
	}
	return	-1 ;
}

/*	[UԂɐݒ肷B
 */
BOOL
skkLocalJisyo_initJisyo (
	SkkLocalJisyo*	pJisyo)
{
	TBufStringMarker		mk ;
	const char*	ptr ;
	static const char		sstrInitialJisyoContent []	= OKURIARI_STR "\n" OKURINASI_STR "\n" ;
	static int				snOkuriNasiPos	= -1 ;

	assert (pJisyo != NULL) ;

	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, 0, &mk)))
		return	FALSE ;

	ptr	= sstrInitialJisyoContent ;
	while (*ptr != '\0') {
		if (TFAILED (TBufStringMarker_InsertChar (&mk, Char_MakeAscii (*ptr), 1)))
			return	FALSE ;
		ptr	++ ;
	}
	pJisyo->m_nOkuriAriPos	= 0 ;
	/*	ƋꂵR[h({̓RpCɕ
	 *	ǁA𔲂Ĉx strlen 𑖂点B*/
	if (snOkuriNasiPos < 0)
		snOkuriNasiPos		= strlen (OKURIARI_STR "\n") ;
	pJisyo->m_nOkuriNasiPos	= snOkuriNasiPos ;

	pJisyo->m_fUnicode		= TRUE ;	/* VKɍꍇɂ unicode */
	return	TRUE ;
}

/*	
 */
int
skkLocalJisyo_findCandidate (
	SkkLocalJisyo*	pJisyo,
	int			nStartPos,
	int			nEndPos,
	const Char*	pString,
	int			nString)
{
	TBufStringMarker		mk ;
	int			nPos, nptr ;
	const Char*	ptr ;
	Char			chTop, cc ;

	assert (pJisyo  != NULL) ;
	assert (pString != NULL) ;

	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nStartPos, &mk)))
		return	-1 ;

	/*	s當񂪈v邩ǂ𔻒肷B*/
	nPos	= nStartPos ;
	chTop	= *pString ;
	while (cc = TBufStringMarker_GetChar (&mk), 
		   nPos < nEndPos && cc != (Char) -1) {
		if (cc == chTop) {
			TBufStringMarker_Forward (&mk, 1) ;
			nPos	++ ;

			ptr		= pString + 1 ;
			nptr	= nString - 1 ;
			while (nptr > 0) {
				cc	= TBufStringMarker_GetChar (&mk) ;
				if (cc != *ptr)
					break ;
				TBufStringMarker_Forward (&mk, 1) ;
				nPos	++ ;
				ptr		++ ;
				nptr	-- ;
			}
			if (nptr <= 0) {
				/*	tB*/
				cc	= TBufStringMarker_GetChar (&mk) ;
				if (cc == ' ')
					return	nPos ;
			}
		}
		
		/*	̍s܂ňړB*/
		while (cc != '\n' && cc != (Char) -1 && nPos < nEndPos) {
			TBufStringMarker_Forward (&mk, 1) ;
			nPos	++ ;
			cc = TBufStringMarker_GetChar (&mk) ;
		}
		TBufStringMarker_Forward (&mk, 1) ;
		nPos	++ ;
	}
	return	-1 ;
}

/*	okuri-search p̌B
 */
int
skkLocalJisyo_findOkuriCandidate (
	SkkLocalJisyo*	pJisyo,
	int			nStartPos,
	int			nEndPos,
	const Char*	pString,
	int			nString,
	int*		piMatchLen)
{
	TBufStringMarker		mk ;
	int			nPos, nptr ;
	const Char*	ptr ;
	Char			chTop, cc ;

	assert (pJisyo  != NULL) ;
	assert (pString != NULL) ;

	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nStartPos, &mk)))
		return	-1 ;

	/*	s當񂪈v邩ǂ𔻒肷B*/
	nPos	= nStartPos ;
	chTop	= *pString ;
	while (cc = TBufStringMarker_GetChar (&mk), nPos < nEndPos && cc != (Char) -1) {
		if (cc == chTop) {
			TBufStringMarker_Forward (&mk, 1) ;
			nPos	++ ;

			ptr		= pString + 1 ;
			nptr	= nString - 1 ;
			while (nptr > 0) {
				cc	= TBufStringMarker_GetChar (&mk) ;
				if (cc != *ptr) {
					if (Char_IsAscii (cc) && 'a' <= cc && cc <= 'z') {
						TBufStringMarker_Forward (&mk, 1) ;
						nPos	++ ;
						cc	= TBufStringMarker_GetChar (&mk) ;
						if (cc == ' ') {
							if (piMatchLen != NULL)
								*piMatchLen	= ptr - pString ;
							return	nPos ;
						}
					}
					break ;
				}
				TBufStringMarker_Forward (&mk, 1) ;
				nPos	++ ;
				ptr		++ ;
				nptr	-- ;
			}
		}
		
		/*	̍s܂ňړB*/
		while (cc != '\n' && cc != (Char) -1 && nPos < nEndPos) {
			TBufStringMarker_Forward (&mk, 1) ;
			nPos	++ ;
			cc = TBufStringMarker_GetChar (&mk) ;
		}
		TBufStringMarker_Forward (&mk, 1) ;
		nPos	++ ;
	}
	return	-1 ;
}

BOOL
skkLocalJisyo_findCompletion (
	TBufStringMarker*	pMarker,
	TVarbuffer*		pvbuf)
{
	Char	cc ;
	int	nPos ;
	Char			rszBuffer [16] ;

	nPos	= 0 ;
	while (cc = TBufStringMarker_GetChar (pMarker),
		   cc != (Char)-1 && cc != ' ' && cc != '[' && cc != '\n' && cc != '\r') {
		if (nPos >= ARRAYSIZE (rszBuffer)) {
			if (TFAILED (TVarbuffer_Add (pvbuf, rszBuffer, nPos)))
				return	FALSE ;
			nPos	= 0 ;
		}
		TBufStringMarker_Forward (pMarker, 1) ;
		rszBuffer [nPos ++]	= cc ;
	}
	if (nPos > 0) {
		if (TFAILED (TVarbuffer_Add (pvbuf, rszBuffer, nPos)))
			return	FALSE ;
	}
	while (cc != '\n' && cc != '\r' && cc != (Char) -1) {
		TBufStringMarker_Forward (pMarker, 1) ;
		cc	= TBufStringMarker_GetChar (pMarker) ;
	}
	while (cc == '\n' || cc == '\r') {
		TBufStringMarker_Forward (pMarker, 1) ;
		cc	= TBufStringMarker_GetChar (pMarker) ;
	}
	return	TRUE ;
}

BOOL
skkLocalJisyo_prepareDirectory (
	LPCTSTR	pPath)
{
	LPCTSTR	ptr ;
	LPCTSTR	pLast ;
	LPTSTR		pDir ;
	TCHAR				szDir [MAX_PATH + 1] ;

	if (pPath == NULL)
		return	FALSE ;

	pLast	= NULL ;
	ptr		= pPath ;
	pDir	= szDir ;
	while (*ptr != TEXT ('\0')) {
		while (*ptr != TEXT ('\0') && *ptr != TEXT ('\\')) {
			*pDir ++	= *ptr ++ ;
		}
		if (*ptr == TEXT ('\\')) {
			*pDir		= TEXT ('\0') ;
#if defined (DEBUG) || 1
			DebugPrintfToFile (TEXT ("CreateDierctory => \"%s\"\n"), szDir) ;
			if (! CreateDirectory (szDir, NULL)) {
				DWORD	dwError	= GetLastError () ;
				DebugPrintfToFile (TEXT ("CreateDirectory failed (%d/0x%0x)\n"), dwError, dwError) ;
			}
#else
			CreateDirectory (szDir, NULL) ;
#endif
			*pDir ++	= *ptr ++ ;
		}
	}
	return	TRUE ;
}

/*	w肳ꂽʒus܂łw肳ꂽʒuɈړB
 */
BOOL
skkLocalJisyo_moveLineToTop (
	SkkLocalJisyo*	pJisyo,
	int			nFrom,
	BOOL			fOkuri)
{
	TVarbuffer			vbufLine ;
	Char		cc ;
	Char				rCBuf [16] ;
	Char*		ptr ;
	int		nptr, nDelete, nTo ;
	BOOL		fRetval	= FALSE ;
	TBufStringMarker	mk ;
	TBufStringMarker	mkDest ;

	DEBUGPRINTF ((TEXT ("moveLineToTop (%d, %d)\n"), nFrom, fOkuri)) ;

	if (TFAILED (TVarbuffer_Initialize (&vbufLine, sizeof (Char))))
		return	FALSE ;
	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nFrom, &mk))) {
		DEBUGPRINTF ((TEXT ("Can't solve marker (%d)\n"))) ;
		return	FALSE ;
	}

	ptr		= rCBuf ;
	nptr	= ARRAYSIZE (rCBuf) ;
	nDelete	= 0 ;
	while (cc = TBufStringMarker_GetChar (&mk), cc != (Char)-1) {
		*ptr	++ = cc ;
		nptr	-- ;
		if (nptr <= 0) {
			/*	ꕶǉ`JԂƏd16
			 *	܂Ƃ߂Ăi[邱ƂɂĂ݂B*/
			if (TFAILED (TVarbuffer_Add (&vbufLine, rCBuf, ARRAYSIZE (rCBuf)))) {
				DEBUGPRINTF ((TEXT ("Add Line Failure\n"))) ;
				return	FALSE ;
			}
			ptr		= rCBuf ;
			nptr	= ARRAYSIZE (rCBuf) ;
		}
		TBufStringMarker_Forward (&mk, 1) ;
		nDelete	++ ;
		/*	ړɂ͉s܂߂̂ŁAsR[hoꍇ
		 *	[v𔲂̂́ÄʒuB*/
		if (cc == '\n')
			break ;
	}
	if (ptr != rCBuf)
		if (TFAILED (TVarbuffer_Add (&vbufLine, rCBuf, ptr - rCBuf))) {
			DEBUGPRINTF ((TEXT ("Add Line Failure\n"))) ;
			return	FALSE ;
		}

	/*	ړ̍s폜B*/
	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nFrom, &mk)) ||
		TFAILED (TBufStringMarker_DeleteChar   (&mk, nDelete))) {
		DEBUGPRINTF ((TEXT ("Delete or Solev failure\n"))) ;
		goto	Exit_Func ;
	}

	nTo	= (fOkuri)? pJisyo->m_nOkuriAriPos	: pJisyo->m_nOkuriNasiPos ;
	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nTo, &mkDest))) {
		DEBUGPRINTF ((TEXT ("Can't solve marker (%d)\n"), nTo)) ;
		goto	Exit_Func ;
	}
	while (cc = TBufStringMarker_GetChar (&mkDest), 
		   cc != (Char) -1) {
		TBufStringMarker_Forward (&mkDest, 1) ;
		if (cc == '\n')
			break ;
	}
	if (cc != '\n') {
		DEBUGPRINTF ((TEXT ("Insertion Point Error? (0x%lx)\n"), cc)) ;
		goto	Exit_Func ;
	}
	/*	ړɃRs[Ăs̒g}B*/
	ptr		= TVarbuffer_GetBuffer (&vbufLine) ;
	nptr	= TVarbuffer_GetUsage  (&vbufLine) ;
#if defined (DEBUG)
	/*	肭ĂȂ悤Ȃ̂ŁAǂ̕񂪈ړ
	 *	̂܂\B*/
	{
		TCHAR	wbuf [512] ;

		ConvertCStrToTStr (ptr, nptr, wbuf, ARRAYSIZE (wbuf)) ;
		wbuf [ARRAYSIZE (wbuf) - 1]	= L'\0' ;
		DEBUGPRINTF ((TEXT ("MoveLine = \"%s\"\n"), wbuf)) ;
	}
#endif
	fRetval	= TBufStringMarker_InsertString (&mkDest, ptr, nptr) ;
 Exit_Func:
	TVarbuffer_Uninitialize (&vbufLine) ;
	return	fRetval ;
}

BOOL
skkLocalJisyo_bDeleteOkuriEntry (
	SkkLocalJisyo*	pJisyo,
	int				nPos,
	const Char*		pOkurigana,
	int				nOkurigana,
	const Char*		pResult,
	int				nResult,
	int*			pnRemove)
{
	int					nRemove ;
	Char				cc ;
	TBufStringMarker	mk ;
	BOOL				bRetval	= FALSE ;

	nRemove	= 0 ;
	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk)))
		goto	exit_loop ;

	cc	= TBufStringMarker_GetChar (&mk) ;
	while (cc != (Char)-1 && cc != '\n') {
		Char	ccNext ;
		TBufStringMarker_Forward (&mk, 1) ;
		ccNext	= TBufStringMarker_GetChar (&mk) ;
		if (cc == '/' && ccNext == '[') 
			break ;
		cc		= ccNext ;
	}
	if (cc == (Char)-1 || cc == '\n')
		goto	exit_loop ;

	/*	[ sxA艼v邩ׂB
	 */
	for ( ; ; ) {
		int		nCount ;

		nCount	= skkLocalJisyo_iCompareMarkerWithString (&mk, pOkurigana, nOkurigana) ;
		if (nCount == 0) {
			cc = TBufStringMarker_GetChar (&mk) ;
			if (cc == '/') {
				int		nWords = 0 ;

				/* ̂ŁAx͒Pꂪ݂邩ǂ`FbNB*/
				TBufStringMarker_Forward (&mk, 1) ;
				for ( ; ; ) {
					nCount	= skkLocalJisyo_iCompareMarkerWithString (&mk, pResult, nResult) ;
					if (nCount == 0) {
						cc = TBufStringMarker_GetChar (&mk) ;
						if (cc == '/') {
							/* result + '/' Ȃ΂ȂȂB*/
							TBufStringMarker_Backward   (&mk, nResult) ;
							TBufStringMarker_DeleteChar (&mk, nResult + 1) ;
							nRemove	+= nResult + 1 ;

							/*	A entry  [艼/] ƂƂɂȂĂ܂̂ȂA
							 *	폜Ȃ΂ȂȂB
							 */
							cc = TBufStringMarker_GetChar (&mk) ;
							if (cc == ']' && nWords == 0) {
								/*	``[艼/]'' 폜B
								 */
								TBufStringMarker_Backward   (&mk, nOkurigana + 2) ;
								if (TBufStringMarker_GetChar (&mk) == '[') {
									TBufStringMarker_DeleteChar (&mk, nOkurigana + 3) ;	/* ``['' + 艼 + ''/]'' */ 
									nRemove	+= nOkurigana + 3 ;
								} else {
									/* ͒vIȃG[ƎvB*/
								}
							}
							/*  result ݂Ȃ΁ASȂƂȂB*/
							bRetval	= skkLocalJisyo_bOkuriEntryExistp (pJisyo, nPos, pResult, nResult) ;
							goto	exit_loop ;
						}
					}
					while (cc != '/' && cc != (Char)-1 && cc != '\n' && cc != ']') {
						TBufStringMarker_Forward (&mk, 1) ;
						cc = TBufStringMarker_GetChar (&mk) ;
					}
					if (cc != '/')
						break ;

					while (cc == '/') {
						TBufStringMarker_Forward (&mk, 1) ;
						cc = TBufStringMarker_GetChar (&mk) ;
					}
					nWords	++ ;
				}
				goto	exit_loop ;
			}
		}
		/*  "]/[" TB*/
		if (! skkLocalJisyo_bSkipNextOkuriEntry (&mk))
			break ;
	}

exit_loop:
	*pnRemove	= nRemove ;
	return	bRetval ;
}

BOOL
skkLocalJisyo_bOkuriEntryExistp (
	SkkLocalJisyo*	pJisyo,
	int				nPos,
	const Char*		pResult,
	int				nResult)
{
	TBufStringMarker	mk ;
	Char				cc ;

	if (TFAILED (TBufString_SolveMarker (&pJisyo->m_bufJisyo, nPos, &mk)))
		return	FALSE ;

	do {
		int		nCount ;

		/*	艼̒g͊mFȂŁAŏ̌܂ňړB
		 */
		cc = TBufStringMarker_GetChar (&mk) ;
		while (cc != '/' && cc != (Char)-1 && cc != '\n' && cc != ']') {
			TBufStringMarker_Forward (&mk, 1) ;
			cc = TBufStringMarker_GetChar (&mk) ;
		}
		if (cc == '/') {
			/* ̂ŁAx͒Pꂪ݂邩ǂ`FbNB*/
			TBufStringMarker_Forward (&mk, 1) ;

			cc	= TBufStringMarker_GetChar (&mk) ;
			while (cc != (Char)-1 && cc != '\n' && cc != ']') {
				nCount	= skkLocalJisyo_iCompareMarkerWithString (&mk, pResult, nResult) ;
				if (nCount == 0) 
					return	TRUE ;

				cc = TBufStringMarker_GetChar (&mk) ;
				while (cc != '/' && cc != (Char)-1 && cc != '\n' && cc != ']') {
					TBufStringMarker_Forward (&mk, 1) ;
					cc = TBufStringMarker_GetChar (&mk) ;
				}
				if (cc != '/') {
					/* vIȃG[݂ƎvB*/
					return	FALSE ;
				}
				TBufStringMarker_Forward (&mk, 1) ;
				cc	= TBufStringMarker_GetChar (&mk) ;
			}
		}
		if (! skkLocalJisyo_bSkipNextOkuriEntry (&mk))
			return	FALSE ;
	}	while (cc != (Char)-1) ;

	return	FALSE ;
}

int
skkLocalJisyo_iCompareMarkerWithString (
	TBufStringMarker*		pmk,
	const Char*				pString,
	int						nString)
{
	Char	cc ;

	while (nString > 0) {
		/*  pString  unicode Ǝ buffer  jisx0208-1983 ̔rŎsĂB*/
		cc = TBufStringMarker_GetChar (pmk) ;
		if (cc != *pString)
			break ;
		TBufStringMarker_Forward (pmk, 1) ;
		pString	++ ;
		nString	-- ;
	}
	return	nString ;
}

BOOL
skkLocalJisyo_bSkipNextOkuriEntry (
	TBufStringMarker*		pmk)
{
	Char	cc ;

	/*  "]/[" TB*/
	cc = TBufStringMarker_GetChar (pmk) ;

	while (cc != (Char)-1 && cc != '\n') {
		while (cc != (Char)-1 && cc != ']' && cc != '\n') {
			TBufStringMarker_Forward (pmk, 1) ;
			cc = TBufStringMarker_GetChar (pmk) ;
		}
		if (cc != ']')
			break ;

		TBufStringMarker_Forward (pmk, 1) ;
		cc = TBufStringMarker_GetChar (pmk) ;
		if (cc != '/') 
			continue ;

		TBufStringMarker_Forward (pmk, 1) ;
		cc = TBufStringMarker_GetChar (pmk) ;
		if (cc == '[') {
			TBufStringMarker_Forward (pmk, 1) ;
			return	TRUE ;
		}
	}
	return	FALSE ;
}

BOOL
skkLocalJisyo_bVerify (
	SkkLocalJisyo*	pJisyo)
{
	BOOL	bRetval	= TRUE ;
	int		nOkuriAri, nOkuriNasi ;

	/*	okuri-ari, okuri-nashi ̓ƂĂ̂`FbNB
	 */
	nOkuriAri	= skkLocalJisyo_findAnsiAtLineTop (pJisyo, 0, OKURIARI_STR) ;
	nOkuriNasi	= skkLocalJisyo_findAnsiAtLineTop (pJisyo, nOkuriAri, OKURINASI_STR) ;
	if (nOkuriAri != pJisyo->m_nOkuriAriPos) {
		DEBUGPRINTF ((TEXT ("OkuriAriPos error (%d != %d)\n"), nOkuriAri, pJisyo->m_nOkuriAriPos)) ;
		bRetval	= FALSE ;
	}
	if (nOkuriNasi != pJisyo->m_nOkuriNasiPos) {
		DEBUGPRINTF ((TEXT ("OkuriNasiPos error (%d != %d)\n"), nOkuriNasi, pJisyo->m_nOkuriNasiPos)) ;
		bRetval	= FALSE ;
	}
	if (nOkuriAri > nOkuriNasi) {
		DEBUGPRINTF ((TEXT ("OkuriNasiPos < OkuriAriPos (%d < %d)\n"), nOkuriNasi, nOkuriAri)) ;
	}
	return	bRetval ;
}

