
;;

(defmacro when (cond &rest body)
  (list 'if cond (cons 'progn body)))

(defmacro unless (cond &rest body)
  (cons 'if (cons cond (cons nil body))))

(defvar save-match-data-internal)
(defmacro save-match-data (&rest body)
  (` (let ((save-match-data-internal (match-data)))
       (unwind-protect
	   (progn (,@ body))
	 (set-match-data save-match-data-internal)))))
	 
;;

(setq skk-num-alist-type1
  '((?0 . "O") (?1 . "P") (?2 . "Q") (?3 . "R")
    (?4 . "S") (?5 . "T") (?6 . "U") (?7 . "V")
    (?8 . "W") (?9 . "X")
    (?. . "D")
    (?  . "")))

(setq skk-num-type-alist
  '((0 . identity)
    (1 . skk-num-jisx0208-latin)
    (2 . skk-num-type2-kanji)
    (3 . skk-num-type3-kanji)
    (4 . skk-num-recompute)
    (5 . skk-num-type5-kanji)
    (9 . skk-num-shogi)))

(setq skk-num-alist-type2
  '((?0 . "Z") (?1 . "") (?2 . "") (?3 . "O")
    (?4 . "l") (?5 . "") (?6 . "Z") (?7 . "")
    (?8 . "") (?9 . "") (?\  . "")))

(setq skk-num-alist-type3
  (append
   '((ju . "\") (hyaku . "S") (sen . "")
     (man . "") (oku . "") (cho . "") (kei . ""))
   skk-num-alist-type2))

(setq skk-num-alist-type5
  '((ju . "E") (hyaku . "S") (sen . "")
    (man . "") (oku . "") (cho . "") (kei . "")
    (?0 . "") (?1 . "") (?2 . "") (?3 . "Q")
    (?4 . "l") (?5 . "") (?6 . "Z") (?7 . "")
    (?8 . "") (?9 . "") (?\  . "")))
    
(defun skk-num-get-suuji (expression alist)
  (cdr (assq expression alist)))

(defun skk-num-int-p (num)
  (not (string-match "\\.[0-9]" num)))
  
(defun skk-num-exp (num type)
  (save-match-data
    (let ((fun (cdr (assq type skk-num-type-alist))))
      (when fun
	(funcall fun num)))))

(defun skk-num-jisx0208-latin (num)
  (let ((candidate
	 (mapconcat (function
		     (lambda (c)
		       (skk-num-get-suuji c skk-num-alist-type1)))
		    num "")))
    (if (not (string= candidate ""))
      candidate)))

(defun skk-num-type2-kanji (num)
  (save-match-data
    (when (skk-num-int-p num)
      (let ((candidate
	     (mapconcat (function
			 (lambda (c)
			   (skk-num-get-suuji
			    c
			    skk-num-alist-type2)))
			num "")))
	(if (not (string= candidate ""))
	  candidate)))))

(defun skk-num-type3-kanji (num)
  (save-match-data
    (when (skk-num-int-p num)
      ;; _܂܂Ȃ
      (skk-num-to-kanji num 'type3))))

(defun skk-num-type5-kanji (num)
  (save-match-data
    (when (skk-num-int-p num)
      ;; _܂܂Ȃ
      (skk-num-to-kanji num 'type5))))

(defun skk-num-to-kanji (num type &optional alist)
  (let ((len (length num))
	(i 0)
	char v num1 v1)
    (unless alist
      (setq alist
	    (symbol-value
	     (intern (format "skk-num-alist-%s" type)))))
    ;; u狞v܂ł͏o͂B
    (when (> len 20)
      (skk-error "ʂ傫܂I" "Too big number!"))
    (setq num (append num nil))
    (cond
     ((<= len 4)
      (while (setq char (car num))
	;; :     \  S  
	;; len:   1   2   3   4
	(cond
	 ((= len 1)
	  ;; ʂ\킷ȊO̊B
	  (unless (eq char ?0)
	    ;; ̈ʂ 0 łȂB
	    (setq v (concat v (skk-num-get-suuji char alist)))))
	 (t
	  ;; ʂ\킷ȊO̊B
	  (unless (or (and (eq type 'type3)
			   (memq char '(?0 ?1)))
		      (and (eq type 'type5)
			   (eq char ?0)))
	    ;; type3 ̂Ƃ́A\̈ʈȏŁA 0, 1 ȊO̐B
	    ;; type5 ̂Ƃ́A\̈ʈȏŁA 0 ȊO̐B
	    (setq v (concat v (skk-num-get-suuji char alist))))
	  ;; ʂ\킷B
	  (when (and (not (eq char ?0)) (< 1 len))
	    (setq v
		  (concat
		   v
		   (skk-num-get-suuji
		    (cond ((eq len 2) 'ju) ; \
			  ((eq len 3) 'hyaku) ; S
			  (t 'sen)) ; 
		    alist))))))
	(setq len (1- len) num (cdr num))))
     (t
      (setq num (nreverse num))
      (while num
	(setq num1 nil)
	(while (and (< (length num1) 4) num)
	  (setq num1 (cons (car num) num1)
		num (cdr num)))
	(when num1
	  (setq v1 (skk-num-to-kanji num1 type alist))
	  (when (string= v1 (skk-num-get-suuji ?0 alist))
	    (setq v1 ""))
	  (when (and (eq type 'type3) (eq i 1) (equal v1 ""))
	    ;; {ł́u牭vƂ\͂ƂɎg邪Au疜vƂ\
	    ;; ͂܂gȂ̂ŁAu疜vɒB
	    (setq v1 (concat "" v1)))
	  (setq
	   v
	   (concat
	    v1
	    (when v1
	      (skk-num-get-suuji
	       (cond ((eq i 0) ?\ )
		     ((eq i 1) 'man) ; 
		     ((eq i 2) 'oku) ; 
		     ((eq i 3) 'cho) ; 
		     ((eq i 4) 'kei)) ; 
	       alist))
	    v)))
	(setq i (1+ i)))))
    ;;
    (or v
	(skk-num-get-suuji ?0 alist))))

(defun skk-num-shogi (num)
  (save-match-data
    (when (and (= (length num) 2)
	       (skk-num-int-p num))
      (let ((candidate
	     (concat (skk-num-get-suuji (aref num 0) skk-num-alist-type1)
		     (skk-num-get-suuji (aref num 1) skk-num-alist-type2))))
	(unless (string= candidate "")
	  candidate)))))
	  
;;

(defun skk-current-date (&optional pp-function format and-time)
  (unless pp-function
    (setq pp-function skk-default-current-date-function))
  (funcall pp-function
	   (skk-current-date-1) format skk-date-ad and-time))

(defun skk-current-date-1 (&optional specified-time)
  (let* ((str (current-time-string specified-time))
	 (date (substring str 8 10)))
    (when (eq (aref date 0) ?\040)
      (setq date (substring date 1)))
    (list
     (substring str 20 24)
     (substring str 4 7)
     date
     (substring str 0 3)
     (substring str 11 13)
     (substring str 14 16)
     (substring str 17 19))))

(defun skk-default-current-date
  (date-information
   format num-type gengo gengo-index month-alist-index dayofweek-alist-index
   &optional and-time)
  (let ((year (car date-information))
	(month (nth 1 date-information))
	(day (nth 2 date-information))
	(day-of-week (nth 3 date-information))
	(hour (nth 4 date-information))
	(minute (nth 5 date-information))
	(second (nth 6 date-information))
	v)
    (when gengo
      (setq v (skk-ad-to-gengo-1 (string-to-number year))))
    (setq year (if gengo
		   (concat (if gengo-index
			       (nth gengo-index (car v))
			     (car (car v)))
			   (skk-num-exp (number-to-string (cdr v))
					num-type))
		 (skk-num-exp year num-type)))
    (when month-alist-index
      (setq month (skk-num-exp (nth month-alist-index
				    (cdr (assoc month skk-month-alist)))
			       num-type)))
    (setq day (skk-num-exp day num-type))
    (when dayofweek-alist-index
      (setq day-of-week (nth dayofweek-alist-index
			     (cdr (assoc day-of-week skk-day-of-week-alist)))))
    (when and-time
      (setq hour (skk-num-exp hour num-type))
      (setq minute (skk-num-exp minute num-type))
      (setq second (skk-num-exp second num-type)))
    (if and-time
	(format (or format "%sN%s%s(%s)%s%s%sb")
		year month day day-of-week hour minute second)
      (format (or format "%sN%s%s(%s)") year month day day-of-week))))

(defun skk-ad-to-gengo (gengo-index &optional divider tail not-gannen)
  ;; ɕϊBIvV divider w肳Ă΁AN
  ;; ̊ԂɁAtail w肳Ă΁A̖ɁAꂼ̕A
  ;; B
  ;; o;
  ;; ꂫ#˂ /(skk-ad-to-gengo 0 nil "N")/(skk-ad-to-gengo 0 " " " N")/
  (let ((v (skk-ad-to-gengo-1 (string-to-number
			       (car skk-num-list))
			      not-gannen)))
    (concat (nth gengo-index (car v))
	    divider
	    (if (not (stringp (cdr v)))
		(number-to-string (cdr v))
	      (cdr v))
	    tail)))

;;;###autoload
(defun skk-ad-to-gengo-1 (ad &optional not-gannen)
  ;; AD is a number and NOT-GANNEN is a boolean optional
  ;; arg.
  ;; return a cons cell of which car is a Gengo list
  ;; gotten from `skk-gengo-alist', and cdr is a number
  ;; of year.
  ;; if NOT-GANNEN is non-nil and calculated year is 1,
  ;; return a value of which cdr is "" (string).
  (when (>= 1866 ad)
    (skk-error "܂" "Unkown year"))
  (cons (cond ((>= 1911 ad)
	       (setq ad (- ad 1867))
	       (cdr (assq 'meiji skk-gengo-alist)))
	      ((>= 1925 ad)
	       (setq ad (- ad 1911))
	       (cdr (assq 'taisho skk-gengo-alist)))
	      ((>= 1988 ad)
	       (setq ad (- ad 1925))
	       (cdr (assq 'showa skk-gengo-alist)))
	      (t
	       (setq ad (- ad 1988))
	       (cdr (assq 'heisei skk-gengo-alist))))
	(cond (not-gannen ad)
	      ((= ad 1) "")
	      (t ad))))

;;;###autoload
(defun skk-gengo-to-ad (&optional head tail)
  ;; 𐼗ɕϊBIvV HEAD, TAIL w肳Ă
  ;; ΁A̕擪AɘAB
  ;; o;
  ;;   傤#˂ /(skk-gengo-to-ad "" "N")/(skk-gengo-to-ad "" " N")/\
  ;;   (skk-gengo-to-ad "" "N")/(skk-gengo-to-ad "" " N")/
  (save-match-data
    (when (string-match (car skk-num-list)
			skk-henkan-key)
      (let ((v (skk-gengo-to-ad-1
		(substring skk-henkan-key 0 (match-beginning 0))
		(string-to-number (car skk-num-list)))))
	(when v
	  (concat head
		  (number-to-string v)
		  tail))))))

;;;###autoload
(defun skk-gengo-to-ad-1 (gengo number)
  ;; GENGO is a string and NUMBER is a number.
  ;; return a year (number) equal to GENGO-NUMBER.
  (+ number
     (cond
      ((eq number 0)
       (skk-error "0 N͂蓾Ȃ"
		  "Cannot convert 0 year"))
      ((member gengo '("ւ" ""))
       1988)
      ((member gengo '("傤" "a"))
       (if (> 64 number)
	   1925
	 (skk-error "a 63 N܂łł"
		    "The last year of Showa is 63")))
      ((member gengo '("傤" "吳"))
       (if (> 15 number)
	   1911
	 (skk-error "吳 14 N܂łł"
		    "The last year of Taisyo is 14")))
      ((member gengo '("߂" ""))
       (if (> 45 number)
	   1867
	 (skk-error " 44 N܂łł"
		    "The last year of Meiji is 44")))
      (t
       (skk-error "ʕs\ȌłI"
		  "Unknown Gengo!")))))

;;;###autoload
(defun skk-calc (operator)
  ;; 2 ̈ operator ̌vZB
  ;; : '/ ͈ƂēnȂ̂ (defalias 'div '/) ȂǂƂAʂ̌`
  ;; skk-calc ɓnB
  ;; o; #*# /(skk-calc '*)/
  (number-to-string (apply operator
			   (mapcar 'string-to-number
				   skk-num-list))))

;;;###autoload
(defun skk-plus ()
  ;; o; #+#+# /(skk-plus)/
  (skk-calc '+))

;;;###autoload
(defun skk-minus ()
  (skk-calc '-))

;;;###autoload
(defun skk-times ()
  (skk-calc '*))

;;; SKK-GADGET.EL related.
(setq skk-gengo-alist
  '((heisei "" "H") (showa "a" "S") (taisho "吳" "T")
    (meiji "" "M")))

(setq skk-month-alist
  '(("Jan" "1" "Januar") ("Feb" "2" "Februar") ("Mar" "3" "Mrz")
    ("Apr" "4" "April") ("May" "5" "Mai")
    ("Jun" "6" "Juni") ("Jul" "7" "Juli") ("Aug" "8" "August")
    ("Sep" "9" "September") ("Oct" "10" "Oktober")
    ("Nov" "11" "November") ("Dec" "12" "Dezember")))

(setq skk-day-of-week-alist
  '(("Sun" "" "So") ("Mon" "" "Mo") ("Tue" "" "Di") ("Wed" "" "Mi")
    ("Thu" "" "Do") ("Fri" "" "Fr") ("Sat" "y" "Sa")))

(setq skk-default-current-date-function
  '(lambda (date-information format gengo and-time)
      (skk-default-current-date date-information nil 0 gengo 0 0 0 and-time)))

(setq skk-date-ad nil)

(setq skk-number-style 1)

(setq skk-units-alist
  '(("mile" ("km" . 1.6093) ("yard" . 1760))
    ("yard" ("feet" . 3) ("cm" . 91.44))
    ("feet" ("inch" . 12) ("cm" . 30.48))
    ("inch" ("feet" . 0.5) ("cm" . 2.54))))


