#if defined (UNITTEST)
#include "StdAfx.h"
#else
#include "local.h"
#endif
#include "ImeDoc.h"
#include "ImeBufferP.h"
#include "TSearchSession.h"
#include "keymap.h"
#include "jstring.h"
#include "TMarker.h"
#include "lmstate.h"
#include "TMSG.h"
#include "ImeConfig.h"

static	int		LM_bSkkInputByCodeOrMenu0		(struct CImeDoc*) ;
static	int		LM_bSkkInputByCodeOrMenuJump	(struct CImeDoc*) ;
static	int		LM_bSkkInputByCodeOrMenu1		(struct CImeDoc*) ;

#if !defined (UNITTEST)
static	BOOL	bParseKutenCode					(LPCWSTR, int, int*, int*) ;
#endif

/*================================================================ skk-input-by-code-or-menu
 */
static	int		LM_bSkkInputByCodeOrMenu_1		(struct CImeDoc*) ;
static	int		LM_bSkkInputByCodeOrMenu_2		(struct CImeDoc*) ;
static	int		LM_bSkkInputByCodeOrMenu_3		(struct CImeDoc*) ;
static	int		LM_bSkkInputByCodeOrMenu_Exit	(struct CImeDoc*) ;

int
LM_bSkkInputByCodeOrMenu (
	struct CImeDoc*				pDoc)
{
	int		iKcodeCharset ;
	WCHAR	bufText   [MAXCOMPLEN] ;
	int		nText ;
	LPWSTR	pwText ;

	iKcodeCharset	= ImeConfig_iGetSkkKcodeCharset () ;

	/*	JIS  EUC Ȃ΋_R[h݂͑邾낤ǁAUnicode ɂ͑ȂBƁAShiftJIS œꂽƂ
	 *	v]͂Ȃ̂낤H
	 *
	 *	JIS  EUC  KUTEN ǉBłȂꍇɂ͒ڃR[ĥ݂ƂAȁBskk-kcode-charset
	 *	 Enum ^ɂāA
	 *
	 *		japanese-jisx0208, katakana-jisx0201
	 *
	 *	cȂقǁAEUC ͂Ȃ̂BAjisx0201 āAǂē͂́H ̂͏CȂƑʖ
	 *	B܂Ajapanese-jis  unicode 炢ɊɂݒcƂH
	 *
	 *	_R[hƂ̂́Axx-xx Ƃ`炵B
	 */
	switch (iKcodeCharset) {
	case	KCODE_CHARSET_KATAKANA_JISX0201:
		/* ͂Ȃ񂶂cB*/
		nText	= wnsprintfW (bufText, ARRAYSIZE (bufText) - 1, L"7/8 bits code for %s (nn or CR for Jump Menu): ", L"katakana-jisx0201") ;
		break ;
	case	KCODE_CHARSET_JAPANESE_JISX0208:
		/* _OK */
		nText	= wnsprintfW (bufText, ARRAYSIZE (bufText) - 1, L"7/8 bits or KUTEN code for %s (00nn or CR for Jump Menu): ", L"japanese-jisx0208") ;
		break ;
	case	KCODE_CHARSET_JAPANESE_JISX0212:
		/* _OK */
		nText	= wnsprintfW (bufText, ARRAYSIZE (bufText) - 1, L"7/8 bits or KUTEN code for %s (00nn or CR for Jump Menu): ", L"japanese-jisx0212") ;
		break ;
	case	KCODE_CHARSET_UNICODE:
	default:
		/* _͎gȂB*/
		nText	= wnsprintfW (bufText, ARRAYSIZE (bufText) - 1, L"8 bits code for %s (00nn or CR for Jump Menu): ", L"unicode") ;
		break ;
	}
	pwText		= ImeDoc_pAlloca (pDoc, nText		* sizeof (WCHAR), __alignof (WCHAR)) ;
	if (pwText == NULL)
		return	LMR_ERROR ;
	if (! ImeDoc_bPushReg (pDoc, LMREG_0))
		return	LMR_ERROR ;
	memcpy (pwText, bufText, nText * sizeof (WCHAR)) ;
	ImeDoc_vSetRegConstString (pDoc, LMREGARG_0,	pwText,		nText) ;
	ImeDoc_vSetRegBool (pDoc, LMREGARG_1, LFALSE) ;
	if (! ImeDoc_bCall (pDoc, LM_bReadFromMinibuffer, LM_bSkkInputByCodeOrMenu_1))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkInputByCodeOrMenu_Exit (
	struct CImeDoc*		pDoc)
{
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

int
LM_bSkkInputByCodeOrMenu_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPCWSTR	pReadTop, pReadEnd, pRead ;
	LPCWSTR	pwResult ;
	int		nRead, nResult ;

	if (ImeDoc_bSignalp (pDoc)) {
		if (ImeDoc_nGetSignal (pDoc) == LMSIGNAL_QUIT) 
			ImeDoc_vClearSignals (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL || 
		! ImeDoc_bGetRegConstString (pDoc, LMREGARG_RETVAL, &pwResult, &nResult)) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_Exit) ;
		return	LMR_CONTINUE ;
	}

	/* Ő󔒂͏ׂȁcB*/
	pReadEnd	= pwResult + nResult - 1 ;
	pReadTop	= pwResult ;
	while (pReadEnd >= pReadTop && iswspace (*pReadEnd))
		pReadEnd	-- ;
	while (pReadTop <= pReadEnd && iswspace (*pReadTop))
		pReadTop	++ ;
	pRead		= pReadTop ;
	nRead		= (pReadTop <= pReadEnd)? ((pReadEnd - pReadTop) + 1) : 0 ;
	if (nRead > 0) {
		int		n1, n2, ch, iKcodeCharset ;
		BOOL	bKuten, bSingleByte ;
		WCHAR	wch ;

		iKcodeCharset	= ImeConfig_iGetSkkKcodeCharset () ;
		switch (iKcodeCharset) {
		case	KCODE_CHARSET_KATAKANA_JISX0201:
			/* ͂Ȃ񂶂cB*/
			bKuten		= FALSE ;
			bSingleByte	= TRUE ;
			break ;
		case	KCODE_CHARSET_JAPANESE_JISX0208:
			/* _OK */
			bKuten		= TRUE ;
			bSingleByte	= FALSE ;
			break ;
		case	KCODE_CHARSET_JAPANESE_JISX0212:
			/* _OK */
			bKuten		= TRUE ;
			bSingleByte	= FALSE ;
			break ;
		case	KCODE_CHARSET_UNICODE:
		default:
			/* _͎gȂB*/
			bKuten		= FALSE ;
			bSingleByte	= FALSE ;
			break ;
		}

		/* ̏ꍇɂ́Aparse B*/
		if (bKuten && 
#if defined (UNITTEST)
			_snwscanf_s (pRead, nRead, L"%d-%d", &n1, &n2) == 2
#else
			bParseKutenCode (pRead, nRead, &n1, &n2)
#endif
			) {
			n1	+= 32 + 128 ;
			n2	+= 32 + 128 ;
		} else {
			if (bSingleByte) {
				if (! bSkkHexChar (pRead [0]) || ! bSkkHexChar (pRead [1])) {
					n1	= -1 ;
				} else {
					n1	= iSkkCharToHex (pRead [0]) * 16 + iSkkCharToHex (pRead [1]) ;
				}
				n2	= 0 ;
			} else {
				if (nRead < 4) {
					n1	= 0 ;
				} else {
					if (! bSkkHexChar (pRead [0]) || ! bSkkHexChar (pRead [1])) {
						n1	= -1 ;
					} else {
						n1	= iSkkCharToHex (pRead [0]) * 16 + iSkkCharToHex (pRead [1]) ;
					}
				}
				if (nRead < 2) {
					n2	= 0 ;
				} else {
					if (! bSkkHexChar (pRead [2]) || ! bSkkHexChar (pRead [3])) {
						n1	= -1 ;
					} else {
						n2	= iSkkCharToHex (pRead [2]) * 16 + iSkkCharToHex (pRead [3]) ;
					}
				}
			}
			if (n1 < 0 || n2 < 0 || n1 >= 256 || n2 >= 256) {
				ImeDoc_bSetMessage (pDoc, L"Invalid code") ;
				ImeDoc_vSetSignalError (pDoc) ;
				ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_Exit) ;
				return	LMR_CONTINUE ;
			}
		}
		/* ͉\ȕ낤H */
		if (! bMakeUnicodeChar (iKcodeCharset, n1, n2, &ch)) {
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_0, n1) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_1, n2) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkInputByCodeOrMenu0, LM_bSkkInputByCodeOrMenu_2))
				return	LMR_ERROR ;
			return	LMR_CONTINUE ;
		}
		wch	= ch ;
		if (! ImeBuffer_bInsertAndInherit (pBuffer, &wch, 1)) {
			ImeDoc_vSetSignalError (pDoc) ;
			ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_Exit) ;
			return	LMR_CONTINUE ;
		}
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_3) ;
		return	LMR_CONTINUE ;
	} else {
		/* ڃR[h͂ɂȂB*/
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_0, 0) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_1, 0) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkInputByCodeOrMenu0, LM_bSkkInputByCodeOrMenu_2))
			return	LMR_ERROR ;
		return	LMR_CONTINUE ;
	}
}

int
LM_bSkkInputByCodeOrMenu_2 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	WCHAR	wch ;
	int		nCH ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegInteger (pDoc, LMREGARG_RETVAL, &nCH)) {
		ImeDoc_bSetMessage (pDoc, L"Invalid code") ;
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_Exit) ;
		return	LMR_CONTINUE ;
	}
	wch	= nCH ;
	if (! ImeBuffer_bInsertAndInherit (pBuffer, &wch, 1)) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_Exit) ;
		return	LMR_CONTINUE ;
	}
	ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_3) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkInputByCodeOrMenu_3 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_Exit) ;
		return	LMR_CONTINUE ;
	}
	/* ... */
	ImeDoc_vClearMessage (pDoc) ;

	if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkInputByCodeOrMenu_Exit))
			return	LMR_ERROR ;
	} else {
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu_Exit) ;
	}
	return	LMR_CONTINUE ;
}

/*================================================================ skk-input-by-code-or-menu0
 */
static	int		LM_bSkkInputByCodeOrMenu0_1		(struct CImeDoc*) ;

int
LM_bSkkInputByCodeOrMenu0 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	BOOL	bRetval	= FALSE ;
	int		n1, n2 ;

	if (ImeDoc_bSignalp (pDoc)) {
		return	LMR_RETURN ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bGetRegInteger (pDoc, LMREGARG_0, &n1) || 
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_1, &n2)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	pBuffer->m_bSkkInputByCodeOrMenuMode	= TRUE ;
	switch (ImeConfig_iGetSkkKcodeCharset ()) {
	case	KCODE_CHARSET_KATAKANA_JISX0201:
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_0, n1) ;
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_1, n2) ;
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_2, 0xA1) ;
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_3, 0xDF) ;
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_4, 0x00) ;
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_5, 0x00) ;
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_6, 0xA1) ;
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_7, 0xDF ) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkInputByCodeOrMenu1, LM_bSkkInputByCodeOrMenu0_1))
			return	LMR_ERROR ;
		return	LMR_CONTINUE ;

	case	KCODE_CHARSET_JAPANESE_JISX0208:
	case	KCODE_CHARSET_JAPANESE_JISX0212:
		if (n1 == 0) {
			/* 94 WȂ̂ŁA0x21  94 B*/
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_0, n2) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_1, 0x2121 | 0x8080) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_2, 0x7F7F | 0x8080) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_3, 0x80 + 0x21) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_4, 0x80 + 0x21 + 94 - 1) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_5, 0x80 + 0x21) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_6, 0x80 + 0x21 + 94 - 1) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkInputByCodeOrMenuJump, LM_bSkkInputByCodeOrMenu0_1))
				return	LMR_ERROR ;
		} else {
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_0, n1) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_1, n2) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_2, 0x2121 | 0x8080) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_3, 0x7F7F | 0x8080) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_4, 0x80 + 0x21) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_5, 0x80 + 0x21 + 94 - 1) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_6, 0x80 + 0x21) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_7, 0x80 + 0x21 + 94 - 1) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkInputByCodeOrMenu1, LM_bSkkInputByCodeOrMenu0_1))
				return	LMR_ERROR ;
		}
		return	LMR_CONTINUE ;

	case	KCODE_CHARSET_UNICODE:
		/* 0x0021 - 0xFFEE */
		if (n1 == 0) {
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_0, n2) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_1, 0x0021) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_2, 0xFFFD) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_3, 0x00) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_4, 0xFF) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_5, 0x00) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_6, 0xFF) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkInputByCodeOrMenuJump, LM_bSkkInputByCodeOrMenu0_1))
				return	LMR_ERROR ;
		} else {
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_0, n1) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_1, n2) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_2, 0x0021) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_3, 0xFFFD) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_4, 0x00) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_5, 0xFF) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_6, 0x00) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_7, 0xFF) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkInputByCodeOrMenu1, LM_bSkkInputByCodeOrMenu0_1))
				return	LMR_ERROR ;
		}
		return	LMR_CONTINUE ;

	default:
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_RETVAL, -1) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu0_1) ;
		return	LMR_CONTINUE ;
	}
}

int
LM_bSkkInputByCodeOrMenu0_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) {
		pBuffer->m_bSkkInputByCodeOrMenuMode	= FALSE ;
	}
	return	LMR_RETURN ;
}

/*================================================================ skk-input-by-code-or-menu-jump
 */
static	int		LM_bSkkInputByCodeOrMenuJump_1		(struct CImeDoc*) ;
static	int		LM_bSkkInputByCodeOrMenuJump_2		(struct CImeDoc*) ;
static	int		LM_bSkkInputByCodeOrMenuJump_Exit	(struct CImeDoc*) ;

/*
	int			n1,
	int			nCodeMin, 
	int			nCodeMax,
	int			nSkkCodeN1Min,
	int			nSkkCodeN1Max,
	int			nSkkCodeN2Min,
	int			nSkkCodeN2Max,
*/
int
LM_bSkkInputByCodeOrMenuJump (
	struct CImeDoc*			pDoc)
{
	int		n1, nCodeMin, nCodeMax, nSkkCodeN1Min, nSkkCodeN1Max, nSkkCodeN2Min, nSkkCodeN2Max ;
	int*	pnData ;
#if !defined (UNITTEST)
	int		iKcodeCharset ;
	struct CImeBuffer*	pBuffer ;
#endif

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	if (! ImeDoc_bGetRegInteger (pDoc, LMREGARG_0, &n1) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_1, &nCodeMin) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_2, &nCodeMax) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_3, &nSkkCodeN1Min) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_4, &nSkkCodeN1Max) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_5, &nSkkCodeN2Min) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_6, &nSkkCodeN2Max)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	pnData	= ImeDoc_pAlloca (pDoc, sizeof (int) * 10, __alignof (int)) ;
	if (pnData == NULL)
		return	LMR_ERROR ;

#if !defined (UNITTEST)
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	iKcodeCharset	= ImeConfig_iGetSkkKcodeCharset () ;
	if (! imeBuffer_bShowInputByCodeOrMenuJump (pBuffer, iKcodeCharset, n1, nCodeMin, nCodeMax, nSkkCodeN1Min, nSkkCodeN1Max, nSkkCodeN2Min, nSkkCodeN2Max)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
#endif

#if !defined (UNITTEST)
	pnData [0]	= (n1 < nSkkCodeN1Min)? nSkkCodeN1Min : n1 ;
	pnData [1]	= nSkkCodeN2Min ;	/* n2 */
#else
	pnData [0]	= n1 ;
	pnData [1]	= 0 ;	/* n2 */
#endif
	pnData [2]	= n1 ;	/* n1orig */
	pnData [3]	= 0 ;	/* n2orig */
	pnData [4]	= nCodeMin ;
	pnData [5]	= nCodeMax ;
	pnData [6]	= nSkkCodeN1Min ;
	pnData [7]	= nSkkCodeN1Max ;
	pnData [8]	= nSkkCodeN2Min ;
	pnData [9]	= nSkkCodeN2Max ;
	if (! ImeDoc_bPushReg (pDoc, LMREG_0))
		return	LMR_ERROR ;
	ImeDoc_vSetRegPointer (pDoc, LMREG_0, pnData) ;
	ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenuJump_1) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkInputByCodeOrMenuJump_Exit (
	struct CImeDoc*			pDoc)
{
#if !defined (UNITTEST)
	TVarbuffer*		pvbufCandInfo ;

	/*	CLOSE  CREATE ɗ̂Av낤H
	 *	ށAkeyboard quit ̖肪ȁB̎ close ͕KvȁB
	 *	Ƃ킯ŁA Close B
	 */
	ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_CLOSE_CANDIDATELIST) ;

	pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
	if (pvbufCandInfo != NULL)
		TVarbuffer_Clear (pvbufCandInfo) ;
#endif
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	return	LMR_RETURN ;
}

int
LM_bSkkInputByCodeOrMenuJump_1 (
	struct CImeDoc*			pDoc)
{
#if defined (UNITTEST)
	struct CImeBuffer*	pBuffer ;
	int			nKeys1 ;
	int*		pnData ;
	LPCWSTR		pwKeys1 ;
	int			n1, n2, nCodeMin, nCodeMax, nSkkCodeN1Min, nSkkCodeN1Max, nSkkCodeN2Min, nSkkCodeN2Max ;
	int			nSkkCodeN2Last, iKcodeCharset, nN1Org, nN2Org, i ;
	WCHAR		wbuf [512] ;
	LPWSTR		pDest, pDestEnd ;
#endif

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenuJump_Exit) ;
		return	LMR_CONTINUE ;
	}

#if defined (UNITTEST)
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	pwKeys1	= ImeConfig_pGetSkkInputByCodeMenuKeys1 (&nKeys1) ;
	if (pBuffer == NULL || pwKeys1 == NULL || nKeys1 <= 0) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenuJump_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegPointer (pDoc, LMREG_0, (void**)&pnData)) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenuJump_Exit) ;
		return	LMR_CONTINUE ;
	}

	n1				= pnData [0] ;
	nCodeMin		= pnData [4] ;
	nCodeMax		= pnData [5] ;
	nSkkCodeN1Min	= pnData [6] ;
	nSkkCodeN1Max	= pnData [7] ;
	nSkkCodeN2Min	= pnData [8] ;
	nSkkCodeN2Max	= pnData [9] ;

	if (n1 < nSkkCodeN1Min) {
		n1	= nSkkCodeN1Min ;	/* skk-input-by-code-or-menu-jump-default */ ;
	}
	n2				= nSkkCodeN2Min ;
	nSkkCodeN2Last	= nSkkCodeN2Min + ((nSkkCodeN2Max - nSkkCodeN1Min) & 0xFFFFFF00U) ;
	pDestEnd		= wbuf + ARRAYSIZE (wbuf) ;
	iKcodeCharset	= ImeConfig_iGetSkkKcodeCharset () ;

	nN1Org			= n1 ;
	nN2Org			= n2 ;

	pDest	= wbuf ;
	for (i = 0 ; i < nKeys1 && pDest < pDestEnd ; i ++) {
		int		ch ;
#define	SAFE_COPY(ptr,chara)	{ if (ptr < pDestEnd) { *ptr ++ = (chara) ; } } 
		SAFE_COPY (pDest, towupper (pwKeys1 [i])) ;
		SAFE_COPY (pDest, L':') ;
		if (bMakeUnicodeChar (iKcodeCharset, n1, n2, &ch)) {
			SAFE_COPY (pDest, ch) ;
		} else {
			SAFE_COPY (pDest, L'?') ;
			SAFE_COPY (pDest, L'?') ;
			SAFE_COPY (pDest, L'?') ;
		}
		SAFE_COPY (pDest, L' ') ;
		SAFE_COPY (pDest, L' ') ;
#undef	SAFE_COPY

		n2	+= 16 ;
		if (n2 > nSkkCodeN2Max) {
			n1	++ ;
			if (n1 > nSkkCodeN1Max)
				n1	= nSkkCodeN1Min ;
			n2	=	nSkkCodeN2Min ;
		}
		/* unicode ̎ɕKvB*/
		if (((n1 << 8) | n2) > nCodeMax) {
			n1	= (nCodeMin >> 8) & 0xFF ;
			n2	= (nCodeMin     ) & 0xFF ;
		}
	}
	ImeDoc_bSetMessageN (pDoc, wbuf, pDest - wbuf) ;
	pnData [0]	= n1 ;
	pnData [1]	= n1 ;
	pnData [2]	= nN1Org ;
	pnData [3]	= nN2Org ;
#endif
	if (! ImeDoc_bCall (pDoc, LM_bNextCommandEvent, LM_bSkkInputByCodeOrMenuJump_2))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkInputByCodeOrMenuJump_2 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	struct TMSG	msg ;
	int			nCH, nKeys1 ;
	int*		pnData ;
	LPCWSTR		pwKeys1 ;
	int			n1, n2, nCodeMin, nCodeMax, nSkkCodeN1Min, nSkkCodeN1Max, nSkkCodeN2Min, nSkkCodeN2Max ;
#if defined (UNITTEST)
	int			nN1Org, nN2Org ;
#endif
#if !defined (UNITTEST)
	TVarbuffer*	pvbufCandInfo ;
#endif
	WCHAR		bufText [256] ;
	int			nText ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenuJump_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	pwKeys1	= ImeConfig_pGetSkkInputByCodeMenuKeys1 (&nKeys1) ;
	if (pBuffer == NULL || pwKeys1 == NULL || nKeys1 <= 0) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenuJump_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegMsg (pDoc, LMREGARG_RETVAL, &msg) ||
		! ImeDoc_bGetRegPointer (pDoc, LMREG_0, (void**)&pnData)) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenuJump_Exit) ;
		return	LMR_CONTINUE ;
	}

	n1				= pnData [0] ;
	n2				= pnData [1] ;
#if defined (UNITTEST)
	nN1Org			= pnData [2] ;
	nN2Org			= pnData [3] ;
#endif
	nCodeMin		= pnData [4] ;
	nCodeMax		= pnData [5] ;
	nSkkCodeN1Min	= pnData [6] ;
	nSkkCodeN1Max	= pnData [7] ;
	nSkkCodeN2Min	= pnData [8] ;
	nSkkCodeN2Max	= pnData [9] ;

	if (! ImeDoc_bEventToCharacter (&msg, &nCH)) {
		/* "`%s' is not valid here!" */
		int	iFuncNo ;

		if (ImeDoc_bLookupKeymap (pDoc, &msg, &iFuncNo) && 
			(iFuncNo == NFUNC_ABORT_RECURSIVE_EDIT || iFuncNo == NFUNC_KEYBOARD_QUIT)) {
			ImeDoc_vSetSignal (pDoc, LMSIGNAL_QUIT) ;
			ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
			ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenuJump_Exit) ;
			return	LMR_CONTINUE ;
		}
#if defined (UNITTEST)
		n1	= nN1Org ;
		n2	= nN2Org ;
#endif
	} else {
		int		i, nSkkCodeN2Last ;

		nSkkCodeN2Last	= nSkkCodeN2Min + ((nSkkCodeN2Max - nSkkCodeN2Min) & 0xFFFFFFF0U) ;

		/* skk-input-by-code-menu-keys1 ɂ邩ǂ`FbNB*/
		for (i = 0 ; i < nKeys1 ; i ++) {
			if (towlower (nCH) == towlower (pwKeys1 [i]))
				break ;
		}
		if (i >= nKeys1) {
			if (ImeDoc_bSkkPreviousCandidateCharp (pDoc, nCH)) {
				/* n1, n2 ߂B*/
#if defined (UNITTEST)
				n1	= nN1Org ;
				n2	= nN2Org ;
#endif
				for (i = 0 ; i < nKeys1 ; i ++) {
					n2	-= 16 ;
					if (n2 < nSkkCodeN2Min) {
						n1	-- ;
						if (n1 < nSkkCodeN1Min)
							n1	= nSkkCodeN1Max ;
						n2	= nSkkCodeN2Last ;
					}
					/* unicode ̎ɕKvB*/
					if (((n1 << 8) | n2) < nCodeMin) {
						n1	= (nCodeMax >> 8) & 0xFF ;
						n2	= (nCodeMax     ) & 0xFF ;
					}
				}
#if !defined (UNITTEST)
				pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
				if (pvbufCandInfo != NULL && TVarbuffer_GetUsage (pvbufCandInfo) >= sizeof (MYCAND)) {
					LPMYCAND	pMyCand ;

					pMyCand		= (LPMYCAND) TVarbuffer_GetBuffer (pvbufCandInfo) ;
					if (pMyCand != NULL) {
						LPCANDIDATELIST	pCandidateList	= (LPCANDIDATELIST)&(pMyCand->cl) ;
						if (pCandidateList->dwPageStart >= (DWORD)nKeys1) {
							pCandidateList->dwPageStart	-= nKeys1 ;
						} else {
							pCandidateList->dwPageStart	= (pCandidateList->dwCount + pCandidateList->dwPageStart - nKeys1) ;
						}
						pCandidateList->dwSelection	= pCandidateList->dwPageStart ;
					}
					ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_UPDATE_CANDIDATELIST) ;
				}
#endif
			} else if (ImeConfig_bSkkStartHenkanCharp (nCH)) {
				/*  n1, n2  ok */
#if !defined (UNITTEST)
				for (i = 0 ; i < nKeys1 ; i ++) {
					n2	+= 16 ;
					if (n2 > nSkkCodeN2Max) {
						n1	++ ;
						if (n1 > nSkkCodeN1Max)
							n1	= nSkkCodeN1Min ;
						n2	= nSkkCodeN2Min ;
					}
					/* unicode ̎ɕKvB*/
					if (((n1 << 8) | n2) > nCodeMax) {
						n1	= (nCodeMin  >> 8) & 0xFF ;
						n2	= (nCodeMin      ) & 0xFF ;
					}
				}
				pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
				if (pvbufCandInfo != NULL && TVarbuffer_GetUsage (pvbufCandInfo) >= sizeof (MYCAND)) {
					LPMYCAND	pMyCand ;

					pMyCand		= (LPMYCAND) TVarbuffer_GetBuffer (pvbufCandInfo) ;
					if (pMyCand != NULL) {
						LPCANDIDATELIST	pCandidateList	= (LPCANDIDATELIST)&(pMyCand->cl) ;
						pCandidateList->dwPageStart	+= nKeys1 ;
						if (pCandidateList->dwPageStart >= pCandidateList->dwCount) {
							pCandidateList->dwPageStart	-= pCandidateList->dwCount ;
						}
						pCandidateList->dwSelection	= pCandidateList->dwPageStart ;
					}
					ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_UPDATE_CANDIDATELIST) ;
				}
#endif
			} else if (nCH == L'?') {
				int		iKcodeCharset, wch ;
#if defined (UNITTEST)
				n1	= nN1Org ;
				n2	= nN2Org ;
				/* code \ hit any key ƁB*/
#endif
				iKcodeCharset	= ImeConfig_iGetSkkKcodeCharset () ;
				if (bMakeUnicodeChar (iKcodeCharset, n1, n2, &wch)) {
					nText	= wnsprintf (bufText, ARRAYSIZE (bufText), L"`%c' EUC: %2x%2x (%3d, %3d), JIS: %2x%2x (%3d, %3d), UNICODE: %04x",
						wch, n1, n2, n1, n2, (n1 & 0x7F), (n2 & 0x7F), (n1 & 0x7F), (n2 & 0x7F), wch) ;
				} else {
					nText	= wnsprintf (bufText, ARRAYSIZE (bufText), L"`?' EUC: %2x%2x (%3d, %3d), JIS: %2x%2x (%3d, %3d)",
						n1, n2, n1, n2, (n1 & 0x7F), (n2 & 0x7F), (n1 & 0x7F), (n2 & 0x7F)) ;
				}
				ImeDoc_bSetMessageN (pDoc, bufText, nText) ;
			} else {
				int	iFuncNo ;

				/* Abort ̏B*/
				if (ImeDoc_bLookupKeymap (pDoc, &msg, &iFuncNo) && 
					(iFuncNo == NFUNC_ABORT_RECURSIVE_EDIT || iFuncNo == NFUNC_KEYBOARD_QUIT)) {
					ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
					ImeDoc_vSetSignal (pDoc, LMSIGNAL_QUIT) ;
					ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenuJump_Exit) ;
					return	LMR_CONTINUE ;
				}

				/* ȃL[B*/
#if defined (UNITTEST)
				n1	= nN1Org ;
				n2	= nN2Org ;
#endif
				if (0x21 <= nCH && nCH < 0x7F) {
					nText	= wnsprintf (bufText, ARRAYSIZE (bufText), L"`%c' is not valid here!", nCH) ;
				} else {
					nText	= wnsprintf (bufText, ARRAYSIZE (bufText), L"`0x%0x' is not valid here!", nCH) ;
				}
				ImeDoc_bSetMessageN (pDoc, bufText, nText) ;
			}
		} else {
#if !defined (UNITTEST)
			TVarbuffer*	pvbufCandInfo ;
#endif

#if defined (UNITTEST)
			n1	= nN1Org ;
			n2	= nN2Org ;
#endif
			while (i > 0) {
				n2	+= 16 ;
				if (n2 > nSkkCodeN2Max) {
					n1	++ ;
					if (n1 > nSkkCodeN1Max)
						n1	= nSkkCodeN1Min ;
					n2	=	nSkkCodeN2Min ;
				}
				if (((n1 << 8) | n2) > nCodeMax) {
					n1	= (nCodeMin >> 8) & 0xFF ;
					n2	= (nCodeMin     ) & 0xFF ;
				}
				i	-- ;
			}

#if ! defined (UNITTEST)
			ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_CLOSE_CANDIDATELIST) ;

			pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
			if (pvbufCandInfo != NULL)
				TVarbuffer_Clear (pvbufCandInfo) ;
#endif
			// skk-input-by-code-or-menu-jump-default  n1 ɍXVB
			//  n1, n2  skk-input-by-code-or-menu-1 JB
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_0, n1) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_1, n2) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_2, nCodeMin) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_3, nCodeMax) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_4, nSkkCodeN1Min) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_5, nSkkCodeN1Max) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_6, nSkkCodeN2Min) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_7, nSkkCodeN2Max) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkInputByCodeOrMenuJump_Exit, LM_bSkkInputByCodeOrMenu1))
				return	LMR_ERROR ;
			return	LMR_CONTINUE ;
		}
	}
	pnData [0]	= n1 ;
	pnData [1]	= n2 ;
	ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenuJump_1) ;
	return	LMR_CONTINUE ;
}

/*================================================================ skk-input-by-code-or-menu-1
 */
static	int		LM_bSkkInputByCodeOrMenu1_1		(struct CImeDoc*) ;
static	int		LM_bSkkInputByCodeOrMenu1_2		(struct CImeDoc*) ;
static	int		LM_bSkkInputByCodeOrMenu1_Exit	(struct CImeDoc*) ;

/*
	int			n1,
	int			n2,
	int			nCodeMin, 
	int			nCodeMax,
	int			nSkkCodeN1Min,
	int			nSkkCodeN1Max,
	int			nSkkCodeN2Min,
	int			nSkkCodeN2Max,
*/
int
LM_bSkkInputByCodeOrMenu1 (
	struct CImeDoc*			pDoc)
{
	int		n1, n2, nCodeMin, nCodeMax, nSkkCodeN1Min, nSkkCodeN1Max, nSkkCodeN2Min, nSkkCodeN2Max ;
	int*	pnData ;
	struct CImeBuffer*	pBuffer ;
#if !defined (UNITTEST)
	int		iKcodeCharset ;
#endif

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (! ImeDoc_bGetRegInteger (pDoc, LMREGARG_0, &n1) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_1, &n2) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_2, &nCodeMin) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_3, &nCodeMax) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_4, &nSkkCodeN1Min) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_5, &nSkkCodeN1Max) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_6, &nSkkCodeN2Min) ||
		! ImeDoc_bGetRegInteger (pDoc, LMREGARG_7, &nSkkCodeN2Max)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	pnData	= ImeDoc_pAlloca (pDoc, sizeof (int) * 10, __alignof (int)) ;
	if (pnData == NULL)
		return	LMR_ERROR ;

#if !defined (UNITTEST)
	iKcodeCharset	= ImeConfig_iGetSkkKcodeCharset () ;
	if (! imeBuffer_bShowInputByCodeOrMenu1Jump (pBuffer, iKcodeCharset, n1, n2, nCodeMin, nCodeMax, nSkkCodeN1Min, nSkkCodeN1Max, nSkkCodeN2Min, nSkkCodeN2Max)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
#endif
	pBuffer->m_bSkkInputByCodeOrMenu1Mode	= TRUE ;

	pnData [0]	= n1 ;
	pnData [1]	= n2 ;	/* n2 */
	pnData [2]	= n1 ;	/* n1orig */
	pnData [3]	= n2 ;	/* n2orig */
	pnData [4]	= nCodeMin ;
	pnData [5]	= nCodeMax ;
	pnData [6]	= nSkkCodeN1Min ;
	pnData [7]	= nSkkCodeN1Max ;
	pnData [8]	= nSkkCodeN2Min ;
	pnData [9]	= nSkkCodeN2Max ;
	if (! ImeDoc_bPushReg (pDoc, LMREG_0))
		return	LMR_ERROR ;

	ImeDoc_vSetRegPointer (pDoc, LMREG_0, pnData) ;
	ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu1_1) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkInputByCodeOrMenu1_Exit (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
#if !defined (UNITTEST)
	TVarbuffer*	pvbufCandInfo ;
#endif

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) {
		pBuffer->m_bSkkInputByCodeOrMenu1Mode	= FALSE ;
	}

	ImeDoc_vPopReg (pDoc, LMREG_0) ;

#if !defined (UNITTEST)
	/*	CLOSE  CREATE ɗ̂Av낤H
	 *	ށAkeyboard quit ̖肪ȁB̎ close ͕KvȁB
	 *	Ƃ킯ŁA Close B
	 */
	ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_CLOSE_CANDIDATELIST) ;

	pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
	if (pvbufCandInfo != NULL)
		TVarbuffer_Clear (pvbufCandInfo) ;
#endif
	return	LMR_RETURN ;
}

int
LM_bSkkInputByCodeOrMenu1_1 (
	struct CImeDoc*			pDoc)
{
#if defined (UNITTEST)
	LPCWSTR			pwKeys2 ;
	int				nKeys2 ;
	int				i, nN1Org, nN2Org, iKcodeCharset ;
	WCHAR			wbuf [512] ;
	LPWSTR			pDest, pDestEnd ;
	int*			pnData ;
	int		n1, n2, nCodeMin, nCodeMax, nSkkCodeN1Min, nSkkCodeN1Max, nSkkCodeN2Min, nSkkCodeN2Max ;
#endif

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu1_Exit) ;
		return	LMR_CONTINUE ;
	}

#if defined (UNITTEST)
	if (! ImeDoc_bGetRegPointer (pDoc, LMREG_0, (void**)&pnData)) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu1_Exit) ;
		return	LMR_CONTINUE ;
	}
	pwKeys2	= ImeConfig_pGetSkkInputByCodeMenuKeys2 (&nKeys2) ;
	if (pwKeys2 == NULL || nKeys2 <= 0) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu1_Exit) ;
		return	LMR_CONTINUE ;
	}

	iKcodeCharset	= ImeConfig_iGetSkkKcodeCharset () ;
	pDestEnd		= wbuf + ARRAYSIZE (wbuf) ;
	
	n1				= pnData [0] ;
	n2				= pnData [1] ;
	nCodeMin		= pnData [4] ;
	nCodeMax		= pnData [5] ;
	nSkkCodeN1Min	= pnData [6] ;
	nSkkCodeN1Max	= pnData [7] ;
	nSkkCodeN2Min	= pnData [8] ;
	nSkkCodeN2Max	= pnData [9] ;

	nN1Org		= n1 ;
	nN2Org		= n2 ;
	pDest		= wbuf ;
	for (i = 0 ; i < nKeys2 && pDest < pDestEnd ; i ++) {
		int		ch ;
#define	SAFE_COPY(ptr,chara)	{ if (ptr < pDestEnd) { *ptr ++ = (chara) ; } } 
		SAFE_COPY (pDest, towupper (pwKeys2 [i])) ;
		SAFE_COPY (pDest, L':') ;
		if (bMakeUnicodeChar (iKcodeCharset, n1, n2, &ch)) {
			SAFE_COPY (pDest, ch) ;
		} else {
			SAFE_COPY (pDest, L'?') ;
			SAFE_COPY (pDest, L'?') ;
			SAFE_COPY (pDest, L'?') ;
		}
		SAFE_COPY (pDest, L' ') ;
		SAFE_COPY (pDest, L' ') ;
#undef	SAFE_COPY

		n2	++ ;
		if (n2 > nSkkCodeN2Max) {
			n1	++ ;
			if (n1 > nSkkCodeN1Max)
				n1	= nSkkCodeN1Min ;
			n2	=	nSkkCodeN2Min ;
		}
		if (((n1 << 8) | n2) > nCodeMax) {
			n1	= (nCodeMin >> 8) & 0xFF ;
			n2	= (nCodeMin     ) & 0xFF ;
		}
	}
	ImeDoc_bSetMessageN (pDoc, wbuf, pDest - wbuf) ;

	pnData [0]	= n1 ;
	pnData [1]	= n2 ;
	pnData [2]	= nN1Org ;
	pnData [3]	= nN2Org ;
#endif
	if (! ImeDoc_bCall (pDoc, LM_bNextCommandEvent, LM_bSkkInputByCodeOrMenu1_2))
		return	LMR_ERROR ;

	return	LMR_CONTINUE ;
}

int
LM_bSkkInputByCodeOrMenu1_2 (
	struct CImeDoc*			pDoc)
{
	struct TMSG		msg ;
	LPCWSTR			pwKeys2 ;
	int*	pnData ;
	int		n1, n2, nCodeMin, nCodeMax, nSkkCodeN1Min, nSkkCodeN1Max, nSkkCodeN2Min, nSkkCodeN2Max ;
	int		nCH, nKeys2 ;
#if defined (UNITTEST)
	int		nN1Org, nN2Org ;
#endif
#if !defined (UNITTEST)
	TVarbuffer*	pvbufCandInfo ;
#endif
	WCHAR	bufText [256] ;
	int		nText ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu1_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegPointer (pDoc, LMREG_0, (void**)&pnData) ||
		! ImeDoc_bGetRegMsg (pDoc, LMREGARG_RETVAL, &msg)) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu1_Exit) ;
		return	LMR_CONTINUE ;
	}
	pwKeys2	= ImeConfig_pGetSkkInputByCodeMenuKeys2 (&nKeys2) ;
	if (pwKeys2 == NULL || nKeys2 <= 0) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu1_Exit) ;
		return	LMR_CONTINUE ;
	}

	n1				= pnData [0] ;
	n2				= pnData [1] ;
#if defined (UNITTEST)
	nN1Org			= pnData [2] ;
	nN2Org			= pnData [3] ;
#endif
	nCodeMin		= pnData [4] ;
	nCodeMax		= pnData [5] ;
	nSkkCodeN1Min	= pnData [6] ;
	nSkkCodeN1Max	= pnData [7] ;
	nSkkCodeN2Min	= pnData [8] ;
	nSkkCodeN2Max	= pnData [9] ;
	DEBUGPRINTFEX (99, (TEXT ("[1] LM_bSkkInputByCodeOrMenu1_2(n1:%d(min:%d,max:%d), n2:%d(min:%d,max:%d)\n"),
		n1, nSkkCodeN1Min, nSkkCodeN1Max, n2, nSkkCodeN2Min, nSkkCodeN2Max)) ;

	if (! ImeDoc_bEventToCharacter (&msg, &nCH)) {
		int	iFuncNo ;

		/* Abort ̏B*/
		if (ImeDoc_bLookupKeymap (pDoc, &msg, &iFuncNo) && 
			(iFuncNo == NFUNC_ABORT_RECURSIVE_EDIT || iFuncNo == NFUNC_KEYBOARD_QUIT)) {
			ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
			ImeDoc_vSetSignal (pDoc, LMSIGNAL_QUIT) ;
			ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu1_Exit) ;
			return	LMR_CONTINUE ;
		}

		/* "`%s' is not valid here!" */
#if defined (UNITTEST)
		n1	= nN1Org ;
		n2	= nN2Org ;
#endif
	} else {
		int		i ;

		/* skk-input-by-code-menu-keys1 ɂ邩ǂ`FbNB*/
		for (i = 0 ; i < nKeys2 ; i ++) {
			if (towlower (nCH) == towlower (pwKeys2 [i]))
				break ;
		}
		if (i >= nKeys2) {
			if (ImeDoc_bSkkPreviousCandidateCharp (pDoc, nCH)) {
				/* n1, n2 ߂B*/
#if defined (UNITTEST)
				n1	= nN1Org ;
				n2	= nN2Org ;
#endif
				for (i = 0 ; i < nKeys2 ; i ++) {
					n2	-- ;
					if (n2 < nSkkCodeN2Min) {
						n1	-- ;
						if (n1 < nSkkCodeN1Min)
							n1	= nSkkCodeN1Max ;
						n2	= nSkkCodeN2Max ;
					}
					/* unicode ̎ɕKvB*/
					if (((n1 << 8) | n2) < nCodeMin) {
						n1	= ((nCodeMax - 1) >> 8) & 0xFF ;
						n2	= ((nCodeMax - 1)     ) & 0xFF ;
					}
				}
#if !defined (UNITTEST)
				pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
				if (pvbufCandInfo != NULL && TVarbuffer_GetUsage (pvbufCandInfo) >= sizeof (MYCAND)) {
					LPMYCAND	pMyCand ;

					pMyCand		= (LPMYCAND) TVarbuffer_GetBuffer (pvbufCandInfo) ;
					if (pMyCand != NULL) {
						LPCANDIDATELIST	pCandidateList	= (LPCANDIDATELIST)&(pMyCand->cl) ;
						if (pCandidateList->dwPageStart >= (DWORD)nKeys2) {
							pCandidateList->dwPageStart	-= nKeys2 ;
						} else {
							pCandidateList->dwPageStart	= (pCandidateList->dwCount + pCandidateList->dwPageStart - nKeys2) ;
						}
						pCandidateList->dwSelection	= pCandidateList->dwPageStart ;
					}
					ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_UPDATE_CANDIDATELIST) ;
				}
#endif
			} else if (ImeConfig_bSkkStartHenkanCharp (nCH)) {
				/*  n1, n2  ok */
#if !defined (UNITTEST)
				for (i = 0 ; i < nKeys2 ; i ++) {
					n2	++ ;
					if (n2 > nSkkCodeN2Max) {
						n1	++ ;
						if (n1 > nSkkCodeN1Max)
							n1	= nSkkCodeN1Min ;
						n2	= nSkkCodeN2Min ;
					}
					/* unicode ̎ɕKvB*/
					if (((n1 << 8) | n2) > nCodeMax) {
						n1	= (nCodeMin  >> 8) & 0xFF ;
						n2	= (nCodeMin      ) & 0xFF ;
					}
				}
				pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
				if (pvbufCandInfo != NULL && TVarbuffer_GetUsage (pvbufCandInfo) >= sizeof (MYCAND)) {
					LPMYCAND	pMyCand ;

					pMyCand		= (LPMYCAND) TVarbuffer_GetBuffer (pvbufCandInfo) ;
					if (pMyCand != NULL) {
						LPCANDIDATELIST	pCandidateList	= (LPCANDIDATELIST)&(pMyCand->cl) ;
						pCandidateList->dwPageStart	+= nKeys2 ;
						if (pCandidateList->dwPageStart >= pCandidateList->dwCount) {
							pCandidateList->dwPageStart	-= pCandidateList->dwCount ;
						}
						pCandidateList->dwSelection	= pCandidateList->dwPageStart ;
					}
					ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_UPDATE_CANDIDATELIST) ;
				}
#endif
			} else if (nCH == L'?') {
				int		iKcodeCharset, wch ;
#if defined (UNITTEST)
				n1	= nN1Org ;
				n2	= nN2Org ;
				/* code \ hit any key ƁB*/
#endif
				iKcodeCharset	= ImeConfig_iGetSkkKcodeCharset () ;
				if (bMakeUnicodeChar (iKcodeCharset, n1, n2, &wch)) {
					nText	= wnsprintf (bufText, ARRAYSIZE (bufText), L"`%c' EUC: %2x%2x (%3d, %3d), JIS: %2x%2x (%3d, %3d), UNICODE: %04x",
						wch, n1, n2, n1, n2, (n1 & 0x7F), (n2 & 0x7F), (n1 & 0x7F), (n2 & 0x7F), wch) ;
				} else {
					nText	= wnsprintf (bufText, ARRAYSIZE (bufText), L"`?' EUC: %2x%2x (%3d, %3d), JIS: %2x%2x (%3d, %3d)",
						n1, n2, n1, n2, (n1 & 0x7F), (n2 & 0x7F), (n1 & 0x7F), (n2 & 0x7F)) ;
				}
				ImeDoc_bSetMessageN (pDoc, bufText, nText) ;
			} else {
				int	iFuncNo ;

				/* Abort ̏B*/
				if (ImeDoc_bLookupKeymap (pDoc, &msg, &iFuncNo) && 
					(iFuncNo == NFUNC_ABORT_RECURSIVE_EDIT || iFuncNo == NFUNC_KEYBOARD_QUIT)) {
					ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
					ImeDoc_vSetSignal (pDoc, LMSIGNAL_QUIT) ;
					ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu1_Exit) ;
					return	LMR_CONTINUE ;
				}

				/* ȃL[B*/
#if defined (UNITTEST)
				n1	= nN1Org ;
				n2	= nN2Org ;
#endif
				if (0x21 <= nCH && nCH < 0x7F) {
					nText	= wnsprintf (bufText, ARRAYSIZE (bufText), L"`%c' is not valid here!", nCH) ;
				} else {
					nText	= wnsprintf (bufText, ARRAYSIZE (bufText), L"`0x%0x' is not valid here!", nCH) ;
				}
				ImeDoc_bSetMessageN (pDoc, bufText, nText) ;
			}
		} else {
			int		nRetval, iKcodeCharset ;

#if defined (UNITTEST)
			n1	= nN1Org ;
			n2	= nN2Org ;
#endif
			while (i > 0) {
				n2	++ ;
				if (n2 >= nSkkCodeN2Max) {
					n1	++ ;
					if (n1 >= nSkkCodeN1Max)
						n1	= nSkkCodeN1Min ;
					n2	=	nSkkCodeN2Min ;
				}
				if (((n1 << 8) | n2) >= nCodeMax) {
					n1	= (nCodeMin >> 8) & 0xFF ;
					n2	= (nCodeMin     ) & 0xFF ;
				}
				i	-- ;
			}
			iKcodeCharset	= ImeConfig_iGetSkkKcodeCharset () ;
			/* codeset Ƃđʖڂ灬ɂB*/
			if (! bMakeUnicodeChar (iKcodeCharset, n1, n2, &nRetval)) {
				ImeDoc_vSetRegInteger (pDoc, LMREGARG_RETVAL, 0) ;
			} else {
				ImeDoc_vSetRegInteger (pDoc, LMREGARG_RETVAL, nRetval) ;
			}
			ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu1_Exit) ;
			return	LMR_CONTINUE ;
		}
	}
	DEBUGPRINTFEX (99, (TEXT ("[2] LM_bSkkInputByCodeOrMenu1_2(n1:%d(min:%d,max:%d), n2:%d(min:%d,max:%d)\n"),
		n1, nSkkCodeN1Min, nSkkCodeN1Max, n2, nSkkCodeN2Min, nSkkCodeN2Max)) ;
	pnData [0]	= n1 ;
	pnData [1]	= n2 ;
	ImeDoc_vJump (pDoc, LM_bSkkInputByCodeOrMenu1_1) ;
	return	LMR_CONTINUE ;
}


#if !defined (UNITTEST)
BOOL
bParseKutenCode (
	LPCWSTR		wstrKuten,
	int			nKutenLen,
	int*		pn1, 
	int*		pn2) 
{
	LPCWSTR		ptr, ptrEnd, ptrFirstEnd, ptrSecond, ptrSecondEnd ;
	int			nN1, nN2 ;

	ptr			= wstrKuten ;
	ptrEnd		= wstrKuten + nKutenLen ;
	while (ptr < ptrEnd) {
		if (! (L'0' <= *ptr && *ptr <= L'9')) 
			break ;
		ptr	++ ;
	}
	ptrFirstEnd	= ptr ;
	if (ptr == wstrKuten)
		return	FALSE ;

	if (ptr >= ptrEnd || *ptr != L'-')
		return	FALSE ;
	ptr	++ ;
	ptrSecond	= ptr ;
	while (ptr < ptrEnd) {
		if (! (L'0' <= *ptr && *ptr <= L'9')) 
			break ;
		ptr	++ ;
	}
	ptrSecondEnd	= ptr ;
	if (ptrSecondEnd == ptrSecond)
		return	FALSE ;

	ptr			= wstrKuten ;
	nN1			= 0 ;
	while (ptr < ptrFirstEnd) {
		nN1	= nN1 * 10 + (*ptr - '0') ;
		ptr	++ ;
	}
	ptr			= ptrSecond ;
	nN2			= 0 ;
	while (ptr < ptrSecondEnd) {
		nN2	= nN2 * 10 + (*ptr - '0') ;
		ptr	++ ;
	}
	if (pn1 != NULL)
		*pn1	= nN1 ;
	if (pn2 != NULL)
		*pn2	= nN2 ;
	return	TRUE ;
}

#endif



