/* # SKKIME1.0 (Simple Kana-Kanji Input Method Editor for Windows2000)
 * Toascii.c
 * This file is part of skkime1.0.
 * Copyright (C) 1999
 * Takashi SAKAMOTO (tatari_sakamoto@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "skki1_5.h"

/*************************************************************************
 *	ImeToAsciiEx 
 *	ImeToAsciiEx ֐ hIMC p[^ɏ] IME ϊGWʂĕ
 *	ʂ𐶐܂B
 *	UINT
 *		ImeToAsciiEx(
 *		UINT uVirKey,
 *		UINT uScanCode,
 *		CONST LPBYTE lpbKeyState,
 *		LPDWORD lpdwTransBuf,
 *		UINT fuState,
 *		HIMC hIMC
 *	)
 *	(p[^)
 *	uVirKey
 *		ϊ鉼zL[R[hw肵܂BvpeB 
 *		IME_PROP_KBD_CHAR_FIRST rbgIAzL[̏ʃoCg
 *		͕⏕LN^R[h(?)ɂȂ܂B
 *		Unicode ɂẮAIME_PROP_KBD_CHAR_FIRST rbgI 
 *		uVirKey ̏ʃ[h Unicode ܂݂܂B
 *	uScanCode
 *		ϊL[̃n[hEFAXLR[hw肵܂B
 *	lpbKeyState
 *		݂̃L[{[h̏Ԃ܂256oCg̔zւ̃|C^łB
 *		IME ͂̓eύXĂ͂܂B
 *	lpdwTransBuf
 *		ϊʂ󂯎 DWORD ̃obt@w|C^łB̏́A
 *		[bZ[Wobt@̒][bZ[W1][wParam1][lParam1]
 *		[bZ[W2][wParam2][lParam2][...[...[...]]] ̂悤ɂȂ܂B
 *	fuState
 *		Active menu flag. H
 *	hIMC
 *		Input context handle.
 *	(Ԃl)
 *		Ԃl̓bZ[W̐\܂B̐obt@TCY
 *		傫΃bZ[Wobt@͏\ł͂ȂłcēRł́c(^^;;
 *		VXe hMsgBuf `FbNăbZ[W𓾂܂B
 *	(See Also)
 *		ImmToAsciiEx
 *************************************************************************/
UINT	WINAPI
ImeToAsciiEx (
	register UINT			uVKey,
	register UINT			uScanCode,
	register CONST LPBYTE	lpbKeyState,
	register LPTRANSMSGLIST	lpTransBuf,
	register UINT			fuState,
	register HIMC			hIMC)
{
	LPARAM			lParam ;
	LPINPUTCONTEXT	lpIMC ;
	BOOL			fOpen ;

	DEBUGPRINTF ((TEXT ("ImeToAsciiEx (VKEY:%x, SCANCODE:%x, PKEY:%p, PTRANS:%p, STATE:%x, HIMC%x)\n"), uVKey, uScanCode, lpbKeyState, lpTransBuf, fuState, hIMC)) ;
	
	g_lpCurTransKey	= lpTransBuf ;
	lParam			= ((DWORD)uScanCode << 16) + 1L ;
	
	/* Init uNumTransKey here. */
	g_uNumTransKey	= 0 ;

	/* if hIMC is NULL, this means DISABLE IME. */
	if (!hIMC)
		return	0 ;

	lpIMC	= ImmLockIMC(hIMC) ;
	if (!lpIMC)
	   	return	0 ;

	fOpen			= lpIMC->fOpen ;

	ImmUnlockIMC (hIMC) ;

	/* The current status of IME is "closed". */
	/*
	if (!fOpen) {
		DEBUGPRINTF ((TEXT ("ImeToAsciiEx (VKEY:%x, SCANCODE:%x, PKEY:%p, PTRANS:%p, STATE:%x, HIMC%x) -- !fOpen\n"), uVKey, uScanCode, lpbKeyState, lpTransBuf, fuState, hIMC)) ;
		SKKCloseText (hIMC) ;
		goto itae_exit ;
	}
	*/
	if (uScanCode & 0x8000){
		IMEKeyupHandler (hIMC, uVKey, lParam, lpbKeyState, fOpen) ;
	} else {
		IMEKeydownHandler (hIMC, uVKey, lParam, lpbKeyState, fOpen) ;
	}
	/* Clear static value, no more generated message! */
	g_lpCurTransKey		= NULL ;

	//itae_exit:
	DEBUGPRINTF ((TEXT ("ImeToAsciiEx (VKEY:%x, SCANCODE:%x, PKEY:%p, PTRANS:%p, STATE:%x, HIMC%x) -- leave(%d)\n"), uVKey, uScanCode, lpbKeyState, lpTransBuf, fuState, hIMC, (int)g_uNumTransKey)) ;

	/*
	 *	If trans key buffer that is allocated by USER.EXE full up,
	 *	the return value is the negative number.
	 */
	if (g_bOverTransKey)
		return	(int)g_uNumTransKey ;

	return	(int)g_uNumTransKey ;
}


/*
 *	  GenerateMessageToTransKey ()
 *
 *	  Update the transrate key buffer.
 */
BOOL	PASCAL
GenerateMessageToTransKey (
	register LPTRANSMSGLIST		lpTransBuf,
	register LPTRANSMSG			lpGeneMsg)
{
	register LPTRANSMSG	lpgmT0 ;
	register BOOL		fRetval ;

	DEBUGPRINTF ((TEXT ("GenerateMessageToTransKey (%p, %p) -- enter\n"),lpTransBuf, lpGeneMsg)) ;

	g_uNumTransKey	++ ;
	if (g_uNumTransKey >= lpTransBuf->uMsgCount){
		DEBUGPRINTFEX (101, (TEXT ("GenerateMessageToTransKey(): over(%d >= %d)\n"), g_uNumTransKey, lpTransBuf->uMsgCount)) ;
		g_bOverTransKey	= TRUE ;
		fRetval			= FALSE ;
	} else {
		lpgmT0		= lpTransBuf->TransMsg + (g_uNumTransKey - 1) ;
		*lpgmT0		= *lpGeneMsg ;
		fRetval		= TRUE ;
	}
	DEBUGPRINTF ((TEXT ("GenerateMessageToTransKey (%p, %p) -- leave(%d)\n"), lpTransBuf, lpGeneMsg, fRetval)) ;
	return	fRetval ;
}

