/* # SKKIME1.0 (Simple Kana-Kanji Input Method Editor for Windows2000)
 * uistate.c
 * This file is part of skkime1.0.
 * Copyright (C) 1999
 * Takashi SAKAMOTO (tatari_sakamoto@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "skki1_5.h"
#include "skkui.h"
#include "resource.h"

enum {
	INDEX_STATUS_BMP_DI	= 0,
	INDEX_STATUS_BMP_KANA,
	INDEX_STATUS_BMP_KATA,
	INDEX_STATUS_BMP_ZENEI,
	INDEX_STATUS_BMP_ASCII,
	INDEX_STATUS_BMP_HANKANA,
	INDEX_STATUS_BMP_HANKANA_ROMAN,
	NUM_STATUS_BMP_MODES,

	INDEX_STATUS_BMP_LU		= NUM_STATUS_BMP_MODES,
	INDEX_STATUS_BMP_RB,
	INDEX_STATUS_BMP_RB2,
	INDEX_STATUS_BMP_LU2,
	NUM_STATUS_BMPS,
} ;

enum {
	NUM_HEIGHT_STATUS_BMP	= NUM_STATUS_BMPS,
	NUM_WIDTH_STATUS_BMP	= 2,
} ;

/*	ONOFF.BMP ̕ƍB*/
#define	UISTATE_ONOFFBMP_SRC_WIDTH	36
#define	UISTATE_ONOFFBMP_SRC_HEIGHT	100

/*========================================================
 *	vg^Cv錾B
 */
static	void	PASCAL	uiStatus_vPaint					(HWND, HDC, DWORD) ;
static	void	PASCAL	uiStatus_vOnEraseBkGround		(HWND, HDC) ;
static	void	PASCAL	uiStatus_vOnButton				(HWND, UINT, WPARAM, LPARAM) ;
static	BOOL			uiStatus_bOnLeftMenu			(HIMC, HWND) ;
static	BOOL			uiStatus_bOnRightMenu			(HIMC, HWND) ;
static	int				uiStatus_iShowLeftMenu			(HIMC, HWND) ;
static	int				uiStatus_iShowRightMenu			(HIMC, HWND) ;
static	BOOL			uiStatus_bInsertInputModeMenu	(HIMC, HMENU, int FAR*) ;
static	HBITMAP			uiStatus_hCreateCModeBitmap		(HWND) ;
static	HBITMAP			uiStatus_hCreateOnOffBitmap		(HWND) ;
static	DWORD	PASCAL	uiStatus_dwCheckPushedStatus	(HWND, LPPOINT) ;
static	void	PASCAL	uiStatus_vChangeCursor			(HWND) ;

/*========================================================
 *	IMECMODE_XXX  Bitmap Data ̉摜̈ʒȗΉ\D
 */
static const int	uiStatus_rButtonYTable []	= {
	BTY_KANA,
	BTY_KATAKANA,
	BTY_ZENEI,
	BTY_JISX0201KANA,
	BTY_JISX0201ROMAN,
	BTY_ASCII,
	BTY_DIRECTINPUT,
} ;


/*========================================================
 *	Global Functions
 */

/*
 *	Status Window  WINDOWPROCB
 *(Rg)
 *	Status Window Window Message͂̊֐ɑė܂̂ŁA
 *	Window Message ɑΉs܂B̊֐̂ Window
 *	Message ̐U蕪łǁB
 *()
 *	hWnd	Window Message 󂯎 Composition Window  Handle
 *	message	Window Message
 *	wParam	Window Message ̈1
 *	lParam	Window Message ̈2
 */
LRESULT CALLBACK
UIStatus_WndProc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT	ps ;
	HWND		hUIWnd ;
	HDC			hDC ;
	HBITMAP		hbmpStatus ;
	HIMC		hIMC ;

	switch (message){
	case	WM_UI_UPDATE:
		InvalidateRect (hWnd, NULL, FALSE) ;
		break ;

	case	WM_PAINT:
		hDC = BeginPaint (hWnd, &ps) ;
		uiStatus_vPaint (hWnd, hDC, 0) ;
		EndPaint (hWnd, &ps) ;
		break ;

	case	WM_ERASEBKGND:
		uiStatus_vOnEraseBkGround (hWnd, (HDC) wParam) ;
		break ;

	case	WM_MOUSEACTIVATE:
		return	MA_NOACTIVATE ;

	case	WM_NCMOUSEMOVE:
	case	WM_MOUSEMOVE:
	case	WM_LBUTTONUP:
	case	WM_RBUTTONUP:
	case	WM_LBUTTONDOWN:
	case	WM_RBUTTONDOWN:
	case	WM_SETCURSOR:
		uiStatus_vOnButton (hWnd, message, wParam, lParam) ;
		break ;

	case	WM_MOVE:
		{
			RECT	rc ;

			if (GetWindowRect (hWnd, &rc)) {
				HANDLE	hMutex	= TSkkIme_hCreateMutex (SKKIME_MUTEX_NAME) ;
				if (hMutex != NULL) {
					DWORD	dw = WaitForSingleObject (hMutex, INFINITE) ;
					if (dw != WAIT_FAILED && dw != WAIT_TIMEOUT) {
						g_ptUIStatus.x	= rc.left ;
						g_ptUIStatus.y	= rc.top ;
						ReleaseMutex (hMutex) ;
					}
					CloseHandle (hMutex) ;
				}
			}
			hUIWnd = (HWND)GetWindowLongPtr (hWnd, FIGWL_UISTATE_SVRWND) ;
			if (IsWindow (hUIWnd))
				SendMessage (hUIWnd, WM_UI_STATEMOVE, wParam, lParam) ;
		}
		break ;

	case	WM_CREATE:
#ifdef STATUSWINDOW_LARGE
		SetWindowLongPtr (hWnd, FIGWL_UISTATE_STATUSORGBMP,	(LONG_PTR)LoadBitmap (g_hInst, TEXT ("STATUSBMP"))) ;
		SetWindowLongPtr (hWnd, FIGWL_UISTATE_CLOSEORGBMP,	(LONG_PTR)LoadBitmap (g_hInst, TEXT ("ONOFFBMP"))) ;
#else
		SetWindowLongPtr (hWnd, FIGWL_UISTATE_STATUSORGBMP,	(LONG_PTR)LoadBitmap (g_hInst, TEXT ("STATUSSBMP"))) ;
		SetWindowLongPtr (hWnd, FIGWL_UISTATE_CLOSEORGBMP,	(LONG_PTR)LoadBitmap (g_hInst, TEXT ("ONOFFSBMP"))) ;
#endif
		SetWindowLongPtr (hWnd, FIGWL_UISTATE_STATUSBMP,	(LONG_PTR)uiStatus_hCreateCModeBitmap (hWnd)) ;
		SetWindowLongPtr (hWnd, FIGWL_UISTATE_CLOSEBMP,		(LONG_PTR)uiStatus_hCreateOnOffBitmap (hWnd)) ;
		SetWindowLongPtr (hWnd, FIGWL_UISTATE_CURSOR,		(LONG_PTR)0) ;
		SetWindowLong (hWnd, FIGWL_UISTATE_MOUSE,			0L) ;
		SetWindowLong (hWnd, FIGWL_UISTATE_PUSHSTATUS,		0L) ;
		SetWindowLong (hWnd, FIGWL_UISTATE_PREVCMODE,		0L) ;
		break ;

	case	WM_SYSCOLORCHANGE:
		hbmpStatus = (HBITMAP)GetWindowLongPtr (hWnd, FIGWL_UISTATE_STATUSBMP) ;
		if (hbmpStatus)
			DeleteObject (hbmpStatus) ;
		hbmpStatus = (HBITMAP)GetWindowLongPtr (hWnd, FIGWL_UISTATE_CLOSEBMP) ;
		if (hbmpStatus)
			DeleteObject (hbmpStatus) ;
		SetWindowLongPtr (hWnd, FIGWL_UISTATE_STATUSBMP,	(LONG_PTR)uiStatus_hCreateCModeBitmap (hWnd)) ;
		SetWindowLongPtr (hWnd, FIGWL_UISTATE_CLOSEBMP,		(LONG_PTR)uiStatus_hCreateOnOffBitmap (hWnd)) ;
		break ;

	case	WM_DESTROY:
#if 1
		hbmpStatus = (HBITMAP)GetWindowLongPtr (hWnd, FIGWL_UISTATE_STATUSORGBMP) ;
		if (hbmpStatus)
			DeleteObject (hbmpStatus) ;
		hbmpStatus = (HBITMAP)GetWindowLongPtr (hWnd, FIGWL_UISTATE_CLOSEORGBMP) ;
		if (hbmpStatus)
			DeleteObject (hbmpStatus) ;
#endif
		hbmpStatus = (HBITMAP)GetWindowLongPtr (hWnd, FIGWL_UISTATE_STATUSBMP) ;
		if (hbmpStatus)
			DeleteObject (hbmpStatus) ;
		hbmpStatus = (HBITMAP)GetWindowLongPtr (hWnd, FIGWL_UISTATE_CLOSEBMP) ;
		if (hbmpStatus)
			DeleteObject (hbmpStatus) ;
		SetWindowLong (hWnd, FIGWL_UISTATE_MOUSE, 0L) ;
		SetWindowLong (hWnd, FIGWL_UISTATE_PUSHSTATUS, 0L) ;
		break ;

	case	WM_UI_HIDE:
		ShowWindow (hWnd, SW_HIDE) ;
		return	0 ;

	default:
		if (!MyIsIMEMessage(message))
			return	DefWindowProc (hWnd, message, wParam, lParam) ;
		break ;
	}
	return 0 ;
}

void	PASCAL
UIStatus_Update (
	LPUIEXTRA	lpUIExtra)
{
	if (IsWindow (lpUIExtra->uiStatus.hWnd))
		SendMessage (lpUIExtra->uiStatus.hWnd, WM_UI_UPDATE, 0, 0L) ;
	return ;
}

void	PASCAL
UIStatus_Show (
	LPUIEXTRA		lpUIExtra)
{
	HANDLE	hMutex ;

	if (lpUIExtra == NULL || ! IsWindow (lpUIExtra->uiStatus.hWnd))
		return ;

	hMutex	= TSkkIme_hCreateMutex (SKKIME_MUTEX_NAME) ;
	if (hMutex != NULL) {
		DWORD	dw = WaitForSingleObject (hMutex, INFINITE) ;
		if (dw != WAIT_FAILED && dw != WAIT_TIMEOUT) {
			if (g_ptUIStatus.x != -1) {
				lpUIExtra->uiStatus.pt.x	= g_ptUIStatus.x ;
				lpUIExtra->uiStatus.pt.y	= g_ptUIStatus.y ;
			} 
			ReleaseMutex (hMutex) ;
		}
		CloseHandle (hMutex) ;
	}
	SetWindowPos (lpUIExtra->uiStatus.hWnd, HWND_TOP, lpUIExtra->uiStatus.pt.x, lpUIExtra->uiStatus.pt.y, 1, 1, 
			SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOSENDCHANGING | SWP_NOOWNERZORDER | SWP_SHOWWINDOW) ;
	return ;
}

void	PASCAL
UIStatus_Move (
	LPUIEXTRA		lpUIExtra,
	const POINTS*	pPT)
{
	RECT	rc ;

	if (! IsWindow (lpUIExtra->uiStatus.hWnd))
		return ;

	GetWindowRect (lpUIExtra->uiStatus.hWnd, &rc) ;
	MoveWindow (lpUIExtra->uiStatus.hWnd, pPT->x, pPT->y, rc.right - rc.left, rc.bottom - rc.top, FALSE) ;
	return ;
}

void
UIStatus_AdjustPosition (
	LPUIEXTRA		lpUIExtra)
{
	RECT	rcWorkArea, rcWnd ;

	if (! IsWindow (lpUIExtra->uiStatus.hWnd))
		return ;

	SystemParametersInfo (SPI_GETWORKAREA, 0, &rcWorkArea, FALSE) ;
	if (GetWindowRect (lpUIExtra->uiStatus.hWnd, &rcWnd)) {
		POINT	pt ;
		pt.x	= rcWnd.left ;
		pt.y	= rcWnd.top ;
		if (rcWnd.left < rcWorkArea.left) 
			pt.x	= rcWorkArea.left ;
		if (rcWnd.right > rcWorkArea.right) 
			pt.x	= rcWorkArea.right - (rcWnd.right - rcWnd.left) ;
		if (rcWnd.top  < rcWorkArea.top)
			pt.y	= rcWorkArea.top ;
		if (rcWnd.bottom > rcWorkArea.bottom)
			pt.y	= rcWorkArea.bottom - (rcWnd.bottom - rcWnd.top) ;
		if (pt.x != rcWnd.left || pt.y != rcWnd.top) {
			MoveWindow (lpUIExtra->uiStatus.hWnd, pt.x, pt.y, rcWnd.right - rcWnd.left, rcWnd.bottom - rcWnd.top, TRUE) ;
			lpUIExtra->uiStatus.pt.x	= rcWnd.left ;
			lpUIExtra->uiStatus.pt.y	= rcWnd.top ;
		}
	}
	return ;
}

/*========================================================================*
 *	private  functions
 */
void	PASCAL
uiStatus_vPaint (
	HWND		hStatusWnd,
	HDC			hDC,
	DWORD		dwPushedStatus)
{
	HIMC			hIMC ;
	LPINPUTCONTEXT	lpIMC ;
	HDC				hMemDC ;
	HBITMAP			hbmpOld ;
	int				src_x, src_y ;
	int				iConversionMode ;
	HWND			hSvrWnd ;
	POINT			pt ;
	HBITMAP			hbmpStatus ;
	HBRUSH			hOldBrush, hBrush ;

	hSvrWnd		= (HWND)GetWindowLongPtr (hStatusWnd,	FIGWL_UISTATE_SVRWND) ;
	hIMC		= (HIMC)GetWindowLongPtr (hSvrWnd,		IMMGWLP_IMC) ;

	if (hIMC){
		lpIMC	= ImmLockIMC (hIMC) ;
	} else {
		lpIMC	= NULL ;
	}
	hMemDC	= CreateCompatibleDC (hDC) ;

	GetCursorPos (&pt) ;
	ScreenToClient (hStatusWnd, &pt) ;

	if (hIMC){
		/* Table QƂɂǂ̂ȁD*/
		iConversionMode	= iGetConversionModeFromLPIMC (lpIMC) ;
	} else {
		iConversionMode	= (int)GetWindowLong (hStatusWnd, FIGWL_UISTATE_PREVCMODE) ;
	}

	hbmpStatus	= (HBITMAP)GetWindowLongPtr (hStatusWnd, FIGWL_UISTATE_CLOSEBMP) ;
	hbmpOld		= (HBITMAP)SelectObject (hMemDC, hbmpStatus) ;
	if (hIMC){
		if (iConversionMode == IMECMODE_OFF){
			src_y	= 0 ;
		} else {
			src_y	= STATUSONOFF_BTNY ;
		}
		if (dwPushedStatus & PUSHED_STATUS_ONOFF){
			src_x	= STATUSONOFF_BTNX * 2 ;
		} else if (STATUSONOFF_X <= pt.x && pt.x < STATUSCMODE_X &&
				   STATUSONOFF_Y <= pt.y && pt.y < 10){
			src_x	= STATUSONOFF_BTNX * 1 ;
		} else {
			src_x	= 0 ;
		}
	}  else {
		src_x	= 0 ;
		src_y	= 0 ;
	}
	BitBlt (hDC, STATUSONOFF_X, STATUSONOFF_Y, STATUSONOFF_BTNX, STATUSONOFF_BTNY, hMemDC, src_x, src_y, SRCCOPY) ;
	SelectObject (hMemDC, hbmpOld) ;

	hbmpStatus	= (HBITMAP)GetWindowLongPtr (hStatusWnd, FIGWL_UISTATE_STATUSBMP) ;
	hbmpOld		= (HBITMAP)SelectObject (hMemDC, hbmpStatus) ;
	if (hIMC){
		if (dwPushedStatus & PUSHED_STATUS_MODE){
			src_x	= STATUSCMODE_BTNX * 2 ;
		} else if (STATUSCMODE_X <= pt.x && pt.x < STATUS_BTNX &&
				   STATUSCMODE_Y <= pt.y && pt.y < STATUS_BTNY){
			src_x	= STATUSCMODE_BTNX * 1 ;
		} else {
			src_x	= 0 ;
		}
	} else {
		src_x	= STATUSCMODE_BTNX * 3 ;
	}

	BitBlt (hDC, STATUSCMODE_X, STATUSCMODE_Y, STATUSCMODE_BTNX, STATUSCMODE_BTNY, hMemDC, src_x, uiStatus_rButtonYTable [iConversionMode], SRCCOPY) ;
	SelectObject (hMemDC, hbmpOld) ;
	DeleteDC (hMemDC) ;
	if (hIMC){
		ImmUnlockIMC (hIMC) ;
		SetWindowLong (hStatusWnd, FIGWL_UISTATE_PREVCMODE, (LONG)iConversionMode) ;
	}
	return ;
}

void	PASCAL
uiStatus_vOnEraseBkGround (
	HWND		hStatusWnd,
	HDC		hDC)
{
	HDC			hMemDC ;
	HBITMAP		hbmpOld ;
	int			iConversionMode ;
	HWND		hSvrWnd ;
	HBITMAP		hbmpStatus ;
	HBRUSH		hOldBrush, hBrush ;

	hSvrWnd	= (HWND)GetWindowLongPtr (hStatusWnd,	FIGWL_UISTATE_SVRWND) ;
	hMemDC	= CreateCompatibleDC (hDC) ;

	hbmpStatus	= (HBITMAP)GetWindowLongPtr (hStatusWnd, FIGWL_UISTATE_CLOSEBMP) ;
	hbmpOld		= (HBITMAP)SelectObject (hMemDC, hbmpStatus) ;
	BitBlt (hDC, STATUSONOFF_X, STATUSONOFF_Y, STATUSONOFF_BTNX, STATUSONOFF_BTNY,
			hMemDC, 0, 0, SRCCOPY) ;
	SelectObject (hMemDC, hbmpOld) ;

	hbmpStatus		= (HBITMAP)GetWindowLongPtr (hStatusWnd, FIGWL_UISTATE_STATUSBMP) ;
	hbmpOld			= (HBITMAP)SelectObject (hMemDC, hbmpStatus) ;
	iConversionMode	= (int)GetWindowLong (hStatusWnd, FIGWL_UISTATE_PREVCMODE) ;
	BitBlt (hDC, STATUSCMODE_X, STATUSCMODE_Y, STATUSCMODE_BTNX, STATUSCMODE_BTNY, hMemDC, STATUSCMODE_BTNX * 3, uiStatus_rButtonYTable [iConversionMode], SRCCOPY) ;
	SelectObject (hMemDC, hbmpOld) ;
	DeleteDC (hMemDC) ;
	return ;
}

void	PASCAL
uiStatus_vOnButton (
	HWND		hStatusWnd, 
	UINT		message,
	WPARAM		wParam,
	LPARAM		lParam)
{
	POINT			pt ;
	HDC				hDC ;
	DWORD			dwMouse, dwPushedStatus, dwTemp, fdwConversion ;
	HIMC			hIMC ;
	HWND			hSvrWnd ;
	BOOL			fOpen ;
	HMENU			hMenu ;
	static	POINT	ptdif ;
	static	RECT	drc ;
	static	RECT	rc ;

	hDC	= GetDC (hStatusWnd) ;

	if (message == WM_SETCURSOR && (HWND) wParam == hStatusWnd)
		message	= HIWORD (lParam) ;

	switch (message){
	case WM_RBUTTONDOWN:
	case WM_LBUTTONDOWN:
		dwMouse			= GetWindowLong (hStatusWnd, FIGWL_UISTATE_MOUSE) ;
		DEBUGPRINTFEX (104, (TEXT ("uiStatus_vOnButton (message:%d, dwMouse:0x%x)\n"), message, dwMouse)) ;
		if (!(dwMouse & FIM_FOCUS))
			break ;
		GetCursorPos (&pt) ;
		dwPushedStatus	= uiStatus_dwCheckPushedStatus (hStatusWnd, &pt) ;
		GetWindowRect (hStatusWnd, &drc) ;
		ptdif.x			= pt.x - drc.left ;
		ptdif.y			= pt.y - drc.top ;
		rc				= drc ;
		rc.right		-= rc.left ;
		rc.bottom		-= rc.top ;
		SetWindowLong (hStatusWnd, FIGWL_UISTATE_MOUSE,			dwMouse | FIM_CAPUTURED) ;
		SetWindowLong (hStatusWnd, FIGWL_UISTATE_PUSHSTATUS,	(DWORD)dwPushedStatus) ;
		uiStatus_vPaint (hStatusWnd, hDC, dwPushedStatus) ;
		break ;

	case	WM_NCMOUSEMOVE:
	case	WM_MOUSEMOVE:
	case	WM_SETCURSOR:
		dwMouse			= GetWindowLong (hStatusWnd, FIGWL_UISTATE_MOUSE) ;
		dwPushedStatus	= (DWORD)GetWindowLong (hStatusWnd, FIGWL_UISTATE_PUSHSTATUS) ;
		DEBUGPRINTFEX (104, (TEXT ("uiStatus_vOnButton (message:%d, dwMouse:0x%x, Pushed:0x%x)\n"), message, dwMouse, dwPushedStatus)) ;
		if (dwPushedStatus){
			if (dwPushedStatus & PUSHED_STATUS_MOVE){
				if (dwMouse & FIM_MOVED){
					DrawUIBorder (&drc) ;
					GetCursorPos (&pt) ;
					drc.left	= pt.x - ptdif.x ;
					drc.top		= pt.y - ptdif.y ;
					drc.right	= drc.left + rc.right ;
					drc.bottom	= drc.top + rc.bottom ;
					DrawUIBorder (&drc) ;
				} else if (dwMouse & FIM_CAPUTURED){
					DrawUIBorder (&drc) ;
					SetWindowLong (hStatusWnd, FIGWL_UISTATE_MOUSE, dwMouse | FIM_MOVED) ;
				}
			}
		} else {
			GetCursorPos (&pt) ;
			ScreenToClient (hStatusWnd, &pt) ;
			GetClientRect (hStatusWnd, &rc) ;
			if (pt.x < rc.left || pt.y < rc.top || pt.x >= rc.right || pt.y >= rc.bottom){
				if (dwMouse & FIM_FOCUS){
					/* Mouse Cursor  Window ɓėB*/
					ReleaseCapture () ;
					dwMouse	= dwMouse & ~FIM_FOCUS ;
					SetWindowLong (hStatusWnd, FIGWL_UISTATE_MOUSE, dwMouse) ;
				}
			} else {
				if (!(dwMouse & FIM_FOCUS)){
					/* Mouse Cursor  Window ɓėB*/
					SetCapture (hStatusWnd) ;
					dwMouse	= dwMouse | FIM_FOCUS ;
					SetWindowLong (hStatusWnd, FIGWL_UISTATE_MOUSE, dwMouse) ;
				}
			}
			uiStatus_vChangeCursor (hStatusWnd) ;
			uiStatus_vPaint (hStatusWnd, hDC, 0) ;
		}
		break ;

	case WM_RBUTTONUP:
		dwMouse	= GetWindowLong (hStatusWnd, FIGWL_UISTATE_MOUSE) ;
		DEBUGPRINTFEX (104, (TEXT ("uiStatus_vOnButton (WM_RBUTTONUP, dwMouse:0x%x)\n"), dwMouse)) ;
		if (dwMouse & FIM_CAPUTURED){
			if (dwMouse & FIM_MOVED){
				DrawUIBorder (&drc) ;
				GetCursorPos (&pt) ;
				MoveWindow (hStatusWnd,
							pt.x - ptdif.x,
							pt.y - ptdif.y,
							rc.right,
							rc.bottom, TRUE) ;
			}
		}
		hSvrWnd	= (HWND)GetWindowLongPtr (hStatusWnd,	FIGWL_UISTATE_SVRWND) ;
		hIMC	= (HIMC)GetWindowLongPtr (hSvrWnd,		IMMGWLP_IMC) ;
		if (hIMC)
			uiStatus_bOnRightMenu (hIMC, hStatusWnd) ;
		uiStatus_vPaint (hStatusWnd, hDC, 0) ;
		SetWindowLong (hStatusWnd, FIGWL_UISTATE_MOUSE, dwMouse & FIM_FOCUS) ;
		SetWindowLong (hStatusWnd, FIGWL_UISTATE_PUSHSTATUS, 0L) ;
		break ;

	case WM_LBUTTONUP:
		dwMouse	= GetWindowLong (hStatusWnd, FIGWL_UISTATE_MOUSE) ;
		DEBUGPRINTFEX (104, (TEXT ("uiStatus_vOnButton (WM_LBUTTONUP, dwMouse:0x%x)\n"), dwMouse)) ;
		if (dwMouse & FIM_CAPUTURED){
			if (dwMouse & FIM_MOVED){
				DrawUIBorder (&drc) ;
				GetCursorPos (&pt) ;
				MoveWindow (hStatusWnd,
							pt.x - ptdif.x,
							pt.y - ptdif.y,
							rc.right,
							rc.bottom, TRUE) ;
			}
		}
		hSvrWnd	= (HWND)GetWindowLongPtr (hStatusWnd,	FIGWL_UISTATE_SVRWND) ;
		hIMC	= (HIMC)GetWindowLongPtr (hSvrWnd,		IMMGWLP_IMC) ;
		if (hIMC){
			GetCursorPos (&pt) ;
			dwPushedStatus	= GetWindowLong (hStatusWnd, FIGWL_UISTATE_PUSHSTATUS) ;
			dwPushedStatus	&= uiStatus_dwCheckPushedStatus (hStatusWnd, &pt) ;
			if (dwPushedStatus & PUSHED_STATUS_MODE){
				uiStatus_bOnLeftMenu (hIMC, hStatusWnd) ;
			} else if (dwPushedStatus & PUSHED_STATUS_ONOFF){
				BOOL	fOpen	= ImmGetOpenStatus (hIMC) ;
				fOpen	= !fOpen ;
				ImmSetOpenStatus (hIMC, fOpen) ;
			}
		}
		uiStatus_vPaint (hStatusWnd, hDC, 0) ;
		SetWindowLong (hStatusWnd, FIGWL_UISTATE_MOUSE, dwMouse & FIM_FOCUS) ;
		SetWindowLong (hStatusWnd, FIGWL_UISTATE_PUSHSTATUS, 0L) ;
		break ;
	default:
		break ;
	}
	ReleaseDC (hStatusWnd, hDC) ;
	return ;
}

/*
 *	status window NbNꂽ̃j[֐B
 *-----
 *(Rg)
 *	uiStatus_iShowLeftMenu () Ăяo PopupMenu \B
 *	uiStatus_iShowLeftMenu () ̕Ԃl PopupMenu ̂ǂ̍ڂI
 *	̂Aē̓[h̐؂ւsB
 *()
 *	HWND 	hwnd		̃j[ Owner ƂȂ WindowB
 *	HIMC	hIMC		Handle of Input Method Context ȁH
 */
BOOL
uiStatus_bOnLeftMenu (
	HIMC		hIMC,
	HWND		hwnd)
{
	BOOL	fOpen ;
	int	nCmd ;

	/* |bvAbvj[\B*/
	nCmd	= uiStatus_iShowLeftMenu (hIMC, hwnd) ;
	if (nCmd < 0 || nCmd == IDCANCEL)
		return	TRUE ;

	return	ChangeConversionModeWithIDM (hIMC, nCmd) ;
}

/*
 *	status window ENbNꂽ̃j[֐B
 *-----
 *()
 *	HWND 	hwnd		̃j[ Owner ƂȂ WindowB
 *	HIMC	hIMC		Handle of Input Method Context ȁH
 */
BOOL
uiStatus_bOnRightMenu (
	HIMC		hIMC,
	HWND		hwnd)
{
	HWND		hUIWnd ;
	BOOL		fOpen ;
	int		nCmd ;

	nCmd	= uiStatus_iShowRightMenu (hIMC, hwnd) ;
	if (nCmd < 0 || nCmd == IDCANCEL)
		return	TRUE ;
	switch (nCmd){
	case	IDM_SHOW_PROPERTY_WINDOW:
	{
		ImeConfigure (GetKeyboardLayout(0), 0, IME_CONFIG_GENERAL, 0) ;
		break ;
	}
	case	IDM_HIDE_TOOLBAR:
	{
		SetRegDwordValue (TEXT ("\\Window"), TEXT (REGKEY_HIDETOOLBAR), 1) ;
		hUIWnd	= (HWND)GetWindowLongPtr (hwnd, FIGWL_UISTATE_SVRWND) ;
		if (IsWindow (hUIWnd))
			SendMessage (hUIWnd, WM_UI_STATEHIDE, (WPARAM)0, (LPARAM)0) ;
		TSkkIme_vUpdateIndicIcon (hIMC) ;
		break ;
	}
	case	IDM_RECONVERT:
	{
		DWORD	dwSize	= (DWORD) MyImmRequestMessage (hIMC, IMR_RECONVERTSTRING, 0) ;
		if (dwSize) {
			LPRECONVERTSTRING	lpRS ;

			lpRS			= (LPRECONVERTSTRING) GlobalAlloc (GPTR, dwSize) ;
			lpRS->dwSize	= dwSize ;
			if (dwSize = (DWORD) MyImmRequestMessage (hIMC, IMR_RECONVERTSTRING, (LPARAM) lpRS)) {
				LPINPUTCONTEXT			lpIMC ;

				lpIMC	= ImmLockIMC (hIMC) ;
				if (lpIMC != NULL) {
					LPCOMPOSITIONSTRING	lpCompStr ;
					if (ImmGetIMCCSize (lpIMC->hCompStr) < sizeof (MYCOMPSTR))
						goto	pass_1 ;
					/*	ϊ[hIɉ̓[hɐݒ肷B*/
					lpCompStr	= (LPCOMPOSITIONSTRING)ImmLockIMCC (lpIMC->hCompStr) ;
					if (lpCompStr != NULL) {
						if (! ImmGetOpenStatus (hIMC))
							ImmSetOpenStatus (hIMC, TRUE) ;
						TSkkIme_bSetReconvertStr (hIMC, lpIMC, lpCompStr, lpRS, TRUE) ; // ?
						ImmUnlockIMCC (lpIMC->hCompStr) ;
					}
				pass_1:
					ImmUnlockIMC (hIMC) ;
				}
				MyImmRequestMessage(hIMC, IMR_CONFIRMRECONVERTSTRING, (LPARAM)lpRS);
			} else {
				GlobalFree((HANDLE)lpRS);
			}
		}
	}
	default:
		ChangeConversionModeWithIDM (hIMC, nCmd) ;
		break ;
	}
	return	TRUE ;
}

/*
 *	Status Window NbÑj[쐬A[ȖIʂ𓾂֐B
 *----
 *(Rg)
 *	j[쐬Ƃ̂́ACreatePopupMenu () ƂƂŁA
 *	[ȖIʂ𓾂Ƃ̂́ATrackPopupMenu () Ă܂ƂƁB
 *	̊֐ TrackPopupMenu ̕ԂlԂ܂B
 *()
 *	HWND 	hwnd		̃j[ Owner ƂȂ WindowB
 *	HIMC	hIMC		Handle of Input Method Context ȁH
 */
int
uiStatus_iShowLeftMenu (
	HIMC		hIMC,
	HWND		hwnd)
{
	static	MYMENUITEMINFO	myLeftMenuItemInfoTbl []	= {
	  { MIIM_TYPE,	MFT_SEPARATOR, 0, NULL, },
	  { MIIM_STATE | MIIM_ID | MIIM_STRING, 0, IDCANCEL, TEXT ("LZ"), },
	} ;
	HMENU			hMenu ;
	POINT			pos ;
	int				nCmd ;
	MENUITEMINFO	menuItemInfo ;
	LPINPUTCONTEXT	lpIMC ;
	int				i ;
	int				iPosition ;

	GetCursorPos ((LPPOINT)&pos) ;

	/* |bvAbvj[쐬B*/
	hMenu	= CreatePopupMenu () ;
	if (!hMenu)
		return	-1 ;

	iPosition	= 0 ;
	uiStatus_bInsertInputModeMenu (hIMC, hMenu, &iPosition) ;
	/* |bvAbvj[̍ڂB*/
	for (i = 0 ; i < sizeof (myLeftMenuItemInfoTbl) / sizeof (MYMENUITEMINFO) ; i ++){
		int		nResult ;

		menuItemInfo.cbSize			= sizeof (MENUITEMINFO) ;
		menuItemInfo.fMask			= myLeftMenuItemInfoTbl [i].m_fMask ;
		menuItemInfo.fType			= myLeftMenuItemInfoTbl [i].m_fType ;
		menuItemInfo.wID			= myLeftMenuItemInfoTbl [i].m_wID ;
		if (menuItemInfo.fMask & MIIM_STATE)
			menuItemInfo.fState	= MFS_ENABLED ;
		menuItemInfo.hbmpChecked	= NULL ;
		menuItemInfo.hbmpUnchecked	= NULL ;
		//menuItemInfo.hbmpItem		= NULL ;
		menuItemInfo.dwTypeData		= myLeftMenuItemInfoTbl [i].m_dwTypeData ;
		if (!menuItemInfo.dwTypeData){
			menuItemInfo.cch		= 0 ;
		} else {
			menuItemInfo.cch		= lstrlen (menuItemInfo.dwTypeData) ;
		}
		nResult	= InsertMenuItem (hMenu, iPosition, TRUE, (LPMENUITEMINFO)&menuItemInfo) ;
		if (nResult == 0 && (menuItemInfo.fMask & MIIM_STRING) != 0) {
			menuItemInfo.fMask	&= ~MIIM_STRING ;
			menuItemInfo.fMask	|= MIIM_TYPE ;
			menuItemInfo.fType	|= MFT_STRING ;
			nResult	= InsertMenuItem (hMenu, iPosition, TRUE, (LPMENUITEMINFO)&menuItemInfo) ;
		}
		if (nResult != 0)
			iPosition	++ ;
	}
	/* |bvAbvj[\AIʂ𓾂B*/
	nCmd	= TrackPopupMenu (hMenu, TPM_RETURNCMD | TPM_RIGHTALIGN | TPM_BOTTOMALIGN, pos.x, pos.y, 0, hwnd, NULL) ;
	DestroyMenu (hMenu) ;
	return	nCmd ;
}

/*
 *	Status Window ENbÑj[쐬A[ȖIʂ𓾂֐B
 *----
 *(Rg)
 *	uiStatus_iShowLeftMenu QƂ̂ƁB
 *()
 *	HWND 	hwnd		̃j[ Owner ƂȂ WindowB
 *	HIMC	hIMC		Handle of Input Method Context ȁH
 */
int
uiStatus_iShowRightMenu (
	HIMC		hIMC,
	HWND		hwnd)
{
	static	MYMENUITEMINFOEX	myRightMenuItemInfoTbl []	= {
	  { MIIM_STATE | MIIM_ID | MIIM_STRING,				0,
		IDM_HIDE_TOOLBAR,			TEXT ("^XNo[ɓ(&I)"),	NULL },
	  { MIIM_TYPE,										MFT_SEPARATOR,	
		0,							NULL,								NULL },
	  { MIIM_STATE | MIIM_ID | MIIM_STRING,				0,
		IDM_SHOW_PROPERTY_WINDOW,	TEXT ("vpeB(&R)"), 			NULL },
	  { MIIM_TYPE,										MFT_SEPARATOR,
		0,							NULL,								NULL },
	  { MIIM_STATE | MIIM_ID | MIIM_STRING,				0,
		IDM_RECONVERT,				TEXT ("ĕϊ(&C)"),				NULL },
	  { MIIM_TYPE,										MFT_SEPARATOR,	
		0,							NULL,								NULL },
	  { MIIM_STATE | MIIM_SUBMENU | MIIM_STRING,		0,
		0,							TEXT ("̓[h(&N)"),			NULL },
	  { MIIM_TYPE,										MFT_SEPARATOR,	
		0,							NULL,								NULL },
	  { MIIM_STATE | MIIM_ID | MIIM_STRING,				0,
		IDCANCEL,					TEXT ("LZ"),				NULL },
	} ;
	HMENU	hMenu ;
	HMENU	hSubMenu ;
	int	nCmd ;
	int				iMenuItem ;
	POINT			pos ;
	MENUITEMINFO	menuItemInfo ;

	GetCursorPos ((LPPOINT)&pos) ;
	hMenu		= CreatePopupMenu () ;
	if (!hMenu)
		return	-1 ;
	hSubMenu	= CreatePopupMenu () ;
	if (!hSubMenu){
		DestroyMenu (hMenu) ;
		return	-1 ;
	}
	iMenuItem	= 0 ;
	uiStatus_bInsertInputModeMenu (hIMC, hSubMenu, &iMenuItem) ;

	iMenuItem	= 0 ;
	while (iMenuItem < ARRAYSIZE (myRightMenuItemInfoTbl)) {
		int		nResult ;

		menuItemInfo.cbSize			= sizeof (MENUITEMINFO) ;
		menuItemInfo.fMask			= myRightMenuItemInfoTbl [iMenuItem].m_fMask ;
		menuItemInfo.fType			= myRightMenuItemInfoTbl [iMenuItem].m_fType ;
		menuItemInfo.wID			= myRightMenuItemInfoTbl [iMenuItem].m_wID ;
		menuItemInfo.fState			= (myRightMenuItemInfoTbl [iMenuItem].m_pGetStateProc != NULL)?  (myRightMenuItemInfoTbl [iMenuItem].m_pGetStateProc)() : 0 ;
		if (menuItemInfo.fMask & MIIM_STATE)
			menuItemInfo.fState		|= MFS_ENABLED ;
		menuItemInfo.hSubMenu		= (menuItemInfo.fMask & MIIM_SUBMENU)? hSubMenu : 0 ;
		menuItemInfo.hbmpChecked	= NULL ;
		menuItemInfo.hbmpUnchecked	= NULL ;
		//menuItemInfo.hbmpItem		= NULL ;
		menuItemInfo.dwTypeData		= myRightMenuItemInfoTbl [iMenuItem].m_dwTypeData ;
		if (!menuItemInfo.dwTypeData){
			menuItemInfo.cch		= 0 ;
		} else {
			menuItemInfo.cch		= lstrlen (myRightMenuItemInfoTbl [iMenuItem].m_dwTypeData) ;
		}
		nResult	= InsertMenuItem (hMenu, iMenuItem, TRUE, (LPMENUITEMINFO)&menuItemInfo) ;
		if (nResult == 0 && (menuItemInfo.fMask & MIIM_STRING) != 0) {
			/*	Windows2000  MIIM_STRING ̈sBMIIM_TYPE ĂƓȂƂ
			 *	svcȓ悤ɌB
			 */
			menuItemInfo.fMask	&= ~MIIM_STRING ;
			menuItemInfo.fMask	|= MIIM_TYPE ;
			menuItemInfo.fType	|= MFT_STRING ;
			nResult	= InsertMenuItem (hMenu, iMenuItem, TRUE, (LPMENUITEMINFO)&menuItemInfo) ;
		}
		if (nResult != 0)
			iMenuItem	++ ;
	}
	nCmd	= TrackPopupMenu (hMenu, TPM_RETURNCMD | TPM_RIGHTALIGN | TPM_BOTTOMALIGN, pos.x, pos.y, 0, hwnd, NULL) ;
	DestroyMenu (hMenu) ;
	DestroyMenu (hSubMenu) ;
	return	nCmd ;
}

BOOL
uiStatus_bInsertInputModeMenu (
	HIMC		hIMC,
	HMENU		hMenu,
	int FAR*	lpiPosition)
{
	static	MYMENUITEMINFO	myMenuItemInfoTbl []	= {
	  {	MIIM_STATE | MIIM_TYPE | MIIM_ID | MIIM_STRING,
		MFT_RADIOCHECK,
		IDM_CMODE_TO_ASCII,			TEXT ("AXL[(&S)"), },
	  { MIIM_STATE | MIIM_TYPE | MIIM_ID | MIIM_STRING,
		MFT_RADIOCHECK,
		IDM_CMODE_TO_ROMANHIRA,		TEXT ("[}(&H)"), },
	  {	MIIM_STATE | MIIM_TYPE | MIIM_ID | MIIM_STRING,
		MFT_RADIOCHECK,
		IDM_CMODE_TO_ROMANKATA,		TEXT ("[}Љ(&K)"), },
	  { MIIM_STATE | MIIM_TYPE | MIIM_ID | MIIM_STRING,
		MFT_RADIOCHECK,
		IDM_CMODE_TO_JISX0201KANA,	TEXT ("pЉ(&A)"), },
	  { MIIM_STATE | MIIM_TYPE | MIIM_ID | MIIM_STRING,
		MFT_RADIOCHECK,
		IDM_CMODE_TO_JISX0201ROMAN,	TEXT ("pp(&R)"), },
	  { MIIM_STATE | MIIM_TYPE | MIIM_ID | MIIM_STRING,
		MFT_RADIOCHECK,
		IDM_CMODE_TO_ZENEI,			TEXT ("Spp(&E)"), },
	  { MIIM_STATE | MIIM_TYPE | MIIM_ID | MIIM_STRING,
		MFT_RADIOCHECK,
		IDM_CMODE_TO_DIRECTINPUT,	TEXT ("ړ(&C)"), },
	} ;
	/* IME_CMODE_XXX  myMenuItemInfoTbl ̑Ή\D*/
	static const int	myInputModeTbl []	= {
		1,	/* [}(&H) */
		2,	/* [}Љ(&K) */
		5,	/* Spp(&E) */
		3,	/* pЉ(&A) */
		4,	/* pp(&R) */
		0,	/* ASCII (&S) */
		6,	/* ړ(&C) */
	} ;
	LPINPUTCONTEXT	lpIMC ;
	int			iNum ;
	int			i ;
	int			iPosition ;
	MENUITEMINFO			menuItemInfo ;

	/* ̐`FbNB*/
	if (!hMenu || !lpiPosition)
		return	FALSE ;

	/* ݂̓̓[h𓾂B*/
	lpIMC	= (LPINPUTCONTEXT)ImmLockIMC (hIMC) ;
	if (!lpIMC)
		return	FALSE ;

	iNum	= myInputModeTbl [iGetConversionModeFromLPIMC (lpIMC)] ;
	ImmUnlockIMC (hIMC) ;
	iPosition	= *lpiPosition ;
	/* j[ǉB*/
	for (i = 0 ; i < sizeof (myMenuItemInfoTbl) / sizeof (MYMENUITEMINFO) ; i ++){
		int	nResult ;

		menuItemInfo.cbSize			= sizeof (MENUITEMINFO) ;
		menuItemInfo.fMask			= myMenuItemInfoTbl [i].m_fMask ;
		menuItemInfo.fType			= myMenuItemInfoTbl [i].m_fType ;
		menuItemInfo.wID			= myMenuItemInfoTbl [i].m_wID ;
		menuItemInfo.fState			= (i == iNum)? MFS_CHECKED : MFS_UNCHECKED ;
		menuItemInfo.hbmpChecked	= NULL ;
		menuItemInfo.hbmpUnchecked	= NULL ;
		//menuItemInfo.hbmpItem		= NULL ;
		menuItemInfo.dwTypeData		= myMenuItemInfoTbl [i].m_dwTypeData ;
		menuItemInfo.cch			= lstrlen (menuItemInfo.dwTypeData) ;

		nResult	= InsertMenuItem (hMenu, iPosition, TRUE, (LPMENUITEMINFO)&menuItemInfo) ;
		if (nResult == 0 && (menuItemInfo.fMask & MIIM_STRING) != 0) {
			menuItemInfo.fMask	&= ~MIIM_STRING ;
			menuItemInfo.fMask	|= MIIM_TYPE ;
			menuItemInfo.fType	|= MFT_STRING ;
			nResult	= InsertMenuItem (hMenu, iPosition, TRUE, (LPMENUITEMINFO)&menuItemInfo) ;
		}
		if (nResult != 0)
			iPosition	++ ;
	}
	*lpiPosition	= iPosition ;
	return	TRUE ;
}

/*
 *	ϊ[h̃rbg}bv쐬֐B
 *-----
 *(Rg)
 *	ϊ[h = [hAړ̓[hȂ \̂Ɏg bitmap 
 *	Resource ǂݍށB
 *()
 *	HWND	hwnd		̃rbg}bv𗘗p WindowB
 *
 *	[ށAȂƂȂcKvAB
 */
HBITMAP
uiStatus_hCreateCModeBitmap (HWND hwnd)
{
	HDC		hDC, hMemDC, hSrcMemDC, hOrgMemDC ;
	HBRUSH		hbrushOld ;
	HBITMAP	hbmpStatus, hbmpOld ;
	HBITMAP	hbmpSrcOld,	hbmpSrc ;
	HBITMAP	hbmpOrgOld, hbmpOrg ;
	int		y, i ;

	hDC			= CreateDC (TEXT ("DISPLAY"), NULL, NULL, NULL) ;
	hbmpStatus	= CreateCompatibleBitmap (hDC, STATUSCMODE_BTNX * 4, STATUSCMODE_BTNY * NUM_STATUS_BMP_MODES) ;
	hMemDC		= CreateCompatibleDC (hDC) ;
	hbmpOld		= SelectObject (hMemDC, hbmpStatus) ;
	hbrushOld	= SelectObject (hMemDC, GetSysColorBrush (COLOR_3DFACE)) ;
	PatBlt (hMemDC, 0, 0, STATUSCMODE_BTNX * 4, STATUSCMODE_BTNY * NUM_STATUS_BMP_MODES, PATCOPY) ;
	SelectObject (hMemDC, hbrushOld) ;

	hbmpOrg		= (HBITMAP)GetWindowLongPtr (hwnd, FIGWL_UISTATE_STATUSORGBMP) ;
	hOrgMemDC	= CreateCompatibleDC (hDC) ;
	hbmpOrgOld	= SelectObject (hOrgMemDC, hbmpOrg) ;
	hbmpSrc		= CreateCompatibleBitmap (hDC, STATUSCMODE_BTNX * NUM_WIDTH_STATUS_BMP, STATUSCMODE_BTNY * NUM_HEIGHT_STATUS_BMP) ;
	hSrcMemDC	= CreateCompatibleDC (hDC) ;
	hbmpSrcOld	= SelectObject (hSrcMemDC, hbmpSrc) ;
	BitBlt (hSrcMemDC, 0,					0, STATUSCMODE_BTNX, STATUSCMODE_BTNY * NUM_HEIGHT_STATUS_BMP, hOrgMemDC, 0, 0, SRCCOPY) ;
	BitBlt (hSrcMemDC, STATUSCMODE_BTNX,	0, STATUSCMODE_BTNX, STATUSCMODE_BTNY * NUM_HEIGHT_STATUS_BMP, hOrgMemDC, 0, 0, NOTSRCCOPY) ;

	for (i = 0 ; i < NUM_STATUS_BMP_MODES ; i ++){
		BitBlt (hMemDC, STATUSCMODE_BTNX * i, 0, STATUSCMODE_BTNX, STATUSCMODE_BTNY * INDEX_STATUS_BMP_LU, hSrcMemDC, 0, 0, SRCAND) ;
	}
	for (y = 0 ; y < NUM_STATUS_BMP_MODES ; y ++){
		BitBlt (hMemDC, STATUSCMODE_BTNX * 1, STATUSCMODE_BTNY * y, STATUSCMODE_BTNX, STATUSCMODE_BTNY, hSrcMemDC, 0, STATUSCMODE_BTNY * INDEX_STATUS_BMP_LU, SRCINVERT) ;
		BitBlt (hMemDC, STATUSCMODE_BTNX * 1, STATUSCMODE_BTNY * y, STATUSCMODE_BTNX, STATUSCMODE_BTNY, hSrcMemDC, 0, STATUSCMODE_BTNY * INDEX_STATUS_BMP_RB, SRCINVERT) ;
		BitBlt (hMemDC, STATUSCMODE_BTNX * 2, STATUSCMODE_BTNY * y, STATUSCMODE_BTNX, STATUSCMODE_BTNY, hSrcMemDC, 0, STATUSCMODE_BTNY * INDEX_STATUS_BMP_RB2, SRCINVERT) ;
		BitBlt (hMemDC, STATUSCMODE_BTNX * 2, STATUSCMODE_BTNY * y, STATUSCMODE_BTNX, STATUSCMODE_BTNY, hSrcMemDC, 0, STATUSCMODE_BTNY * INDEX_STATUS_BMP_LU2, SRCINVERT) ;
	}

	/*	̒iKł 0,1,6-9  temporary ̍Ɨ̈ƂėpB
	 *	ʂ́A0-5ɍ쐬̂ŁA̕Ȃ΂ȂȂB
	 */
	hbrushOld	= SelectObject (hSrcMemDC, GetSysColorBrush (COLOR_3DLIGHT)) ;
	PatBlt (hSrcMemDC, 0, STATUSCMODE_BTNY * 0, STATUSCMODE_BTNX, STATUSCMODE_BTNY, PATCOPY) ;
	SelectObject (hSrcMemDC, hbrushOld) ;
	hbrushOld	= SelectObject (hSrcMemDC, GetSysColorBrush (COLOR_3DSHADOW)) ;
	PatBlt (hSrcMemDC, 0, STATUSCMODE_BTNY * 1, STATUSCMODE_BTNX, STATUSCMODE_BTNY, PATCOPY) ;
	SelectObject (hSrcMemDC, hbrushOld) ;
	BitBlt (hSrcMemDC, 0, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 0), STATUSCMODE_BTNX, STATUSCMODE_BTNY,
			hSrcMemDC, 0, STATUSCMODE_BTNY * 0, SRCAND) ;
	BitBlt (hSrcMemDC, 0, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 1), STATUSCMODE_BTNX, STATUSCMODE_BTNY,
			hSrcMemDC, 0, STATUSCMODE_BTNY * 1, SRCAND) ;
	BitBlt (hSrcMemDC, 0, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 2), STATUSCMODE_BTNX, STATUSCMODE_BTNY,
			hSrcMemDC, 0, STATUSCMODE_BTNY * 0, SRCAND) ;
	BitBlt (hSrcMemDC, 0, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 3), STATUSCMODE_BTNX, STATUSCMODE_BTNY,
			hSrcMemDC, 0, STATUSCMODE_BTNY * 1, SRCAND) ;
	for (y = 0 ; y < NUM_STATUS_BMP_MODES ; y ++){
		BitBlt (hMemDC, STATUSCMODE_BTNX * 1, STATUSCMODE_BTNY * y, STATUSCMODE_BTNX, STATUSCMODE_BTNY, hSrcMemDC, 0, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 0), SRCPAINT) ;
		BitBlt (hMemDC, STATUSCMODE_BTNX * 1, STATUSCMODE_BTNY * y, STATUSCMODE_BTNX, STATUSCMODE_BTNY, hSrcMemDC, 0, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 1), SRCPAINT) ;
		BitBlt (hMemDC, STATUSCMODE_BTNX * 2, STATUSCMODE_BTNY * y, STATUSCMODE_BTNX, STATUSCMODE_BTNY, hSrcMemDC, 0, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 2), SRCPAINT) ;
		BitBlt (hMemDC, STATUSCMODE_BTNX * 2, STATUSCMODE_BTNY * y, STATUSCMODE_BTNX, STATUSCMODE_BTNY, hSrcMemDC, 0, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 3), SRCPAINT) ;
	}

	hbrushOld	= SelectObject (hSrcMemDC, GetSysColorBrush (COLOR_GRAYTEXT)) ;
	PatBlt (hSrcMemDC, 0, STATUSCMODE_BTNY * 0, STATUSCMODE_BTNX, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 0), PATCOPY) ;
	SelectObject (hSrcMemDC, hbrushOld) ;
	BitBlt (hSrcMemDC, 0, 0, STATUSCMODE_BTNX * 1, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 0), hSrcMemDC, STATUSCMODE_BTNX, 0, SRCAND) ;
	BitBlt (hMemDC, STATUSCMODE_BTNX * 3, 0, STATUSCMODE_BTNX, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 0), hSrcMemDC, 0, 0, SRCPAINT) ;
	BitBlt (hMemDC, STATUSCMODE_BTNX * 3, 0, STATUSCMODE_BTNX, STATUSCMODE_BTNY * (NUM_STATUS_BMP_MODES + 0), hOrgMemDC, STATUSCMODE_BTNX * 1, 0, SRCPAINT) ;
	SelectObject (hOrgMemDC, hbmpOrgOld) ;
	DeleteDC (hOrgMemDC) ;

	SelectObject (hSrcMemDC, hbmpSrcOld) ;
	DeleteDC (hSrcMemDC) ;
	DeleteObject (hbmpSrc) ;
	SelectObject (hMemDC, hbmpOld) ;
	DeleteDC (hMemDC) ;
	DeleteDC (hDC) ;
	return	hbmpStatus ;
}

HBITMAP
uiStatus_hCreateOnOffBitmap (
	HWND		hwnd)
{
	HDC		hDC, hMemDC, hSrcMemDC, hOrgMemDC ;
	HBRUSH		hbrushOld ;
	HBITMAP	hbmpStatus, hbmpOld ;
	HBITMAP	hbmpSrcOld,	hbmpSrc ;
	HBITMAP	hbmpOrgOld, hbmpOrg ;

	hDC			= CreateDC (TEXT ("DISPLAY"), NULL, NULL, NULL) ;
	hbmpStatus	= CreateCompatibleBitmap (hDC, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 2) ;
	hMemDC		= CreateCompatibleDC (hDC) ;
	hbmpOld		= SelectObject (hMemDC, hbmpStatus) ;
	hbrushOld	= SelectObject (hMemDC, GetSysColorBrush (COLOR_3DFACE)) ;
	PatBlt (hMemDC, 0, 0, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 2, PATCOPY) ;
	SelectObject (hMemDC, hbrushOld) ;

	hbmpOrg		= (HBITMAP)GetWindowLongPtr (hwnd, FIGWL_UISTATE_CLOSEORGBMP) ;
	hOrgMemDC	= CreateCompatibleDC (hDC) ;
	hbmpOrgOld	= SelectObject (hOrgMemDC, hbmpOrg) ;
	hbmpSrc		= CreateCompatibleBitmap (hDC, UISTATE_ONOFFBMP_SRC_WIDTH, UISTATE_ONOFFBMP_SRC_HEIGHT) ;
	hSrcMemDC	= CreateCompatibleDC (hDC) ;
	hbmpSrcOld	= SelectObject (hSrcMemDC, hbmpSrc) ;
	BitBlt (hSrcMemDC, 0, 0, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 5, hOrgMemDC, 0, 0, SRCCOPY) ;
	SelectObject (hOrgMemDC, hbmpOrgOld) ;
	DeleteDC (hOrgMemDC) ;

	BitBlt (hMemDC, 0, STATUSONOFF_BTNY * 0, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1,
			hSrcMemDC, 0, STATUSONOFF_BTNY * 2, SRCINVERT) ;
	BitBlt (hMemDC, 0, STATUSONOFF_BTNY * 1, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1,
			hSrcMemDC, 0, STATUSONOFF_BTNY * 2, SRCINVERT) ;
	BitBlt (hMemDC, 0, STATUSONOFF_BTNY * 0, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1,
			hSrcMemDC, 0, STATUSONOFF_BTNY * 0, SRCPAINT) ;
	BitBlt (hMemDC, 0, STATUSONOFF_BTNY * 1, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1,
			hSrcMemDC, 0, STATUSONOFF_BTNY * 1, SRCPAINT) ;
	hbrushOld	= SelectObject (hSrcMemDC, GetSysColorBrush (COLOR_3DLIGHT)) ;
	PatBlt (hSrcMemDC, 0, STATUSONOFF_BTNY * 0, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1, PATCOPY) ;
	SelectObject (hSrcMemDC, hbrushOld) ;
	hbrushOld	= SelectObject (hSrcMemDC, GetSysColorBrush (COLOR_3DSHADOW)) ;
	PatBlt (hSrcMemDC, 0, STATUSONOFF_BTNY * 1, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1, PATCOPY) ;
	SelectObject (hSrcMemDC, hbrushOld) ;
	BitBlt (hMemDC, 0, STATUSONOFF_BTNY * 0, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1,
			hSrcMemDC, 0, STATUSONOFF_BTNY * 3, SRCINVERT) ;
	BitBlt (hMemDC, 0, STATUSONOFF_BTNY * 1, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1,
			hSrcMemDC, 0, STATUSONOFF_BTNY * 3, SRCINVERT) ;
	BitBlt (hMemDC, 0, STATUSONOFF_BTNY * 0, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1,
			hSrcMemDC, 0, STATUSONOFF_BTNY * 4, SRCINVERT) ;
	BitBlt (hMemDC, 0, STATUSONOFF_BTNY * 1, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1,
			hSrcMemDC, 0, STATUSONOFF_BTNY * 4, SRCINVERT) ;
	BitBlt (hSrcMemDC, 0, STATUSONOFF_BTNY * 3, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 2,
			hSrcMemDC, 0, STATUSONOFF_BTNY * 0, SRCAND) ;
	BitBlt (hMemDC, 0, STATUSONOFF_BTNY * 0, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 2, 
			hSrcMemDC, 0, STATUSONOFF_BTNY * 3, SRCPAINT) ;
	BitBlt (hMemDC, 0, STATUSONOFF_BTNY * 0, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1, 
			hSrcMemDC, 0, STATUSONOFF_BTNY * 4, SRCPAINT) ;
	BitBlt (hMemDC, 0, STATUSONOFF_BTNY * 1, STATUSONOFF_BTNX * 3, STATUSONOFF_BTNY * 1, 
			hSrcMemDC, 0, STATUSONOFF_BTNY * 3, SRCPAINT) ;
	SelectObject (hSrcMemDC, hbmpSrcOld) ;
	DeleteObject (hbmpSrc) ;
	DeleteDC (hSrcMemDC) ;
	SelectObject (hMemDC, hbmpOld) ;
	DeleteDC (hMemDC) ;
	DeleteDC (hDC) ;
	return	hbmpStatus ;
}

/*
 *	Status Window NbNꂽɂ̏ꏊɊ蓖Ăꂽ@\𒲂ׂ֐B
 *----
 */
DWORD	PASCAL
uiStatus_dwCheckPushedStatus (
	HWND			hStatusWnd,
	LPPOINT		lppt)
{
	POINT	pt ;
	RECT	rc ;

	if (lppt){
		pt	= *lppt ;
		ScreenToClient (hStatusWnd, &pt) ;
		GetClientRect  (hStatusWnd, &rc) ;
		if (!PtInRect (&rc, pt))
			return	0L ;
		if (pt.x < STATUSONOFF_BTNX){
			if (pt.y < 10){
				return	PUSHED_STATUS_ONOFF ;
			} else {
				return	PUSHED_STATUS_MOVE ;
			}
		}
		return	PUSHED_STATUS_MODE ;
	}
	return	0L ;
}

void	PASCAL
uiStatus_vChangeCursor (
	HWND		hwnd)
{
	POINT	pt ;
	HCURSOR	hCursor ;
	BOOL	fIn ;
	RECT	rc ;

	/* }EXJ[\̈ʒu𓾂B*/
	GetCursorPos (&pt) ;
	GetWindowRect (hwnd, &rc) ;
	fIn	= PtInRect (&rc, pt) ;
	ScreenToClient (hwnd, &pt) ;

	hCursor	= (HCURSOR)GetWindowLongPtr (hwnd, FIGWL_UISTATE_CURSOR) ;
	/* }EXJ[\̈ʒuJ[\̌`ǂ̂悤ɂ邩肷B*/
	if (STATUSONOFF_X <= pt.x && pt.x < STATUSCMODE_X &&
		10            <= pt.y && pt.y < STATUSONOFF_BTNY){
		/* ݂̃J[\ۑB񂫂B*/
		if (!hCursor)
			SetWindowLongPtr (hwnd, FIGWL_UISTATE_CURSOR, (LONG_PTR)GetCursor ()) ;
		/* J[\̌`ύXB*/
		(void)SetCursor (LoadCursor (NULL, IDC_SIZEALL)) ;
	} else {
		/* ȑÕJ[\ɖ߂B*/
		hCursor	= hCursor? hCursor : LoadCursor (NULL, IDC_ARROW) ;
		if (hCursor) 
			SetCursor (hCursor) ;
		SetWindowLongPtr (hwnd, FIGWL_UISTATE_CURSOR, (LONG_PTR)0) ;
	}
	return ;
}


