/* # SKKIME1.0 (Simple Kana-Kanji Input Method Editor for Windows2000)
 * imm.c
 * This file is part of skkime1.0.
 * Copyright (C) 1999
 * Takashi SAKAMOTO (tatari_sakamoto@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "skki1_5.h"
#include "resource.h"
#include "ImeDoc.h"
#include "TSearchSession.h"
#include "TLispSession.h"
#include "skkui.h"
#include "jstring.h"
#include "ImeConfig.h"


/*
 *	Private Prototypes
 */
static	DWORD	PASCAL	getImeLeftMenuItems		(HIMC, DWORD, DWORD, LPMYIMEMENUITEMINFO, LPMYIMEMENUITEMINFO, DWORD) ;
static	DWORD	PASCAL	getImeRightMenuItems	(HIMC, DWORD, DWORD, LPMYIMEMENUITEMINFO, LPMYIMEMENUITEMINFO, DWORD) ;
static	DWORD	PASCAL	getImeRightSubMenuItems	(HIMC, DWORD, DWORD, LPMYIMEMENUITEMINFO, LPMYIMEMENUITEMINFO, DWORD) ;
static	long	PASCAL	insertStringMenuItem	(LPMYIMEMENUITEMINFO, const LPMYSTRTYPEPAIR, UINT) ;
static	long	PASCAL	insertStringMenuItemEx	(LPMYIMEMENUITEMINFO, const LPMYSTRTYPEPAIREX, UINT) ;
static	long	PASCAL	insertSeparatorMenuItem	(LPMYIMEMENUITEMINFO) ;
static	long	PASCAL	insertInputModeMenuItems(HIMC, LPMYIMEMENUITEMINFO) ;

/***********************************************************************
 *	ImeInquire ()
 *		For Windows 95, Windows 98, and Windows NT 3.51
 *	ImeInquire ֐ IME ̏舵B܂AImeInquire ֐
 *	IMEINFO \̋y IME  UI class name ԂB
 *	BOOL
 *		ImeInquire(
 *		LPIMEINFO lpIMEInfo,
 *		LPTSTR lpszWndClass,
 *		LPCTSTR lpszData
 *	)
 *	Parameters
 *	lpIMEInfo
 *		IME info \̂w|C^B
 *	lpszWndClass
 *		IME ɂ window class l߂B̖O IME  UI
 *		class łB
 *	lpszData
 *		IME IvVubNB̃o[Wł NULLB
 *
 *		For Windows NT 4.0 and Windows 2000
 *	BOOL
 *		ImeInquire(
 *		LPIMEINFO lpIMEInfo,
 *		LPTSTRlpszWndClass,
 *		DWORD dwSystemInfoFlags
 *	)
 *	Parameters
 *	lpIMEInfo
 *		IME info \̂w|C^B
 *	lpszWndClass
 *		IME ɂ window class l߂B̖O IME  UI
 *		class łB
 *	dwSystemInfoFlags
 *		VXeɂĒ񋟂VXeύXB̂悤ȃtO
 *		^B
 *		IME_SYSINFO_WINLOGON
 *			IME ɃNCAgvZX Winlogon vZXł邱Ƃ IME
 *			ɒm点BIME ͂̃tOw肳ꂽ IME ̐ݒ[U
 *			ׂł͂ȂB
 *		IME_SYSINFO_WOW16
 *			IME ɃNCAgvZX 16-bit AvP[Vł邱
 *			m点B
 *	Return Values
 *		΁ATRUEBłȂ΁AFALSEB
 ***********************************************************************/
BOOL	WINAPI
ImeInquire (
	LPIMEINFO	lpIMEInfo,
	LPTSTR		lpszClassName,
	DWORD		dwSystemInfoFlags)
{
#if defined (_DEBUG)
	static	TCHAR	tszLastUser [UNLEN + 1] = { TEXT ('\0') } ;
#endif
	/*
	 * IMEINFO \̂B
	 */
	lpIMEInfo->dwPrivateDataSize 	= sizeof (UIEXTRA) ;
	lpIMEInfo->fdwProperty			= IME_PROP_KBD_CHAR_FIRST |
#if defined (MIXED_UNICODE_ANSI) || defined (UNICODE)
									  IME_PROP_UNICODE |
#endif
									  IME_PROP_AT_CARET ;
	lpIMEInfo->fdwConversionCaps	= IME_CMODE_LANGUAGE |
									  IME_CMODE_FULLSHAPE |
									  IME_CMODE_ROMAN |
							   		  IME_CMODE_CHARCODE ;
	lpIMEInfo->fdwSentenceCaps		= 0L ;
	lpIMEInfo->fdwUICaps			= UI_CAP_2700 ;

	/*
	 *	fdwSCSCaps
	 *		fdwSCSCaps rbg SetCompositionString ̔\͂w肷B
	 * 		----------------------------------------------------------
	 *		SCS_CAP_COMPSTR
	 *			IME  SCS_SETSTR  composition string 𐶐łB
	 *		SCS_CAP_MAKEREAD
	 *			ImmSetCompositionString  SCS_SETSTR ŌĂяo
	 *			IME  compositoin string  reading string  lpRead 
	 *			Ȃɐ邱ƂłB̔\͂̂ IME ̉
	 *			̓AvP[V SCS_SETSTR  lpRead ݒ肷
	 *			KvȂB
	 *		SCS_CAP_CONVERTSTRING
	 *			IME  reconvert @\ĂB
	 * 		----------------------------------------------------------
	 *		SKK  composition string 𒼐ڑ삳悤ȃC^[
	 *		tF[XƖꒃǂȂ̂ŁA̔\͎͂
	 *		B
	 */
	lpIMEInfo->fdwSCSCaps		= SCS_CAP_COMPSTR | SCS_CAP_MAKEREAD | SCS_CAP_SETRECONVERTSTRING ;
	lpIMEInfo->fdwSelectCaps	= SELECT_CAP_CONVERSION ;

	/*
	 * Ăяo IME UI class m邱Ƃł悤 lpszClassName
	 *  UI class B
	 */
#if defined (MIXED_UNICODE_ANSI)
	memcpy ((LPWSTR)lpszClassName, (LPWSTR)g_wszUIClassName, (lstrlenW((LPWSTR)g_wszUIClassName) + 1) * sizeof(WCHAR)) ;
#else
	lstrcpy (lpszClassName, (LPTSTR)g_szUIClassName) ;
#endif

	/*	Logon Session ǂ */
	g_bSkkImeSecure	= ((dwSystemInfoFlags & IME_SYSINFO_WINLOGON) != 0) ;
	return	TRUE ;
}

/***********************************************************************
 *	ImeConversionList
 *	ImeConversionList ֐ɂĕʂ͕̕̕ϊʂ̃X
 *	g𓾂邱ƂłB
 *		DWORD
 *			IMEConversionList(
 *			HIMC			hIMC,
 *			LPCTSTR			lpSrc,
 *			LPCANDIDATELIST	lpDst,
 *			DWORD			dwBufLen,
 *			UINT			uFlag
 *			)
 *	Parameters
 *	hIMC
 *		̓ReLXg̃nhB
 *	lpSrc
 *		ϊ镶
 *	lpDst
 *		ϊʂi[obt@w|C^B
 *	dwBufLen
 *		ϊʂi[obt@̑傫B
 *	uFlag
 *		݂͎3̃tO1ݒ肷邱ƂłB
 *		EGCL_CONVERSION
 *			ǂݕ lpSrc p[^ɐݒ肷BIME  lpDst p
 *			[^ɂ̕ϊʂԂB
 *		EGCL_REVERSECONVERSION
 *			ϊʕ lpSrc p[^ɐݒ肷BIME ͂̓ǂ݂
 *			lpDstp[^ɕԂB
 *		EGCL_REVERSE_LENGTH
 *			lpSrc p[^ɕϊʕݒ肷BIME 
 *			GCL_REVERSECONVERSION ̈钷ԂBႦ΁AIME 
 *			sentence period ϊʂǂ݂ɕϊ邱Ƃł
 *			BʂƂāAsentence period oCgPʂŕ
 *			ԂB
 *	Return Values
 *		ϊʕ񃊃Xg̃oCgԂB
 *	Comments
 *		̊֐̓AvP[V IME-related bZ[W
 *		Ȃ IME ɂČĂяo邱ƂӐ}ĂB䂦ɁAIME
 *		͂̊֐ IME-related bZ[W𐶐ׂłȂB
 ***********************************************************************/
DWORD	WINAPI
ImeConversionList (
	HIMC				hIMC,
	LPCTSTR			lpSource,
	LPCANDIDATELIST	pCandList,
	DWORD				dwBufLen,
	UINT				uFlags)
{
#if defined (MIXED_UNICODE_ANSI)
	return	0 ;
#else
	struct CTSearchSession*	pHenkanSession ;
	LPCWSTR		wstrResult ;
	int			nNeedLength, nCandidate, iAnnotationType ;
	DWORD		dwSize ;
	WCHAR		bufTemp [256] ;

	/* tϊ͎słȂB
	if (uFlags == GCL_REVERSECONVERSION || uFlags == GCL_REVERSE_LENGTH)
		return	0 ;
	*/
	if (uFlags != GCL_CONVERSION)
		return	0 ;

	/*	lϊ̐ݒ肪܂Ȃ̂͂܂cB
	 */
	pHenkanSession	= THenkanSession_pCreate (lpSource, lstrlen (lpSource), SEARCH_OKURI_NASHI, ImeConfig_bSkkNumericConversionp (), ImeConfig_bSkkNumConvertFloatp ()) ;
	if (pHenkanSession == NULL)
		return	0 ;
	iAnnotationType	= ImeConfig_iGetSkkShowAnnotationType () ;
	nCandidate		= 0 ;
	nNeedLength		= 0 ;
	for ( ; ; ) {
		wstrResult	= TSearchSession_pGetReferCandidate (pHenkanSession) ;
		if (wstrResult == NULL)
			wstrResult	= TSearchSession_pGetCandidate (pHenkanSession) ;
		if (wstrResult != NULL) {
			int			nResultLen ;

			/*	Lisp Conversion  Annotation ̏B
			 */
			nResultLen	= lstrlenW (wstrResult) ;
			if (iAnnotationType != DISABLE_ANNOTATION) {
				LPCWSTR	pwSrc, pwNote ;
				int		nWordLen, nNoteLen ;

				pwSrc	= wstrResult + nResultLen - 1 ;
				while (pwSrc >= wstrResult && *pwSrc != L';') 
					pwSrc	-- ;
				if (pwSrc >= wstrResult && *pwSrc == L';') {
					pwNote		= pwSrc + 1 ;
					nNoteLen	= nResultLen - (pwNote - wstrResult) ;
					nWordLen	= pwSrc - wstrResult ;
				} else {
					nNoteLen	= 0 ;
					nWordLen	= nResultLen ;
				}
				if (nWordLen > 1 && wstrResult [0] == L'(' && wstrResult [nWordLen - 1] == L')') {
					if (TLispSession_bEval (wstrResult, nWordLen, bufTemp, ARRAYSIZE (bufTemp) - 1)) {
						bufTemp [ARRAYSIZE (bufTemp) - 1]	= L'\0' ;
						nWordLen	= lstrlenW (bufTemp) ;
					}
				}
				if (nNoteLen > 1 && pwNote [0] == L'(' && pwNote [nNoteLen - 1] == L')') {
					if (TLispSession_bEval (pwNote, nNoteLen, bufTemp, ARRAYSIZE (bufTemp) - 1)) {
						bufTemp [ARRAYSIZE (bufTemp) - 1]	= L'\0' ;
						nNoteLen	= lstrlenW (bufTemp) ;
					}
				}
				nResultLen	= nWordLen + nNoteLen + ((nNoteLen > 0)? 1 : 0)  ;
			}
			nNeedLength	+= nResultLen + 1 ; 
			nCandidate	++ ;
		}
		if (! TSearchSession_bNextCandidate (pHenkanSession))
			break ;
	}
	TSearchSession_bRewind (pHenkanSession) ;

	dwSize	= sizeof (CANDIDATELIST) + sizeof (DWORD) * nCandidate + sizeof (MYCHAR) * nNeedLength ;
	if (pCandList != NULL && dwBufLen >= sizeof (CANDIDATELIST)) {
		pCandList->dwSize		= dwSize ;
		pCandList->dwStyle		= IME_CAND_READ ;
		pCandList->dwCount		= (DWORD) nCandidate ;
		pCandList->dwPageSize	= (nCandidate < MAX_SKKCANDPAGESIZE)? nCandidate : MAX_SKKCANDPAGESIZE ;
		pCandList->dwSelection	= 0 ;
		pCandList->dwPageStart	= 0 ;
	}
	if (pCandList != NULL && dwBufLen >= dwSize) {
		LPMYSTR		pCandidateStr ;
		DWORD*		pdwCandidateOffset ;
		DWORD*		pdwAnnotationOffset ;
		int			i ;

		pCandidateStr			= (LPMYSTR)((LPSTR)pCandList + sizeof (CANDIDATELIST) + sizeof (DWORD) * nCandidate * 2) ;
		pdwCandidateOffset		= pCandList->dwOffset ;
		pdwAnnotationOffset		= pdwCandidateOffset + nCandidate ;
		for (i = 0 ; i < nCandidate ; i ++) {
			*pdwCandidateOffset		= (DWORD)((LPSTR)pCandidateStr - (LPSTR)pCandList) ;
			*pdwAnnotationOffset	= 0 ;
			
			wstrResult	= TSearchSession_pGetReferCandidate (pHenkanSession) ;
			if (wstrResult == NULL)
				wstrResult	= TSearchSession_pGetCandidate (pHenkanSession) ;
			if (wstrResult != NULL) {
				int			nResultLen ;

				/*	Lisp Eval Kvǂ`FbN邽߂ Annotation Ɩ{̂𕪗B
				 */
				nResultLen	= lstrlenW (wstrResult) ;
				if (iAnnotationType != DISABLE_ANNOTATION) {
					LPCWSTR	pwSrc, pwNote, pwWord ;
					int		nWordLen, nNoteLen ;

					pwSrc	= wstrResult + nResultLen - 1 ;
					pwWord	= wstrResult ;
					while (pwSrc >= wstrResult && *pwSrc != L';') 
						pwSrc	-- ;
					if (pwSrc >= wstrResult && *pwSrc == L';') {
						pwNote		= pwSrc + 1 ;
						nNoteLen	= nResultLen - (pwNote - wstrResult) ;
						nWordLen	= pwSrc - wstrResult ;
					} else {
						nNoteLen	= 0 ;
						nWordLen	= nResultLen ;
					}
					/* Word  Eval Kv邩ǂ`FbNB*/
					if (nWordLen > 1 && wstrResult [0] == L'(' && wstrResult [nWordLen - 1] == L')') {
						if (TLispSession_bEval (wstrResult, nWordLen, bufTemp, ARRAYSIZE (bufTemp) - 1)) {
							bufTemp [ARRAYSIZE (bufTemp) - 1]	= L'\0' ;
							nWordLen	= lstrlenW (bufTemp) ;
							pwWord		= bufTemp ;
						}
					}
					DEBUGPRINTFEX (103, (TEXT ("candidate (%d) -> \"%s\"(%d)\n"), i, pwWord, nWordLen)) ;
					if (nWordLen > 0) 
						memcpy (pCandidateStr, pwWord, sizeof (WCHAR) * nWordLen) ;
					pCandidateStr [nWordLen]	= L'\0' ;
					pCandidateStr				+= nWordLen + 1 ;

					/* Annotation  Eval Kv邩ǂ`FbNB*/
					if (nNoteLen > 1 && pwNote [0] == L'(' && pwNote [nNoteLen - 1] == L')') {
						if (TLispSession_bEval (pwNote, nNoteLen, bufTemp, ARRAYSIZE (bufTemp) - 1)) {
							bufTemp [ARRAYSIZE (bufTemp) - 1]	= L'\0' ;
							nNoteLen	= lstrlenW (bufTemp) ;
							pwNote		= bufTemp ;
						}
					}
					if (nNoteLen > 0) {
						memcpy (pCandidateStr, pwNote, sizeof (WCHAR) * nNoteLen) ;
						pCandidateStr [nNoteLen]	= L'\0' ;
						*pdwAnnotationOffset		= (DWORD)((LPSTR)pCandidateStr - (LPSTR)pCandList) ;
						pCandidateStr				+= nNoteLen + 1 ;
					}
				} else {
					DEBUGPRINTFEX (103, (TEXT ("candidate (%d) -> \"%s\"(%d)\n"), i, wstrResult, nResultLen)) ;
					memcpy (pCandidateStr, wstrResult, sizeof (WCHAR) * nResultLen) ;
					pCandidateStr [nResultLen]	= L'\0' ;
					pCandidateStr		+= nResultLen + 1 ;
				}
				pdwCandidateOffset	++ ;
				pdwAnnotationOffset	++ ;
			}
			TSearchSession_bNextCandidate (pHenkanSession) ;
		}
	}
	TSearchSession_vDestroy (pHenkanSession) ;
	return	dwSize ;
#endif
}

/***********************************************************************
 *	ImeDestroy()
 *	ImeDestroy ֐ IME gIB
 *	BOOL
 *		ImeDestroy(
 *			UINT uReserved
 *		)
 *	Parameters
 *		uReserved
 *			\񂳂Ă邯ǁA݂ 0 łꂼB̃o[W
 *			́A0łȂ FALSE ԂƁB
 *	Return Values
 *		֐ATRUEBȂAFALSEB
 ***********************************************************************/
BOOL	WINAPI
ImeDestroy (
	UINT	uForce)
{
	DEBUGPRINTFEX (99, (MYTEXT ("ImeDestroy (uReserved:%x)\n"), uForce)) ;

	if (uForce)
		return	FALSE ;
	/*
	 * IׂȂ̂cInputContext Ɖׂ
	 * Ȃ悤ȁcBǉłȂ̂ł́H
	 * IME ŜŋLĂ鎑͑S skkiserv.exe ɓn̂ŁA
	 * 邱ƂȂH
	 */
	return	TRUE ;
}

/***********************************************************************
 *	ImeEscape
 *	ImeEscape ֐̓AvP[VɁA IMM ֐ł͒ڂɗpł
 *	Ȃ悤 IME ̓ɃANZX邱ƂB
 *	 IME  country-specific ֐ private ֐ɑ΂ĂȂ
 *	Ă͂ȂȂ̂łc炵B
 *	LRESULT
 *		ImeEscape(
 *			HIMC hIMC,
 *			UINT uEscape,
 *			LPVOID lpData
 *		)
 *	Parameters
 *		hIMC
 *			̓ReLXg̃nhB
 *		uEscape
 *			sGXP[v֐w肷B
 *		lpData
 *			w肳ꂽGXP[vɕKvȃf[^w|C^B
 *	ImeEscape ֐͎̂悤ȃGXP[v֐T|[gĂB
 *	IME_ESC_QUERY _SUPPORT
 *		Ă邩ǂ`FbNBÃGXP[v
 *		ĂȂԂB
 *	IME_ESC_RESERVED_FIRST
 *		IME_ESC_RESERVED_FIRST  IME_ESC_RESERVED_LAST ̊Ԃ̃GXP[v
 *		VXeɗ\񂳂B
 *	IME_ESC_RESERVED_LAST
 *		IME_ESC_RESERVED_FIRST  IME_ESC_RESERVED_LAST ̊Ԃ̃GXP[v
 *		VXeɗ\񂳂B
 *	IME_ESC_PRIVATE_FIRST
 *		IME_ESC_PRIVATE_FIRST  IME_ESC_PRIVATE_LAST ̊Ԃ̃GXP[v
 *		IME ɗ\񂳂BIME ͎RɎ̖ړIɉẴGXP[
 *		v֐gƂłB
 *	IME_ESC_PRIVATE_LAST
 *		IME_ESC_PRIVATE_FIRST  IME_ESC_PRIVATE_LAST ̊Ԃ̃GXP[v
 *		IME ɗ\񂳂BIME ͎RɎ̖ړIɉẴGXP[
 *		v֐gƂłB
 *	IME_ESC_SEQUENCE_TO_INTERNAL
 *		ŗL̃GXP[vBFar East vbgtH[̉œ
 *		vAvP[V͂gׂł͂ȂBEUDC
 *		GfB^̂߂̂̂łB*(LPWORD)lpData  sequence code 
 *		łAԂl͂ sequence code ɑ΂镶R[hłB
 *	IME_ESC_GETHELPFILENAME
 *		IME ̃wvt@C̖O𓾂GXP[vB֐߂
 *		(LPTSTR)lpData  IME wvt@C̃tpXɂȂĂB
 *		pX 80 * sizeof (TCHAR) ZB̊֐ Windows'98
 *		Windows 2000 ɒǉĂB
 *	IME_ESC_PRIVATE_HOTKEY
 *		lpData  hot key ID i[Ă DWORD ւ̃|C^łB
 *		(IME_HOTKEY_PRIVATE_FIRST IME_HOTKEY_PRIVATE_LAST ͈̔)
 *		VXe͈͓̔ hot key 󂯂ƂAIMM 
 *		ImeEscape ֐g IME ɃfBXpb`BWindowsR95 
 *		̓T|[gĂȂB
 *	Return Values
 *		s 0 ɂȂBȂ΃GXP[v֐ɏ]ĕԂl
 *		܂B
 *	Comments
 *		p[^ǂ͊eX̃GXP[v֐ɂB
 *		uEscape  IME_ESC_QUERY_SUPPORT ̎AlpData  IME GXP[
 *		vl܂񂾕ϐւ̃|C^łB̃Tv́AIME  
 *		IME_ESC_GETHELPFILENAME T|[gĂ邩ǂ肷
 *		̂ɗpłB
 *		DWORD	dwEsc	= IME_ESC_GETHELPFILENAME ;
 *		LRESULT lRet	= ImmEscape(hKL, hIMC, IME_ESC_QUERYSUPPORT, (LPVOID)&dwEsc) ;
 *	See Also
 *		ImmEscape
 ***********************************************************************/
LRESULT	WINAPI
ImeEscape (
	HIMC	hIMC,
	UINT	uSubFunc,
	LPVOID	lpData)
{
	LRESULT	lRet	= FALSE ;
	
	DEBUGPRINTFEX (99, (TEXT ("ImeEscape (hIMC: %ld, uSubFunc: %x, lpData: %p)"), hIMC, uSubFunc, lpData)) ;

	switch (uSubFunc){
	case IME_ESC_QUERY_SUPPORT:
		switch (*(LPUINT)lpData){
		case IME_ESC_QUERY_SUPPORT:
		case IME_ESC_PRI_GETDWORDTEST:
		case IME_ESC_GETHELPFILENAME:
			lRet	= TRUE ;
			break ;
		default:
			lRet	= FALSE ;
			break ;
		}
		break ;

	case IME_ESC_PRI_GETDWORDTEST:
		lRet	= 0x12345678 ;
		break ;

	case IME_ESC_GETHELPFILENAME:
		Mylstrcpy ((LPMYSTR)lpData, MYTEXT("SKKI1_0.HLP")) ;
		lRet	= TRUE ;
		break ;

	default:
		lRet	= FALSE ;
		break ;
	}
	return	lRet ;
}

/***********************************************************************
 *	ImeSetActiveContext ()
 *	ImeSetActiveContext ֐ current IME  Input Context  active 
 *	ȂƂm点B
 *	BOOL
 *		ImeSetActiveContext(
 *		HIMC hIMC,
 *		BOOL fFlag
 *	)
 *	Parameters
 *		hIMC
 *			Input Context Handle
 *		fFlag
 *			TRUE ȂANeBu FALSE ȂANeBuɂȂƂ
 *			B
 *		Return Values
 *			 TRUE AȂ FALSE ԂB
 *	Comments
 *		IME ͐VIꂽ Input Context ɂĂ̊֐Œm炳
 *		BIME ͏sĂ悢AvĂ킯ł͂ȂB
 *	See Also
 *		ImeSetActiveContext
 ***********************************************************************/
BOOL	WINAPI
ImeSetActiveContext (
	HIMC		hIMC,
	BOOL		fFlag)
{
	TRANSMSG		GnMsg ;
	LPINPUTCONTEXT	lpIMC ;

	DEBUGPRINTFEX (106, (TEXT ("ImeSetActiveContext (hIMC: %lx, fFlag:%d)\n"), hIMC, fFlag)) ;

	if (!hIMC) {
#if !defined (NO_TSF)
		if (fFlag) 
			vTSF_ActivateLanguageBar (hIMC, TRUE) ;
#endif
		return	TRUE ;
	}

	lpIMC		= (LPINPUTCONTEXT)ImmLockIMC (hIMC) ;
	if (lpIMC){
		if (ImmGetIMCCSize (lpIMC->hCompStr) >= sizeof (MYCOMPSTR)) {
			GnMsg.message	= WM_IME_NOTIFY ;
			GnMsg.wParam	= IMN_PRIVATE ;
			if (fFlag){
				GnMsg.lParam	= IMN_PRIVATE_ACTIVATECONTEXT ;
				GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;
				TSkkIme_vUpdateConfig (hIMC) ;
			} else {
				GnMsg.lParam	= IMN_PRIVATE_DEACTIVATECONTEXT ;
				GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;
			}

			/*
			 *	݂̂ƂADEACTIVATECONTEXT ͗pĂȂ̂ŁAȉ̃R[h̓Rg
			 *	ƂB(Wed Nov 09 21:44:01 2005)
			 *
			if (fFlag){
				GnMsg.lParam	= IMN_PRIVATE_ACTIVATECONTEXT ;
				GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;
				TSkkIme_vInvalidateConfig (hIMC, TRUE) ;
			} else {
				GnMsg.lParam	= IMN_PRIVATE_DEACTIVATECONTEXT ;
				GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;
			}
			*/
		}
		ImmUnlockIMC (hIMC) ;
	}
	return	TRUE ;
}

/***********************************************************************
 *	ImeProcessKey ()
 *	ImeProcessKey ֐ IMM ʂė^ꂽSẴL[Xg[NO
 *	āÃL[^ꂽ Input Context  IME ɕKvȂ̂
 *	 TRUE ԂB
 *	BOOL
 *		ImeProcessKey(
 *		HIMC hIMC,
 *		UINT uVirKey,
 *		DWORD lParam,
 *		CONST LPBYTE lpbKeyState
 *	)
 *	Parameters
 *		hIMC
 *			Input context handle
 *		uVirKey
 *			ׂzL[B
 *		lParam
 *			L[bZ[W lParamB(WM_KEYDOWN,WM_KEYUP  LPARAM)
 *		lpbKeyState
 *			݂̃L[{[h̏Ԃ܂256oCg̔zw|C^B
 *			IME ͂̓eύXׂł͂ȂB
 *	Return Values
 *		Ȃ TRUEBłȂ FALSEB
 *	Comments
 *		VXe̓L[ IME ɂĎ舵ׂۂ̊֐
 *		ĂяoƂɂČ肵ĂBAvP[VL[bZ[
 *		W󂯎Oɂ̊֐ TRUE Ԃ΁AIME ͂̃L[
 *		BVXe ImeToAsciiEx ֐ĂяoB
 *		FALSE ԂȂ΁AVXe͂̃L[ IME ɂď
 *		ȂƂ̂ŃL[bZ[W̓AvP[VɑB
 ***********************************************************************/
BOOL	WINAPI
ImeProcessKey (
	HIMC			hIMC,
	UINT			vKey,
	LPARAM			lKeyData,
	CONST LPBYTE	lpbKeyState)
{
	LPINPUTCONTEXT	lpIMC ;
	BOOL			fOpen ;
	BOOL 					fEaten		= FALSE ;
	LPCOMPOSITIONSTRING		lpCompStr ;
	LPMYCOMPSTR				lpMyCompStr ;
	
	if (lKeyData & 0x80000000)
		return	FALSE ;

	lpIMC	= ImmLockIMC (hIMC) ;
	if (!lpIMC)
		return	FALSE ;

	fOpen		 = lpIMC->fOpen ;
	lpCompStr	= (LPCOMPOSITIONSTRING)ImmLockIMCC (lpIMC->hCompStr) ;
	if (lpCompStr != NULL) {
		struct TMSG	msg ;
		BOOL		fToggleIME ;

		lpMyCompStr	= (LPMYCOMPSTR) lpCompStr ;
		msg.m_nMessage	= WM_KEYDOWN ;
		msg.m_wParam	= vKey ;		/* wParam, lParam ̐ݒ͂ VK_KEYDOWN ɑ̂H ȃbZ[Wɂ̂H */
		msg.m_lParam	= lKeyData ;
		msg.m_rParam	= 0 ;
		if (lpbKeyState != NULL) {
			msg.m_rParam	|= (lpbKeyState [VK_LCONTROL]	& 0x80)? KEYMASK_LCONTROL	: 0 ;
			msg.m_rParam	|= (lpbKeyState [VK_RCONTROL]	& 0x80)? KEYMASK_RCONTROL	: 0 ;
			msg.m_rParam	|= (lpbKeyState [VK_LMENU]		& 0x80)? KEYMASK_LMENU		: 0 ;
			msg.m_rParam	|= (lpbKeyState [VK_RMENU]		& 0x80)? KEYMASK_RMENU		: 0 ;
			msg.m_rParam	|= (lpbKeyState [VK_LSHIFT]		& 0x80)? KEYMASK_LSHIFT		: 0 ;
			msg.m_rParam	|= (lpbKeyState [VK_RSHIFT]		& 0x80)? KEYMASK_RSHIFT		: 0 ;
			msg.m_rParam	|= (lpbKeyState [VK_SCROLL]		& 0x80)? KEYMASK_SCROLL		: 0 ;
			msg.m_rParam	|= (lpbKeyState [VK_NUMLOCK]	& 0x80)? KEYMASK_NUMLOCK	: 0 ;
		} else {
			msg.m_rParam	|= (GetKeyState (VK_LCONTROL))?	KEYMASK_LCONTROL	: 0 ;
			msg.m_rParam	|= (GetKeyState (VK_RCONTROL))?	KEYMASK_RCONTROL	: 0 ;
			msg.m_rParam	|= (GetKeyState (VK_LMENU))?	KEYMASK_LMENU		: 0 ;
			msg.m_rParam	|= (GetKeyState (VK_RMENU))?	KEYMASK_RMENU		: 0 ;
			msg.m_rParam	|= (GetKeyState (VK_LSHIFT))?	KEYMASK_LSHIFT		: 0 ;
			msg.m_rParam	|= (GetKeyState (VK_RSHIFT))?	KEYMASK_RSHIFT		: 0 ;
			msg.m_rParam	|= (GetKeyState (VK_SCROLL))?	KEYMASK_SCROLL		: 0 ;
			msg.m_rParam	|= (GetKeyState (VK_NUMLOCK))?	KEYMASK_NUMLOCK		: 0 ;
		}
		msg.m_nTime		= GetTickCount () ;
		msg.m_pt.x		= msg.m_pt.y	= 0 ;
		if (! ImeDoc_bQueryToggleIMEEvent (&lpMyCompStr->_Doc, &msg, &fEaten))
			fEaten	= FALSE ;
		if (fOpen && ! fEaten)
			ImeDoc_bQueryFilterEvent (&lpMyCompStr->_Doc, &msg, &fEaten) ;
		ImmUnlockIMCC (lpIMC->hCompStr) ;
	}
	ImmUnlockIMC (hIMC) ;
	return	fEaten ;
}

/***********************************************************************
 *	ImeSelect
 *	ImeSelect ֐ IME private context ̏yє񏉊s
 *	̂ɗpB
 *	BOOL
 *		ImeSelect(
 *			HIMC hIMC,
 *			BOOL fSelect
 *		)
 *	Parameters
 *		hIMC
 *			Input context handle
 *		fSelect
 *			TRUE ȂΏAFALSE Ȃ񏉊(\[X̉)
 *			B
 *	Return Values
 *		 TRUEBȂ FALSEB
 ***********************************************************************/
BOOL	WINAPI
ImeSelect (
	HIMC		hIMC,
	BOOL		fSelect)
{
	LPINPUTCONTEXT		lpIMC ;
	LPCOMPOSITIONSTRING	lpCompStr ;
	TRANSMSG			GnMsg ;
	LOGFONT				lf ;

	/* it's NULL context. */
	if (!hIMC)
		return TRUE ;

	lpIMC	= ImmLockIMC(hIMC) ;
	if (lpIMC){
		DEBUGPRINTFEX (106, (TEXT ("ImeSelect (hIMC: %lx, fSelect:%d)\n"), hIMC, fSelect)) ;
		if (fSelect){
			/* Initialize */
			LPCANDIDATEINFO	lpCandInfo ;

			/*	MS-IME  SKKIME  Toolbar Ő؂ւƁA
			 *	ImeSelect ɂ͗邪AActiveContext ɂ͗ȂƂ
			 *	ۂBConfig ̓ǂݍ݂̃^C~O	ActiveContext
			 *	ɍiȂcB
			 *
			 *	ŃWXgɐG邱Ƃ UserEnv ̃G[ɂȂ̂낤
			 *	H
			 */
			ImeConfig_vLoad () ;

			/* Init the general member of IMC. */
			if (!(lpIMC->fdwInit & INIT_LOGFONT)){
#if defined (UNICODE)
				/* Ő̂H Ƃ .A.  */
				lpIMC->lfFont.W.lfCharSet	= SHIFTJIS_CHARSET ;
#else
				lpIMC->lfFont.A.lfCharSet	= SHIFTJIS_CHARSET ;
#endif
				lpIMC->fdwInit				|= INIT_LOGFONT ;
			}
			if (!(lpIMC->fdwInit & INIT_CONVERSION)){
				lpIMC->fdwConversion	= IME_CMODE_ROMAN | IME_CMODE_FULLSHAPE | IME_CMODE_NATIVE ;
				lpIMC->fdwInit			|= INIT_CONVERSION ;
			}
			if (!(lpIMC->fdwInit & INIT_COMPFORM)) {
				lpIMC->cfCompForm.dwStyle			= 1 ; // H
				lpIMC->cfCompForm.ptCurrentPos.x	= 0 ;
				lpIMC->cfCompForm.ptCurrentPos.y	= 0 ;
				lpIMC->fdwInit				|= INIT_COMPFORM ;
			}

			lpIMC->hCompStr = ImmReSizeIMCC (lpIMC->hCompStr, sizeof (MYCOMPSTR)) ;
			lpCompStr		= (LPCOMPOSITIONSTRING)ImmLockIMCC (lpIMC->hCompStr) ;
			if (lpCompStr){
				LPMYCOMPSTR	lpMyCompStr	= (LPMYCOMPSTR) lpCompStr ;

				lpCompStr->dwSize	= sizeof (MYCOMPSTR) ;
				TSkkIme_vInitCompStr (lpCompStr, CLR_RESULT_AND_UNDET) ;
				ImeDoc_bInit (&lpMyCompStr->_Doc) ;
				lpMyCompStr->_bCandMode				= FALSE ;
				lpMyCompStr->_fComposing			= FALSE ;
				lpMyCompStr->cfCandBase.hWnd		= (HWND) 0 ;
				lpMyCompStr->cfCandBase.dwIndex		= (DWORD) -1 ;
				lpMyCompStr->cfCandBase.dwStyle		= (DWORD) 0 ;
				ImmUnlockIMCC (lpIMC->hCompStr) ;
			}
			TSkkIme_bSetConversionMode (hIMC, lpIMC->fdwConversion) ;

			lpIMC->hCandInfo	= ImmReSizeIMCC (lpIMC->hCandInfo, sizeof (MYCAND)) ;
			lpCandInfo			= (LPCANDIDATEINFO)ImmLockIMCC (lpIMC->hCandInfo) ;
			if (lpCandInfo){
				TSkkIme_vClearCandidate (lpCandInfo) ;
				ImmUnlockIMCC (lpIMC->hCandInfo) ;
			}

			/*	ImeSelect (TRUE)  Text Service Framework 𗘗p application ł
			 *	Ă΂Ă܂B̂߁ALangBarItemMgr  LangBarItem o^ׂ
			 *	ꏊƂēK؂ł͂ȂB
			 *	ImeActiveContext (TRUE) œo^K؂ł悤ɌB
			 *
			 *	---- 폜ꂽR[h (Sun Aug 14 02:21:27 2005)
#if !defined (NO_TSF)	
			vTSF_ActivateLanguageBar (hIMC, TRUE) ;//commented out
#endif
			 */
			MakeInfoGuideLine (hIMC, NULL, 0, -1) ;
		} else {
			/* Uninitialize */
			/*  DEBUGPRINTF ((TEXT ("ImeSelect: (%ld) -> MakeInfoGuideLine (0)\n"), hIMC)) ;
			 *	MakeInfoGuideLine (hIMC, NULL, 0, -1) ;	
			 *	 MakeInfoGuideLine ͕svB
			 *	 GenerateMessage s application I
			 *	hangup ̂BӁII
			 */
			if (lpIMC->hCompStr){
				lpCompStr	= (LPCOMPOSITIONSTRING)ImmLockIMCC (lpIMC->hCompStr) ;
				if (lpCompStr && lpCompStr->dwSize == sizeof(MYCOMPSTR)){
					ImeDoc_vClear (&((LPMYCOMPSTR)lpCompStr)->_Doc) ;
					ImmUnlockIMCC (lpIMC->hCompStr) ;
				}
			}
			ImeConfig_vUnload () ;

			/*	LangBarItemMgr  LangBarItem 폜̂ɓKꏊ͂炾Ǝv
			 *	BImeActiveContext (FALSE) ̒iKł́A{ׂłȂꍇ(
			 *	Application  focus ړ)ɂ LangBarItem 폜Ă܂
			 *	ƂɂȂB
			 */
			/*	CloseStatusWindow ̈ʒuɒǉAɂcƂɂB
			 *		(Fri Dec 09 14:28:56 2005)
			 *
			 *	̑͗ǂȂ悤BI TSF Manager 폜ĂH
			 *	Ŗɍ폜悤ƂƁAɈُɎԂĂ܂Ƃ
			 *	B
			 *		(Thu Apr 26 21:45:29 2007)
			 *	L͂܂R[hB
#if !defined (NO_TSF)
				vTSF_ActivateLanguageBar (hIMC, FALSE) ;//commented out
#endif
			 *
			 *	ǂĂĂȂBThreadMgr݂鎞ɂ͏Ƃ̂H
			 *	ł܂̂H
			 *	CĂ݂R[hB
			 *
#if !defined (NO_TSF)
			if (bTSF_DoesThreadMgrExistp ()) {
				vTSF_ActivateLanguageBar (hIMC, FALSE) ;//commented out
			}
#endif
			 */
#if !defined (NO_TSF)
			if (bTSF_DoesThreadMgrExistp ()) {
				vTSF_ActivateLanguageBar (hIMC, FALSE) ;
			}
#endif
		}
		ImmUnlockIMC (hIMC) ;
	}
	return	TRUE ;
}

/***********************************************************************
 *	ImeSetCompositionString ()
 *	ImeSetCompositionString ֐̓AvP[V lpComp lpRead p
 *	[^ɓꂽf[^IMEcomposition string \̂ɐݒ肷
 *	̂ɗpBIME  WM_IME_COMPOSITION bZ[W𐶐B
 *	BOOL WINAPI
 *		ImeSetCompositionString(
 *		HIMC hIMC,
 *		DWORD dwIndex,
 *		LPCVOID lpComp,
 *		DWORD dwCompLen,
 *		LPCVOID lpRead,
 *		DWORD dwReadLen
 *	);
 *	Parameters
 *		hIMC
 *			Input context handle. 
 *		dwIndex
 *			dwIndex ɗ^l͎̒ʂłB
 *			------------------------------------------------------------
 *	SCS_SETSTR
 *		AvP[V composition string  reading string ܂
 *		̗͂ZbgBȂƂ lpComp  lpRead p[^
 *		1͐ȕw|C^łȂ΂ȂȂB
 *		񂪒̂ȂAIME ͂؂l߁B
 *	SCS_CHANGEATTR
 *		AvP[V composition string  reading string 
 *		̗҂̐ߏݒ肷BȂƂ lpComp  lpRead 1
 *		 valid Ȑߏ̔zw|C^łȂ΂ȂȂB
 *	SCS_QUERYRECONVERTSTRING
 *		AvP[V IME ɂ RECONVERTSTRINGSTRUCTURE 𒲐
 *		Ă悤˗BAvP[V̒lݒ肵
 *		ImeSetCompositionString ĂяoȂAIME  RECONVERTSTRING
 *		\̂𒲐Ȃ΂ȂȂBAvP[V͂̎
 *		ꂽ RECONVERTSTRING \̂ SCS_RECONVERTSTRING łāA
 *		̊֐ɓnƂłBIME  WM_IMECOMPOSITION bZ[W
 *		𐶐Ă͂ȂȂB
 *	SCS_SETRECONVERTSTRING
 *		AvP[V IME  RECONVERTSTRING \̂Ɋ܂܂
 *		 reconvert 悤Ɉ˗B
 *			------------------------------------------------------------
 *		lpComp
 *			XVꂽ܂ރobt@w|C^B̌^
 *			dwIndex ̒lɂČ肳B
 *		dwCompLen
 *			obt@̃oCgB
 *		lpRead
 *			XVꂽ܂ރobt@w|C^B̌^
 *			dwIndex ̒lɂČ肳BdwIndex ̒l
 *			SCS_SETRECONVERTSTRING  SCS_QUERYRECONVERTSTRING 
 *			΁AlpRead ͍XVꂽ reading string ܂
 *			RECONVERTSTRING \̂w|C^łBIꂽ
 *			IME  SCS_CAP_MAKEREAD Ă΁ANULL ɂȂB
 *		dwReadLen
 *			obt@̃oCgB
 *	Comments
 *		Unicode ɂẮAƂ SCS_SETSTR 肳Ă Unicode 
 *		܂łĂAdwCompLen  dwReadLen ̓obt@̃oCg
 *		łB
 *		SCS_SETRECONVERTSTRING ܂ SCS_QUERYRECONVERTSTRING 
 *		SCS_CAP_CONVERTSTRING vpeB IME ɂ̂ݗpłB
 *		̃vpeB ImmGetProperty ֐gƂœ邱ƂłB
 ***********************************************************************/
BOOL	WINAPI
ImeSetCompositionString(
	HIMC			hIMC,
	DWORD			dwIndex,
	LPVOID			lpComp,
	DWORD			dwComp,
	LPVOID			lpRead,
	DWORD			dwRead)
{
	LPINPUTCONTEXT			lpIMC ;
	LPCOMPOSITIONSTRING	lpCompStr ;

	switch (dwIndex) {
	case	SCS_SETSTR:
	{
		if (lpComp == NULL || dwComp <= 0)
			break ;

		lpIMC	= ImmLockIMC (hIMC) ;
		if (!lpIMC)
			break ;

		lpCompStr	= (LPCOMPOSITIONSTRING)ImmLockIMCC (lpIMC->hCompStr) ;
		if (lpCompStr != NULL) {
			TSkkIme_vInitCompStr (lpCompStr, CLR_RESULT_AND_UNDET) ;
			ImeDoc_bSetConversionString (&((LPMYCOMPSTR)lpCompStr)->_Doc, lpComp, dwComp) ;
			TSkkIme_bUpdateComposition (hIMC, lpIMC, lpCompStr) ;
			ImmUnlockIMCC (lpIMC->hCompStr) ;
		}
		ImmUnlockIMC (hIMC) ;
		return	TRUE ;
	}
	case	SCS_CHANGEATTR:
	case	SCS_CHANGECLAUSE:
		break ;
	case	SCS_QUERYRECONVERTSTRING:
	{
		LPRECONVERTSTRING		pRStr		= (LPRECONVERTSTRING)lpComp ;

#if defined (DEBUG) || defined (DBG)
		DebugPrintf (TEXT ("SCS_QUERYRECONVERTSTRING:\n")) ;
		if (lpComp != NULL) {
			DebugPrintf (TEXT ("lpComp:\n")) ;
			DebugDumpRS (lpComp) ;
		}
		if (lpRead != NULL) {
			DebugPrintf (TEXT ("lpRead:\n")) ;
			DebugDumpRS (lpRead) ;
		}
#endif
		if (pRStr != NULL) {
			/*	TCY܂̏ꍇɂ́A芸 dwStrLen 
			 *	킹Bkakasi Ƃ̈ʒuŘAĝ
			 *	ȂB*/
			if (pRStr->dwCompStrLen   <= 0 ||
				pRStr->dwTargetStrLen <= 0) {
				pRStr->dwCompStrOffset		= 0 ;
				pRStr->dwTargetStrOffset	= 0 ;
				pRStr->dwCompStrLen			= pRStr->dwStrLen ;
				pRStr->dwTargetStrLen		= pRStr->dwStrLen ;
			}
			/*	TCY傫߂Ȃ璲Ȃ΁B*/
			if (pRStr->dwCompStrLen >= MAXCOMPSIZE) 
				pRStr->dwCompStrLen			= MAXCOMPSIZE - 1 ;
			if (pRStr->dwTargetStrLen >= MAXCOMPSIZE) 
				pRStr->dwTargetStrLen		= MAXCOMPSIZE - 1 ;
		}
		return	TRUE ;
	}
	case	SCS_SETRECONVERTSTRING:
	{
		LPRECONVERTSTRING		pRStr		= (LPRECONVERTSTRING)lpComp ;

#if defined (DEBUG) || defined (DBG)
		DebugPrintf (TEXT ("SCS_SETRECONVERTSTRING:\n")) ;
		if (lpComp != NULL) {
			DebugPrintf (TEXT ("lpComp:\n")) ;
			DebugDumpRS (lpComp) ;
		}
		if (lpRead != NULL) {
			DebugPrintf (TEXT ("lpRead:\n")) ;
			DebugDumpRS (lpRead) ;
		}
#endif
		if (!pRStr)
			break ;

		lpIMC	= ImmLockIMC (hIMC) ;
		if (!lpIMC) {
			break ;
		}
		if (ImmGetIMCCSize (lpIMC->hCompStr) < sizeof (MYCOMPSTR)){
			ImmUnlockIMC (hIMC) ;
			break ;
		}
		/*	ϊ[hIɉ̓[hɐݒ肷B*/
		//SKKChangeConversionMode (hIMC, IDM_CMODE_TO_ROMANHIRA) ;
		lpCompStr	= (LPCOMPOSITIONSTRING)ImmLockIMCC (lpIMC->hCompStr) ;
		if (lpCompStr != NULL) {
			if (! ImmGetOpenStatus (hIMC))
				ImmSetOpenStatus (hIMC, TRUE) ;
			TSkkIme_bSetReconvertStr (hIMC, lpIMC, lpCompStr, pRStr, TRUE) ;
			ImmUnlockIMCC (lpIMC->hCompStr) ;
		}
		ImmUnlockIMC (hIMC) ;
		return	TRUE ;
	}
	default:
		break ;
	}
	return	FALSE ;
}

/***********************************************************************
 *	ImeGetImeMenuItems
 *	ImeGetImeMenuItems ֐ IME menu ɓo^Ă menu item 𓾂B
 *	DWORD WINAPI
 *		ImeGetImeMenuItems(
 *		HIMC hIMC,
 *		DWORD dwFlags,
 *		DWORD dwType,
 *		LPIMEMENUITEMINFO lpImeParentMenu,
 *		LPIMEMENUITEMINFO lpImeMenu,
 *		DWORD dwSize
 *	)
 *	Parameters
 *		hIMC
 *			lpMenuItem ͂ input context Ɋ֘AtĂ menu
 *			item ܂ށB
 *		dwFlags
 *			rbg̑gݍ킹ȂB
 *			------------------------------------------------------------
 *			IGIMIF_RIGHTMENU	
 *				 1 ȂÅ֐͉ENbN Context menu ɑ΂
 *				menu item ԂB
 *			------------------------------------------------------------
 *		dwType
 *			rbg̑gݍ킹ȂB
 *			------------------------------------------------------------
 *			IGIMII_CMODE
 *				conversion mode Ɋ֘AtĂ menu item ԂB
 *			IGIMII_SMODE
 *				sentence mode Ɋ֘AtĂ menu item ԂB
 *			IGIMII_CONFIGURE
 *				IME ̐ݒɊ֘AtĂ menu item ԂB
 *			IGIMII_TOOLS
 *				IME c[Ɋ֘AtĂ menu item ԂB
 *			IGIMII_HELP
 *				IME wvɊ֘AtĂ menu item ԂB
 *			IGIMII_OTHER
 *				̑ menu item ԂB
 *			IGIMII_INPUTTOOLS
 *				gꂽ͕̓@񋟂 IME ̓c[Ɋ֘A
 *				tꂽ menu item ԂB
 *			------------------------------------------------------------
 *		lpImeParentMenu
 *			fType  MFT_SUBMENU  IMEMENUINFO \̂w|C
 *			^BImeGetMenuItems ͂ menu item  submenu item ԂB
 *			 NULL Ȃ lpImeMenu  top-level  IME menu item 
 *			łB
 *		lpImeMenu
 *			menu item ̓e󂯎obt@w|C^B̃ob
 *			t@ IMEMENUINFO \̂̔złB NULL Ȃ
 *			ImeGetImeMenuItems ͓o^ menu item ̐ԂB
 *		dwSize
 *			IMEMENUITEMINFO \̂󂯂obt@̃TCYB
 *	Return Values
 *		lpIM ɐݒ肳ꂽ menu item ̐AB lpImeMenu  NULL 
 *		Ȃ ImeMenuItems ͎w肳ꂽ hKL ɓo^Ă menu item 
 *		̐ԂB
 *	ImeGetImeMenuItems  Windows'98  Windows 2000 p̐V֐łB
 ***********************************************************************/
DWORD	WINAPI
ImeGetImeMenuItems (
	HIMC		hIMC,
	DWORD		dwFlags,
	DWORD		dwType,
	LPMYIMEMENUITEMINFO	lpImeParentMenu,
	LPMYIMEMENUITEMINFO	lpImeMenu,
	DWORD					dwSize)
{
	/*	dwType  MSIME ̓`FbNĂȂ悤Bɍ킹B
	 *	ATSF ĂƁA method ͏ɓ삵ĂȂ悤Ɏv邪B
	 *
	 *	if (IsTSFEnabled ())
	 *		return	0 ;
	 *
	 *	̃R[hLɂ邩ǂ͔B
	 *	[Sun Nov 23 00:26:08 2003]
	 */
	dwType	|= IGIMII_CMODE | IGIMII_SMODE | IGIMII_CONFIGURE | IGIMII_TOOLS | IGIMII_HELP | IGIMII_OTHER ;
	if (dwFlags & IGIMIF_RIGHTMENU){
		return	getImeRightMenuItems (hIMC, dwFlags, dwType, lpImeParentMenu, lpImeMenu, dwSize) ;
	} else {
		return	getImeLeftMenuItems (hIMC, dwFlags, dwType, lpImeParentMenu, lpImeMenu, dwSize) ;
	}
}

DWORD	PASCAL
getImeLeftMenuItems (
	HIMC		hIMC,
	DWORD		dwFlags,
	DWORD		dwType,
	LPMYIMEMENUITEMINFO	lpImeParentMenu,
	LPMYIMEMENUITEMINFO	lpImeMenu,
	DWORD		dwSize)
{
	static MYSTRTYPEPAIR	myLeftMenuItemInfoTbl []	= {
		{ IDIM_SHOW_TOOLBAR,	MYTEXT ("c[o[\(&I)"), },
		{ IDIM_CANCEL,			MYTEXT ("LZ"), },
	} ;
	long					lUsage ;
	int						iMenuItem ;

	if (! TSkkIme_bIsHideToolbarp ())
		return	0 ;

	iMenuItem		= 0 ;
	/*	c[o[\(&I) j[ɒǉB*/
	if (dwType & IGIMII_OTHER) {
		if (lpImeMenu != NULL) {
			insertStringMenuItem (lpImeMenu ++, &myLeftMenuItemInfoTbl [0], IMFS_ENABLED) ;
			insertSeparatorMenuItem (lpImeMenu ++) ;
		}
		iMenuItem		+= 2 ;
	}
	if (dwType & IGIMII_CMODE) {
		/*	̓[h̕ύXj[ǉB*/
		lUsage			= insertInputModeMenuItems (hIMC, lpImeMenu) ;
		if (lUsage > 0) {
			if (lpImeMenu != NULL) {
				lpImeMenu		+= lUsage ;
				insertSeparatorMenuItem (lpImeMenu ++) ;
			}
			iMenuItem		+= lUsage + 1 ;
		}
	}
	/*	LZǉB*/
	if (lpImeMenu != NULL)
		insertStringMenuItem (lpImeMenu ++, &myLeftMenuItemInfoTbl [1], IMFS_ENABLED) ;
	iMenuItem		++ ;

	DEBUGPRINTFEX (102, (TEXT ("getImeLeftMenuItems return (%d), lUsage = %d\n"),
						 iMenuItem, (int) lUsage)) ;
	return	iMenuItem ;
}

DWORD	PASCAL
getImeRightMenuItems (
	HIMC			hIMC,
	DWORD			dwFlags,
	DWORD			dwType, 
	LPMYIMEMENUITEMINFO	lpImeParentMenu, 
	LPMYIMEMENUITEMINFO	lpImeMenu,
	DWORD			dwSize)
{
	static MYSTRTYPEPAIREX	myRightMenuItemInfoTbl [NUM_ROOT_MENU_R]	= {
		{ IDIM_SHOW_TOOLBAR,	IGIMII_OTHER,		MYTEXT ("c[o[\(&I)"), },
		{ 0,					IGIMII_OTHER,		0, },
		{ IDIM_PROPERTY,		IGIMII_CONFIGURE,	MYTEXT ("vpeB(&R)"), },
		{ 0,					IGIMII_CONFIGURE | IGIMII_OTHER,	0, },
		{ IDIM_INPUTMODE,		IGIMII_CMODE,		MYTEXT ("̓[h(&N)"), },
		{ 0,					IGIMII_CMODE,		0, },
		{ IDIM_CANCEL,			(DWORD)-1,			MYTEXT ("LZ"), },
	} ;
	LPMYSTRTYPEPAIREX	ptr ;
	int				iMenuItem, nCount ;
	BOOL				fInsert ;

	if (lpImeParentMenu)
		return	getImeRightSubMenuItems (hIMC, dwFlags, dwType, lpImeParentMenu, lpImeMenu, dwSize) ;
	if (! TSkkIme_bIsHideToolbarp ())
		return	0 ;

	iMenuItem		= 0 ;
	ptr				= myRightMenuItemInfoTbl ;
	for (nCount = 0 ; nCount < ARRAYSIZE (myRightMenuItemInfoTbl) ; nCount ++) {
		fInsert	= (ptr->m_dwType != (DWORD) -1)? ((ptr->m_dwType & dwType) != 0) : TRUE ;
		if (fInsert) {
			if (lpImeMenu != NULL) {
				if (! ptr->m_lpString){
					insertSeparatorMenuItem (lpImeMenu) ;
				} else {
					/*	j[ǉB*/
					insertStringMenuItemEx (lpImeMenu, ptr, IMFS_ENABLED) ;
					
					/*	̓[h̏ꍇɂ submenu ݂B*/
					if (ptr->m_wID == IDIM_INPUTMODE) 
						lpImeMenu->fType	= IMFT_SUBMENU ;
				}
				lpImeMenu	++ ;
			}
			iMenuItem	++ ;
		}
		ptr			++ ;
	}
	return	iMenuItem ;
}

DWORD	PASCAL
getImeRightSubMenuItems (
	HIMC		hIMC,
	DWORD		dwFlags,
	DWORD		dwType,
	LPMYIMEMENUITEMINFO	lpImeParentMenu,
	LPMYIMEMENUITEMINFO	lpImeMenu,
	DWORD		dwSize)
{
	if ((lpImeParentMenu->wID != IDIM_INPUTMODE) ||
		(dwType & IGIMII_CMODE) == 0)
		return	0 ;
	if (!lpImeMenu)
		return	NUM_INPUTMODE_MENU ;
	return	insertInputModeMenuItems (hIMC, lpImeMenu) ;
}

long	PASCAL
insertStringMenuItem (
	LPMYIMEMENUITEMINFO		lpImeMenu,
	const LPMYSTRTYPEPAIR		lpMyStrTypePair,
	UINT						uValue)
{
	if (!lpImeMenu || !lpMyStrTypePair)
		return	0L ;

	lpImeMenu->cbSize			= sizeof(IMEMENUITEMINFO) ;
	lpImeMenu->fType			= 0 ;
	lpImeMenu->fState			= uValue ;
	lpImeMenu->wID				= lpMyStrTypePair->m_wID ;
	Mylstrcpy (lpImeMenu->szString, lpMyStrTypePair->m_lpString) ;
	lpImeMenu->hbmpChecked		= 0 ;
	lpImeMenu->hbmpUnchecked	= 0 ;
	lpImeMenu->hbmpItem			= 0 ;
	return	1L ;
}

long	PASCAL
insertStringMenuItemEx (
	LPMYIMEMENUITEMINFO		lpImeMenu,
	const LPMYSTRTYPEPAIREX	lpMyStrTypePair,
	UINT						uValue)
{
	if (!lpImeMenu || !lpMyStrTypePair)
		return	0L ;

	lpImeMenu->cbSize			= sizeof(IMEMENUITEMINFO) ;
	lpImeMenu->fType			= 0 ;
	lpImeMenu->fState			= uValue ;
	lpImeMenu->wID				= lpMyStrTypePair->m_wID ;
	Mylstrcpy (lpImeMenu->szString, lpMyStrTypePair->m_lpString) ;
	lpImeMenu->hbmpChecked		= 0 ;
	lpImeMenu->hbmpUnchecked	= 0 ;
	lpImeMenu->hbmpItem			= 0 ;
	return	1L ;
}

long	PASCAL
insertSeparatorMenuItem (
	LPMYIMEMENUITEMINFO	lpImeMenu)
{
	if (!lpImeMenu)
		return	0L ;
	lpImeMenu->cbSize			= sizeof(IMEMENUITEMINFO) ;
	lpImeMenu->fType			= IMFT_SEPARATOR ;
	lpImeMenu->fState			= 0 ;
	lpImeMenu->wID				= 0 ;
	lpImeMenu->szString [0]		= MYTEXT ('\0') ;
	lpImeMenu->hbmpChecked		= 0 ;
	lpImeMenu->hbmpUnchecked	= 0 ;
	lpImeMenu->hbmpItem			= 0 ;
	return	1L ;
}

long	PASCAL
insertInputModeMenuItems (
	HIMC		hIMC,
	LPMYIMEMENUITEMINFO	lpImeMenu)
{
	static MYSTRTYPEPAIR	inputModeMenuItemTbl [NUM_INPUTMODE_MENU]	= {
		{ IDM_CMODE_TO_ASCII,			MYTEXT ("AXL[(&S)"), },
		{ IDM_CMODE_TO_ROMANHIRA,		MYTEXT ("[}(&H)"), },
		{ IDM_CMODE_TO_ROMANKATA,		MYTEXT ("[}Љ(&K)"), },
		{ IDM_CMODE_TO_JISX0201KANA,	MYTEXT ("pЉ(&A)"), },
		{ IDM_CMODE_TO_JISX0201ROMAN,	MYTEXT ("pp(&R)"), },
		{ IDM_CMODE_TO_ZENEI,			MYTEXT ("Spp(&E)"), },
		{ IDM_CMODE_TO_DIRECTINPUT,	MYTEXT ("ړ(&C)"), },
	} ;
	DWORD					dwMyInputModeTbl [NUM_INPUTMODE_MENU]	= {
		IDM_CMODE_TO_ROMANHIRA,
		IDM_CMODE_TO_ROMANKATA,
		IDM_CMODE_TO_ZENEI,
		IDM_CMODE_TO_JISX0201KANA,
		IDM_CMODE_TO_JISX0201ROMAN,
		IDM_CMODE_TO_ASCII,
		IDM_CMODE_TO_DIRECTINPUT,
	} ;
	LPMYSTRTYPEPAIR			lpMyMenuItemInfo ;
	long					lCounter ;
	DWORD					dwCMode ;
	LPINPUTCONTEXT			lpIMC ;

	/* ݂̓̓[h𓾂B*/
	lpIMC	= (LPINPUTCONTEXT)ImmLockIMC (hIMC) ;
	if (!lpIMC)
		return	0 ;
	
	/*	IMECMODE_XXX  IDM_CMODE_ ɕϊB
	 */
	dwCMode		= dwMyInputModeTbl [iGetConversionModeFromLPIMC (lpIMC)] ;
	ImmUnlockIMC (hIMC) ;

	if (lpImeMenu == NULL) 
		return	sizeof (inputModeMenuItemTbl) / sizeof (MYSTRTYPEPAIR) ;

	lpMyMenuItemInfo	= inputModeMenuItemTbl ;
	lCounter			= 0 ;
	while (lCounter < sizeof (inputModeMenuItemTbl) / sizeof (MYSTRTYPEPAIR)){
		lpImeMenu->cbSize			= sizeof(IMEMENUITEMINFO) ;
		lpImeMenu->fType			= IMFT_RADIOCHECK ;
		lpImeMenu->wID				= lpMyMenuItemInfo->m_wID ;
		if (lpMyMenuItemInfo->m_wID == dwCMode){
			lpImeMenu->fState		= IMFS_ENABLED | IMFS_CHECKED ;
		} else {
			lpImeMenu->fState		= IMFS_ENABLED | IMFS_UNCHECKED ;
		}
		Mylstrcpy (lpImeMenu->szString, lpMyMenuItemInfo->m_lpString) ;
		lpImeMenu->hbmpChecked		= 0 ;
		lpImeMenu->hbmpUnchecked	= 0 ;
		lpImeMenu->hbmpItem			= 0 ;
		lpMyMenuItemInfo	++ ;
		lpImeMenu	++ ;
		lCounter	++ ;
	}
	return	lCounter ;
}

