/* # SKKIME1.0 (Simple Kana-Kanji Input Method Editor for Windows2000)
 * Skkui.c
 * This file is part of skkime1.0.
 * Copyright (C) 1999
 * Takashi SAKAMOTO (tatari_sakamoto@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "skki1_5.h"
#include "skkui.h"
#include "ImeBuffer.h"
#include "resource.h"
#include "jstring.h"
#include "ImeConfig.h"

static	void	TSkkIme_vGenerateMsgOpenCandidate	(HIMC, LPINPUTCONTEXT, LPCOMPOSITIONSTRING) ;
static	void	TSkkIme_vGenerateMsgUpdateCandidate	(HIMC, LPINPUTCONTEXT) ;
static	void	TSkkIme_vGenerateMsgCloseCandidate	(HIMC, LPINPUTCONTEXT, BOOL) ;

/*************************************************************************
 *	TSkkIme_bHandleKeyDown
 *		ToAsciiEx Ăяo֐BSKKIME ɑ΂L[͂B
 *	BOOL	PASCAL
 *		TSkkIme_bHandleKeyDown (
 *		HIMC	hIMC,
 *		WPARAM	wParam,
 *		LPARAM	lParam,
 *		LPBYTE	lpbKeyState
 *	)
 *	[p[^]
 *		hIMC
 *			Input Context Handle
 *		wParam
 *			 16 rbgɉꂽ̕R[hA
 *			 16 rbgɉꂽL[̉zL[R[hĂ܂B
 *		lParam
 *			WM_KEYDOWN ɗ^ lParam ƓłB
 *		lpbKeyState
 *			݂̃L[{[h̏Ԃ܂256oCg̔zւ̃|C^łB
 *			IME ͂̓eύXĂ͂܂B
 *	[Ԃl]
 *		G[ꍇɂ FALSE AłȂ TRUE Ԃ܂B
 *	[Rg]
 *		AvP[Vւ̃L[͂́A
 *			ImeProcessKey ---> ToAsciiEx ---> ImeKeyDownHandler --->
 *			TSkkIme_bHandleKeyDown
 *		̏Ԃɏ܂B
 *		̊֐ł͉ꂽL[݂̃L[{[h̃CAEgg
 *		Ή镶R[hɕϊ(܂ł͏ɂĂ邯)A
 *		L[}bvēꂽ@\s֐Ăяo܂B
 *************************************************************************/
BOOL	PASCAL
TSkkIme_bHandleKeyDown (
	register HIMC		hIMC,
	register WPARAM		wParam,
	register LPARAM		lParam,
	register LPBYTE		lpbKeyState,
	register BOOL		fOpen)
{
	LPINPUTCONTEXT		lpIMC ;
	LPCOMPOSITIONSTRING	lpCompStr ;
	LPMYCOMPSTR			lpMyCompStr ;
	WORD				woChara, woVKey ;
	struct CImeDoc*		pDoc ;
	struct TMSG			msg ;
	BOOL				fEaten ;
	BOOL				fToggleIME ;

	if (!hIMC)
		return	FALSE ;

#if defined (MIXED_UNICODE_ANSI) || defined (UNICODE)
	/* Unicode ȂΓ͂ WORD ɂȂB*/
	woChara	= HIWORD (wParam) ;
#else
	/* Unicode łȂA͂ char */
	woChara	= (WORD)((BYTE)HIBYTE (wParam)) ;
#endif
	/*	SHIFT ACONTROL ꂽꍇɂ filter B
	 */
	woVKey	= LOWORD (wParam) ;
	if ((woVKey == VK_SHIFT || woVKey == VK_CONTROL) && woChara == 0)
		return	TRUE ;

	lpIMC 	= ImmLockIMC (hIMC) ;
	if (!lpIMC)
		return	FALSE ;

	lpCompStr		= (LPCOMPOSITIONSTRING)ImmLockIMCC (lpIMC->hCompStr) ;
	lpMyCompStr		= (LPMYCOMPSTR) lpCompStr ;
	pDoc			= &lpMyCompStr->_Doc ;
	
	/*	lpbKeyState nƂĂȂA͖Ȃ̂H
	 */
	msg.m_nMessage	= WM_IME_KEYDOWN ;
	msg.m_wParam	= wParam ;
	msg.m_lParam	= lParam ;
	msg.m_rParam	= 0 ;
	/* rParam ɗ~񂾂Rs[B*/
	if (lpbKeyState != NULL) {
		msg.m_rParam	|= (lpbKeyState [VK_LCONTROL]	& 0x80)? KEYMASK_LCONTROL	: 0 ;
		msg.m_rParam	|= (lpbKeyState [VK_RCONTROL]	& 0x80)? KEYMASK_RCONTROL	: 0 ;
		msg.m_rParam	|= (lpbKeyState [VK_LMENU]		& 0x80)? KEYMASK_LMENU		: 0 ;
		msg.m_rParam	|= (lpbKeyState [VK_RMENU]		& 0x80)? KEYMASK_RMENU		: 0 ;
		msg.m_rParam	|= (lpbKeyState [VK_LSHIFT]		& 0x80)? KEYMASK_LSHIFT		: 0 ;
		msg.m_rParam	|= (lpbKeyState [VK_RSHIFT]		& 0x80)? KEYMASK_RSHIFT		: 0 ;
		msg.m_rParam	|= (lpbKeyState [VK_SCROLL]		& 0x80)? KEYMASK_SCROLL		: 0 ;
		msg.m_rParam	|= (lpbKeyState [VK_NUMLOCK]	& 0x80)? KEYMASK_NUMLOCK	: 0 ;
	} else {
		msg.m_rParam	|= (GetAsyncKeyState (VK_LCONTROL))?	KEYMASK_LCONTROL	: 0 ;
		msg.m_rParam	|= (GetAsyncKeyState (VK_RCONTROL))?	KEYMASK_RCONTROL	: 0 ;
		msg.m_rParam	|= (GetAsyncKeyState (VK_LMENU))?		KEYMASK_LMENU		: 0 ;
		msg.m_rParam	|= (GetAsyncKeyState (VK_RMENU))?		KEYMASK_RMENU		: 0 ;
		msg.m_rParam	|= (GetAsyncKeyState (VK_LSHIFT))?		KEYMASK_LSHIFT		: 0 ;
		msg.m_rParam	|= (GetAsyncKeyState (VK_RSHIFT))?		KEYMASK_RSHIFT		: 0 ;
		msg.m_rParam	|= (GetAsyncKeyState (VK_SCROLL))?		KEYMASK_SCROLL		: 0 ;
		msg.m_rParam	|= (GetAsyncKeyState (VK_NUMLOCK))?		KEYMASK_NUMLOCK		: 0 ;
	}
	msg.m_pt.x		= 0 ;
	msg.m_pt.y		= 0 ;
	msg.m_nTime		= GetTickCount () ;
	if (! ImeDoc_bQueryToggleIMEEvent (pDoc, &msg, &fToggleIME)) 
		fToggleIME	= FALSE ;
	if (! fOpen) {
		if (! fToggleIME) {
			DEBUGPRINTFEX (103, (TEXT ("TSkkIme_bHandleKeyDown -> TSkkIme_vCloseText ()\n"))) ;
			TSkkIme_vCloseText (hIMC) ;
			goto	exit_func ;
		} else {
			ImmSetOpenStatus (hIMC, TRUE) ;
			goto	exit_func ;
		}
	} else {
		if (fToggleIME) {
			DEBUGPRINTFEX (103, (TEXT ("TSkkIme_bHandleKeyDown -> TSkkIme_vCloseText ()\n"))) ;
			TSkkIme_vCloseText (hIMC) ;
			ImmSetOpenStatus (hIMC, FALSE) ;
			goto	exit_func ;
		}
	}
	if (ImeDoc_bFilterEvent (pDoc, &msg)) 
		TSkkIme_bUpdateComposition (hIMC, lpIMC, lpCompStr) ;

  exit_func:
	ImmUnlockIMCC (lpIMC->hCompStr) ;
	ImmUnlockIMC (hIMC) ;
	return	TRUE ;
}

/*************************************************************************
 *	TSkkIme_bHandleKeyUp
 *		ToAsciiEx Ăяo֐BL[ꂽ̏sB
 *	BOOL	PASCAL
 *		TSkkIme_bHandleKeyUp (
 *		HIMC	hIMC,
 *		WPARAM	wParam,
 *		LPARAM	lParam,
 *		LPBYTE	lpbKeyState
 *	)
 *	[p[^]
 *		hIMC
 *			Input Context Handle
 *		wParam
 *			 16 rbgɉꂽ̕R[hA
 *			 16 rbgɉꂽL[̉zL[R[hĂ܂B
 *		lParam
 *			WM_KEYDOWN ɗ^ lParam ƓłB
 *		lpbKeyState
 *			݂̃L[{[h̏Ԃ܂256oCg̔zւ̃|C^łB
 *			IME ͂̓eύXĂ͂܂B
 *	[Ԃl]
 *		G[ꍇɂ FALSE AłȂ TRUE Ԃ܂B
 *	[Rg]
 *		Rg[L[ꂽ Result String ꂽ肽
 *		Ƃ]܂B
 *		̂ɂ]܂ꂽƐ\܂ƁAControl-Key Ȃ
 *		 Result String Ԃ܂ Result String  Control-Key map 
 *		Ԃ Client ݂邩łBʐ^邱Ƃ
 *		o܂B
 *		łǃRg[L[w𗣂ȂŃtH[JXړ
 *		ςɂȂĂ܂܂B(;_;)
 *		AWM_IME_KEYUP 𑗂ĂAvP[V WM_KEYUP 
 *		 (VK_PROCESSKEY Ƃ̂zL[R[hɓĂ܂) Ƃ
 *		̂ SendMessage Œ WM_KEYUP 𑗂ƂĂ܂B
 *		̊֐ skkinput_delay_result_string  TRUE łȂΌĂ΂
 *		Ƃ͂܂B
 *************************************************************************/
BOOL	PASCAL
TSkkIme_bHandleKeyUp (
	register HIMC		hIMC,
	register WPARAM		wParam,
	register LPARAM		lParam,
	register LPBYTE		lpbKeyState,
	register BOOL		fOpen)
{
	return	TRUE ;
}

/*************************************************************************
 *	SKKConversinoStart
 *		ϊJnȌ WM_IME_STARTCONVERSION ̔ssB
 *	BOOL	PASCAL
 *		TSkkIme_bStartConversion (
 *		HIMC	hIMC,
 *		LPINPUTCONTEXT	lpIMC,
 *	)
 *	[p[^]
 *		hIMC
 *			Input Context Handle
 *		lpIMC
 *			Input Context
 *	[Ԃl]
 *		 TRUE
 *	[Rg]
 *		WM_IME_STARTCONVERSION 𔭍sAUser Interface ̂߂̏
 *		ȂĂȂcƂ悤ɏǉ
 *		Ă킯łB
 *************************************************************************/
BOOL	PASCAL
TSkkIme_bStartConversion (
	HIMC				hIMC,
	LPINPUTCONTEXT		lpIMC)
{
	TRANSMSG 		GnMsg ;
	LPMYCOMPSTR		lpMyCompStr ;
	BOOL			fModeInit ;

	DEBUGPRINTFEX (99, (TEXT ("TSkkIme_bStartConversion ()\n"))) ;

	if (ImmGetIMCCSize (lpIMC->hCompStr) < sizeof (MYCOMPSTR)){
		LPCOMPOSITIONSTRING	lpCompStr ;
		DWORD				dwSize ;

		dwSize				= sizeof (MYCOMPSTR) ;
		lpIMC->hCompStr		= ImmReSizeIMCC (lpIMC->hCompStr, dwSize) ;
		lpCompStr			= (LPCOMPOSITIONSTRING)ImmLockIMCC (lpIMC->hCompStr) ;
		lpCompStr->dwSize	= dwSize ;
		lpMyCompStr			= (LPMYCOMPSTR)lpCompStr ;
		fModeInit			= TRUE ;
		lpMyCompStr->_fComposing	= FALSE ;
		ImeConfig_vLoad () ;

	} else {
		lpMyCompStr			= (LPMYCOMPSTR)ImmLockIMCC (lpIMC->hCompStr) ;
		fModeInit			= FALSE ;
	}
	if (! lpMyCompStr->_fComposing) {
		TSkkIme_vInitCompStr ((LPCOMPOSITIONSTRING)lpMyCompStr, CLR_RESULT_AND_UNDET) ;
		if (fModeInit){
			ImeDoc_bInit (&lpMyCompStr->_Doc) ;
			lpMyCompStr->_bCandMode			= FALSE ;
		}
		GnMsg.message	= WM_IME_STARTCOMPOSITION ;
		GnMsg.wParam	= 0 ;
		GnMsg.lParam	= 0 ;
		GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;
		lpMyCompStr->_fComposing	= TRUE ;

		/*	̃R[h mozilla ΍łBmozilla  GCS_RESULT 
		 *	GCS_COMPSTR ɗƑΉłȂB
		 *	Amozilla(1.3betamF)̃oOłBnswindow.cpp
		 *	 mIMECursorPosition ĂȂ߂łB
		 */
		if (fModeInit) {
			GnMsg.message	= WM_IME_COMPOSITION ;
			GnMsg.lParam	= GCS_COMPALL | GCS_CURSORPOS ;
			GnMsg.wParam	= 0 ;
			GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;
		}
	}
	return	TRUE ;
}

BOOL	PASCAL
SKKTerminateComposing (
	register HIMC					hIMC,
	register LPINPUTCONTEXT			lpIMC)
{
	TRANSMSG				GnMsg ;
	register BOOL			fRetval ;
	register LPMYCOMPSTR	lpMyCompStr ;

	GnMsg.message	= WM_IME_ENDCOMPOSITION ;
	GnMsg.wParam	= 0 ;
	GnMsg.lParam	= 0 ;
	GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;

	if (ImmGetIMCCSize (lpIMC->hCompStr) < sizeof (MYCOMPSTR))
		return	FALSE ;

	lpMyCompStr		= (LPMYCOMPSTR)ImmLockIMCC (lpIMC->hCompStr) ;
	if (lpMyCompStr != NULL) {
		lpMyCompStr->_fComposing	= FALSE ;
		ImmUnlockIMCC (lpIMC->hCompStr) ;
		fRetval	= TRUE ;
	} else {
		fRetval	= FALSE ;
	}
	return	fRetval ;
}

/*	Input Context ̏ԂXV֐B
 *
 *	GCS_COMPALL ݒ肵Ă̂ŁA
 *		GCS_COMPSTR			-> dwCompStrLen			x
 *		GCS_COMPATTR		-> dwCompAttrLen		x
 *		GCS_COMPREADSTR		-> dwCompReadStrLen		x
 *		GCS_COMPREADATTR	-> dwCompReadAttrLen	x
 *		GCS_COMPCLAUSE		-> dwCompClauseLen		x
 *		GCS_COMPREADCLAUSE	-> dwCompReadClauseLen	x
 *	ݒ肳Ȃ΂ȂȂB
 *
 *	ƁA
 *		GCS_CURSORPOS		-> dwCursorPos	,		x
 *		GCS_DELTASTART		-> dwDeltaStart ( CompStrLen ̒lz
 *							Ă͂ȂȂ)			x
 *		GCS_RESULTSTR		-> dwResultStrLen		x
 */
BOOL
TSkkIme_bUpdateComposition (
	register HIMC					hIMC,
	register LPINPUTCONTEXT			lpIMC,
	register LPCOMPOSITIONSTRING	lpCompStr)
{
	struct  CImeDoc*		pDoc	= GetDocumentFromCompStr (lpCompStr) ;
	LPCWSTR					wstrPText ;
	LPMYCOMPSTR				lpMyCompStr	= (LPMYCOMPSTR) lpCompStr ;
	int						nPText, nRawShift, nRawCursor, nText, nDeltaStart ;
	unsigned int			uUpdateFlag, uCompFlag ;
	BOOL					fContinue, fComposing, fRetval ;
	WCHAR					wszReadingText [MAXCOMPLEN] ;
	int						nReadingText, nReadingShift ;
	const struct TMSG*		pUnprocessMsg ;
	LPMYCAND				pNewCand ;
	DWORD					dwCandSize ;
	TRANSMSG				GnMsg ;

	/*	fEaten ̍\͗vύXBbZ[W邩ǂ̔ eaten ǂ߂B
	 */
	BOOL					fEaten	= TRUE ;

	fComposing	= lpMyCompStr->_fComposing ;
	uCompFlag	= 0 ;

	wstrPText	= ImeDoc_pGetPreeditText (pDoc, &nPText) ;
	if (! fComposing) 
		TSkkIme_bStartConversion (hIMC, lpIMC) ;

	if (! ImeDoc_bQueryUpdateContext (pDoc, &nRawShift, &nRawCursor, &fContinue)) 
		goto	Exit ;
	DEBUGPRINTFEX (103, (TEXT ("Text:%p/(%d), Shift:%d, Cursor:%d, Continue:%d\n"), wstrPText, nPText, nRawShift, nRawCursor, fContinue)) ;

	nReadingShift	= nRawShift ;
	nReadingText	= ImeDoc_iGetReadingText (pDoc, wszReadingText, ARRAYSIZE (wszReadingText), nRawShift, &nReadingShift) ;
#if defined (DEBUG) || defined (DBG)
	if (nReadingText < ARRAYSIZE (wszReadingText)) {
		wszReadingText [nReadingText]	= L'\0' ;
	} else {
		wszReadingText [ARRAYSIZE (wszReadingText) - 1]	= L'\0' ;
	}
	DEBUGPRINTFEX (103, (TEXT ("ReadingText:\"%s\"/(%d), Shift:%d\n"), wszReadingText, nReadingText, nReadingShift)) ;
#endif

	GnMsg.message	= WM_IME_COMPOSITION ;
	GnMsg.lParam	= 0 ;
	GnMsg.wParam	= 0 ;
	/*	ϊʂݒ肷B*/
	if (nRawShift > 0) {
		memcpy (GETLPRESULTSTR (lpCompStr), wstrPText, MIN (MAXCOMPSIZE, nRawShift) * sizeof (MYCHAR)) ;
		lpCompStr->dwResultStrLen	= MIN (MAXCOMPSIZE, nRawShift) ;
		GnMsg.lParam	|= GCS_RESULTALL ;		/* dwResultStrLen ݒ */
#if defined (DEBUG) || defined (DBG)
		{
			WCHAR	wszBuffer [128] ;
			int		n = (lpCompStr->dwResultStrLen < 127)? lpCompStr->dwResultStrLen : 127 ;
			memcpy (wszBuffer, GETLPRESULTSTR (lpCompStr), n * sizeof (WCHAR)) ;
			wszBuffer [n]	= TEXT ('\0') ;
			DEBUGPRINTFEX (99, (TEXT ("[a]RawShiftString = \"%s\"(%d)\n"), wszBuffer, n)) ;
			memcpy (wszBuffer, wstrPText, n * sizeof (WCHAR)) ;
			wszBuffer [n]	= TEXT ('\0') ;
			DEBUGPRINTFEX (99, (TEXT ("[b]RawShiftString = \"%s\"(%d)\n"), wszBuffer, n)) ;
		}
#endif
		/*	dwResultClauseLen ݒ肵ȂƁAMSMONEY ANZXᔽŗĂ܂B
		 *	 GCS_RESULTALL ł͂Ȃ GCS_RESULTSTR ɂĂĂ悤
		 *	vB
		 */
		lpCompStr->dwResultClauseLen	= 8 ;
		SetClause (GETLPRESULTCLAUSE     (lpCompStr), lpCompStr->dwResultStrLen) ;
	} else {
		lpCompStr->dwResultStrLen		= 0 ;
		lpCompStr->dwResultClauseLen	= 0 ;
		/*  0 łĂg̐ݒ͂ĂB*/
		SetClause (GETLPRESULTCLAUSE     (lpCompStr), lpCompStr->dwResultStrLen) ;
		nRawShift	= 0 ;
	}

	/*	ϊʑ̓ǂ݂ݒ肷B*/
	if (nReadingShift > 0) {
		memcpy (GETLPRESULTREADSTR (lpCompStr), wszReadingText, MIN (MAXCOMPSIZE, nReadingShift) * sizeof (MYCHAR)) ;
		lpCompStr->dwResultReadStrLen	= MIN (MAXCOMPSIZE, nReadingShift) ;

		lpCompStr->dwResultReadClauseLen	= 8 ;
		SetClause (GETLPRESULTREADCLAUSE (lpCompStr), lpCompStr->dwResultReadStrLen) ;
	} else {
		lpCompStr->dwResultReadStrLen		= 0 ;
		lpCompStr->dwResultReadClauseLen	= 0 ;
		/*  0 łĂg̐ݒ͂ĂB*/
		SetClause (GETLPRESULTREADCLAUSE (lpCompStr), lpCompStr->dwResultReadStrLen) ;
	}

	/*	Composition String ݒ肷B*/
	nText	= nPText - nRawShift ;
	DEBUGPRINTFEX (99, (TEXT ("Text:%d, Shift:%d, Cursor:%d\n"), nText, nRawShift, nRawCursor)) ;
	if (nText > 0) {
		int		nStart, nEnd ;
		
		nDeltaStart	= wmemindex (GETLPCOMPSTR (lpCompStr), wstrPText + nRawShift, ((int)lpCompStr->dwCompStrLen < nText)? (int)lpCompStr->dwCompStrLen : nText) ;
		memcpy (GETLPCOMPSTR (lpCompStr), wstrPText + nRawShift, MIN (MAXCOMPSIZE, nText) * sizeof (WCHAR)) ;
		memset ((LPSTR)GETLPCOMPATTR(lpCompStr), ATTR_INPUT, MIN (MAXCOMPSIZE, nText) * sizeof (BYTE)) ;
		if (ImeDoc_bGetConvertedRegion (pDoc, &nStart, &nEnd) && nStart < nEnd && 0 <= nStart && nEnd <= nText) 
			memset ((LPSTR)GETLPCOMPATTR (lpCompStr) + nStart, ATTR_TARGET_CONVERTED, MIN (MAXCOMPSIZE, (nEnd - nStart)) * sizeof (BYTE)) ;
		uCompFlag	|= SKKUI_UPDATE_COMPOSITIONSTRING ;
		uCompFlag	|= (lpCompStr->dwCursorPos != nRawCursor)? SKKUI_UPDATE_CURSOR : 0 ;
	} else {
#if 0
		uCompFlag	|= (lpCompStr->dwCompStrLen > 0)? SKKUI_UPDATE_COMPOSITIONSTRING : 0 ;
#else
		uCompFlag	|= (fContinue)? SKKUI_UPDATE_COMPOSITIONSTRING : 0 ;
#endif
		uCompFlag	|= (lpCompStr->dwCursorPos != nRawCursor)? SKKUI_UPDATE_CURSOR : 0 ;
		TSkkIme_vInitCompStr (lpCompStr, CLR_UNDET) ;
		nDeltaStart	= 0 ;
	}

	/*	Composition String ̓ǂ݂ݒ肷B*/
	if (nReadingText > nReadingShift) {
		int	n	= nReadingText - nReadingShift ;
		memcpy (GETLPCOMPREADSTR (lpCompStr), wszReadingText + nReadingShift, MIN (MAXCOMPSIZE, n) * sizeof (WCHAR)) ;
		memset (GETLPCOMPREADATTR (lpCompStr), 0, MIN (MAXCOMPSIZE, n) * sizeof (BYTE)) ;
		lpCompStr->dwCompReadStrLen		= MIN (MAXCOMPSIZE, n) ;
		lpCompStr->dwCompReadAttrLen	= MIN (MAXCOMPSIZE, n) ;	/* ǂ݂ attribute ͕svH */
	} else {
		lpCompStr->dwCompReadStrLen		= 0 ;
		lpCompStr->dwCompReadAttrLen	= 0 ;
	}
	lpCompStr->dwCursorPos			= MIN (MAXCOMPSIZE, nRawCursor) ;
	lpCompStr->dwCompStrLen			= MIN (MAXCOMPSIZE, nText) ;

	SetClause (GETLPCOMPCLAUSE     (lpCompStr), lpCompStr->dwCompStrLen) ;
	lpCompStr->dwCompClauseLen		= (lpCompStr->dwCompStrLen > 0)? 8 : 0 ;
	SetClause (GETLPCOMPREADCLAUSE (lpCompStr), lpCompStr->dwCompReadStrLen) ;
	lpCompStr->dwCompReadClauseLen	= (lpCompStr->dwCompReadStrLen > 0)? 8 : 0 ;

	/*	mozilla ͂ꂪ 0 łȂƗcBfakeime ͂ꂪ 8
	 *	ɂȂĂƎv̂AǂĂővȂ̂낤H > MS-IME
	 *	t attribute ݒ肳Ă΁cA͖{]|B
	 */
	lpCompStr->dwCompAttrLen		= MIN (MAXCOMPSIZE, nText) ;
	lpCompStr->dwDeltaStart			= MIN (MAXCOMPSIZE, nDeltaStart) ;

	/*	fContinue == FALSE ̏ꍇłAOpenStatus Ȃ application ݂(H)Ȃc
	 *	AfContinue == FALSE ̎ GCS_COMPALL ݒ肳 MSMONEY ŃJ[\ʒu
	 *	ȂB̓IȏǏ́AuJ[\eLXg̐擪ɌŒ肳vłB
	 */
	if (fContinue) 
		GnMsg.lParam	|= GCS_COMPALL | GCS_DELTASTART | GCS_CURSORPOS ;
	GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;

	if (! ImeDoc_bUpdateContext (pDoc))
		goto	Exit ;

	pUnprocessMsg	= ImeDoc_pGetUnprocessEvent (pDoc) ;
	if (pUnprocessMsg != NULL) {
		GnMsg.message	= WM_IME_KEYDOWN ;
		switch (pUnprocessMsg->m_nMessage) {
		case	WM_IME_KEYDOWN:
			/*	wParam ɂ VK_ CODE ȂƂȂB
			 *	AKeyEvent  wParam 𒼐ړ邱Ƃ͂łȂB
			 *	̂ł͂܂̂ɁA܂Ă܂Ƃ́B
			 *	LOWORD (wParam) łB
			 */
			GnMsg.wParam	= (WPARAM) LOWORD (pUnprocessMsg->m_wParam) ;
			GnMsg.lParam	= (LPARAM) pUnprocessMsg->m_lParam ;
			GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;
			break ;
		case	WM_KEYDOWN:
			GnMsg.wParam	= (WPARAM) pUnprocessMsg->m_wParam ;
			GnMsg.lParam	= (LPARAM) pUnprocessMsg->m_lParam ;
			GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;
			break ;
		case	WM_CHAR:
		default:
			break ;
		}
	}

	pNewCand		= ImeDoc_pGetStatusText (pDoc, &dwCandSize) ;
	if (pNewCand != NULL && dwCandSize > sizeof (MYCAND)) {
		LPMYCAND	pMyCand ;

		DWORD	dwPrevCandSize	= ImmGetIMCCSize (lpIMC->hCandInfo) ;

		if (dwPrevCandSize != dwCandSize) {
			lpIMC->hCandInfo	= ImmReSizeIMCC (lpIMC->hCandInfo, dwCandSize) ;
		}
		pMyCand	= (LPMYCAND)ImmLockIMCC (lpIMC->hCandInfo) ;
		memcpy (pMyCand, pNewCand, dwCandSize) ;

		if (! lpMyCompStr->_bCandMode) {
			TSkkIme_vGenerateMsgOpenCandidate (hIMC, lpIMC, lpCompStr) ;
			lpMyCompStr->_bCandMode	= TRUE ;
		} else {
			TSkkIme_vGenerateMsgUpdateCandidate (hIMC, lpIMC) ;
		}
	} else {
		if (lpMyCompStr->_bCandMode) {
			TSkkIme_vGenerateMsgCloseCandidate (hIMC, lpIMC, TRUE) ;
			lpMyCompStr->_bCandMode	= FALSE ;
		}
	}

	uUpdateFlag	= ImeDoc_uGetUpdateFlag (pDoc) ;
	if (uUpdateFlag & SKKUI_UPDATE_CONVERSIONMODE) 
		TSkkIme_bNotifyChangeConversionMode (hIMC) ;
	lpMyCompStr->dwUpdateFlag	= uCompFlag ;
	ImeDoc_vClearUpdateFlag (pDoc) ;

	if (! fContinue) 
		SKKTerminateComposing (hIMC, lpIMC) ;
	fRetval	= TRUE ;

  Exit:
	return	fRetval ;
}

void
TSkkIme_vGenerateMsgOpenCandidate (
	HIMC				hIMC,
	LPINPUTCONTEXT		lpIMC,
	LPCOMPOSITIONSTRING	lpCompStr)
{
	LPMYCOMPSTR	lpMyCompStr	= (LPMYCOMPSTR) lpCompStr ;
	TRANSMSG	GnMsg ;

	GnMsg.message	= WM_IME_NOTIFY ;
	GnMsg.wParam	= IMN_OPENCANDIDATE ;
	GnMsg.lParam	= 1L ;
	GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;

	DEBUGPRINTFEX (99, (TEXT ("WM_IME_NOTIFY/IMN_OPENCANDIDATE\n"))) ;
	return ;
}

void
TSkkIme_vGenerateMsgUpdateCandidate (
	HIMC			hIMC,
	LPINPUTCONTEXT	lpIMC)
{
	TRANSMSG		GnMsg ;

	GnMsg.message	= WM_IME_NOTIFY ;
	GnMsg.wParam	= IMN_CHANGECANDIDATE ;
	GnMsg.lParam	= 1L ;
	GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;
	return ;
}

void
TSkkIme_vGenerateMsgCloseCandidate (
	HIMC			hIMC,
	LPINPUTCONTEXT	lpIMC,
	BOOL			bClear)
{
	TRANSMSG		GnMsg ;

	if (bClear) {
		LPCANDIDATEINFO	pCandInfo ;

		pCandInfo		= (LPCANDIDATEINFO)ImmLockIMCC (lpIMC->hCandInfo) ;
		if (pCandInfo != NULL) {
			TSkkIme_vClearCandidate (pCandInfo) ;
			ImmUnlockIMCC (lpIMC->hCandInfo) ;
		}
	}
	GnMsg.message		= WM_IME_NOTIFY ;
	GnMsg.wParam		= IMN_CLOSECANDIDATE ;
	GnMsg.lParam		= 1L ;
	GenerateMessage (hIMC, lpIMC, g_lpCurTransKey, (LPTRANSMSG)&GnMsg) ;

	DEBUGPRINTFEX (99, (TEXT ("WM_IME_NOTIFY/IMN_CLOSECANDIDATE\n"))) ;
	return ;
}


