#if defined (UNITTEST)
#include "StdAfx.h"
#else
#include "local.h"
#endif
/*#include "c:\Program Files\Microsoft Platform SDK\include\lmcons.h"*/
#include <lmcons.h>
#include "TCommuSession.h"
#include "TPacket.h"
#include "TProtocol.h"
#include "debug.h"

#define	SKKIME_PIPE_PREFIX_NUM	(29)
#define	OPENSERVER_MAX_RETRY	(128)

#define	SKKISERV_TIMEOUT		(3000)
#define	WAITPIPE_TIMEOUT		(100)
#define	WAITPIPE_SLEEPTIME		(100)

#if defined (UNITTEST)
#define	SKKISERVER_PIPE_BASENAME	L"\\\\.\\pipe\\skkiserv010500_pipe_%s"
#define	PROGRAMF_DIR				L"IME"
#define	SKKIMEProg_S_DIR			L"SKKIM15"
#define	SKKIMEServerFile			L"skkiserv.exe"
#endif

/*========================================================================
 *	prototypes
 */
static	BOOL	tCommunicateSession_bInitPipeName	(void) ;
static	BOOL	tCommunicateSession_bStartServer	(void) ;

/*========================================================================
 *	global variables
 */
static	WCHAR	m_rszSkkImePipeName [MAX_PATH]	= SKKISERVER_PIPE_BASENAME ;

/*========================================================================
 *	Public functions
 */
HANDLE
TCommunicateSession_hOpenServer (void)
{
	register HANDLE		hPipe ;
	register int		nRetry ;
	register BOOL		fServerStart ;

	hPipe			= INVALID_HANDLE_VALUE ;
	nRetry			= OPENSERVER_MAX_RETRY ;
	fServerStart	= FALSE ;
	for ( ; ; ) {
		register DWORD	dwError ;

		hPipe	= CreateFileW (m_rszSkkImePipeName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL) ;
		dwError	= GetLastError () ;
		if (hPipe != INVALID_HANDLE_VALUE) {
			break ;
		}
		if (dwError != ERROR_PIPE_BUSY) {
			DEBUGPRINTF ((TEXT ("CreateFileW failure (error:0x%lx)\n"), dwError)) ;
			/*	ERROR_FILE_NOT_FOUND ƂƂ́Askkiserv ȂAor
			 *	skkiserv  CreatePipe  server pipe 쐬ĂȂB
			 *
			 *	pipe server  multi-instance łȂƁApipe ̍쐬̊Ԃ˂
			 *	ERROR_FILE_NOT_FOUND 邱ƂB
			 */
			if (dwError == ERROR_FILE_NOT_FOUND) {
				if (! fServerStart) {
#if !defined (UNITTEST)
					if (! tCommunicateSession_bStartServer ())
						break ;
					fServerStart	= TRUE ;
#else
					break ;
#endif
				}
				/*	skkiserv NĂ̂҂BSleep ȂƑq˂
				 *	sB
				 */
				Sleep (WAITPIPE_SLEEPTIME) ;
				continue ;
			}
			/*	ERROR_FILE_NOT_FOUND ȊȌꍇɂ͑zOȂ̂ŔB*/
			break ;
		}
		/*	PIPE  BUSY łȂȂ̂҂BA҂
		 *	PIPE  BUSY łȂȂ̂ɐ䂪߂ȂƂ̂
		 *	(^C~OˑH)
		 *	莞Ԃŏ𔲂ABUSY Ȃēx҂AƂƂJ
		 *	ԂB
		 */
		if (WaitNamedPipeW (m_rszSkkImePipeName, WAITPIPE_TIMEOUT) == 0) {
			dwError	= GetLastError () ;
#if defined (DBG)
			DEBUGPRINTF ((TEXT ("WaitNamedPipeW failed (0x%lx)\n"), dwError)) ;
#endif
			if (dwError != ERROR_SEM_TIMEOUT && dwError != ERROR_FILE_NOT_FOUND)
				break ;
		}
	}
	return	hPipe ;
}

BOOL
TCommunicateSession_bSend (
	HANDLE							hPipe,
	struct CTPacket*				pPacket)
{
	register const BYTE*	ptr ;
	register int		nptr ;
	DWORD				dwWrite ;

	ptr			= TPacket_pGetData (pPacket) ;
	nptr		= TPacket_iGetDataSize (pPacket) ;
	while (nptr > 0) {
		if (! WriteFile (hPipe, ptr, nptr, &dwWrite, NULL)) {
#if defined (DBG)
			DWORD	dwError	= GetLastError () ;
			DEBUGPRINTF ((TEXT ("_Send (data:%d) fail (%ld)\n"), nptr, dwWrite)) ;
#endif
			return	FALSE ;
		}
		ptr		+= dwWrite ;
		nptr	-= dwWrite ;
	}
	FlushFileBuffers (hPipe) ;
	return	TRUE ;
}

BOOL
TCommunicateSession_bRecv (
	HANDLE							hPipe,
	struct CTPacket*				pPacket)
{
	DWORD	dwRead ;

	if (! ReadFile (hPipe, TPacket_pGetData (pPacket), TPACKET_MSGBUFSIZE, &dwRead, NULL)) {
#if defined (DBG)
		DWORD	dwError	= GetLastError () ;
		DEBUGPRINTF ((TEXT ("_Recv () error (%ld)\n"), dwError)) ;
#endif
		return	FALSE ;
	}
	TPacket_vSetDataSize (pPacket, dwRead) ;
	return	TRUE ;
}

BOOL
TCommunicateSession_bClassInit (void)
{
	WCHAR		rszUserName [UNLEN + 1] ;
	ULONG		ulUserNameLen ;
	LPWSTR		ptr ;

	ulUserNameLen	= sizeof (rszUserName) / sizeof (rszUserName [0]) - 1 ;
	if (! GetUserNameW (rszUserName, &ulUserNameLen)) 
		return	FALSE ;
	rszUserName [UNLEN]	= TEXT ('\0') ;

	/*	VXe[U͍ȂcƂB
	 *
	 *	VXeƒʏ̃[U̓vZXԂL̂ŁA skkiserv
	 *	NƁAsystem ̃vtB[œĂ skkiserv Ɉʃ[
	 *	U query 邱ƂɂȂ...ʃ[U̐ݒ肪fȂȂ
	 *	܂B
	 *	``SYSTEM'' 炵ǁAcase ĔrB
	 */
	if (! lstrcmpiW (rszUserName, L"system")) 
		return	FALSE ;

	ptr	= rszUserName ;
	while (*ptr != L'\0') {
		/*	pXƂĎgpłȂ͎̂̂ĂB*/
		if (*ptr == L'\\' || *ptr == L'*' || *ptr == L'?' ||
			*ptr == L':'  || *ptr == L'.' || *ptr == L'/') {
			*ptr = L'-' ;
		}
		ptr	++ ;
	}
	wnsprintf (m_rszSkkImePipeName, MAX_PATH - 1, SKKISERVER_PIPE_BASENAME, rszUserName) ;
	m_rszSkkImePipeName [MAX_PATH - 1]	= L'\0' ;
	DEBUGPRINTF ((TEXT ("skimic: pipename = \"%s\"\n"), m_rszSkkImePipeName)) ;
	return	TRUE ;
}

BOOL
tCommunicateSession_bStartServer (void)
{
	WCHAR					rszWindowsDir [MAX_PATH + 1] ;
	WCHAR					rszCmd [MAX_PATH + 1] ;
    STARTUPINFOW			si ;
    PROCESS_INFORMATION		pi ;
	register int			nWindowsDir, n ;
	register BOOL			fRetval ;

	nWindowsDir	= GetWindowsDirectoryW (rszWindowsDir, ARRAYSIZE (rszWindowsDir) - 1) ;
	rszWindowsDir [nWindowsDir]		= L'\0' ;
	n			= wnsprintf (rszCmd, ARRAYSIZE (rszCmd) - 1, L"%s" L"\\" PROGRAMF_DIR L"\\" SKKIMEProg_S_DIR L"\\" SKKIMEServerFile, rszWindowsDir) ;
	rszCmd [n]	= L'\0' ;


	ZeroMemory (&si, sizeof(si)) ;
	si.cb	= sizeof (si) ;

	/*	 pipename n\ɖ߂Ă݂B(PTNVQS())
	 *	Windows2000sp4  skkiserv  connect s錻ۂ̂ŁB
	 *	͂̏낤H 
	 *	̂ skkiserv  pipename ɍĂƗ\zĂ鎞
	 *	if (! CreateProcessW (rszCmd, L"", NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
	 *	B
	 */
	if (! CreateProcessW (rszCmd, m_rszSkkImePipeName + SKKIME_PIPE_PREFIX_NUM, NULL, NULL, FALSE, 0, NULL, rszWindowsDir, &si, &pi)) {
		fRetval	= FALSE ;
	} else {
		WaitForInputIdle (pi.hProcess, SKKISERV_TIMEOUT) ;
		
		/* svɂȂnhB*/
		CloseHandle (pi.hProcess) ;
		CloseHandle (pi.hThread) ;
		fRetval	= TRUE ;
	}
	return	fRetval ;
}

