/* # SKKIME1.0 (Simple Kana-Kanji Input Method Editor for Windows2000)
 * uiminibuf.c
 * This file is part of skkime1.0.
 * Copyright (C) 1999
 * Takashi SAKAMOTO (tatari_sakamoto@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "skki1_5.h"
#include "resource.h"
#include "skkui.h"

/*
 *	vg^Cv錾B
 */
static	void	PASCAL	uiMinibuf_vPaint			(HWND) ;
static	BOOL	PASCAL	uiMinibuf_bDrag				(HWND, UINT) ;
static	BOOL	PASCAL	uiMinibuf_bHandleContextMenu(HWND, HIMC) ;
static	int		PASCAL	uiMinibuf_iPopupMenu		(HWND, HIMC) ;
static	DWORD	PASCAL	uiMinibuf_dwGetCursorPos	(HDC, HIMC, LPPOINT) ;
static	BOOL	PASCAL	uiMinibuf_bInvalidate		(LPUIEXTRA, LPINPUTCONTEXT) ;
static	void	PASCAL	uiMinibuf_vMoveDefault		(HWND, LPUIEXTRA, LPINPUTCONTEXT) ;
#if defined (does_not_work_well)
static	void	PASCAL	uiMinibuf_vMoveExclude		(HWND, LPUIEXTRA, LPINPUTCONTEXT, CANDIDATEFORM*) ;
static	void	PASCAL	uiMinibuf_vMovePosition		(HWND, LPUIEXTRA, LPINPUTCONTEXT, CANDIDATEFORM*) ;
#endif
static	BOOL	PASCAL	uiMinibuf_bAdjustWindowSize	(HWND, LPINPUTCONTEXT, LPSIZE) ;

/*
 *	---  ֐̒` ---  ֐̒` ---  ֐̒` ---
 */

LRESULT	CALLBACK
UIMinibuf_WndProc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HWND	hUIWnd ;

	switch (message){
	case WM_CREATE:
		SetWindowLong (hWnd, FIGWL_UIMINIBUF_MOUSE, 0L) ;
		break ;

	case WM_PAINT:
		uiMinibuf_vPaint (hWnd) ;
		break;

	case WM_SETCURSOR:
		uiMinibuf_bDrag (hWnd, HIWORD (lParam)) ;
		break ;

	case WM_LBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_MOUSEMOVE:
	case WM_NCMOUSEMOVE:
	case WM_LBUTTONUP:
	case WM_RBUTTONUP:
		uiMinibuf_bDrag (hWnd, message) ;
		break ;

	case WM_MOVE:
		hUIWnd	= (HWND)GetWindowLongPtr (hWnd, FIGWL_UIMINIBUF_SVRWND) ;
		if (IsWindow (hUIWnd))
			SendMessage (hUIWnd, WM_UI_CANDMOVE, wParam, lParam) ;
		break;

	default:
		if (!MyIsIMEMessage (message))
			return	DefWindowProc (hWnd, message, wParam, lParam) ;
		break ;
	}
	return	0L ;
}

void	PASCAL
UIMinibuf_Create (
	HWND				hUIWnd,
	LPUIEXTRA			lpUIExtra,
	LPINPUTCONTEXT		lpIMC)
{
	LPCOMPOSITIONSTRING	lpCompStr ;
	POINT							pt ;
	HINSTANCE				hInstance ;

	if (lpUIExtra->uiGuide.rc.left == -1) {
		RECT	rc ;
		GetWindowRect (lpIMC->hWnd, &rc) ;
		lpUIExtra->uiGuide.rc.left	= rc.left;
		lpUIExtra->uiGuide.rc.top	= rc.bottom;
	}
	hInstance	= (HINSTANCE)GetWindowLongPtr (hUIWnd, GWLP_HINSTANCE) ;
	if (!IsWindow (lpUIExtra->uiGuide.hWnd)){
		HDC			hdcIC ;
		TEXTMETRIC	tm ;
		int			dx, dy ;
		RECT		rcWork ;

		/*	Window ̑傫 guess B
		 */
		hdcIC	= CreateIC (TEXT ("DISPLAY"), NULL, NULL, NULL) ;
		GetTextMetrics (hdcIC, &tm) ;
		dx	= tm.tmAveCharWidth * DEFAULT_MINIBUF_CHARWIDTH ;
		dy 	= tm.tmHeight + tm.tmExternalLeading ;
		DeleteDC (hdcIC) ;

		lpUIExtra->uiGuide.rc.right		= dx + 2 * GetSystemMetrics(SM_CXBORDER)
			+ 2 * GetSystemMetrics(SM_CXEDGE) ;
		lpUIExtra->uiGuide.rc.bottom	= dy + 2 * GetSystemMetrics(SM_CYBORDER)
			+ 2 * GetSystemMetrics(SM_CYEDGE) ;

		/*	\ʒuA\TCY Window ʊOɏoȂǂ
		 *	`FbNB
		 */
		SystemParametersInfo (SPI_GETWORKAREA, 0, &rcWork, FALSE) ;
		if ((lpUIExtra->uiGuide.rc.left + lpUIExtra->uiGuide.rc.right) > rcWork.right) {
			lpUIExtra->uiGuide.rc.left	= rcWork.right - lpUIExtra->uiGuide.rc.right ;
			if (lpUIExtra->uiGuide.rc.left < 0) {
				lpUIExtra->uiGuide.rc.left	= 0 ;
				lpUIExtra->uiGuide.rc.right	= rcWork.right ;
			}
		}
		if ((lpUIExtra->uiGuide.rc.top + lpUIExtra->uiGuide.rc.bottom) > rcWork.bottom) {
			lpUIExtra->uiGuide.rc.top	= rcWork.bottom - lpUIExtra->uiGuide.rc.bottom ;
			if (lpUIExtra->uiGuide.rc.top < 0)
				lpUIExtra->uiGuide.rc.top	= 0 ;
		}
		lpUIExtra->uiGuide.hWnd = 
			CreateWindowEx (WS_EX_WINDOWEDGE,
							(LPTSTR)g_szGuideClassName, NULL,
							WS_COMPDEFAULT | WS_DLGFRAME,
							lpUIExtra->uiGuide.rc.left,
							lpUIExtra->uiGuide.rc.top,
							lpUIExtra->uiGuide.rc.right,
							lpUIExtra->uiGuide.rc.bottom,
							hUIWnd, NULL, hInstance, NULL) ;
	}
	SetWindowLongPtr (lpUIExtra->uiGuide.hWnd, FIGWL_UIMINIBUF_SVRWND, 		(LONG_PTR)hUIWnd) ;
	SetWindowLong    (lpUIExtra->uiGuide.hWnd, FIGWL_UIMINIBUF_PREVCURSOR,	(LONG)0L) ;
	ShowWindow (lpUIExtra->uiGuide.hWnd, SW_HIDE) ;
	lpUIExtra->uiGuide.bShow	= FALSE ;

	DEBUGPRINTFEX (99, (TEXT ("UIMinibuf_Create (x:%d, y:%d, w:%d, h:%d)\n"),
						lpUIExtra->uiGuide.rc.left,
						lpUIExtra->uiGuide.rc.top,
						lpUIExtra->uiGuide.rc.right,
						lpUIExtra->uiGuide.rc.bottom)) ;
	return ;
}

void	PASCAL
UIMinibuf_Hide (LPUIEXTRA lpUIExtra)
{
	if (IsWindow (lpUIExtra->uiGuide.hWnd)){
		ShowWindow (lpUIExtra->uiGuide.hWnd, SW_HIDE) ;
		lpUIExtra->uiGuide.bShow	= FALSE ;
	}
	return ;
}

void	PASCAL
UIMinibuf_NotifyComposition (
	HWND				hUIWnd,
	LPUIEXTRA			lpUIExtra,
	LPINPUTCONTEXT		lpIMC)
{
	RECT	rc ;
	SIZE	sz ;

	if (! IsWindow (lpUIExtra->uiGuide.hWnd))
		return ;
	if (! (lpUIExtra->dwShowStyle & ISC_SHOWUIGUIDELINE))
		return ;

	GetWindowRect (lpUIExtra->uiGuide.hWnd, &rc) ;
	sz.cx	= rc.right  - rc.left ;
	sz.cy	= rc.bottom - rc.top ;
	/*	Window ̃TCYύXKv邩ǂ`FbNB*/
	if (uiMinibuf_bAdjustWindowSize (hUIWnd, lpIMC, &sz)) {
		if (sz.cx != (rc.right - rc.left) || sz.cy != (rc.bottom - rc.top)) {
			DEBUGPRINTFEX (99, (TEXT ("uiMinibuf_bAdjustWindowSize (%d!=%d or %d!=%d)\n"), 
				sz.cx, (rc.right - rc.left), sz.cy, (rc.bottom - rc.top))) ;
			UIMinibuf_Move (hUIWnd, lpUIExtra, lpIMC) ;
		}
	}
	InvalidateRect (lpUIExtra->uiGuide.hWnd, NULL, FALSE) ;
	return ;
}

void	PASCAL
UIMinibuf_Move (
	HWND				hUIWnd,
	LPUIEXTRA			lpUIExtra,
	LPINPUTCONTEXT		lpIMC)
{
	CANDIDATEFORM	acf ;

	if (! (lpUIExtra->dwShowStyle & ISC_SHOWUIGUIDELINE))
		return ;

#if defined (does_not_work_well)
	/*	Office2003 ŎĂ݂ƂACaret ړĂ Exclude ̏񂪍XVȂB
	 *	炩̎i(ȂAOpenCandidate/CloseCandidate ̐ĂяoȂɂ́Ap
	 *	łȂ悤B(Mon Nov 14 21:16:31 2005)
	 *	̂Ƃ̓RgAEgB
	 */
	if ((lpUIExtra->dwShowStyle & ISC_SHOWUIALLCANDIDATEWINDOW) != 0 &&
		lpUIExtra->hIMC != NULL) {
		if (ImmRequestMessage (lpUIExtra->hIMC, IMR_CANDIDATEWINDOW, (LPARAM)&acf) != 0) {
			if (acf.dwIndex != -1 ){
				if (acf.dwStyle == CFS_EXCLUDE){
					uiMinibuf_vMoveExclude (hUIWnd, lpUIExtra, lpIMC, &acf) ;
				}  else if (acf.dwStyle == CFS_CANDIDATEPOS){
					uiMinibuf_vMovePosition (hUIWnd, lpUIExtra, lpIMC, &acf) ;
				}
				return ;
			}
		}
	}
#endif
	uiMinibuf_vMoveDefault (hUIWnd, lpUIExtra, lpIMC) ;
	return ;
}

/*========================================================================
 *	private functions
 */
/*
 *	Minibuffer Window  WM_PAINT bZ[W֐B
 *-----
 */
void	PASCAL
uiMinibuf_vPaint (
	HWND	hMinibufWnd)
{
	PAINTSTRUCT		ps ;
	HIMC			hIMC ;
	LPINPUTCONTEXT	lpIMC ;
	LPMYCOMPSTR		lpMyCompStr ;
	HBRUSH			hbr, hBrushParent ;
	HDC				hDC, hPDC ;
	HFONT			hOldFont ;
	RECT			rc ;
	HWND			hSvrWnd ;
	int				nRegionStart, nRegionEnd ;
	int				x, y, nDY, nText, nTextPos ;
	LPMYSTR			pText ;
	long			lLength, lCursorPos, lPrevCursorPos ;
	BOOL			fRegionSelected ;
	DWORD			dwLevel, dwSize ;
	TEXTMETRIC		tm ;
	TCHAR			bufGuide [MAXGLCHAR] ;

	hDC		= BeginPaint (hMinibufWnd, &ps) ;
	hSvrWnd	= (HWND)GetWindowLongPtr (hMinibufWnd,	FIGWL_UIMINIBUF_SVRWND) ;
	hIMC	= (HIMC)GetWindowLongPtr (hSvrWnd,		IMMGWLP_IMC) ;
	if (! hIMC)
		goto	exit_func ;

	dwLevel	= ImmGetGuideLine (hIMC, GGL_LEVEL, NULL, 0) ;
	if (! dwLevel)
		goto	exit_func ;
	dwSize	= ImmGetGuideLine (hIMC, GGL_STRING, NULL, 0) ;
	if (dwSize <= 0)
		goto	exit_func ;

	lpIMC	= ImmLockIMC (hIMC) ;
	if (! lpIMC)
		goto	exit_func ;

	if (dwSize >= (sizeof (bufGuide) - sizeof (TCHAR)))
		dwSize	= sizeof (bufGuide) - sizeof (TCHAR) ;

	dwSize	= ImmGetGuideLine (hIMC, GGL_STRING, bufGuide, dwSize) ;
	lLength	= dwSize / sizeof (TCHAR) ;
	hOldFont= TSkkIme_hCheckNativeCharset (hDC) ;

	/* Minibuffer Window ̑傫𓾂B*/
	GetClientRect (hMinibufWnd, &rc) ;
		
	/* eWindow ̔wiF𓾂B*/
	hPDC 			= GetDC (GetParent (hMinibufWnd)) ;
	hBrushParent	= CreateSolidBrush (GetBkColor (hPDC)) ;
	lPrevCursorPos	= GetWindowLong (hMinibufWnd, FIGWL_UIMINIBUF_PREVCURSOR) ;

	GetTextMetrics (hDC, &tm) ;
	nDY			= tm.tmHeight ;

	SetBkMode (hDC, OPAQUE) ;
	SetBkColor (hDC, GetBkColor (hPDC)) ;
	SetTextColor (hDC, RGB (0, 0, 0)) ;

	DeleteObject (hBrushParent) ;
	ReleaseDC (GetParent (hMinibufWnd), hPDC) ;

	if (hOldFont) 
		DeleteObject(SelectObject(hDC, hOldFont));

  exit_func:
	EndPaint (hMinibufWnd, &ps) ;
	return ;
}

BOOL	PASCAL
uiMinibuf_bDrag (
	HWND	hWnd,
	UINT	uMessage)
{
	return	TRUE ;
}

void	PASCAL
uiMinibuf_vMoveDefault (
	HWND				hUIWnd,
	LPUIEXTRA			lpUIExtra,
	LPINPUTCONTEXT		lpIMC)
{
	RECT	rc, rcWork ;
	SIZE	szGuidWnd ;

	GetWindowRect (lpIMC->hWnd, &rc) ;
	lpUIExtra->uiGuide.rc.left	= rc.left;
	lpUIExtra->uiGuide.rc.top	= rc.bottom;

	szGuidWnd.cx	= lpUIExtra->uiGuide.rc.right ;
	szGuidWnd.cy	= lpUIExtra->uiGuide.rc.bottom ;
	if (uiMinibuf_bAdjustWindowSize (hUIWnd, lpIMC, &szGuidWnd)) {
		lpUIExtra->uiGuide.rc.right		= szGuidWnd.cx ;
		lpUIExtra->uiGuide.rc.bottom	= szGuidWnd.cy ;
	}

	SystemParametersInfo (SPI_GETWORKAREA, 0, &rcWork, FALSE) ;
	if ((lpUIExtra->uiGuide.rc.left + lpUIExtra->uiGuide.rc.right) > rcWork.right) {
		lpUIExtra->uiGuide.rc.left	= rcWork.right - lpUIExtra->uiGuide.rc.right ;
		if (lpUIExtra->uiGuide.rc.left < 0) {
			lpUIExtra->uiGuide.rc.left	= 0 ;
			lpUIExtra->uiGuide.rc.right	= rcWork.right ;
		}
	}
	if ((lpUIExtra->uiGuide.rc.top + lpUIExtra->uiGuide.rc.bottom) > rcWork.bottom) {
		lpUIExtra->uiGuide.rc.top	= rcWork.bottom - lpUIExtra->uiGuide.rc.bottom ;
		if (lpUIExtra->uiGuide.rc.top < 0)
			lpUIExtra->uiGuide.rc.top	= 0 ;
	}
	if (!IsWindow (lpUIExtra->uiGuide.hWnd))
		return ;

	MoveWindow (lpUIExtra->uiGuide.hWnd,
				lpUIExtra->uiGuide.rc.left,
				lpUIExtra->uiGuide.rc.top,
				lpUIExtra->uiGuide.rc.right,
				lpUIExtra->uiGuide.rc.bottom,
				TRUE) ;
	return ;
}

#if defined (does_not_work_well)
void	PASCAL
uiMinibuf_vMoveExclude (
	HWND			hUIWnd,
	LPUIEXTRA		lpUIExtra,
	LPINPUTCONTEXT	lpIMC,
	CANDIDATEFORM*	pCF)
{
	RECT	rc ;
	RECT	rcWork ;
	RECT	rcExclude ;
	RECT	rcUIWnd ;
	POINT	pt ;
	SIZE	szGuidWnd ;

	/* Windows ̍Ɨ̈𓾂B(Desktop ̑傫) */
	SystemParametersInfo (SPI_GETWORKAREA, 0, &rcWork, FALSE) ;

	if (IsWindow(lpUIExtra->uiGuide.hWnd)){
		GetClientRect (lpUIExtra->uiGuide.hWnd, &rc) ;
	} else {
		rc.left	= rc.top	= rc.bottom	= rc.right	= 0 ;
	}

	/* ⑋\Ă͂Ȃ̈𓾂B*/
	pt.x	= pCF->rcArea.left ;
	pt.y	= pCF->rcArea.top ;
	ClientToScreen (lpIMC->hWnd, &pt) ;
	rcExclude.left		= pt.x ;
	rcExclude.top		= pt.y ;
	pt.x	= pCF->rcArea.right ;
	pt.y	= pCF->rcArea.bottom ;
	ClientToScreen (lpIMC->hWnd, &pt) ;
	rcExclude.right		= pt.x ;
	rcExclude.bottom	= pt.y ;

	/* ⑋̈ʒuƑ傫肷B*/
	GetClientRect (lpIMC->hWnd, &rcUIWnd) ;
	szGuidWnd.cx	= rcUIWnd.right - rcUIWnd.left + 2 * GetSystemMetrics (SM_CXEDGE) ;
	szGuidWnd.cy	= rc.bottom - rc.top + 2 * GetSystemMetrics (SM_CYEDGE) ;
	uiMinibuf_bAdjustWindowSize (hUIWnd, lpIMC, &szGuidWnd) ;

	pt.x	= 0 ;
	pt.y	= szGuidWnd.cy ;
	ClientToScreen (lpIMC->hWnd, &pt) ;

	/* Window ʊOɏoĂ܂Ƃ悤ɒB*/
	if ((pt.x + szGuidWnd.cx) > rcWork.right){
		pt.x	= rcWork.right - szGuidWnd.cx ;
		if (pt.x < 0)
			pt.x	= 0 ;
	}
	if ((pt.y + szGuidWnd.cy) > rcWork.bottom){
		pt.y = rcWork.bottom - szGuidWnd.cy ;
		if (pt.y < 0)
			pt.y	= 0 ;
	}
	/* Window w肵̈ɂĂȂǂ`FbNB*/
	if (IsCrossRectangle (&pt, &szGuidWnd, &rcExclude)){
		pt.y	= rcExclude.bottom ;
		if ((pt.y + szGuidWnd.cy) > rcWork.bottom)
			pt.y	= rcExclude.top - szGuidWnd.cy ;
	}

	if (IsWindow (lpUIExtra->uiGuide.hWnd)){
		GetWindowRect (lpUIExtra->uiGuide.hWnd, &rc) ;
		MoveWindow (lpUIExtra->uiGuide.hWnd, pt.x, pt.y, szGuidWnd.cx, szGuidWnd.cy, TRUE) ;
	}
	return ;
}

void	PASCAL
uiMinibuf_vMovePosition (
	HWND			hUIWnd,
	LPUIEXTRA		lpUIExtra,
	LPINPUTCONTEXT	lpIMC, 
	CANDIDATEFORM*	pCF)
{
	RECT	rc ;
	RECT	rcUIWnd ;
	RECT	rcWork ;
	SIZE	szGuidWnd ;
	POINT	pt ;

	pt.x	= pCF->ptCurrentPos.x ;
	pt.y	= pCF->ptCurrentPos.y ;
	ClientToScreen (lpIMC->hWnd, &pt) ;
		
	/* Windows ̍Ɨ̈𓾂B(Desktop ̑傫) */
	SystemParametersInfo (SPI_GETWORKAREA, 0, &rcWork, FALSE) ;
	if (IsWindow (lpUIExtra->uiGuide.hWnd)) {
		GetWindowRect (lpUIExtra->uiGuide.hWnd, &rc) ;
	} else {
		rc.left = rc.right = rc.top = rc.bottom	= 0 ;
	}

	/*	Window ʊOɏoƂӐ}Ă邩H 
	 *
	 *	ǂʊOw肷邱ƂŁA\Ӑ}Ă
	 *	AvP[V݂悤B
	 * ()
	 *	PSO
	 */
	if ((pt.x + 1) >= rcWork.right && (pt.y + 1) >= rcWork.bottom) {
		/*	̏ꍇɂ͉ʊOɏoƂӐ}ĂƔfB*/
		if (IsWindow (lpUIExtra->uiGuide.hWnd))
			ShowWindow (lpUIExtra->uiGuide.hWnd, SW_HIDE) ;
		return ;
	}

	/* ⑋̑傫vZB*/
	GetClientRect (lpIMC->hWnd, &rcUIWnd) ;
	szGuidWnd.cx	= rcUIWnd.right - rcUIWnd.left + 2 * GetSystemMetrics (SM_CXEDGE) ;
	szGuidWnd.cy	= rc.bottom - rc.top + 2 * GetSystemMetrics (SM_CYEDGE) ;
	uiMinibuf_bAdjustWindowSize (hUIWnd, lpIMC, &szGuidWnd) ;

	/*	Window ʊOɏoĂ܂Ƃ悤ɒB
	 */
	if ((pt.x + szGuidWnd.cx) > rcWork.right){
		pt.x	= rcWork.right - szGuidWnd.cx ;
		if (pt.x < 0)
			pt.x	= 0 ;
	}
	if ((pt.y + szGuidWnd.cy) > rcWork.bottom){
		pt.y = rcWork.bottom - szGuidWnd.cy ;
		if (pt.y < 0)
			pt.y	= 0 ;
	}
	if (IsWindow (lpUIExtra->uiGuide.hWnd)){
		MoveWindow (lpUIExtra->uiGuide.hWnd, pt.x, pt.y, szGuidWnd.cx, szGuidWnd.cy, TRUE) ;
	}
	return ;
}
#endif

BOOL	PASCAL
uiMinibuf_bAdjustWindowSize (
	HWND			hwnd,
	LPINPUTCONTEXT	lpIMC,
	LPSIZE			lpSize)
{
	HDC				hDC ;
	HFONT			hOldFont ;
	HIMC			hIMC ;
	HWND			hSvrWnd ;
	SIZE			sz ;
	DWORD			dwLevel, dwSize ;
	int				nDefaultWidth, nDefaultHeight, nFrameWidth, nFrameHeight ;
	int				nWidth, nHeight ;
	TCHAR			bufGuide [MAXGLCHAR] ;
	TEXTMETRIC		tm ;

	hSvrWnd	= (HWND)GetWindowLongPtr (hwnd,		FIGWL_UIMINIBUF_SVRWND) ;
	hIMC	= (HIMC)GetWindowLongPtr (hSvrWnd,	IMMGWLP_IMC) ;
	if (hIMC == NULL || lpIMC == NULL)
		return	FALSE ;

	hDC			= GetDC (hwnd) ;
	if (!hDC)
		return	FALSE ;

	hOldFont	= TSkkIme_hCheckNativeCharset (hDC) ;
	GetTextMetrics (hDC, &tm) ;

	nFrameWidth		= GetSystemMetrics (SM_CXDLGFRAME) * 2 ;
	nFrameHeight	= GetSystemMetrics (SM_CYDLGFRAME) * 2 ;
	nDefaultWidth	= tm.tmAveCharWidth * DEFAULT_MINIBUF_CHARWIDTH  + nFrameWidth ;
	nDefaultHeight	= tm.tmHeight + nFrameHeight ;

	nWidth	= nDefaultWidth ;
	nHeight	= nDefaultHeight ;

	dwLevel	= ImmGetGuideLine (hIMC, GGL_LEVEL, NULL, 0) ;
	if (dwLevel != 0) {
		dwSize	= ImmGetGuideLine (hIMC, GGL_STRING, NULL, 0) ;
		if (dwSize > 0) {
			long			lLength ;

			if (dwSize >= sizeof (bufGuide)) 
				dwSize	= sizeof (bufGuide) - sizeof (TCHAR) ;
			dwSize		= ImmGetGuideLine (hIMC, GGL_STRING, bufGuide, dwSize) ;
			lLength		= dwSize / sizeof (TCHAR) ;
			hOldFont	= TSkkIme_hCheckNativeCharset (hDC) ;

			if (lLength > 0) {
				LPCMYSTR	pText	= bufGuide ;
				int			nText	= lLength ;
				SIZE		sz ;

				nWidth	= (lpSize != NULL && nDefaultWidth < lpSize->cx)? lpSize->cx : nDefaultWidth ;
				nHeight	= 0 ;
				while (nText > 0) {
					int			nChar ;
					for (nChar = 1 ; nChar <= nText ; nChar ++) {
						MyGetTextExtentPoint (hDC, pText, nChar, &sz) ;
						/* FrameWidth ͕`̈ƐĂ͂Ȃ̂ŊOB*/
						if (sz.cx >= (nWidth - nFrameWidth)) 
							break ;
					}
					nChar	-- ;
					if (nChar <= 0)
						break ;
					nHeight	+= tm.tmHeight ;
					pText	+= nChar ;
					nText	-= nChar ;
				}
				nHeight	= (nHeight <= 0)? nDefaultHeight : (nHeight + nFrameHeight) ;
			}
		}
	}
	DeleteObject (SelectObject (hDC, hOldFont)) ;
	ReleaseDC (hwnd, hDC) ;
	if (lpSize) {
		DEBUGPRINTFEX (99, (TEXT ("uiMinibuf_bAdjustWindowSize (%dx%d <-> %dx%d)\n"), 
			lpSize->cx, lpSize->cy, nWidth, nHeight)) ;
		if (lpSize->cx < nWidth)
			lpSize->cx	= nWidth ;
		lpSize->cy	= nHeight ;
	}
	return	TRUE ;
}

