#if defined (UNITTEST)
#include "StdAfx.h"
#else
#include "local.h"
#endif
#include "ImeDoc.h"
#include "ImeBufferP.h"
#include "TMarker.h"
#include "RuleTreeNode.h"
#include "keymap.h"
#include "TMSG.h"
#include "TSearchSession.h"
#include "TJisyoUpdateSession.h"
#include "lmstate.h"
#include "jstring.h"
#include "ImeConfig.h"

/*=================================================================================
 *	prototypes
 */
static	int		LM_bSkkProcessPrefixOrSuffix	(struct CImeDoc*) ;
static	int		LM_bSkkUpdateJisyo				(struct CImeDoc*) ;
static	int		LM_bSkkModeExit					(struct CImeDoc*) ;
static	int		LM_bSkkChangeMarker				(struct CImeDoc*) ;
static	int		LM_bSkkHenkan1					(struct CImeDoc*) ;
static	int		LM_bSkkHenkanInMinibuff			(struct CImeDoc*) ;
static	int		LM_bSkkHenkanShowCandidates		(struct CImeDoc*) ;
static	int		LM_bSkkRemoveCommon				(struct CImeDoc*) ;


/*================================================================ skk-pre-command
 */
int
LM_bSkkPreCommand (
	struct CImeDoc*		pDoc)
{
	static BYTE	_srSkkKanaCleanupCommandList []	= {
		NFUNC_SKK_UNDO,
		NFUNC_SKK_KAKUTEI,
		NFUNC_SKK_DELETE_BACKWARD_CHAR,
		NFUNC_SKK_INSERT,
		NFUNC_SKK_PREVIOUS_CANDIDATE,
	} ;
	int		i ;

	/*
	 *	(defun skk-pre-command ()
	 *  (when (and (memq last-command
	 *		   '(skk-insert skk-previous-candidate))
	 *	     (null (memq this-command
	 *			 skk-kana-cleanup-command-list)))
	 *    (skk-kana-cleanup t)))
	*/
	if (ImeDoc_iGetLastCommand (pDoc) != NFUNC_SKK_INSERT && 
		ImeDoc_iGetLastCommand (pDoc) != NFUNC_SKK_PREVIOUS_CANDIDATE) {
		return	LMR_RETURN ;
	}
	for (i = 0 ; i < ARRAYSIZE (_srSkkKanaCleanupCommandList) ; i ++)
		if (_srSkkKanaCleanupCommandList [i] == ImeDoc_iGetThisCommand (pDoc))
			return	LMR_RETURN ;
	ImeDoc_vSetRegBool (pDoc, LMREGARG_0, TRUE) ;
	ImeDoc_vJump (pDoc, LM_bSkkKanaCleanup) ;
	return	LMR_CONTINUE ;
}

/*================================================================ skk-post-command
 */
int
LM_bSkkPostCommand (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int		nPrefix ;
	struct CSkkRuleTreeIterator*	piteSkkRuleTree ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	/*
	 * (defun skk-after-point-move ()
	 *	(when (and (not (and skk-previous-point
	 *				(= skk-previous-point (point))))
	 *				(skk-get-prefix skk-current-rule-tree))
	 *		(skk-with-point-move
	 *		(skk-erase-prefix 'clean))))
	 */
	piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;

	if (! (pBuffer->m_pmkSkkPreviousPoint != NULL &&
		TMarker_iGetPosition (pBuffer->m_pmkSkkPreviousPoint) == TMarker_iGetPosition (pBuffer->m_pmkPoint)) && 
		(! SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree) || SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) == NULL)) {
		imeBuffer_bSkkErasePrefix (pBuffer, TRUE) ;
		imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkPreviousPoint, MARKER_SKK_PREVIOUS_POINT, pBuffer->m_pmkPoint) ;
	}

	/*	completion ̉B̃^C~OŎsׂH
	 */
	if (pBuffer->m_pSkkCompSession != NULL && ImeDoc_iGetThisCommand (pDoc) != NFUNC_SKK_COMP_DO) {
		TSearchSession_vDestroy (pBuffer->m_pSkkCompSession) ;
		pBuffer->m_pSkkCompSession	= NULL ;
	}
	return	LMR_RETURN ;
}

/*================================================================ skk-insert
 */
static	int		LM_bSkkInsert_Exit				(struct CImeDoc*) ;

int
LM_bSkkInsert (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	BOOL	bRetval	= TRUE ;
	WCHAR	wch ;
	int		nPrefix ;
	struct CSkkRuleTreeIterator*	piteSkkRuleTree ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bGetLastCommandChar (pDoc, &wch)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;

	if (pBuffer->m_bSkkHenkanMode && ImeConfig_bSkkSpecialMidashiCharp (wch)) {
		if (! ImeDoc_bCall (pDoc, LM_bSkkProcessPrefixOrSuffix, LM_bSkkInsert_Exit))
			return	LMR_ERROR ;
	} else if (ImeConfig_bSkkSetHenkanPointKeyp (wch) &&
		(pBuffer->m_bSkkOkurigana || 
		//pBuffer->m_pSkkCurrentRuleTree == NULL ||
		 ! SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree) ||
		/*	(null (skk-get-prefix skk-current-rule-tree)) ̂́Askk-rule-tree 
		 *	root wĂꍇłB
		 */
		 SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) == NULL ||
		 ! SkkRuleTreeIterator_bHaveSelectBranchp (piteSkkRuleTree, wch))) {
//		TSkkRuleTreeNode_pGetPrefix (pBuffer->m_pSkkCurrentRuleTree, &nPrefix) == NULL || 
//		TSkkRuleTreeNode_pSelectBranch (pBuffer->m_pSkkCurrentRuleTree, wch) == NULL)) {
		if (! ImeDoc_bCall (pDoc, LM_bSkkSetHenkanPoint, LM_bSkkInsert_Exit))
			return	LMR_ERROR ;
	} else if (pBuffer->m_bSkkHenkanMode && ImeConfig_bSkkStartHenkanCharp (wch)) {
		if (! ImeDoc_bCall (pDoc, LM_bSkkStartHenkan, LM_bSkkInsert_Exit))
			return	LMR_ERROR ;
	} else if (pBuffer->m_bSkkHenkanMode != LON) {
		if (! ImeDoc_bCall (pDoc, LM_bSkkKanaInput, LM_bSkkInsert_Exit))
			return	LMR_ERROR ;
	} else if (pBuffer->m_bSkkHenkanMode == LON && ImeConfig_bSkkTryCompletionCharp (wch)) {
		/* completion ̊JnB*/
		ImeDoc_vSetRegBool (pDoc, LMREGARG_0, (ImeDoc_iGetLastCommand (pDoc) != NFUNC_SKK_COMP_DO)? TRUE : FALSE) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkCompDo, LM_bSkkInsert_Exit))
			return	LMR_ERROR ;
	} else if (pBuffer->m_bSkkHenkanMode == LON && 
		(ImeConfig_bSkkNextCompletionCharp (wch) || ImeConfig_bSkkPreviousCompletionCharp (wch)) &&
		ImeDoc_iGetLastCommand (pDoc) == NFUNC_SKK_COMP_DO) {
		/* completion ̑B*/
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_0, wch) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkCompPreviousNext, LM_bSkkInsert_Exit))
			return	LMR_ERROR ;
	} else {
		/* ȊOB*/
		if (! ImeDoc_bCall (pDoc, LM_bSkkKanaInput, LM_bSkkInsert_Exit))
			return	LMR_ERROR ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkInsert_Exit (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	/* skk-point-move ̌ʁB*/
	imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkPreviousPoint, MARKER_SKK_PREVIOUS_POINT, pBuffer->m_pmkPoint) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}


/*================================================================ skk-process-prefix-or-suffix
 */
static	int		LM_bSkkProcessPrefixOrSuffix_1	(struct CImeDoc*) ;
static	int		LM_bSkkProcessPrefixOrSuffix_2	(struct CImeDoc*) ;
static	int		LM_bSkkProcessPrefixOrSuffix_3	(struct CImeDoc*) ;

int
LM_bSkkProcessPrefixOrSuffix (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	WCHAR	wch ;

	if (ImeDoc_bSignalp (pDoc)) {
		return	LMR_RETURN ;
	}

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkProcessPrefixOrSuffix_1))
			return	LMR_ERROR ;
		return	LMR_CONTINUE ;
	} else if (pBuffer->m_bSkkHenkanMode == LON) {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_0, TRUE) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKanaCleanup, LM_bSkkProcessPrefixOrSuffix_3))
			return	LMR_ERROR ;
		return	LMR_CONTINUE ;
	} else if (ImeDoc_bGetLastCommandChar (pDoc, &wch)) {
		LPWSTR	pwString ;

		pwString	= ImeDoc_pAlloca (pDoc, sizeof (WCHAR), __alignof (WCHAR)) ;
		if (pwString == NULL)
			return	LMR_ERROR ;
		*pwString	= wch ;
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwString, 1) ;
		ImeDoc_vJump (pDoc, LM_bSkkInsertStr) ;
		return	LMR_CONTINUE ;
	} else {
		ImeDoc_vJump (pDoc, LM_bSkkEmulateOriginalMap) ;
		return	LMR_CONTINUE ;
	}
}

int
LM_bSkkProcessPrefixOrSuffix_1 (
	struct CImeDoc*		pDoc)
{
	if (! ImeDoc_bCall (pDoc, LM_bSkkSetHenkanPointSubr, LM_bSkkProcessPrefixOrSuffix_2))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkProcessPrefixOrSuffix_2 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) {
		return	LMR_RETURN ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	ImeBuffer_bInsertAndInherit (pBuffer, L">", 1) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

int
LM_bSkkProcessPrefixOrSuffix_3 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int		nPoint, nHenkanStartPoint ;

	if (ImeDoc_bSignalp (pDoc)) {
		return	LMR_RETURN ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	ImeBuffer_bInsertAndInherit (pBuffer, L">", 1) ;
	imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkHenkanEndPoint, MARKER_SKK_HENKAN_END_POINT, pBuffer->m_pmkPoint) ;
	pBuffer->m_iSkkHenkanCount	= 0 ;
	nPoint				= (pBuffer->m_pmkPoint != NULL)? TMarker_iGetPosition (pBuffer->m_pmkPoint) : 0 ;
	nHenkanStartPoint	= (pBuffer->m_pmkSkkHenkanStartPoint != NULL)? TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) : 0 ;
	pBuffer->m_nSkkHenkanKeyLen	= nPoint - nHenkanStartPoint ;
	if (pBuffer->m_nSkkHenkanKeyLen > 0) {
		memcpy (pBuffer->m_bufSkkHenkanKey, pBuffer->m_bufComp + nHenkanStartPoint, (nPoint - nHenkanStartPoint) * sizeof (WCHAR));
	} else {
		pBuffer->m_nSkkHenkanKeyLen	= 0 ;
	}
	pBuffer->m_nSkkPrefixLen	= 0 ;
	pBuffer->m_bSkkAfterPrefix	= LTRUE ;
	ImeDoc_vJump (pDoc, LM_bSkkHenkan) ;
	return	LMR_CONTINUE ;
}

/*================================================================ skk-kana-input
 */
static	int		LM_bSkkKanaInput_1		(struct CImeDoc*) ;
static	int		LM_bSkkKanaInput_1_1	(struct CImeDoc*) ;
static	int		LM_bSkkKanaInput_3		(struct CImeDoc*) ;
static	int		LM_bSkkKanaInput_3_1	(struct CImeDoc*) ;
static	int		LM_bSkkKanaInput_3_2	(struct CImeDoc*) ;
static	int		LM_bSkkKanaInput_4		(struct CImeDoc*) ;
static	int		LM_bSkkKanaInput_4_0	(struct CImeDoc*) ;
static	int		LM_bSkkKanaInput_4_0_1	(struct CImeDoc*) ;
static	int		LM_bSkkKanaInput_4_1	(struct CImeDoc*) ;
static	int		LM_bSkkKanaInput_Exit	(struct CImeDoc*) ;

/*	LMREG_0				nQueueLen
 *	LMREG_1				pwQueue
 *	LMREG_2				pData
 *	LMREG_3				pNext
 */
int
LM_bSkkKanaInput (
	struct CImeDoc*		pDoc)
{
	WCHAR*		pwQueue ;
	LPWSTR		pwBuffer ;
	int			nQueue ;

	pwQueue	= ImeDoc_pAlloca (pDoc, 64 * sizeof (WCHAR), __alignof (WCHAR)) ;
	if (pwQueue == NULL)
		return	LMR_ERROR ;

	pwBuffer	= ImeDoc_pAlloca (pDoc, sizeof (WCHAR) * MAXCOMPLEN, __alignof (WCHAR)) ;
	if (pwBuffer == NULL)
		return	LMR_ERROR ;

	if (! ImeDoc_bPushReg (pDoc, LMREG_0) || ! ImeDoc_bPushReg (pDoc, LMREG_1) ||
		! ImeDoc_bPushReg (pDoc, LMREG_2) || ! ImeDoc_bPushReg (pDoc, LMREG_3) ||
		! ImeDoc_bPushReg (pDoc, LMREG_4) || ! ImeDoc_bPushReg (pDoc, LMREG_5))
		return	LMR_ERROR ;

	nQueue				= 0 ;
	pwQueue [nQueue ++]	= ImeDoc_iGetLastCommandChar (pDoc) ;
	ImeDoc_vSetRegInteger	(pDoc, LMREG_0, nQueue) ;
	ImeDoc_vSetRegString	(pDoc, LMREG_1, pwQueue,	64) ;
	ImeDoc_vSetRegString	(pDoc, LMREG_5, pwBuffer,	MAXCOMPLEN) ;
	ImeDoc_vJump (pDoc, LM_bSkkKanaInput_1) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKanaInput_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	WCHAR*		pwQueue ;
	int			nQueueLen, nQueueSize ;
//	LPCWSTR		pNextState ;
	LPCWSTR		pPrefix ;
	int			nNextState, i, nPrefix ;
	WCHAR		bufNextState [64] ;
	WCHAR		bufTempNextState [16] ;
	struct CSkkRuleTreeNode*	pNext ;
	const struct CSkkRuleTreeNodeOutput*	pData	= NULL ;
	struct CSkkRuleTreeIterator*	piteSkkRuleTree ;

	/* Exit-State ͕KoRȂƂȂB*/
	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_0, &nQueueLen))
		nQueueLen	= 0 ;
	if (! ImeDoc_bGetRegString  (pDoc, LMREG_1, &pwQueue, &nQueueSize)) {
		pwQueue		= NULL ;
		nQueueSize	= 0 ;
	}
	if (nQueueLen <= 0) {
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}

	piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;
	if (! SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree) || SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) == NULL) {
//	if (pBuffer->m_pSkkCurrentRuleTree == NULL || TSkkRuleTreeNode_pGetPrefix (pBuffer->m_pSkkCurrentRuleTree, &nPrefix) == NULL) {
		/* ̏ꏊ̊JnʒuƂB*/
		imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkKanaStartPoint, MARKER_SKK_KANA_START_POINT, pBuffer->m_pmkPoint) ;
		/* Rule-Tree ̍ɖ߂B*/
		SkkRuleTreeIterator_vMoveToRoot (piteSkkRuleTree) ;
//		pBuffer->m_pSkkCurrentRuleTree	= ImeConfig_pGetSkkRuleTree () ;
		pPrefix	= NULL ;
		nPrefix	= 0 ;
	} else {
		imeBuffer_bSkkErasePrefix (pBuffer, FALSE) ;
		pPrefix	= SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) ;
	}
	/* skk-prefix  skk-get-prefix + last-command-char */
	if ((nPrefix + 1) > MAXLEN_PREFIX) {
		/* vtBNXobt@[ꂽꍇcB*/
		imeBuffer_bSkkErasePrefix (pBuffer, FALSE) ;
		pBuffer->m_nSkkPrefixLen		= 0 ;
		SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
//		pBuffer->m_pSkkCurrentRuleTree	= NULL ;
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (nPrefix > 0) 
		memcpy (pBuffer->m_bufSkkPrefix, pPrefix, sizeof (WCHAR) * nPrefix) ;
	pBuffer->m_bufSkkPrefix [nPrefix]	= ImeDoc_iGetLastCommandChar (pBuffer->m_pDoc) ;
	pBuffer->m_nSkkPrefixLen			= nPrefix + 1 ;

//	pNext	= TSkkRuleTreeNode_pSelectBranch (pBuffer->m_pSkkCurrentRuleTree, pwQueue [0]) ;
	pData	= NULL ;
	if (SkkRuleTreeIterator_bHaveSelectBranchp (piteSkkRuleTree, pwQueue [0])) {
//	if (pNext != NULL) {
		if (SkkRuleTreeIterator_bNextHasBranchListp (piteSkkRuleTree, pwQueue [0])) {
			int		iCH ;
//		if (TSkkRuleTreeNode_pGetBranchList (pNext) != NULL) {
			if (pBuffer->m_bSkkHenkanMode == LACTIVE &&
				ImeConfig_bSkkKakuteiEarlyp () && !ImeConfig_bSkkProcessOkuriEarlyp ()) {
				/* !! */
				ImeDoc_vSetRegConstString	(pDoc, LMREGARG_0, NULL, 0) ;
				ImeDoc_vSetRegInteger		(pDoc, LMREG_0, nQueueLen) ;
				ImeDoc_vSetRegConstPointer	(pDoc, LMREG_2, pData) ;
				ImeDoc_vSetRegInteger		(pDoc, LMREG_3, pwQueue [0]) ;
//				ImeDoc_vSetRegPointer		(pDoc, LMREG_3, pNext) ;
				if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkKanaInput_1_1))
					return	LMR_ERROR ;
				return	LMR_CONTINUE ;
			}
			iCH	= pwQueue [0] ;
			memmove (pwQueue, pwQueue + 1, sizeof (pwQueue [0]) * (nQueueLen - 1)) ;
			nQueueLen	-- ;
			SkkRuleTreeIterator_vWalk (piteSkkRuleTree, iCH) ;
//			pBuffer->m_pSkkCurrentRuleTree	= pNext ;
		} else {
//			LPCWSTR	pNextState ;
			int		nNextState, i, nNextTree ;

			SkkRuleTreeIterator_vWalk (piteSkkRuleTree, pwQueue [0]) ;
			pData		= SkkRuleTreeIterator_pGetOutput (piteSkkRuleTree) ;
			nNextState	= SkkRuleTreeIterator_iGetNextState (piteSkkRuleTree, bufNextState, ARRAYSIZE (bufNextState), &nNextTree) ;
			SkkRuleTreeIterator_vMoveTree (piteSkkRuleTree, nNextTree) ;
//			pData		= TSkkRuleTreeNode_pGetOutput (pNext) ;
//			pNextState	= TSkkRuleTreeNode_pGetNextState (pNext, &nNextState) ;
			if ((nNextState + nQueueLen - 1) < nQueueSize) {
				memmove (pwQueue + nNextState, pwQueue + 1, sizeof (pwQueue [0]) * (nQueueLen - 1)) ;
				for (i = 0 ; i < nNextState ; i ++)
					pwQueue [i]	= bufNextState [i] ;
				nQueueLen	= nNextState + nQueueLen - 1 ;
			} else {
				/* L[ꂽꍇcB̏̂ǂ͓B*/
				SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
//				pBuffer->m_pSkkCurrentRuleTree	= NULL ;

				/* exit state ցB*/
				ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
				return	LMR_CONTINUE ;
			}
			SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
//			pBuffer->m_pSkkCurrentRuleTree	= NULL ;
		}
	} else {
		const struct CSkkRuleTreeNodeOutput*	pD ;

		pD	= SkkRuleTreeIterator_pGetOutput (piteSkkRuleTree) ;
//		pD	= pBuffer->m_pSkkCurrentRuleTree != NULL? TSkkRuleTreeNode_pGetOutput (pBuffer->m_pSkkCurrentRuleTree) : NULL ;
		if (pD != NULL) {
			int		nNextTree ;

			nNextState	= 0 ;
			pData		= pD ;
			if (SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree)) {
				nNextState	= SkkRuleTreeIterator_iGetNextState (piteSkkRuleTree, bufNextState, ARRAYSIZE (bufNextState), &nNextTree) ;
			} else {
				nNextTree	= SkkRuleTreeIterator_iGetRuleTree (piteSkkRuleTree) ;
			}
			SkkRuleTreeIterator_vMoveTree (piteSkkRuleTree, nNextTree) ;

//			pNextState	= NULL ;
//			if (pBuffer->m_pSkkCurrentRuleTree != NULL) 
//				pNextState	= TSkkRuleTreeNode_pGetNextState (pBuffer->m_pSkkCurrentRuleTree, &nNextState) ;
			if ((nNextState + nQueueLen) < nQueueSize) {
				if (nNextState > 0) {
					memmove (pwQueue + nNextState, pwQueue, sizeof (pwQueue [0]) * nQueueLen) ;
					for (i = 0 ; i < nNextState ; i ++)
						pwQueue [i]	= bufNextState [i] ;
					nQueueLen	= nNextState + nQueueLen ;
				}
			} else {
				/* L[ꂽꍇcB̏̂ǂ͓B*/
				SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
//				pBuffer->m_pSkkCurrentRuleTree	= NULL ;

				/* exit state ցB*/
				ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
				return	LMR_CONTINUE ;
			}
			SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
//			pBuffer->m_pSkkCurrentRuleTree	= NULL ;
		} else {
#if 0
			const struct CSkkRuleTreeNodeOutput*	pDD	= NULL ;

			/*	skk-kana-input-search-function  ``oh'' ϊ邩ۂ check ɕύXB
			 *	Cӂ lisp function s`Ƃ͔̂B
			 *	 oh ɂĂ preceding-char KvƂ錵@ł邽߁cȂ
			 *	B
			 *	(defcustom skk-kana-input-search-function
			 *	  #'(lambda ()
			 *		  (save-match-data
			 *		(and (string-match "^h\\([bcdfghjklmnpqrstvwxz]\\)$" skk-prefix)
			 *			 (member (char-to-string (preceding-char)) '("" "I"))
			 *			 (cons '("I" . "") (match-string 1 skk-prefix)))))
			 *
			 *	 ad hoc Ȍ` code ɖߍłcB
			 */
			if (TRUE && 
				pBuffer->m_nSkkPrefixLen == 2 && pBuffer->m_bufSkkPrefix [0] == L'h' && (L'b' <= pBuffer->m_bufSkkPrefix [1] && pBuffer->m_bufSkkPrefix [1] <= L'z') &&
				pBuffer->m_bufSkkPrefix [1] != L'e' && pBuffer->m_bufSkkPrefix [1] != L'i' && pBuffer->m_bufSkkPrefix [1] != L'o' && pBuffer->m_bufSkkPrefix [1] != L'u') {
				/*	 preceding-char 邱ƂɂȂB
				 */
				int		nPoint			= TMarker_iGetPosition (pBuffer->m_pmkPoint) ;
				int		nPrecedingChar	= -1 ;

				if (0 < nPoint && nPoint <= pBuffer->m_nbufComp) {
					nPrecedingChar	= pBuffer->m_bufComp [nPoint - 1] ;
				} else if (nPoint == 0) {
					/*	OɃAvP[Vɓn̍ŌƂƂB
					 *	ÁuOɁvNA@́cpӂ؂̂낤H
					 */
					nPrecedingChar	= (pBuffer->m_nLatestShiftTextLen > 0)? pBuffer->m_bufLatestShiftText [pBuffer->m_nLatestShiftTextLen - 1] : -1 ;
				}
				if (nPrecedingChar == L'' || nPrecedingChar == L'I') {
					pDD			= ImeConfig_pGetOhRuleTreeNodeOutput () ;
					/*pDD		= TSkkRuleTreeNodeOutput_pCreateStringPairInstance (L"", 1, L"I", 1) ;*/

					pNextState	= bufTempNextState ;
					nNextState	= 1 ;
					bufTempNextState [0]	= pBuffer->m_bufSkkPrefix [1] ;
				}
			} else {
				pDD	= NULL ;
			}
			if (pDD != NULL) {
				/*					  (setq data (car dd)
										queue (nconc (string-to-char-list (cdr dd))
													 (cdr queue))
				 */
				if ((nNextState + nQueueLen - 1) < nQueueSize) {
					memmove (pwQueue + nNextState, pwQueue + 1, sizeof (pwQueue [0]) * (nQueueLen - 1)) ;
					for (i = 0 ; i < nNextState ; i ++)
						pwQueue [i]	= pNextState [i] ;
					nQueueLen	= nQueueLen + nNextState - 1 ;
				} else {
					SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
//					pBuffer->m_pSkkCurrentRuleTree	= NULL ;

					/* exit state ցB*/
					ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
					return	LMR_CONTINUE ;
				}
				pData							= pDD ;
				SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
//				pBuffer->m_pSkkCurrentRuleTree	= NULL ;
			} else if (SkkRuleTreeIterator_bRootp (piteSkkRuleTree)) {
//			} else if (pBuffer->m_pSkkCurrentRuleTree == ImeConfig_pGetSkkRuleTree ()) {
				nQueueLen						= 0 ;
				SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
//				pBuffer->m_pSkkCurrentRuleTree	= NULL ;
			} else {
				SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
//				pBuffer->m_pSkkCurrentRuleTree	= NULL ;
			}
#else
			/*	oh Ȃ΁ApDD == NULL ͕KB
			 */
			if (SkkRuleTreeIterator_bRootp (piteSkkRuleTree)) {
				nQueueLen						= 0 ;
				SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
			} else {
				SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
			}
#endif
		}
	}


	ImeDoc_vSetRegInteger		(pDoc, LMREG_0, nQueueLen) ;
	ImeDoc_vSetRegConstPointer	(pDoc, LMREG_2, pData) ;

	/* pData nȂƂȂB*/
	ImeDoc_vJump (pDoc, LM_bSkkKanaInput_3) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKanaInput_1_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	WCHAR*		pwQueue ;
	int			nQueueLen, nQueueSize ;
	struct CSkkRuleTreeIterator*	piteSkkRuleTree ;
	int			wch ;
//	struct CSkkRuleTreeNode*	pNext ;

	/* Exit-State ͕KoRȂƂȂB*/
	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_0, &nQueueLen))
		nQueueLen	= 0 ;
	if (! ImeDoc_bGetRegString  (pDoc, LMREG_1, &pwQueue, &nQueueSize)) {
		pwQueue		= NULL ;
		nQueueSize	= 0 ;
	}

	piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;
	if (ImeDoc_bGetRegInteger (pDoc, LMREG_3, &wch)) {
		SkkRuleTreeIterator_vWalk (piteSkkRuleTree, wch) ;
	} else {
		SkkRuleTreeIterator_vReset (piteSkkRuleTree) ;
	}
//	if (! ImeDoc_bGetRegPointer (pDoc, LMREG_3, &pNext)) 
//		pNext	= NULL ;

	/* !! */
	imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkKanaStartPoint, MARKER_SKK_KANA_START_POINT, pBuffer->m_pmkPoint) ;

	memmove (pwQueue, pwQueue + 1, sizeof (pwQueue [0]) * (nQueueLen - 1)) ;
	nQueueLen	-- ;
//	pBuffer->m_pSkkCurrentRuleTree	= pNext ;

	ImeDoc_vSetRegInteger (pDoc, LMREG_0, nQueueLen) ;
	ImeDoc_vJump (pDoc, LM_bSkkKanaInput_3) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKanaInput_3 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*		pBuffer ;
	int						nQueueLen ;
	LPCWSTR					pPrefix ;
	int						nPrefix ;
	struct CSkkRuleTreeIterator*			piteSkkRuleTree ;
	const struct CSkkRuleTreeNodeOutput*	pData	= NULL ;

	/* Exit-State ͕KoRȂƂȂB*/
	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;

	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_0, &nQueueLen))
		nQueueLen	= 0 ;
	if (! ImeDoc_bGetRegConstPointer (pDoc, LMREG_2, &pData))
		pData	= NULL ;

	if (pData == NULL) {
		pPrefix	= SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) ;
//		pPrefix	= (pBuffer->m_pSkkCurrentRuleTree != NULL)? TSkkRuleTreeNode_pGetPrefix (pBuffer->m_pSkkCurrentRuleTree, &nPrefix) : NULL ;
		if (pPrefix != NULL && nPrefix > 0) {
			memcpy (pBuffer->m_bufSkkPrefix, pPrefix, sizeof (WCHAR) * nPrefix) ;
			pBuffer->m_nSkkPrefixLen	= nPrefix ;
			imeBuffer_bSkkInsertPrefix (pBuffer, pBuffer->m_bufSkkPrefix, pBuffer->m_nSkkPrefixLen) ;
		} else {
			if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
				/* !! */
				ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
				if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkKanaInput_3_1))
					return	LMR_ERROR ;
				/* !! */
			} else {
				ImeDoc_vJump (pDoc, LM_bSkkKanaInput_3_1) ;
			}
			return	LMR_CONTINUE ;
		}
		//ImeDoc_vJump (pDoc, LM_bSkkKanaInput_1) ;
		//return	LMR_CONTINUE ;
	} else {
		int		nType ;

		pBuffer->m_nSkkPrefixLen	= 0 ;
		/*
		 *	  (when (functionp data)
		 *		(setq data (funcall data (skk-make-raw-arg arg))))
		 */
		/*	ǂĝYނǁAdata  stringp ̂Ƃ̎ł͍lȂB
		 *	Ŕc炻̃R[hɂ́A񂪑ȂB܂̂Aǂ
		 *	ǂ̂AォēxB
		 */
		nType	= TSkkRuleTreeNodeOutput_iGetType (pData) ;
		if (nType == NTYPE_MACRO) {
			int		nFuncNo ;

			/* cB*/
			if (TSkkRuleTreeNodeOutput_bGetMacro (pData, &nFuncNo)) {
				PLMFUNC	pPC ;
				/* !! */
				pPC		= ImeDoc_pLookupLMState (pDoc, nFuncNo) ;
				if (pPC != NULL) {
					/*	argument  clear ĂKvƎvcB
					 *	Aarg ͓IɂgȂ call-interactive  function
					 *	 nil Ǝvł邵vH
					 *	(Fri Jun 09 12:01:33 2006)
					 */
					if (! ImeDoc_bCall (pDoc, pPC, LM_bSkkKanaInput_3_2))
						return	LMR_ERROR ;
					return	LMR_CONTINUE ;
				}
				/* !! */
			}
			/*	II  loop ̓ɖ߂̂͗ǂȂ̂ȂǁcB
			 *	 '("" . "") Ȍʂ macro ԂƂĂȂ̂Łc
			 *	AԂlp̃WX^ƂpӂȂƖ{͑ʖڂcB
			 *	ƂȂƁALMR_RETURN 鎞 set ȂƂȂȁcB
			 */
			ImeDoc_vJump (pDoc, LM_bSkkKanaInput_1) ;
			return	LMR_CONTINUE ;
		}
		if (nType == NTYPE_STRINGPAIR || nType == NTYPE_STRING) {
			LPCWSTR		wptr ;
			int			nwptr ;

			if (nType == NTYPE_STRINGPAIR) {
				if (pBuffer->m_bSkkKatakana) {
					if (! TSkkRuleTreeNodeOutput_bGetRString (pData, &wptr, &nwptr)) {
						wptr	= NULL ;
						nwptr	= 0 ;
					}
				} else {
					if (! TSkkRuleTreeNodeOutput_bGetLString (pData, &wptr, &nwptr)) {
						wptr	= NULL ;
						nwptr	= 0 ;
					}
				}
			} else {
				if (! TSkkRuleTreeNodeOutput_bGetString (pData, &wptr, &nwptr)) {
					wptr	= NULL ;
					nwptr	= 0 ;
				}
			}
			/* insert-paren  through cB*/
			if (pBuffer->m_bSkkHenkanMode == LACTIVE &&
				ImeConfig_bSkkKakuteiEarlyp () && !ImeConfig_bSkkProcessOkuriEarlyp ()) {
				/* !! */
				ImeDoc_vSetRegString (pDoc, LMREGARG_0, NULL, 0) ;
				/* wptr, nwptr  heap ̏wĂ̂ŁAł͉ȂƍlėǂB*/
				ImeDoc_vSetRegConstString (pDoc, LMREG_4, wptr, nwptr) ;
				if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkKanaInput_4))
					return	LMR_ERROR ;
				return	LMR_CONTINUE ;
				/* !! */
			}
			ImeDoc_vSetRegConstString (pDoc, LMREG_4, wptr, nwptr) ;
			ImeDoc_vJump (pDoc, LM_bSkkKanaInput_4) ;
			return	LMR_CONTINUE ;
		}
	}
	/*
	 */
	ImeDoc_vJump (pDoc, LM_bSkkKanaInput_1) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKanaInput_3_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	int						nQueueLen ;

	/* Exit-State ͕KoRȂƂȂB*/
	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}

	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_0, &nQueueLen))
		nQueueLen	= 0 ;

	pBuffer->m_nSkkPrefixLen	= 0 ;
	if (! (nQueueLen > 0 || (ImeDoc_iGetThisCommand (pBuffer->m_pDoc) != NFUNC_SKK_INSERT && pBuffer->m_bSkkHenkanMode))) {
		/* !! */
		if (! ImeDoc_bCall (pDoc, LM_bSkkEmulateOriginalMap, LM_bSkkKanaInput_1))
			return	LMR_ERROR ;
		/* !! */
		return	LMR_CONTINUE ;
	}
	/* while (nQueueLen > 0) ̃[v̓ւƖ߂B*/
	ImeDoc_vJump (pDoc, LM_bSkkKanaInput_1) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKanaInput_3_2 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	LPWSTR					pwBuffer ;
	LPWSTR					pwString = NULL, pwLeft = NULL, pwRight = NULL ;
	LPCWSTR					pwResult ;
	int						nString, nLeft, nRight, n, nResultLen, nBufferSize ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}

	pwResult	= NULL ;
	nResultLen	= 0 ;
	if (! ImeDoc_bGetRegString (pDoc, LMREG_5, &pwBuffer, &nBufferSize)) {
		pwBuffer	= NULL ;
		nBufferSize	= 0 ;
	}

	/*	͂́upwStringv͒Nmۂ̈Ȃ̂ƂƂB̊֐OɗpӂĂ
	 *	Ƃ͓̂Bargument ɓƂ̂́uCӂ function Ă΂邩v̏󋵂ł
	 *	ɂȂBheap ̂悤Ȃ̂pӂKv̂낤cB
	 *
	 *	return buffer Ƃ̂pӂ邩BRETVAL ̊gBLMREGARG_RETVAL Ɠx̊Ԓx
	 *	ێȂ funcall Ă̎󂯓n̂߂ɂg̈B
	 *	Areturn Ė߂͑ my heap  my stack ɃRs[ăf[^ی삵ȂƂ
	 *	ȂB
	 *	return buffer  document class 1ĂB
	 */
	if (ImeDoc_bGetRegConstString (pDoc, LMREGARG_RETVAL, &pwString, &nString)) {
		/*	 pwString  m_bufRETURN wĂ̂ŁAFREE ͂ȂŗǂB(Ƃɂ)
		 */
		if (pwString != NULL && nString > 0) {
			n			= MIN (nString, nBufferSize) ;
			memcpy (pwBuffer, pwString, n * sizeof (WCHAR)) ;
			pwResult	= pwBuffer ;
			nResultLen	= n ;
		}
	} else if (ImeDoc_bGetRegConstStringPair (pDoc, LMREGARG_RETVAL, &pwLeft, &nLeft, &pwRight, &nRight)) {
		/*	 pwLeft, pwRight  m_bufRETURN wĂ̂ŁAFREE ͂ȂŗǂB(Ƃ
		 *	)
		 */
		pwString	= (pBuffer->m_bSkkKatakana)? pwRight : pwLeft ;
		nString		= (pBuffer->m_bSkkKatakana)? nLeft   : nRight ;
		if (pwString != NULL && nString > 0) {
			n			= MIN (nString, nBufferSize) ;
			memcpy (pwBuffer, pwString, n * sizeof (WCHAR)) ;
			pwResult	= pwBuffer ;
			nResultLen	= n ;
		}
	} else {
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_1) ;
		return	LMR_CONTINUE ;
	}
	/* insert-paren  through cB*/
	if (pBuffer->m_bSkkHenkanMode == LACTIVE &&
		ImeConfig_bSkkKakuteiEarlyp () && !ImeConfig_bSkkProcessOkuriEarlyp ()) {
		ImeDoc_vSetRegString (pDoc, LMREGARG_0, NULL, 0) ;
		/* wptr, nwptr  heap ̏wĂ̂ŁAł͉ȂƍlėǂB*/
		ImeDoc_vSetRegConstString (pDoc, LMREG_4, pwString, nString) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkKanaInput_4))
			return	LMR_ERROR ;
	} else {
		ImeDoc_vSetRegConstString (pDoc, LMREG_4, pwString, nString) ;
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_4) ;
	}
	return	LMR_CONTINUE ;
}

/*	skk-kana-input  skk-insert-str ̏Bdata m肵 buffer ɕ񂪑}B
 */
int
LM_bSkkKanaInput_4 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	LPCWSTR					wptr ;
	int						nwptr ;

	/* Exit-State ͕KoRȂƂȂB*/
	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}

	if (! ImeDoc_bGetRegConstString (pDoc, LMREG_4, &wptr, &nwptr)) {
		wptr	= NULL ;
		nwptr	= 0 ;
	}
	ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, wptr, nwptr) ;
	if (! ImeDoc_bCall (pDoc, LM_bSkkInsertStr, LM_bSkkKanaInput_4_0))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

/*	skk-kana-input rł skk-auto-insert-paren ̏Bbuffer ɑ}ꂽāAparen
 *	Ȃ pair ǉB
 */
int
LM_bSkkKanaInput_4_0 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	LPCWSTR					wptr ;
	int						nwptr ;

	/* Exit-State ͕KoRȂƂȂB*/
	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}

	if (! ImeDoc_bGetRegConstString (pDoc, LMREG_4, &wptr, &nwptr)) {
		wptr	= NULL ;
		nwptr	= 0 ;
	}
	/*	skk-auto-insert-paren ̏B
	 */
	if (nwptr > 0 && ImeConfig_bSkkAutoInsertParen ()) {
		WCHAR	bufPairStr [MAXCOMPLEN] ;
		int		nPairStr ;

			/*
		      (when pair
				(while (> count1 0)
				  (if (not (string= pair (char-to-string (following-char))))
				      (progn
						(setq inserted (1+ inserted))
						(skk-insert-str pair)))
				  (setq count1 (1- count1)))
				(unless (= inserted 0)
				  (backward-char inserted)))
				  */
		nPairStr	= ImeConfig_iSkkAssocSkkAutoParenStringAlist (wptr, nwptr, bufPairStr, ARRAYSIZE (bufPairStr)) ;
		if (nPairStr > 0) {
			if (! ImeBuffer_bFollowingStringEqualp (pBuffer, bufPairStr, nPairStr)) {
				LPWSTR	pwBuffer ;
				int		n, nBufferSize ;

				/*	LMREG_5 ɊmۂĂobt@𗘗pB
				 *	́Aparen Ɏg result ێĂꏊł邪A
				 *	 result ͎gȂ̂ŁB
				 */
				if (! ImeDoc_bGetRegString (pDoc, LMREG_5, &pwBuffer, &nBufferSize)) {
					pwBuffer	= NULL ;
					nBufferSize	= 0 ;
				}
				n	= MIN (nBufferSize, nPairStr) ;
				if (n > 0)
					memcpy (pwBuffer, bufPairStr, n * sizeof (WCHAR)) ;
				ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwBuffer, n) ;

				/*	݂̃J[\ʒuLcB{̓}[JĝcX^bN
				 *	}[JƉł邩ǂ̂ŁB(G[AO̓r)
				 *
				 *	 REG_4  integer ɕςĂ܂ĂB
				 */
				ImeDoc_vSetRegInteger (pDoc, LMREG_4, TMarker_iGetPosition (pBuffer->m_pmkPoint)) ;

				if (! ImeDoc_bCall (pDoc, LM_bSkkInsertStr, LM_bSkkKanaInput_4_0_1))
					return	LMR_ERROR ;
				return	LMR_CONTINUE ;
			}
		}
	}
	ImeDoc_vJump (pDoc, LM_bSkkKanaInput_4_1) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKanaInput_4_0_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	int	nPoint ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL && ImeDoc_bGetRegInteger (pDoc, LMREG_4, &nPoint)) {
		int		nCurPoint ;
		nCurPoint	= TMarker_iGetPosition (pBuffer->m_pmkPoint) ;
		ImeBuffer_bBackwardChar (pBuffer, nCurPoint - nPoint) ;
	}
	ImeDoc_vJump (pDoc, LM_bSkkKanaInput_4_1) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKanaInput_4_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	int						nQueueLen ;

	/* Exit-State ͕KoRȂƂȂB*/
	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_Exit) ;
		return	LMR_CONTINUE ;
	}

	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_0, &nQueueLen))
		nQueueLen	= 0 ;

	if (pBuffer->m_bSkkOkurigana && nQueueLen <= 0) {
		/* !! */
		if (! ImeDoc_bCall (pDoc, LM_bSkkSetOkuriganaAdJisx0201, LM_bSkkKanaInput_1))
			return	LMR_ERROR ;
		/* !! */
	} else {
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput_1) ;
	}
	return	LMR_CONTINUE ;
}


int
LM_bSkkKanaInput_Exit (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
	} else {
		/* skk-point-move ̌ʁB*/
		imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkPreviousPoint, MARKER_SKK_PREVIOUS_POINT, pBuffer->m_pmkPoint) ;
	}

	/*  pop  signal ̗LɊւ炸sȂ΂ȂȂB*/
	ImeDoc_vPopReg (pDoc, LMREG_5) ;
	ImeDoc_vPopReg (pDoc, LMREG_4) ;
	ImeDoc_vPopReg (pDoc, LMREG_3) ;
	ImeDoc_vPopReg (pDoc, LMREG_2) ;
	ImeDoc_vPopReg (pDoc, LMREG_1) ;
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-start-henkan
 */
static	int	LM_bSkkStartHenkan_1	(struct CImeDoc*) ;
static	int	LM_bSkkStartHenkan_2	(struct CImeDoc*) ;

int
LM_bSkkStartHenkan (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
		pBuffer->m_iSkkHenkanCount	++ ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkan) ;
	} else {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_0, TRUE) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKanaCleanup, LM_bSkkStartHenkan_1))
			return	LMR_ERROR ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkStartHenkan_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int		nPrefix, nPos, nHenkanStartPoint ;
	struct CSkkRuleTreeIterator*	piteSkkRuleTree ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;
	if (SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree) && SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) != NULL) {
//	if (pBuffer->m_pSkkCurrentRuleTree != NULL && TSkkRuleTreeNode_pGetPrefix (pBuffer->m_pSkkCurrentRuleTree, &nPrefix) != NULL) {
		/* skk-error "Have unfixed skk-prefix" */
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	nPos	= TMarker_iGetPosition (pBuffer->m_pmkPoint) ;
	if (pBuffer->m_pmkSkkHenkanStartPoint == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	nHenkanStartPoint	= TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) ;
	if (nPos < nHenkanStartPoint) {
		/* skk-error "Henkan end point must be after henkan start point" */
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	pBuffer->m_nSkkHenkanKeyLen	= nPos - nHenkanStartPoint ;
	memcpy (pBuffer->m_bufSkkHenkanKey, pBuffer->m_bufComp + nHenkanStartPoint, pBuffer->m_nSkkHenkanKeyLen * sizeof (WCHAR)) ;
	if (pBuffer->m_bSkkKatakana) {
		pBuffer->m_nSkkHenkanKeyLen	= iSkkKatakanaToHiragana (pBuffer->m_bufSkkHenkanKey, ARRAYSIZE (pBuffer->m_bufSkkHenkanKey), pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen, FALSE) ;
	}
	if (pBuffer->m_bSkkOkurigana) {
		int		i ;

		for (i = pBuffer->m_nSkkHenkanKeyLen - 1 ; i >= 0 ; i --) {
			if (pBuffer->m_bufSkkHenkanKey [i] != L' ')
				break ;
		}
		if (i < 0 || pBuffer->m_bufSkkHenkanKey [i] == L'*') {
			/* skk-error "No okurgana!" */
			ImeDoc_vSetSignalError (pDoc) ;
			return	LMR_RETURN ;
		}
	}
	if (FALSE /* skk-allow-spaces-newlines-and-tabs */) {
		LPWSTR	wpSrc, wpDest, wptrEnd ;
		wpSrc	= pBuffer->m_bufSkkHenkanKey ;
		wpDest	= pBuffer->m_bufSkkHenkanKey ;
		wptrEnd	= pBuffer->m_bufSkkHenkanKey + pBuffer->m_nSkkHenkanKeyLen ;
		while (wpSrc < wptrEnd) {
			if (*wpSrc == L' ' || *wpSrc == L'\n' || *wpSrc == L'\r' || *wpSrc == L'\t') {
				while (wpSrc < wptrEnd && (*wpSrc == L' ' || *wpSrc == L'\n' || *wpSrc == L'\r' || *wpSrc == L'\t'))
					wpSrc	++ ;
				if (wpSrc >= wptrEnd)
					break ;
			}
			if (wpDest != wpSrc)
				*wpDest ++	= *wpSrc ++ ;
		}
		pBuffer->m_nSkkHenkanKeyLen	= wpDest - pBuffer->m_bufSkkHenkanKey ;
	} else {
		LPWSTR	wpSrc, wptrEnd ;
		wpSrc	= pBuffer->m_bufSkkHenkanKey ;
		wptrEnd	= pBuffer->m_bufSkkHenkanKey + pBuffer->m_nSkkHenkanKeyLen ;
		while (wpSrc < wptrEnd && (*wpSrc != L'\n' && *wpSrc != L'\r'))
			wpSrc	++ ;
		if (wpSrc < wptrEnd) {
			/* skk-error "Henkan key may not contain a new line character" */
			ImeDoc_vSetSignalError (pDoc) ;
			return	LMR_RETURN ;
		}
		wpSrc	= pBuffer->m_bufSkkHenkanKey ;
		while (wpSrc < wptrEnd && (*wpSrc != L' ' && *wpSrc != L'\t'))
			wpSrc	++ ;
		pBuffer->m_nSkkHenkanKeyLen	= wpSrc - pBuffer->m_bufSkkHenkanKey ;
	}
	imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkHenkanEndPoint, MARKER_SKK_HENKAN_END_POINT, pBuffer->m_pmkPoint) ;
	pBuffer->m_iSkkHenkanCount	= 0 ;

	if (! ImeDoc_bCall (pDoc, LM_bSkkHenkan, LM_bSkkStartHenkan_2))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkStartHenkan_2 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (pBuffer->m_bSkkAbbrevMode && pBuffer->m_bSkkHenkanMode == LACTIVE) {
		imeBuffer_bSkkJModeOn (pBuffer, pBuffer->m_bSkkKatakana) ;
		pBuffer->m_bSkkAbbrevMode	= LTRUE ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-previous-candidate
 */
static	int		LM_bSkkPreviousCandidate_1		(struct CImeDoc*) ;
static	int		LM_bSkkPreviousCandidate_Exit	(struct CImeDoc*) ;

int
LM_bSkkPreviousCandidate (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (pBuffer->m_bSkkHenkanMode != LACTIVE) {
		if (ImeDoc_iGetLastCommand (pDoc) != NFUNC_SKK_KAKUTEI_HENKAN) {
			/* last-command-char ͍̂Ƃ characterp H */
			if (! ImeDoc_bCall (pDoc, LM_bSkkKanaInput, LM_bSkkPreviousCandidate_Exit))
				return	LMR_ERROR ;
			return	LMR_CONTINUE ;
		} else {
			if (! ImeBuffer_bDeleteRegion (pBuffer, TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint), TMarker_iGetPosition (pBuffer->m_pmkPoint))) {
				ImeDoc_vSetSignalError (pDoc) ;
				ImeDoc_vJump (pDoc, LM_bSkkPreviousCandidate_Exit) ;
				return	LMR_CONTINUE ;
			}
			if (! ImeDoc_bCall (pDoc, LM_bSkkSetHenkanPointSubr, LM_bSkkPreviousCandidate_1))
				return	LMR_ERROR ;
			return	LMR_CONTINUE ;
		}
	} else if (pBuffer->m_nSkkHenkanKeyLen == 0) {
		ImeDoc_vJump (pDoc, LM_bSkkPreviousCandidate_Exit) ;
		return	LMR_CONTINUE ;
	} else {
		struct TMarker*	pmkMark		= NULL ;

		if (TMarker_iGetPosition (pBuffer->m_pmkPoint) != pBuffer->m_nbufComp) {
			pmkMark		= ImeBuffer_pMakeMarker (pBuffer, TRUE) ;
			if (pmkMark != NULL) {
				TMarker_bSetPosition (pmkMark, pBuffer->m_pmkPoint) ;
				TMarker_bForward (pmkMark, 1) ;
			}
		}
		if (pBuffer->m_iSkkHenkanCount == 0) {
			int	nSkkPrefixLen ;

			if (pBuffer->m_nSkkOkuriCharLen > 0) {
				if (pBuffer->m_nSkkHenkanKeyLen > 0)
					pBuffer->m_nSkkHenkanKeyLen	-- ;
			}
			if (pBuffer->m_bSkkKatakana) {
				pBuffer->m_nSkkHenkanKeyLen	= iSkkHiraganaToKatakana (pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen, pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen, FALSE) ;
			}
			pBuffer->m_iSkkHenkanCount			= -1 ;
			/* skk-henkan-in-minibuff-flag = nil */
			pBuffer->m_bSkkHenkanInMinibuffFlag	= LFALSE ;
			/* skk-henkan-list = nil */
			pBuffer->m_nSkkOkuriCharLen			= 0 ;
			pBuffer->m_iSkkOkuriIndexMin		= -1 ;
			pBuffer->m_iSkkOkuriIndexMax		= -1 ;
			pBuffer->m_bSkkOkurigana			= LFALSE ;

			/*	skk-process-okueri-early ̏ꍇAusv̒iKŕϊ\ɂȂ̂ŁA
			 *	skk-prefix ݂Ԃ̂܂ skk-previous-candidates ɗ邱Ƃ
			 *	\ɂȂB̎Askk-prefix  nil ɖ߂Ă܂ẮAusv
			 *	cĂ܂ƂɂȂB
			 */
			nSkkPrefixLen	= 0 ;
			if (pBuffer->m_nSkkPrefixLen > 0) {
				if (pBuffer->m_pmkSkkKanaStartPoint != NULL && TMarker_bIsValidp (pBuffer->m_pmkSkkKanaStartPoint)) {
					int	nKanaStartPoint ;

					nKanaStartPoint	= TMarker_iGetPosition (pBuffer->m_pmkSkkKanaStartPoint) ;
					if (nKanaStartPoint < 0 || (nKanaStartPoint + pBuffer->m_nSkkPrefixLen) > pBuffer->m_nbufComp) {
						/* error */
					} else {
						if (memcmp (pBuffer->m_bufComp + nKanaStartPoint, pBuffer->m_bufSkkPrefix, pBuffer->m_nSkkPrefixLen * sizeof (WCHAR)) == 0) {
							nSkkPrefixLen	= pBuffer->m_nSkkPrefixLen ;
						}
					}
				}
			}
			pBuffer->m_nSkkPrefixLen		= nSkkPrefixLen ;

			/* (when (skk-numeric-p) ... */
			/*	ȊJnʒu delete-region ɂĔj󂳂ꂽAinsert ŕłȂ
			 *	̂ňꎞIɃJ[\}[JɕύXB
			 */
			(void) TMarker_bSetCursor (pBuffer->m_pmkSkkKanaStartPoint, TRUE) ;
			ImeBuffer_bDeleteRegion (pBuffer, TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint), TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanEndPoint)) ;
			ImeBuffer_iInsert (pBuffer, pBuffer->m_pmkSkkHenkanEndPoint, pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen) ;
			(void) TMarker_bSetCursor (pBuffer->m_pmkSkkKanaStartPoint, FALSE) ;
			imeBuffer_bSkkChangeMarkerToWhite (pBuffer) ;
		} else {
			pBuffer->m_iSkkHenkanCount	-- ;
			/*	skk-insert-new-word ... skk-get-current-candidate ... skk-henkan-list ̌`CɂȂȁB
			 *	ʂ skk-get-current-candidate-1  car  cdr ƂƂǂȂH
			 */
			if (pBuffer->m_pSkkCurrentSearchProgSession != NULL) {
				LPCWSTR	wstrNewWord ;

				wstrNewWord	= imeBuffer_pSkkGetCurrentCandidate (pBuffer) ;
				if (wstrNewWord != NULL) {
					if (! imeBuffer_bSkkInsertNewWord (pBuffer, wstrNewWord, lstrlenW (wstrNewWord)))
						return	FALSE ;
				}
			}
		}
		if (pmkMark != NULL) {
			TMarker_bSetPosition (pBuffer->m_pmkPoint, pmkMark) ;
			TMarker_bBackward (pBuffer->m_pmkPoint, 1) ;
			ImeBuffer_bDeleteMarker (pBuffer, pmkMark) ;
		} else {
			/* (point-max) ܂ cursor ړB*/
			TMarker_bForward (pBuffer->m_pmkPoint, pBuffer->m_nbufComp - TMarker_iGetPosition (pBuffer->m_pmkPoint)) ;
		}
		if (pBuffer->m_bSkkAbbrevMode && pBuffer->m_iSkkHenkanCount == -1) {
			/* skk-abbrev-mode-on */
			imeBuffer_bSkkAbbrevModeOn (pBuffer) ;
		}
		ImeDoc_vJump (pDoc, LM_bSkkPreviousCandidate_Exit) ;
		return	LMR_CONTINUE ;
	}
}

int
LM_bSkkPreviousCandidate_1 (
	struct CImeDoc*		pDoc)
{
	/* skk-get-last-henkan-datum ͂Ȃc
	 */
	ImeDoc_bSetThisCommand (pDoc, NFUNC_SKK_UNDO_KAKUTEI_HENKAN) ;
	ImeDoc_vJump (pDoc, LM_bSkkPreviousCandidate_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkPreviousCandidate_Exit (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) {
		/* skk-point-move ̌ʁB*/
		imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkPreviousPoint, MARKER_SKK_PREVIOUS_POINT, pBuffer->m_pmkPoint) ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-henkan
 */
/*	recursive-edit ɓ^C~ONeBJB
 */
static	int		LM_bSkkHenkan_1		(struct CImeDoc*) ;
static	int		LM_bSkkHenkan_3		(struct CImeDoc*) ;
static	int		LM_bSkkHenkan_4		(struct CImeDoc*) ;
static	int		LM_bSkkHenkan_Exit	(struct CImeDoc*) ;

int
LM_bSkkHenkan (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	struct TMarker*		pmkMark			= NULL ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (pBuffer->m_nSkkHenkanKeyLen == 0) {
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
		ImeDoc_vJump (pDoc, LM_bSkkKakuteiAdJisx0201) ;
		return	LMR_CONTINUE ;
	} 

	if (! ImeDoc_bPushReg (pDoc, LMREG_1))
		return	LMR_ERROR ;

	if (TMarker_iGetPosition (pBuffer->m_pmkPoint) != pBuffer->m_nbufComp) {
		pmkMark		= ImeBuffer_pMakeMarker (pBuffer, TRUE) ;
		if (pmkMark != NULL) {
			TMarker_bSetPosition (pmkMark, pBuffer->m_pmkPoint) ;
			TMarker_bForward (pmkMark, 1) ;
		}
	}
	ImeDoc_vSetRegPointer (pDoc, LMREG_1, pmkMark) ;

	if (pBuffer->m_bSkkHenkanMode != LACTIVE) {
		if (! ImeDoc_bCall (pDoc, LM_bSkkChangeMarker, LM_bSkkHenkan_1))
			return	LMR_ERROR ;
	} else {
		if (! ImeDoc_bCall (pDoc, LM_bSkkHenkan1, LM_bSkkHenkan_3))
			return	LMR_ERROR ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkan_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int					iSearchType ;
	LPCWSTR				pwOkurigana ;
	int					nOkurigana ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkHenkan_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkan_Exit) ;
		return	LMR_CONTINUE ;
	}

	/* (skk-current-search-prog-list skk-search-prog-list) */
	if (pBuffer->m_pSkkCurrentSearchProgSession != NULL) {
		TSearchSession_vDestroy (pBuffer->m_pSkkCurrentSearchProgSession) ;
	}
	/* 艼̏ꍇ͒Ӂc*/
	/*  okurigana (or skk-henkan-okurigana skk-okuri-char)) */
	if (pBuffer->m_nSkkHenkanOkuriganaLen > 0) {
		pwOkurigana	= pBuffer->m_bufSkkHenkanOkurigana ;
		nOkurigana	= pBuffer->m_nSkkHenkanOkuriganaLen ;
	} else {
		pwOkurigana	= pBuffer->m_bufSkkOkuriChar ;
		nOkurigana	= pBuffer->m_nSkkOkuriCharLen ;
	}
	if (nOkurigana > 0) {
		/*	!  ImeDoc_bSkkHenkanOkuriStrictlyp ̌ʂKvI ƂLĂBWed Jul 05 22:06:11 2006
		 */
		if (ImeConfig_bSkkHenkanOkuriStrictlyp ()) {
			iSearchType	= SEARCH_OKURI_STRICT ;
		} else if (ImeConfig_bSkkHenkanStrictOkuriPrecedencep ()) {
			iSearchType	= SEARCH_OKURI_STRICT_PRECEDENCE ;
		} else {
			iSearchType	= SEARCH_OKURI_ARI ;
		}
		pBuffer->m_pSkkCurrentSearchProgSession	= TOkuriHenkanSession_pCreate (pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen, pwOkurigana, nOkurigana, iSearchType, ImeConfig_bSkkNumericConversionp (), ImeConfig_bSkkNumConvertFloatp ()) ;
	} else {
		if (ImeConfig_bSkkAutoOkuriProcessp ()) {
			/* ʏ̌ + 艼邩Ȃ̑gݍ킹B*/
			iSearchType	= SEARCH_OKURI_AUTO ;
		} else {
			iSearchType	= SEARCH_OKURI_NASHI ;
		}
		pBuffer->m_pSkkCurrentSearchProgSession	= THenkanSession_pCreate (pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen, iSearchType, ImeConfig_bSkkNumericConversionp (), ImeConfig_bSkkNumConvertFloatp ()) ;
	}
	if (! ImeDoc_bCall (pDoc, LM_bSkkHenkan1, LM_bSkkHenkan_3))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkan_3 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int			nResultLen ;
	LPCWSTR		pwResult ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkHenkan_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkan_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegConstString (pDoc, LMREGARG_RETVAL, &pwResult, &nResultLen))
		nResultLen	= -1 ;

	if (nResultLen < 0) {
		/* signal 󂯂ꍇ < 0 ߂ƂB*/
		ImeDoc_vJump (pDoc, LM_bSkkHenkan_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (nResultLen > 0) {
		/*	ImeDoc_vSetRegConstString (pDoc, LMREGARG_RETVAL, pwResult, nResultLen) ;
		 *	LMREGARG_RETVAL ͂̂܂܂ LM_bSkkHenkan_4 ւƔԁB
		 */
		ImeDoc_vJump (pDoc, LM_bSkkHenkan_4) ;
	} else {
		/* ̏ꍇ́Askk-henkan-in-minibuff ɓB*/
		if (! ImeDoc_bCall (pDoc, LM_bSkkHenkanInMinibuff, LM_bSkkHenkan_4))
			return	LMR_ERROR ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkan_4 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPCWSTR				pNewWord ;
	int					nNewWordLen ;
	struct TMarker*		pmkMark	= NULL ;
	BOOL				bKakuteiHenkan	= FALSE ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkHenkan_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkan_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegConstString (pDoc, LMREGARG_RETVAL, &pNewWord, &nNewWordLen)) {
		pNewWord	= NULL ;
		nNewWordLen	= 0 ;
	} else {
		/*	skk-henkan-in-minibuff ̕Ԃl m_bufRETURN 𗘗p邪ALM_bSkkHenkan_3 
		 *	~ėꍇɂ́c
		 */
		LPWSTR		pwBuffer ;

		if (nNewWordLen > 0) {
			pwBuffer	= ImeDoc_pAlloca (pDoc, sizeof (WCHAR) * nNewWordLen, __alignof (WCHAR)) ;
			if (pwBuffer == NULL)
				return	LMR_ERROR ;
			memcpy (pwBuffer, pNewWord, nNewWordLen * sizeof (WCHAR)) ;
			pNewWord	= pwBuffer ;
		} else {
			pNewWord	= NULL ;
		}
		nNewWordLen	= nNewWordLen ;

		bKakuteiHenkan	= pBuffer->m_bSkkKakuteiFlag ;
	}
	if (! ImeDoc_bGetRegPointer (pDoc, LMREG_1, &pmkMark))
		pmkMark			= NULL ;

	if (nNewWordLen > 0) {
		/* skk-insert-new-word new-word */
		imeBuffer_bSkkInsertNewWord (pBuffer, pNewWord, nNewWordLen) ;
	}
	if (pmkMark != NULL) {
		TMarker_bSetPosition (pBuffer->m_pmkPoint, pmkMark) ;
		TMarker_bBackward (pBuffer->m_pmkPoint, 1) ;
		ImeBuffer_bDeleteMarker (pBuffer, pmkMark) ;
	} else {
		/* (point-max) ܂ cursor ړB*/
		TMarker_bForward (pBuffer->m_pmkPoint, pBuffer->m_nbufComp - TMarker_iGetPosition (pBuffer->m_pmkPoint)) ;
	}
	if (bKakuteiHenkan) {
		/* (skk-kakutei new-word) */
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pNewWord, nNewWordLen) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkHenkan_Exit))
			return	LMR_ERROR ;
	} else {
		ImeDoc_vJump (pDoc, LM_bSkkHenkan_Exit) ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkan_Exit (
	struct CImeDoc*			pDoc)
{
	ImeDoc_vPopReg (pDoc, LMREG_1) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-set-henkan-point-subr
 */
static	int		LM_bSkkSetHenkanPointSubr_1 (struct CImeDoc*) ;

int
LM_bSkkSetHenkanPointSubr (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) {
		return	LMR_RETURN ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	/* cancel ͂ȂB*/
	if (pBuffer->m_bSkkHenkanMode) {
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkSetHenkanPointSubr_1))
			return	LMR_ERROR ;
	} else {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_0, FALSE) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKanaCleanup, LM_bSkkSetHenkanPointSubr_1))
			return	LMR_ERROR ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkSetHenkanPointSubr_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int		nPrefix ;
	struct CSkkRuleTreeIterator*	piteSkkRuleTree ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;
	if (! SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree) || SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) == NULL) {
//	if (pBuffer->m_pSkkCurrentRuleTree == NULL || TSkkRuleTreeNode_pGetPrefix (pBuffer->m_pSkkCurrentRuleTree, &nPrefix) == NULL) {
		ImeBuffer_bInsertAndInherit (pBuffer, L"", 1) ;
	} else {
		imeBuffer_bSkkErasePrefix (pBuffer, FALSE) ;
		ImeBuffer_bInsertAndInherit (pBuffer, L"", 1) ;
		imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkKanaStartPoint, MARKER_SKK_KANA_START_POINT, pBuffer->m_pmkPoint) ;
		imeBuffer_bSkkInsertPrefix (pBuffer, pBuffer->m_bufSkkPrefix, pBuffer->m_nSkkPrefixLen) ;
	}
	pBuffer->m_bSkkHenkanMode		= LON ;
	pBuffer->m_pmkSkkHenkanEndPoint	= NULL ;
	imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkHenkanStartPoint, MARKER_SKK_HENKAN_START_POINT, pBuffer->m_pmkPoint) ;

	/*	̏̓G[ĂKsȂ΂ȂȂI
	 */
	imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkPreviousPoint, MARKER_SKK_PREVIOUS_POINT, pBuffer->m_pmkPoint) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-set-henkan-point
 */
static	int		LM_bSkkSetHenkanPoint_1		(struct CImeDoc*) ;
static	int		LM_bSkkSetHenkanPoint_2		(struct CImeDoc*) ;
static	int		LM_bSkkSetHenkanPoint_3		(struct CImeDoc*) ;
static	int		LM_bSkkSetHenkanPoint_4		(struct CImeDoc*) ;
static	int		LM_bSkkSetHenkanPoint_Exit	(struct CImeDoc*) ;

int
LM_bSkkSetHenkanPoint (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int		iLastCommandChar, iLastChar	;
	BOOL	bNormal, bSokuon, bHenkanActive ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	iLastCommandChar	= ImeDoc_iGetLastCommandChar (pBuffer->m_pDoc) ;
	iLastChar			= (L'A' <= iLastCommandChar && iLastCommandChar <= L'Z')? (iLastCommandChar + (L'a' - L'A')) : iLastCommandChar ;
	bNormal				= (iLastChar != iLastCommandChar) ;
	bSokuon				= (pBuffer->m_nSkkPrefixLen == 1 && pBuffer->m_bufSkkPrefix [0] == iLastChar && iLastChar != L'o') ;
	bHenkanActive		= (pBuffer->m_bSkkHenkanMode == LACTIVE) ;

	if (! ImeDoc_bPushReg (pDoc, LMREG_1) ||
		! ImeDoc_bPushReg (pDoc, LMREG_2) ||
		! ImeDoc_bPushReg (pDoc, LMREG_3))
		return	LMR_ERROR ;

	ImeDoc_vSetRegInteger (pDoc, LMREG_1, iLastChar) ;
	ImeDoc_vSetRegBool    (pDoc, LMREG_3, bNormal) ;

	if (pBuffer->m_bSkkHenkanMode != LON) {
		if (bNormal) {
			if (! ImeDoc_bCall (pDoc, LM_bSkkSetHenkanPointSubr, LM_bSkkSetHenkanPoint_Exit))
				return	LMR_ERROR ;
		} else {
			/* ! bNormal */
			if (pBuffer->m_bSkkHenkanMode) {
				if (bHenkanActive) {
					if (! ImeDoc_bCall (pDoc, LM_bSkkSetHenkanPointSubr, LM_bSkkSetHenkanPoint_4))
						return	LMR_ERROR ;
				} else {
					if (! ImeDoc_bCall (pDoc, LM_bSkkSetHenkanPointSubr, LM_bSkkSetHenkanPoint_Exit))
						return	LMR_ERROR ;
				}
			} else {
				if (bHenkanActive) {
					if (! ImeDoc_bCall (pDoc, LM_bSkkEmulateOriginalMap, LM_bSkkSetHenkanPoint_Exit))
						return	LMR_ERROR ;
				}
			}
		}
	} else if (! bNormal) {
		WCHAR	wch	= iLastChar ;
		int		nSkkHenkanStartPoint, nPoint ;

		ImeBuffer_bInsertAndInherit (pBuffer, &wch, 1) ;
		imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkHenkanStartPoint, MARKER_SKK_HENKAN_START_POINT, pBuffer->m_pmkPoint) ;
		pBuffer->m_iSkkHenkanCount		= 0 ;
		if (pBuffer->m_pmkSkkHenkanStartPoint == NULL) {
			ImeDoc_vSetSignalError (pDoc) ;
			return	LMR_RETURN ;
		}
		nSkkHenkanStartPoint	= TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) ;
		nPoint					= TMarker_iGetPosition (pBuffer->m_pmkPoint) ;
		if (nPoint >= nSkkHenkanStartPoint) {
			pBuffer->m_nSkkHenkanKeyLen	= nPoint - nSkkHenkanStartPoint ;
			memcpy (pBuffer->m_bufSkkHenkanKey, pBuffer->m_bufComp + nSkkHenkanStartPoint, pBuffer->m_nSkkHenkanKeyLen * sizeof (WCHAR)) ;
		} else {
			pBuffer->m_nSkkHenkanKeyLen	= 0 ;
		}
		pBuffer->m_nSkkPrefixLen	= 0 ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkHenkan, LM_bSkkSetHenkanPoint_Exit))
			return	LMR_ERROR ;
	} else {
		BOOL	bProcess	= FALSE ;
		int		nPrefix ;
		struct CSkkRuleTreeIterator*	piteSkkRuleTree ;

		piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;
		if (! SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree) || SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) == NULL) {
//		if (pBuffer->m_pSkkCurrentRuleTree == NULL || TSkkRuleTreeNode_pGetPrefix (pBuffer->m_pSkkCurrentRuleTree, &nPrefix) == NULL) {
			ImeDoc_vSetRegInteger (pDoc, LMREG_2, bSokuon) ;
			ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_1) ;
		} else if (pBuffer->m_pmkSkkKanaStartPoint != NULL && pBuffer->m_pmkSkkHenkanStartPoint != NULL && TMarker_iGetPosition (pBuffer->m_pmkSkkKanaStartPoint) != TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint)) {
			ImeDoc_vSetRegInteger (pDoc, LMREG_2, bSokuon) ;
			if (! bSokuon) {
				ImeDoc_vSetRegBool    (pDoc, LMREGARG_0, FALSE) ;
				if (! ImeDoc_bCall (pDoc, LM_bSkkKanaCleanup, LM_bSkkSetHenkanPoint_1))
					return	LMR_ERROR ;
			} else {
				ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_1) ;
			}
		} else {
			ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_Exit) ;
		}
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkSetHenkanPoint_Exit (
	struct CImeDoc*			pDoc)
{
	BOOL	bNormal ;
	int		iLastChar ;

	if (! ImeDoc_bGetRegBool (pDoc, LMREG_3, &bNormal)) 
		bNormal	= FALSE ;
	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_1, &iLastChar)) {
		bNormal		= FALSE ;
		iLastChar	= -1 ;
	}
	ImeDoc_vPopReg (pDoc, LMREG_3) ;
	ImeDoc_vPopReg (pDoc, LMREG_2) ;
	ImeDoc_vPopReg (pDoc, LMREG_1) ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	if (bNormal) {
		ImeDoc_bSetLastCommandChar (pDoc, iLastChar) ;
		ImeDoc_vJump (pDoc, LM_bSkkKanaInput) ;
		return	LMR_CONTINUE ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

int
LM_bSkkSetHenkanPoint_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	BOOL	bProcess		= FALSE ;
	BOOL	bSokuon			= FALSE ;
	BOOL	bNormal			= TRUE ;
	int		iLastChar		= -1 ;

	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_1, &iLastChar))
		iLastChar	= -1 ;
	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_2, &bSokuon))
		bSokuon		= FALSE ;
	if (! ImeDoc_bGetRegBool (pDoc, LMREG_3, &bNormal))
		bNormal		= TRUE ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_Exit) ;
		return	LMR_CONTINUE ;
	}

	if (pBuffer->m_bSkkOkurigana) {
		ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (pBuffer->m_pmkSkkHenkanStartPoint != NULL && TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) == TMarker_iGetPosition (pBuffer->m_pmkPoint)) {
		bProcess	= TRUE ;
	} else {
		WCHAR	wch ;	/* char-before ͌ɂ͓삵ȂB*/
		int		nPoint ;

		nPoint	= TMarker_iGetPosition (pBuffer->m_pmkPoint) ;
		wch		= (nPoint > 0)? pBuffer->m_bufComp [nPoint - 1] : L'\0' ;
		if ((L'0' <= wch && wch <= L'9') || wch == L'' || (L'O' <= wch && wch <= L'X')) {
			bProcess	= FALSE ;
		} else {
			bProcess	= TRUE ;
		}
	}
	if (bProcess) {
		if (ImeConfig_bSkkProcessOkuriEarlyp ()) {
			WCHAR	buf [2] ;
			LPWSTR	pDest, pDestEnd ;
			int		n ;

			imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkHenkanEndPoint, MARKER_SKK_HENKAN_END_POINT, pBuffer->m_pmkPoint) ;

			/* skk-okuri-char ̐ݒB*/
			buf [0]				= iLastChar ;
			pBuffer->m_nSkkOkuriCharLen	= ImeConfig_iGetSkkOkuriChar (buf, 1, pBuffer->m_bufSkkOkuriChar, ARRAYSIZE (pBuffer->m_bufSkkOkuriChar)) ;
			if (pBuffer->m_nSkkOkuriCharLen <= 0) {
				pBuffer->m_bufSkkOkuriChar [0]	= iLastChar ;
				pBuffer->m_nSkkOkuriCharLen		= 1 ;
			}
			pDest		= pBuffer->m_bufSkkHenkanKey ;
			pDestEnd	= pBuffer->m_bufSkkHenkanKey + ARRAYSIZE (pBuffer->m_bufSkkHenkanKey) ;
			if (bSokuon) {
				n			= MIN (pDestEnd - pDest, TMarker_iGetPosition (pBuffer->m_pmkSkkKanaStartPoint) - TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint)) ;
				if (n < 0) {
					ImeDoc_vSetSignalError (pDoc) ;
					ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_Exit) ;
					return	LMR_CONTINUE ;
				}
				if (n > 0)
					memcpy (pDest, pBuffer->m_bufComp + TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint), n * sizeof (WCHAR)) ;
				pDest		+= n ;
				n			= MIN (pDestEnd - pDest, 1) ;
				if (n > 0)
					memcpy (pDest, (pBuffer->m_bSkkKatakana? L"b" : L""), 1 * sizeof (WCHAR)) ;
				pDest		+= n ;
				n			= MIN (pDestEnd - pDest, pBuffer->m_nSkkOkuriCharLen) ;
				if (n > 0)
					memcpy (pDest, pBuffer->m_bufSkkOkuriChar, n * sizeof (WCHAR)) ;
				pDest		+= n ;
				pBuffer->m_nSkkHenkanKeyLen	= pDest - pBuffer->m_bufSkkHenkanKey ;
				imeBuffer_bSkkErasePrefix (pBuffer, FALSE) ;
				ImeBuffer_bInsertAndInherit (pBuffer, pBuffer->m_bSkkKatakana? L"b" : L"", 1) ;
				pBuffer->m_nSkkPrefixLen	= 0 ;
				pBuffer->m_iSkkHenkanCount	= 0 ;
				if (! ImeDoc_bCall (pDoc, LM_bSkkHenkan, LM_bSkkSetHenkanPoint_2))
					return	LMR_ERROR ;

				return	LMR_CONTINUE ;
			} else {
				n			= MIN (pDestEnd - pDest, TMarker_iGetPosition (pBuffer->m_pmkPoint) - TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint)) ;
				if (n < 0) {
					ImeDoc_vSetSignalError (pDoc) ;
					ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_Exit) ;
					return	LMR_CONTINUE ;
				}
				if (n > 0)
					memcpy (pDest, pBuffer->m_bufComp + TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint), n * sizeof (WCHAR)) ;
				pDest		+= n ;
				n			= MIN (pDestEnd - pDest, pBuffer->m_nSkkOkuriCharLen) ;
				if (n > 0)
					memcpy (pDest, pBuffer->m_bufSkkOkuriChar, n * sizeof (WCHAR)) ;
				pDest		+= n ;
				pBuffer->m_nSkkHenkanKeyLen	= pDest - pBuffer->m_bufSkkHenkanKey ;
				ImeBuffer_bInsertAndInherit (pBuffer, L" ", 1) ;
				pBuffer->m_nSkkPrefixLen	= 0 ;
				pBuffer->m_iSkkHenkanCount	= 0 ;
				if (! ImeDoc_bCall (pDoc, LM_bSkkHenkan, LM_bSkkSetHenkanPoint_2))
					return	LMR_ERROR ;

				return	LMR_CONTINUE ;
			}
			/* ... */
		} else {
			if (TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) != TMarker_iGetPosition (pBuffer->m_pmkPoint)) {
				int		nPrefix ;
				struct CSkkRuleTreeIterator*	piteSkkRuleTree ;

				if (bSokuon) {
					imeBuffer_bSkkErasePrefix (pBuffer, TRUE) ;
					ImeBuffer_bInsertAndInherit (pBuffer, pBuffer->m_bSkkKatakana? L"b" : L"", 1) ;
				}

				/*	 "IwakU" Ɠ͂āA"IwaKu" Ɠ̓鏈B
				 */
				piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;
				if (! bSokuon && 
					SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree) &&
					SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) != NULL &&
//					pBuffer->m_pSkkCurrentRuleTree != NULL && 
//					TSkkRuleTreeNode_pGetPrefix (pBuffer->m_pSkkCurrentRuleTree, &nPrefix) != NULL &&
					bNormal &&
					pBuffer->m_nSkkPrefixLen > 0 &&
					TMarker_iGetPosition (pBuffer->m_pmkPoint) >= pBuffer->m_nSkkPrefixLen &&
					memcmp (pBuffer->m_bufComp + TMarker_iGetPosition (pBuffer->m_pmkPoint) - pBuffer->m_nSkkPrefixLen, pBuffer->m_bufSkkPrefix, pBuffer->m_nSkkPrefixLen * sizeof (WCHAR)) == 0) {

					/*	]̕@ƁAMakSu ̂悤Ȍ͂ƁA
					 *	(1) "k|" (: S) (| ̓J[\ʒu)
					 *	(2) "s|" 
					 *	(3) "s|" (: u)
					 *	(4) "s"
					 *	ƂϊĂ܂̂ŁACB́Askk-kana-input ̑ɔf
					 *	䂾˂ׂŁA okurigana ƎvĂ܂̂͂悭Ȃ悤B
					 *	ƁAMatsU ̂悤ȓ͂\ɁBprefix-len == 1̔͂͂B
					 */
					if (pBuffer->m_bSkkOkurigana) {
						if (pBuffer->m_pmkSkkOkuriganaStartPoint != NULL && 
							TMarker_bIsValidp (pBuffer->m_pmkSkkOkuriganaStartPoint) && 
							pBuffer->m_bufComp [TMarker_iGetPosition (pBuffer->m_pmkSkkOkuriganaStartPoint)] == L'*') {
							ImeBuffer_bDeleteRegion (pBuffer, TMarker_iGetPosition (pBuffer->m_pmkSkkOkuriganaStartPoint), TMarker_iGetPosition (pBuffer->m_pmkSkkOkuriganaStartPoint) + 1) ;
						}
						pBuffer->m_bSkkOkurigana	= LFALSE ;
					}
					if (pBuffer->m_pmkSkkKanaStartPoint != NULL && TMarker_bIsValidp (pBuffer->m_pmkSkkKanaStartPoint)) {
						ImeBuffer_bDeleteRegion (pBuffer, TMarker_iGetPosition (pBuffer->m_pmkSkkKanaStartPoint), TMarker_iGetPosition (pBuffer->m_pmkPoint)) ;
					}
					imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkOkuriganaStartPoint, MARKER_SKK_OKURIGANA_START_POINT, pBuffer->m_pmkPoint) ;
					ImeBuffer_bInsertAndInherit (pBuffer, L"*", 1) ;
					imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkKanaStartPoint, MARKER_SKK_KANA_START_POINT, pBuffer->m_pmkPoint) ;
					imeBuffer_bSkkInsertPrefix (pBuffer, pBuffer->m_bufSkkPrefix, pBuffer->m_nSkkPrefixLen) ;

					pBuffer->m_nSkkOkuriCharLen		= pBuffer->m_nSkkPrefixLen ;
					memcpy (pBuffer->m_bufSkkOkuriChar, pBuffer->m_bufSkkPrefix, sizeof (WCHAR) * pBuffer->m_nSkkPrefixLen) ;
					pBuffer->m_bSkkOkurigana		= LTRUE ;
				} else {
					imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkOkuriganaStartPoint, MARKER_SKK_OKURIGANA_START_POINT, pBuffer->m_pmkPoint) ;
					ImeBuffer_bInsertAndInherit (pBuffer, L"*", 1) ;
					imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkKanaStartPoint, MARKER_SKK_KANA_START_POINT, pBuffer->m_pmkPoint) ;
					pBuffer->m_nSkkOkuriCharLen		= 1 ;
					pBuffer->m_bufSkkOkuriChar [0]	= iLastChar ;
					pBuffer->m_bSkkOkurigana		= LTRUE ;
				}
			}
		}
	}
	ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkSetHenkanPoint_2 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	BOOL			bSokuon ;

	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_2, &bSokuon)) 
		bSokuon	= FALSE ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_Exit) ;
		return	LMR_CONTINUE ;
	}

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeBuffer_bDeleteBackwardChar (pBuffer, bSokuon? 2 : 1)) {
		ImeDoc_vSetSignalError (pDoc) ;
	}
	ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkSetHenkanPoint_3 (
	struct CImeDoc*			pDoc)
{
	if (! ImeDoc_bCall (pDoc, LM_bSkkSetCharBeforeAsOkurigana, LM_bSkkSetHenkanPoint_Exit))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkSetHenkanPoint_4 (
	struct CImeDoc*		pDoc)
{
	ImeDoc_vJump (pDoc, LM_bSkkSetHenkanPoint_Exit) ;
	return	LMR_CONTINUE ;
}

/*================================================================ skk-mode
 */
static	int		LM_bSkkMode_1	(struct CImeDoc*) ;

int
LM_bSkkMode (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (! pBuffer->m_bSkkMode) {
		if (! ImeDoc_bCall (pDoc, LM_bSkkModeExit, LM_bSkkMode_1))
			return	LMR_ERROR ;
	} else {
		if (! pBuffer->m_bSkkModeInvoked)
			imeBuffer_bSkkModeInvoke (pBuffer) ;
		ImeDoc_vJump (pDoc, LM_bSkkMode_1) ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkMode_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	/* ... pre-command-hook, post-command-hook ̓x^ɎƂāA
	 *	run-hooks ͕s\B*/
	imeBuffer_bSkkJModeOn (pBuffer, LFALSE) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-mode-exit
 */
static	int		LM_bSkkModeExit_1 (struct CImeDoc*) ;

int
LM_bSkkModeExit (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (! ImeDoc_bPushReg (pDoc, LMREG_0))
		return	LMR_ERROR ;

	/* 0  LM_bSkkKakuteiAdJisx0201 ւ̈B*/
	ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
	ImeDoc_vSetRegBool   (pDoc, LMREG_0, pBuffer->m_bSkkMode) ;
	pBuffer->m_bSkkMode	= TRUE ;
	if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkModeExit_1))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkModeExit_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	BOOL				bSkkMode ;

	if (! ImeDoc_bGetRegBool (pDoc, LMREG_0, &bSkkMode))
		bSkkMode	= FALSE ;
	ImeDoc_vPopReg (pDoc, LMREG_0) ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	pBuffer->m_bSkkMode	= bSkkMode ;
	imeBuffer_bSkkModeOff (pBuffer) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-latin-mode
 */
static	int		LM_bSkkLatinMode_1 (struct CImeDoc*) ;

int
LM_bSkkLatinMode (
	struct CImeDoc*				pDoc)
{
	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
	if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkLatinMode_1))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkLatinMode_1 (
	struct CImeDoc*				pDoc)
{
	struct CImeBuffer*		pBuffer ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	imeBuffer_bSkkLatinModeOn (pBuffer) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-jisx0208-latin-mode
 */
static	int		LM_bSkkJisx0208LatinMode_1 (struct CImeDoc*) ;

int
LM_bSkkJisx0208LatinMode (
	struct CImeDoc*				pDoc)
{
	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
	if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkJisx0208LatinMode_1))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkJisx0208LatinMode_1 (
	struct CImeDoc*				pDoc)
{
	struct CImeBuffer*		pBuffer ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	imeBuffer_bSkkJisx0208LatinModeOn (pBuffer) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-abbrev-mode
 */
static	int		LM_bSkkAbbrevMode_1 (struct CImeDoc*) ;
static	int		LM_bSkkAbbrevMode_2 (struct CImeDoc*) ;

int
LM_bSkkAbbrevMode (
	struct CImeDoc*				pDoc)
{
	struct CImeBuffer*		pBuffer ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkAbbrevMode_1))
			return	LMR_ERROR ;
	} else if (pBuffer->m_bSkkHenkanMode == LON) {
		/* "Ɂ[hɓĂ܂" "Already in  mode" */
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	} else {
		ImeDoc_vJump (pDoc, LM_bSkkAbbrevMode_1) ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkAbbrevMode_1 (
	struct CImeDoc*				pDoc)
{
	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	if (! ImeDoc_bCall (pDoc, LM_bSkkSetHenkanPointSubr, LM_bSkkAbbrevMode_2))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkAbbrevMode_2 (
	struct CImeDoc*				pDoc)
{
	struct CImeBuffer*		pBuffer ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	imeBuffer_bSkkAbbrevModeOn (pBuffer) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-set-okurigana
 */
static	int		LM_bSkkSetOkurigana_1 (struct CImeDoc*) ;

int
LM_bSkkSetOkurigana (
	struct CImeDoc*				pDoc)
{
	struct CImeBuffer*	pBuffer ;
	WCHAR	wbuf [32] ;
	int		nOkuriStartPos, nPoint, nHenkanStartPoint, nHenkanEndPoint, nOkuriganaPrefix ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkHenkanEndPoint, MARKER_SKK_HENKAN_END_POINT, pBuffer->m_pmkSkkOkuriganaStartPoint) ;
	/* save-point Ȃ̂ marker  point ۑB*/
	nOkuriStartPos	= TMarker_iGetPosition (pBuffer->m_pmkSkkOkuriganaStartPoint) ;
	if (pBuffer->m_bufComp [nOkuriStartPos] != L'*') {
		ImeBuffer_iInsertByPosition (pBuffer, nOkuriStartPos, L"*", 1) ;
	}
	nPoint			= TMarker_iGetPosition (pBuffer->m_pmkPoint) ;

	/* 艼 start point ɂ '*' ݂B*/
	pBuffer->m_nSkkHenkanOkuriganaLen	= nPoint - nOkuriStartPos - 1 ;
	if (pBuffer->m_nSkkHenkanOkuriganaLen > 0) {
		pBuffer->m_nSkkHenkanOkuriganaLen	= (pBuffer->m_nSkkHenkanOkuriganaLen > MAXLEN_HENKAN_OKURIGANA)? MAXLEN_HENKAN_OKURIGANA : pBuffer->m_nSkkHenkanOkuriganaLen ;
		memcpy (pBuffer->m_bufSkkHenkanOkurigana, pBuffer->m_bufComp + nOkuriStartPos + 1, sizeof (WCHAR) * pBuffer->m_nSkkHenkanOkuriganaLen) ;
	}

	nHenkanStartPoint	= TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) ;
	nHenkanEndPoint		= TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanEndPoint) ;
	pBuffer->m_nSkkHenkanKeyLen		= nHenkanEndPoint - nHenkanStartPoint ;
	memcpy (pBuffer->m_bufSkkHenkanKey, pBuffer->m_bufComp + nHenkanStartPoint, sizeof (WCHAR) * (nHenkanEndPoint - nHenkanStartPoint)) ;

	nOkuriganaPrefix	= imeBuffer_iSkkOkuriganaPrefix (pBuffer, pBuffer->m_bufSkkHenkanOkurigana, pBuffer->m_nSkkHenkanOkuriganaLen, wbuf, ARRAYSIZE (wbuf)) ;
	if (nOkuriganaPrefix > 0) {
		if ((pBuffer->m_nSkkHenkanKeyLen + nOkuriganaPrefix) <= MAXLEN_HENKAN_KEY) {
			memcpy (pBuffer->m_bufSkkHenkanKey + pBuffer->m_nSkkHenkanKeyLen, wbuf, nOkuriganaPrefix * sizeof (WCHAR)) ;
			pBuffer->m_nSkkHenkanKeyLen	+= nOkuriganaPrefix ;
		}
	} else {
		if ((pBuffer->m_nSkkHenkanKeyLen + pBuffer->m_nSkkOkuriCharLen) <= MAXLEN_HENKAN_KEY) {
			memcpy (pBuffer->m_bufSkkHenkanKey + pBuffer->m_nSkkHenkanKeyLen, pBuffer->m_bufSkkOkuriChar, pBuffer->m_nSkkOkuriCharLen * sizeof (WCHAR)) ;
			pBuffer->m_nSkkHenkanKeyLen	+= pBuffer->m_nSkkOkuriCharLen ;
		}
	}
	pBuffer->m_nSkkPrefixLen	= 0 ;

	if (pBuffer->m_bSkkKatakana) {
		pBuffer->m_nSkkHenkanKeyLen			= iSkkKatakanaToHiragana (pBuffer->m_bufSkkHenkanKey, ARRAYSIZE (pBuffer->m_bufSkkHenkanKey), pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen, FALSE) ;
		pBuffer->m_nSkkHenkanOkuriganaLen	= iSkkKatakanaToHiragana (pBuffer->m_bufSkkHenkanOkurigana, ARRAYSIZE (pBuffer->m_bufSkkHenkanOkurigana), pBuffer->m_bufSkkHenkanOkurigana, pBuffer->m_nSkkHenkanOkuriganaLen, FALSE) ;
	}
	ImeBuffer_bDeleteRegion (pBuffer, TMarker_iGetPosition (pBuffer->m_pmkSkkOkuriganaStartPoint), TMarker_iGetPosition (pBuffer->m_pmkSkkOkuriganaStartPoint) + 1) ;
	pBuffer->m_iSkkHenkanCount	= 0 ;

	if (! ImeDoc_bCall (pDoc, LM_bSkkHenkan, LM_bSkkSetOkurigana_1))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkSetOkurigana_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	pBuffer->m_bSkkOkurigana	= LFALSE ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-delete-backward-char
 */
static	int		LM_bSkkDeleteBackwardChar_1		(struct CImeDoc*) ;
static	int		LM_bSkkDeleteBackwardChar_Exit	(struct CImeDoc*) ;

int
LM_bSkkDeleteBackwardChar (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
		if (! ImeConfig_bSkkDeleteImplesKakuteip () && TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanEndPoint) == TMarker_iGetPosition (pBuffer->m_pmkPoint)) {
			if (! ImeDoc_bCall (pDoc, LM_bSkkPreviousCandidate, LM_bSkkDeleteBackwardChar_Exit))
				return	LMR_ERROR ;
			/* overwrite-mode ͂ȂB*/
			return	LMR_CONTINUE ;
		} else {
			if (! ImeDoc_bCall (pDoc, LM_bSkkEmulateOriginalMap, LM_bSkkDeleteBackwardChar_1))
				return	LMR_ERROR ;
			return	LMR_CONTINUE ;
		}
	} else if (pBuffer->m_bSkkHenkanMode && TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) >= TMarker_iGetPosition (pBuffer->m_pmkPoint)) {
		pBuffer->m_iSkkHenkanCount	= 0 ;
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkDeleteBackwardChar_Exit))
			return	LMR_ERROR ;
		return	LMR_CONTINUE ;
	} else {
		struct CSkkRuleTreeIterator*	piteSkkRuleTree ;
		int		nPrefix ;

		imeBuffer_bSkkDeleteOkuriMark (pBuffer) ;
		piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;
		if (SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree) && SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) != NULL) {
//		if (pBuffer->m_pSkkCurrentRuleTree != NULL && TSkkRuleTreeNode_pGetPrefix (pBuffer->m_pSkkCurrentRuleTree, &nPrefix) != NULL) {
			imeBuffer_bSkkErasePrefix (pBuffer, TRUE) ;
		} else {
			pBuffer->m_pmkSkkKanaStartPoint	= NULL ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkEmulateOriginalMap, LM_bSkkDeleteBackwardChar_Exit))
				return	LMR_ERROR ;
			return	LMR_CONTINUE ;
		}
	}
	ImeDoc_vJump (pDoc, LM_bSkkDeleteBackwardChar_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkDeleteBackwardChar_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	int		nCount	= 1 ;	/* num-arg = 1 ƉB*/

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkDeleteBackwardChar_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkDeleteBackwardChar_Exit) ;
		return	LMR_CONTINUE ;
	}

	if (pBuffer->m_nSkkPrefixLen > nCount) {
		pBuffer->m_nSkkPrefixLen	-- ;
	} else {
		pBuffer->m_nSkkPrefixLen	= 0 ;
	}
	if (TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanEndPoint) >= TMarker_iGetPosition (pBuffer->m_pmkPoint)) {
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkDeleteBackwardChar_Exit))
			return	LMR_ERROR ;
	} else {
		ImeDoc_vJump (pDoc, LM_bSkkDeleteBackwardChar_Exit) ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkDeleteBackwardChar_Exit (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) {
		/* skk-point-move ̌ʁB*/
		imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkPreviousPoint, MARKER_SKK_PREVIOUS_POINT, pBuffer->m_pmkPoint) ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-kakutei
 */
static	int		LM_bSkkKakutei_1 (struct CImeDoc*) ;
static	int		LM_bSkkKakutei_2 (struct CImeDoc*) ;
static	int		LM_bSkkKakutei_3 (struct CImeDoc*) ;
static	int		LM_bSkkKakutei_4 (struct CImeDoc*) ;
static	int		LM_bSkkKakutei_5 (struct CImeDoc*) ;

/*	skk-update-jisyo ̒ minibuffer-flag  true Ȃ yes-or-no-p Ă΂Ă܂߂
 *	 skk-kakutei ͂̈ʒuɂȂ΂ȂȂB
 */
int
LM_bSkkKakutei (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPCWSTR				wstrWord ;
	int					nWordLen ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	/*	mword ς܂Ă邩ǂ`FbNB*/
	if (! ImeDoc_bGetRegConstString (pDoc, LMREGARG_0, &wstrWord, &nWordLen)) {
		wstrWord	= NULL ;
		nWordLen	= 0 ;
	}
	if (pBuffer->m_bSkkHenkanMode) {
		LPWSTR	pwKakuteiMidasi,   pwKakuteiWord ;
		int		nKakuteiMidasiLen, nKakuteiWordLen ;

		pwKakuteiMidasi	= (LPWSTR) ImeDoc_pAlloca (pDoc, MAXLEN_KAKUTEI_MIDASI * sizeof (WCHAR), __alignof (WCHAR)) ;
		pwKakuteiWord	= (LPWSTR) ImeDoc_pAlloca (pDoc, MAXLEN_KAKUTEI_WORD   * sizeof (WCHAR), __alignof (WCHAR)) ;
		if (pwKakuteiMidasi == NULL || pwKakuteiWord == NULL)
			return	LMR_ERROR ;

		if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
			/* ̍XVꂱB͎͂ȂB̂B*/
			/* (skk-update-jisyo kakutei-word) */
			if (nWordLen <= 0 && pBuffer->m_pSkkCurrentSearchProgSession != NULL) {
				wstrWord	= TSearchSession_pGetReferCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;
				if (wstrWord == NULL)
					wstrWord	= TSearchSession_pGetCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;
				nWordLen	= (wstrWord != NULL)? (int)wcslen (wstrWord) : 0 ;
			}
		/*
		(when (or (and (not skk-search-excluding-word-pattern-function)
				       kakutei-word)
				  (and kakutei-word
				       skk-search-excluding-word-pattern-function
				       (not (funcall
						     skk-search-excluding-word-pattern-function
						     kakutei-word))))
		  (skk-update-jisyo kakutei-word)

		  */
			/*	_bSkkUpdateJisyo ̌ĂяoŒO̊m(o+P)XVĂ܂̂
			 *	ŕۑĂB
			 */
			nKakuteiMidasiLen	= pBuffer->m_nKakuteiMidasiLen ;
			if (nKakuteiMidasiLen > 0) {
				memcpy (pwKakuteiMidasi, pBuffer->m_bufKakuteiMidasi, nKakuteiMidasiLen * sizeof (WCHAR)) ;
				nKakuteiWordLen		= pBuffer->m_nKakuteiWordLen ;
				memcpy (pwKakuteiWord,  pBuffer->m_bufKakuteiWord,    nKakuteiWordLen * sizeof (WCHAR)) ;
			} else {
				nKakuteiWordLen		= 0 ;
			}

			if (nWordLen > 0) {
				LPWSTR	pDest ;

				/*	́A wstrWord, nWordLen  LM_bSkkUpdateJisyo ɓnȂ΂ȂȂA
				 *	̎wĂ̈悪ۏ؂ȂȂƂƂB
				 *
				 *	z@B stack ̈mۂāAɕۑcB alignment B
				 */
				pDest	= (LPWSTR) ImeDoc_pAlloca (pDoc, nWordLen * sizeof (WCHAR), __alignof (WCHAR)) ;
				if (pDest == NULL) {
					/* fatal error */
					return	LMR_ERROR ;
				}
				memcpy (pDest, wstrWord, sizeof (WCHAR) * nWordLen) ;

				if (! ImeDoc_bPushReg (pDoc, LMREG_0) || ! ImeDoc_bPushReg (pDoc, LMREG_1))
					return	LMR_ERROR ;
				ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pDest, nWordLen) ;
				ImeDoc_vSetRegBool   (pDoc, LMREGARG_1, FALSE) ;
				ImeDoc_vSetRegString (pDoc, LMREG_0, pwKakuteiWord,   nKakuteiWordLen) ;
				ImeDoc_vSetRegString (pDoc, LMREG_1, pwKakuteiMidasi, nKakuteiMidasiLen) ;
				if (! ImeDoc_bCall (pDoc, LM_bSkkUpdateJisyo, LM_bSkkKakutei_1)) 
					return	LMR_ERROR ;

				/*	̕t߂œǂ݉Ƃ MidashiWord uĂƂƂKvɂȂɈႢȂcB
				 *	̏ shift č폜ꂽ肵AƂłB
				 *
				 *	ɑǁAvmFI
				 *
				 *	Aǂ݉ĝ͊mł͂ȂϊȂ̂ł͂ȂHƂ^₪B
				 *	^C~OŕKvƂ̂؂ȂƂȂB
				 */
				if (! ImeDoc_bRecursiveEditp (pDoc)) {
					ImeBuffer_bSetReadingProperty (pBuffer, pBuffer->m_pmkSkkHenkanStartPoint, pBuffer->m_pmkSkkHenkanEndPoint, pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen) ;
				}
				return	LMR_CONTINUE ;
			} else {
				if (! ImeDoc_bPushReg (pDoc, LMREG_0) || ! ImeDoc_bPushReg (pDoc, LMREG_1))
					return	LMR_ERROR ;
				ImeDoc_vSetRegString (pDoc, LMREG_0, pwKakuteiWord,   nKakuteiWordLen) ;
				ImeDoc_vSetRegString (pDoc, LMREG_1, pwKakuteiMidasi, nKakuteiMidasiLen) ;
				ImeDoc_vJump (pDoc, LM_bSkkKakutei_1) ;
				return	LMR_CONTINUE ;
			}
		} else {
			int		nHenkanStartPoint ;

			/* pBuffer->m_bSkkHenkanMode ==LON */
			/*
			 *		;; [hŊm肵ꍇB֋XIɌ݂̃|Cg܂ło
			 *		;; ėXVB
			 *		(when (and (> skk-kakutei-history-limit 0)
			 *				   (< skk-henkan-start-point (point))
			 *				   (skk-save-point
			 *				    (goto-char skk-henkan-start-point)
			 *				    (eq (skk-what-char-type) 'hiragana)))
			 *		  (skk-update-kakutei-history
			 *		   (buffer-substring-no-properties
			 *		    skk-henkan-start-point (point))))))
			 */
			if (pBuffer->m_pmkSkkHenkanStartPoint != NULL && TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) < TMarker_iGetPosition (pBuffer->m_pmkPoint)) {
				nHenkanStartPoint	= TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) ;
				if (0 <= nHenkanStartPoint && nHenkanStartPoint < pBuffer->m_nbufComp && L'' <= pBuffer->m_bufComp [nHenkanStartPoint] && pBuffer->m_bufComp [nHenkanStartPoint] <= L'') {
					imeBuffer_bSkkUpdateKakuteiHistory (pBuffer, pBuffer->m_bufComp + nHenkanStartPoint, TMarker_iGetPosition (pBuffer->m_pmkPoint) - nHenkanStartPoint, NULL, 0) ;
				}
			}
		}
		ImeDoc_vJump (pDoc, LM_bSkkKakutei_4) ;
		return	LMR_CONTINUE ;
	}
	ImeDoc_vJump (pDoc, LM_bSkkKakutei_5) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKakutei_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPWSTR	pwCombWord		= NULL ;
	LPWSTR	pwSkkHenkanKey	= NULL ;
	LPWSTR	pwKakuteiMidasi,   pwKakuteiWord ;
	int		nKakuteiMidasiLen, nKakuteiWordLen ;
	int		nCombWordLen, nSkkHenkanKeyLen, nHenkanStartPoint ;

	/*	m茩oƃWX^̕A܂ł͕KsB
	 */
	if (! ImeDoc_bGetRegString (pDoc, LMREG_0, &pwKakuteiWord, &nKakuteiWordLen)) {
		pwKakuteiWord	= NULL ;
		nKakuteiWordLen	= 0 ;
	}
	if (! ImeDoc_bGetRegString (pDoc, LMREG_1, &pwKakuteiMidasi, &nKakuteiMidasiLen)) {
		pwKakuteiMidasi		= NULL ;
		nKakuteiMidasiLen	= 0 ;
	}
	ImeDoc_vPopReg (pDoc, LMREG_1) ;
	ImeDoc_vPopReg (pDoc, LMREG_0) ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	pwSkkHenkanKey	= (LPWSTR) ImeDoc_pAlloca (pDoc, pBuffer->m_nSkkHenkanKeyLen * sizeof (WCHAR), __alignof (WCHAR)) ;
	pwCombWord		= (LPWSTR) ImeDoc_pAlloca (pDoc, MAXCOMPLEN * sizeof (WCHAR), __alignof (WCHAR)) ;
	if (pwSkkHenkanKey == NULL || pwCombWord == NULL)
		return	LMR_ERROR ;

	if (pBuffer->m_bSkkAfterPrefix && ! bSettoji (pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen)) {

		if (nKakuteiWordLen > 0 && bSettoji (pwKakuteiMidasi, nKakuteiMidasiLen)) {
			/* skk-henkan-key ̑ޔB*/
			if (pBuffer->m_nSkkHenkanKeyLen > 0) {
				memcpy (pwSkkHenkanKey, pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen * sizeof (WCHAR)) ;
			}
			nSkkHenkanKeyLen	= pBuffer->m_nSkkHenkanKeyLen ;

			/*	skk-henkan-start-point ̌ʂA(nth 1 list2) ɓcB
			 */
			nHenkanStartPoint	= TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) - pBuffer->m_nKakuteiWordLen - 1 /* ``''̕B*/ ;

			/*	O̕ϊ application ɓnĂ\B̏ꍇȀ͓Ă͂܂ȂcB
			 */
			if (nHenkanStartPoint >= 0 && (nHenkanStartPoint + nKakuteiWordLen) <= pBuffer->m_nbufComp && wcsncmp (pBuffer->m_bufComp + nHenkanStartPoint, pwKakuteiWord, nKakuteiWordLen) == 0) {
				pBuffer->m_nSkkHenkanKeyLen	= iConcatString (pBuffer->m_bufSkkHenkanKey, ARRAYSIZE (pBuffer->m_bufSkkHenkanKey), pwKakuteiMidasi, nKakuteiMidasiLen - 1, pBuffer->m_bufKakuteiMidasi, pBuffer->m_nKakuteiMidasiLen, NULL) ;
				nCombWordLen				= iConcatString (pwCombWord,                 MAXCOMPLEN,                             pwKakuteiWord,   nKakuteiWordLen,       pBuffer->m_bufKakuteiWord,   pBuffer->m_nKakuteiWordLen,   NULL) ;

				/*	́A wstrWord, nWordLen  LM_bSkkUpdateJisyo ɓnȂ΂ȂȂA
				 *	̎wĂ̈悪ۏ؂ȂȂƂƂB
				 *
				 *	z@B stack ̈mۂāAɕۑcB alignment B
				 */
				if (! ImeDoc_bPushReg (pDoc, LMREG_0) || ! ImeDoc_bPushReg (pDoc, LMREG_1))
					return	LMR_ERROR ;
				ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwCombWord, nCombWordLen) ;
				ImeDoc_vSetRegBool   (pDoc, LMREGARG_1, FALSE) ;
				ImeDoc_vSetRegString (pDoc, LMREG_0, pwSkkHenkanKey, nSkkHenkanKeyLen) ;
				ImeDoc_vSetRegBool   (pDoc, LMREG_1, TRUE) ;
				if (! ImeDoc_bCall (pDoc, LM_bSkkUpdateJisyo, LM_bSkkKakutei_2)) {
					return	LMR_ERROR ;
				}
				return	LMR_CONTINUE ;
			}
		}
		pBuffer->m_bSkkAfterPrefix	= LFALSE ;

	} else if (pBuffer->m_nKakuteiWordLen > 0 && bSetsubiji (pBuffer->m_bufKakuteiMidasi, pBuffer->m_nKakuteiMidasiLen)) {

		/* skk-henkan-key ̑ޔB*/
		if (pBuffer->m_nSkkHenkanKeyLen > 0)
			memcpy (pwSkkHenkanKey, pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen * sizeof (WCHAR)) ;
		nSkkHenkanKeyLen	= pBuffer->m_nSkkHenkanKeyLen ;

		pBuffer->m_nSkkHenkanKeyLen	= iConcatString (pBuffer->m_bufSkkHenkanKey, ARRAYSIZE (pBuffer->m_bufSkkHenkanKey), pwKakuteiMidasi, nKakuteiMidasiLen, pBuffer->m_bufKakuteiMidasi + 1, pBuffer->m_nKakuteiMidasiLen - 1, NULL) ;
		nCombWordLen				= iConcatString (pwCombWord,                 MAXCOMPLEN,                             pwKakuteiWord,   nKakuteiWordLen,   pBuffer->m_bufKakuteiWord,       pBuffer->m_nKakuteiWordLen,       NULL) ;

		/*	skk-update-jisyo ̌ĂяoB
		 */
		if (! ImeDoc_bPushReg (pDoc, LMREG_0) || ! ImeDoc_bPushReg (pDoc, LMREG_1))
			return	LMR_ERROR ;
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwCombWord, nCombWordLen) ;
		ImeDoc_vSetRegBool   (pDoc, LMREGARG_1, FALSE) ;
		ImeDoc_vSetRegString (pDoc, LMREG_0, pwSkkHenkanKey, nSkkHenkanKeyLen) ;
		ImeDoc_vSetRegBool   (pDoc, LMREG_1, FALSE) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkUpdateJisyo, LM_bSkkKakutei_2)) {
			return	LMR_ERROR ;
		}
		return	LMR_CONTINUE ;
	}

	ImeDoc_vJump (pDoc, LM_bSkkKakutei_3) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKakutei_2 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPWSTR				pwSkkHenkanKey ;
	int					nSkkHenkanKeyLen ;
	int					nRetval	= LMR_RETURN ;
	BOOL				bAfterPrefix ;

	if (ImeDoc_bSignalp (pDoc)) 
		goto	exit_func ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		goto	exit_func ;
	}
	/*	 REG[2] ̒lۑĂ邩ۂɐsB܂j󂷂ꍇɂ͕ۑĂAƁH
	 */
	if (! ImeDoc_bGetRegString (pDoc, LMREG_0, &pwSkkHenkanKey, &nSkkHenkanKeyLen)) {
		pwSkkHenkanKey		= NULL ;
		nSkkHenkanKeyLen	= 0 ;
	}
	if (! ImeDoc_bGetRegBool (pDoc, LMREG_1, &bAfterPrefix)) {
		bAfterPrefix		= FALSE ;
	}
	if (bAfterPrefix) {
		pBuffer->m_bSkkAfterPrefix	= LFALSE ;
	}

	/* skk-henkan-key ̕B*/
	if (nSkkHenkanKeyLen > 0)
		memcpy (pBuffer->m_bufSkkHenkanKey, pwSkkHenkanKey, nSkkHenkanKeyLen * sizeof (WCHAR)) ;
	pBuffer->m_nSkkHenkanKeyLen	= nSkkHenkanKeyLen ;

	ImeDoc_vJump (pDoc, LM_bSkkKakutei_3) ;
	nRetval	= LMR_CONTINUE ;

exit_func:
	ImeDoc_vPopReg (pDoc, LMREG_1) ;
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	nRetval ;
}

/*
 *	(when (skk-numeric-p)
 *		(setq converted (skk-get-current-candidate))
 *		(skk-num-update-jisyo kakutei-word converted))))
 */
int
LM_bSkkKakutei_3 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! imeBuffer_bSkkNumUpdateJisyo (pBuffer)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	ImeDoc_vJump (pDoc, LM_bSkkKakutei_4) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKakutei_4 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (pBuffer->m_bSkkMode) {
		imeBuffer_bSkkKakuteiCleanupBuffer (pBuffer) ;
		/* skk-kakutei-end-function ͎\͂ȂB*/
		/* skk-kakutei-initialize? */
		imeBuffer_bSkkKakuteiInitialize (pBuffer, NULL, 0) ;
	}
	ImeDoc_vJump (pDoc, LM_bSkkKakutei_5) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKakutei_5 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	/* skk-do-auto-fill */
	if (pBuffer->m_bSkkMode) {
		if (! (pBuffer->m_bSkkJMode || pBuffer->m_bSkkJisx0201Mode)) {
			imeBuffer_bSkkJModeOn (pBuffer, pBuffer->m_bSkkKatakana) ;
		} 
	} else {
		ImeDoc_vJump (pDoc, LM_bSkkModeAdJisx0201) ;
		return	LMR_CONTINUE ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-emulate-original-map
 */
int
LM_bSkkEmulateOriginalMap (
	struct CImeDoc*		pDoc)
{
	const struct TMSG*	pMsg ;
	int					nFuncNo, nOrigFunc ;
	PLMFUNC				pPC	= NULL ;

	/*	major-map  lookup  function ǉ call B
	 */
	pMsg	= ImeDoc_pGetLastCommandEvent (pDoc) ;
	if (pMsg == NULL)
		return	LMR_RETURN ;

	if (! ImeDoc_bLookupKeymap (pDoc, pMsg, &nOrigFunc))
		nOrigFunc	= NFUNC_INVALID_CHAR ;
	if (! ImeDoc_bLookupMajorKeymap (pDoc, pMsg, &nFuncNo)) 
		nFuncNo	= NFUNC_INVALID_CHAR ;

	/*	emulate-original-map ̌Ăяo[vɂȂȂ悤ɁA
	 *	ƂƂ̌Ăяo original-map ̌ĂяovĂꍇɂ́A
	 *	ȂB
	 */
	if (nFuncNo == nOrigFunc && nFuncNo != NFUNC_INVALID_CHAR) {
		ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
		return	LMR_RETURN ;
	}

	pPC		= ImeDoc_pLookupLMState (pDoc, nFuncNo) ;
	if (pPC == NULL) {
		ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
		return	LMR_RETURN ;	/* 삪蓖ĂĂȂꍇB*/
	}
	ImeDoc_vJump (pDoc, pPC) ;
	return	LMR_CONTINUE ;
}

/*================================================================ skk-kana-clean-up
 */
static	int		LM_bSkkHenkan1_1 (struct CImeDoc*) ;

/*	ԂBm_bufRETURN, LMREGARG_RETVAL ̃RrB
 */
int
LM_bSkkHenkan1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPCWSTR				wstrNewWord	= NULL ;
	int					nRetval		= 0 ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (pBuffer->m_iSkkHenkanCount == 0) {
		/* kakutei jisyo  search ȂƂ̂́cǂ䂷ׂH */

		if (pBuffer->m_pSkkCurrentSearchProgSession != NULL) {
			do {
				wstrNewWord		= TSearchSession_pGetCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;
				if (wstrNewWord != NULL)
					break ;
			}	while (TSearchSession_bNextCandidate (pBuffer->m_pSkkCurrentSearchProgSession)) ;

			if (wstrNewWord != NULL && pBuffer->m_bSkkKakuteiFlag) {
				ImeDoc_bSetLastCommand (pBuffer->m_pDoc, NFUNC_SKK_KAKUTEI_HENKAN) ;
			}
		}
	} else {
		if (pBuffer->m_pSkkCurrentSearchProgSession != NULL) {
			int		iCount ;

			while (TSearchSession_bNextCandidate (pBuffer->m_pSkkCurrentSearchProgSession)) {
				wstrNewWord		= TSearchSession_pGetCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;
				if (wstrNewWord != NULL)
					break ;
			}
			/* ͏uĂB*/
			iCount	= ImeConfig_iGetCountHenkanShowChange () ;
			if (iCount < SHOWCANDLIST_COUNT_NOSHOW) {
				if (wstrNewWord != NULL && pBuffer->m_iSkkHenkanCount > iCount) {
					ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates) ;
					/* LMREGARG_RETVAL  LM_bSkkHenkanShowCandidates ̒l̂܂܎gB*/
					return	LMR_CONTINUE ;
				}
			}
		}
	}
	if (wstrNewWord != NULL) {
		LPWSTR	pwRetbuff ;
		int		nNewWordLen	= lstrlenW (wstrNewWord) ;
		int		nRetbuffSize ;

		pwRetbuff	= ImeDoc_pGetReturnBuffer (pDoc, &nRetbuffSize) ;
		if (pwRetbuff == NULL)
			return	LMR_ERROR ;

		nRetval	= (nNewWordLen < nRetbuffSize)? nNewWordLen : nRetbuffSize ;
		if (nRetval > 0)
			memcpy (pwRetbuff, wstrNewWord, nRetval * sizeof (WCHAR)) ;
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_RETVAL, pwRetbuff, nRetval) ;
	} else {
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_RETVAL, NULL, 0) ;
	}
	return	LMR_RETURN ;
}

/*================================================================ skk-henkan-in-minibuff
 */
static	int		LM_bSkkHenkanInMinibuff_1		(struct CImeDoc*) ;
static	int		LM_bSkkHenkanInMinibuff_S2		(struct CImeDoc*) ;
static	int		LM_bSkkHenkanInMinibuff_S3		(struct CImeDoc*) ;
static	int		LM_bSkkHenkanInMinibuff_S4		(struct CImeDoc*) ;
static	int		LM_bSkkHenkanInMinibuff_Exit	(struct CImeDoc*) ;

/*	ƂāALMREGARG_0 ɓ͌ʂ̕i[邽߂̗̈^邱ƁB
 *	ԂĺA͌ʂ̕񒷂 LMREGARG_RETVAL ɐ^œB
 */
int
LM_bSkkHenkanInMinibuff (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPWSTR	pwText ;
	int		nText, iDepth ;
	WCHAR	bufKeyword [MAXCOMPLEN] ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_RETVAL, -1) ;
		return	LMR_RETURN ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_RETVAL, -1) ;
		return	LMR_RETURN ;
	}
	pwText	= (LPWSTR) ImeDoc_pAlloca (pDoc, MAXCOMPLEN * sizeof (WCHAR), __alignof (WCHAR)) ;
	if (pwText == NULL)
		return	LMR_ERROR ;
	if (! ImeDoc_bPushReg (pDoc, LMREG_0) ||
		! ImeDoc_bPushReg (pDoc, LMREG_1) ||
		! ImeDoc_bPushReg (pDoc, LMREG_2))
		return	LMR_ERROR ;
	ImeDoc_vSetRegString (pDoc, LMREG_0, pwText, MAXCOMPLEN) ;

	if (ImeConfig_bSkkNumericConversionp () && TSearchSession_bNumericp (pBuffer->m_pSkkCurrentSearchProgSession)) {
		nText	= iSkkNumComputeHenkanKey (pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen, bufKeyword, MAXCOMPLEN, NULL, 0, ImeConfig_bSkkNumConvertFloatp ()) ;
	} else if (pBuffer->m_nSkkOkuriCharLen > 0) {
		nText	= iSkkComputeHenkanKey2 (bufKeyword, MAXCOMPLEN, pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen, pBuffer->m_bufSkkHenkanOkurigana, pBuffer->m_nSkkHenkanOkuriganaLen) ;
	} else {
		nText	= (pBuffer->m_nSkkHenkanKeyLen < MAXCOMPLEN)? pBuffer->m_nSkkHenkanKeyLen : MAXCOMPLEN ;
		memcpy (bufKeyword, pBuffer->m_bufSkkHenkanKey, nText * sizeof (WCHAR)) ;
	}
	if (nText < MAXCOMPLEN) {
		bufKeyword [nText]	= L'\0' ;
	} else {
		bufKeyword [MAXCOMPLEN - 1]	= L'\0' ;
	}
	iDepth	= ImeDoc_iGetMinibufferDepth (pDoc) ;
	if (iDepth > 8)
		iDepth	= 8 ;
	nText	= wnsprintfW (pwText, MAXCOMPLEN, L"%so^%s %s ", L"[[[[[[[[[" + (8-iDepth), L"]]]]]]]]]" + (8-iDepth), bufKeyword) ;

	ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwText, nText) ;
	ImeDoc_vSetRegBool (pDoc, LMREGARG_1, pBuffer->m_bSkkJMode) ;
	if (! ImeDoc_bCall (pDoc, LM_bReadFromMinibuffer, LM_bSkkHenkanInMinibuff_1))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkanInMinibuff_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*		pBuffer ;
	LPCWSTR					pwRead = NULL ;
	int						nRead ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (ImeDoc_bSignalp (pDoc)) {
		if (ImeDoc_nGetSignal (pDoc) != LMSIGNAL_QUIT) {
			ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
			return	LMR_CONTINUE ;
		} else {
			/* condition-case  signal Ă̂ŁB*/
			ImeDoc_vClearSignals (pDoc) ;
			nRead	= 0 ;
		}
	} else {
		if (! ImeDoc_bGetRegConstString (pDoc, LMREGARG_RETVAL, &pwRead, &nRead))
			nRead	= -1 ;
	}
	if (nRead < 0) {
		/* buffer 쐬sH */
		ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
		return	LMR_CONTINUE ;
	}

	if (ImeDoc_iGetLastCommand (pDoc) == NFUNC_ABORT_RECURSIVE_EDIT) {
		/* quit */
		nRead	= 0 ;
	}
	if (nRead == 0) {
		if (pBuffer->m_bSkkHenkanShowCandidatesMode) { /* skk-exit-show-candidates */
		} else {
			pBuffer->m_iSkkHenkanCount	-- ;
		}
		if (pBuffer->m_iSkkHenkanCount == -1) {
			pBuffer->m_nSkkHenkanOkuriganaLen	= 0 ;
			pBuffer->m_bSkkOkurigana			= LFALSE ;
			pBuffer->m_nSkkOkuriCharLen			= 0 ;
			imeBuffer_bSkkChangeMarkerToWhite (pBuffer) ;
		}
	} else {
		LPCWSTR	wptr ;
		int		nptr ;

		wptr	= pwRead + nRead - 1 ;
		nptr	= nRead ;
		while (nptr > 0 && 
#if defined (UNICODE) || defined (_UNICODE)
			(*wptr == TEXT(' ') || *wptr == TEXT('@'))
#else
			(*wptr == L' ' || *wptr == L'@')
#endif
			) {
			wptr	-- ;
			nptr	-- ;
		}
		nRead	= wptr - pwRead + 1 ;

		/*	 #4 ̏Ȃ΂ȂȂ̂Řb͂ȒPł͂ȂB#4 ̏A#4  replace
		 *	 candidate  insert ܂ŎsȂ΂ȂȂB
		 */
		/*	#4 ܂܂ꂽł邩H Aconcat ň͂܂ꂽ #4 ͖ȂƂȂ̂H
		 *	͊ȒPł͂Ȃ̂cBƂAconcat ̌ʂXɐlϊƂ combo ͂̂H
		 *	 concat (lisp Lq) #4 ͑gݍ킹ȂƂuƂɂB
		 */
		if (nRead > 0 && TSearchSession_bNumericp (pBuffer->m_pSkkCurrentSearchProgSession) &&  !bStringLispFormatp (pwRead, nRead)) {
			LPWSTR		pwText, pwDest, pwDestEnd ;
			LPCWSTR		pwNumeric, pwEnd ;
			/*	ʂ #4 ݂̑`FbNł邩B
			 */
			/*	MAXCOMPLEN ȏ͓͂̕łȂ̂ŁAwpText Ƃ MAXCOMPLEN mۂ
			 *	ΖȂƎvB
			 */
			pwText		= ImeDoc_pAlloca (pDoc, sizeof (WCHAR) * (nRead + MAXCOMPLEN + 1), __alignof (WCHAR)) ;
			if (pwText == NULL)
				return	LMR_ERROR ;	/* critical error */

			pwNumeric	= TSearchSession_pGetNumericList (pBuffer->m_pSkkCurrentSearchProgSession) ;
			pwDest		= pwText + nRead ;
			pwDestEnd	= pwDest + MAXCOMPLEN ;
			wptr		= pwRead ;
			pwEnd		= pwRead + nRead ;
			while (wptr < pwEnd) {
				if (*wptr == L'#') {
					LPCWSTR	wptrBase ;

					wptr	++ ;
					/*	̏ꍇɂ #4 łȂɊ֌WȂ numeric-list 1 skip ȂƂȂB
					 */
					wptrBase	= wptr ;
					while (wptr < pwEnd && (L'0' <= *wptr && *wptr <= L'9'))
						wptr	++ ;
					if (wptr == (wptrBase + 1) && wptrBase < pwEnd && *wptrBase == L'4') {
						/*	̍ځB*/
						/*	Ή鐔l݂ȂΖB*/
						if (*pwNumeric != L'\0') {
							LPWSTR	wptrDestBak	= pwDest ;
							/*	wptrNumeric ɑ΂ĐVɐlo^ȂƂȂB*/
							vCopyString (&pwDest, pwDestEnd, pwNumeric) ;
							if (pwDest >= pwDestEnd) {
								/*	obt@I[o[ꍇɁAI[o[钼O܂ł͔F߂̂A
								 *	SSĵH
								 */
								pwDest	= wptrDestBak ;
								break ;
							}
							*pwDest ++	= L'\0' ;
						}
					}
					/*	wptrNumeric ̍ڂֈړB
					 */
					if (*pwNumeric != L'\0') {
						pwNumeric	+= lstrlenW (pwNumeric) + 1 ;
					}
				} else {
					wptr	++ ;
				}
			}
			if (pwDest < pwDestEnd) {
				*pwDest ++	= L'\0' ;

				/*	#4 o^̏ւƕ򂷂B*/
				TS4Mapping_vClearList (pBuffer->m_plstSkkS4NumericMapping) ;
				pBuffer->m_plstSkkS4NumericMapping	= NULL ;

				memcpy (pwText, pwRead, nRead * sizeof (WCHAR)) ;
				ImeDoc_vSetRegConstString (pDoc, LMREG_1, pwText, nRead) ;
				ImeDoc_vSetRegConstString (pDoc, LMREG_2, pwText + nRead, pwDest - (pwText + nRead)) ;
				ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_S2) ;
				return	LMR_CONTINUE ;
			}
		}

		/* (setq skk-henkan-list (nconc skk-henkan-list (list new-one))) B*/
		if (nRead > 0 && pBuffer->m_pSkkCurrentSearchProgSession != NULL) {
			/*	VKɌ insert 鎞 #4 expand  control ł邾낤H ZKv낤B
			 */
			if (! TSearchSession_bInsertCandidate (pBuffer->m_pSkkCurrentSearchProgSession, pwRead, nRead)) {
				ImeDoc_vSetSignalError (pDoc) ;
				ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
				return	LMR_CONTINUE ;
			}
		}
		pBuffer->m_bSkkKakuteiFlag	= TRUE ;
	}
	ImeDoc_vSetRegConstString (pDoc, LMREGARG_RETVAL, pwRead, nRead) ;
	ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkanInMinibuff_Exit (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) {
		TS4Mapping_vClearList (pBuffer->m_plstSkkS4NumericMapping) ;
		pBuffer->m_plstSkkS4NumericMapping	= NULL ;
	}
	ImeDoc_vPopReg (pDoc, LMREG_2) ;
	ImeDoc_vPopReg (pDoc, LMREG_1) ;
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	return	LMR_RETURN ;
}

int
LM_bSkkHenkanInMinibuff_S2 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPWSTR		pwText ;
	LPCWSTR		pwNumeric ;
	int			nTextSize, nNumeric, iDepth, nText ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegString (pDoc, LMREG_0, &pwText, &nTextSize)) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegConstString (pDoc, LMREG_2, &pwNumeric, &nNumeric)) {
		pwNumeric	= NULL ;
		nNumeric	= 0 ;
	}
	if (nNumeric <= 0 || pwNumeric == NULL || *pwNumeric == L'\0') {
		ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_S4) ;
		return	LMR_CONTINUE ;
	}

	/*	"[#4o^] %s "
	 */
	iDepth	= ImeDoc_iGetMinibufferDepth (pDoc) ;
	if (iDepth > 8)
		iDepth	= 8 ;
	nText	= wnsprintfW (pwText, nTextSize, L"%so^(#4)%s %s ", L"[[[[[[[[[" + (8-iDepth), L"]]]]]]]]]" + (8-iDepth), pwNumeric) ;
	ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwText, nText) ;

	ImeDoc_vSetRegBool (pDoc, LMREGARG_1, pBuffer->m_bSkkJMode) ;
	if (! ImeDoc_bCall (pDoc, LM_bReadFromMinibuffer, LM_bSkkHenkanInMinibuff_S3))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkanInMinibuff_S3 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*		pBuffer ;
	LPCWSTR					pwRead = NULL ;
	int						nRead ;
	LPCWSTR					pwNumeric ;
	int						nNumericSize, nNumericLen ;
	struct CTS4Mapping*		pNode ;

	/*	 quit ̔܂ł́A bSkkHenkanInMinibuff_1 ƓłB
	 */
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (ImeDoc_bSignalp (pDoc)) {
		if (ImeDoc_nGetSignal (pDoc) != LMSIGNAL_QUIT) {
			ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
			return	LMR_CONTINUE ;
		} else {
			/* condition-case  signal Ă̂ŁB*/
			ImeDoc_vClearSignals (pDoc) ;
			nRead	= 0 ;
		}
	} else {
		if (! ImeDoc_bGetRegConstString (pDoc, LMREGARG_RETVAL, &pwRead, &nRead))
			nRead	= -1 ;
	}

	if (! ImeDoc_bGetRegConstString (pDoc, LMREG_2, &pwNumeric, &nNumericSize)) {
		pwNumeric		= NULL ;
		nNumericSize	= 0 ;
	}

	if (nRead < 0 || nNumericSize <= 0) {
		/* buffer 쐬sH */
		ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
		return	LMR_CONTINUE ;
	}

	if (ImeDoc_iGetLastCommand (pDoc) == NFUNC_ABORT_RECURSIVE_EDIT) {
		/* quit */
		nRead	= 0 ;
	}

	/*	 #4 B
	 */
	if (nRead > 0) {
		LPCWSTR	wptr ;
		int		nptr ;

		wptr	= pwRead + nRead - 1 ;
		nptr	= nRead ;
		while (nptr > 0 && 
#if defined (UNICODE) || defined (_UNICODE)
			(*wptr == TEXT(' ') || *wptr == TEXT('@'))
#else
			(*wptr == L' ' || *wptr == L'@')
#endif
			) {
			wptr	-- ;
			nptr	-- ;
		}
		nRead	= wptr - pwRead + 1 ;
	}

	/*	#4  mapping LB*/
	pNode	= TS4Mapping_pCreate (pwNumeric, pwRead, nRead) ;
	if (pNode == NULL) {
		/* fatal */
		return	LMR_ERROR ;
	}
	/*	buffer-local variable ƂƂŁApbuffer ̈ꕔƂċLB
	 *	REG ̏ゾƁAꂸɂЂǂƂɂȂ肻̂ŁB
	 */
	if (! TS4Mapping_bInsertLast (&pBuffer->m_plstSkkS4NumericMapping, pNode)) {
		TS4Mapping_vClearList (pBuffer->m_plstSkkS4NumericMapping) ;
		pBuffer->m_plstSkkS4NumericMapping	= NULL ;
		return	LMR_ERROR ;
	}
	
	/*	 numeric ւƈړB
	 */
	nNumericLen		= lstrlenW (pwNumeric) ;
	pwNumeric		+= (nNumericLen + 1) ;
	nNumericSize	-= (nNumericLen + 1) ;
	if (nNumericSize <= 0) {
		ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_S4) ;
	} else {
		ImeDoc_vSetRegConstString (pDoc, LMREG_2, pwNumeric, nNumericSize) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_S2) ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkanInMinibuff_S4 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPCWSTR		pwRead ;
	int			nRead ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
		return	LMR_CONTINUE ;
	}

	if (! ImeDoc_bGetRegConstString (pDoc, LMREG_1, &pwRead, &nRead)) {
		pwRead	= NULL ;
		nRead	= 0 ;
	}

	/* (setq skk-henkan-list (nconc skk-henkan-list (list new-one))) B*/
	if (nRead > 0 && pBuffer->m_pSkkCurrentSearchProgSession != NULL) {
		BOOL	bFound	= FALSE ;

		/*	VKɌ insert 鎞 #4 expand  control ł邾낤H ZKv낤B
		 */
		if (! TSearchSession_bInsertCandidateWithS4Mapping (pBuffer->m_pSkkCurrentSearchProgSession, pwRead, nRead, pBuffer->m_plstSkkS4NumericMapping)) {
			ImeDoc_vSetSignalError (pDoc) ;
			ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
			return	LMR_CONTINUE ;
		}

		/*	lϊ̖肩ǂݒKvB
//		pwRead	= TSearchSession_pGetReferCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;
//		if (pwRead == NULL)
		 */
		pwRead	= TSearchSession_pGetCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;
		if (pwRead != NULL) {
			nRead	= lstrlenW (pwRead) ;
			if (nRead > 0) {
				LPWSTR		pwRetbuf ;
				int			nRetbufSize ;

				pwRetbuf	= ImeDoc_pGetReturnBuffer (pDoc, &nRetbufSize) ;
				if (pwRetbuf != NULL && nRetbufSize > 0) {
					nRead	= MIN (nRetbufSize, nRead) ;
					memcpy (pwRetbuf, pwRead, nRead * sizeof (WCHAR)) ;
					pwRead	= pwRetbuf ;
					bFound	= TRUE ;
				}
			}
		}
		if (! bFound) {
			pwRead	= NULL ;
			nRead	= 0 ;
		}
	} else {
		pwRead	= NULL ;
		nRead	= 0 ;
	}
	ImeDoc_vSetRegConstString (pDoc, LMREGARG_RETVAL, pwRead, nRead) ;
	ImeDoc_vJump (pDoc, LM_bSkkHenkanInMinibuff_Exit) ;
	return	LMR_CONTINUE ;
}

/*================================================================ skk-kana-clean-up
 */
static	int		LM_bSkkTryCompletion_1		(struct CImeDoc*) ;
static	int		LM_bSkkTryCompletion_Exit	(struct CImeDoc*) ;

int
LM_bSkkTryCompletion (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (pBuffer->m_bSkkHenkanMode == LON) {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_0, (ImeDoc_iGetLastCommand (pDoc) != NFUNC_SKK_COMP_DO)? TRUE : FALSE) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkCompDo, LM_bSkkTryCompletion_1))
			return	LMR_ERROR ;
	} else {
		if (! ImeDoc_bCall (pDoc, LM_bSkkEmulateOriginalMap, LM_bSkkTryCompletion_Exit)) 
			return	LMR_ERROR ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkTryCompletion_1 (
	struct CImeDoc*		pDoc)
{
	if (! ImeDoc_bSignalp (pDoc)) {
		ImeDoc_bSetThisCommand (pDoc, NFUNC_SKK_COMP_DO) ;
	}
	ImeDoc_vJump (pDoc, LM_bSkkTryCompletion_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkTryCompletion_Exit (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) {
		/* skk-point-move ̌ʁB*/
		imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkPreviousPoint, MARKER_SKK_PREVIOUS_POINT, pBuffer->m_pmkPoint) ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-abbrev-comma
 */
int
LM_bSkkAbbrevComma (
	struct CImeDoc*		pDoc)
{
	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	if (ImeDoc_iGetLastCommand (pDoc) == NFUNC_SKK_COMP_DO) {
		if (! ImeDoc_bCall (pDoc, LM_bSkkCompPrevious, LM_bSkkTryCompletion_1))
			return	LMR_ERROR ;
	} else {
		if (! ImeDoc_bCall (pDoc, LM_bSkkEmulateOriginalMap, LM_bSkkTryCompletion_Exit)) 
			return	LMR_ERROR ;
	}
	return	LMR_CONTINUE ;
}

/*================================================================ skk-abbrev-period
 */
int
LM_bSkkAbbrevPeriod (
	struct CImeDoc*		pDoc)
{
	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	if (ImeDoc_iGetLastCommand (pDoc) == NFUNC_SKK_COMP_DO) {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_0, FALSE) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkCompDo, LM_bSkkTryCompletion_1))
			return	LMR_ERROR ;
	} else {
		if (! ImeDoc_bCall (pDoc, LM_bSkkEmulateOriginalMap, LM_bSkkTryCompletion_Exit)) 
			return	LMR_ERROR ;
	}
	return	LMR_CONTINUE ;
}

/*================================================================ skk-set-char-before-as-okurigana
 */
int
LM_bSkkSetCharBeforeAsOkurigana (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	BOOL				bNoSokuon	= FALSE ;
	int					iPt1, iOkuri = -1, iSokuon = -1 ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	iPt1	= TMarker_iGetPosition (pBuffer->m_pmkPoint) ;
	if (iPt1 > 0) {
		iOkuri	= pBuffer->m_bufComp [iPt1 - 1] ;
	} else {
		iOkuri	= -1 ;
	}
	if (iOkuri >= 0) {
		int		iOkuriganaStartPoint, iBufferEnd ;
		struct TMarker*	pmkBufferEnd	= NULL ;

		if (! bNoSokuon) {
			iSokuon	= (iPt1 > 1)? pBuffer->m_bufComp [iPt1 - 2] : -1 ;
			if (
#if defined (UNICODE) || defined (_UNICODE)
				iSokuon != TEXT('') && iSokuon != TEXT('b')
#else
				iSokuon != L'' && iSokuon != L'b'
#endif
				)
				iSokuon	= -1 ;
		}
		if (! ImeBuffer_bSetMarker (pBuffer, &pBuffer->m_pmkSkkOkuriganaStartPoint, MARKER_SKK_OKURIGANA_START_POINT, pBuffer->m_pmkPoint) ||
			pBuffer->m_pmkSkkOkuriganaStartPoint == NULL) {
			ImeDoc_vSetSignalError (pDoc) ;
			return	LMR_RETURN ;
		}
		(void) TMarker_bBackward (pBuffer->m_pmkSkkOkuriganaStartPoint, iSokuon >= 0? 2 : 1) ;

		iOkuriganaStartPoint	= TMarker_iGetPosition (pBuffer->m_pmkSkkOkuriganaStartPoint) ;
		if (! ImeBuffer_bGetMarker (pBuffer, MARKER_BUFFEREND, &pmkBufferEnd) || pmkBufferEnd == NULL) {
			iBufferEnd	= pBuffer->m_nbufComp ;
		} else {
			iBufferEnd	= TMarker_iGetPosition (pmkBufferEnd) ;
		}
		pBuffer->m_nSkkOkuriCharLen	= imeBuffer_iSkkOkuriganaPrefix (pBuffer, pBuffer->m_bufComp + iOkuriganaStartPoint, iBufferEnd - iOkuriganaStartPoint, pBuffer->m_bufSkkOkuriChar, ARRAYSIZE (pBuffer->m_bufSkkOkuriChar)) ;

		/*	original ɂ́A skk-okurigana  t ɂR[h͂ȂAǉB
		 *	skk-henkan ɑΉcode skk-okurigana Ă̂ŁAnil ̂܂܂ƑϊɂȂȂB
		 */
		pBuffer->m_bSkkOkurigana	= LTRUE ;

		/*	͕̏Kv낤H
		if (pBuffer->m_pSkkCurrentSearchProgSession == NULL) {
		}
		*/
		ImeDoc_vJump (pDoc, LM_bSkkSetOkuriganaAdJisx0201) ;
		return	LMR_CONTINUE ;
	}
	return	LMR_RETURN ;
}


/*================================================================ skk-toggle-kutouten
 */
int
LM_bSkkToggleKutouten (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int					iCount ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	/*	Ǔ_^Cv𑝂₵Ƃl͂̂낤H 
	 *	Wł 'en, 'jp ȂAalist ƂƂ͒ǉłƂƂŁc
	 *	Ȁꍇɂ toggle-kutoten ꎩ̂Ē`ȂƂȂǁB
	 */
	iCount	= ImeConfig_iGetNumberOfKutotens () ;
	if (iCount > 0) {
		pBuffer->m_bSkkCurrentKutotenType	= (pBuffer->m_bSkkCurrentKutotenType + 1) % iCount ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-current-kuten
 */
/*	interactive-function ł͂Ȃ̂cB
 */
int
LM_bSkkCurrentKuten (
	struct CImeDoc*		pDoc)
{
	LPCWSTR		pwKuten ;
	int			nKutenLen ;
//	LPWSTR		pwResult ;
//	int			nResultSize ;

	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
#if 0
	pwResult	= ImeDoc_pGetReturnBuffer (pDoc, &nResultSize) ;
	pwKuten		= ImeDoc_pGetCurrentKuten (pDoc, pBuffer->m_bSkkCurrentKutotenType, &nKutenLen) ;
	if (pwKuten != NULL && pwResult != NULL) {
		int		n	= MIN (nResultSize, nKutenLen) ;
		memcpy (pwResult, pwKuten, nKutenLen * sizeof (WCHAR)) ;
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_RETVAL, pwResult, nKutenLen) ;
	} else {
		ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	}
#else
	pwKuten		= ImeConfig_pGetCurrentKuten (pBuffer->m_bSkkCurrentKutotenType, &nKutenLen) ;
	if (pwKuten != NULL) {
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_RETVAL, pwKuten, nKutenLen) ;
	} else {
		ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	}
#endif
	return	LMR_RETURN ;
}

/*================================================================ skk-current-touten
 */
/*	interactive-function ł͂Ȃ̂cǂ邩B
 */
int
LM_bSkkCurrentTouten (
	struct CImeDoc*		pDoc)
{
	LPCWSTR		pwTouten ;
	int			nToutenLen ;
//	LPWSTR		pwResult ;
//	int			nResultSize ;

	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
#if 0
	pwResult	= ImeDoc_pGetReturnBuffer (pDoc, &nResultSize) ;
	pwTouten	= ImeDoc_pGetCurrentTouten (pDoc, pBuffer->m_bSkkCurrentKutotenType, &nToutenLen) ;
	if (pwTouten != NULL && pwResult != NULL) {
		int		n	= MIN (nResultSize, nToutenLen) ;
		memcpy (pwResult, pwTouten, nToutenLen * sizeof (WCHAR)) ;
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_RETVAL, pwResult, nToutenLen) ;
	} else {
		ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	}
#else
	/*	q[v pwTouten mۂĂ̂ŁAReturn Buffer ɃRs[Kv͂ȂB
	 */
	pwTouten	= ImeConfig_pGetCurrentTouten (pBuffer->m_bSkkCurrentKutotenType, &nToutenLen) ;
	if (pwTouten != NULL) {
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_RETVAL, pwTouten, nToutenLen) ;
	} else {
		ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	}
#endif
	return	LMR_RETURN ;
}

/*================================================================ skk-toggle-characters
 */
static	int		LM_bSkkToggleCharacters_1 (struct CImeDoc*) ;
static	int		LM_bSkkToggleCharacters_2 (struct CImeDoc*) ;

int
LM_bSkkToggleCharacters (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bPushReg (pDoc, LMREG_0))
		return	LMR_ERROR ;

	if (pBuffer->m_bSkkHenkanMode == LON) {
		struct TMarker*		pmkHenkanStartPoint ;
		struct TMarker*		pmkPoint ;
		int					nEndPos, nPosition, nCH ;
		int					(*pFunc)(struct CImeDoc*) ;
		BOOL				bVContract	= FALSE ;

		if (! ImeBuffer_bGetMarker (pBuffer, MARKER_SKK_HENKAN_START_POINT, &pmkHenkanStartPoint) || pmkHenkanStartPoint == NULL ||
			! ImeBuffer_bGetMarker (pBuffer, MARKER_POINT, &pmkPoint) || pmkPoint == NULL) {
			ImeDoc_vSetSignalError (pDoc) ;
			return	LMR_RETURN ;
		}
		nEndPos		= TMarker_iGetPosition (pmkPoint) ;
		nPosition	= TMarker_iGetPosition (pmkHenkanStartPoint) ;
		pFunc		= NULL ;
		while (nPosition < nEndPos && nPosition < pBuffer->m_nbufComp) {
			nCH	= pBuffer->m_bufComp [nPosition] ;
			if (bHiraganaCharp (nCH)) {
				pFunc		= &LM_bSkkKatakanaRegion ;
				bVContract	= TRUE ;
				break ;
			} else if (bKatakanaCharp (nCH)) {
				pFunc		= &LM_bSkkHiraganaRegion ;
				break ;
			} else if (bJisx0208LatinCharp (nCH)) {
				pFunc		= &LM_bSkkLatinRegion ;
				break ;
			} else if (bSkkAsciiCharp (nCH)) {
				pFunc		= &LM_bSkkJisx0208LatinRegion ;
				break ;
			}
			nPosition	++ ;
		}
		ImeDoc_vSetRegPointer (pDoc, LMREGARG_0, pFunc) ;
		ImeDoc_vSetRegBool (pDoc, LMREGARG_1, bVContract) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkHenkanSkkRegionByFunc, LM_bSkkToggleCharacters_1))
			return	LMR_ERROR ;
		return	LMR_CONTINUE ;
	} else if (ImeDoc_bRecursiveEditp (pDoc) && ! pBuffer->m_bSkkJMode) {
		imeBuffer_bSkkJModeOn (pBuffer, FALSE) ;
	} else {
		pBuffer->m_bSkkKatakana	= ! pBuffer->m_bSkkKatakana ;
	}
	ImeDoc_vJump (pDoc, LM_bSkkToggleCharacters_1) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkToggleCharacters_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) 
		ImeDoc_vSetRegBool   (pDoc, LMREG_0, pBuffer->m_bSkkKatakana) ;
	ImeDoc_vSetRegString (pDoc, LMREGARG_0, NULL, 0) ;
	if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkToggleCharacters_2))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkToggleCharacters_2 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	BOOL					bSkkKatakana ;

	if (! ImeDoc_bGetRegBool (pDoc, LMREG_0, &bSkkKatakana))
		bSkkKatakana	= FALSE ;
	ImeDoc_vPopReg (pDoc, LMREG_0) ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (pBuffer->m_bSkkJMode) {
		imeBuffer_bSkkJModeOn (pBuffer, bSkkKatakana) ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-henkan-skk-region-by-func
 */
static	int		LM_bSkkHenkanSkkRegionByFunc_1		(struct CImeDoc*) ;
static	int		LM_bSkkHenkanSkkRegionByFunc_Exit	(struct CImeDoc*) ;

int
LM_bSkkHenkanSkkRegionByFunc (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
		ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
		return	LMR_RETURN ;
	} else if (pBuffer->m_bSkkHenkanMode == LON) {
		int		(*pFunc)(struct CImeDoc*) ;
		int		nHenkanStartPoint, nPoint, nPrefix ;
		BOOL	bVContract ;
		struct CSkkRuleTreeIterator*	piteSkkRuleTree ;

		imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkHenkanEndPoint, MARKER_SKK_HENKAN_END_POINT, pBuffer->m_pmkPoint) ;
		nHenkanStartPoint	= TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) ;
		nPoint				= TMarker_iGetPosition (pBuffer->m_pmkPoint) ;
		if (nHenkanStartPoint < nPoint && 0 <= nHenkanStartPoint && nHenkanStartPoint < pBuffer->m_nbufComp &&
			bHiraganaCharp (pBuffer->m_bufComp [nHenkanStartPoint])) {
			imeBuffer_bSkkUpdateKakuteiHistory (pBuffer, pBuffer->m_bufComp + nHenkanStartPoint, nPoint - nHenkanStartPoint, NULL, 0) ;
		}
		piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;
		if (SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree)) {
//		if (TSkkRuleTreeNode_pGetPrefix (pBuffer->m_pSkkCurrentRuleTree, &nPrefix) != NULL) {
			/* ͓r̉vtBNX܂B*/
			ImeDoc_bSetMessage (pDoc, L"͓r̉vtBNX܂") ;
			ImeDoc_vSetSignalError (pDoc) ;
			ImeDoc_vJump (pDoc, LM_bSkkHenkanSkkRegionByFunc_Exit) ;
			return	LMR_CONTINUE ;
		}
		if (nPoint < nHenkanStartPoint) {
			ImeDoc_bSetMessage (pDoc, L"J[\ϊJnn_Oɂ܂") ;
			ImeDoc_vSetSignalError (pDoc) ;
			ImeDoc_vJump (pDoc, LM_bSkkHenkanSkkRegionByFunc_Exit) ;
			return	LMR_CONTINUE ;
		}
		if (! ImeConfig_bSkkAllowsSpacesNewlinesAndTabs ()) {
			int		nPos ;

			for (nPos = nHenkanStartPoint ; nPos < nPoint && nPos < pBuffer->m_nbufComp ; nPos ++) {
				int		nCH	= pBuffer->m_bufComp [nPos] ;
				if (nCH == L' ' || nCH == L'\t' || nCH == L'\n' || nCH == L'\r') {
					ImeDoc_bSetMessage (pDoc, L"ϊL[ɉs܂܂Ă܂") ;
					ImeDoc_vSetSignalError (pDoc) ;
					ImeDoc_vJump (pDoc, LM_bSkkHenkanSkkRegionByFunc_Exit) ;
					return	LMR_CONTINUE ;
				}
			}
		}
		if (! ImeDoc_bGetRegPointer (pDoc, LMREGARG_0, (void**)&pFunc)) {
			ImeDoc_vSetSignalError (pDoc) ;
			ImeDoc_vJump (pDoc, LM_bSkkHenkanSkkRegionByFunc_Exit) ;
			return	LMR_CONTINUE ;
		}
		if (! ImeDoc_bGetRegBool (pDoc, LMREGARG_1, &bVContract))
			bVContract	= FALSE ;
		if (pFunc != NULL) {
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_0, nHenkanStartPoint) ;
			ImeDoc_vSetRegInteger (pDoc, LMREGARG_1, nPoint) ;
			ImeDoc_vSetRegBool (pDoc, LMREGARG_2, bVContract) ;
			if (! ImeDoc_bCall (pDoc, pFunc, LM_bSkkHenkanSkkRegionByFunc_1))
				return	LMR_ERROR ;
		} else {
			ImeDoc_vJump (pDoc, LM_bSkkHenkanSkkRegionByFunc_1) ;
		}
		return	LMR_CONTINUE ;
	} else {
		if (! ImeDoc_bCall (pDoc, LM_bSkkEmulateOriginalMap, LM_bSkkHenkanSkkRegionByFunc_Exit))
			return	LMR_ERROR ;
		return	LMR_CONTINUE ;
	}
}

int
LM_bSkkHenkanSkkRegionByFunc_1 (
	struct CImeDoc*			pDoc)
{
	if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkHenkanSkkRegionByFunc_Exit))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkanSkkRegionByFunc_Exit (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	/* skk-point-move ̌ʁB*/
	imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkPreviousPoint, MARKER_SKK_PREVIOUS_POINT, pBuffer->m_pmkPoint) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-katakana-region
 */
int
LM_bSkkKatakanaRegion (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int		nStartPos, nEndPos, nLength, nReplaceText ;
	WCHAR	bufTemp [MAXCOMPLEN] ;
	BOOL	bVContract	= FALSE ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bGetRegInteger (pDoc, LMREGARG_0, &nStartPos) || ! ImeDoc_bGetRegInteger (pDoc, LMREGARG_1, &nEndPos)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! imeBuffer_bValidRegionp (pBuffer, nStartPos, nEndPos)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bGetRegBool (pDoc, LMREGARG_2, &bVContract))
		bVContract	= FALSE ;

	nLength	= iSkkHiraganaToKatakana (bufTemp, ARRAYSIZE (bufTemp), pBuffer->m_bufComp + nStartPos, nEndPos - nStartPos, bVContract) ;

	nReplaceText	= nEndPos - nStartPos ;
	if (nLength < nReplaceText) {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nLength) ;
		ImeBuffer_bDeleteRegion (pBuffer, nStartPos + nLength, nEndPos) ;
	} else if (nLength > nReplaceText) {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nReplaceText) ;
		ImeBuffer_iInsertByPosition (pBuffer, nEndPos, bufTemp + nReplaceText, nLength - nReplaceText) ;
	} else {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nLength) ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-hiragana-region
 */
int
LM_bSkkHiraganaRegion (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int		nStartPos, nEndPos, nLength, nReplaceText ;
	WCHAR	bufTemp [MAXCOMPLEN] ;
	BOOL	bVExtract ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bGetRegInteger (pDoc, LMREGARG_0, &nStartPos) || ! ImeDoc_bGetRegInteger (pDoc, LMREGARG_1, &nEndPos)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! imeBuffer_bValidRegionp (pBuffer, nStartPos, nEndPos)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bGetRegBool (pDoc, LMREGARG_2, &bVExtract))
		bVExtract	= FALSE ;
	nLength	= iSkkKatakanaToHiragana (bufTemp, ARRAYSIZE (bufTemp), pBuffer->m_bufComp + nStartPos, nEndPos - nStartPos, bVExtract) ;

	nReplaceText	= nEndPos - nStartPos ;
	if (nLength < nReplaceText) {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nLength) ;
		ImeBuffer_bDeleteRegion (pBuffer, nStartPos + nLength, nEndPos) ;
	} else if (nLength > nReplaceText) {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nReplaceText) ;
		ImeBuffer_iInsertByPosition (pBuffer, nEndPos, bufTemp + nReplaceText, nLength - nReplaceText) ;
	} else {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nLength) ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-jisx0208-latin-region
 */
int
LM_bSkkJisx0208LatinRegion (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int		nStartPos, nEndPos, nLength, nReplaceText ;
	WCHAR	bufTemp [MAXCOMPLEN] ;
	LPWSTR	pDest, pDestEnd ;
	LPCWSTR	pSrc, pSrcEnd ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bGetRegInteger (pDoc, LMREGARG_0, &nStartPos) || ! ImeDoc_bGetRegInteger (pDoc, LMREGARG_1, &nEndPos)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! imeBuffer_bValidRegionp (pBuffer, nStartPos, nEndPos)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	pDest		= bufTemp ;
	pDestEnd	= bufTemp + ARRAYSIZE (bufTemp) ;
	pSrc		= pBuffer->m_bufComp + nStartPos ;
	pSrcEnd		= pBuffer->m_bufComp + nEndPos ; 
	while (pSrc < pSrcEnd && pDest < pDestEnd) {
		if (L' ' <= *pSrc && *pSrc <= L'~') {
			WCHAR	bufCH [MAXCOMPLEN] ;
			int		nLen	= ImeConfig_iSkkARefSkkJisx0208LatinVector (*pSrc, bufCH, ARRAYSIZE (bufCH)) ;

			if (nLen >= 0) {
				LPCWSTR	ptr ;

				ptr		= bufCH ;
				while (pDest < pDestEnd && nLen > 0) {
					*pDest ++	= *ptr ++ ;
				}
			} else {
				*pDest ++	= *pSrc ++ ;
			}
		} else {
			*pDest ++	= *pSrc ++ ;
		}
	}

	nLength			= pDest - bufTemp ;
	nReplaceText	= nEndPos - nStartPos ;
	if (nLength < nReplaceText) {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nLength) ;
		ImeBuffer_bDeleteRegion (pBuffer, nStartPos + nLength, nEndPos) ;
	} else if (nLength > nReplaceText) {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nReplaceText) ;
		ImeBuffer_iInsertByPosition (pBuffer, nEndPos, bufTemp + nReplaceText, nLength - nReplaceText) ;
	} else {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nLength) ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-latin-region
 */
int
LM_bSkkLatinRegion (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int		nStartPos, nEndPos, nLength, nReplaceText ;
	WCHAR	bufTemp [MAXCOMPLEN] ;
	LPWSTR	pDest, pDestEnd ;
	LPCWSTR	pSrc,  pSrcEnd ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bGetRegInteger (pDoc, LMREGARG_0, &nStartPos) || ! ImeDoc_bGetRegInteger (pDoc, LMREGARG_1, &nEndPos)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! imeBuffer_bValidRegionp (pBuffer, nStartPos, nEndPos)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	pDest		= bufTemp ;
	pDestEnd	= bufTemp + ARRAYSIZE (bufTemp) ;
	pSrc		= pBuffer->m_bufComp + nStartPos ;
	pSrcEnd		= pBuffer->m_bufComp + nEndPos ; 
	while (pSrc < pSrcEnd && pDest < pDestEnd) {
		if (L'@' <= *pSrc && *pSrc <= L'`') {
			WCHAR	bufCH [16] ;
			int		nLen ;

			/*	skk-jisx0208-latin-vector tɈB*/
			nLen	= ImeConfig_iSkkRevRefSkkJisx0208LatinVector (*pSrc, bufCH, ARRAYSIZE (bufCH)) ;
			if (nLen >= 0) {
				LPCWSTR	ptr ;

				ptr		= bufCH ;
				while (pDest < pDestEnd && nLen > 0) {
					*pDest ++	= *ptr ++ ;
				}
			} else {
				*pDest ++	= *pSrc ++ ;
			}
		} else {
			*pDest ++	= *pSrc ++ ;
		}
	}

	nLength			= pDest - bufTemp ;
	nReplaceText	= nEndPos - nStartPos ;
	if (nLength < nReplaceText) {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nLength) ;
		ImeBuffer_bDeleteRegion (pBuffer, nStartPos + nLength, nEndPos) ;
	} else if (nLength > nReplaceText) {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nReplaceText) ;
		ImeBuffer_iInsertByPosition (pBuffer, nEndPos, bufTemp + nReplaceText, nLength - nReplaceText) ;
	} else {
		ImeBuffer_iOverwriteByPosition (pBuffer, nStartPos, bufTemp, nLength) ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-jisx0208-latin-insert
 */
static	int		LM_bSkkJisx0208LatinInsert_1	(struct CImeDoc*) ;
static	int		LM_bSkkJisx0208LatinInsert_2	(struct CImeDoc*) ;
static	int		LM_bSkkJisx0208LatinInsert_Exit	(struct CImeDoc*) ;

int
LM_bSkkJisx0208LatinInsert (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	WCHAR				bufResult [MAXCOMPLEN] ;
	LPWSTR				pwResult ;
	int					iCH ;
	int					nResult ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	iCH		= ImeDoc_iGetLastCommandChar (pDoc) ;
	nResult	= ImeConfig_iSkkARefSkkJisx0208LatinVector (iCH, bufResult, ARRAYSIZE (bufResult)) ;
	if (nResult < 0) {
		if (! ImeDoc_bCall (pDoc, LM_bSkkEmulateOriginalMap, LM_bSkkJisx0208LatinInsert_Exit))
			return	LMR_ERROR ;
		return	LMR_CONTINUE ;
	}
	pwResult	= ImeDoc_pAlloca (pDoc, nResult * sizeof (WCHAR), __alignof (WCHAR)) ;
	if (pwResult == NULL)
		return	LMR_ERROR ;
	if (! ImeDoc_bPushReg (pDoc, LMREG_0))
		return	LMR_ERROR ;

	memcpy (pwResult, bufResult, sizeof (WCHAR) * nResult) ;
	ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwResult, nResult) ;
	ImeDoc_vSetRegConstString (pDoc, LMREG_0,    pwResult, nResult) ;
	if (! ImeDoc_bCall (pDoc, LM_bSkkInsertStr, LM_bSkkJisx0208LatinInsert_1))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkJisx0208LatinInsert_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPCWSTR				pwResult ;
	int					nResult ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkJisx0208LatinInsert_Exit) ;
		return	LMR_CONTINUE ;
	}

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkJisx0208LatinInsert_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegConstString (pDoc, LMREG_0, &pwResult, &nResult)) {
		pwResult	= NULL ;
		nResult		= 0 ;
	}

	/*	yÁcB܂Lł͂Ȃ̂ǂ͂ĂB
	 */
	if (nResult > 0 && ImeConfig_bSkkAutoInsertParen ()) {
		WCHAR	bufPairStr [MAXCOMPLEN] ;
		int		nPairStr ;

		nPairStr	= ImeConfig_iSkkAssocSkkAutoParenStringAlist (pwResult, nResult, bufPairStr, ARRAYSIZE (bufPairStr)) ;
		if (nPairStr > 0) {
			LPWSTR			pwPairStr ;
			struct TMarker*	pmkPoint ;
			struct TMarker*	pmkMark ;

			pwPairStr	= ImeDoc_pAlloca (pDoc, nPairStr * sizeof (WCHAR), __alignof (WCHAR)) ;
			if (pwPairStr == NULL)
				return	LMR_ERROR ;

			pmkMark		= ImeBuffer_pMakeMarker (pBuffer, FALSE) ;
			if (pmkMark == NULL || ! ImeBuffer_bGetMarker (pBuffer, MARKER_POINT, &pmkPoint) || pmkPoint == NULL) {
				ImeDoc_vSetSignalError (pDoc) ;
				ImeDoc_vJump (pDoc, LM_bSkkJisx0208LatinInsert_Exit) ;
				return	LMR_CONTINUE ;
			}
			TMarker_bSetPosition (pmkMark, pmkPoint) ;

			memcpy (pwPairStr, bufPairStr, sizeof (WCHAR) * nPairStr) ;
			ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwPairStr, nPairStr) ;
			ImeDoc_vSetRegPointer (pDoc, LMREG_0, pmkMark) ;

			/*	̃R[h insert 񐔂𐔂āAJ[\ backward Ă邪cB insert
			 *	񐔂́Aۂ insert ꂽʂƈقȂ\ČB
			 *	marker g cursor ʒuL̂ǂB
			 */
			if (! ImeDoc_bCall (pDoc, LM_bSkkInsertStr, LM_bSkkJisx0208LatinInsert_2))
				return	LMR_ERROR ;
			return	LMR_CONTINUE ;
		}
	}
	ImeDoc_vJump (pDoc, LM_bSkkJisx0208LatinInsert_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkJisx0208LatinInsert_2 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	struct TMarker*		pmkMark ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		/*  buffer s݂ł̓}[J̔jłȂB*/
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkJisx0208LatinInsert_Exit) ;
		return	LMR_CONTINUE ;
	}
	/*	signal Ԃł낤ƂȂ낤ƁAmarker jȂƂ܂B
	 */
	if (! ImeDoc_bGetRegPointer (pDoc, LMREG_0, (void**)&pmkMark)) {
		pmkMark	= NULL ;
	}
	if (pmkMark != NULL) {
		TMarker_bSetPosition (pBuffer->m_pmkPoint, pmkMark) ;
		ImeBuffer_bDeleteMarker (pBuffer, pmkMark) ;
	}
	ImeDoc_vJump (pDoc, LM_bSkkJisx0208LatinInsert_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkJisx0208LatinInsert_Exit (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) {
		/* skk-point-move ̌ʁB*/
		imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkPreviousPoint, MARKER_SKK_PREVIOUS_POINT, pBuffer->m_pmkPoint) ;
	}
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-kana-clean-up
 */
static	int		LM_bSkkKanaCleanup_1 (struct CImeDoc*) ;
static	int		LM_bSkkKanaCleanup_2 (struct CImeDoc*) ;

int
LM_bSkkKanaCleanup (
	struct CImeDoc*		pDoc)
{
	const struct CSkkRuleTreeNodeOutput*	pData	= NULL ;
	struct CImeBuffer*	pBuffer ;
	LPCWSTR				wstrKana ;
	int					nKanaLen, nNextState ;
	BOOL				bForce ;
	struct CSkkRuleTreeIterator*	piteSkkRuleTree ;

	if (ImeDoc_bSignalp (pDoc)) {
		return	LMR_RETURN ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	/*	Register oRňn̂cH
	 */
	if (! ImeDoc_bGetRegBool (pDoc, LMREGARG_0, &bForce)) {
		bForce	= FALSE ;
	}

	piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;
	/* 	 BOOL bForce ǂĎ󂯎̂cB*/
	if (SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree) && ! SkkRuleTreeIterator_bHaveNextState (piteSkkRuleTree)) {
//	if (pBuffer->m_pSkkCurrentRuleTree != NULL && TSkkRuleTreeNode_pGetNextState (pBuffer->m_pSkkCurrentRuleTree, &nNextState) == NULL) {
		pData	= SkkRuleTreeIterator_pGetOutput (piteSkkRuleTree) ;
//		pData	= TSkkRuleTreeNode_pGetOutput (pBuffer->m_pSkkCurrentRuleTree) ;
	} else {
		pData	= NULL ;
	}
	if (bForce || pData != NULL) {
		imeBuffer_bSkkErasePrefix (pBuffer, TRUE) ;

		wstrKana	= NULL ;
		nKanaLen	= 0 ;
		if (pData != NULL) {
			int		nType	= TSkkRuleTreeNodeOutput_iGetType (pData) ;
			if (nType == NTYPE_MACRO) {
				int		nFuncNo ;
				PLMFUNC	pPC	= NULL ;

				/*	ށAskk-kana-cleanup ܂̂B execute ȂƂȂ̂ŁAxԂȂ΁cB
				 */
				TSkkRuleTreeNodeOutput_bGetMacro (pData, &nFuncNo) ;
				pPC	= ImeDoc_pLookupLMState (pDoc, nFuncNo) ;
				if (pPC == NULL) {
					ImeDoc_vSetSignalError (pDoc) ;
					return	LMR_RETURN ;
				}
				if (! ImeDoc_bCall (pDoc, pPC, LM_bSkkKanaCleanup_1))
					return	LMR_ERROR ;
				return	LMR_CONTINUE ;
			} else {
				if (nType == NTYPE_STRINGPAIR) {
					if (pBuffer->m_bSkkKatakana) {
						TSkkRuleTreeNodeOutput_bGetRString (pData, &wstrKana, &nKanaLen) ;
					} else {
						TSkkRuleTreeNodeOutput_bGetLString (pData, &wstrKana, &nKanaLen) ;
					}
				} else if (nType == NTYPE_STRING) {
					TSkkRuleTreeNodeOutput_bGetString (pData, &wstrKana, &nKanaLen) ;
				}
			}
		}
		if (wstrKana != NULL && nKanaLen > 0) {
			LPWSTR	pwString ;

			pwString	= ImeDoc_pAlloca (pDoc, nKanaLen * sizeof (WCHAR), __alignof (WCHAR)) ;
			if (pwString == NULL) 
				return	LMR_ERROR ;
			memcpy (pwString, wstrKana, nKanaLen * sizeof (WCHAR)) ;
			ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwString, nKanaLen) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkInsertStr, LM_bSkkKanaCleanup_2))
				return	LMR_ERROR ;
			return	LMR_CONTINUE ;
		}
		pBuffer->m_pmkSkkKanaStartPoint	= NULL ;
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

int
LM_bSkkKanaCleanup_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPCWSTR				pwString, pwLeft, pwRight ;
	int					nStringLen, nLeft, nRight, nResultLen ;
	LPWSTR				pwResult ;

	if (ImeDoc_bSignalp (pDoc)) {
		return	LMR_RETURN ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	/*	Macro s̏B{͂̌ʂ string Ł`Ƃ\B
	 */
	if (ImeDoc_bGetRegConstString (pDoc, LMREGARG_RETVAL, &pwString, &nStringLen)) {
		/* ... */
	} else if (ImeDoc_bGetRegConstStringPair (pDoc, LMREGARG_RETVAL, &pwLeft, &nLeft, &pwRight, &nRight)) {
		pwString	= (pBuffer->m_bSkkKatakana)? pwRight : pwLeft ;
		nStringLen	= (pBuffer->m_bSkkKatakana)? nRight  : nLeft ;
	} else {
		pwString	= NULL ;
		nStringLen	= 0 ;
	}
	if (pwString != NULL && nStringLen > 0) {
		pwResult	= ImeDoc_pAlloca (pDoc, nStringLen * sizeof (WCHAR), __alignof (WCHAR)) ;
		if (pwResult == NULL)
			return	LMR_ERROR ;
		memcpy (pwResult, pwString, nStringLen * sizeof (WCHAR)) ;
		nResultLen	= nStringLen ;
	} else {
		pwResult	= NULL ;
		nResultLen	= 0 ;
	}
	ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwResult, nResultLen) ;
	if (! ImeDoc_bCall (pDoc, LM_bSkkInsertStr, LM_bSkkKanaCleanup_2))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKanaCleanup_2 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) {
		return	LMR_RETURN ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	pBuffer->m_pmkSkkKanaStartPoint	= NULL ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-insert-str
 */
/*
(defun skk-insert-str (str)
  "STR }BKvł `self-insert-after-hook' R[B
`overwrite-mode' ł΁AK؂ɏ㏑sB"
  (insert-and-inherit str)
  (if (eq skk-henkan-mode 'on)
      ;;
      (when (and skk-auto-start-henkan
				 (not skk-okurigana))
		(skk-auto-start-henkan str))
    ;;
    (when (and (boundp 'self-insert-after-hook)
		       self-insert-after-hook)
      (funcall self-insert-after-hook
		       (- (point) (length str))
		       (point)))
    (when overwrite-mode
      (skk-del-char-with-pad (skk-ovwrt-len (string-width str)))))
  ;; SKK 9.6 ł͂̃^C~O fill sĂASKK 10 ł͍sĂ
  ;; ȂB
  (when (and skk-j-mode
		     (not skk-henkan-mode))
    (skk-do-auto-fill)))
		*/
int
LM_bSkkInsertStr (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPCWSTR				pwString ;
	int					nStringLen ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bGetRegConstString (pDoc, LMREGARG_0, &pwString, &nStringLen)) {
		pwString	= NULL ;
		nStringLen	= 0 ;
	}
	ImeBuffer_bInsertAndInherit (pBuffer, pwString, nStringLen) ;
	if (pBuffer->m_bSkkHenkanMode == LON) {
		if (ImeConfig_bSkkAutoStartHenkanp () && ! pBuffer->m_bSkkOkurigana) {
			ImeDoc_vJump (pDoc, LM_bSkkAutoStartHenkan) ;
			return	LMR_CONTINUE ;
		}
	} else {
		/* self-insert-after-hook ̎͂ȂB*/
		/* overwrite-mode ̎ȂB*/
	}
	/* skk-do-auto-fill  editor ̋@\łƎv̂Ŏ͂ȂB*/
	return	LMR_RETURN ;
}

/*================================================================ skk-auto-start-henkan
 */
static	int		LM_bSkkAutoStartHenkan_1	(struct CImeDoc*) ;
static	int		LM_bSkkAutoStartHenkan_Exit	(struct CImeDoc*) ;

int
LM_bSkkAutoStartHenkan (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPCWSTR				pwString ;
	int					nStringLen ;
	struct TMarker*		pmkMark ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bPushReg (pDoc, LMREG_0) || ! ImeDoc_bPushReg (pDoc, LMREG_1))
		return	LMR_ERROR ;

	/*	skk-save-point Ă̂YĂI
	 */
	pmkMark	= ImeBuffer_pMakeMarker (pBuffer, TRUE) ;
	if (pmkMark == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	TMarker_bSetPosition (pmkMark, pBuffer->m_pmkPoint) ;
	ImeDoc_vSetRegPointer (pDoc, LMREG_0, pmkMark) ;

	if (! ImeDoc_bGetRegConstString (pDoc, LMREGARG_0, &pwString, &nStringLen)) {
		pwString	= NULL ;
		nStringLen	= 0 ;
	}
	if (ImeConfig_bSkkAutoStartHenkanKeywordp (pwString, nStringLen)) {
		if (! ImeBuffer_bBackwardChar (pBuffer, 1)) {
			ImeDoc_vSetSignalError (pDoc) ;
			ImeDoc_vJump (pDoc, LM_bSkkAutoStartHenkan_Exit) ;
			return	LMR_CONTINUE ;
		}
		if (pBuffer->m_pmkSkkHenkanStartPoint != NULL && TMarker_bIsValidp (pBuffer->m_pmkSkkHenkanStartPoint) &&
			TMarker_iGetPosition (pBuffer->m_pmkPoint) > TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint)) {
			LPWSTR		pwSkkPrefix ;

			/* let skk-prefix ""  emulationBskk-prefix ꎞIɑޔB*/
			if (pBuffer->m_nSkkPrefixLen > 0) {
				pwSkkPrefix	= ImeDoc_pAlloca (pDoc, pBuffer->m_nSkkPrefixLen * sizeof (WCHAR), __alignof (WCHAR)) ;
				if (pwSkkPrefix == NULL)
					return	LMR_ERROR ;
				memcpy (pwSkkPrefix, pBuffer->m_bufSkkPrefix, sizeof (WCHAR) * pBuffer->m_nSkkPrefixLen) ;
			} else {
				pwSkkPrefix	= NULL ;
			}
			if (! ImeDoc_bPushReg (pDoc, LMREG_1))
				return	LMR_ERROR ;
			ImeDoc_vSetRegConstString (pDoc, LMREG_1, pwSkkPrefix, pBuffer->m_nSkkPrefixLen) ;

			pBuffer->m_nSkkPrefixLen	= 0 ;
			ImeDoc_vSetRegNil (pDoc, LMREGARG_0) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkStartHenkan, LM_bSkkAutoStartHenkan_1))
				return	LMR_ERROR ;
			return	LMR_CONTINUE ;
		}
	}
	ImeDoc_vJump (pDoc, LM_bSkkAutoStartHenkan_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkAutoStartHenkan_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPCWSTR				pwSkkPrefix ;
	int					nSkkPrefix ;

	if (ImeDoc_bSignalp (pDoc)) 
		goto	exit_func ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		goto	exit_func ;
	}
	/* skk-prefix ̕AB*/
	if (! ImeDoc_bGetRegConstString (pDoc, LMREG_1, &pwSkkPrefix, &nSkkPrefix)) {
		pwSkkPrefix	= NULL ;
		nSkkPrefix	= 0 ;
	}
	if (nSkkPrefix > 0) 
		memcpy (pBuffer->m_bufSkkPrefix, pwSkkPrefix, nSkkPrefix * sizeof (WCHAR)) ;
	pBuffer->m_nSkkPrefixLen	= nSkkPrefix ;

exit_func:
	ImeDoc_vJump (pDoc, LM_bSkkAutoStartHenkan_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkAutoStartHenkan_Exit (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;

	/*	J[\}[J̈ʒuB
	 */
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) {
		struct TMarker*	pmkMark ;

		/*	ۑĂ}[JɃ|Cg킹B
		 */
		if (ImeDoc_bGetRegPointer (pDoc, LMREG_0, (void**)&pmkMark) && pmkMark != NULL && TMarker_bIsValidp (pmkMark)) {
			TMarker_bSetPosition (pBuffer->m_pmkPoint, pmkMark) ;
			ImeBuffer_bDeleteMarker (pBuffer, pmkMark) ;
		}
	}
	ImeDoc_vPopReg (pDoc, LMREG_1) ;
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-change-marker
 */
static	int		LM_bSkkChangeMarker_1 (struct CImeDoc*) ;

int
LM_bSkkChangeMarker (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int					nPosition ;

	if (ImeDoc_bSignalp (pDoc)) {
		return	LMR_RETURN ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (pBuffer->m_pmkSkkHenkanStartPoint == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	nPosition	= TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) - 1 ;
	if (nPosition < 0) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (pBuffer->m_bufComp [nPosition] != L'') {
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkChangeMarker_1))
			return	LMR_ERROR ;
		return	LMR_CONTINUE ;
	}
	TMarker_bSetPosition (pBuffer->m_pmkPoint, pBuffer->m_pmkSkkHenkanStartPoint) ;
	TMarker_bBackward (pBuffer->m_pmkPoint, 1) ;
	ImeBuffer_bInsertAndInherit (pBuffer, L"", 1) ;
	ImeBuffer_bDeleteRegion (pBuffer, TMarker_iGetPosition (pBuffer->m_pmkPoint), TMarker_iGetPosition (pBuffer->m_pmkPoint) + 1) ;
	pBuffer->m_bSkkHenkanMode	= LACTIVE ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

int
LM_bSkkChangeMarker_1 (
	struct CImeDoc*		pDoc)
{
	/* (skk-error "It seems that you have delete ") */
	ImeDoc_vSetSignalError (pDoc) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-purge-from-jisyo
 */
/*
;;; jisyo related functions
(defun skk-purge-from-jisyo (&optional arg)
  "[hŌ݂̌obt@B"
  (interactive "*P")
  (skk-with-point-move
   (cond
    ((not (eq skk-henkan-mode 'active))
     (skk-emulate-original-map arg))
    ((and (eq skk-henkan-mode 'active)
		  (not (string= skk-henkan-key ""))
		  (yes-or-no-p
		   (format
		    (if skk-japanese-message-and-error
				"%s /%s/%s폜܂BǂłH"
		      "Really purge \"%s /%s/%s\"?")
		    skk-henkan-key
		    (skk-get-current-candidate)
		    (cond
		     ((not (and skk-henkan-okurigana
						(or skk-henkan-okuri-strictly
						    skk-henkan-strict-okuri-precedence)))
		      " ")
		     (skk-japanese-message-and-error
		      (format " (艼: %s) " skk-henkan-okurigana))
		     (t
		      (format " (okurigana: %s) " skk-henkan-okurigana))))))
     ;; skk-henkan-start-point  point ܂ō폜Ă܂ĂAϊ
     ;;  (J[\𓮂ƂȂ) skk-purge-from-jisyo ĂׂΖȂ
     ;; AJ[\ႤꏊֈړĂꍇ́A폜ׂłȂ̂
     ;; ō폜Ă܂\BŁA艼΂̒
     ;; ߂ end ߁A̕ϊɊ֘A𐳊mɐ؂悤
     ;; B
     (let ((end (if skk-henkan-okurigana
				    (+ (length skk-henkan-okurigana)
				       skk-henkan-end-point)
				  skk-henkan-end-point))
		   (word (skk-get-current-candidate)))
       (skk-update-jisyo word 'purge)
       ;; Emacs 19.28  Overlay ĂȂƁA insert 
       ;; skk-henkan-key ɉ̂ Overlay Ă܂B
       (when skk-use-face
		 (skk-henkan-face-off))
       (delete-region skk-henkan-start-point end)
       (skk-change-marker-to-white)
       (skk-kakutei)))))
  nil) */

static	int		LM_bSkkPurgeFromJisyo_1		(struct CImeDoc*) ;
static	int		LM_bSkkPurgeFromJisyo_2		(struct CImeDoc*) ;
static	int		LM_bSkkPurgeFromJisyo_3		(struct CImeDoc*) ;
static	int		LM_bSkkPurgeFromJisyo_4		(struct CImeDoc*) ;
static	int		LM_bSkkPurgeFromJisyo_5		(struct CImeDoc*) ;
static	int		LM_bSkkPurgeFromJisyo_Exit	(struct CImeDoc*) ;

int
LM_bSkkPurgeFromJisyo (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (! ImeDoc_bPushReg (pDoc, LMREG_0) ||
		! ImeDoc_bPushReg (pDoc, LMREG_1) ||
		! ImeDoc_bPushReg (pDoc, LMREG_2) ||
		! ImeDoc_bPushReg (pDoc, LMREG_3))
		return	LMR_ERROR ;

	if (pBuffer->m_bSkkHenkanMode != LACTIVE) {
		if (! ImeDoc_bCall (pDoc, LM_bSkkEmulateOriginalMap, LM_bSkkPurgeFromJisyo_Exit))
			return	LMR_ERROR ;
	} else {
		if (pBuffer->m_bSkkHenkanMode == LACTIVE && pBuffer->m_nSkkHenkanKeyLen > 0) {
			LPWSTR	pwText ;
			WCHAR	bufOkurigana [128] ;
			LPCWSTR	pwCandidate ;
			int		n ;

			pwCandidate	= TSearchSession_pGetCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;
			if (pwCandidate == NULL) {
				ImeDoc_vSetSignalError (pDoc) ;
				return	LMR_RETURN ;
			}

#define	TEMP_TEXTBUFSIZE	384
			pwText	= ImeDoc_pAlloca (pDoc, sizeof (WCHAR) * TEMP_TEXTBUFSIZE, __alignof (WCHAR)) ;
			if (pwText == NULL)
				return	LMR_ERROR ;
			ImeDoc_vSetRegString (pDoc, LMREG_0, pwText, TEMP_TEXTBUFSIZE) ;

			if (! (pBuffer->m_bSkkOkurigana && (ImeConfig_bSkkHenkanOkuriStrictlyp () || ImeConfig_bSkkHenkanStrictOkuriPrecedencep ()))) {
				bufOkurigana [0]	= L' ' ;
				bufOkurigana [1]	= L'\0' ;
			} else {
				LPWSTR	pwDest, pwDestEnd ;

				pwDest		= bufOkurigana ;
				pwDestEnd	= bufOkurigana + ARRAYSIZE (bufOkurigana) - 1 ;
				vCopyString  (&pwDest, pwDestEnd, L" (艼: ") ;
				vCopyStringN (&pwDest, pwDestEnd, pBuffer->m_bufSkkHenkanOkurigana, pBuffer->m_nSkkHenkanOkuriganaLen) ;
				vCopyString  (&pwDest, pwDestEnd, L") ") ;
				*pwDest		= L'\0' ;
			}

			n	= MIN (TEMP_TEXTBUFSIZE - 1, pBuffer->m_nSkkHenkanKeyLen) ;
			memcpy (pwText, pBuffer->m_bufSkkHenkanKey, n * sizeof (WCHAR)) ;
			wnsprintfW (pwText + n, TEMP_TEXTBUFSIZE - n - 1, L" /%s/%s폜܂B낵łH", pwCandidate, bufOkurigana) ;
			pwText [TEMP_TEXTBUFSIZE - 1]	= L'\0' ;
#undef	TEMP_TEXTBUFSIZE

			ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwText, lstrlenW (pwText)) ;
			/*	yes-or-no ĂяoB
			 */
			if (! ImeDoc_bCall (pDoc, LM_bYesOrNop, LM_bSkkPurgeFromJisyo_1))
				return	LMR_ERROR ;
		} else {
			ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_Exit) ;
		}
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkPurgeFromJisyo_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	BOOL				bRetval ;

	if (ImeDoc_bSignalp (pDoc)) {
		if (ImeDoc_nGetSignal (pDoc) == LMSIGNAL_QUIT) 
			ImeDoc_vClearSignals (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegBool (pDoc, LMREGARG_RETVAL, &bRetval)) {
		bRetval	= FALSE ;
	}
	if (bRetval) {
		LPCWSTR		pwWord ;
		int			nWordLen ;

		/*	skk-henkan-okurigana ̏ skk-update-jisyo ̌cĂ̂Ɖ肵
		 *	ł͐ȂB
		 */
		pwWord		= TSearchSession_pGetCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;
		nWordLen	= lstrlenW (pwWord) ;

		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwWord, nWordLen) ;
		ImeDoc_vSetRegBool        (pDoc, LMREGARG_1, TRUE) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkUpdateJisyo, LM_bSkkPurgeFromJisyo_2))
			return	LMR_ERROR ;
		return	LMR_CONTINUE ;
	} else {
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_2) ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkPurgeFromJisyo_2 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! TSearchSession_bNumericp (pBuffer->m_pSkkCurrentSearchProgSession) ||
		TSearchSession_pGetReferCandidate (pBuffer->m_pSkkCurrentSearchProgSession) == NULL) {
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_5) ;
		return	LMR_CONTINUE ;
	}
	/*	 #4 ϊꍇɂ #4 Ή̍폜q˂Kv邪c
	 */
	ImeDoc_vSetRegInteger (pDoc, LMREG_1, 0) ;
	ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_3) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkPurgeFromJisyo_3 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	struct CTS4Candidate*	pS4Node ;
	int						nTextSize, num, i ;
	LPWSTR					pwText, pwDest, pwDestEnd ;
	LPCWSTR					pwNumKeyword, pwNumResult ;
	int						nNumKeywordLen, nNumResultLen ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}

	if (! ImeDoc_bGetRegString (pDoc, LMREG_0, &pwText, &nTextSize) || 
		! ImeDoc_bGetRegInteger (pDoc, LMREG_1, &num)) {
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}

	/* #4 ̃`FbNsB*/
	pS4Node	= TSearchSession_pGetNumericLink (pBuffer->m_pSkkCurrentSearchProgSession) ;
	
	for (i = 0 ; i < num && pS4Node != NULL ; i ++) {
		pS4Node	= TSearchSession_pGetNextNumericLink	(pBuffer->m_pSkkCurrentSearchProgSession, pS4Node) ;
	}
	while (pS4Node != NULL) {
		/*	"[#4폜] %s /%s/ 폜܂H"
		 *	cdꍇɂ͂ƂB
		 */
		pwNumKeyword	= TSearchSession_pGetNumericKeyword (pBuffer->m_pSkkCurrentSearchProgSession, pS4Node, &nNumKeywordLen) ;
		pwNumResult		= TSearchSession_pGetNumericResult  (pBuffer->m_pSkkCurrentSearchProgSession, pS4Node, &nNumResultLen) ;
		if (pwNumKeyword != NULL && nNumKeywordLen > 0 && pwNumResult != NULL && nNumResultLen > 0)
			break ;
		pS4Node	= TSearchSession_pGetNextNumericLink	(pBuffer->m_pSkkCurrentSearchProgSession, pS4Node) ;
		num	++ ;
	}
	if (pS4Node == NULL) {
		/* IB*/
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_5) ;
		return	LMR_CONTINUE ;
	}

	ImeDoc_vSetRegInteger     (pDoc, LMREG_1, num) ;
	ImeDoc_vSetRegConstString (pDoc, LMREG_2, pwNumKeyword, nNumKeywordLen) ;
	ImeDoc_vSetRegConstString (pDoc, LMREG_3, pwNumResult,  nNumResultLen) ;

	pwDest			= pwText ;
	pwDestEnd		= pwText + nTextSize ;
	vCopyString  (&pwDest, pwDestEnd, L"[#4폜] ") ;
	vCopyStringN (&pwDest, pwDestEnd, pwNumKeyword, nNumKeywordLen) ;
	vCopyString  (&pwDest, pwDestEnd, L" /") ;
	vCopyStringN (&pwDest, pwDestEnd, pwNumResult, nNumResultLen) ;
	vCopyString  (&pwDest, pwDestEnd, L"/ 폜܂B낵łH") ;
	ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pwText, pwDest - pwText) ;
	if (! ImeDoc_bCall (pDoc, LM_bYesOrNop, LM_bSkkPurgeFromJisyo_4))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkPurgeFromJisyo_4 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	LPWSTR					pwText ;
	LPCWSTR					pwNumKeyword, pwNumResult ;
	int						nNumKeywordLen, nNumResultLen, num, nTextSize ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}

	/*	ϊL[ꎞIɕۑB*/
	if (! ImeDoc_bGetRegString (pDoc, LMREG_0, &pwText, &nTextSize) || 
		! ImeDoc_bGetRegInteger (pDoc, LMREG_1, &num) ||
		! ImeDoc_bGetRegConstString (pDoc, LMREG_2, &pwNumKeyword, &nNumKeywordLen) ||
		! ImeDoc_bGetRegConstString (pDoc, LMREG_3, &pwNumResult,  &nNumResultLen)) {
		/* error */
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}

	/*  skk-update-jisyo ĂяoB
	 *	͕s\B
	 *	ŔcHenkanKey  numeric-convertion Ă܂ # ɗƂĂ܂B
	 */
	TPurgeSession_bUpdate (pwNumKeyword, nNumKeywordLen, pwNumResult, nNumResultLen, NULL, 0, FALSE) ;

	/*   S4Node wB*/
	num	++ ;
	ImeDoc_vSetRegInteger (pDoc, LMREG_1, num) ;

	ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_3) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkPurgeFromJisyo_5 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int					nStart, nEnd ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkPurgeFromJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}

	nStart	= TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint) ;
	nEnd	= TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanEndPoint) ; 
	if (pBuffer->m_nSkkHenkanOkuriganaLen > 0) {
		nEnd	+= pBuffer->m_nSkkHenkanOkuriganaLen ;
	}
	ImeBuffer_bDeleteRegion (pBuffer, nStart, nEnd) ;
	imeBuffer_bSkkChangeMarkerToWhite (pBuffer) ;

	ImeDoc_vSetRegNil (pDoc, LMREGARG_0) ;
	if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkPurgeFromJisyo_Exit))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkPurgeFromJisyo_Exit (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;

	ImeDoc_vPopReg (pDoc, LMREG_3) ;
	ImeDoc_vPopReg (pDoc, LMREG_2) ;
	ImeDoc_vPopReg (pDoc, LMREG_1) ;
	ImeDoc_vPopReg (pDoc, LMREG_0) ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	/* skk-point-move ̌ʁB*/
	imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkPreviousPoint, MARKER_SKK_PREVIOUS_POINT, pBuffer->m_pmkPoint) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-update-jisyo
 */
static	int		LM_bSkkUpdateJisyo_1	(struct CImeDoc*) ;
static	int		LM_bSkkUpdateJisyo_2	(struct CImeDoc*) ;
static	int		LM_bSkkUpdateJisyo_Exit	(struct CImeDoc*) ;

/*	LMARG_0:		pWord, nWordLen (STRING)
 *	LMARG_1:		bPurge (BOOL)
 */
int
LM_bSkkUpdateJisyo (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPWSTR	pMidashi ;
	int		nMidashiLen ;
	LPCWSTR	pWord ;
	int		nWordLen ;
	BOOL	bPurge ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	if (! ImeDoc_bGetRegConstString (pDoc, LMREGARG_0, &pWord, &nWordLen)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	/* bPurge ͏ȗ\B*/
	if (! ImeDoc_bGetRegBool (pDoc, LMREGARG_1, &bPurge)) {
		bPurge	= FALSE ;
	}

	if (! ImeDoc_bPushReg (pDoc, LMREG_0) || ! ImeDoc_bPushReg (pDoc, LMREG_1))
		return	LMR_ERROR ;

	/*	ꂾƖ{ɐlϊsĂ邩炩Ȃ悤ɎvȁcB
	 */
	if (ImeConfig_bSkkNumericConversionp () && (bStringNumericWordp (pWord, nWordLen) || bStringLispFormatp (pWord, nWordLen))) {
		pMidashi	= ImeDoc_pAlloca (pDoc, MAXCOMPLEN * sizeof (WCHAR), __alignof (WCHAR)) ;
		if (pMidashi == NULL) 
			return	LMR_ERROR ;
		nMidashiLen	= iSkkNumComputeHenkanKey (pBuffer->m_bufSkkHenkanKey, pBuffer->m_nSkkHenkanKeyLen, pMidashi, MAXCOMPLEN, NULL, 0, ImeConfig_bSkkNumConvertFloatp ()) ;
	} else {
		pMidashi	= pBuffer->m_bufSkkHenkanKey ;
		nMidashiLen	= pBuffer->m_nSkkHenkanKeyLen ;
	}
	if (pBuffer->m_iSkkOkuriIndexMin > -1) {
		LPWSTR	pNewWord ;
		pNewWord	= ImeDoc_pAlloca (pDoc, MAXCOMPLEN * sizeof (WCHAR), __alignof (WCHAR)) ;
		if (pNewWord == NULL)
			LMR_ERROR ;
		ImeDoc_vSetRegString (pDoc, LMREG_0, pNewWord, MAXCOMPLEN) ;
		ImeDoc_vSetRegBool   (pDoc, LMREG_1, bPurge) ;

		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pWord, nWordLen) ;
		ImeDoc_vSetRegString (pDoc, LMREGARG_1, pNewWord, MAXCOMPLEN) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkRemoveCommon, LM_bSkkUpdateJisyo_1))
			return	LMR_ERROR ;
	} else {
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pWord,    nWordLen) ;
		ImeDoc_vSetRegBool        (pDoc, LMREGARG_1, bPurge) ;
		ImeDoc_vSetRegConstString (pDoc, LMREGARG_2, pMidashi, nMidashiLen) ;
		ImeDoc_vJump (pDoc, LM_bSkkUpdateJisyo_2) ;
	}
	return	LMR_CONTINUE ;
}

int
LM_bSkkUpdateJisyo_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPWSTR		pWord ;
	int			nMaxWordSize, nWordLen ;
	LPCWSTR		pMidashi ;
	int			nMidashiLen ;
	BOOL		bPurge ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkUpdateJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkUpdateJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegInteger (pDoc, LMREGARG_RETVAL, &nWordLen)) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkUpdateJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegString (pDoc, LMREG_0, &pWord, &nMaxWordSize)) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkUpdateJisyo_Exit) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegBool (pDoc, LMREG_1, &bPurge))
		bPurge	= FALSE ;

	pMidashi	= pBuffer->m_bufSkkHenkanKey ;
	nMidashiLen	= pBuffer->m_nSkkHenkanKeyLen ;
	ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, pWord, nWordLen) ;
	ImeDoc_vSetRegBool        (pDoc, LMREGARG_1, bPurge) ;
	ImeDoc_vSetRegConstString (pDoc, LMREGARG_2, pMidashi, nMidashiLen) ;
	ImeDoc_vJump (pDoc, LM_bSkkUpdateJisyo_2) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkUpdateJisyo_2 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	LPWSTR	pWord ;
	LPCWSTR	pMidashi, pOkurigana ;
	int		nWordLen, nMidashiLen, nOkuriganaLen ;
	BOOL	bPurge, bOkuri ;
	BOOL	(*pUpdateProc)(LPCWSTR, int, LPCWSTR, int, LPCWSTR, int, BOOL) = NULL ;

	if (ImeDoc_bSignalp (pDoc))
		goto	exit_func ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		goto	exit_func ;
	}
	if (! ImeDoc_bGetRegConstString (pDoc, LMREGARG_0, &pWord, &nWordLen)) {
		ImeDoc_vSetSignalError (pDoc) ;
		goto	exit_func ;
	}
	if (! ImeDoc_bGetRegBool (pDoc, LMREGARG_1, &bPurge)) 
		bPurge	= FALSE ;
	if (! ImeDoc_bGetRegConstString (pDoc, LMREGARG_2, &pMidashi, &nMidashiLen)) {
		ImeDoc_vSetSignalError (pDoc) ;
		goto	exit_func ;
	}
	if (pBuffer->m_nSkkHenkanOkuriganaLen > 0) {
		pOkurigana		= pBuffer->m_bufSkkHenkanOkurigana ;
		nOkuriganaLen	= pBuffer->m_nSkkHenkanOkuriganaLen ;
	} else if (pBuffer->m_nSkkOkuriCharLen > 0) {
		pOkurigana		= pBuffer->m_bufSkkOkuriChar ;
		nOkuriganaLen	= pBuffer->m_nSkkOkuriCharLen ;
	} else {
		pOkurigana		= NULL ;
		nOkuriganaLen	= 0 ;
	}
	/* okurigana, midashi, word, purge/record ̃yAœo^B*/
	if (! bPurge) {
		pUpdateProc	= TRecordSession_bUpdate ;
	} else {
		pUpdateProc	= TPurgeSession_bUpdate ;
	}
	if (pUpdateProc == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		goto	exit_func ;
	}
	bOkuri	= FALSE ;
	if (nOkuriganaLen > 0) {
		/*	skk-process-okuri-early ̎ɂ́A艼ł̂ȂB
		 */
		if (ImeConfig_bSkkProcessOkuriEarlyp ()) {
			pOkurigana		= L"" ; 
			nOkuriganaLen	= 0 ;
		}
		bOkuri	= TRUE ;
	} else {
		/*	skk-process-okuri-early ̎ɂ́A艼ł̂ȂB
		 */
		if (ImeConfig_bSkkProcessOkuriEarlyp () && 
			pBuffer->m_bSkkJMode && 
			pBuffer->m_nSkkHenkanKeyLen > 1 && 
			(ImeConfig_bSkkSetHenkanPointKeyp (towupper (pBuffer->m_bufSkkHenkanKey [pBuffer->m_nSkkHenkanKeyLen - 1])) ||
			ImeConfig_bSkkSetHenkanPointKeyp (towlower (pBuffer->m_bufSkkHenkanKey [pBuffer->m_nSkkHenkanKeyLen - 1])))) {
			pOkurigana		= L"" ; 
			nOkuriganaLen	= 0 ;
			bOkuri			= TRUE ;
		}
	}
	if ((*pUpdateProc) (pMidashi, nMidashiLen, pWord, nWordLen, pOkurigana, nOkuriganaLen, bOkuri)) {
		/* ̌o Word ŒO kakutei-history ƂċLB*/
		if (! bOkuri) 
			imeBuffer_bSkkUpdateKakuteiHistory (pBuffer, pMidashi, nMidashiLen, pWord, nWordLen) ;
	}
exit_func:
	ImeDoc_vJump (pDoc, LM_bSkkUpdateJisyo_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkUpdateJisyo_Exit (
	struct CImeDoc*		pDoc)
{
	ImeDoc_vPopReg (pDoc, LMREG_1) ;
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-remove-common
 */

static	int	LM_bSkkRemoveCommon_1 (struct CImeDoc*) ;

/*	Ԃl: LMREGARG_RETVAL  WordResult  length
 */
int
LM_bSkkRemoveCommon (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*		pBuffer ;
	LPCWSTR		pWord ;
	int			nWordLen ;
	LPWSTR		pwWordResult ;
	int			nWordResultSize ;
	LPCWSTR		pMidashi ;
	int			nMidashiLen, nNewWordLen ;

	if (ImeDoc_bSignalp (pDoc))
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bGetRegConstString (pDoc, LMREGARG_0, &pWord, &nWordLen) ||
		! ImeDoc_bGetRegString      (pDoc, LMREGARG_1, &pwWordResult, &nWordResultSize)) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if ((pBuffer->m_pSkkCurrentSearchProgSession != NULL && TSearchSession_bNumericp (pBuffer->m_pSkkCurrentSearchProgSession)) ||
		pBuffer->m_bSkkAbbrevMode ||
		! (pBuffer->m_bSkkHenkanInMinibuffFlag || (pBuffer->m_iSkkOkuriIndexMin <= pBuffer->m_iSkkHenkanCount && pBuffer->m_iSkkHenkanCount <= pBuffer->m_iSkkOkuriIndexMax))) {
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_RETVAL, 0) ;
		return	LMR_RETURN ;
	}

	pMidashi	= pBuffer->m_bufSkkHenkanKey ;
	nMidashiLen	= pBuffer->m_nSkkHenkanKeyLen ;
	if (nMidashiLen >= 2 && nWordLen >= 2) {
		int		nMidashiTail, nWordTail ;

		if (bSkkAsciiCharp (pMidashi [nMidashiLen - 1]) && pMidashi [nMidashiLen - 1] == pWord [nWordLen - 1]) {
			nMidashiLen	-- ;
			nWordLen	-- ;
		}
		nMidashiTail	= pMidashi [nMidashiLen - 1] ;
		nWordTail		= pWord [nWordLen - 1] ;
		if (nMidashiTail == nWordTail && ((L'' <= nMidashiTail && nMidashiTail <= L'') || nMidashiTail == L'A' || nMidashiTail == L'B' || nMidashiTail == L'C' || nMidashiTail == L'D')) {
			int		nPos, nPos2, nChar ;

			nPos	= nWordLen - 1 ;
			while (nPos > 0) {
				nChar	= pWord [nPos - 1] ;
				if (bCJKKanjip (nChar)) { /* unicode Ŋǂācǂ邩B*/
					break ;
				}
				nPos	-- ;
			}
			nPos2	= nMidashiLen - (nWordLen - nPos) ;

			if (nMidashiLen == nWordLen && memcmp (pMidashi + nPos2, pWord + nPos, sizeof (WCHAR) * nMidashiLen) == 0) {
				WCHAR	bufHenkanOkurigana	[2] ;
				WCHAR	bufNewSkkHenkanKey	[MAXCOMPLEN] ;
				WCHAR	bufNewWord			[MAXCOMPLEN] ;
				WCHAR	bufNewSkkOkuriChar	[MAXCOMPLEN] ;
				int		nOkuriFirst, nHenkanOkuriganaLen, nNewSkkOkuriCharLen, nNewSkkHenkanKeyLen ;

				nOkuriFirst				= pWord [nPos] ;
				bufHenkanOkurigana [0]	= nOkuriFirst ;
				nHenkanOkuriganaLen		= 1 ;
				if (nOkuriFirst == L'' && (nPos + 2) <= nWordLen) {
					bufHenkanOkurigana [1]	= pWord [nPos + 1] ;
					nHenkanOkuriganaLen	++ ;
				}

				nNewWordLen			= nPos ;
				if (nNewWordLen >= ARRAYSIZE (bufNewWord)) {
					ImeDoc_vSetSignalError (pDoc) ;
					return	LMR_RETURN ;
				}
				memcpy (bufNewWord, pWord, sizeof (WCHAR) * nNewWordLen) ;
				bufNewWord [nNewWordLen]	= L'\0' ;
				nNewSkkOkuriCharLen	= imeBuffer_iSkkOkuriganaPrefix (pBuffer, bufHenkanOkurigana, nHenkanOkuriganaLen, bufNewSkkOkuriChar, ARRAYSIZE (bufNewSkkOkuriChar)) ;

				nNewSkkHenkanKeyLen	= nPos2 + nNewSkkOkuriCharLen ;
				if ((nPos2 + nNewSkkOkuriCharLen) >= ARRAYSIZE (bufNewSkkHenkanKey)) {
					ImeDoc_vSetSignalError (pDoc) ;
					return	LMR_RETURN ;
				}
				memcpy (bufNewSkkHenkanKey,         pMidashi + 0,       sizeof (WCHAR) * nPos2) ;
				memcpy (bufNewSkkHenkanKey + nPos2, bufNewSkkOkuriChar, sizeof (WCHAR) * nNewSkkOkuriCharLen) ;
				bufNewSkkHenkanKey [nNewSkkHenkanKeyLen]	= L'\0' ;

				if (nNewWordLen > nWordResultSize) {
					ImeDoc_vSetSignalError (pDoc) ;
					return	LMR_RETURN ;
				}

				if (!pBuffer->m_bSkkHenkanInMinibuffFlag) {
					memcpy (pBuffer->m_bufSkkHenkanKey, bufNewSkkHenkanKey, sizeof (WCHAR) * nNewSkkHenkanKeyLen) ;
					pBuffer->m_nSkkHenkanKeyLen	= nNewSkkHenkanKeyLen ;
					memcpy (pBuffer->m_bufSkkOkuriChar, bufNewSkkOkuriChar, sizeof (WCHAR) * nNewSkkOkuriCharLen) ;
					pBuffer->m_nSkkOkuriCharLen	= nNewSkkOkuriCharLen ;

					memcpy (pwWordResult, bufNewWord, sizeof (WCHAR) * nNewWordLen) ;
				} else {
					WCHAR	bufText [256] ;
					int		nText ;
					LPWSTR	pwText, pwNewWord, pwNewSkkHenkanKey, pwNewSkkOkuriChar ;

					/* Shall I register this as okuri-ari word: %s /%s/ ? */
					nText	= wnsprintfW (bufText, ARRAYSIZE (bufText), L"Shall I register this as okuri-ari word: %s /%s/ ? ", bufNewSkkHenkanKey, bufNewWord) ;
					if (! ImeDoc_bPushReg (pDoc, LMREG_0) || ! ImeDoc_bPushReg (pDoc, LMREG_1) ||
						! ImeDoc_bPushReg (pDoc, LMREG_2) || ! ImeDoc_bPushReg (pDoc, LMREG_3))
						return	LMR_ERROR ;

#define	SAFE_ALLOCA_STRING(pDest,pSrc,nSize,nAlign)	\
	if ((nSize) > 0) { \
		(pDest) = ImeDoc_pAlloca (pDoc, (nSize), (nAlign)) ; \
		if ((pDest) == NULL) \
			return LMR_ERROR ; \
		memcpy ((pDest), (pSrc), (nSize)) ; \
	} else { \
		(pDest) = NULL ; \
	}
					SAFE_ALLOCA_STRING (pwNewWord,			bufNewWord,			nNewWordLen				* sizeof (WCHAR), __alignof (WCHAR)) ;
					SAFE_ALLOCA_STRING (pwNewSkkHenkanKey,	bufNewSkkHenkanKey,	nNewSkkHenkanKeyLen		* sizeof (WCHAR), __alignof (WCHAR)) ;
					SAFE_ALLOCA_STRING (pwNewSkkOkuriChar,	bufNewSkkOkuriChar,	nNewSkkOkuriCharLen		* sizeof (WCHAR), __alignof (WCHAR)) ;
					SAFE_ALLOCA_STRING (pwText,				bufText,			nText					* sizeof (WCHAR), __alignof (WCHAR)) ;
#undef	SAFE_ALLOCA_STRING
					ImeDoc_vSetRegConstString (pDoc, LMREG_0, pwNewWord,			nNewWordLen) ;
					ImeDoc_vSetRegConstString (pDoc, LMREG_1, pwNewSkkHenkanKey,	nNewSkkHenkanKeyLen) ;
					ImeDoc_vSetRegConstString (pDoc, LMREG_2, pwNewSkkOkuriChar,	nNewSkkOkuriCharLen) ;
					ImeDoc_vSetRegString      (pDoc, LMREG_3, pwWordResult,			nWordResultSize) ;
					ImeDoc_vSetRegString (pDoc, LMREGARG_0, pwText, nText) ;
					if (! ImeDoc_bCall (pDoc, LM_bYesOrNop, LM_bSkkRemoveCommon_1))
						return	LMR_ERROR ;
					return	LMR_CONTINUE ;
				}
			}
		}
	}
	ImeDoc_vSetRegInteger (pDoc, LMREGARG_RETVAL, nNewWordLen) ;
	return	LMR_RETURN ;
}

/*
					// yes-or-no-p Ŏs skk-henkan-okurigana  nil  skk-okuri-char  nil B
					if (! ImeDoc_bYesOrNop (pBuffer->m_pDoc, bufText, nText, &bRetval)) {
						return	FALSE ;
					}
					if (bRetval) {
						memcpy (pBuffer->m_bufSkkHenkanKey, bufNewSkkHenkanKey, sizeof (WCHAR) * nNewSkkHenkanKeyLen) ;
						pBuffer->m_nSkkHenkanKeyLen	= nNewSkkHenkanKeyLen ;
						memcpy (pBuffer->m_bufSkkOkuriChar, bufNewSkkOkuriChar, sizeof (WCHAR) * nNewSkkOkuriCharLen) ;
						pBuffer->m_nSkkOkuriCharLen	= nNewSkkOkuriCharLen ;

						memcpy (pwWordResult, bufNewWord, sizeof (WCHAR) * nNewWordLen) ;
					} else {
						pBuffer->m_nSkkHenkanOkuriganaLen	= 0 ;
						pBuffer->m_nSkkOkuriCharLen			= 0 ;
					}
					ImeDoc_vClearMessage (pBuffer->m_pDoc) ;
 */
int
LM_bSkkRemoveCommon_1 (
	struct CImeDoc*		pDoc)
{
	struct CImeBuffer*	pBuffer ;
	LPCWSTR	pNewSkkHenkanKey, pNewSkkOkuriChar, pNewWord ;
	int		nNewSkkHenkanKeyLen, nNewSkkOkuriCharLen, nNewWordLen ;
	LPWSTR	pwWordResult ;
	int		nWordResultSize ;
	BOOL	bYesOrNo ;

	if (ImeDoc_bSignalp (pDoc))
		goto	exit_func ;
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		goto	exit_func ;
	}
	if (! ImeDoc_bGetRegBool (pDoc, LMREGARG_RETVAL, &bYesOrNo)) 
		goto	exit_func ;
	if (! ImeDoc_bGetRegConstString (pDoc, LMREG_0, &pNewWord, &nNewWordLen)) {
		pNewWord			= NULL ;
		nNewWordLen			= 0 ;
	}
	if (! ImeDoc_bGetRegConstString (pDoc, LMREG_1, &pNewSkkHenkanKey, &nNewSkkHenkanKeyLen)) {
		pNewSkkHenkanKey	= NULL ;
		nNewSkkHenkanKeyLen	= 0 ;
	}
	if (! ImeDoc_bGetRegConstString (pDoc, LMREG_2, &pNewSkkOkuriChar, &nNewSkkOkuriCharLen)) {
		pNewSkkOkuriChar	= NULL ;
		nNewSkkOkuriCharLen	= 0 ;
	}
	if (! ImeDoc_bGetRegString (pDoc, LMREG_3, &pwWordResult, &nWordResultSize)) {
		pwWordResult	= NULL ;
		nWordResultSize	= 0 ;
	}
	if (bYesOrNo) {
		if (nNewSkkHenkanKeyLen > 0)
			memcpy (pBuffer->m_bufSkkHenkanKey, pNewSkkHenkanKey, sizeof (WCHAR) * nNewSkkHenkanKeyLen) ;
		pBuffer->m_nSkkHenkanKeyLen	= nNewSkkHenkanKeyLen ;
		if (nNewSkkOkuriCharLen > 0)
			memcpy (pBuffer->m_bufSkkOkuriChar, pNewSkkOkuriChar, sizeof (WCHAR) * nNewSkkOkuriCharLen) ;
		pBuffer->m_nSkkOkuriCharLen	= nNewSkkOkuriCharLen ;

		if (nNewWordLen > 0) {
			nNewWordLen	= MIN (nNewWordLen, nWordResultSize) ;
			if (nNewWordLen > 0 && pwWordResult != NULL)
				memcpy (pwWordResult, pNewWord, sizeof (WCHAR) * nNewWordLen) ;
		}
	} else {
		pBuffer->m_nSkkHenkanOkuriganaLen	= 0 ;
		pBuffer->m_nSkkOkuriCharLen			= 0 ;
		nNewWordLen							= 0 ;
	}
	ImeDoc_vClearMessage (pDoc) ;
	ImeDoc_vSetRegInteger (pDoc, LMREGARG_RETVAL, nNewWordLen) ;

exit_func:
	ImeDoc_vPopReg (pDoc, LMREG_3) ;
	ImeDoc_vPopReg (pDoc, LMREG_2) ;
	ImeDoc_vPopReg (pDoc, LMREG_1) ;
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	return	LMR_RETURN ;
}

/*================================================================ skk-henkan-show-candidates
 */
/*	̓͗ǂȂB̂Ȃꗗ\@\ Windows  interface 邩B
 *	܂A͌ł킹ƂƂŃX[B
 */
static	int		LM_bSkkHenkanShowCandidates_1		(struct CImeDoc*) ;
static	int		LM_bSkkHenkanShowCandidates_2		(struct CImeDoc*) ;
static	int		LM_bSkkHenkanShowCandidates_3		(struct CImeDoc*) ;
static	int		LM_bSkkHenkanShowCandidates_4		(struct CImeDoc*) ;
static	int		LM_bSkkHenkanShowCandidates_5		(struct CImeDoc*) ;
static	int		LM_bSkkHenkanShowCandidates_Exit_0	(struct CImeDoc*) ;
static	int		LM_bSkkHenkanShowCandidates_Exit	(struct CImeDoc*) ;

/*	Ԃl: LMREGARG_RETVAL ɑI̕
 *			LMREGARG_0	ɗ^ꂽobt@ɑI(̕)Rs[
 */
int
LM_bSkkHenkanShowCandidates (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;

	if (ImeDoc_bSignalp (pDoc)) 
		return	LMR_RETURN ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}
	if (! ImeBuffer_bDeleteRegion (pBuffer, TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint), TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanEndPoint))) {
		ImeDoc_vSetSignalError (pDoc) ;
		return	LMR_RETURN ;
	}

	//strCandidateKey	= ImeDoc_pGetSkkHenkanShowCandidatesKeys (pBuffer->m_pDoc, &nCandidateKey) ;
	//nMaxShow		= (nCandidateKey < 7)? nCandidateKey : 7 ;	/* candidate-key 7ŒɂׂH */
	//nLoop			= 0 ;

	pBuffer->m_bSkkHenkanShowCandidatesMode	= TRUE ;

	if (! ImeDoc_bPushReg (pDoc, LMREG_1) || ! ImeDoc_bPushReg (pDoc, LMREG_2) ||
		! ImeDoc_bPushReg (pDoc, LMREG_3) || ! ImeDoc_bPushReg (pDoc, LMREG_4))
		return	LMR_ERROR ;

	ImeDoc_vSetRegInteger	(pDoc, LMREG_1, 0) ;			/* nLoop */
	ImeDoc_vSetRegBool		(pDoc, LMREG_2, FALSE) ;		/* bUpdateShow */
	ImeDoc_vSetRegInteger	(pDoc, LMREG_4, 0) ;			/* nShow */
	ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_1) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkanShowCandidates_Exit (
	struct CImeDoc*			pDoc)
{
	ImeDoc_vPopReg (pDoc, LMREG_4) ;
	ImeDoc_vPopReg (pDoc, LMREG_3) ;
	ImeDoc_vPopReg (pDoc, LMREG_2) ;
	ImeDoc_vPopReg (pDoc, LMREG_1) ;
	return	LMR_RETURN ;
}

int
LM_bSkkHenkanShowCandidates_1 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	int		nCandidateKey, nLoop, nMaxShow ;
	WCHAR	buffer [512] ;
	LPWSTR	pDest, pDestEnd ;
	LPCWSTR	pCandidateKey ;
	BOOL	bSearchContinue ;
	int		nShow, nHenkanCount, iShowAnnotation ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0) ;
		return	LMR_CONTINUE ;
	}
	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_1, &nLoop))
		nLoop		= 0 ;

	pCandidateKey	= ImeConfig_pGetSkkHenkanShowCandidatesKeys (&nCandidateKey) ;
	nMaxShow		= (nCandidateKey < IMEDOC_NUMHENKANSHOWCANDIDATES)? nCandidateKey : IMEDOC_NUMHENKANSHOWCANDIDATES ;	/* candidate-key 7ŒɂׂH */

	/* P7\Ηǂ̂悤BGɏƂ͂ȂB*/
	nHenkanCount	= pBuffer->m_iSkkHenkanCount + nLoop * nMaxShow ;
	pDest			= buffer ;
	pDestEnd		= buffer + ARRAYSIZE (buffer) ;
	nShow			= 0 ;
	iShowAnnotation	= ImeConfig_iGetSkkShowAnnotationType () ;

	buffer [0]	= L'\0' ;
	if (pBuffer->m_pSkkCurrentSearchProgSession != NULL) {
#if defined (UNITTEST)
		LPCWSTR		pSrc ;
		int			nLeft ;
		WCHAR		buf [64] ;

		while (nShow < nMaxShow) {
			LPCWSTR		wstrNewWord ;
			if (pDest < pDestEnd) 
				*pDest ++	= *pCandidateKey ++ ;
			if (pDest < pDestEnd)
				*pDest ++	= L':' ;
			wstrNewWord		=TSearchSession_pGetCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;
			if (wstrNewWord != NULL) {
				int		nNewWordLen ;

				nNewWordLen	= lstrlenW (wstrNewWord) ;
				/* lisp function Ă邩ǂ`FbNB*/
				/* Annotation OKvB*/
				if (iShowAnnotation != DISABLE_ANNOTATION) {
					pSrc	= wstrNewWord + nNewWordLen - 1 ;
					while (pSrc >= wstrNewWord && *pSrc != L';')
						pSrc -- ;
					if (pSrc >= wstrNewWord && *pSrc == L';') 
						nNewWordLen	= pSrc - wstrNewWord ;
				}
				pSrc		= wstrNewWord ;
				while (pDest < pDestEnd && nNewWordLen > 0) {
					*pDest ++	= *pSrc ++ ;
					nNewWordLen	-- ;
				}
			} else {
				/* ̏ꍇׂ͖B*/
			}
			if (pDest < pDestEnd)
				*pDest ++	= L' ' ;
				nShow	++ ;
			if (!TSearchSession_bNextCandidate (pBuffer->m_pSkkCurrentSearchProgSession))
				break ;
		}
		nLeft	= TSearchSession_iGetNumberOfCandidate (pBuffer->m_pSkkCurrentSearchProgSession, &bSearchContinue) - nHenkanCount - nShow ;
		wnsprintf (buf, ARRAYSIZE (buf) - 1, L"[c %d%s]", nLeft, (bSearchContinue? L"+" : L"")) ;
		buf [ARRAYSIZE (buf) - 1]	= L'\0' ;
		pSrc	= buf ;
		while (pDest < pDestEnd && *pSrc != L'\0') {
			*pDest ++	= *pSrc ++ ;
		}
		ImeDoc_bSetMessageN (pDoc, buffer, pDest - buffer) ;
#else
		if (! imeBuffer_bInitializeHenkanCandidateList (pBuffer, nHenkanCount)) {
			/* fatal error */
			return	LMR_ERROR ;
		}
		/*
		 *	ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_UPDATE_CANDIDATES or IMEDOC_UPDATE_CANDIDATES) ;
		 *	͍쐬ƍXVȂ̂ŁB
		 */
		ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_UPDATE_CANDIDATELIST) ;
#endif
	} else {
		ImeDoc_vClearMessage (pDoc) ;	/* ̓G[ł́H */
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_RETVAL, -1) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0) ;
		return	LMR_CONTINUE ;
	}
	ImeDoc_vSetRegInteger (pDoc, LMREG_4, nShow) ;
	ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_2) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkanShowCandidates_2 (
	struct CImeDoc*			pDoc)
{
	ImeDoc_vSetRegBool (pDoc, LMREG_2, FALSE) ;
	ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_3) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkanShowCandidates_3 (
	struct CImeDoc*			pDoc)
{
	int			nLoop ;
	BOOL		bUpdateShow ;

	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_1, &nLoop))
		nLoop	= 0 ;
	if (! ImeDoc_bGetRegBool    (pDoc, LMREG_2, &bUpdateShow))
		bUpdateShow	= FALSE ;
#if !defined (UNITTEST)
#else
	if (bUpdateShow) {
		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_1) ;
		return	LMR_CONTINUE ;
	}
#endif
	if (! ImeDoc_bCall (pDoc, LM_bNextCommandEvent, LM_bSkkHenkanShowCandidates_4))
		return	LMR_ERROR ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkanShowCandidates_4 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	struct TMSG				ev ;
	int						n, nCH, nShow, nCandidateKey, nLoop, nMaxShow ;
	LPCWSTR					pCandidateKey ;
	BOOL					bKeyChar, bUpdateShow ;

	if (! ImeDoc_bGetRegMsg (pDoc, LMREGARG_RETVAL, &ev)) {
		/* FALSE exit */
		ImeDoc_vSetRegInteger (pDoc, LMREGARG_RETVAL, -1) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0) ;
		return	LMR_CONTINUE ;
	}
	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0) ;
		return	LMR_CONTINUE ;
	}

	/* CxgLN^[ɕϊB*/
	bKeyChar	= ImeDoc_bEventToCharacter (&ev, &nCH) ;
	if (! bKeyChar) {
		/* original map ̎śH */
		nCH	= -1 ;
//		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_3) ;
//		return	LMR_CONTINUE ;
	}

	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_4, &nShow))
		nShow	= 0 ;

	/*  skk-key-num-alist ƌׂāA₪I΂ꂽ`FbNB*/
	pCandidateKey	= ImeConfig_pGetSkkHenkanShowCandidatesKeys (&nCandidateKey) ;
	nMaxShow		= (nCandidateKey < IMEDOC_NUMHENKANSHOWCANDIDATES)? nCandidateKey : IMEDOC_NUMHENKANSHOWCANDIDATES ;	/* candidate-key 7ŒɂׂH */
	n				= 0 ;
	bUpdateShow		= FALSE ;

#if !defined (UNITTEST)
	/*	nShow ͍ČvZȂ̂ŁAMaxShow ɌŒ肵Ă܂B(!UNITETEST _)
	 */
	nShow			= nMaxShow ;
#endif
	while (n < nShow && n < nCandidateKey) {
		if (towlower (pCandidateKey [n]) == towlower (nCH)) {
			/* found */
			break ;
		}
		n	++ ;
	}
	if (! ImeDoc_bGetRegInteger (pDoc, LMREG_1, &nLoop))
		nLoop	= 0 ;
	if (n < nShow && n < nCandidateKey) {
		int		nBack ;
		LPCWSTR	wstrNewWord ;
#if !defined (UNITTEST)
		TVarbuffer*	pvbufCandInfo ;
#endif

		/* IB*/
#if !defined (UNITTEST)
		nBack	= n ;
		while (nBack > 0 && TSearchSession_bNextCandidate (pBuffer->m_pSkkCurrentSearchProgSession))
			nBack	-- ;
		if (nBack != 0) {
			/*	ĂʂȃL[B*/
			while (nBack < n && TSearchSession_bPreviousCandidate (pBuffer->m_pSkkCurrentSearchProgSession))
				nBack	++ ;
			goto	exit_func ;
		}
#else
		nBack	= nShow - n - 1 ;
		while (nBack > 0 && TSearchSession_bPreviousCandidate (pBuffer->m_pSkkCurrentSearchProgSession)) 
			nBack	-- ;
#endif

		wstrNewWord	= TSearchSession_pGetCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;
		if (wstrNewWord != NULL) {
			LPWSTR		pwRetbuff ;
			int			n, nNewWordLen, nRetbuffSize ;

			pwRetbuff	= ImeDoc_pGetReturnBuffer (pDoc, &nRetbuffSize) ;
			if (pwRetbuff == NULL)
				return	LMR_ERROR ;

			nNewWordLen	= lstrlenW (wstrNewWord) ;
			n			= MIN (nNewWordLen, nRetbuffSize) ;
			if (n > 0)
				memcpy (pwRetbuff, wstrNewWord, n * sizeof (WCHAR)) ;
			ImeDoc_vSetRegConstString (pDoc, LMREGARG_RETVAL, pwRetbuff, n) ;

			/*
			 */
			pBuffer->m_bSkkKakuteiFlag	= TRUE ;
		} else {
			ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
		}
#if !defined (UNITTEST)
		/* _Exit_ őSċzׂH */
		ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_CLOSE_CANDIDATELIST) ;

		pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
		if (pvbufCandInfo != NULL)
			TVarbuffer_Clear (pvbufCandInfo) ;
#endif
		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0) ;
		return	LMR_CONTINUE ;
	} else {
		int		iFuncNo ;

		if (! ImeDoc_bLookupKeymap (pDoc, &ev, &iFuncNo)) 
			iFuncNo	= NFUNC_NOFUNCTION ;

		/* ȊÕL[̏ꍇBK SPC ͌B*/
		if (nCH == L' ' || iFuncNo == NFUNC_SKK_START_HENKAN) {
#if !defined (UNITTEST)
			/*	NextCandidate ̐s̔ minibuffer-mode Ɉڍsׂǂ
			 *	`FbNłȂBnMaxShow 񂵂Ă݂KvB
			 */
			n	= 0 ;
			while (n <= nMaxShow && TSearchSession_bNextCandidate (pBuffer->m_pSkkCurrentSearchProgSession)) 
				n	++ ;
			if (n == (nMaxShow + 1)) {
				TVarbuffer*	pvbufCandInfo ;

				nLoop	++ ;
				(void) TSearchSession_bPreviousCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;

				pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
				if (pvbufCandInfo != NULL) {
					LPMYCAND	pMyCand ;

					pMyCand		= (LPMYCAND) TVarbuffer_GetBuffer (pvbufCandInfo) ;
					if (pMyCand != NULL) {
						LPCANDIDATELIST	pCandidateList ;

						pCandidateList	= (LPCANDIDATELIST)&(pMyCand->cl) ;
						pCandidateList->dwPageStart	+= nMaxShow ;
						pCandidateList->dwSelection	+= nMaxShow ;
					}
#if !defined (UNITTEST)
					ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_UPDATE_CANDIDATELIST) ;
#endif	
				}
			} else {
#if !defined (UNITTEST)
				TVarbuffer*	pvbufCandInfo ;
#endif
				/*	ʖځBċAo^[hɈڍs邱ƁB
				/*	ċA[hł́ACandidate-List ͕BA߂ė̏󋵎ł
				 *	ēx邱ƂɂȂ邪cB
				 */
				ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_CLOSE_CANDIDATELIST) ;
#if !defined (UNITTEST)
				pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
				if (pvbufCandInfo != NULL)
					TVarbuffer_Clear (pvbufCandInfo) ;
#endif
				ImeDoc_vSetRegInteger (pDoc, LMREG_3, n) ;
				if (! ImeDoc_bCall (pDoc, LM_bSkkHenkanInMinibuff, LM_bSkkHenkanShowCandidates_5))
					return	LMR_ERROR ;
				return	LMR_CONTINUE ;
			}
#else
			if (! TSearchSession_bNextCandidate (pBuffer->m_pSkkCurrentSearchProgSession)) {
#if 1
				ImeDoc_vSetRegInteger (pDoc, LMREG_3, nShow) ;
				/*	ċAo^[hɈړB*/
				/*	ċAo^[h߂ėɂ܂ skk-henkan-show-candidates-mode ł
				 *	ꍇɂ́ACANDIDATELIST č\zȂƂȂ > Windows IME
				 *	̂܂܏p肵ȂƁB
				 */
				if (! ImeDoc_bCall (pDoc, LM_bSkkHenkanInMinibuff, LM_bSkkHenkanShowCandidates_5))
					return	LMR_ERROR ;

				/*	IWi̓́A
				 *	1. skk-exit-show-candidates  skk-henkan-count ݒ肵A
				 *	   skk-henkan1 ֓͂ nil łƂĖ߂B
				 *	2. skk-henkan1 ͕Ԃl nil  skk-henkan ւƖ߂Askk-henkan ͕Ԃl nil 
				 *	   ꍇɂ́Askk-henkan-in-minibuff ĂяoB
				 *	3. skk-henkan-in-minibuff  cancel  (keyboard-quit Ȃreturn)A
				 *	   skk-exit-show-candidates QƂāAݒ肳Ă΂̒llƂ
				 *	   skk-henkan-show-candidates ĂяoB
				 *
				 *	 skk-henkan oR skk-henkan-in-minibuff  skk-henkan-show-candidates 
				 *	ēxĂяo\ǂȂ͕̂ȂAstack ]vɕKvɂȂ肻
				 *	ŁA skk-henkan-in-minibuff ĂяoĂB
				 */
#else
				/*
				       (let ((last-showed-index (+ 4 (* loop max-candidates))))
						 (setq skk-exit-show-candidates
						       ;; cdr ́Ao^ɓOɍŌɕ\
						       ;; Q̒ōŏ̌wCfNX
						       (cons loop last-showed-index))
						 ;; o^ɓBskk-henkan-count 
						 ;; skk-henkan-list ̍Ō̌̎ (݂Ȃ
						 ;; --- nil)wB
						 (setq skk-henkan-count (+ last-showed-index n)
						       loop nil))))
							   */
				int	nLastShowedIndex ;

				nLastShowedIndex	= 4 + nLoop * nMaxShow ;
				pBuffer->m_iSkkHenkanCount	= nLastShowedIndex + n ;
				ImeDoc_vSetRegInteger (pDoc, LMREGARG_0, 0) ;
				ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0) ;
#endif
				return	LMR_CONTINUE ;
			} else {
				nLoop	++ ;
				(void)TSearchSession_bPreviousCandidate (pBuffer->m_pSkkCurrentSearchProgSession) ;
				/* ̌̕\ցcB*/
			}
#endif
			bUpdateShow	= TRUE ;
		} else if (nCH == L'.') {
#if !defined (UNITTEST)
			TVarbuffer*	pvbufCandInfo ;
#endif
			/* o^ɈړL[̏ꍇBdefault  '.' */
#if !defined (UNITTEST)
			ImeDoc_vSetRegInteger (pDoc, LMREG_3, 0) ;	/* ړĂȂA0ŗǂH */
#else
			ImeDoc_vSetRegInteger (pDoc, LMREG_3, nShow) ;
#endif
			/*	ċAo^[hɈړB*/
			/*	ċAo^[h߂ėɂ܂ skk-henkan-show-candidates-mode ł
			 *	ꍇɂ́ACANDIDATELIST č\zȂƂȂ > Windows IME
			 *	̂܂܏p肵ȂƁB
			 */
			if (! ImeDoc_bCall (pDoc, LM_bSkkHenkanInMinibuff, LM_bSkkHenkanShowCandidates_5))
				return	LMR_ERROR ;
#if !defined (UNITTEST)
			/*	ċA[hł́ACandidate-List ͕BA߂ė̏󋵎ł
			 *	ēx邱ƂɂȂ邪cB
			 */
			ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_CLOSE_CANDIDATELIST) ;

			pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
			if (pvbufCandInfo != NULL)
				TVarbuffer_Clear (pvbufCandInfo) ;
#endif
			return	LMR_CONTINUE ;
		} else if (nCH == L'x' || iFuncNo == NFUNC_SKK_PREVIOUS_CANDIDATE) {
#if !defined (UNITTEST)
			TVarbuffer*	pvbufCandInfo ;
#endif
			/* 1Oɖ߂L[̏ꍇB(skk-previous-candidate-char) default  'x' */
			if (nLoop <= 0) {
				struct TMSG	msg ;

				pBuffer->m_iSkkHenkanCount	= ImeConfig_iGetCountHenkanShowChange () + 1 ;

				/* skk-previous-candidate  bind ꂽ unread-event ɐݒ肵Ă悤cB*/
#if 1
				msg.m_nMessage	= WM_LM_COMMAND ;
				msg.m_nTime		= GetTickCount () ;
				msg.m_wParam	= NFUNC_SKK_PREVIOUS_CANDIDATE ;
				msg.m_lParam	= 0 ;
				msg.m_rParam	= 0 ;
				msg.m_pt.x		= msg.m_pt.y	= 0 ;
				ImeDoc_bSetUnreadCommandEvent (pBuffer->m_pDoc, &msg) ;
#else
				/*	LookupKeymap  WM_CHAR 𗝉ȂB͂܂BA'x'  previous-candidate
				 *	 bind ĂȂ\B܂B
				 */
				ImeDoc_bSetUnreadCommandChar (pBuffer->m_pDoc, L'x') ;
#endif
				ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
				ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0);
#if !defined (UNITTEST)
				ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_CLOSE_CANDIDATELIST) ;

				pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
				if (pvbufCandInfo != NULL)
					TVarbuffer_Clear (pvbufCandInfo) ;
#endif
				return	LMR_CONTINUE ;
			} else {
				int		nBack ;
#if !defined (UNITTEST)
				TVarbuffer*	pvbufCandInfo ;

				/*	\鎞 NextCandidate 𗘗p SearchSession XVĂȂ̂ŁA
				 *	߂ʂ nMaxShow łB
				 *
				 *		-------------------------- O
				 *				nMaxShow ̌
				 *		-------------------------- 
				 *				nShow \
				 *		-------------------------- ݎwĂʒuB
				 *
				 *	ꂪA
				 *		-------------------------- O
				 *				nMaxShow ̌
				 *		-------------------------- 񂩂ݎwĂʒuɂȂ (!UNITTEST)
				 */
				nBack	= nMaxShow ;
				while (nBack > 0 && TSearchSession_bPreviousCandidate (pBuffer->m_pSkkCurrentSearchProgSession)) 
					nBack	-- ;

				/*	Ȁ̒gŃG[Aǂ񂾂낤H 
				 *	Lock sƂB
				 */
				pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
				if (pvbufCandInfo != NULL) {
					LPMYCAND	pMyCand ;

					pMyCand		= (LPMYCAND) TVarbuffer_GetBuffer (pvbufCandInfo) ;
					if (pMyCand != NULL) {
						LPCANDIDATELIST	pCandidateList	= (LPCANDIDATELIST)&(pMyCand->cl) ;
						pCandidateList->dwPageStart	-= nMaxShow ;
						pCandidateList->dwSelection	-= nMaxShow ;
					}
				}
				ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_UPDATE_CANDIDATELIST) ;
#else
				nBack	= nMaxShow + nShow ;
				while (nBack > 0 && TSearchSession_bPreviousCandidate (pBuffer->m_pSkkCurrentSearchProgSession)) 
					nBack	-- ;
#endif
				nLoop	-- ;
				bUpdateShow	= TRUE ;
			}
		} else if (nCH == 0x07 || (iFuncNo == NFUNC_KEYBOARD_QUIT || iFuncNo == NFUNC_ABORT_RECURSIVE_EDIT)) {
#if !defined (UNITTEST)
			TVarbuffer*	pvbufCandInfo ;
#endif
			struct TMSG	msg ;

			/* quit  bind ĂꍇB*/
			/* signal g keyboard-quit ĂяoĂ悤cB*/
#if 1
			msg.m_nMessage	= WM_LM_COMMAND ;
			msg.m_wParam	= NFUNC_KEYBOARD_QUIT ;
			msg.m_lParam	= 0 ;
			msg.m_pt.x		= msg.m_pt.y	= 0 ;
			msg.m_nTime		= GetTickCount () ;
			ImeDoc_bSetUnreadCommandEvent (pDoc, &msg) ;
#else
			ImeDoc_bSetUnreadCommandChar (pDoc, 0x07) ;
#endif
			ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
			ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0);
#if !defined (UNITTEST)
			ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_CLOSE_CANDIDATELIST) ;

			pvbufCandInfo	= ImeDoc_pGetCandidateInfoBuffer (pDoc) ;
			if (pvbufCandInfo != NULL)
				TVarbuffer_Clear (pvbufCandInfo) ;
#endif
			pBuffer->m_bSkkKakuteiFlag	= FALSE ;
			return	LMR_CONTINUE ;
		} else {
			/*	ȊO͖ȃL[łƕ\̂Asit-for ݂łȂ(applicationɐ䂪߂)̂
			 *	\邱Ƃ̓[U[ɂƂĎזłȂB
			 */
			WCHAR	bufText [256] ;
			int		nText ;
			if (0x21 <= nCH && nCH < 0x7E) {
				nText	= wnsprintfW (bufText, ARRAYSIZE (bufText), L"`%c' is not valid here!", nCH) ;
			} else {
				nText	= wnsprintfW (bufText, ARRAYSIZE (bufText), L"0x%0x is not valid here!", nCH) ;
			}
			ImeDoc_bSetMessageN (pDoc, bufText, nText) ;
		}
	}
exit_func:
	ImeDoc_vSetRegInteger (pDoc, LMREG_1, nLoop) ;
	ImeDoc_vSetRegBool    (pDoc, LMREG_2, bUpdateShow) ;
	ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_3) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkanShowCandidates_5 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	LPCWSTR		pwResult ;
	int			nShow, nResultLen ;

	if (ImeDoc_bSignalp (pDoc)) {
		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0) ;
		return	LMR_CONTINUE ;
	}
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0) ;
		return	LMR_CONTINUE ;
	}
	if (ImeDoc_bGetRegConstString (pDoc, LMREGARG_RETVAL, &pwResult, &nResultLen) && nResultLen > 0) {
		/* return n */
		/* ImeDoc_vSetRegInteger (pDoc, LMREGARG_RETVAL, n) ; */
		ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit_0) ;
		return	LMR_CONTINUE ;
	}
	if (ImeDoc_bGetRegInteger (pDoc, LMREG_4, &nShow)) {
		while (nShow > 1 && TSearchSession_bPreviousCandidate (pBuffer->m_pSkkCurrentSearchProgSession)) 
			nShow	-- ;
	}
	ImeDoc_vSetRegBool (pDoc, LMREG_2, TRUE) ;
	ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_1) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkHenkanShowCandidates_Exit_0 (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) {
		pBuffer->m_bSkkHenkanShowCandidatesMode	= FALSE ;
		/* skk-point-move ̌ʁB*/
		imeBuffer_bSkkSetMarker (pBuffer, &pBuffer->m_pmkSkkPreviousPoint, MARKER_SKK_PREVIOUS_POINT, pBuffer->m_pmkPoint) ;
	}
	ImeDoc_vClearMessage (pDoc) ;
	ImeDoc_vJump (pDoc, LM_bSkkHenkanShowCandidates_Exit) ;
	return	LMR_CONTINUE ;
}

/*================================================================ ad-keyboard-quit (around)
 */
static	int		LM_bSkkKeyboardQuit_PostprocessPreviousCandidateCharForDeleteOkuriHenkanQuit (struct CImeDoc*) ;
static	int		LM_bSkkKeyboardQuit_1 (struct CImeDoc*) ;

/*	around (Ƃ̊֐̎s܂) emulationBad-do-it ͂Ƃ̊֐̎sɓ邱Ƃ
 *	B
 *	Ԃl: LMREGARG_RETVAL: ad-do-it ǂB{ ad-do-it ĂԈʒu͔CӂȂ̂ŕԂl
 *			͓̂K؂ł͂ȂcB
 */
int
LM_bSkkAdKeyboardQuit (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	struct CSkkRuleTreeIterator*	piteSkkRuleTree	= NULL ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, FALSE) ;
		return	LMR_RETURN ;
	}
	if (! pBuffer->m_bSkkMode) {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, FALSE) ;
		return	LMR_RETURN ;
	}
	if (! pBuffer->m_bSkkHenkanMode) {
		int		nPrefix ;

		piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;
		if (SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree) && SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) != NULL) {
//		if (pBuffer->m_pSkkCurrentRuleTree != NULL && TSkkRuleTreeNode_pGetPrefix (pBuffer->m_pSkkCurrentRuleTree, &nPrefix) != NULL) {
			imeBuffer_bSkkErasePrefix (pBuffer, TRUE) ;
		} else {
			ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, FALSE) ;
			return	LMR_RETURN ;
		}
	} else if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
		pBuffer->m_iSkkHenkanCount	= 0 ;
		if (ImeConfig_bSkkDeleteOkuriWhenQuit () && pBuffer->m_nSkkHenkanOkuriganaLen > 0) {
			/*	艼̒LĂB
			 */
			if (! ImeDoc_bPushReg (pDoc, LMREG_0))
				return	LMR_ERROR ;
			ImeDoc_vSetRegInteger (pDoc, LMREG_0, pBuffer->m_nSkkHenkanOkuriganaLen) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkPreviousCandidate, LM_bSkkKeyboardQuit_PostprocessPreviousCandidateCharForDeleteOkuriHenkanQuit))
				return	LMR_ERROR ;
		} else {
			if (! ImeDoc_bCall (pDoc, LM_bSkkPreviousCandidate, LM_bSkkKeyboardQuit_1))
				return	LMR_ERROR ;
		}
		return	LMR_CONTINUE ;
	} else {
		if (ImeDoc_iGetLastCommand (pDoc) == NFUNC_SKK_COMP_DO) {
			if (pBuffer->m_pmkSkkHenkanStartPoint != NULL)
				ImeBuffer_bDeleteRegion (pBuffer, TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint), TMarker_iGetPosition (pBuffer->m_pmkPoint)) ;
			/* skk-comp-key */
			if (pBuffer->m_pSkkCompSession != NULL) {
				LPCWSTR	pSkkCompKey ;
				int		nSkkCompKey ;

				pSkkCompKey	= TSearchSession_pGetKeyword (pBuffer->m_pSkkCompSession, &nSkkCompKey) ;
				ImeBuffer_iInsert (pBuffer, pBuffer->m_pmkPoint, pSkkCompKey, nSkkCompKey) ;

				TSearchSession_vDestroy (pBuffer->m_pSkkCompSession) ;
				pBuffer->m_pSkkCompSession	= NULL ;
			}
		} else {
			imeBuffer_bSkkErasePrefix (pBuffer, TRUE) ;
			if (pBuffer->m_pmkSkkHenkanStartPoint != NULL && TMarker_iGetPosition (pBuffer->m_pmkPoint) > TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint)) {
				ImeBuffer_bDeleteRegion (pBuffer, TMarker_iGetPosition (pBuffer->m_pmkPoint), TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint)) ;
			}
			ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkKeyboardQuit_1))
				return	LMR_ERROR ;
			return	LMR_CONTINUE ;
		}
	}
	ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, TRUE) ;
	return	LMR_RETURN ;
}

int
LM_bSkkKeyboardQuit_PostprocessPreviousCandidateCharForDeleteOkuriHenkanQuit (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*	pBuffer ;
	int					nCount ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetSignalError (pDoc) ;
	} else {
		if (! ImeDoc_bGetRegInteger (pDoc, LMREG_0, &nCount))
			nCount	= 0 ;
		if (! ImeBuffer_bDeleteBackwardChar (pBuffer, nCount)) {
			ImeDoc_vSetSignalError (pDoc) ;
		}
	}
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	ImeDoc_vJump (pDoc, LM_bSkkKeyboardQuit_1) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkKeyboardQuit_1 (
	struct CImeDoc*			pDoc)
{
	/* ... */
	ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, TRUE) ;
	return	LMR_RETURN ;
}

/*================================================================ ad-abort-recursive-edit (around)
 */
#define	LM_bSkkAbortRecursiveEdit_PostprocessPreviousCandidateCharForDeleteOkuriHenkanQuit	LM_bSkkKeyboardQuit_PostprocessPreviousCandidateCharForDeleteOkuriHenkanQuit
#define	LM_bSkkAbortRecursiveEdit_1	LM_bSkkKeyboardQuit_1

/*	around (Ƃ̊֐̎s܂) emulation
 */
int
LM_bSkkAdAbortRecursiveEdit (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	struct CSkkRuleTreeIterator*	piteSkkRuleTree	= NULL ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, FALSE) ;
		return	LMR_RETURN ;
	}
	/*	ȉ hook 菜삪KvHF
	 *  (skk-remove-minibuffer-setup-hook
	 *   'skk-j-mode-on 'skk-setup-minibuffer
	 *   #'(lambda () (add-hook 'pre-command-hook 'skk-pre-command nil 'local)))
	 */
	if (! pBuffer->m_bSkkMode) {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, FALSE) ;
		return	LMR_RETURN ;
	}
	if (! pBuffer->m_bSkkHenkanMode) {
		int		nPrefix ;

		piteSkkRuleTree	= ImeBuffer_pSkkGetSkkRuleTreeIterator (pBuffer) ;
		if (SkkRuleTreeIterator_bHavePrefixp (piteSkkRuleTree) && SkkRuleTreeIterator_pGetPrefix (piteSkkRuleTree, &nPrefix) != NULL) {
//		if (pBuffer->m_pSkkCurrentRuleTree != NULL && TSkkRuleTreeNode_pGetPrefix (pBuffer->m_pSkkCurrentRuleTree, &nPrefix) != NULL) {
			imeBuffer_bSkkErasePrefix (pBuffer, TRUE) ;
		} else {
			ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, FALSE) ;
			return	LMR_RETURN ;
		}
	} else if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
		pBuffer->m_iSkkHenkanCount	= 0 ;
		if (ImeConfig_bSkkDeleteOkuriWhenQuit () && pBuffer->m_nSkkHenkanOkuriganaLen > 0) {
			/*	艼̒LĂB
			 */
			if (! ImeDoc_bPushReg (pDoc, LMREG_0))
				return	LMR_ERROR ;
			ImeDoc_vSetRegInteger (pDoc, LMREG_0, pBuffer->m_nSkkHenkanOkuriganaLen) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkPreviousCandidate, LM_bSkkAbortRecursiveEdit_PostprocessPreviousCandidateCharForDeleteOkuriHenkanQuit))
				return	LMR_ERROR ;
		} else {
			if (! ImeDoc_bCall (pDoc, LM_bSkkPreviousCandidate, LM_bSkkAbortRecursiveEdit_1))
				return	LMR_ERROR ;
		}
		return	LMR_CONTINUE ;
	} else {
		if (ImeDoc_iGetLastCommand (pDoc) == NFUNC_SKK_COMP_DO) {
			if (pBuffer->m_pmkSkkHenkanStartPoint != NULL)
				ImeBuffer_bDeleteRegion (pBuffer, TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint), TMarker_iGetPosition (pBuffer->m_pmkPoint)) ;
			/* skk-comp-key */
			if (pBuffer->m_pSkkCompSession != NULL) {
				LPCWSTR	pSkkCompKey ;
				int		nSkkCompKey ;

				pSkkCompKey	= TSearchSession_pGetKeyword (pBuffer->m_pSkkCompSession, &nSkkCompKey) ;
				ImeBuffer_iInsert (pBuffer, pBuffer->m_pmkPoint, pSkkCompKey, nSkkCompKey) ;

				TSearchSession_vDestroy (pBuffer->m_pSkkCompSession) ;
				pBuffer->m_pSkkCompSession	= NULL ;
			}
		} else {
			imeBuffer_bSkkErasePrefix (pBuffer, TRUE) ;
			if (pBuffer->m_pmkSkkHenkanStartPoint != NULL && TMarker_iGetPosition (pBuffer->m_pmkPoint) > TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint)) {
				ImeBuffer_bDeleteRegion (pBuffer, TMarker_iGetPosition (pBuffer->m_pmkPoint), TMarker_iGetPosition (pBuffer->m_pmkSkkHenkanStartPoint)) ;
			}
			ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkAbortRecursiveEdit_1))
				return	LMR_ERROR ;
			return	LMR_CONTINUE ;
		}
	}
	ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, TRUE) ;
	return	LMR_RETURN ;
}

#undef	LM_bSkkAbortRecursiveEdit_PostprocessPreviousCandidateCharForDeleteOkuriHenkanQuit
#undef	LM_bSkkAbortRecursiveEdit_1

/*================================================================ ad-exit-minibuffer (around)
 */
static	int		LM_bSkkExitMinibuffer_1		(struct CImeDoc*) ;

/*	around (Ƃ̊֐̎s܂) emulation
 */
int
LM_bSkkAdExitMinibuffer (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, FALSE) ;
		return	LMR_RETURN ;
	}
	/*  (skk-remove-minibuffer-setup-hook
	 *   'skk-j-mode-on 'skk-setup-minibuffer
	 *   #'(lambda ()
	 *       (add-hook 'pre-command-hook 'skk-pre-command nil 'local)))
	 *
	 *	̈ǂׂcB
	 */
	if (! (pBuffer->m_bSkkJMode || pBuffer->m_bSkkJisx0201Mode || pBuffer->m_bSkkAbbrevMode)) {
		/* through */
	} else {
		BOOL	bNoNewline ;

		bNoNewline	= (ImeConfig_bSkkEggLikeNewline () && pBuffer->m_bSkkHenkanMode) ;
		if (! ImeDoc_bPushReg (pDoc, LMREG_0))
			return	LMR_ERROR ;

		ImeDoc_vSetRegBool (pDoc, LMREG_0, bNoNewline) ;
		if (pBuffer->m_bSkkMode) {
			ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
			if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkExitMinibuffer_1))
				return	LMR_ERROR ;
		} else {
			ImeDoc_vJump (pDoc, LM_bSkkExitMinibuffer_1) ;
		}
		return	LMR_CONTINUE ;
	}
	ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, FALSE) ;	/* ad-do-it */
	return	LMR_RETURN ;
}

int
LM_bSkkExitMinibuffer_1 (
	struct CImeDoc*			pDoc)
{
	BOOL					bNoNewline ;

	if (! ImeDoc_bGetRegBool (pDoc, LMREG_0, &bNoNewline))
		bNoNewline	= TRUE ;
	ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, bNoNewline) ;
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	return	LMR_RETURN ;
}

/*================================================================ ad-next-line (before)
 */
/*
 *	(defadvice next-line (before skk-ad activate)
 *	  (when (eq skk-henkan-mode 'active)
 *	    (skk-kakutei)))
 */
int
LM_bSkkAdNextLine (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;

	/*	before advice Ȃ̂ ad-do-it ̎w͂ȂBƂAK{͎̂sB
	 */
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) {
		if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
			ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
			ImeDoc_vJump (pDoc, LM_bSkkKakuteiAdJisx0201) ;
			return	LMR_CONTINUE ;
		}
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ ad-previous-line (before)
 */
/*
 *	(defadvice previous-line (before skk-ad activate)
 *	  (when (eq skk-henkan-mode 'active)
 *	    (skk-kakutei)))
 */
int
LM_bSkkAdPreviousLine (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;

	/*	before advice Ȃ̂ ad-do-it ̎w͂ȂBƂAK{͎̂sB
	 */
	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer != NULL) {
		if (pBuffer->m_bSkkHenkanMode == LACTIVE) {
			ImeDoc_vSetRegConstString (pDoc, LMREGARG_0, NULL, 0) ;
			ImeDoc_vJump (pDoc, LM_bSkkKakuteiAdJisx0201) ;
			return	LMR_CONTINUE ;
		}
	}
	ImeDoc_vSetRegNil (pDoc, LMREGARG_RETVAL) ;
	return	LMR_RETURN ;
}

/*================================================================ ad-newline (around)
 */
static	int		LM_bSkkAdNewline_1		(struct CImeDoc*) ;
static	int		LM_bSkkAdNewline_Exit	(struct CImeDoc*) ;
	
	/*
(skk-defadvice newline (around skk-ad activate)
  "`skk-egg-like-newline' Aϊ͊m̂ݍsAsȂB"
  (interactive "*P")
  (if (not (or skk-j-mode
	       skk-jisx0201-mode
	       skk-abbrev-mode))
      ad-do-it
    (let (;;(arg (ad-get-arg 0))
	  ;; `skk-kakutei' s `skk-henkan-mode' ̒l
	  ;;  nil ɂȂ̂ŁAۑĂKvB
	  (no-newline (and skk-egg-like-newline
			   skk-henkan-mode))
	  (auto-fill-function (if (interactive-p)
				  auto-fill-function
				nil)))
      ;; fill Ă nil AĂ :-<
      ;;(if (skk-kakutei)
      ;;    (setq arg (1- arg)))
      ;;(if skk-mode
      ;;    (let ((opos (point)))
      ;;      ;; skk-kakutei (skk-do-auto-fill) ɂčs܂Ԃꂽ
      ;;      ;; arg  1 炷B
      ;;      (skk-kakutei)
      ;;      (if (and (not (= opos (point))) (integerp arg))
      ;;          (ad-set-arg 0 (1- arg)))))
      (when skk-mode
	(skk-kakutei))
      (undo-boundary)
      (unless no-newline
	ad-do-it))))*/
int
LM_bSkkAdNewline (
	struct CImeDoc*			pDoc)
{
	struct CImeBuffer*		pBuffer ;
	BOOL		bNoNewline ;

	pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
	if (pBuffer == NULL) {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, FALSE) ;
		ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_NEWLINE) ;
		return	LMR_RETURN ;
	}
	if (! (pBuffer->m_bSkkJMode || pBuffer->m_bSkkJisx0201Mode || pBuffer->m_bSkkAbbrevMode)) {
		ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, FALSE) ;
		ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_NEWLINE) ;
		return	LMR_RETURN ;
	}
	if (! ImeDoc_bPushReg (pDoc, LMREG_0))
		return	LMR_ERROR ;

	/*	newline őSm@\̂ȂlȂ΁B
	 */
	bNoNewline	= (ImeConfig_bSkkEggLikeNewline () && pBuffer->m_bSkkHenkanMode)? TRUE : FALSE ;
	ImeDoc_vSetRegBool (pDoc, LMREG_0, bNoNewline) ;

	if (pBuffer->m_bSkkMode) {
		ImeDoc_vSetRegNil (pDoc, LMREGARG_0) ;
		if (! ImeDoc_bCall (pDoc, LM_bSkkKakuteiAdJisx0201, LM_bSkkAdNewline_1))
			return	LMR_ERROR ;
		return	LMR_CONTINUE ;
	}
	ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, FALSE) ;
	ImeDoc_vJump (pDoc, LM_bSkkAdNewline_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkAdNewline_1 (
	struct CImeDoc*			pDoc)
{
	BOOL	bNoNewline ;

	if (! ImeDoc_bGetRegBool (pDoc, LMREG_0, &bNoNewline))
		bNoNewline	= TRUE ;

	/*	newline őSmɂꍇɂ QueryShiftCount ̐ skk-kakutei ̒ł
	 *	Ƃӂ݂āApoint-marker ŏIʒuֈړΗǂB
	 */
	if (ImeConfig_bSkkIsNewlineKakuteiAllp () && ! ImeDoc_bIsStatusActivep (pDoc)) {
		struct CImeBuffer*	pBuffer ;

		pBuffer	= ImeDoc_pGetCurrentBuffer (pDoc) ;
		if (pBuffer != NULL && TMarker_bIsValidp (pBuffer->m_pmkPoint)) {
			struct TMarker*	pmkBufferEnd	= NULL ;

			/*	point marker sւƈړB*/
			if (ImeBuffer_bGetMarker (pBuffer, MARKER_BUFFEREND, &pmkBufferEnd) && pmkBufferEnd != NULL) {
				TMarker_bSetPosition (pBuffer->m_pmkPoint, pmkBufferEnd) ;
			} else {
				int		nPoint	= TMarker_iGetPosition (pBuffer->m_pmkPoint) ;
				if (0 <= nPoint && nPoint < pBuffer->m_nbufComp) {
					TMarker_bForward (pBuffer->m_pmkPoint, pBuffer->m_nbufComp - nPoint) ;
				}
			}
		}
	}
	ImeDoc_vSetRegBool (pDoc, LMREGARG_RETVAL, bNoNewline) ;
	ImeDoc_vJump (pDoc, LM_bSkkAdNewline_Exit) ;
	return	LMR_CONTINUE ;
}

int
LM_bSkkAdNewline_Exit (
	struct CImeDoc*			pDoc)
{
	BOOL	bNoNewline ;

	if (! ImeDoc_bGetRegBool (pDoc, LMREG_0, &bNoNewline))
		bNoNewline	= TRUE ;
	if (! bNoNewline) {
		/*	newline  unprocess event ɕϊB*/
		ImeDoc_bUnprocessEvent (pDoc, ImeDoc_pGetLastCommandEvent (pDoc)) ;
	}
	ImeDoc_vPopReg (pDoc, LMREG_0) ;
	ImeDoc_vSetUpdateFlag (pDoc, IMEDOC_NEWLINE) ;
	return	LMR_RETURN ;
}




